/*
DisChoco: Distributed Constraint Reasoning Solver 
Copyright (C) 2007-2013 Redouane Ezzahir & Mohamed Wahbi 

DisChoco is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DisChoco is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

To contact the authors: <http://dischoco.sourceforge.net/>
*/ 
package benchmark;

import static simulation.parser.InstanceTokens.BENCHMARK;
import static simulation.parser.InstanceTokens.CONSTRAINT_MODEL;
import static simulation.parser.InstanceTokens.FORMAT;
import static simulation.parser.InstanceTokens.INSTANCE;
import static simulation.parser.InstanceTokens.MODEL;
import static simulation.parser.InstanceTokens.NAME;
import static simulation.parser.InstanceTokens.PRESENTATION;
import static simulation.parser.InstanceTokens.TYPE;
import static simulation.parser.InstanceTokens.XDisCSP_1_0;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import org.jdom2.Document;
import org.jdom2.Element;
import org.jdom2.output.Format;
import org.jdom2.output.XMLOutputter;
import util.Commands;

/**
 * The class Generator 
 * @author Mohamed Wahbi
 * @since 2012/10/17
 */
public abstract class Generator implements Runnable {

	public static final String COMMA = ", ";
	/** the type of the benchmark (e.g., Random DisCSP, Random DCOP, etc) */
	protected String typeOfBenchmark;
	
	/** The model of the problem. <tt>Simple</tt> when agents control exactly one 
	 * variable per agent or <tt>Complex</tt> otherwise */
	protected String model;
	
	/** The model of constraints. Constraints may be Partially of Totally 
	 * Known Constraints (TKC or PKC). */
	protected String constraintsModel;
	
	protected File experimentation = null;
	
	/**
	 * Create the generator instance 
	 * @param benchmark : type of the benchmark
	 * @param model : agents model, i.e., <tt>Simple</tt> or <tt>Complex</tt>.
	 * @param constraintsModel : the model of constraints, i.e., <tt>TKC</tt> or <tt>PKC</tt>.
	 */
	public Generator(String benchmark, String model, String constraintsModel){
		this.typeOfBenchmark = benchmark;
		this.model = model;
		this.constraintsModel = constraintsModel;
	}
	
	/**
	 * Create the thread for running experiments 
	 * @param experimentation : the configuration file of experimentation
	 */
	public Generator(File experimentation){
		this.experimentation = experimentation;
	}

	/**
	 * This method saves the given document <tt>document</tt> into a file with the given name <tt>fileName</tt>
	 * 
	 * @param fileName : the file name
	 * @param document : the document to save
	 */
	protected void save(String fileName, Document document) {
		try {
			FileWriter fw = new FileWriter(fileName);
			new XMLOutputter(Format.getPrettyFormat()).output(document, fw);
			Process p = Runtime.getRuntime().exec(Commands.BZIP2+" " + fileName);
			p.waitFor();
			p.exitValue();
			p.destroy();
		} catch (Exception e) {
			e.printStackTrace();
			throw new IllegalAccessError("\n"+Commands.BZIP2+" is not installed in your machine. " +
					"Please install it!");
		}
	}
	
	/**
	 * This method saves the document into the a file (with the given name) in the given directory. 
	 * @param directory : the directory 
	 * @param fileName : the file name
	 * @param document : the document to save
	 */
	protected void save(String directory, String fileName, Document document) {
		try {
			XMLOutputter out = new XMLOutputter(Format.getPrettyFormat());
			out.output(document, new FileOutputStream(directory + File.separator + fileName));
			Process p = Runtime.getRuntime().exec(Commands.BZIP2+" " + directory + File.separator + fileName);
			p.waitFor();
			p.exitValue();
			p.destroy();
		} catch (Exception e) {
			e.printStackTrace();
			throw new IllegalAccessError("\n"+Commands.BZIP2+" is not installed in your machine. " +
					"Please install it!");
		}
	}
	
	/**
	 * This method saves the document into the a file (with the given name) in the given directory. 
	 * @param directory : the directory 
	 * @param fileName : the file name
	 * @param document : the document to save
	 */
	protected void saveXML(String directory, String fileName, Document document) {
		try {
			XMLOutputter out = new XMLOutputter(Format.getPrettyFormat());
			out.output(document, new FileOutputStream(directory + File.separator + fileName, false));
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * This method return the presentation of the generated instance 
	 * 
	 * @param nbOfInstance : the number of the instance 
	 * @return org.jdom2..Element the presentation element
	 */
	protected Element getPresentation(int nbOfInstance, String type) {
		Element presentation = new Element(PRESENTATION);
		presentation.setAttribute(NAME, INSTANCE + nbOfInstance);
		presentation.setAttribute(TYPE, type);
		presentation.setAttribute(BENCHMARK, this.typeOfBenchmark);		// ...
		presentation.setAttribute(MODEL, this.model); 			// Simple or Complex
		presentation.setAttribute(CONSTRAINT_MODEL, this.constraintsModel); // TKC or PKC
		presentation.setAttribute(FORMAT, XDisCSP_1_0);
		return presentation;
	}
	
	public abstract void generate();
	
}
