/*
DisChoco: Distributed Constraint Reasoning Solver 
Copyright (C) 2007-2013 Redouane Ezzahir & Mohamed Wahbi 

DisChoco is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DisChoco is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

To contact the authors: <http://dischoco.sourceforge.net/>
*/ 
package benchmark.graphcoloring;

import static simulation.parser.InstanceTokens.AGENT;
import static simulation.parser.InstanceTokens.AGENTS;
import static simulation.parser.InstanceTokens.ARITY;
import static simulation.parser.InstanceTokens.BENCHMARK;
import static simulation.parser.InstanceTokens.CONFLICTS;
import static simulation.parser.InstanceTokens.CONSTRAINT;
import static simulation.parser.InstanceTokens.CONSTRAINTS;
import static simulation.parser.InstanceTokens.DCOP;
import static simulation.parser.InstanceTokens.DESCRIPTION;
import static simulation.parser.InstanceTokens.DOMAIN;
import static simulation.parser.InstanceTokens.DOMAINS;
import static simulation.parser.InstanceTokens.DisCSP;
import static simulation.parser.InstanceTokens.ID;
import static simulation.parser.InstanceTokens.INSTANCE;
import static simulation.parser.InstanceTokens.INSTANCES;
import static simulation.parser.InstanceTokens.MODEL;
import static simulation.parser.InstanceTokens.NAME;
import static simulation.parser.InstanceTokens.NB_AGENTS;
import static simulation.parser.InstanceTokens.NB_CONSTRAINTS;
import static simulation.parser.InstanceTokens.NB_DOMAINS;
import static simulation.parser.InstanceTokens.NB_INSTANCES;
import static simulation.parser.InstanceTokens.NB_RELATIONS;
import static simulation.parser.InstanceTokens.NB_TUPLES;
import static simulation.parser.InstanceTokens.NB_VALUES;
import static simulation.parser.InstanceTokens.NB_VARIABLES;
import static simulation.parser.InstanceTokens.REFERENCE;
import static simulation.parser.InstanceTokens.RELATION;
import static simulation.parser.InstanceTokens.RELATIONS;
import static simulation.parser.InstanceTokens.SCOPE;
import static simulation.parser.InstanceTokens.SEMANTICS;
import static simulation.parser.InstanceTokens.TYPE;
import static simulation.parser.InstanceTokens.VARIABLE;
import static simulation.parser.InstanceTokens.VARIABLES;
import static util.Commands.DomainSize;
import static util.Commands.GuiPresentation;
import static util.Commands.TIGHTNESS;
import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Random;
import kernel.order.AgentOrdering;
import org.jdom2.Document;
import org.jdom2.Element;
import org.jdom2.input.SAXBuilder;
import simulation.DisProblem;
import simulation.DisSolverSimulator;
import simulation.parser.InstanceParser;
import util.Commands;
import util.Logger;
import util.Tools;
import benchmark.Generator;
import benchmark.ResultPrinter;
import dischoco.DisChoco;

public class URGCGenerator extends Generator {

	private int minNbAgents = 5; // number of variable in the problem
	private int stepNbAgents = 1; // number of variable in the problem
	private int maxNbAgents = 5; // number of variable in the problem
	private int minNbColours = 3; // the number of colours = domain size
	private int stepNbColours = 1; // the number of colours = domain size
	private int maxNbColours = 3; // the number of colours = domain size
	private int minDensity = 5;
	private int stepDensity = 1;
	private int maxDensity = 5;
	private int nbInstances = 100;
	private static final String NODE = "Node";
	private static final String DENSITY = "density";
	private static final String MIN_DENSITY = "MinDensity";
	private static final String MAX_DENSITY = "MaxDensity";
	private static final String STEP = "step";
	private static final String GraphColoring = Commands.GraphColoring;
	public static final String GC = Commands.GC;
	private static final String NB_FOLDERS = "nbFolders";
	private static final String DIRECTORY = "directory";
	private static final String NB_NODES = "nbAgents";
	private static final String NB_COLORES = "nbColores";

	private File directory = new File("./" + GC + "/");
	private int nbVarsPerAgent = 1;
	public final static String fixKeys = DENSITY; 
	public final static int nbHeaderLine = 4;

	public URGCGenerator(File directory, int nbInstances, String model) {
		super(GraphColoring, model, Commands.TKC);
		this.directory = directory;
		if (!this.directory.isDirectory()) {
			this.directory.mkdir();
		}
		this.nbInstances = nbInstances;
	}

	/**
	 * Constructor 
	 * @param directory
	 * @param nbInstances
	 * @param minNbAgents
	 * @param stepNbAgents
	 * @param maxNbAgents
	 * @param minDensity
	 * @param stepDensity
	 * @param maxDensity
	 * @param minNbColors
	 * @param stepNbColors
	 * @param maxNbColors
	 */
	public URGCGenerator(File directory, int nbInstances, String model, 
			int minNbAgents, int stepNbAgents, int maxNbAgents, 
			int minDensity, int stepDensity, int maxDensity, 
			int minNbColors, int stepNbColors, int maxNbColors) {
		this(directory, nbInstances, model);
		this.setNbAgents(minNbAgents, stepNbAgents, maxNbAgents);
		this.setDensity(minDensity, stepDensity, maxDensity);
		this.setNbCoulors(minNbColors, stepNbColors, maxNbColors);
	}
	
	/**
	 * Create the thread for running experiments 
	 * @param experimentation : the configuration file of experimentation
	 */
	public URGCGenerator(File file) {
		super(file);
	}

	public void setNbVarsPerAgent(Integer nbVarsPerAgent) {
		this.nbVarsPerAgent = nbVarsPerAgent;
	}

	public void setNbAgents(int minNbAgents, int stepNbAgents, int maxNbAgents) {
		this.minNbAgents = minNbAgents;
		this.stepNbAgents = stepNbAgents;
		this.maxNbAgents = maxNbAgents;
	}

	public void setDensity(int minDensity, int stepDensity, int maxDensity) {
		this.minDensity = minDensity;
		this.stepDensity = stepDensity;
		this.maxDensity = maxDensity;
	}

	public void setNbCoulors(int minNbColours, int stepNbColours, int maxNbColours) {
		this.minNbColours = minNbColours;
		this.stepNbColours = stepNbColours;
		this.maxNbColours = maxNbColours;
	}

	public Element getAgents(int nbAgents) {
		Element agents = new Element(AGENTS);
		agents.setAttribute(NB_AGENTS, "" + nbAgents);
		for (int id = 1 ; id <= nbAgents ; id++) {
			Element agent = new Element(AGENT);
			agent.setAttribute(NAME, "A" + id);
			agent.setAttribute(ID, "" + id);
			agent.setAttribute(DESCRIPTION, NODE + id);
			agents.addContent(agent);
		}
		return agents;
	}

	public Element getDomaines(int nbColours) {
		Element domains = new Element(DOMAINS);
		domains.setAttribute(NB_DOMAINS, "1");
		Element domain = new Element("domain");
		domain.setAttribute(NAME, "D1");
		domain.setAttribute(NB_VALUES, "" + nbColours);
		domain.setText("1.." + nbColours);
		domains.addContent(domain);
		return domains;
	}

	public Element getVariables(int nbAgents) {
		Element variables = new Element(VARIABLES);
		variables.setAttribute(NB_VARIABLES, "" + nbAgents);
		for (int id = 1 ; id <= nbAgents ; id++) {
			Element variable = new Element(VARIABLE);
			variable.setAttribute(AGENT, "A" + id);
			variable.setAttribute(NAME, "X" + id + ".1");
			variable.setAttribute(ID, "0");
			variable.setAttribute(DOMAIN, "D1");
			variable.setAttribute(DESCRIPTION, "X" + id + ".1");
			variables.addContent(variable);
		}
		return variables;
	}

	private Element getRelations(int nbColours) {
		Element relations = new Element(RELATIONS);
		relations.setAttribute(NB_RELATIONS, "1");
		Element relation = new Element(RELATION);
		relation.setAttribute(NAME, "R1");
		relation.setAttribute(ARITY, "2");
		relation.setAttribute(NB_TUPLES, "" + nbColours);
		relation.setAttribute(SEMANTICS, CONFLICTS);
		String tuples = "";
		for (int value = 1 ; value < nbColours ; value++) {
			tuples += value + " " + value + "|";
		}
		tuples += nbColours + " " + nbColours;
		relation.setText(tuples);
		relations.addContent(relation);
		return relations;
	}

	/**
	 * @param instance
	 * @param nbI
	 */
	private Element getConstraints(int nbAgents, int nbConstraints) {
		Random rand = new Random();
		rand.setSeed(System.currentTimeMillis());
		Element constraints = new Element(CONSTRAINTS);
		constraints.setAttribute(NB_CONSTRAINTS, "" + nbConstraints);
		/* CT means "constraint" */
		int PossibleCTs = nbAgents * (nbAgents - 1) / 2;
		/* Create an array for each possible binary constraint. */
		long CTarray[] = new long[PossibleCTs];
		long selectedCT;
		int i = 0;
		int var1, var2;

		/*
		 * Initialize the CTarray. Each entry has one var in the high two bytes, and the other in the low two bytes.
		 */
		for (var1 = 0 ; var1 < (nbAgents - 1) ; var1++) {
			for (var2 = var1 + 1 ; var2 < nbAgents ; var2++) {
				CTarray[i++] = (var1 << 16) | var2;
			}
		}
		/* Select this.nbConstraint constraints. */
		for (int c = 0 ; c < nbConstraints ; c++) {
			/* Choose a random number between c and PossibleCTs - 1, inclusive. */
			int r = c + (int) (rand.nextFloat() * (PossibleCTs - c));
			/* Swap elements [c] and [r]. */
			selectedCT = CTarray[r];
			CTarray[r] = CTarray[c];
			CTarray[c] = selectedCT;

			Element constraint = new Element(CONSTRAINT);
			constraint.setAttribute(NAME, "C" + (c + 1));
			constraint.setAttribute(MODEL, Commands.TKC);
			constraint.setAttribute(ARITY, "2");
			/* Broadcast the constraint. */
			var1 = (int) (CTarray[c] >> 16) + 1;
			var2 = (int) (CTarray[c] & 0x0000FFFF) + 1;
			// constraint.setAttribute(SCOPE_AGENTS, "A"+var1+" A"+var2);
			constraint.setAttribute(SCOPE, "X" + var1 + ".1" + " X" + var2 + ".1");
			constraint.setAttribute(REFERENCE, "R1");
			constraints.addContent(constraint);
		}
		return constraints;
	}


	/**
	 * This method return the description of the instance. This will be used in the 
	 * Graphical User Interface. 
	 * 
	 * @param nbOfInstance : the number of instance
	 * @param p1 : the network connectivity defined as the ratio of existing binary constraints,
	 * @param p2 : is the constraint tightness defined as the ratio of forbidden value pairs. 
	 * @param nbConstraint : the number of constraint in the instance
	 * @param nbTuples : the number of tuples 
	 * @return
	 */
	private Element getGuiPresentation(int numInstance, int nbAgents, int domainSize, int p1, int nbConstraint, int nbTuples) {
		Element presentation = new Element(GuiPresentation);
		presentation.setAttribute(TYPE, DCOP);
		presentation.setAttribute(BENCHMARK, this.typeOfBenchmark);
		presentation.setAttribute(NAME, INSTANCE + (numInstance));
		presentation.setAttribute(MODEL, this.constraintsModel);
		presentation.setAttribute(NB_AGENTS, Integer.toString(nbAgents));
		presentation.setAttribute(DomainSize, Integer.toString(domainSize));
		presentation.setAttribute(DENSITY, Integer.toString(p1));
		presentation.setAttribute(TIGHTNESS, Integer.toString(100/domainSize));
		presentation.setAttribute(NB_CONSTRAINTS, Integer.toString(nbConstraint));
		presentation.setAttribute(NB_TUPLES, Integer.toString(nbTuples));
		return presentation;
	}

	public Element getInstance(int nbAgents, int nbColours, int p1, int numInstance) {
		int nbConstraint = (p1 * nbAgents * (nbAgents - 1)) / 200;
		Element instance = new Element(INSTANCE);
		instance.addContent(this.getPresentation(numInstance, DisCSP));
		instance.addContent(this.getAgents(nbAgents));
		instance.addContent(this.getDomaines(nbColours));
		instance.addContent(this.getVariables(nbAgents));
		instance.addContent(this.getRelations(nbColours));
		instance.addContent(this.getConstraints(nbAgents, nbConstraint));
		instance.addContent(this.getGuiPresentation(numInstance, nbAgents, nbColours, p1, nbConstraint, nbColours));
		return instance;
	}

	/**
	 * 
	 */
	public void generate() {
		String path = this.directory.getPath();
		Element experiments = new Element(Commands.Experiments);
		Document confDocument = new Document(experiments);
		experiments.setAttribute(TYPE,  DisCSP);
		experiments.setAttribute(BENCHMARK, this.typeOfBenchmark);
		Element instances = new Element(INSTANCES);
		experiments.addContent(instances);
		instances.setAttribute(NAME, GC);
		instances.setAttribute(TYPE,  DisCSP);
		instances.setAttribute(BENCHMARK, this.typeOfBenchmark);
		int nbN = ((this.maxNbAgents - this.minNbAgents) / this.stepNbAgents + 1);
		int nbC = ((this.maxNbColours - this.minNbColours) / this.stepNbColours + 1);
		instances.setAttribute(NB_FOLDERS, "" + (nbN * nbC));
		// n=number of nodes , k=number of colours , c=number of constraint and
		// i= numeral of instance
		for (int n = this.minNbAgents ; n <= this.maxNbAgents ; n += this.stepNbAgents) {
			for (int k = this.minNbColours ; k <= this.maxNbColours ; k += this.stepNbColours) {
				File folder = new File(path + File.separator + n + "_" + k);
				if (!folder.isDirectory()) {
					folder.mkdir();
				}
				Element dir = new Element(DIRECTORY);
				instances.addContent(dir);
				dir.setAttribute(NB_NODES, "" + n);
				dir.setAttribute(NB_COLORES, "" + k);
				dir.setAttribute(NB_INSTANCES, "" + this.nbInstances);
				Element density = new Element(DENSITY);
				density.setAttribute(MIN_DENSITY, this.minDensity + "");
				density.setAttribute(MAX_DENSITY, this.maxDensity + "");
				density.setAttribute(STEP, this.stepDensity + "");
				dir.addContent(density);
				for (int c = this.minDensity ; c <= this.maxDensity ; c += this.stepDensity) {
					String d = (c < 10? "0" + c : "" + c);
					File subfolder = new File(folder + File.separator + d);
					if (!subfolder.isDirectory()) {
						subfolder.mkdir();
					}
					for (int i = 1 ; i <= this.nbInstances ; i++) {
						Element instance = null;
						if(nbVarsPerAgent==1) {
							instance = this.getInstance(n, k, c, i);
						} else {
							//TODO implement a generator for complex graph coloring problems
							throw new UnsupportedOperationException("Not yet implemented");
						}
						String fileName = folder + File.separator + d + File.separator + Commands.GC + "_" + i + ".xml";
						this.save(fileName, new Document(instance));
					}
				}
			}
		}
		String experimentsConfFileName = Commands.Experiments + "_" + this.nbInstances + ".xml";
		this.saveXML(path, experimentsConfFileName, confDocument);
	}
	
	@Override
	public void run() {
		int nbInstance, nbAgents, nbColours;
		String filePath = this.experimentation.getParent();
		ArrayList<String> keys = null;
		// //////////////////////////////////////////////////////////////////////////////////////////////////////////
		Document document = null;
		SAXBuilder sxb = new SAXBuilder();
		try {
			document = sxb.build(this.experimentation);
		} catch (Exception e) {}
		Element xp = document.getRootElement();
		Element instances = xp.getChild(INSTANCES);
		List<Element> communications = xp.getChild(Commands.Communications).getChildren();
		List<Element> algorithms = xp.getChild(Commands.Algorithms).getChildren();
		String[] protocols = new String[algorithms.size()];
		String[] orderings = new String[algorithms.size()];
		int a = 0;
		Iterator<Element> it = algorithms.iterator();
		while (it.hasNext()) {
			Element algorithm = it.next();
			protocols[a] = algorithm.getText();
			orderings[a] = algorithm.getAttributeValue(Commands.AO);
			a++;
		}
		it = communications.iterator();
		while (it.hasNext()) {
			// /////////////////////////////////////////////////////////////////
			Element comm = it.next();
			String commType = comm.getAttributeValue(TYPE);
			String channelType = comm.getAttributeValue(Commands.channelType);
			int minURC = new Integer(comm.getAttributeValue(Commands.MinURC));
			int maxURC = new Integer(comm.getAttributeValue(Commands.MaxURC));
			int nbRun = new Integer(comm.getAttributeValue(Commands.NbRun));
			long maxNCCCs = new Long(comm.getAttributeValue(Commands.MaxNCCCs));
			// /////////////////////////////////////////////////////////////////
			String name = instances.getAttribute(NAME).getValue();
			int nbFolders = new Integer(instances.getAttributeValue(NB_FOLDERS));
			List<Element> config = instances.getChildren();
			for (int f = 0 ; f < nbFolders ; f++) {
				Element dir = config.get(f);
				nbAgents = new Integer(dir.getAttributeValue(NB_NODES));
				nbColours = new Integer(dir.getAttributeValue(NB_COLORES));
				nbInstance = new Integer(dir.getAttributeValue(NB_INSTANCES));
				Element density = dir.getChild(DENSITY);
				String directory = filePath + File.separator + Commands.Results+"_"+nbInstance+"_"+nbRun;
				try {
					File d = new File(directory);
					if (!d.exists()) {
						d.mkdir();
					}
				} catch (Exception e) {}
				directory += File.separator + commType + "_" + channelType;
				try {
					File thefile = new File(directory);
					if (!thefile.exists()) {
						thefile.mkdir();
					}
				} catch (Exception e) {}
				String subDirectory = directory + File.separator + nbAgents + "_" + nbColours;
				try {
					File thefile = new File(subDirectory);
					if (!thefile.exists()) {
						thefile.mkdir();
					}
				} catch (Exception e) {}
				for (int p = 0 ; p < protocols.length ; p++) { // Protocols
					Logger allResults = new Logger(subDirectory, "All_" + protocols[p] + "-" + orderings[p] + ".csv");
					Logger results = new Logger(subDirectory, protocols[p] + "-" + orderings[p] + ".csv");
					int minDensity = new Integer(density.getAttributeValue(MIN_DENSITY));
					int maxDensity = new Integer(density.getAttributeValue(MAX_DENSITY));
					int stepDensity = new Integer(density.getAttributeValue(STEP));
					String path = filePath + File.separator + nbAgents + "_" + nbColours;
					for (int d = minDensity ; d <= maxDensity ; d += stepDensity) {
						String den = (d < 10? "0" + d : "" + d);
						HashMap<String, Double> densityStatistics = new HashMap<String, Double>();
						for (int i = 1 ; i <= nbInstance ; i++) {
							String subFolder = (d < 10? "0" + d : "" + d);
							String problem = path + File.separator + subFolder + File.separator + name + "_" + i + ".xml.bz2";
							HashMap<String, Double> instanceStatistics = new HashMap<String, Double>();
							for (int r = 1 ; r <= nbRun ; r++) {
								DisProblem disProb = new DisProblem(problem, protocols[p]);
								DisSolverSimulator solver = new DisSolverSimulator(disProb);
								solver.setCutOffNCCCs(maxNCCCs);
								solver.addCommunicationMetric();
								solver.setMessageDelay(minURC, maxURC);
								// solver.CheckTheReturnedSolution(true);
								solver.setCentralizedAO(AgentOrdering.getCentralizedAO(orderings[p]));
								solver.solve();
								if(keys == null ){
									keys = disProb.master.getStatisticsKeys();
								}
								HashMap<String, Double> runStatistics = solver.getStatistics();
								if (d == minDensity && i == 1 && r == 1 && d == minDensity) {
									String header = Commands.GraphColoring + COMMA;
									header += Commands.Algorithm + COMMA + protocols[p] + COMMA + Commands.AO + COMMA + orderings[p] + "\n";
									header += NB_NODES + COMMA + nbAgents + COMMA + NB_COLORES + COMMA + nbColours + "\n";
									header += Commands.channelType + COMMA + channelType + COMMA + Commands.Communication + COMMA ;
									header += commType + COMMA + Commands.MinURC + COMMA + minURC + COMMA + Commands.MaxURC + COMMA + maxURC +  "\n\n";
									header += DENSITY + COMMA + INSTANCE + COMMA + Commands.NbRun + COMMA + Tools.toString(keys);
									allResults.printToLogln(header);
									header = Commands.GraphColoring + COMMA;
									header += Commands.Algorithm + COMMA + protocols[p] + COMMA + Commands.AO + COMMA + orderings[p] + "\n";
									header += NB_NODES + COMMA + nbAgents + COMMA + NB_COLORES + COMMA + nbColours + "\n";
									header += Commands.channelType + COMMA + channelType + COMMA + Commands.Communication + COMMA + commType ;
									header += COMMA + Commands.MinURC + COMMA + minURC + COMMA + Commands.MaxURC + COMMA + maxURC + "\n\n";
									header += DENSITY + COMMA + Tools.toString(keys);
									results.printToLogln(header);
								}
								allResults.printToLogln("0." +den + COMMA + i + COMMA + r + COMMA + Tools.statisticsToString(keys, runStatistics));
								instanceStatistics = Tools.merge(runStatistics, instanceStatistics);
								disProb.setAgents(null);
								System.gc();
							}
							Tools.average(instanceStatistics, nbRun);
							densityStatistics = Tools.merge(instanceStatistics, densityStatistics);
						}
						Tools.average(densityStatistics, nbInstance);
						results.printToLogln("0." +den + COMMA + Tools.statisticsToString(keys, densityStatistics));
					}
					results.close();
					allResults.close();
					keys = null;
				}
				String pathName = subDirectory;
				ResultPrinter printer = new ResultPrinter(protocols, orderings, nbHeaderLine);
				printer.setFixKeys(fixKeys);
				ArrayList<String> paths = printer.printToGnuPlot(pathName, "", commType);
				printer.plot(pathName, paths);
				//////
				String header = Commands.channelType + COMMA + channelType + COMMA + Commands.Communication + COMMA ; 
				header += commType + COMMA + Commands.MinURC + COMMA + minURC + COMMA + Commands.MaxURC + COMMA + maxURC;
				printer.printAllResults(pathName, header);
			}
		}
	}

	public void lunch() {
		int nbInstance, nbAgents, nbColours;
		String filePath = new File(DisChoco.filePath).getParent();
		ArrayList<String> keys = null;
		String directory = filePath + File.separator + Commands.Results;
		try {
			File d = new File(directory);
			if (!d.exists()) {
				d.mkdir();
			}
		} catch (Exception e) {}
		// //////////////////////////////////////////////////////////////////////////////////////////////////////////
		Document document = null;
		SAXBuilder sxb = new SAXBuilder();
		try {
			document = sxb.build(DisChoco.filePath);
		} catch (Exception e) {}
		Element xp = document.getRootElement();
		Element instances = xp.getChild(INSTANCES);
		List<Element> communications = xp.getChild(Commands.Communications).getChildren();
		List<Element> algorithms = xp.getChild(Commands.Algorithms).getChildren();
		String[] protocols = new String[algorithms.size()];
		String[] orderings = new String[algorithms.size()];
		int a = 0;
		Iterator<Element> it = algorithms.iterator();
		while (it.hasNext()) {
			Element algorithm = it.next();
			protocols[a] = algorithm.getText();
			orderings[a] = algorithm.getAttributeValue(Commands.AO);
			a++;
		}
		it = communications.iterator();
		while (it.hasNext()) {
			// /////////////////////////////////////////////////////////////////
			Element comm = it.next();
			String commType = comm.getAttributeValue(TYPE);
			String channelType = comm.getAttributeValue(Commands.channelType);
			int minURC = new Integer(comm.getAttributeValue(Commands.MinURC));
			int maxURC = new Integer(comm.getAttributeValue(Commands.MaxURC));
			int nbRun = new Integer(comm.getAttributeValue(Commands.NbRun));
			long maxNCCCs = new Long(comm.getAttributeValue(Commands.MaxNCCCs));
			// /////////////////////////////////////////////////////////////////
			directory = filePath + File.separator + Commands.Results + File.separator + commType + "_" + channelType;
			try {
				File f = new File(directory);
				if (!f.exists()) {
					f.mkdir();
				}
			} catch (Exception e) {}
			String name = instances.getAttribute(NAME).getValue();
			int nbFolders = new Integer(instances.getAttributeValue(NB_FOLDERS));
			List<Element> config = instances.getChildren();
			for (int p = 0 ; p < protocols.length ; p++) { // Protocols
				for (int f = 0 ; f < nbFolders ; f++) {
					Element dir = config.get(f);
					nbAgents = new Integer(dir.getAttributeValue(NB_NODES));
					nbColours = new Integer(dir.getAttributeValue(NB_COLORES));
					Logger allResults =
							new Logger(directory + File.separator + nbAgents + "_" + nbColours, "All_" + protocols[p] + "-" + orderings[p] + ".csv");
					Logger results = new Logger(directory + File.separator + nbAgents + "_" + nbColours, protocols[p] + "-" + orderings[p] + ".csv");
					nbInstance = new Integer(dir.getAttributeValue(NB_INSTANCES));
					Element density = dir.getChild(DENSITY);
					int minDensity = new Integer(density.getAttributeValue(MIN_DENSITY));
					int maxDensity = new Integer(density.getAttributeValue(MAX_DENSITY));
					int stepDensity = new Integer(density.getAttributeValue(STEP));
					String path = filePath + File.separator + nbAgents + "_" + nbColours;
					for (int d = minDensity ; d <= maxDensity ; d += stepDensity) {
						String den = (d < 10? "0" + d : "" + d);
						HashMap<String, Double> stat1 = new HashMap<String, Double>();
						DisSolverSimulator solver = null;
						for (int i = 1 ; i <= nbInstance ; i++) {
							String subFolder = (d < 10? "0" + d : "" + d);
							DisProblem disProb = null;
							InstanceParser parser = new InstanceParser(name + "_" + i);
							String problem = path + File.separator + subFolder + File.separator + name + "_" + i + ".xml.bz2";
							parser.loadInstance(problem);
							parser.parse(false);
							HashMap<String, Double> stat2 = new HashMap<String, Double>();
							for (int r = 1 ; r <= nbRun ; r++) {
								AgentOrdering.getCentralizedAO(orderings[p]);
								disProb = new DisProblem(parser, protocols[p]);
								solver = new DisSolverSimulator(disProb);
								solver.setCutOffNCCCs(maxNCCCs);
								solver.solve();
								if(keys == null ){
									keys = disProb.master.getStatisticsKeys();
								}
								HashMap<String, Double> stat = solver.getStatistics();
								if (d == minDensity && i == 1 && r == 1 && d == minDensity) {
									String header = Commands.GraphColoring + COMMA;
									header += Commands.Algorithm + COMMA + protocols[p] + COMMA + Commands.AO + COMMA + orderings[p] + "\n";
									header += Commands.channelType + COMMA + channelType + COMMA + Commands.Communication + COMMA + commType;
									header += COMMA + Commands.MinURC + COMMA + minURC + COMMA + Commands.MaxURC + COMMA + maxURC + "\n\n";
									header += NB_NODES + COMMA + NB_COLORES + COMMA + DENSITY + COMMA + INSTANCE + COMMA + Commands.NbRun + COMMA + keys;
									allResults.printToLogln(header);
									header = Commands.GraphColoring + COMMA;
									header += Commands.Algorithm + COMMA + protocols[p] + COMMA + Commands.AO + COMMA + orderings[p] + "\n";
									header += Commands.channelType + COMMA + channelType + COMMA + Commands.Communication + COMMA + commType ;
									header += COMMA + Commands.MinURC + COMMA + minURC + COMMA + Commands.MaxURC + COMMA + maxURC + "\n\n";
									header += DENSITY + COMMA + keys;
									results.printToLogln(header);
								}
								stat2 = Tools.merge(stat, stat2);
								allResults.printToLogln("0." +den + COMMA + i + COMMA + r + COMMA + Tools.statisticsToString(keys, stat));
								solver = null;
								System.gc();
							}
							Tools.average(stat2, nbRun);
							stat1 = Tools.merge(stat2, stat1);
							parser = null;
						}
						Tools.average(stat1, nbInstance);
						results.printToLogln("0." +den + COMMA + Tools.statisticsToString(keys, stat1));
					}
					results.close();
					allResults.close();
					keys = null;
				}
			}
			String header =
					Commands.channelType + COMMA + channelType + COMMA + Commands.Communication + COMMA + commType + COMMA + Commands.MinURC + COMMA + minURC
							+ COMMA + Commands.MaxURC + COMMA + maxURC;
			for (int f = 0 ; f < nbFolders ; f++) {
				Element dir = config.get(f);
				nbAgents = new Integer(dir.getAttributeValue(NB_NODES));
				nbColours = new Integer(dir.getAttributeValue(NB_COLORES));
				ResultPrinter printer = new ResultPrinter(protocols, orderings, nbHeaderLine);
				printer.setFixKeys(fixKeys);
				String subFolder = nbAgents + "_" + nbColours;
				ArrayList<String> paths = printer.printToGnuPlot(directory + File.separator + subFolder, "", commType);
				printer.plot(directory + File.separator + subFolder, paths);
				printer.printAllResults(directory + File.separator + subFolder, header);
			}
		}
	}
	
}
