/*
 * DisChoco: Distributed Constraint Reasoning Solver
 * Copyright (C) 2007-2013 Redouane Ezzahir & Mohamed Wahbi
 * DisChoco is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * DisChoco is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 * To contact the authors: <http://dischoco.sourceforge.net/>
 */
package benchmark.meetingscheduling;

import static simulation.parser.InstanceTokens.AGENT;
import static simulation.parser.InstanceTokens.AGENTS;
import static simulation.parser.InstanceTokens.ARITY;
import static simulation.parser.InstanceTokens.BENCHMARK;
import static simulation.parser.InstanceTokens.COMPLEX_MODEL;
import static simulation.parser.InstanceTokens.CONSTRAINT;
import static simulation.parser.InstanceTokens.CONSTRAINTS;
import static simulation.parser.InstanceTokens.COST_SEPARATOR;
import static simulation.parser.InstanceTokens.DCOP;
import static simulation.parser.InstanceTokens.DEFAULT_COST;
import static simulation.parser.InstanceTokens.DESCRIPTION;
import static simulation.parser.InstanceTokens.DOMAIN;
import static simulation.parser.InstanceTokens.DOMAINS;
import static simulation.parser.InstanceTokens.EXPRESSION;
import static simulation.parser.InstanceTokens.FUNCTIONAL;
import static simulation.parser.InstanceTokens.ID;
import static simulation.parser.InstanceTokens.INFINITY;
import static simulation.parser.InstanceTokens.INITIAL_COST;
import static simulation.parser.InstanceTokens.INSTANCE;
import static simulation.parser.InstanceTokens.INSTANCES;
import static simulation.parser.InstanceTokens.MAXIMAL_COST;
import static simulation.parser.InstanceTokens.MODEL;
import static simulation.parser.InstanceTokens.NAME;
import static simulation.parser.InstanceTokens.NB_AGENTS;
import static simulation.parser.InstanceTokens.NB_CONSTRAINTS;
import static simulation.parser.InstanceTokens.NB_DOMAINS;
import static simulation.parser.InstanceTokens.NB_INSTANCES;
import static simulation.parser.InstanceTokens.NB_PREDICATES;
import static simulation.parser.InstanceTokens.NB_RELATIONS;
import static simulation.parser.InstanceTokens.NB_TUPLES;
import static simulation.parser.InstanceTokens.NB_VALUES;
import static simulation.parser.InstanceTokens.NB_VARIABLES;
import static simulation.parser.InstanceTokens.PARAMETERS;
import static simulation.parser.InstanceTokens.PREDICATE;
import static simulation.parser.InstanceTokens.PREDICATES;
import static simulation.parser.InstanceTokens.REFERENCE;
import static simulation.parser.InstanceTokens.RELATION;
import static simulation.parser.InstanceTokens.RELATIONS;
import static simulation.parser.InstanceTokens.SCOPE;
import static simulation.parser.InstanceTokens.SEMANTICS;
import static simulation.parser.InstanceTokens.SOFT;
import static simulation.parser.InstanceTokens.TKC;
import static simulation.parser.InstanceTokens.TUPLES_SEPARATOR;
import static simulation.parser.InstanceTokens.TYPE;
import static simulation.parser.InstanceTokens.VALUE_SEPARATOR;
import static simulation.parser.InstanceTokens.VARIABLE;
import static simulation.parser.InstanceTokens.VARIABLES;
import static simulation.parser.InstanceTokens.VARIABLE_IDENTIFIER_SEPARATOR;
import static simulation.parser.InstanceTokens.domainToString;
import static simulation.parser.InstanceTokens.getAgentNameFor;
import static simulation.parser.InstanceTokens.getConstraintNameFor;
import static simulation.parser.InstanceTokens.getDomainNameFor;
import static simulation.parser.InstanceTokens.getMeetingNameFor;
import static simulation.parser.InstanceTokens.getRelationNameFor;
import static util.Commands.DIRECTORY;
import static util.Commands.DISPONIBILITY;
import static util.Commands.GuiPresentation;
import static util.Commands.MAX_AGENTS_PER_MEETING;
import static util.Commands.MEETING;
import static util.Commands.MEETINGS;
import static util.Commands.MEETINGS_DURATION;
import static util.Commands.MIN_AGENTS_PER_MEETING;
import static util.Commands.NB_AGENTS_PER_MEETING;
import static util.Commands.NB_DAYS;
import static util.Commands.NB_HOURS;
import static util.Commands.NB_MEETINGS;
import static util.Commands.STEP_AGENTS_PER_MEETING;
import java.io.File;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;
import kernel.order.AgentOrdering;
import org.jdom2.Document;
import org.jdom2.Element;
import org.jdom2.input.SAXBuilder;
import simulation.DisProblem;
import simulation.DisSolverSimulator;
import simulation.parser.PredicateTokens;
import util.Commands;
import util.Logger;
import util.Tools;
import benchmark.Generator;
import benchmark.ResultPrinter;

/** 
 * This class implements a Distributed Meeting Scheduling Problem generator of the Distributed Multi-Event 
 * Scheduling (DiMES) [R. T. Maheswaran et al; 2004] <i>``Taking DCOP to the Real World: Efficient Complete 
 * Solutions for Distributed Multi-Event Scheduling''</i>.
 * 
 * @author Mohamed Wahbi
 */
public class DCOPMeetingScheduling extends Generator {

	/** The mode of encoding */
	public enum Mode {
		/** Events As (private) Variables */
		EAV,
		/** Private Events As (private) Variables */
		PEAV,
		/** Events As Shared Variables */
		EASV
	}

	public final static String fixKeys = NB_AGENTS_PER_MEETING;

	/** For each meeting, its list of attendees */
	private ArrayList<ArrayList<Integer>> attendees;

	/** For each participant, its list of meetings */
	private ArrayList<ArrayList<Integer>> meetings;

	/** The cost that each participant assigns to having a meeting in each time slot */
	private ArrayList<ArrayList<Integer>> preferences;

	/** Number of available time slots for each meeting */

	public static int nbHeaderLine = 4;
	private static int agentsPerMeet;
	private static int instance;
	private static int run;
	/** the number of instances to generate for each class */
	private int nbInstances = 25;
	/** the ranges of number of agents */
	private int minNbAgents, maxNbAgents, stepNbAgents;
	/** the ranges of number of meetings */
	private int minNbMeetings, maxNbMeetings, stepNbMeetings;
	/** the ranges of number of meeting per agent */
	private int minNbAgentsPerMeeting, maxNbAgentsPerMeeting, stepNbAgentsPerMeeting = 0;
	/** for each agent there are a slots */
	private Slot[] slots;
	private int meetingsDuration = 1;
	private File directory = new File(Commands.MSP);
	/** The maximum value of the costs */
	private Integer maxCost;
	/**  numbers of days to put in the slots */
	private int nbDays = 5;
	/**  numbers of hours in the day to put in the slots */
	private int nbHours = 10;
	/**  percentage of disponibility over <i>numbers of hours</i> <code>x</code> <i>numbers of days</i> */
	private int disponibility = 100;
	private final Mode mode;
	private String constraintModel = TKC;
	private boolean intension = true; // intension:true; extension:false

	/** 
	 * Constructor
	 * @param directory 			the directory that will contain all generated instances
	 * @param totallyKnownConstraint the totally known constraint model or partially known constraint one
	 * @param intension				the intension or extension
	 * 
	 * @param nbInstances			number of instances to generate
	 * @param meetingsDuration		the duration of meetings in the number of slots
	 * 
	 * @param nbDays 				number of days
	 * @param nbHours 				number of hours per day
	 * @param disponibility 		percentage of free slots in the (nbDays * nbHours) slots.
	 * 
	 * @param minNbAgents 			minimum number of agents
	 * @param maxNbAgents 			maximum number of agents
	 * @param stepNbAgents 			step over the number of agents
	 * 
	 * @param minNbMeetings 		minimum number of meetings
	 * @param maxNbMeetings 		maximum number of meetings
	 * @param stepNbMeetings 		step over the number of meetings
	 * 
	 * @param maxCost 				if not null, each attendee assigns a random cost in [0, value] to having any meeting at each time slot
	 */
	public DCOPMeetingScheduling(String constraintsModel, 
			File directory, boolean intension, int nbInstances, int meetingsDuration,
			int nbDays, int nbHours, int disponibility, int minNbAgents, int maxNbAgents, int stepNbAgents, int minNbMeetings, int maxNbMeetings,
			int stepNbMeetings, int minNbAgentsPerMeeting, int maxNbAgentsPerMeeting, int stepNbAgentsPerMeeting, Integer maxCost, Mode mode) {
		super(Commands.DIMES, COMPLEX_MODEL, constraintsModel);
		this.directory = directory;
		if (!this.directory.isDirectory()) {
			this.directory.mkdir();
		}
		this.constraintModel = constraintsModel;
		this.intension = intension;
		this.nbInstances = nbInstances;
		this.meetingsDuration = meetingsDuration;

		this.nbDays = nbDays;
		this.nbHours = nbHours;
		this.disponibility = disponibility;

		this.minNbAgents = minNbAgents;
		this.maxNbAgents = maxNbAgents;
		this.stepNbAgents = stepNbAgents;

		this.minNbMeetings = minNbMeetings;
		this.maxNbMeetings = maxNbMeetings;
		this.stepNbMeetings = stepNbMeetings;

		this.minNbAgentsPerMeeting = minNbAgentsPerMeeting;
		this.maxNbAgentsPerMeeting = maxNbAgentsPerMeeting;
		this.stepNbAgentsPerMeeting = stepNbAgentsPerMeeting;

		this.maxCost = maxCost;
		this.mode = mode;
	}
	
	/**
	 * Create the thread for running experiments 
	 * @param experimentation : the configuration file of experimentation
	 */
	public DCOPMeetingScheduling(File file) {
		super(file);
		this.mode = null;
	}

	/** 
	 * @param nbAgents number of agents
	 * @param nbMeetings number of meetings
	 * @param nbAgentsPerMeeting number of agents per meeting
	 */
	public Element generateInstance(int numInstance, int nbAgents, int nbMeetings, int nbAgentsPerMeeting) {
		Random rand = new Random();

		// Create time slots
		this.slots = new Slot[nbAgents];
		for(int a = 0 ; a < nbAgents ; a++) {
			this.slots[a] = Slot.generateWeek(this.nbDays, this.nbHours, this.disponibility);
		}

		// Create the agents' preferences
		if (this.maxCost != null) { // so that rand(0, maxCost) is in [0, maxCost]
			this.preferences = new ArrayList<ArrayList<Integer>>(nbAgents);
			for(int a = 0 ; a < nbAgents ; a++) {
				int nbSlots = slots[a].getNbSlot();
				ArrayList<Integer> pref = new ArrayList<Integer>(nbSlots);
				for(int j = nbSlots - 1 ; j >= 0 ; j--) {
					pref.add(rand.nextInt(maxCost));
				}
				this.preferences.add(pref);
			}
		}
		int cmp = 0;
		HashSet<Integer> scheduledAgents = null;
		ArrayList<Integer> agents = null;
		do {
			scheduledAgents = new HashSet<Integer>();
			// Create the pool of agents
			agents = new ArrayList<Integer>(nbAgents);
			this.meetings = new ArrayList<ArrayList<Integer>>(nbAgents);
			for(int a = 0 ; a < nbAgents ; a++) {
				agents.add(a);
				this.meetings.add(new ArrayList<Integer>(nbMeetings));
			}

			// Create the meetings
			this.attendees = new ArrayList<ArrayList<Integer>>(nbMeetings);
			for(int m = 0 ; m < nbMeetings ; m++) {
				// Randomly choose participants
				ArrayList<Integer> participants = new ArrayList<Integer>(nbAgentsPerMeeting);
				this.attendees.add(participants);
				LinkedList<Integer> selectedAgents = new LinkedList<Integer>(agents);
				Collections.shuffle(selectedAgents);
				for(int j = 0 ; j < nbAgentsPerMeeting ; j++) {
					Integer agent = selectedAgents.removeFirst();
					participants.add(agent);
					this.meetings.get(agent).add(m);
				}
				scheduledAgents.addAll(participants);
			}
		} while (!scheduledAgents.containsAll(agents) && cmp < 10e4);
		if (!scheduledAgents.containsAll(agents)) {
			throw new UnsupportedOperationException("Impossible to generate uniform instance with the given parameters.\n"
					+ "Please check parameters: nbAgents=" + nbAgents + " nbMeetings=" + nbMeetings + " nbAgentsPerMeeting=" + nbAgentsPerMeeting);
		}
		return this.getInstance(numInstance, nbMeetings, nbAgents, nbAgentsPerMeeting);
	}

	/**
	 * 
	 * @param nbMeetings
	 * @param nbAgents
	 * @param nbAgentsPerMeeting
	 * @return
	 */
	public Element getInstance(int numInstance, int nbMeetings, int nbAgents, int nbAgentsPerMeeting) {
		Element instance = new Element(INSTANCE);
		instance.addContent(this.getPresentation(numInstance, DCOP));
		instance.addContent(this.getAgents(nbAgents));
		instance.addContent(this.getDomaines(nbAgents));
		instance.addContent(this.getVariables(nbMeetings));
		if (this.intension && mode == Mode.PEAV) {
			instance.addContent(this.getPredicates());
		}
		instance.addContent(this.getRelations());
		instance.addContent(this.getConstraints(nbAgents));
		instance.addContent(this.getGuiPresentation(nbMeetings, nbAgents, nbAgentsPerMeeting, this.nbInstances));
		return instance;
	}

	/**
	 * 
	 * @param nbAgents
	 * @return
	 */
	private Element getAgents(int nbAgents) {
		Element agents = new Element(AGENTS);
		agents.setAttribute(NB_AGENTS, Integer.toString(nbAgents));
		for(int id = 1 ; id <= nbAgents ; id++) {
			Element agent = new Element(AGENT);
			agent.setAttribute(NAME, getAgentNameFor(id));
			agent.setAttribute(ID, Integer.toString(id));
			agent.setAttribute(DESCRIPTION, AGENT + id);
			agents.addContent(agent);
		}
		return agents;
	}

	private Element getDomaines(int nbAgents) {
		Element domains = new Element(DOMAINS);
		domains.setAttribute(NB_DOMAINS, Integer.toString(nbAgents));
		for(int agId = 1 ; agId <= nbAgents ; agId++) {
			Element domain = new Element(DOMAIN);
			domain.setAttribute(NAME, getDomainNameFor(agId));
			Integer[] values = this.slots[agId - 1].toIntDomain();
			domain.setAttribute(NB_VALUES, Integer.toString(values.length));
			domain.setText(domainToString(values));
			domains.addContent(domain);
		}
		return domains;
	}

	private Element getVariables(int nbMeetings) {
		Element variables = new Element(VARIABLES);
		for(int m = 0 ; m < nbMeetings ; m++) {
			ArrayList<Integer> participants = this.attendees.get(m);
			Element variable = new Element(VARIABLE);
			variables.addContent(variable);
			Integer agId = participants.get(0) + 1;
			String agentName = getAgentNameFor(agId);
			int meeting = m + 1;
			String varName = "", id = Integer.toString(1);
			if (mode == Mode.PEAV) {
				varName = getMeetingNameFor(agId + VARIABLE_IDENTIFIER_SEPARATOR + (meeting));
				id = Integer.toString(meeting);
			} else {
				varName = getMeetingNameFor(meeting + VARIABLE_IDENTIFIER_SEPARATOR + Integer.toString(1));
			}
			if (mode != Mode.EASV) {
				variable.setAttribute(AGENT, agentName);
			}
			variable.setAttribute(NAME, varName);
			variable.setAttribute(ID, id);
			variable.setAttribute(DOMAIN, getDomainNameFor(agId));
			variable.setAttribute(DESCRIPTION, "M_" + (id));
			if (mode == Mode.PEAV) {
				for(int i = 1 ; i < participants.size() ; i++) {
					variable = new Element(VARIABLE);
					variables.addContent(variable);
					agId = participants.get(i) + 1;
					agentName = getAgentNameFor(agId);
					varName = getMeetingNameFor(agId + VARIABLE_IDENTIFIER_SEPARATOR + meeting);
					variable.setAttribute(AGENT, agentName);
					variable.setAttribute(NAME, varName);
					variable.setAttribute(ID, id);
					variable.setAttribute(DOMAIN, getDomainNameFor(agId));
					variable.setAttribute(DESCRIPTION, "M_" + (id));
				}
			}
		}
		variables.setAttribute(NB_VARIABLES, Integer.toString(variables.getContentSize()));
		return variables;
	}

	/**
	 * 
	 */
	private Element getPredicates() {
		Element predicates = new Element(PREDICATES);
		predicates.addContent(getEqualPredicate());
		predicates.addContent(getNonEqualPredicate());
		predicates.setAttribute(NB_PREDICATES, Integer.toString(predicates.getContentSize()));
		return predicates;
	}

	/**
	 * 
	 */
	private Element getEqualPredicate() {
		Element eq = new Element(PREDICATE);
		eq.setAttribute(NAME, "EQ");
		Element params = new Element(PARAMETERS);
		params.setText("int X int Y");
		eq.addContent(params);
		Element expression = new Element(EXPRESSION);
		Element functional = new Element(FUNCTIONAL);
		functional.setText(PredicateTokens.EQ + "(X,Y)");
		expression.addContent(functional);
		eq.addContent(expression);
		return eq;
	}

	/**
	 * 
	 */
	private Element getNonEqualPredicate() {
		Element ne = new Element(PREDICATE);
		ne.setAttribute(NAME, "NE");
		Element params = new Element(PARAMETERS);
		params.setText("int X int Y");
		ne.addContent(params);
		Element expression = new Element(EXPRESSION);
		Element functional = new Element(FUNCTIONAL);
		functional.setText(PredicateTokens.NE + "(X, Y)");
		expression.addContent(functional);
		ne.addContent(expression);
		return ne;
	}

	private Element getRelations() {
		Element relations = new Element(RELATIONS);
		createPreferenceRelations(relations);
		if (!this.intension) {
			// we need binary NEQ and EQ relations
			Element relation = new Element(RELATION);
			relation.setAttribute(NAME, "NE");
			relation.setAttribute(ARITY, Integer.toString(2));
			relation.setAttribute(SEMANTICS, SOFT);
			relation.setAttribute(DEFAULT_COST, "0");
			Integer[] values = Slot.generateWeek(nbDays, nbHours, 100).toIntDomain();
			int nbSlots = values.length;
			relation.setAttribute(NB_TUPLES, Integer.toString(nbSlots));
			Integer value = values[0];
			StringBuilder sb = new StringBuilder(INFINITY + COST_SEPARATOR + Integer.toString(value) + VALUE_SEPARATOR + Integer.toString(value));
			for(int i = 1 ; i < values.length ; i++) {
				value = values[i];
				sb.append(TUPLES_SEPARATOR);
				sb.append(Integer.toString(value));
				sb.append(VALUE_SEPARATOR);
				sb.append(Integer.toString(value));
			}
			relation.setText(sb.toString());
			relations.addContent(relation);
			if (mode == Mode.PEAV) { // we also need the binary EQ relation
				relation = new Element(RELATION);
				relation.setAttribute(NAME, "EQ");
				relation.setAttribute(ARITY, Integer.toString(2));
				relation.setAttribute(SEMANTICS, SOFT);
				relation.setAttribute(DEFAULT_COST, INFINITY);
				relation.setAttribute(NB_TUPLES, Integer.toString(nbSlots));
				value = values[0];
				sb = new StringBuilder("0" + COST_SEPARATOR + Integer.toString(value) + VALUE_SEPARATOR + Integer.toString(value));
				for(int i = 1 ; i < values.length ; i++) {
					value = values[i];
					sb.append(TUPLES_SEPARATOR);
					sb.append(Integer.toString(value));
					sb.append(VALUE_SEPARATOR);
					sb.append(Integer.toString(value));
				}
				relation.setText(sb.toString());
				relations.addContent(relation);
			}
		}
		relations.setAttribute(NB_RELATIONS, Integer.toString(relations.getContentSize()));
		return relations;
	}

	private void createPreferenceRelations(Element relations) {
		if (this.preferences != null) { // each agent expresses preferences over the time slots
			// Loop over the agents
			for(int a = 0 ; a < this.meetings.size() ; a++) {
				int agId = a + 1;
				Integer[] values = this.slots[a].toIntDomain();
				Element relation = new Element(RELATION);
				// String agentName = getAgentNameFor(agId);
				relation.setAttribute(NAME, getRelationNameFor(agId));
				relation.setAttribute(ARITY, Integer.toString(1));
				int nbSlots = values.length;
				relation.setAttribute(NB_TUPLES, Integer.toString(nbSlots));
				relation.setAttribute(SEMANTICS, SOFT);
				relation.setAttribute(DEFAULT_COST, INFINITY);

				ArrayList<Integer> pref = this.preferences.get(a);
				StringBuilder sb = new StringBuilder();
				int s = 0;
				while (s < nbSlots - 1) {
					sb.append(Integer.toString(pref.get(s)));
					sb.append(COST_SEPARATOR);
					sb.append(values[s]);
					sb.append(TUPLES_SEPARATOR);
					s++;
				}
				sb.append(Integer.toString(pref.get(s)));
				sb.append(COST_SEPARATOR);
				sb.append(values[s]);
				relation.setText(sb.toString());
				relations.addContent(relation);
			}
		}
		if (mode != Mode.PEAV && this.preferences == null) {
			// we need trivial unary constraints for agents involved in a single meeting
			Element relation = new Element(RELATION);
			relation.setAttribute(NAME, "trivial1");
			relation.setAttribute(ARITY, Integer.toString(1));
			relation.setAttribute(NB_TUPLES, Integer.toString(1));
			relation.setAttribute(SEMANTICS, SOFT);
			relation.setAttribute(DEFAULT_COST, Integer.toString(0));
			relation.addContent("0:1");
			relations.addContent(relation);
		}
	}

	/**
	 * 
	 * @param constraints
	 */
	private void createPreferenceConstraints(Element constraints) {
		if (this.preferences != null) { // soft preference constraints
			// Loop over the agents
			for(int a = 0 ; a < this.meetings.size() ; a++) {
				int agent = (a + 1);
				// Loop over the meetings this agent participates in
				for(int m : this.meetings.get(a)) {
					int meeting = m + 1;
					Element constraint = new Element(CONSTRAINT);
					String constraintName = getConstraintNameFor(agent + VARIABLE_IDENTIFIER_SEPARATOR + meeting);
					constraint.setAttribute(NAME, constraintName);
					constraint.setAttribute(MODEL, this.constraintModel);
					String varName = "";
					if (mode == Mode.PEAV) {
						varName = getMeetingNameFor(agent + VARIABLE_IDENTIFIER_SEPARATOR + meeting);
					} else {
						varName = getMeetingNameFor(meeting + VARIABLE_IDENTIFIER_SEPARATOR + Integer.toString(1));
					}
					constraint.setAttribute(SCOPE, varName);
					constraint.setAttribute(ARITY, Integer.toString(1));
					constraint.setAttribute(REFERENCE, getRelationNameFor(agent));
					constraints.addContent(constraint);
				}
			}
		}
	}

	/**
	 * 
	 * @param constraints
	 */
	private void createEqualityConstraints(Element constraints) {
		for(int i = 0 ; i < this.attendees.size() ; i++) {
			ArrayList<Integer> participants = this.attendees.get(i);
			int meeting = (i + 1);
			// Loop over the pairs of participants
			for(int j = 0 ; j < participants.size() - 1 ; j++) {
				int Aj = participants.get(j) + 1;
				String varJ = getMeetingNameFor(Aj + VARIABLE_IDENTIFIER_SEPARATOR + meeting);
				for(int k = j + 1 ; k < participants.size() ; k++) {
					int Ak = participants.get(k) + 1;
					String varK = getMeetingNameFor(Ak + VARIABLE_IDENTIFIER_SEPARATOR + meeting);
					Element constraint = new Element(CONSTRAINT);
					String constraintName = getConstraintNameFor(varJ + "_" + varK);
					constraint.setAttribute(NAME, constraintName);
					constraint.setAttribute(MODEL, this.constraintModel);
					constraint.setAttribute(ARITY, Integer.toString(2));
					constraint.setAttribute(SCOPE, varJ + VALUE_SEPARATOR + varK);
					constraint.setAttribute(REFERENCE, "EQ");
					if (intension) {
						Element parameters = new Element(PARAMETERS);
						parameters.setText(varJ + VALUE_SEPARATOR + varK);
						constraint.addContent(parameters);
					}
					constraints.addContent(constraint);
				}
			}
		}
	}

	/**
	 * 
	 * @param nbAgents
	 * @return
	 */
	private Element getConstraints(int nbAgents) {
		int nbr = 0;
		Element constraints = new Element(CONSTRAINTS);
		createPreferenceConstraints(constraints);
		if (mode == Mode.PEAV) {
			// Binary equality constraints: loop over the meetings
			// (i) all agents attending a meeting must agree on when it will occur
			createEqualityConstraints(constraints);
		}
		// allDifferent constraints: loop over the agents
		// (ii) an agent cannot attend two meetings at same time
		if (!intension) {
			// create several NE constraints
			for(int a = 0 ; a < nbAgents ; a++) {
				int agent = a + 1;
				ArrayList<Integer> myMeetings = this.meetings.get(a);
				int size = myMeetings.size();
				if (size > 1 || this.preferences == null) {
					for(int i = 0 ; i < size - 1 ; i++) {
						int mi = myMeetings.get(i) + 1;
						for(int j = i + 1 ; j < size ; j++) {
							int mj = myMeetings.get(j) + 1;
							Element constraint = new Element(CONSTRAINT);
							nbr++;
							String constraintName = getConstraintNameFor("_availability"+nbr);
							constraint.setAttribute(NAME, constraintName);
							constraint.setAttribute(MODEL, this.constraintModel);
							constraint.setAttribute(ARITY, Integer.toString(2));
							String scope = "";
							if (mode == Mode.PEAV) {
								scope += getMeetingNameFor(agent + VARIABLE_IDENTIFIER_SEPARATOR + mi + VALUE_SEPARATOR);
								scope += getMeetingNameFor(agent + VARIABLE_IDENTIFIER_SEPARATOR + mj);
							} else {
								scope += getMeetingNameFor(mi + VARIABLE_IDENTIFIER_SEPARATOR + Integer.toString(1) + VALUE_SEPARATOR);
								scope += getMeetingNameFor(mj + VARIABLE_IDENTIFIER_SEPARATOR + Integer.toString(1));
							}
							constraint.setAttribute(SCOPE, scope);
							constraint.setAttribute(REFERENCE, "NE");
							constraints.addContent(constraint);
						}
					}
				}
			}
		} else {
			for(int a = 0 ; a < nbAgents ; a++) {
				int agent = a + 1;
				ArrayList<Integer> myMeetings = this.meetings.get(a);
				if (myMeetings.size() > 1 || this.preferences == null) {
					// FIXME global:allDifferent to several NEQ constraints
					Element constraint = new Element(CONSTRAINT);
					String constraintName = getConstraintNameFor(agent + "_availability");
					constraint.setAttribute(NAME, constraintName);
					constraint.setAttribute(MODEL, this.constraintModel);
					constraint.setAttribute(ARITY, Integer.toString(myMeetings.size()));
					String scope = "";
					for(Integer m : myMeetings) {
						int meeting = m + 1;
						if (mode == Mode.PEAV) {
							scope += getMeetingNameFor(agent + VARIABLE_IDENTIFIER_SEPARATOR + meeting + VALUE_SEPARATOR);
						} else {
							scope += getMeetingNameFor(meeting + VARIABLE_IDENTIFIER_SEPARATOR + Integer.toString(1) + VALUE_SEPARATOR);
						}
					}
					constraint.setAttribute(SCOPE, scope.trim());
					constraint.setAttribute(REFERENCE, "global:allDifferent");
					Element parameters = new Element(PARAMETERS);
					parameters.setText("[ " + scope + "]");
					constraint.addContent(parameters);
					constraints.addContent(constraint);
				}
			}
		}
		constraints.setAttribute(NB_CONSTRAINTS, Integer.toString(constraints.getContentSize()));
		constraints.setAttribute(INITIAL_COST, Integer.toString(0));
		constraints.setAttribute(MAXIMAL_COST, INFINITY);
		return constraints;

	}

	private Element getGuiPresentation(int nbMeetings, int nbAgents, int nbAgentsPerMeeting, int numInstance) {
		Element presentation = new Element(GuiPresentation);
		presentation.setAttribute(NAME, INSTANCE + numInstance);
		presentation.setAttribute(TYPE, DCOP);
		presentation.setAttribute(BENCHMARK, this.typeOfBenchmark);
		presentation.setAttribute(NB_MEETINGS, Integer.toString(nbMeetings));
		presentation.setAttribute(NB_AGENTS, Integer.toString(nbAgents));
		presentation.setAttribute(NB_AGENTS_PER_MEETING, Integer.toString(nbAgentsPerMeeting));
		presentation.setAttribute(DISPONIBILITY, Integer.toString(this.disponibility));
		presentation.setAttribute(NB_DAYS, Integer.toString(this.nbDays));
		presentation.setAttribute(NB_HOURS, Integer.toString(this.nbHours));
		Element meetings = new Element(MEETINGS);
		presentation.addContent(meetings);
		for(int m = 0 ; m < this.attendees.size() ; m++) {
			ArrayList<Integer> participants = this.attendees.get(m);
			for(int i = 0 ; i < participants.size() ; i++) {
				participants.set(i, participants.get(i) + 1);
			}
			int meetingID = (m + 1);
			Element meeting = new Element(MEETING);
			meetings.addContent(meeting);
			meeting.setAttribute(NAME, getMeetingNameFor(meetingID));
			meeting.setText(participants.toString());
		}
		Element agents = new Element(AGENTS);
		presentation.addContent(agents);
		for(int a = 0 ; a < this.meetings.size() ; a++) {
			ArrayList<Integer> meettings = this.meetings.get(a);
			for(int i = 0 ; i < meettings.size() ; i++) {
				meettings.set(i, meettings.get(i) + 1);
			}
			int agentID = (a + 1);
			Element agent = new Element(AGENT);
			agents.addContent(agent);
			agent.setAttribute(NAME, getAgentNameFor(agentID));
			agent.setText(meettings.toString());
		}
		return presentation;
	}

	/**
	 * 
	 */
	public void generate() {
		if (this.mode != Mode.PEAV) {
			throw new UnsupportedOperationException("Not yet implemented for " + this.mode);
		}
		String path = this.directory.getPath();
		Element experiments = new Element(Commands.Experiments);
		Document confDocument = new Document(experiments);
		experiments.setAttribute(TYPE, DCOP);
		experiments.setAttribute(BENCHMARK, this.typeOfBenchmark);
		Element instances = new Element(INSTANCES);
		experiments.addContent(instances);
		instances.setAttribute(TYPE, DCOP);
		instances.setAttribute(BENCHMARK, this.typeOfBenchmark);
		instances.setAttribute(MEETINGS_DURATION, Integer.toString(this.meetingsDuration));
		instances.setAttribute(NB_DAYS, Integer.toString(this.nbDays));
		instances.setAttribute(NB_HOURS, Integer.toString(this.nbHours));
		instances.setAttribute(DISPONIBILITY, Integer.toString(this.disponibility));
		String name = Commands.MSP + "_";
		instances.setAttribute(NAME, name);
		for(int nbMeetings = this.minNbMeetings ; nbMeetings <= this.maxNbMeetings ; nbMeetings += this.stepNbMeetings) {
			for(int nbAgents = this.minNbAgents ; nbAgents <= this.maxNbAgents ; nbAgents += this.stepNbAgents) {
				String subDensity = getMeetingNameFor(nbMeetings) + "_" + getAgentNameFor(nbAgents);
				File folder = new File(path + File.separator + subDensity);
				if (!folder.exists()) {
					folder.mkdir();
				}
				Element dir = new Element(DIRECTORY);
				dir.setAttribute(NAME, subDensity);
				dir.setAttribute(NB_MEETINGS, Integer.toString(nbMeetings));
				dir.setAttribute(NB_AGENTS, Integer.toString(nbAgents));
				dir.setAttribute(MIN_AGENTS_PER_MEETING, Integer.toString(this.minNbAgentsPerMeeting));
				dir.setAttribute(MAX_AGENTS_PER_MEETING, Integer.toString(this.maxNbAgentsPerMeeting));
				dir.setAttribute(STEP_AGENTS_PER_MEETING, Integer.toString(this.stepNbAgentsPerMeeting));
				for(int nbAperM = this.minNbAgentsPerMeeting ; nbAperM <= this.maxNbAgentsPerMeeting ; nbAperM += this.stepNbAgentsPerMeeting) {
					File subFolder = new File(this.directory + File.separator + subDensity + File.separator + nbAperM);
					if (!subFolder.exists()) {
						subFolder.mkdir();
					}
					for(int numInstance = 1 ; numInstance <= this.nbInstances ; numInstance++) {
						Document document = new Document(this.generateInstance(numInstance, nbMeetings, nbAgents, nbAperM));
						this.save(subFolder.getPath(), name + numInstance + ".xml", document);
					}
				}
				dir.setAttribute(NB_INSTANCES, Integer.toString(this.nbInstances));
				instances.addContent(dir);
			}
		}
		String experimentsConfFileName = Commands.DIMES + "_" + nbInstances + ".xml";
		this.saveXML(path, experimentsConfFileName, confDocument);
	}
	
	
	@Override
	public void run() {
		ArrayList<String> keys = null;
		String filePath = this.experimentation.getParent();
		Document document = null;
		SAXBuilder sxb = new SAXBuilder();
		try {
			document = sxb.build(this.experimentation);
		} catch (Exception e) {}
		Element xp = document.getRootElement();
		Element instances = xp.getChild(INSTANCES);
		List<Element> communications = xp.getChild(Commands.Communications).getChildren();
		List<Element> algorithms = xp.getChild(Commands.Algorithms).getChildren();
		String[] protocols = new String[algorithms.size()];
		String[] orderings = new String[algorithms.size()];
		int a = 0;
		Iterator<Element> it = algorithms.iterator();
		while (it.hasNext()) {
			Element algorithm = it.next();
			protocols[a] = algorithm.getText();
			orderings[a] = algorithm.getAttributeValue(Commands.AO);
			a++;
		}
		int meetingDuration = new Integer(instances.getAttributeValue(MEETINGS_DURATION));
		int nbDays = new Integer(instances.getAttributeValue(NB_DAYS));
		int nbHours = new Integer(instances.getAttributeValue(NB_HOURS));
		int disponibility = new Integer(instances.getAttributeValue(DISPONIBILITY));
		String name = instances.getAttribute(NAME).getValue();
		it = communications.iterator();
		while (it.hasNext()) {
			// /////////////////////////////////////////////////////////////////
			Element comm = it.next();
			String commType = comm.getAttributeValue(TYPE);
			String channelType = comm.getAttributeValue(Commands.channelType);
			int minURC = new Integer(comm.getAttributeValue(Commands.MinURC));
			int maxURC = new Integer(comm.getAttributeValue(Commands.MaxURC));
			int nbRun = new Integer(comm.getAttributeValue(Commands.NbRun));
			long maxNCCCs = new Long(comm.getAttributeValue(Commands.MaxNCCCs));
			// /////////////////////////////////////////////////////////////////
			List<Element> lines = instances.getChildren();
			for(int l = 0 ; l < lines.size() ; l++) {
				Element line = lines.get(l);
				int nbMeetings = new Integer(line.getAttributeValue(NB_MEETINGS));
				int nbAgents = new Integer(line.getAttributeValue(NB_AGENTS));
				int minAgentsPerMeet = new Integer(line.getAttributeValue(MIN_AGENTS_PER_MEETING));
				int maxAgentsPerMeet = new Integer(line.getAttributeValue(MAX_AGENTS_PER_MEETING));
				int stepAgentsPerMeet = new Integer(line.getAttributeValue(STEP_AGENTS_PER_MEETING));
				int nbInstance = new Integer(line.getAttributeValue(NB_INSTANCES));
				String problemsFolder = filePath + File.separator + line.getAttributeValue(NAME);
				String resultsDirectory = filePath + File.separator + Commands.Results + "_" + nbInstance + "_" + nbRun;
				try {
					File f = new File(resultsDirectory);
					if (!f.exists()) {
						f.mkdir();
					}
				} catch (Exception e) {}
				resultsDirectory += File.separator + commType + "_" + channelType;
				try {
					File f = new File(resultsDirectory);
					if (!f.exists()) {
						f.mkdir();
					}
				} catch (Exception e) {}
				String subFolder = resultsDirectory + File.separator + line.getAttributeValue(NAME);
				try {
					File f = new File(subFolder);
					if (!f.exists()) {
						f.mkdir();
					}
				} catch (Exception e) {}
				for(int p = 0 ; p < protocols.length ; p++) { // Protocols
					Logger allResults = new Logger(subFolder, "All_" + protocols[p] + "-" + orderings[p] + ".csv");
					Logger results = new Logger(subFolder, protocols[p] + "-" + orderings[p] + ".csv");
					for(agentsPerMeet = minAgentsPerMeet ; agentsPerMeet <= maxAgentsPerMeet ; agentsPerMeet += stepAgentsPerMeet) {
						HashMap<String, Double> densityStatistics = new HashMap<String, Double>();
						for(instance = 1 ; instance <= nbInstance ; instance++) {
							String problem = problemsFolder + File.separator + agentsPerMeet + File.separator + name + instance + ".xml.bz2";
							HashMap<String, Double> instanceStatistics = new HashMap<String, Double>();
							for(run = 1 ; run <= nbRun ; run++) {
								DisProblem disProb = new DisProblem(problem, protocols[p]);
								DisSolverSimulator solver = new DisSolverSimulator(disProb);
								solver.setCutOffNCCCs(maxNCCCs);
								solver.addCommunicationMetric();
								solver.setMessageDelay(minURC, maxURC);
								solver.setCentralizedAO(AgentOrdering.getCentralizedAO(orderings[p]));
								solver.solve();
								if (keys == null) {
									keys = disProb.master.getStatisticsKeys();
								}
								HashMap<String, Double> runStatistics = solver.getStatistics();
								if (instance == 1 && run == 1 && agentsPerMeet == minAgentsPerMeet) {
									String header = Commands.MeetingScheduling + COMMA;
									header += Commands.Algorithm + COMMA + protocols[p] + COMMA + Commands.AO + COMMA + orderings[p] + "\n";
									header += NB_MEETINGS + COMMA + nbMeetings + COMMA + NB_AGENTS + COMMA + nbAgents + COMMA;
									header += MEETINGS_DURATION + COMMA + meetingDuration + COMMA + NB_DAYS + COMMA + nbDays + COMMA;
									header += NB_HOURS + COMMA + nbHours + COMMA + DISPONIBILITY + COMMA + disponibility + COMMA;
									header += NB_INSTANCES + COMMA + nbInstance + COMMA + Commands.NbRun + COMMA + nbRun + "\n";
									header += Commands.channelType + COMMA + channelType + COMMA + Commands.Communication + COMMA + commType + COMMA;
									header += Commands.MinURC + COMMA + minURC + COMMA + Commands.MaxURC + COMMA + maxURC + "\n\n";
									header += NB_AGENTS_PER_MEETING + COMMA;
									allResults.printToLogln(header + INSTANCE + COMMA + Commands.NbRun + COMMA + Tools.toString(keys));
									results.printToLogln(header + Tools.toString(keys));
								}
								allResults.printToLogln(agentsPerMeet + COMMA + instance + COMMA + run + COMMA + Tools.statisticsToString(keys, runStatistics));
								instanceStatistics = Tools.merge(runStatistics, instanceStatistics);
								disProb.setAgents(null);
								System.gc();
							}
							Tools.average(instanceStatistics, nbRun);
							densityStatistics = Tools.merge(instanceStatistics, densityStatistics);
						}
						Tools.average(densityStatistics, nbInstance);
						results.printToLogln(agentsPerMeet + COMMA + Tools.statisticsToString(keys, densityStatistics));
					}
					results.close();
					allResults.close();
					keys = null;
				} // Protocols
				ResultPrinter printer = new ResultPrinter(protocols, orderings, nbHeaderLine);
				printer.setFixKeys(fixKeys);
				ArrayList<String> paths = printer.printToGnuPlot(subFolder, "", commType);
				printer.plot(subFolder, paths);

				String header = Commands.MeetingScheduling + COMMA;
				header += NB_MEETINGS + COMMA + nbMeetings + COMMA + NB_AGENTS + COMMA + nbAgents + COMMA;
				header += MEETINGS_DURATION + COMMA + meetingDuration + COMMA + NB_DAYS + COMMA + nbDays + COMMA;
				header += NB_HOURS + COMMA + nbHours + COMMA + DISPONIBILITY + COMMA + disponibility + COMMA;
				header += NB_INSTANCES + COMMA + nbInstance + COMMA + Commands.NbRun + COMMA + nbRun + "\n";
				header += Commands.channelType + COMMA + channelType + COMMA + Commands.Communication + COMMA + commType + COMMA;
				header += Commands.MinURC + COMMA + minURC + COMMA + Commands.MaxURC + COMMA + maxURC + "";
				printer.printAllResults(subFolder, header);
			}
		}
	}
}
