/*
DisChoco: Distributed Constraint Reasoning Solver 
Copyright (C) 2007-2013 Redouane Ezzahir & Mohamed Wahbi 

DisChoco is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DisChoco is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

To contact the authors: <http://dischoco.sourceforge.net/>
*/ 
package benchmark.meetingscheduling;

import static simulation.parser.InstanceTokens.AGENT;
import static simulation.parser.InstanceTokens.AGENTS;
import static simulation.parser.InstanceTokens.ARITY;
import static simulation.parser.InstanceTokens.BENCHMARK;
import static simulation.parser.InstanceTokens.COMPLEX_MODEL;
import static simulation.parser.InstanceTokens.CONSTRAINT;
import static simulation.parser.InstanceTokens.CONSTRAINTS;
import static simulation.parser.InstanceTokens.DESCRIPTION;
import static simulation.parser.InstanceTokens.DOMAIN;
import static simulation.parser.InstanceTokens.DOMAINS;
import static simulation.parser.InstanceTokens.DisCSP;
import static simulation.parser.InstanceTokens.EXPRESSION;
import static simulation.parser.InstanceTokens.FUNCTIONAL;
import static simulation.parser.InstanceTokens.ID;
import static simulation.parser.InstanceTokens.INSTANCE;
import static simulation.parser.InstanceTokens.INSTANCES;
import static simulation.parser.InstanceTokens.MODEL;
import static simulation.parser.InstanceTokens.NAME;
import static simulation.parser.InstanceTokens.NB_AGENTS;
import static simulation.parser.InstanceTokens.NB_CONSTRAINTS;
import static simulation.parser.InstanceTokens.NB_DOMAINS;
import static simulation.parser.InstanceTokens.NB_INSTANCES;
import static simulation.parser.InstanceTokens.NB_PREDICATES;
import static simulation.parser.InstanceTokens.NB_VALUES;
import static simulation.parser.InstanceTokens.NB_VARIABLES;
import static simulation.parser.InstanceTokens.PARAMETERS;
import static simulation.parser.InstanceTokens.PREDICATE;
import static simulation.parser.InstanceTokens.PREDICATES;
import static simulation.parser.InstanceTokens.REFERENCE;
import static simulation.parser.InstanceTokens.SCOPE;
import static simulation.parser.InstanceTokens.TYPE;
import static simulation.parser.InstanceTokens.VARIABLE;
import static simulation.parser.InstanceTokens.VARIABLES;
import static simulation.parser.InstanceTokens.domainToString;
import static simulation.parser.InstanceTokens.getAgentNameFor;
import static simulation.parser.InstanceTokens.getConstraintNameFor;
import static simulation.parser.InstanceTokens.getDomainNameFor;
import static simulation.parser.InstanceTokens.getMeetingNameFor;
import static util.Commands.DIRECTORY;
import static util.Commands.DISPONIBILITY;
import static util.Commands.GuiPresentation;
import static util.Commands.MAX_GRID_SIZE;
import static util.Commands.MAX_MEETING_PER_AGENTS;
import static util.Commands.MEETING;
import static util.Commands.MEETINGS;
import static util.Commands.MEETINGS_DURATION;
import static util.Commands.MIN_GRID_SIZE;
import static util.Commands.MIN_MEETING_PER_AGENTS;
import static util.Commands.NB_DAYS;
import static util.Commands.NB_HOURS;
import static util.Commands.NB_MEETINGS;
import static util.Commands.NB_MEETINGS_PER_AGENT;
import static util.Commands.STEP_GRID_SIZE;
import static util.Commands.STEP_MEETING_PER_AGENTS;
import static util.Commands.SUBDIRECTORY;
import static util.Commands.X_axis;
import static util.Commands.Y_axis;
import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Random;
import kernel.order.AgentOrdering;
import org.jdom2.Document;
import org.jdom2.Element;
import org.jdom2.input.SAXBuilder;
import simulation.DisProblem;
import simulation.DisSolverSimulator;
import simulation.parser.PredicateTokens;
import util.Commands;
import util.IntArray;
import util.Logger;
import util.Tools;
import benchmark.Generator;
import benchmark.ResultPrinter;

/**
 * This class implements the Uniform Random Meeting Scheduling 
 * Problem Generator as described in <tt>[Wahbi et al., 2013]</tt> <br/>
 * <b>M. Wahbi, R. Ezzahir, C. Bessiere, E. H. Bouyakhf</b>. 
 * <i>Nogood-Based Asynchronous Forward-Checking Algorithms</i>. 
 * Constraints (2013) 
 * 
 * @author Mohamed Wahbi
 *
 */
public class DCSPMeetingScheduling extends Generator {

	private boolean DEBUG = false;
	public final static String GRIDSIZE = util.Commands.GridSize;
	public final static String fixKeys = GRIDSIZE;
	private static final String ARRIVAL_TIME = "ArrivalTime";

	public static int nbHeaderLine = 4;
	private static int meetPerAgent;
	private static int gridSize;
	private static int instance;
	private static int run;
	/** the number of instances to generate for each class */
	private int nbInstances = 25;
	/** the ranges of number of agents */
	private int minNbAgents, maxNbAgents, stepNbAgents;
	/** the ranges of number of meetings */
	private int minNbMeetings, maxNbMeetings, stepNbMeetings;
	/** the ranges of number of meeting per agent */
	private int minNbMeetingsPerAgent, maxNbMeetingsPerAgent, stepNbMeetingsPerAgent = 0;
	/** The geography that defines cities of meetings locations and distances */
	private Geography geography;
	/** for each meeting there are a set of agents that attend it (attendances) */
	private short[][] attendances; 
	/** to store the edges of the meeting graph */
	private short[][] meetingsAdjacences;
	/** for each agent there are a slots */
	private Slot[] slots;
	/**  for each meeting there is a place */
	private Place[] places;
	private short minParticipantsPerMeet = 2;
	private int meetingsDuration = 1;
	private File directory = new File(Commands.MSP);
	/**  numbers of days to put in the slots */
	private int nbDays = 5;
	/**  numbers of hours in the day to put in the slots */
	private int nbHours = 10;
	/**  percentage of disponibility over <i>numbers of hours</i> <code>x</code> <i>numbers of days</i> */
	private int disponibility = 100;
	private int minGridSize = 3; 
	private int maxGridSize = 21; 
	private int stepGridSize = 2;

	public DCSPMeetingScheduling(File directory, int nbInstances, int meetingsDuration) {
		super(Commands.RMSP, COMPLEX_MODEL, Commands.TKC);
		this.directory = directory;
		if (!this.directory.isDirectory()) {
			this.directory.mkdir();
		}
		this.nbInstances = nbInstances;
		this.meetingsDuration = meetingsDuration;
	}

	public DCSPMeetingScheduling(File directory, int nbInstances, int meetingsDuration, 
			int nbMeetings, int nbAgents, int nbMeetingsPerAgent, Geography geography,
			Slot[] slots) {
		this(directory, nbInstances, meetingsDuration);
		this.geography = geography;
		this.slots = slots;
		if (!this.directory.isDirectory()) {
			this.directory.mkdir();
		}
	}

	/**
	 * We encode the DMSP in DisCSP as follows. Each DisCSP agent represents a real 
	 * agent and contains k variables representing the k meetings to which the agent 
	 * participates. These k meetings are selected randomly among the m meetings. 
	 * The domain of each variable contains the d * h slots where a meeting can be 
	 * scheduled. A slot is one hour long, and there are h slots per day and d days. 
	 * There is an equality constraint for each pair of variables corresponding to 
	 * to the same agent. The arrival-time constraint between two variables m_i and m_j 
	 * agents attending a meeting must schedule it at the same slot (constraint (i)). 
	 * There is an arrival-time constraint between all variables/meetings belonging 
	 * to the same agent. The arrival-time constraint between two variables m_i and m_j 
	 * is <tt>abs(m_i-m_j)-duration > TravelingTime(m_i, m_j)</tt>, 
	 * where duration is the duration of every meeting. This arrival-time constraint 
	 * allows us to express both constraints (ii) and (iii). We place meetings randomly 
	 * on the nodes of a uniform grid of size g*g and the traveling time between two 
	 * adjacent nodes is 1 hour. Thus, the traveling time between two meetings equals 
	 * the Euclidean distance between nodes representing the locations where they will 
	 * be held. For varying the tightness of the arrival-time constraint we vary the 
	 * size of the grid on which meetings are placed.
	 * 
	 * @param directory : the directory where the generated instances will be saved
	 * @param nbInstances : the number of instances to generate
	 * @param meetingsDuration : the duration of meetings
	 * @param minNbMeetings : the minimal number of meetings
	 * @param maxNbMeetings : the maximal number of meetings
	 * @param stepNbMeetings : the step of number of meetings 
	 * @param minNbAgents : the minimal number of agents 
	 * @param maxNbAgents : the maximal number of agents 
	 * @param stepNbAgents : the step of number of agents
	 * @param minMeetingsPerAgent : the minimal number of meetings per agent 
	 * @param maxMeetingsPerAgent : the maximal number of meetings per agent 
	 * @param stepMeetingsPerAgent : the step of number of meetings per agent 
	 * @param minGridSize : the minimal grid size
	 * @param maxGridSize : the maximal grid size
	 * @param stepGridSize : the step of grid size
	 * @param nbDays : the number of days 
	 * @param nbHours : the number of hours per a day
	 * @param disponibility : the disponibility of agents over (nbDays * nbHours) slots
	 */
	public DCSPMeetingScheduling(File directory, int nbInstances, int meetingsDuration, 
			int minNbMeetings, int maxNbMeetings, int stepNbMeetings, 
			int minNbAgents, int maxNbAgents, int stepNbAgents, 
			int minMeetingsPerAgent, int maxMeetingsPerAgent, int stepMeetingsPerAgent, 
			int minGridSize, int maxGridSize, int stepGridSize, 
			int nbDays, int nbHours, int disponibility) {
		this(directory, nbInstances, meetingsDuration);
		this.setMeetingsProperties(minNbMeetings, maxNbMeetings, stepNbMeetings);
		this.setAgentsProperties(minNbAgents, maxNbAgents, stepNbAgents);
		this.setMeetingsPerAgentsProperties(minMeetingsPerAgent, maxMeetingsPerAgent, stepMeetingsPerAgent);
		this.setGeographyProperties(minGridSize, maxGridSize, stepGridSize);
		this.setCalendarProperties(nbDays, nbHours, disponibility);
	}
	
	/**
	 * Create the thread for running experiments 
	 * @param experimentation : the configuration file of experimentation
	 */
	public DCSPMeetingScheduling(File file) {
		super(file);
	}

	public void setMeetingsProperties(int minNbMeetings, int maxNbMeetings, int step) {
		this.minNbMeetings = minNbMeetings;
		this.maxNbMeetings = maxNbMeetings;
		this.stepNbMeetings = step;
	}

	public void setAgentsProperties(int minNbAgents, int maxNbAgents, int step) {
		this.minNbAgents = minNbAgents;
		this.maxNbAgents = maxNbAgents;
		this.stepNbAgents = step;
	}

	public void setMeetingsPerAgentsProperties(int min, int max, int step) {
		this.minNbMeetingsPerAgent = min;
		this.maxNbMeetingsPerAgent = max;
		this.stepNbMeetingsPerAgent = step;
	}

	public void setGeographyProperties(int minGridSize, int maxGridSize, int stepGridSize) {
		this.minGridSize = minGridSize;
		this.maxGridSize = maxGridSize;
		this.stepGridSize = stepGridSize;
	}

	public void setCalendarProperties(int nbDays, int nbHours, int disponibility) {
		this.nbDays = nbDays;
		this.nbHours = nbHours;
		this.disponibility = disponibility;
	}

	public void generateCalendares(int nbAgents) {
		this.slots = new Slot[nbAgents];
		for (int a = 0 ; a < nbAgents ; a++) {
			slots[a] = Slot.generateWeek(this.nbDays, this.nbHours, this.disponibility);
		}
	}

	private void initializer(int nbMeetings, int nbAgents) {
		this.attendances = new short[nbMeetings][nbAgents];
		for (int i = 0 ; i < this.attendances.length ; i++) {
			for (int j = 0 ; j < this.attendances[i].length ; j++) {
				this.attendances[i][j] = 0;
			}
		}
		this.meetingsAdjacences = new short[nbMeetings][nbMeetings];
		for (int i = 0 ; i < meetingsAdjacences.length ; i++) {
			for (int j = 0 ; j < meetingsAdjacences[i].length ; j++) {
				meetingsAdjacences[i][j] = 0;
			}
		}
	}

	/**
	 * This method selects for each meeting a place from the geography
	 */
	private void selectPlacesForMeetings(int nbMeetings) {
		this.places = new Place[nbMeetings];
		for (int p = 0 ; p < this.places.length ; p++) {
			this.places[p] = this.geography.getRandomPlace();
			if (this.DEBUG) {
				System.out.println("M" + (p + 1) + " will be hold at " + this.places[p]);
			}
		}
		if (this.DEBUG) {
			System.out.print("----------------------");
			for (int p1 = 0 ; p1 < this.places.length ; p1++) {
				System.out.print("M" + (p1 + 1) + " ");
			}
			System.out.println("");
			for (int p1 = 0 ; p1 < this.places.length ; p1++) {
				System.out.print("M" + (p1 + 1) + " ");
				for (int p2 = 0 ; p2 < this.places.length ; p2++) {
					System.out.print("" + this.getDistance(p1, p2) + "  ");
				}
				System.out.println("");
			}
		}
	}

	/**
	 * select randomly a clique of k meetings | not all of edges of the clique are already in the network.
	 */
	private boolean generateMeetings(int nbMeetings, int nbAgents, int nbMeetingsPerAgent) {
		this.initializer(nbMeetings, nbAgents);
		Random random = new Random();
		short[] nbParticipantsPerMeet = new short[nbMeetings];
		for (int a = 0 ; a < nbParticipantsPerMeet.length ; a++) {
			nbParticipantsPerMeet[a] = 0;
		}
		if (this.DEBUG) {
			System.out.println("generate meetings:");
		}
		ArrayList<int[]> combinations = getAllCombinations(nbMeetings, nbMeetingsPerAgent);
		int[] index = new int[combinations.size()];
		for (int i = 0 ; i < index.length ; i++) {
			index[i] = i;
		}
		for (int agId = 0 ; agId < nbAgents ; agId++) {
			int selected = random.nextInt(index.length - agId);
			index[selected] = index[index.length - agId - 1];
			int[] meetings = combinations.get(selected);
			for (int m_i = 0 ; m_i < meetings.length ; m_i++) {
				this.attendances[meetings[m_i]][agId] = 1;
				nbParticipantsPerMeet[meetings[m_i]]++;
				for (int m_j = m_i + 1 ; m_j < meetings.length ; m_j++) {
					this.meetingsAdjacences[meetings[m_i]][meetings[m_j]] = 1;
					this.meetingsAdjacences[meetings[m_j]][meetings[m_i]] = 1;
				}
			}
		}
		if(DEBUG) {
			System.err.println(toString(nbAgents));
		}
		for (int a = 0 ; a < nbParticipantsPerMeet.length ; a++) {
			if (nbParticipantsPerMeet[a] < this.minParticipantsPerMeet) {
				// TODO
				return false;
			}
		}
		if (this.DEBUG) {
			this.println(nbAgents);
		}
		return true;
	}

	protected void println(int nbAgents) {
		System.out.println(this.toString());
		for (int a = 0 ; a < nbAgents ; a++) {
			String domain = domainToString(this.slots[a].toIntDomain());
			System.out.println("A" + a + " : " + domain);
		}
	}

	public String toString(int nbAgents) {
		String ans = "------------------------------------------------------\n";
		ans += "Meeting\t | Location | Attending agents\n";
		for (int m = 0 ; m < this.attendances.length ; m++) {
			ans += "M" + (m + 1) + "  \t |  " + this.places[m] + "   | ";
			int[] agents = this.getParticipantsOn(m);
			for (int a = 0 ; a < agents.length ; a++) {
				ans += "A" + (agents[a] + 1);
				if (a < agents.length - 1) {
					ans += ", ";
				}
			}
			ans += "\n";
		}
		ans += "------------------------------------------------------\n";
		ans += "Agents   | Meetings\n";
		for (int a = 0 ; a < nbAgents ; a++) {
			ans += "A" + (a + 1) + "  \t |  ";
			int[] meetings = this.getMeetingsOf(a);
			for (int m = 0 ; m < meetings.length ; m++) {
				ans += "M" + (meetings[m] + 1);
				if (m < meetings.length - 1) {
					ans += ", ";
				}
			}
			ans += "\n";
		}
		ans += "------------------------------------------------------\n";
		return ans;
	}

	private Element generateInstance(int nbMeetings, int nbAgents, int nbMeetingsPerAgent) {
//		if(nbAgents*nbMeetingsPerAgent/2 < nbMeetings){
//			throw new Error("\n" +
//					"Impossible to generate this problem:\n" + "not enough agent!!\n" + 
//					NB_MEETINGS + " = "+ nbMeetings + "\n" + NB_AGENTS + " = " + nbAgents + "\n" + 
//					NB_MEETINGS_PER_AGENT + " = " + nbMeetingsPerAgent);
//		}
		if(Tools.C(nbAgents, minParticipantsPerMeet) < nbMeetings){
			throw new UnsupportedOperationException("\n" +
					"Impossible to generate this problem:\n" + "not enough agent!!\n" + 
					NB_MEETINGS + " = "+ nbMeetings + "\n" + NB_AGENTS + " = " + nbAgents + "\n" + 
					NB_MEETINGS_PER_AGENT + " = " + nbMeetingsPerAgent);
		}
		if(Tools.C(nbMeetings, nbMeetingsPerAgent) < nbAgents){
			throw new UnsupportedOperationException("\n" +
					"Impossible to generate this problem:\n" + "not enough agent!!\n" + 
					NB_MEETINGS + " = "+ nbMeetings + "\n" + NB_AGENTS + " = " + nbAgents + "\n" + 
					NB_MEETINGS_PER_AGENT + " = " + nbMeetingsPerAgent);
		}
		// generate for each agent a calendar;
		this.generateCalendares(nbAgents);
		// select for each meeting a place from the geography
		this.selectPlacesForMeetings(nbMeetings);
		// select randomly a clique of k meetings | not all of edges
		// of the clique are already in the network.
		int i = 0;
		boolean uniform = false;
		while (!uniform) {
			i++;
			uniform = this.generateMeetings(nbMeetings, nbAgents, nbMeetingsPerAgent);
			if (i > 1000000) { 
				System.err.println(toString(nbAgents));
				throw new UnsupportedOperationException("\n" +
						"You can not generate a uniform Meeting scheduling Problem with\n" + 
						NB_MEETINGS + " = "+ nbMeetings + "\n" + NB_AGENTS + " = " + nbAgents + "\n" + 
						NB_MEETINGS_PER_AGENT + " = " + nbMeetingsPerAgent);
			}
		}
		return this.getInstance(nbMeetings, nbAgents, nbMeetingsPerAgent);
	}

	// public void generateOne(){
	// if(nbAgents*this.nbMeetingsPerAgent<this.minParticipantsPerMeet*this.nbMeetings){
	// System.err.println("Impossible to generate this problem:\n"+
	// "not enough agent!!"); // enough
	// return;
	// }
	// //select for each meeting a place from the geography
	// this.selectPlacesForMeetings();
	// //select randomly a clique of k meetings | not all of edges
	// //of the clique are already in the network.
	// boolean uniform = false;
	// while(!uniform){
	// uniform = this.generateMeetings();
	// }
	// Element instance = this.getInstance();
	// Document document = new Document(instance);
	// this.saveInDirectory(document, Commands.MSP+""+".xml");
	// if(this.DEBUG){
	// this.println();
	// }
	// }

	public Element getInstance(int nbMeetings, int nbAgents, int nbMeetingsPerAgent) {
		Element instance = new Element(INSTANCE);
		instance.addContent(this.getPresentation(this.nbInstances, DisCSP));
		instance.addContent(this.getAgents(nbAgents));
		instance.addContent(this.getDomaines(nbAgents));
		instance.addContent(this.getVariables(nbMeetings, nbAgents, nbMeetingsPerAgent));
		instance.addContent(this.getPredicates());
		instance.addContent(this.getConstraints(nbAgents));
		instance.addContent(this.getGuiPresentation(nbMeetings, nbAgents, nbMeetingsPerAgent));
		return instance;
	}

	private Element getGuiPresentation(int nbMeetings, int nbAgents, int nbMeetingsPerAgent) {
		Element presentation = new Element(GuiPresentation);
		presentation.setAttribute(NAME, Commands.RMSP);
		presentation.setAttribute(TYPE, DisCSP);
		presentation.setAttribute(BENCHMARK, this.typeOfBenchmark);
		presentation.setAttribute(NAME, Commands.RMSP);
		presentation.setAttribute(NB_MEETINGS, Integer.toString(nbMeetings));
		presentation.setAttribute(NB_AGENTS, Integer.toString(nbAgents));
		presentation.setAttribute(NB_MEETINGS_PER_AGENT, Integer.toString(nbMeetingsPerAgent));
		presentation.setAttribute(DISPONIBILITY, Integer.toString(this.disponibility));
		presentation.setAttribute(NB_DAYS, Integer.toString(this.nbDays));
		presentation.setAttribute(NB_HOURS, Integer.toString(this.nbHours));
		presentation.setAttribute(X_axis, Integer.toString(this.geography.getRows()));
		presentation.setAttribute(Y_axis, Integer.toString(this.geography.getColumns()));
		Element meetings = new Element(MEETINGS);
		presentation.addContent(meetings);
		for(int m=0 ; m<nbMeetings ; m++){
			Element meeting = new Element(MEETING);
			meetings.addContent(meeting);
			meeting.setAttribute(NAME, getMeetingNameFor(m + 1));
			IntArray agents = new IntArray();
			for(int a = 0 ; a < nbAgents ; a++) {
				if (this.attendances[m][a] == 1){
					agents.add(a + 1);
				}
			}
			meeting.setText(agents.toString());
		}
		Element agents = new Element(AGENTS);
		presentation.addContent(agents);
		for(int a=0 ; a<nbAgents ; a++){
			Element agent = new Element(AGENT);
			agents.addContent(agent);
			agent.setAttribute(NAME, getAgentNameFor(a + 1));
			IntArray meets = new IntArray();
			for(int m=0 ; m<nbMeetings ; m++){
				if (this.attendances[m][a] == 1){
					meets.add(m + 1);
				}
			}
			agent.setText(meets.toString());
		}
		return presentation;
	}

	private Element getAgents(int nbAgents) {
		Element agents = new Element(AGENTS);
		agents.setAttribute(NB_AGENTS, Integer.toString(nbAgents));
		for (int id = 1 ; id <= nbAgents ; id++) {
			Element agent = new Element(AGENT);
			agent.setAttribute(NAME, getAgentNameFor(id));
			agent.setAttribute(ID, Integer.toString(id));
			agent.setAttribute(DESCRIPTION, AGENT + id);
			agents.addContent(agent);
		}
		return agents;
	}

	private Element getDomaines(int nbAgents) {
		Element domains = new Element(DOMAINS);
		domains.setAttribute(NB_DOMAINS, Integer.toString(nbAgents));
		for (int agId = 1 ; agId <= nbAgents ; agId++) {
			Element domain = new Element(DOMAIN);
			domain.setAttribute(NAME, getDomainNameFor(agId));
			Integer[] values = this.slots[agId - 1].toIntDomain();
			domain.setAttribute(NB_VALUES, Integer.toString(values.length));
			domain.setText(domainToString(values));
			domains.addContent(domain);
		}
		return domains;
	}

	private Element getVariables(int nbMeetings, int nbAgents, int nbMeetingsPerAgent) {
		Element variables = new Element(VARIABLES);
		variables.setAttribute(NB_VARIABLES, "" + nbAgents * nbMeetingsPerAgent);
		for (int a = 0 ; a < nbAgents ; a++) {
			int agId = a + 1;
			for (int m = 0 ; m < nbMeetings ; m++) {
				int meeting = m + 1;
				if (this.attendances[m][a] == 1) {
					Element variable = new Element(VARIABLE);
					variable.setAttribute(AGENT, getAgentNameFor(agId));
					variable.setAttribute(NAME, getMeetingNameFor(agId) + "." + meeting);
					// variable.setAttribute(ID, (a+1)+"."+(m+1));
					variable.setAttribute(ID, Integer.toString(meeting));
					variable.setAttribute(DOMAIN, getDomainNameFor(agId));
					variable.setAttribute(DESCRIPTION, "M_" + meeting);
					variables.addContent(variable);
				}
			}
		}
		return variables;
	}

	/**
	 * 
	 */
	private Element getPredicates() {
		Element predicates = new Element(PREDICATES);
		// ///////////////////////////////////////////////////////////////////////
		Element nePredicate = new Element(PREDICATE);
		nePredicate.setAttribute(NAME, ARRIVAL_TIME);
		Element neParameters = new Element(PARAMETERS);
		neParameters.setText("int X int Y int cst");
		nePredicate.addContent(neParameters);
		Element neExpression = new Element(EXPRESSION);
		Element neFunctional = new Element(FUNCTIONAL);
		neFunctional.setText(PredicateTokens.GE+"("+PredicateTokens.ABS+"("+PredicateTokens.SUB+"(X,Y)), cst) "); // it is not
		// necessary to say
		// Mi!=Mj
		neExpression.addContent(neFunctional);
		nePredicate.addContent(neExpression);
		predicates.addContent(nePredicate);
		// ///////////////////////////////////////////////////////////////////////
		Element eqPredicate = new Element(PREDICATE);
		eqPredicate.setAttribute(NAME, "EQ");
		Element eqParameters = new Element(PARAMETERS);
		eqParameters.setText("int X int Y ");
		eqPredicate.addContent(eqParameters);
		Element eqExpression = new Element(EXPRESSION);
		Element eqFunctional = new Element(FUNCTIONAL);
		eqFunctional.setText(PredicateTokens.EQ+"(X, Y)");
		eqExpression.addContent(eqFunctional);
		eqPredicate.addContent(eqExpression);
		predicates.addContent(eqPredicate);
		predicates.setAttribute(NB_PREDICATES, Integer.toString(predicates.getContentSize()));
		return predicates;
	}

	/**
	 * 
	 */
	private Element getConstraints(int nbAgents) {
		Element constraints = new Element(CONSTRAINTS);
		int nbConstraints = 0;
		// ///////////////////////////////////////////////////////////////////////////////////
		for (int m = 0 ; m < this.attendances.length ; m++) {
			int meeting = m +1;
			int[] agents = this.getParticipantsOn(m);
			for (int i = 0 ; i < agents.length - 1 ; i++) {
				for (int j = i + 1 ; j < agents.length ; j++) {
					int Ai = agents[i] + 1;
					int Aj = agents[j] + 1;
					Element constraint = new Element(CONSTRAINT);
					constraint.setAttribute(NAME, getConstraintNameFor(nbConstraints));
					constraint.setAttribute(MODEL, this.constraintsModel);
					constraint.setAttribute(ARITY, "2");
					// constraint.setAttribute(SCOPE_AGENTS,
					// "A"+(a_i+1)+" A"+(a_j+1));
					String scope = getMeetingNameFor(Ai) + "." + meeting + " " + getMeetingNameFor(Aj) + "." + meeting;
					constraint.setAttribute(SCOPE, scope);
					constraint.setAttribute(REFERENCE, "EQ");
					Element parameters = new Element(PARAMETERS);
					parameters.setText(scope + " ");
					constraint.addContent(parameters);
					constraints.addContent(constraint);
					nbConstraints++;
				}
			}
		}
		// ///////////////////////////////////////////////////////////////////////////////////
		for (int a = 0 ; a < nbAgents ; a++) {
			int[] meetings = this.getMeetingsOf(a);
			for (int i = 0 ; i < meetings.length - 1 ; i++) {
				for (int j = i + 1 ; j < meetings.length ; j++) {
					int m_i = meetings[i];
					int m_j = meetings[j];
					Element constraint = new Element(CONSTRAINT);
					constraint.setAttribute(NAME, "C" + (nbConstraints));
					constraint.setAttribute(MODEL, this.constraintsModel);
					constraint.setAttribute(ARITY, "2");
					// constraint.setAttribute(SCOPE_AGENTS,
					// "A"+(a+1)+" A"+(a+1));
					constraint.setAttribute(SCOPE, "M" + (a + 1) + "." + (m_i + 1) + " M" + (a + 1) + "." + (m_j + 1));
					constraint.setAttribute(REFERENCE, ARRIVAL_TIME);
					Element parameters = new Element(PARAMETERS);
					int dd = this.meetingsDuration + this.getDistance(m_i, m_j);
					parameters.setText("M" + (a + 1) + "." + (m_i + 1) + " M" + (a + 1) + "." + (m_j + 1) + " " + dd);
					constraint.addContent(parameters);
					constraints.addContent(constraint);
					nbConstraints++;
				}
			}
		}
		// ///////////////////////////////////////////////////////////////////////////////////
		constraints.setAttribute(NB_CONSTRAINTS, "" + nbConstraints);
		return constraints;
	}

	private int getDistance(int m_i, int m_j) {
		double distance = this.places[m_i].distance(this.places[m_j]);
		if (((double) distance % 1.0) != 0.0) { return (int) distance + 1; }
		return (int) distance;
	}

	private int[] getMeetingsOf(int a) {
		IntArray meetings = new IntArray();
		for (int m = 0 ; m < this.attendances.length ; m++) {
			if (this.attendances[m][a] == 1) {
				meetings.add(m);
			}
		}
		return meetings.getArray();
	}

	private int[] getParticipantsOn(int m) {
		IntArray agents = new IntArray();
		for (int a = 0 ; a < this.attendances[m].length ; a++) {
			if (this.attendances[m][a] == 1) {
				agents.add(a);
			}
		}
		return agents.getArray();
	}

	/**
	 * 
	 */
	public void generate() {
		String path = this.directory.getPath();
		Element experiments = new Element(Commands.Experiments);
		Document confDocument = new Document(experiments);
		experiments.setAttribute(TYPE, DisCSP);
		experiments.setAttribute(BENCHMARK, this.typeOfBenchmark);
		Element instances = new Element(INSTANCES);
		experiments.addContent(instances);
		instances.setAttribute(TYPE, DisCSP);
		instances.setAttribute(BENCHMARK, this.typeOfBenchmark);
		instances.setAttribute(MEETINGS_DURATION, "" + this.meetingsDuration);
		instances.setAttribute(NB_DAYS, "" + this.nbDays);
		instances.setAttribute(NB_HOURS, "" + this.nbHours);
		instances.setAttribute(DISPONIBILITY, "" + this.disponibility);
		String name = Commands.MSP + "_";
		instances.setAttribute(NAME, name);
		for (int nbMeetings = this.minNbMeetings ; nbMeetings <= this.maxNbMeetings ; nbMeetings += this.stepNbMeetings) {
			for (int nbAgents = this.minNbAgents ; nbAgents <= this.maxNbAgents ; nbAgents += this.stepNbAgents) {
				String subDensity = "M-" + nbMeetings + "_A-" + nbAgents;
				File folder = new File(path + File.separator + subDensity);
				if (!folder.exists()) { folder.mkdir(); }
				Element dir = new Element(DIRECTORY);
				dir.setAttribute(NAME, subDensity);
				dir.setAttribute(NB_MEETINGS, "" + nbMeetings);
				dir.setAttribute(NB_AGENTS, "" + nbAgents);
				dir.setAttribute(MIN_MEETING_PER_AGENTS, "" + this.minNbMeetingsPerAgent);
				dir.setAttribute(MAX_MEETING_PER_AGENTS, "" + this.maxNbMeetingsPerAgent);
				dir.setAttribute(STEP_MEETING_PER_AGENTS, "" + this.stepNbMeetingsPerAgent);
				Element subDir = new Element(SUBDIRECTORY);
				subDir.setAttribute(MIN_GRID_SIZE, "" + this.minGridSize);
				subDir.setAttribute(MAX_GRID_SIZE, "" + this.maxGridSize);
				subDir.setAttribute(STEP_GRID_SIZE, "" + this.stepGridSize);
				subDir.setAttribute(NB_INSTANCES, "" + this.nbInstances);
				dir.addContent(subDir);
				for (int nbMperA = this.minNbMeetingsPerAgent ; nbMperA <= this.maxNbMeetingsPerAgent ; nbMperA += this.stepNbMeetingsPerAgent) {
					File subFolder = new File(folder.getAbsolutePath() + File.separator + nbMperA);
					if (!subFolder.exists()) {
						subFolder.mkdir();
					}
					for (int gridSize = this.minGridSize ; gridSize <= this.maxGridSize ; gridSize += this.stepGridSize) {
						this.geography = new Geography(gridSize, gridSize);
						File subsubFolder = new File(subFolder.getAbsolutePath() + File.separator + gridSize);
						if (!subsubFolder.exists()) {
							subsubFolder.mkdir();
						}
						dir.setAttribute(MAX_MEETING_PER_AGENTS, "" + this.maxNbMeetingsPerAgent);
						for (int i = 0 ; i < this.nbInstances ; i++) {
							Document document = new Document(this.generateInstance(nbMeetings, nbAgents, nbMperA));
							this.save(subsubFolder.getPath(), name + (i + 1) + ".xml", document);
						}
					}
				}
				instances.addContent(dir);
			}
		}
		String experimentsConfFileName = Commands.RMSP + "_" + nbInstances + ".xml";
		this.saveXML(path, experimentsConfFileName, confDocument);
	}
	
	@Override
	public void run() {
		ArrayList<String> keys = null;
		String filePath = this.experimentation.getParent();
		// //////////////////////////////////////////////////////////////////////////////////////////////////////////
		Document document = null;
		SAXBuilder sxb = new SAXBuilder();
		try {
			document = sxb.build(this.experimentation);
		} catch (Exception e) {}
		Element xp = document.getRootElement();
		Element instances = xp.getChild(INSTANCES);
		List<Element> communications = xp.getChild(Commands.Communications).getChildren();
		List<Element> algorithms = xp.getChild(Commands.Algorithms).getChildren();
		String[] protocols = new String[algorithms.size()];
		String[] orderings = new String[algorithms.size()];
		int a = 0;
		Iterator<Element> it = algorithms.iterator();
		while (it.hasNext()) {
			Element algorithm = it.next();
			protocols[a] = algorithm.getText();
			orderings[a] = algorithm.getAttributeValue(Commands.AO);
			a++;
		}
		int meetingDuration = new Integer(instances.getAttributeValue(MEETINGS_DURATION));
		int nbDays = new Integer(instances.getAttributeValue(NB_DAYS));
		int nbHours = new Integer(instances.getAttributeValue(NB_HOURS));
		int disponibility = new Integer(instances.getAttributeValue(DISPONIBILITY));
//		int x_axis = new Integer(instances.getAttributeValue(X_axis));
//		int y_axis = new Integer(instances.getAttributeValue(Y_axis));
		String name = instances.getAttribute(NAME).getValue();
		it = communications.iterator();
		while (it.hasNext()) {
			// /////////////////////////////////////////////////////////////////
			Element comm = it.next();
			String commType = comm.getAttributeValue(TYPE);
			String channelType = comm.getAttributeValue(Commands.channelType);
			int minURC = new Integer(comm.getAttributeValue(Commands.MinURC));
			int maxURC = new Integer(comm.getAttributeValue(Commands.MaxURC));
			int nbRun = new Integer(comm.getAttributeValue(Commands.NbRun));
			long maxNCCCs = new Long(comm.getAttributeValue(Commands.MaxNCCCs));
			// /////////////////////////////////////////////////////////////////
			List<Element> lines = instances.getChildren();
			for (Element line : lines) {
				int nbMeetings = new Integer(line.getAttributeValue(NB_MEETINGS));
				int nbAgents = new Integer(line.getAttributeValue(NB_AGENTS));
				int minMeetPerAgent = new Integer(line.getAttributeValue(MIN_MEETING_PER_AGENTS));
				int maxMeetPerAgent = new Integer(line.getAttributeValue(MAX_MEETING_PER_AGENTS));
				int stepMeetPerAgent = new Integer(line.getAttributeValue(STEP_MEETING_PER_AGENTS));
				Element subline = line.getChild(SUBDIRECTORY);
				int minGridSize = new Integer(subline.getAttributeValue(MIN_GRID_SIZE));
				int maxGridSize = new Integer(subline.getAttributeValue(MAX_GRID_SIZE));
				int stepGridSize = new Integer(subline.getAttributeValue(STEP_GRID_SIZE));
				int nbInstance = new Integer(subline.getAttributeValue(NB_INSTANCES));
				String problemsFolder = filePath + File.separator + line.getAttributeValue(NAME);
				String resultsDirectory = filePath + File.separator + Commands.Results + "_" + nbInstance+"_"+nbRun;
				try {
					File f = new File(resultsDirectory);
					if (!f.exists()) { f.mkdir(); }
				} catch (Exception e) {}
				resultsDirectory += File.separator + commType + "_" + channelType;
				try {
					File f = new File(resultsDirectory);
					if (!f.exists()) { f.mkdir(); }
				} catch (Exception e) {}
				String subFolder = resultsDirectory + File.separator + line.getAttributeValue(NAME);
				try {
					File f = new File(subFolder);
					if (!f.exists()) { f.mkdir(); }
				} catch (Exception e) {}
				for (meetPerAgent = minMeetPerAgent ; meetPerAgent <= maxMeetPerAgent ; meetPerAgent += stepMeetPerAgent) {
					String subsubFolder = subFolder + File.separator + meetPerAgent;
					try {
						File f = new File(subsubFolder);
						if (!f.exists()) { f.mkdir(); }
					} catch (Exception e) {}
					for (int p = 0 ; p < protocols.length ; p++) { // Protocols
						Logger allResults = new Logger(subsubFolder, "All_" + protocols[p] + "-" + orderings[p] + ".csv");
						Logger results = new Logger(subsubFolder, protocols[p] + "-" + orderings[p] + ".csv");
						for (gridSize = minGridSize ; gridSize <= maxGridSize ; gridSize += stepGridSize) {
							HashMap<String, Double> densityStatistics = new HashMap<String, Double>();
							for (instance = 1 ; instance <= nbInstance ; instance++) {
								String problem = problemsFolder + File.separator + meetPerAgent + File.separator + gridSize + File.separator + name + instance + ".xml.bz2";
								HashMap<String, Double> instanceStatistics = new HashMap<String, Double>();
								for (run = 1 ; run <= nbRun ; run++) {
									DisProblem disProb = new DisProblem(problem, protocols[p]);
									DisSolverSimulator solver = new DisSolverSimulator(disProb);
									solver.setCutOffNCCCs(maxNCCCs);
									solver.addCommunicationMetric();
									solver.setMessageDelay(minURC, maxURC);
									solver.setCentralizedAO(AgentOrdering.getCentralizedAO(orderings[p]));
									solver.solve();
									if(keys == null ){
										keys = disProb.master.getStatisticsKeys();
									}
									HashMap<String, Double> runStatistics = solver.getStatistics();
									if (instance == 1 && run == 1 && gridSize == minGridSize) {
										String header = Commands.RMSP + COMMA;
										header += Commands.Algorithm + COMMA + protocols[p] + COMMA + Commands.AO + COMMA + orderings[p] + "\n";
										header += NB_MEETINGS + COMMA + nbMeetings + COMMA + NB_AGENTS + COMMA + nbAgents + COMMA ;
										header += MEETINGS_DURATION + COMMA + meetingDuration + COMMA + NB_DAYS + COMMA + nbDays + COMMA ;
										header += NB_HOURS + COMMA + nbHours + COMMA + DISPONIBILITY + COMMA + disponibility + COMMA ;
//										header += X_axis + COMMA + gridSize + COMMA + Y_axis + COMMA + gridSize + COMMA;
										header += NB_INSTANCES+ COMMA +nbInstance + COMMA + Commands.NbRun+ COMMA +nbRun+"\n";
										header += Commands.channelType + COMMA + channelType + COMMA + Commands.Communication + COMMA + commType + COMMA ;
										header += Commands.MinURC + COMMA + minURC + COMMA + Commands.MaxURC + COMMA + maxURC + COMMA;
										header += NB_MEETINGS_PER_AGENT + COMMA + meetPerAgent+ "\n\n";
										header += GRIDSIZE + COMMA ;
										allResults.printToLogln(header + INSTANCE + COMMA + Commands.NbRun + COMMA + Tools.toString(keys));
										results.printToLogln(header + Tools.toString(keys));
									}
									allResults.printToLogln(gridSize + COMMA + instance + COMMA + run + COMMA + Tools.statisticsToString(keys, runStatistics));
									instanceStatistics = Tools.merge(runStatistics, instanceStatistics);
									disProb.setAgents(null);
									System.gc();
								}
								Tools.average(instanceStatistics, nbRun);
								densityStatistics = Tools.merge(instanceStatistics, densityStatistics);
							}
							Tools.average(densityStatistics, nbInstance);
							results.printToLogln(gridSize + COMMA + Tools.statisticsToString(keys, densityStatistics));
						}
						results.close();
						allResults.close();
						keys = null;
					} //Protocols
					ResultPrinter printer = new ResultPrinter(protocols, orderings, nbHeaderLine);
					printer.setFixKeys(fixKeys);
					ArrayList<String> paths = printer.printToGnuPlot(subsubFolder, "", commType);
					printer.plot(subsubFolder, paths);

					String header = Commands.RMSP + COMMA;
					header += NB_MEETINGS + COMMA + nbMeetings + COMMA + NB_AGENTS + COMMA + nbAgents + COMMA ;
					header += MEETINGS_DURATION + COMMA + meetingDuration + COMMA + NB_DAYS + COMMA + nbDays + COMMA ;
					header += NB_HOURS + COMMA + nbHours + COMMA + DISPONIBILITY + COMMA + disponibility + COMMA ;
//					header += X_axis + COMMA + gridSize + COMMA + Y_axis + COMMA + gridSize + COMMA;
					header += NB_INSTANCES+ COMMA +nbInstance + COMMA + Commands.NbRun+ COMMA +nbRun+"\n";
					header += Commands.channelType + COMMA + channelType + COMMA + Commands.Communication + COMMA + commType + COMMA ;
					header += Commands.MinURC + COMMA + minURC + COMMA + Commands.MaxURC + COMMA ;
					header +=  maxURC + NB_MEETINGS_PER_AGENT + COMMA + meetPerAgent + "";
					printer.printAllResults(subsubFolder, header);
				} 
			}
		}
	}
	
	public static boolean nextCombination(int[] combination, int n) {
		int i = combination.length - 1;
		++combination[i];
		while ((i > 0) && (combination[i] >= n - combination.length + 1 + i)) {
			--i;
			++combination[i];
		}
		/* Combination (n-k, n-k+1, ..., n) reached */ 
		if (combination[0] > n - combination.length)
			return false; /* No more combinations can be generated */
		/*
		 * comb now looks like (..., x, n, n, n, ..., n). Turn it into (..., x, x + 1, x + 2, ...)
		 */
		for (i = i + 1 ; i < combination.length ; ++i)
			combination[i] = combination[i - 1] + 1;
		return true;
	}
	
	public static ArrayList<int[]> getAllCombinations(int nbMeetings, int nbMeetingsPerAgent) {
		ArrayList<int[]> list = new ArrayList<int[]>();
		int combination[] = new int[nbMeetingsPerAgent];
		for (int i = 0 ; i < nbMeetingsPerAgent ; ++i) {
			combination[i] = i;
		}
		do {
			list.add(combination.clone());
		} while (nextCombination(combination, nbMeetings));
		return list;
	}

}
