/*
DisChoco: Distributed Constraint Reasoning Solver 
Copyright (C) 2007-2013 Redouane Ezzahir & Mohamed Wahbi 

DisChoco is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DisChoco is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

To contact the authors: <http://dischoco.sourceforge.net/>
*/ 
package benchmark.meetingscheduling;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.Random;

/**
 * 
 * @authors 
 * <a href="mailto:ezzahir@ensa-agadir.ac.ma?subject=DisChoco &cc=wahbi@lirmm.fr &cc=bessiere@lirmm.fr">
 * Redoaune Ezzahir</a> and 
 * <a href="mailto:wahbi@lirmm.fr?subject=DisChoco &cc=ezzahir@ensa-agadir.ac.ma &cc=bessiere@lirmm.fr">
 * Mohamed Wahbi</a>
 *
 */
public class Slot {

	private BitSet[] slots;
	private int nbDays = 5; // 5 days of trhe week
	private int nbHours = 10; // 10 hours per a day from 8h to 18h
	private int startHour = 8;
	private int endHour = 18;
	private Random rand = null;

	public Slot() {
		this.rand = new Random();
	}
	
	public Slot(int nbDays, int startHour, int endHour) {
		this();
		this.nbDays = nbDays;
		this.startHour = startHour;
		this.endHour = endHour;
		this.nbHours = endHour - startHour;
		this.initialize();
	}

	public Slot(int nbDays, int nbHours) {
		this();
		this.nbDays = nbDays;
		this.nbHours = nbHours;
		this.endHour = this.startHour + nbHours;
		this.initialize();
	}

	private void initialize() {
		this.slots = new BitSet[this.nbDays];
		for (int d = 0 ; d < this.slots.length ; d++) {
			this.slots[d] = new BitSet(24);
			for (int h = this.startHour ; h < this.endHour ; h++) {
				this.slots[d].set(h);
			}
			this.slots[d].clear(12);
			this.slots[d].clear(13);
		}
	}

	public ArrayList<Time> getTimes() {
		ArrayList<Time> times = new ArrayList<Time>();
		for (int d = 0 ; d < this.slots.length ; d++) {
			for (int h = this.slots[d].nextSetBit(0) ; h >= 0 ; h = this.slots[d].nextSetBit(h + 1)) {
				times.add(new Time(d + 1, h));
			}
		}
		return times;
	}

	public int getNbtHour() {
		return this.nbHours;
	}

	public int getNbDays() {
		return this.nbDays;
	}

	public String toString() {
		String slots = "";
		for (int d = 0 ; d < this.slots.length ; d++) {
			slots += Time.getDay(d) + "[";
			for (int h = this.slots[d].nextSetBit(0) ; h >= 0 ;) {
				slots += h;
				h = this.slots[d].nextSetBit(h + 1);
				if (h >= 0) {
					slots += ":";
				}
			}
			slots += "]\n";
		}
		return slots;
	}

	private void setSlots(BitSet[] slots) {
		this.slots = slots;
	}

	public static Slot fromString(String str) {
		String[] days = str.split("\n");
		int nbDays = days.length;
		BitSet[] slots = new BitSet[nbDays];
		int startHours = 24;
		int endHours = 0;
		for (int d = 0 ; d < days.length ; d++) {
			slots[d] = new BitSet(24);
			String day = days[d].substring(days[d].indexOf("[") + 1, days[d].lastIndexOf("]"));
			String[] hours = day.split(":");
			for (int h = 0 ; h < hours.length ; h++) {
				int hour = new Integer(hours[h]);
				slots[d].set(hour);
				startHours = Math.min(startHours, hour);
				endHours = Math.max(endHours, hour + 1);
			}
		}
		Slot slot = new Slot(nbDays, startHours, endHours);
		slot.setSlots(slots);
		return slot;
	}

	public String pretty() {
		String slots = "          | ";
		for (int h = this.startHour ; h < this.endHour ; h++) {
			String hour = (h < 10)? "0" + h : "" + h;
			slots += hour + " | ";
		}
		slots += "\n";
		// slots += this.endHour-1+"  |";
		for (int d = 0 ; d < this.slots.length ; d++) {
			slots += Time.getDay(d) + " |";
			for (int h = this.startHour ; h < this.endHour ; h++) {
				if (this.slots[d].get(h)) {
					slots += "    |";
				} else {
					slots += "****|";
				}
			}
			slots += "\n";
		}
		return slots;
	}

	public static Slot generateWeek() {
		return new Slot(5, 8, 18);
	}

	public static Slot generateWeek(int disponibility) {
		Slot slot = new Slot(5, 8, 18);
		slot.remove(100 - disponibility);
		return slot;
	}

	public static Slot generateWeek(int nbDays, int nbHours, int disponibility) {
		Slot slot = new Slot(nbDays, nbHours);
		slot.remove(100 - disponibility);
		return slot;
	}

	public Integer[] toIntDomain() {
		Integer[] domain = new Integer[this.getNbSlot()];
		for (int d = 0 , i = 0 ; d < this.nbDays ; d++) {
			for (int h = 0 ; h < 24 ; h++) {
				if (this.slots[d].get(h)) {
					domain[i] = d * 24 + h;
					i++;
				}
			}
		}
		return domain;
	}

	public void remove(int density) {
		int[] index = new int[this.getNbSlot()];
		int i = 0;
		for (int d = 0 ; d < this.nbDays ; d++) {
			for (int h = 0 ; h < 24 ; h++) {
				if (this.slots[d].get(h)) {
					index[i] = d * 24 + h;
					i++;
				}
			}
		}
		int nbSlotToBeRemoved = (density * index.length) / 100;
		for (i = 0 ; i < nbSlotToBeRemoved ; i++) {
			int selected = this.rand.nextInt(index.length - i);
			int remove = index[selected];
			int d = remove / 24;
			int h = remove % 24;
			index[selected] = index[index.length - i - 1];
			this.slots[d].clear(h);
		}
	}

	public int getNbSlot() {
		int nbSlots = 0;
		for (int d = 0 ; d < this.slots.length ; d++) {
			nbSlots += this.slots[d].cardinality();
		}
		return nbSlots;
	}
}
