/*
 * DisChoco: Distributed Constraint Reasoning Solver
 * Copyright (C) 2007-2013 Redouane Ezzahir & Mohamed Wahbi
 * DisChoco is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * DisChoco is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 * To contact the authors: <http://dischoco.sourceforge.net/>
 */
package benchmark.random;

import static simulation.parser.InstanceTokens.BENCHMARK;
import static simulation.parser.InstanceTokens.DCOP;
import static simulation.parser.InstanceTokens.INFINITY;
import static simulation.parser.InstanceTokens.INITIAL_COST;
import static simulation.parser.InstanceTokens.INSTANCE;
import static simulation.parser.InstanceTokens.INSTANCES;
import static simulation.parser.InstanceTokens.MAXIMAL_COST;
import static simulation.parser.InstanceTokens.MINIMAL_COST;
import static simulation.parser.InstanceTokens.MODEL;
import static simulation.parser.InstanceTokens.NAME;
import static simulation.parser.InstanceTokens.NB_AGENTS;
import static simulation.parser.InstanceTokens.NB_CONSTRAINTS;
import static simulation.parser.InstanceTokens.NB_INSTANCES;
import static simulation.parser.InstanceTokens.NB_TUPLES;
import static simulation.parser.InstanceTokens.SIMPLE_MODEL;
import static simulation.parser.InstanceTokens.TKC;
import static simulation.parser.InstanceTokens.TYPE;
import static util.Commands.AO;
import static util.Commands.Algorithm;
import static util.Commands.Algorithms;
import static util.Commands.Communication;
import static util.Commands.Communications;
import static util.Commands.DENSITY;
import static util.Commands.DomainSize;
import static util.Commands.Experiments;
import static util.Commands.FROM;
import static util.Commands.GuiPresentation;
import static util.Commands.MaxNCCCs;
import static util.Commands.MaxURC;
import static util.Commands.MinURC;
import static util.Commands.NbRun;
import static util.Commands.P_1;
import static util.Commands.RandomDCOP;
import static util.Commands.Results;
import static util.Commands.STEP;
import static util.Commands.TIGHTNESS;
import static util.Commands.TO;
import static util.Commands.channelType;
import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import kernel.order.AgentOrdering;
import org.jdom2.Document;
import org.jdom2.Element;
import org.jdom2.input.SAXBuilder;
import simulation.DisProblem;
import simulation.DisSolverSimulator;
import simulation.parser.InstanceTokens;
import util.Commands;
import util.Logger;
import util.Tools;
import util.graph.UndirectedGraph;
import benchmark.Generator;
import benchmark.ResultPrinter;

/**
 * Uniform Random Binary DisCSP instance generator. 
 * DisCSP are characterized by the 4-tuple <n, d, p1, p2>, where: 
 * <br/>
 * <p>
 * <i>n</i> is the number of agents/variables, each agent owns exactly one variable, 
 * <br/>
 * <i>d</i> is the domain size, i.e., the number of values in the domain of each variable, 
 * <br/>
 * <i>p1</i> the network connectivity defined as the ratio of existing binary constraints,
 * <br/>
 * <i>p2</i> is the constraint tightness defined as the ratio of forbidden value pairs. 
 * </p>
 * @author Mohamed Wahbi
 */
public class DCOPGenerator extends Generator {

	public static final String fixKeys = P_1;

	/** The number of agent/variable in the problem */
	private int nbAgents;
	/** The domain size of each agent */
	private int domainSize;
	/** The graph density (i.e., the ratio of number of constraint to n*(n-1))*/
	private int minP1 = 20, maxP1 = 80, stepP1 = 10;
	private int minCost = 1, maxCost = 100;
	private int nbInstances = 25;
	private String path = ".";

	public static int nbHeaderLine = 4;

	/**
	 * Uniform Random Binary DCOP instance generator. 
	 * DisCSP are characterized by the 4-tuple <n, d, p1, p2>, where: 
	 * <br/>
	 * <i>n</i> is the number of agents/variables, each agent owns exactly one variable, 
	 * <br/>
	 * <i>d</i> is the domain size, i.e., the number of values in the domain of each variable, 
	 * <br/>
	 * <i>p1</i> the network connectivity defined as the ratio of existing binary constraints,
	 * <br/>
	 * <i>p2</i> is the constraint tightness defined as the ratio of forbidden value pairs. 
	 * <br/>
	 * @param path : the path of the directory where instances should be stored 
	 * @param constraintsModel : the model of constraints, i.e., <tt>TKC</tt> or <tt>PKC</tt>.
	 * @param nbInstances : the number of instances to generate 
	 * @param n : the number of agents/variables, each agent owns exactly one variable
	 * @param d : the domain size, i.e., the number of values in the domain of each variable 
	 * @param minP1 : the minimum network connectivity
	 * @param maxP1 : the maximum network connectivity
	 * @param stepP1 : the step of the network connectivity
	 * @param minP2 : the minimum constraint tightness
	 * @param maxP2 : the maximum constraint tightness
	 * @param stepP2 : the step of the constraint tightness
	 * @param minCost : the minimum constraint cost
	 * @param maxCost : the maximum constraint cost
	 */
	public DCOPGenerator(File path, String constraintsModel, 
			Integer nbInstances, Integer nbAgents, Integer domainSize, 
			Integer minP1, Integer maxP1, Integer stepP1,
			Integer minCost, Integer maxCost) {
		super(RandomDCOP, SIMPLE_MODEL, constraintsModel);
		this.path = path.getPath();
		this.nbInstances = nbInstances;
		this.nbAgents = nbAgents;
		this.domainSize = domainSize;
		this.minP1 = minP1;
		this.maxP1 = maxP1;
		this.stepP1 = stepP1;
		this.minCost = minCost;
		this.maxCost = maxCost;
	}
	
	/**
	 * Create the thread for running experiments 
	 * @param experimentation : the configuration file of experimentation
	 */
	public DCOPGenerator(File file) {
		super(file);
	}

	/**
	 * This method return the description of the instance. This will be used in the 
	 * Graphical User Interface. 
	 * 
	 * @param nbOfInstance : the number of instance
	 * @param p1 : the network connectivity defined as the ratio of existing binary constraints,
	 * @param p2 : is the constraint tightness defined as the ratio of forbidden value pairs. 
	 * @param nbConstraint : the number of constraint in the instance
	 * @param nbTuples : the number of tuples 
	 * @return
	 */
	private Element getGuiPresentation(int nbOfInstance, int p1, int nbConstraint, int nbTuples) {
		Element presentation = new Element(GuiPresentation);
		presentation.setAttribute(TYPE, DCOP);
		presentation.setAttribute(BENCHMARK, this.typeOfBenchmark);
		presentation.setAttribute(NAME, INSTANCE + (nbOfInstance));
		presentation.setAttribute(MODEL, this.constraintsModel);
		presentation.setAttribute(NB_AGENTS, Integer.toString(this.nbAgents));
		presentation.setAttribute(DomainSize, Integer.toString(this.domainSize));
		presentation.setAttribute(DENSITY, Integer.toString(p1));
		presentation.setAttribute(TIGHTNESS, "0");
		presentation.setAttribute(NB_CONSTRAINTS, Integer.toString(nbConstraint));
		presentation.setAttribute(NB_TUPLES, Integer.toString(nbTuples));
		return presentation;
	}

	/**
	 * This method generates the instances and store a configuration file for the generated 
	 * problems. 
	 */
	public void generate() {

		assert this.minP1 <= this.maxP1 : "minimum densty must be less or equal than maximum density";
		assert (0 < this.minP1 && this.maxP1 <= 100) : "the density must be between ]0,100]";
		assert ((((this.nbAgents*(this.nbAgents-1)/2) * this.minP1) / 100) >= this.nbAgents-1) : "\nWe can not generate a connected graph with the density "+this.minP1+" such that n="+this.nbAgents;

		Element experiments = new Element(Experiments);
		Document confDocument = new Document(experiments);
		experiments.setAttribute(TYPE, DCOP);
		experiments.setAttribute(BENCHMARK, this.typeOfBenchmark);
		Element instances = new Element(INSTANCES);
		instances.setAttribute(TYPE, DCOP);
		instances.setAttribute(BENCHMARK, this.typeOfBenchmark);
		instances.setAttribute(NB_AGENTS, "" + this.nbAgents);
		instances.setAttribute(DomainSize, "" + this.domainSize);
		instances.setAttribute(MINIMAL_COST, "" + this.minCost);
		instances.setAttribute(MAXIMAL_COST, "" + this.maxCost);
		String filePrefix = this.typeOfBenchmark + "_" + this.nbAgents + "_" + this.domainSize + "_";
		instances.setAttribute(NAME, filePrefix);
		for(int p1 = this.minP1 ; p1 <= this.maxP1 ; p1 += this.stepP1) {
			// String subDensity = Integer.toString(p1); // P_1 + "_" + p1;
			File densityFolder = new File(this.path + File.separator + p1);
			if (!densityFolder.exists()) {
				densityFolder.mkdir();
			}

			for(int i = 1 ; i <= this.nbInstances ; i++) {
				String fileName = densityFolder.getPath() + File.separator + filePrefix + i + ".xml";
				this.save(fileName, new Document(this.generateInstance(i, p1)));
			}
		}
		// The configuration file of the Random DisCSP benchmarks
		Element density = new Element(DENSITY);
		density.setAttribute(FROM, Integer.toString(minP1));
		density.setAttribute(TO, Integer.toString(maxP1));
		density.setAttribute(STEP, Integer.toString(stepP1));
		density.setAttribute(InstanceTokens.NB_INSTANCES, Integer.toString(nbInstances) + "");
		instances.addContent(density);
		experiments.addContent(instances);
		String experimentsConfFileName = Commands.Experiments + "_" + this.nbInstances + ".xml";
		this.saveXML(path, experimentsConfFileName, confDocument);
		
	}

	/**
	 * This method generates a random instance having the p1 as network connectivity 
	 * and p2 as constraint tightness. 
	 * 
	 * @param nbOfInstance the number of instance
	 * @param p1 : the network connectivity defined as the ratio of existing binary constraints,
	 * @param p2 : the constraint tightness defined as the ratio of forbidden value pairs. 
	 * @return org.jdom2.Element : the generated instance 
	 */
	private Element generateInstance(int nbOfInstance, int p1) {
		Element instance = new Element(INSTANCE);
		instance.addContent(this.getPresentation(nbOfInstance, DCOP));
		UndirectedGraph<Integer> graph = UndirectedGraph.GenerateRandomGraph(nbAgents, p1, true);
		instance.addContent(graph.getAgents());
		instance.addContent(graph.getDomaines(domainSize));
		instance.addContent(graph.getVariables());
		int nbConstraint = graph.getNbEdges();
		int nbTuples = domainSize * domainSize;
		if (this.constraintsModel.equals(TKC)) { // Totally Known Constraints
			Element constraints = graph.getConstraints();
			constraints.setAttribute(INITIAL_COST, Integer.toString(this.minCost));
			constraints.setAttribute(MAXIMAL_COST, INFINITY);
			instance.addContent(constraints);
			instance.addContent(graph.getSoftRelations(false, domainSize, 0, minCost, maxCost));
			instance.addContent(getGuiPresentation(nbOfInstance, p1, nbConstraint, nbTuples));
		} else { // Partially Known Constraints
			Element constraints = graph.getPartiallyConstrains();
			constraints.setAttribute(INITIAL_COST, Integer.toString(this.minCost));
			constraints.setAttribute(MAXIMAL_COST, INFINITY);
			instance.addContent(constraints);
			instance.addContent(graph.getPartiallySoftRelations(false, domainSize, 0, minCost, maxCost));
			instance.addContent(getGuiPresentation(nbOfInstance, p1, 2 * nbConstraint, nbTuples / 2));
		}
		return instance;
	}

	@Override
	public void run() {
		int nbInstance, minP1, maxP1, stepP1;
		ArrayList<String> keys = null;
		String filePath = this.experimentation.getParent();
		String directory = "";
		Document document = null;
		SAXBuilder sxb = new SAXBuilder();
		try {
			document = sxb.build(this.experimentation);
		} catch (Exception e) {}
		Element xp = document.getRootElement();
		Element instances = xp.getChild(INSTANCES);
		List<Element> communications = xp.getChild(Communications).getChildren();
		List<Element> algorithms = xp.getChild(Algorithms).getChildren();
		int nbAlgorithms = algorithms.size();
		String[] protocols = new String[nbAlgorithms];
		String[] orderings = new String[nbAlgorithms];
		for(int a = 0 ; a < nbAlgorithms ; a++) {
			Element algorithm = algorithms.get(a);
			protocols[a] = algorithm.getText();
			orderings[a] = algorithm.getAttributeValue(AO);
		}
		int nbCommunications = communications.size();
		int nbAgentss = new Integer(instances.getAttributeValue(NB_AGENTS));
		int domainSize = new Integer(instances.getAttributeValue(DomainSize));
		for(int c = 0 ; c < nbCommunications ; c++) {
			Element communication = communications.get(c);
			String commType = communication.getAttributeValue(TYPE);
			String cht = communication.getAttributeValue(channelType);
			int minURC = new Integer(communication.getAttributeValue(MinURC));
			int maxURC = new Integer(communication.getAttributeValue(MaxURC));
			int nbRun = new Integer(communication.getAttributeValue(NbRun));
			long maxNCCCs = new Long(communication.getAttributeValue(MaxNCCCs));
			String name = instances.getAttribute(NAME).getValue();
			Element graphDensity = instances.getChild(DENSITY);
			minP1 = Integer.parseInt(graphDensity.getAttributeValue(FROM));
			maxP1 = Integer.parseInt(graphDensity.getAttributeValue(TO));
			stepP1 = Integer.parseInt(graphDensity.getAttributeValue(STEP));
			nbInstance = new Integer(graphDensity.getAttributeValue(NB_INSTANCES));
			directory = filePath + File.separator + Results + "_" + nbAgentss + "_" + nbInstance;
			try {
				File f = new File(directory);
				if (!f.exists()) {
					f.mkdir();
				}
			} catch (Exception e) {}
			directory += File.separator + commType + "_" + cht;
			try {
				File f = new File(directory);
				if (!f.exists()) {
					f.mkdir();
				}
			} catch (Exception e) {}
			for(int p = 0 ; p < protocols.length ; p++) { // Protocols
				Logger allResuls = new Logger(directory, "All_" + protocols[p] + "-" + orderings[p] + ".csv");
				Logger results = new Logger(directory, protocols[p] + "-" + orderings[p] + ".csv");
				for(int p1 = minP1 ; p1 <= maxP1 ; p1 += stepP1) {
					HashMap<String, Double> densityStatistics = new HashMap<String, Double>();
					for(int i = 1 ; i <= nbInstance ; i++) {
						String problem = filePath + File.separator + p1 + File.separator + name + i + ".xml.bz2";
						HashMap<String, Double> instanceStatistics = new HashMap<String, Double>();
						for(int r = 1 ; r <= nbRun ; r++) {
							DisProblem disProb = new DisProblem(problem, protocols[p]);
							DisSolverSimulator solver = new DisSolverSimulator(disProb);
							solver.setCutOffNCCCs(maxNCCCs);
							solver.addCommunicationMetric();
							solver.setMessageDelay(minURC, maxURC);
							solver.setCentralizedAO(AgentOrdering.getCentralizedAO(orderings[p]));
							solver.solve();
							if (keys == null) {
								keys = disProb.master.getStatisticsKeys();
							}
							HashMap<String, Double> runStatistics = solver.getStatistics();
							if (i == 1 && r == 1 && p1 == minP1) {
								String header = RandomDCOP + COMMA;
								header += Algorithm + COMMA + protocols[p] + COMMA + AO + COMMA + orderings[p] + "\n";
								header += NB_AGENTS + COMMA + nbAgentss + COMMA + DomainSize + COMMA + domainSize + COMMA;
								header += NB_INSTANCES + COMMA + nbInstance + "\n";
								header += channelType + COMMA + cht + COMMA + Communication + COMMA + commType + COMMA;
								header += MinURC + COMMA + minURC + COMMA + MaxURC + COMMA + maxURC + "\n\n";
								header += P_1 + COMMA + INSTANCE + COMMA + NbRun + COMMA + Tools.toString(keys);
								allResuls.printToLogln(header);
								header = RandomDCOP + COMMA;
								header += Algorithm + COMMA + protocols[p] + COMMA + AO + COMMA + orderings[p] + "\n";
								header += NB_AGENTS + COMMA + nbAgentss + COMMA + DomainSize + COMMA + domainSize + COMMA;
								header += NB_INSTANCES + COMMA + nbInstance + "\n";
								header += channelType + COMMA + cht + COMMA + Communication + COMMA + commType + COMMA;
								header += MinURC + COMMA + minURC + COMMA + MaxURC + COMMA + maxURC + "\n\n";
								header += P_1 + COMMA + Tools.toString(keys);
								results.printToLogln(header);
							}
							allResuls.printToLogln("0."+p1 + COMMA + i + COMMA + r + COMMA + Tools.statisticsToString(keys, runStatistics));
							instanceStatistics = Tools.merge(runStatistics, instanceStatistics);
							disProb.clear();
							disProb.setAgents(null);
							disProb = null;
							System.gc();
						}
						Tools.average(instanceStatistics, nbRun);
						densityStatistics = Tools.merge(instanceStatistics, densityStatistics);
					}
					Tools.average(densityStatistics, nbInstance);
					results.printToLogln("0."+p1 + COMMA + Tools.statisticsToString(keys, densityStatistics));
				}
				results.close();
				allResuls.close();
				keys = null;
			} // protocols
			ResultPrinter printer = new ResultPrinter(protocols, orderings, nbHeaderLine);
			printer.setFixKeys(fixKeys);
			ArrayList<String> paths = printer.printToGnuPlot(directory, "", commType);
			printer.plot(directory, paths);
			String header = RandomDCOP + COMMA;
			header += NB_AGENTS + COMMA + nbAgentss + COMMA + DomainSize + COMMA + domainSize + "\n";
			header += "\n#";
			header += channelType + COMMA + cht + COMMA + Communication + COMMA + commType + COMMA;
			header += MinURC + COMMA + minURC + COMMA + MaxURC + COMMA + maxURC + COMMA + NB_INSTANCES + COMMA + nbInstance;
			printer.printAllResults(directory, header);
		} //Communications
	}

	/**
	 * This method generates the instances and store a configuration file for the generated 
	 * problems. 
	 * 
	 * @param path : the path of the directory where instances should be stored 
	 * @param tkc : true, if constraints should be in Totally Known Constraint model, false otherwise
	 * @param i : the number of instances to generate 
	 * @param n : the number of agents/variables, each agent owns exactly one variable
	 * @param d : the domain size, i.e., the number of values in the domain of each variable 
	 * @param minP1 : the minimum network connectivity
	 * @param maxP1 : the maximum network connectivity
	 * @param stepP1 : the step of the network connectivity
	 * @param minP2 : the minimum constraint tightness
	 * @param maxP2 : the maximum constraint tightness
	 * @param stepP2 : the step of the constraint tightness
	 */
	public static void generate(File path, String constraintModel, Integer i, Integer n, Integer d, Integer minP1, Integer maxP1, Integer stepP1, Integer minP2,
			Integer maxP2, Integer stepP2, Integer minCost, Integer maxCost) {
		DCOPGenerator g = new DCOPGenerator(path, constraintModel, i, n, d, minP1, maxP1, stepP1, minCost, maxCost);
		g.generate();
	}
}
