/*
DisChoco: Distributed Constraint Reasoning Solver 
Copyright (C) 2007-2013 Redouane Ezzahir & Mohamed Wahbi 

DisChoco is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DisChoco is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

To contact the authors: <http://dischoco.sourceforge.net/>
 */
package benchmark.random;

import static simulation.parser.InstanceTokens.BENCHMARK;
import static simulation.parser.InstanceTokens.DCOP;
import static simulation.parser.InstanceTokens.INITIAL_COST;
import static simulation.parser.InstanceTokens.INSTANCE;
import static simulation.parser.InstanceTokens.INSTANCES;
import static simulation.parser.InstanceTokens.MAXIMAL_COST;
import static simulation.parser.InstanceTokens.MODEL;
import static simulation.parser.InstanceTokens.NAME;
import static simulation.parser.InstanceTokens.NB_AGENTS;
import static simulation.parser.InstanceTokens.NB_CONSTRAINTS;
import static simulation.parser.InstanceTokens.NB_INSTANCES;
import static simulation.parser.InstanceTokens.NB_TUPLES;
import static simulation.parser.InstanceTokens.SIMPLE_MODEL;
import static simulation.parser.InstanceTokens.TKC;
import static simulation.parser.InstanceTokens.TYPE;
import static util.Commands.AO;
import static util.Commands.Algorithm;
import static util.Commands.Algorithms;
import static util.Commands.Communication;
import static util.Commands.Communications;
import static util.Commands.DENSITY;
import static util.Commands.DomainSize;
import static util.Commands.Experiments;
import static util.Commands.FROM;
import static util.Commands.GuiPresentation;
import static util.Commands.MaxNCCCs;
import static util.Commands.MaxURC;
import static util.Commands.MinURC;
import static util.Commands.NbRun;
import static util.Commands.P_1;
import static util.Commands.P_2;
import static util.Commands.RandomMaxDisCSP;
import static util.Commands.Results;
import static util.Commands.STEP;
import static util.Commands.TIGHTNESS;
import static util.Commands.TO;
import static util.Commands.channelType;
import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import kernel.order.AgentOrdering;
import org.jdom2.Document;
import org.jdom2.Element;
import org.jdom2.input.SAXBuilder;
import simulation.DisProblem;
import simulation.DisSolverSimulator;
import simulation.parser.InstanceTokens;
import util.Commands;
import util.Logger;
import util.Tools;
import util.graph.UndirectedGraph;
import benchmark.Generator;
import benchmark.ResultPrinter;

/**
 * Uniform Random Binary Max-DisCSP instance generator. 
 * DisCSP are characterized by the 4-tuple <n, d, p1, p2>, where: 
 * <br/>
 * <p>
 * <i>n</i> is the number of agents/variables, each agent owns exactly one variable, 
 * <br/>
 * <i>d</i> is the domain size, i.e., the number of values in the domain of each variable, 
 * <br/>
 * <i>p1</i> the network connectivity defined as the ratio of existing binary constraints,
 * <br/>
 * <i>p2</i> is the constraint tightness defined as the ratio of forbidden value pairs (pair with a cost of 1). 
 * </p>
 * @author Mohamed Wahbi
 */
public class MaxDisCSPGenerator extends Generator {

	public static final String fixKeys = P_2;

	/** The number of agent/variable in the problem */
	private int nbAgents;
	/** The domain size of each agent */
	private int domainSize;
	/** The graph density (i.e., the ratio of number of constraint to n*(n-1))*/
	private int minP1 = 20, maxP1 = 80, stepP1 = 10;
	/** Constraints tightness */
	private int minP2 = 10, maxP2 = 90, stepP2 = 10;
	private int nbInstances = 25;
	private String path = ".";

	public static int nbHeaderLine = 4;

	/**
	 * Uniform Random Binary DisCSP instance generator. 
	 * DisCSP are characterized by the 4-tuple <n, d, p1, p2>, where: 
	 * <br/>
	 * <i>n</i> is the number of agents/variables, each agent owns exactly one variable, 
	 * <br/>
	 * <i>d</i> is the domain size, i.e., the number of values in the domain of each variable, 
	 * <br/>
	 * <i>p1</i> the network connectivity defined as the ratio of existing binary constraints,
	 * <br/>
	 * <i>p2</i> is the constraint tightness defined as the ratio of forbidden value pairs. 
	 * <br/>
	 * @param path : the path of the directory where instances should be stored 
	 * @param constraintsModel : the model of constraints, i.e., <tt>TKC</tt> or <tt>PKC</tt>.
	 * @param nbInstances : the number of instances to generate 
	 * @param nbAgents : the number of agents/variables, each agent owns exactly one variable
	 * @param domainSize : the domain size, i.e., the number of values in the domain of each variable 
	 * @param minP1 : the minimum network connectivity
	 * @param maxP1 : the maximum network connectivity
	 * @param stepP1 : the step of the network connectivity
	 * @param minP2 : the minimum constraint tightness
	 * @param maxP2 : the maximum constraint tightness
	 * @param stepP2 : the step of the constraint tightness
	 */
	public MaxDisCSPGenerator(File file, String constraintsModel, int nbInstances, int nbAgents, int domainSize, int minP1, int maxP1, int stepP1,
			int minP2, int maxP2, int stepP2) {
		super(RandomMaxDisCSP, SIMPLE_MODEL, constraintsModel);
		this.path = file.getPath();
		this.nbInstances = nbInstances;
		this.nbAgents = nbAgents;
		this.domainSize = domainSize;
		this.minP1 = minP1;
		this.maxP1 = maxP1;
		this.stepP1 = stepP1;
		this.minP2 = minP2;
		this.maxP2 = maxP2;
		this.stepP2 = stepP2;
	}
	
	/**
	 * Create the thread for running experiments 
	 * @param experimentation : the configuration file of experimentation
	 */
	public MaxDisCSPGenerator(File file) {
		super(file);
	}

	/**
	 * This method return the description of the instance. This will be used in the 
	 * Graphical User Interface. 
	 * 
	 * @param nbOfInstance : the number of instance
	 * @param p1 : the network connectivity defined as the ratio of existing binary constraints,
	 * @param p2 : is the constraint tightness defined as the ratio of forbidden value pairs. 
	 * @param nbConstraint : the number of constraint in the instance
	 * @param nbTuples : the number of tuples 
	 * @return
	 */
	private Element getGuiPresentation(int nbOfInstance, int p1, int p2, int nbConstraint, int nbTuples) {
		Element presentation = new Element(GuiPresentation);
		presentation.setAttribute(TYPE, DCOP);
		presentation.setAttribute(BENCHMARK, this.typeOfBenchmark);
		presentation.setAttribute(NAME, INSTANCE + (nbOfInstance + 1));
		presentation.setAttribute(MODEL, this.model);
		presentation.setAttribute(NB_AGENTS, "" + this.nbAgents);
		presentation.setAttribute(DomainSize, "" + this.domainSize);
		presentation.setAttribute(DENSITY, "" + p1);
		presentation.setAttribute(TIGHTNESS, "" + p2);
		presentation.setAttribute(NB_CONSTRAINTS, "" + nbConstraint);
		presentation.setAttribute(NB_TUPLES, "" + nbTuples);
		return presentation;
	}

	/**
	 * This method generates the instances and store a configuration file for the generated 
	 * problems. 
	 */
	public void generate() {
		assert this.minP1 <= this.maxP1 : "minimum densty must be less or equal than maximum density";
		assert (0 < this.minP1 && this.maxP1 <= 100) : "the density must be between ]0,100]";
		assert this.minP2 <= this.maxP2 : "minimum tightness must be less or equal than maximum tightness";
		assert (0 < this.minP2 && this.maxP2 <= 100) : "the tightness must be between ]0,100]";
		Element experiments = new Element(Experiments);
		Document confDocument = new Document(experiments);
		experiments.setAttribute(TYPE, DCOP);
		experiments.setAttribute(BENCHMARK, this.typeOfBenchmark);
		Element instances = new Element(INSTANCES);
		instances.setAttribute(TYPE, DCOP);
		instances.setAttribute(BENCHMARK, this.typeOfBenchmark);
		instances.setAttribute(NB_AGENTS, "" + this.nbAgents);
		instances.setAttribute(DomainSize, "" + this.domainSize);
		String filePrefix = this.typeOfBenchmark + "_" + this.nbAgents + "_" + this.domainSize + "_";
		instances.setAttribute(NAME, filePrefix);
		for(int p1 = this.minP1 ; p1 <= this.maxP1 ; p1 += this.stepP1) {
			String subDensity = P_1 + "_" + p1;
			File densityFolder = new File(this.path + File.separator + subDensity);
			if (!densityFolder.exists()) {
				densityFolder.mkdir();
			}
			// this.nbConstraint = this.nbAgents * (this.nbAgents - 1) * p1 / 200;
			for(int p2 = this.minP2 ; p2 <= this.maxP2 ; p2 += this.stepP2) {
				String subTightness = Integer.toString(p2);
				// this.tightness = this.domainSize * this.domainSize * p2 / 100;
				File tightnessFolder = new File(densityFolder.getPath() + File.separator + subTightness);
				if (!tightnessFolder.exists()) {
					tightnessFolder.mkdir();
				}
				for(int i = 1 ; i <= this.nbInstances ; i++) {
					String fileName = tightnessFolder.getPath() + File.separator + filePrefix + i + ".xml";
					this.save(fileName, new Document(this.generateInstance(i, p1, p2)));
				}
			}
			// The configuration file of the Random DisCSP benchmarks
			Element density = new Element(DENSITY);
			density.setAttribute(NAME, subDensity);
			density.setAttribute(P_1, Integer.toString(p1));
			Element tightness = new Element(TIGHTNESS);
			tightness.setAttribute(FROM, Integer.toString(minP2));
			tightness.setAttribute(TO, Integer.toString(maxP2));
			tightness.setAttribute(STEP, Integer.toString(stepP2));
			tightness.setAttribute(InstanceTokens.NB_INSTANCES, Integer.toString(nbInstances) + "");
			density.addContent(tightness);
			instances.addContent(density);
		}
		experiments.addContent(instances);
		String experimentsConfFileName = Commands.Experiments + "_" + this.nbInstances + ".xml";
		this.saveXML(path, experimentsConfFileName, confDocument);
	}

	/**
	 * This method generates a random instance having the p1 as network connectivity 
	 * and p2 as constraint tightness. 
	 * 
	 * @param nbOfInstance the number of instance
	 * @param p1 : the network connectivity defined as the ratio of existing binary constraints,
	 * @param p2 : the constraint tightness defined as the ratio of forbidden value pairs. 
	 * @return org.jdom2.Element : the generated instance 
	 */
	private Element generateInstance(int nbOfInstance, int p1, int p2) {
		Element instance = new Element(INSTANCE);
		instance.addContent(this.getPresentation(nbOfInstance, DCOP));
		UndirectedGraph<Integer> graph = UndirectedGraph.GenerateRandomGraph(nbAgents, p1, true);
		instance.addContent(graph.getAgents());
		instance.addContent(graph.getDomaines(domainSize));
		instance.addContent(graph.getVariables());
		int nbConstraint = graph.getNbEdges();
		int nbTuples = (int) ((domainSize * domainSize * p2) / 100);
		if (this.constraintsModel.equals(TKC)) { // Totally Known Constraints
			Element constraints = graph.getConstraints();
			constraints.setAttribute(INITIAL_COST, Integer.toString(0));
			constraints.setAttribute(MAXIMAL_COST, Integer.toString(nbConstraint + 1));
			instance.addContent(constraints);
			instance.addContent(graph.getSoftRelations(true, domainSize, p2, 0, 1));
			instance.addContent(getGuiPresentation(nbOfInstance, p1, p2, nbConstraint, nbTuples));
		} else { // Partially Known Constraints
			Element constraints = graph.getPartiallyConstrains();
			constraints.setAttribute(INITIAL_COST, Integer.toString(0));
			constraints.setAttribute(MAXIMAL_COST, Integer.toString(2 * nbConstraint + 1));
			instance.addContent(constraints);
			instance.addContent(graph.getPartiallySoftRelations(true, domainSize, p2, 0, 1));
			instance.addContent(getGuiPresentation(nbOfInstance, p1, p2, 2 * nbConstraint, nbTuples / 2));
		}
		return instance;
	}

	@Override
	public void run() {
		int nbInstance, minP2, maxP2, stepP2;
		String p1;
		ArrayList<String> keys = null;
		String filePath = this.experimentation.getParent();
		String directory = "";
		Document document = null;
		SAXBuilder sxb = new SAXBuilder();
		try {
			document = sxb.build(this.experimentation);
		} catch (Exception e) {}
		Element xp = document.getRootElement();
		Element instances = xp.getChild(INSTANCES);
		List<Element> communications = xp.getChild(Communications).getChildren();
		List<Element> algorithms = xp.getChild(Algorithms).getChildren();
		String[] protocols = new String[algorithms.size()];
		String[] orderings = new String[algorithms.size()];
		int a = 0;
		Iterator<Element> it = algorithms.iterator();
		while (it.hasNext()) {
			Element algorithm = it.next();
			protocols[a] = algorithm.getText();
			orderings[a] = algorithm.getAttributeValue(AO);
			a++;
		}
		it = communications.iterator();
		int nbAgents = new Integer(instances.getAttributeValue(NB_AGENTS));
		int domainSize = new Integer(instances.getAttributeValue(DomainSize));
		while (it.hasNext()) {
			// /////////////////////////////////////////////////////////////////
			Element comm = it.next();
			String commType = comm.getAttributeValue(TYPE);
			String cht = comm.getAttributeValue(channelType);
			int minURC = new Integer(comm.getAttributeValue(MinURC));
			int maxURC = new Integer(comm.getAttributeValue(MaxURC));
			int nbRun = new Integer(comm.getAttributeValue(NbRun));
			long maxNCCCs = new Long(comm.getAttributeValue(MaxNCCCs));
			String name = instances.getAttribute(NAME).getValue();
			List<Element> graphDensities = instances.getChildren();
			for(int d = 0 ; d < graphDensities.size() ; d++) {
				Element graphDensity = graphDensities.get(d);
				p1 = graphDensity.getAttributeValue(P_1);
				directory = filePath + File.separator + Results + "_" + nbAgents + "_" + domainSize;
				try {
					File f = new File(directory);
					if (!f.exists()) {
						f.mkdir();
					}
				} catch (Exception e) {}
				directory += File.separator + commType + "_" + cht;
				try {
					File f = new File(directory);
					if (!f.exists()) {
						f.mkdir();
					}
				} catch (Exception e) {}
				String subFolder = directory + File.separator + graphDensity.getAttributeValue(NAME);
				String problemsFolder = filePath + File.separator + graphDensity.getAttributeValue(NAME);
				try {
					File f = new File(subFolder);
					if (!f.exists()) {
						f.mkdir();
					}
				} catch (Exception e) {}
				for(int p = 0 ; p < protocols.length ; p++) { // Protocols
					Logger allResuls = new Logger(subFolder, "All_" + protocols[p] + "-" + orderings[p] + ".csv");
					Logger results = new Logger(subFolder, protocols[p] + "-" + orderings[p] + ".csv");
					boolean headerPrint = true;
					List<Element> tightnesses = graphDensity.getChildren();
					Iterator<Element> t = tightnesses.iterator();
					while (t.hasNext()) {
						Element tightness = t.next();
						minP2 = Integer.parseInt(tightness.getAttributeValue(FROM));
						maxP2 = Integer.parseInt(tightness.getAttributeValue(TO));
						stepP2 = Integer.parseInt(tightness.getAttributeValue(STEP));
						nbInstance = new Integer(tightness.getAttributeValue(NB_INSTANCES));
						for(int _p2 = minP2 ; _p2 <= maxP2 ; _p2 += stepP2) {
							String p2 = String.format("%02d", _p2);
							HashMap<String, Double> densityStatistics = new HashMap<String, Double>();
							for(int i = 1 ; i <= nbInstance ; i++) {
								String problem = problemsFolder + File.separator + p2 + File.separator + name + i + ".xml.bz2";
								HashMap<String, Double> instanceStatistics = new HashMap<String, Double>();
								for(int r = 1 ; r <= nbRun ; r++) {
									DisProblem disProb = new DisProblem(problem, protocols[p]);
									DisSolverSimulator solver = new DisSolverSimulator(disProb);
									solver.setCutOffNCCCs(maxNCCCs);
									solver.addCommunicationMetric();
									solver.setMessageDelay(minURC, maxURC);
									solver.setCentralizedAO(AgentOrdering.getCentralizedAO(orderings[p]));
									solver.solve();
									if (keys == null) {
										keys = disProb.master.getStatisticsKeys();
									}
									HashMap<String, Double> runStatistics = solver.getStatistics();
									if (headerPrint) {
										headerPrint = false;
										String header = RandomMaxDisCSP + COMMA;
										header += Algorithm + COMMA + protocols[p] + COMMA + AO + COMMA + orderings[p] + "\n";
										header += NB_AGENTS + COMMA + nbAgents + COMMA + DomainSize + COMMA + domainSize + COMMA;
										header += P_1 + COMMA + "0." + p1 + COMMA + NB_INSTANCES + COMMA + nbInstance + "\n";
										header += channelType + COMMA + cht + COMMA + Communication + COMMA + commType + COMMA;
										header += MinURC + COMMA + minURC + COMMA + MaxURC + COMMA + maxURC + "\n\n";
										header += P_2 + COMMA + INSTANCE + COMMA + NbRun + COMMA + Tools.toString(keys);
										allResuls.printToLogln(header);
										header = RandomMaxDisCSP + COMMA;
										header += Algorithm + COMMA + protocols[p] + COMMA + AO + COMMA + orderings[p] + "\n";
										header += NB_AGENTS + COMMA + nbAgents + COMMA + DomainSize + COMMA + domainSize + COMMA;
										header += P_1 + COMMA + "0." + p1 + COMMA + NB_INSTANCES + COMMA + nbInstance + "\n";
										header += channelType + COMMA + cht + COMMA + Communication + COMMA + commType + COMMA;
										header += MinURC + COMMA + minURC + COMMA + MaxURC + COMMA + maxURC + "\n\n";
										header += P_2 + COMMA + Tools.toString(keys);
										results.printToLogln(header);
									}
									allResuls.printToLogln("0." + p2 + COMMA + i + COMMA + r + COMMA + Tools.statisticsToString(keys, runStatistics));
									instanceStatistics = Tools.merge(runStatistics, instanceStatistics);
									disProb.clear();
									disProb.setAgents(null);
									disProb = null;
									System.gc();
								}
								Tools.average(instanceStatistics, nbRun);
								densityStatistics = Tools.merge(instanceStatistics, densityStatistics);
							}
							Tools.average(densityStatistics, nbInstance);
							results.printToLogln("0." + p2 + COMMA + Tools.statisticsToString(keys, densityStatistics));
						}
					}
					results.close();
					allResuls.close();
					keys = null;
				}// protocols
				ResultPrinter printer = new ResultPrinter(protocols, orderings, nbHeaderLine);
				printer.setFixKeys(fixKeys);
				ArrayList<String> paths = printer.printToGnuPlot(subFolder, "", commType);
				printer.plot(subFolder, paths);
				String header = RandomMaxDisCSP + COMMA;
				header += NB_AGENTS + COMMA + nbAgents + COMMA + DomainSize + COMMA + domainSize + "\n";
				header += "\n#";
				header += channelType + COMMA + cht + COMMA + Communication + COMMA + commType + COMMA;
				header += MinURC + COMMA + minURC + COMMA + MaxURC + COMMA + maxURC + COMMA + P_1 + COMMA + p1 + COMMA;
				printer.printAllResults(subFolder, header);
			} // density
		}
	}

}
