/*
DisChoco: Distributed Constraint Reasoning Solver 
Copyright (C) 2007-2013 Redouane Ezzahir & Mohamed Wahbi 

DisChoco is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DisChoco is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

To contact the authors: <http://dischoco.sourceforge.net/>
*/ 
package gui.debug;

import gui.graph.presentation.GraphPanel;
import gui.tree.presentation.TreePanel;
import gui.tree.presentation.TreePanel.Orientation;
import java.awt.BorderLayout;
import java.awt.CardLayout;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import kernel.VarIdentifier;
import simulation.parser.InstanceParser;
import simulation.parser.PRelation.RelationTableModel;
import util.Commands;
import util.Instantiation;
import util.VarAssignment;
import util.graph.UndirectedGraph;
import util.graph.Vertex;
import util.tree.AbstractTree;
import dischoco.DisChoco;

@SuppressWarnings("serial")
public class SolutionChecker extends JPanel {

	private final static String TREE = Commands.TREE;
	private final static String GRAPH = Commands.GRAPH;
	private final static int WEST_WIDTH = 225;
	private InstanceParser parser = null;
	private Dimension dimension = null;
	private UndirectedGraph<VarAssignment> graph = null;
	private AbstractTree<VarAssignment> priorityTree = null;
	private Map<String, JTable> mapOfConstraints = null;
	private JPanel westPanel = null;
	private JPanel constraintsPanel = null;
	private JPanel eastPanel = null;
	private JPanel canvasPanel = null;
	private JPanel contentPanel = null;
	private GraphPanel<VarAssignment> graphCanvas = null;
	private TreePanel<VarAssignment> treeCanvas = null;
	private JComboBox typeGraph = null;
	private JButton checkButton = null;
	private JTextField solutionTextField = null;

	public SolutionChecker() {
		super(new BorderLayout());
		initialize();
	}

	/**
	 * Initialize the class.
	 */
	private void initialize() {
		this.contentPanel = new JPanel(new BorderLayout());
		this.dimension = new Dimension(1024, 520);
		contentPanel.setMinimumSize(this.dimension);
		contentPanel.setMaximumSize(this.dimension);
		contentPanel.setPreferredSize(this.dimension);
	}

	/**
	 * 
	 * @param file
	 */
	public void setProblem() {
		initialize();
		this.parseProblem();
		this.createWestPanel();
		this.createEastPanel();
		this.add(contentPanel, BorderLayout.CENTER);
	}

	/**
	 * 
	 */
	private void parseProblem() {
		String fileName = DisChoco.filePath;
		this.parser = new InstanceParser(fileName);
		this.parser.loadInstance(fileName);
		this.parser.parse(false);
		this.graph = parser.getAssignedConstraintGraph();
		this.priorityTree = graph.DFS();
		String type = this.parser.getType();
		if (!(type.equals(Commands.DisCSP) || type.equals(Commands.DCOP))) {
			throw new UnsupportedOperationException("Not yet imlemented for " + type);
		}
		if (this.parser.isComplexModel()) {
			this.parser.convertToSimpleModel();
		}
		this.mapOfConstraints = parser.getConstraintTables();
	}

	/**
	 * This method initializes westPanel
	 * 
	 */
	private void createWestPanel() {
		this.westPanel = new JPanel(new BorderLayout());
		Dimension dim = new Dimension(WEST_WIDTH, dimension.height);
		this.westPanel.setSize(dim);
		this.westPanel.setPreferredSize(dim);
		JPanel north = new JPanel(new FlowLayout());
		JLabel constraintLabel = new JLabel("Constraint ", JLabel.CENTER);
		north.add(constraintLabel);
		String[] constraints = this.mapOfConstraints.keySet().toArray(new String[0]);
		Arrays.sort(constraints);
		this.westPanel.add(north, BorderLayout.NORTH);
		this.constraintsPanel = new JPanel(new CardLayout());
		for(String constraint : this.mapOfConstraints.keySet()) {
			JTable table = this.mapOfConstraints.get(constraint);
			this.constraintsPanel.add(new JScrollPane(table), constraint);
		}
		final JComboBox constraintsComboBox = new JComboBox(constraints);
		north.add(constraintsComboBox);
		constraintsComboBox.addItemListener(new ItemListener() {

			public void itemStateChanged(ItemEvent e) {
				if (e.getStateChange() == ItemEvent.SELECTED) {
					updateTablesPanel(constraintsComboBox);
				}
			}
		});
		this.westPanel.add(constraintsPanel, BorderLayout.CENTER);
		contentPanel.add(makeTitledPanel("Constraints", westPanel), BorderLayout.WEST);
		updateTablesPanel(constraintsComboBox);
	}

	/**
	 * 
	 */
	private void updateTablesPanel(JComboBox constraintsComboBox) {
		String selected = (String) constraintsComboBox.getSelectedItem();
		CardLayout cards = (CardLayout) (constraintsPanel.getLayout());
		cards.show(constraintsPanel, selected);
	}

	/**
	 * This method initializes figurePanel
	 * 
	 * @return javax.swing.JPanel
	 */
	private void createEastPanel() {
		eastPanel = new JPanel(new BorderLayout());
		Dimension dim = new Dimension(dimension.width - WEST_WIDTH, dimension.height);
		eastPanel.setSize(dim);
		eastPanel.setPreferredSize(dim);
		JPanel north = new JPanel(new BorderLayout());
		JLabel typeLabel = new JLabel("Type of Constraint Graph: ", JLabel.LEFT);
		JPanel westNorth = new JPanel(new FlowLayout());
		westNorth.setLayout(new BoxLayout(westNorth, BoxLayout.X_AXIS));
		westNorth.add(typeLabel);
		createTypeGraph();
		westNorth.add(typeGraph);
		JLabel solLabel = new JLabel("Solution: ", JLabel.LEFT);
		westNorth.add(solLabel);
		north.add(westNorth, BorderLayout.WEST);
		solutionTextField = new JTextField(1000);
		north.add(solutionTextField, BorderLayout.CENTER);
		checkButton = new JButton("Check");
		checkButton.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				Instantiation solution = null;
				try {
					solution = Instantiation.fromString(solutionTextField.getText());
				} catch (Exception ex) {
					Container c = getParent();
					while (!(c instanceof JFrame)) {
						c = c.getParent();
					}
					JOptionPane.showMessageDialog(c, "The string is not an Instantiation", Commands.DisChoco_TITLE, JOptionPane.ERROR_MESSAGE, null);
					solutionTextField.setText("");
					return;
				}
				Map<VarIdentifier, Vertex<VarAssignment>> map = new HashMap<VarIdentifier, Vertex<VarAssignment>>();
				for(Vertex<VarAssignment> vertex : graph.getVertexes()) {
					VarAssignment ass = vertex.getData();
					VarIdentifier varId = ass.getVarIdentifier();
					VarAssignment varAss = solution.getVarAssignment(varId);
					if(varAss !=null) {
						int value = varAss.getValue();
						ass.setValue(value);
						map.put(varId, vertex);
					}
				}
				for(Vertex<VarAssignment> vertex : priorityTree.getVertexes()) {
					VarAssignment ass = vertex.getData();
					VarIdentifier varId = ass.getVarIdentifier();
					VarAssignment varAss = solution.getVarAssignment(varId);
					if(varAss !=null) {
						int value = varAss.getValue();
						ass.setValue(value);
						// map.put(varId, vertex);
					}
				}
				Integer cost = new Integer(0);
				for(String key : mapOfConstraints.keySet()) {
					JTable table = mapOfConstraints.get(key);
					int arity = table.getColumnCount() - 1;
					VarIdentifier[] varIds = new VarIdentifier[arity];
					int[] tuple = new int[arity];
					for(int c = 0 ; c < arity ; c++) {
						String var = table.getColumnName(c);
						VarIdentifier varId = VarIdentifier.fromString(var);
						varIds[c] = varId;
						VarAssignment varAss = solution.getVarAssignment(varId);
						if(varAss !=null) {
							tuple[c] = varAss.getValue();
						} else {
							tuple[c] = Integer.MIN_VALUE;
						}
					}
					RelationTableModel model = (RelationTableModel) table.getModel();
					int row = model.binarySearch(tuple);
					if (row >= 0) {
						for(int column = 0 ; column < table.getColumnCount() ; column++) {
							RelationRenderer renderer = (RelationRenderer) table.getCellRenderer(row, column);
							renderer.setSelectedRow(row);
						}
						Integer value = (Integer) table.getValueAt(row, arity);
						cost += value;
						Vertex<VarAssignment> source = map.get(varIds[0]);
						Vertex<VarAssignment> target = map.get(varIds[1]);
						graph.setWeight(source, target, value);
						priorityTree.setWeight(source, target, value);
						System.out.println(key + " " + Arrays.toString(tuple) + " costs " + value);
						model.fireTableDataChanged();
					} else {
						System.out.println(key + " " + Arrays.toString(tuple) + " is not in the table !!!");
					}
				}
				System.out.println("The cost of the solution is: " + cost);
				solutionTextField.setText("");
				graphCanvas.repaint();
				treeCanvas.repaint();
			}
		});
		north.add(checkButton, BorderLayout.EAST);
		eastPanel.add(north, BorderLayout.NORTH);
		createCanvasPanel();
		contentPanel.add(makeTitledPanel("Constraint Graph", eastPanel), BorderLayout.CENTER);
	}

	/**
	 * This method initializes figurePanel
	 * 
	 * @return javax.swing.JPanel
	 */
	private void createCanvasPanel() {
		Dimension dim = new Dimension(dimension.width - 2 * WEST_WIDTH, dimension.height - 100);
		graphCanvas = new GraphPanel<VarAssignment>(graph, dim, Orientation.HORIZONTAL);
		treeCanvas = new TreePanel<VarAssignment>(priorityTree, dim, Orientation.HORIZONTAL);
		dim = new Dimension(dimension.width - WEST_WIDTH, dimension.height - 50);

		canvasPanel = new JPanel(new CardLayout());
		canvasPanel.setSize(dim);
		canvasPanel.setPreferredSize(dim);

		// ------------------------------------------------------------------------------
		// dim = new Dimension(WEST_WIDTH, dimension.height-50);
		// JPanel graphPanel = new JPanel(new BorderLayout());
		// JPanel westGraph = new JPanel();
		// westGraph.setLayout(new BoxLayout(westGraph, BoxLayout.Y_AXIS));
		// westGraph.setPreferredSize(dim);
		// westGraph.setMinimumSize(dim);
		// westGraph.add(getTitledType());
		// westGraph.add(makeTitledPanel("Zoom", graphCanvas.getZoomControls()));
		// westGraph.add(makeTitledPanel("Mouse", graphCanvas.getModeControls()));
		// westGraph.add(makeTitledPanel("Layout", graphCanvas.getLayoutControls()));
		// westGraph.add(makeTitledPanel("Help", graphCanvas.getHelpControls()));
		// graphPanel.add(westGraph, BorderLayout.WEST);
		// JPanel gPanel = new JPanel(new BorderLayout());
		// gPanel.add(graphCanvas, BorderLayout.CENTER);
		// gPanel.setBorder(BorderFactory.createLineBorder(Color.black));
		// graphPanel.add(gPanel, BorderLayout.CENTER);
		canvasPanel.add(graphCanvas, GRAPH);
		// ------------------------------------------------------------------------------
		// JPanel treePanel = new JPanel(new BorderLayout());
		// JPanel westTree = new JPanel();
		// westTree.setLayout(new BoxLayout(westTree, BoxLayout.Y_AXIS));
		// westTree.setPreferredSize(dim);
		// westTree.setMinimumSize(dim);
		// westTree.add(getTitledType());
		// westTree.add(makeTitledPanel("Zoom", treeCanvas.getZoomControls()));
		// westTree.add(makeTitledPanel("Mouse", treeCanvas.getModeControls()));
		// westTree.add(makeTitledPanel("Help", treeCanvas.getHelpControls()));
		// treePanel.add(westTree, BorderLayout.WEST);
		// JPanel tPanel = new JPanel(new BorderLayout());
		// tPanel.add(treeCanvas, BorderLayout.CENTER);
		// tPanel.setBorder(BorderFactory.createLineBorder(Color.black));
		// treePanel.add(tPanel, BorderLayout.CENTER);
		canvasPanel.add(treeCanvas, TREE);

		// ------------------------------------------------------------------------------
		// canvasPanel.add(graphCanvas.getGraphPanel(), GRAPH);
		// canvasPanel.add(treeCanvas.getTreePanel(), TREE);
		eastPanel.add(canvasPanel, BorderLayout.CENTER);
	}

	// public JPanel getTitledType() {
	// if(titledType == null) {
	// titledType = makeTitledPanel("Type", getTypeGraph());
	// }
	// return titledType;
	// }

	public void createTypeGraph() {
		typeGraph = new JComboBox();
		typeGraph.setBounds(new Rectangle(140, 10, 150, 22));
		typeGraph.addItem(Commands.GRAPH);
		typeGraph.addItem(Commands.TREE);
		typeGraph.addItemListener(new ItemListener() {

			public void itemStateChanged(ItemEvent e) {
				if (e.getStateChange() == ItemEvent.SELECTED) {
					plotEastPanel((String) e.getItem());
				}
			}
		});
	}

	private void plotEastPanel(String selectedItem) {
		if (selectedItem.equals(Commands.TREE)) {
			plotTreeAO();
		} else {
			plotGraphAO();
		}
	}

	/**
	 * 
	 */
	private void plotGraphAO() {
		CardLayout cards = (CardLayout) (canvasPanel.getLayout());
		cards.show(canvasPanel, GRAPH);
	}

	/**
	 * 
	 */
	private void plotTreeAO() {
		CardLayout cards = (CardLayout) (canvasPanel.getLayout());
		cards.show(canvasPanel, TREE);
	}

	private static JPanel makeTitledPanel(String title, JComponent component) {
		JPanel panel = new JPanel(new BorderLayout());
		panel.add(component, BorderLayout.CENTER);
		panel.setBorder(BorderFactory.createTitledBorder(title));
		return panel;
	}

}
