/*
DisChoco: Distributed Constraint Reasoning Solver 
Copyright (C) 2007-2013 Redouane Ezzahir & Mohamed Wahbi 

DisChoco is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DisChoco is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

To contact the authors: <http://dischoco.sourceforge.net/>
*/ 
package gui.generators;

import gui.LabelAccessory;
import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.EventQueue;
import java.awt.FlowLayout;
import java.awt.Rectangle;
import java.io.File;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JSpinner;
import javax.swing.SpinnerNumberModel;
import util.Commands;
import benchmark.random.MaxDisCSPGenerator;
import dischoco.DisChoco;

/**
 * The panel for configuring Uniform Random Binary DisCSP benchmark. 
 * @version dischoco 2.5.3<br/>
 * gui.generators.RandomDCSPPanel.java
 * @author Mohamed Wahbi
 * @see DisCSPGenerator
 * @see DisCSPGenerator#DisCSPGenerator(File, boolean, int, int, int, int, int, int, int, int, int)
 */
@SuppressWarnings("serial")
public class RandomMaxDisCSPPanel extends JPanel {

	public static final String TOTALLY_KNOWN_CONSTRAINT = Commands.TOTALLY_KNOWN_CONSTRAINT;
	public static final String PARTIALLY_KNOWN_CONSTRAINT = Commands.PARTIALLY_KNOWN_CONSTRAINT;
	public static final String SUCCEFULLY = Commands.SUCCEFULLY;
	public static final String BROWSE = Commands.BROWSE;
	public static final String GENERATE = Commands.GENERATE;
	public static final String CONSTRAINT_MODEL = Commands.CONSTRAINT_MODEL;
	public static final String NB_INSTANCES = Commands.NumberOfInstances;
	public static final String NB_AGENTS = Commands.NumberOfAgents;
	public static final String DOMAIN_SIZE = Commands.SizeOfDomain;
	public static final String FROM = Commands.FROM;
	public static final String TO = Commands.TO;
	public static final String STEP = Commands.STEP;
	public static final String DENSITY_OF_GRAPH = Commands.DENSITY_OF_GRAPH;
	public static final String CONSTRAINTS_TIGHTNESS = Commands.CONSTRAINTS_TIGHTNESS;

	private JPanel northPanel = null;
	private JPanel centerPanel = null;
	private JPanel parametersPanel = null;
	private JSpinner minP1 = null;
	private JSpinner maxP1 = null;
	private JSpinner stepP1 = null;
	private JSpinner minP2 = null;
	private JSpinner maxP2 = null;
	private JSpinner stepP2 = null;
	private JSpinner nbInstances = null;
	private JSpinner nbAgents = null;
	private JSpinner domainSize = null;
	private JComboBox constraintsModel = null;
	private JButton browseButton;
	private JButton generateButton;
	private JFileChooser fileChooser = null;
	private File file = null;
	private JFrame mainFrame = null;
	
	/**
	 * Create the panel.
	 */
	public RandomMaxDisCSPPanel(JFrame dischoco) {
		super(new BorderLayout());
		this.mainFrame = dischoco;
		this.setPreferredSize(new Dimension(600, 425));
		this.add(createNorthPanel(), BorderLayout.NORTH);
		this.add(createCenterPanel(), BorderLayout.CENTER);
	}

	/**
	 * This method initializes northPanel
	 * @return javax.swing.JPanel 
	 */
	protected JPanel createNorthPanel() {
		northPanel = new JPanel(new FlowLayout());
		browseButton = new JButton();
		browseButton.setText(BROWSE);
		browseButton.addActionListener(new java.awt.event.ActionListener() {
			public void actionPerformed(java.awt.event.ActionEvent e) {
				fileChooser = new JFileChooser(DisChoco.filePath);
				fileChooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
				fileChooser.setAccessory(new LabelAccessory(fileChooser));
				int returnVal = fileChooser.showSaveDialog((Component) e.getSource());
				if (returnVal == JFileChooser.APPROVE_OPTION) {
					file = fileChooser.getSelectedFile();
					if(!file.isDirectory()) {
						file = fileChooser.getCurrentDirectory();
					}
					DisChoco.filePath = file.getAbsolutePath();
					System.out.println("Path : " + file.getAbsolutePath() + "\n");
				}
			}
		});
		northPanel.add(browseButton);
		
		generateButton = new JButton();
		generateButton.setText(GENERATE);
		generateButton.addActionListener(new java.awt.event.ActionListener() {
			public void actionPerformed(java.awt.event.ActionEvent e) {
				if (file == null) {
					JOptionPane.showMessageDialog(mainFrame, "Choose firstly a directoty", "DisChoco", JOptionPane.WARNING_MESSAGE);
					return;
				}
				EventQueue.invokeLater(new Runnable() {

					public void run() {
						try {
							Integer nbInstances = getNbInstances();
							Integer nbAgents = getNbAgents();
							Integer domainSize = getDomainSize();
							Integer minP1 = getMinP1();
							Integer maxP1 = getMaxP1();
							Integer stepP1 = getStepP1();
							Integer minP2 = getMinP2();
							Integer maxP2 = getMaxP2();
							Integer stepP2 = getStepP2();
							String constraintsModel = getConstraintsModel();
							MaxDisCSPGenerator g = new MaxDisCSPGenerator(file, constraintsModel, 
									nbInstances, nbAgents, domainSize, 
									minP1, maxP1, stepP1, 
									minP2, maxP2, stepP2);
							g.generate();
							JOptionPane.showMessageDialog(mainFrame, SUCCEFULLY);
							file = null;
						} catch (Exception e) {
							e.printStackTrace();
						}
					}
				});
			}
		});
		northPanel.add(generateButton);
		
		return northPanel;
	}

	/**
	 * This method initializes parametersPanel
	 * @return javax.swing.JPanel 
	 */
	protected JPanel createCenterPanel() {
		centerPanel = new JPanel(new FlowLayout());
		centerPanel.add(createParametersPanel());
		return centerPanel;
	}


	/**
	 * This method initializes parametersPanel
	 * @return javax.swing.JPanel 
	 */
	protected JPanel createParametersPanel() {
		parametersPanel = new JPanel(null);
		parametersPanel.setPreferredSize(new Dimension(600, 375));
		parametersPanel.setMinimumSize(new Dimension(600, 375));
		parametersPanel.setMaximumSize(new Dimension(600, 375));

		JLabel lblConstraintModel = new JLabel(CONSTRAINT_MODEL, JLabel.RIGHT);
		lblConstraintModel.setBounds(new Rectangle(150, 15, 120, 22));
		parametersPanel.add(lblConstraintModel);
		
		String[] constraintsModels = {TOTALLY_KNOWN_CONSTRAINT, PARTIALLY_KNOWN_CONSTRAINT};
		constraintsModel = new JComboBox(constraintsModels);
		constraintsModel.setSelectedItem(TOTALLY_KNOWN_CONSTRAINT);
		constraintsModel.setBounds(new Rectangle(280, 15, 210, 24));
		parametersPanel.add(constraintsModel);

		JLabel lblNbInstances = new JLabel(NB_INSTANCES, JLabel.RIGHT);
		lblNbInstances.setBounds(new Rectangle(200, 65, 150, 22));
		parametersPanel.add(lblNbInstances);

		nbInstances = new JSpinner(new SpinnerNumberModel(25, 1, 200, 1));
		nbInstances.setBounds(new Rectangle(360, 65, 60, 22));
		parametersPanel.add(nbInstances);

		JLabel lblNbAgents = new JLabel(NB_AGENTS, JLabel.RIGHT);
		lblNbAgents.setBounds(new Rectangle(200, 110, 150, 22));
		parametersPanel.add(lblNbAgents);

		nbAgents = new JSpinner(new SpinnerNumberModel(20, 5, 50, 1));
		nbAgents.setBounds(new Rectangle(360, 110, 60, 22));
		parametersPanel.add(nbAgents);

		JLabel lblDomainSize = new JLabel(DOMAIN_SIZE, JLabel.RIGHT);
		lblDomainSize.setBounds(new Rectangle(200, 150, 150, 22));
		parametersPanel.add(lblDomainSize);

		domainSize = new JSpinner(new SpinnerNumberModel(10, 2, 30, 1));
		domainSize.setBounds(new Rectangle(360, 150, 60, 22));
		parametersPanel.add(domainSize, null);

		JLabel lblFrom = new JLabel(FROM, JLabel.CENTER);
		lblFrom.setBounds(new Rectangle(260, 180, 50, 22));
		parametersPanel.add(lblFrom);

		JLabel lblTo = new JLabel(TO, JLabel.CENTER);
		lblTo.setBounds(new Rectangle(435, 180, 50, 22));
		parametersPanel.add(lblTo);

		JLabel lblStep = new JLabel(STEP, JLabel.CENTER);
		lblStep.setBounds(new Rectangle(350, 180, 50, 22));
		parametersPanel.add(lblStep);
		
		JLabel lblP1 = new JLabel(DENSITY_OF_GRAPH+" (%)", JLabel.RIGHT);
		lblP1.setBounds(new Rectangle(85, 215, 160, 22));
		parametersPanel.add(lblP1);
		
		JLabel lblP2 = new JLabel(CONSTRAINTS_TIGHTNESS+" (%)", JLabel.RIGHT);
		lblP2.setBounds(new Rectangle(85, 255, 160, 22));
		parametersPanel.add(lblP2);
		
		minP1 = new JSpinner(new SpinnerNumberModel(20, 1, 100, 1));
		minP1.setBounds(new Rectangle(255, 215, 60, 22));
		parametersPanel.add(minP1);
		
		maxP1 = new JSpinner(new SpinnerNumberModel(80, 1, 100, 1));
		maxP1.setBounds(new Rectangle(430, 215, 60, 22));
		parametersPanel.add(maxP1);
	
		stepP1 = new JSpinner(new SpinnerNumberModel(30, 1, 100, 1));
		stepP1.setBounds(new Rectangle(345, 215, 60, 22));
		parametersPanel.add(stepP1);

		minP2 = new JSpinner(new SpinnerNumberModel(10, 1, 100, 1));
		minP2.setBounds(new Rectangle(255, 255, 60, 22));
		parametersPanel.add(minP2);

		maxP2 = new JSpinner(new SpinnerNumberModel(90, 1, 100, 1));
		maxP2.setBounds(new Rectangle(430, 255, 60, 22));
		parametersPanel.add(maxP2);

		stepP2 = new JSpinner(new SpinnerNumberModel(20, 1, 100, 1));
		stepP2.setBounds(new Rectangle(345, 255, 60, 22));
		parametersPanel.add(stepP2);

		return parametersPanel;
	}
	
	public Integer getNbInstances() {
		return (Integer) this.nbInstances.getValue();
	}
	
	public Integer getNbAgents() {
		return (Integer) this.nbAgents.getValue();
	}
	
	public Integer getDomainSize() {
		return (Integer) this.domainSize.getValue();
	}
	
	public Integer getMinP1() {
		return (Integer) this.minP1.getValue();
	}
	
	public Integer getMaxP1() {
		return (Integer) this.maxP1.getValue();
	}
	
	public Integer getStepP1() {
		return (Integer) this.stepP1.getValue();
	}
	
	public Integer getMinP2() {
		return (Integer) this.minP2.getValue();
	}
	
	public Integer getMaxP2() {
		return (Integer) this.maxP2.getValue();
	}
	
	public Integer getStepP2() {
		return (Integer) this.stepP2.getValue();
	}

	public String getConstraintsModel() {
		String model = (String) constraintsModel.getSelectedItem();
		if(model.equals(TOTALLY_KNOWN_CONSTRAINT)) {
			return Commands.TKC;
		}
		return Commands.PKC;
	}
}
