package gui.graph.layout;

import util.graph.Vertex;
import gui.graph.presentation.GraphPresentation;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * A {@code Layout} implementation that positions vertices equally spaced on a regular circle.
 *
 * @author Mohamed Wahbi
 */
public class CircleLayout<T extends Comparable<? super T>> extends Layout<T> {

	private double radius;
	private List<Vertex<T>> vertex_ordered_list;
	private Map<T, CircleVertexData> circleVertexDataMap = null;

	/**
	 * Creates an instance for the specified graph.
	 */
	public CircleLayout(GraphPresentation<T> graph) {
		super(graph);
		this.circleVertexDataMap = new HashMap<T, CircleVertexData>();
		for(Vertex<T> vertex : getVertexes()) {
			this.circleVertexDataMap.put(vertex.getData(), new CircleVertexData());
		}
	}

	/**
	 * Returns the radius of the circle.
	 */
	public double getRadius() {
		return radius;
	}

	/**
	 * Sets the radius of the circle.  Must be called before
	 * {@code initialize()} is called.
	 */
	public void setRadius(double radius) {
		this.radius = radius;
	}

	/**
	 * Sets the order of the vertices in the layout according to the ordering
	 * specified by {@code comparator}.
	 */
	public void setVertexOrder(Comparator<Vertex<T>> comparator) {
		if (vertex_ordered_list == null)
			vertex_ordered_list = new ArrayList<Vertex<T>>(getVertexes());
		Collections.sort(vertex_ordered_list, comparator);
	}

	/**
	 * Sets the order of the vertices in the layout according to the ordering
	 * of {@code vertex_list}.
	 */
	public void setVertexOrder(List<Vertex<T>> vertex_list) {
		if (!vertex_list.containsAll(getVertexes()))
			throw new IllegalArgumentException("Supplied list must include " + "all vertices of the graph");
		this.vertex_ordered_list = vertex_list;
	}

	public void reset() {
		initialize();
	}

	public void initialize() {
		this.graph.initialize();

		if (this.size != null) {
			if (vertex_ordered_list == null){
				setVertexOrder(new ArrayList<Vertex<T>>(getVertexes()));
			}
			double height = getHeight();
			double width = getWidth();

			if (radius <= 0) {
				radius = 0.45 * (height < width? height : width);
			}
			int i = 0;
			for(Vertex<T> v : vertex_ordered_list) {
				Point2D coord = transform(v.getData());
				double angle = (2 * Math.PI * i) / vertex_ordered_list.size();
				coord.setLocation(Math.cos(angle) * radius + width / 2, Math.sin(angle) * radius + height / 2);
				CircleVertexData data = getCircleData(v.getData());
				data.setAngle(angle);
				i++;
			}
		}
		updateGraphLocation();
	}

	protected CircleVertexData getCircleData(T v) {
		return circleVertexDataMap.get(v);
	}

	protected static class CircleVertexData {

		private double angle;

		protected double getAngle() {
			return angle;
		}

		protected void setAngle(double angle) {
			this.angle = angle;
		}
	}
}
