package gui.graph.layout;

import util.graph.Vertex;
import gui.graph.presentation.EdgePresentation;
import gui.graph.presentation.GraphPresentation;
import gui.graph.presentation.VertexPresentation;
import java.awt.Dimension;
import java.awt.geom.Point2D;
import java.util.HashMap;
import java.util.Map;

/**
 * This class implements the Fruchterman-Reingold force-directed algorithm for node layout. 
 * 
 * <p>Behavior is determined by the following settable parameters:
 * <ul>
 * <li/> attraction multiplier: how much edges try to keep their vertexes together
 * <li/> repulsion multiplier: how much vertexes try to push each other apart
 * <li/> maximum iterations: how many iterations this algorithm will use before stopping
 * </ul>
 * Each of the first two defaults to 0.75; the maximum number of iterations defaults to 700.
 * 
 * @see "Fruchterman and Reingold, 'Graph Drawing by Force-directed Placement'"
 * 
 * @author Mohamed Wahbi
 */
public class FRLayout<T extends Comparable<? super T>> extends Layout<T> {

	private final double BORDER = VertexPresentation.WIDTH;
	private double EPSILON = 0.000001D;
	private double forceConstant;
	private double temperature;
	private int currentIteration;
	private int maxIterations = 700;
	private double attraction_multiplier = 0.75;
	private double attraction_constant;
	private double repulsion_multiplier = 0.75;
	private double repulsion_constant;
	private double max_dimension;
	private Map<T, FRPoint2D> frLocations = null;
	
	/**
	 * Creates an instance for the specified graph.
	 */
	public FRLayout(GraphPresentation<T> graph) {
		super(graph);
		this.max_dimension = Math.max(getWidth(), getHeight());
		this.frLocations = new HashMap<T, FRPoint2D>();
		for(Vertex<T> vertex : getVertexes()) {
			this.frLocations.put(vertex.getData(), new FRPoint2D());
		}
	}

	/**
	 * Creates an instance of size {@code d} for the specified graph.
	 */
	public FRLayout(GraphPresentation<T> graph, Dimension size) {
		super(graph, size, new RandomLocations<T>(graph.getVertexes(), size));
		this.max_dimension = Math.max(getWidth(), getHeight());
		this.frLocations = new HashMap<T, FRPoint2D>();
		for(Vertex<T> vertex : getVertexes()) {
			this.frLocations.put(vertex.getData(), new FRPoint2D());
		}
		initialize();
	}

	public void reset() {
		doInit();
		while(!done()) {
			step();
		}
		updateGraphLocation();
	}

	public void initialize() {
		this.graph.initialize();
		doInit();
		while(!done()) {
			step();
		}
		updateGraphLocation();
	}

	@Override
	public void setSize(Dimension size) {
		if (initialized == false){
			setInitializer(new RandomLocations<T>(this.graph.getVertexes(), size));
		}
		super.setSize(size);
		this.max_dimension = Math.max(size.height, size.width);
	}

	/**
	 * Sets the attraction multiplier.
	 */
	public void setAttractionMultiplier(double attraction) {
		this.attraction_multiplier = attraction;
	}

	/**
	 * Sets the repulsion multiplier.
	 */
	public void setRepulsionMultiplier(double repulsion) {
		this.repulsion_multiplier = repulsion;
	}

	private void doInit() {
		GraphPresentation<T> graph = getGraphPresentation();
		if (graph != null) {
			currentIteration = 0;
			temperature = getWidth() / 10;
			forceConstant = 5*Math.sqrt(getHeight() * getWidth() / graph.getVertexCount());
			attraction_constant = attraction_multiplier * forceConstant;
			repulsion_constant = repulsion_multiplier * forceConstant;
		}
	}

	/**
	 * Moves the iteration forward one notch, calculation attraction and
	 * repulsion between vertexes and edges and cooling the temperature.
	 */
	public void step() {
		this.currentIteration++;
		
		/**
		 * Calculate repulsion
		 */
		for(VertexPresentation<T> vertex : getVertexPresentations()) {
			calcRepulsion(vertex);
		}
		/**
		 * Calculate attraction
		 */
		for(EdgePresentation<T> edge : getGraphPresentation().getEdgePresentations()) {
			calcAttraction(edge);
		}
		/**
		 * Calculate positions
		 */
		for(VertexPresentation<T> vertex : getVertexPresentations()) {
			if (!isLocked(vertex.getData())) {
				calcPositions(vertex);
			}
		}
		cool();
	}
	
	protected synchronized void calcPositions(VertexPresentation<T> vertex) {
		T v = vertex.getData();
		FRPoint2D fr = getFRLocation(v);
		if (fr == null)
			return;
		double deltaLength = Math.max(EPSILON, Math.hypot(fr.getX(), fr.getY()));

		double disp_x = fr.getX() / deltaLength * Math.min(deltaLength, temperature);
		assert Double.isNaN(disp_x) == false : "Unexpected mathematical result in TreeLayoutTwo:calcPositions";
		double disp_y = fr.getY() / deltaLength * Math.min(deltaLength, temperature);
		assert Double.isNaN(disp_y) == false : "Unexpected mathematical result in TreeLayoutTwo:calcPositions";
		offsetVertex(v, disp_x, disp_y);

		double borderWidth = getWidth() / BORDER;
		double newX = getX(v);
		if (newX < borderWidth) {
			newX = borderWidth + Math.random() * borderWidth * 2.0;
		} else if (newX > (getWidth() - borderWidth)) {
			newX = getWidth() - borderWidth - Math.random() * borderWidth * 2.0;
		}

		double borderHeight = getHeight() / BORDER;
		double newY = getY(v);
		if (newY < borderHeight) {
			newY = borderHeight + Math.random() * borderHeight * 2.0;
		} else if (newY > (getHeight() - borderHeight)) {
			newY = getHeight() - borderHeight - Math.random() * borderHeight * 2.0;
		}
		newX = Math.max(BORDER, Math.min(newX, getWidth()-BORDER));
		newY = Math.max(BORDER, Math.min(newY, getHeight()-BORDER));		
		setLocation(v, newX, newY);
	}

	protected void calcAttraction(EdgePresentation<T> edge) {
		T v1 = edge.getSource().getData();
		T v2 = edge.getTarget().getData();
		if (isLocked(v1) && isLocked(v2)) { // both locked, do nothing
			return;
		}
		Point2D p1 = transform(v1);
		Point2D p2 = transform(v2);
		if (p1 == null || p2 == null)
			return;
		double xDelta = p1.getX() - p2.getX();
		double yDelta = p1.getY() - p2.getY();
		double deltaLength = Math.max(EPSILON, p1.distanceSq(p2));
		double force = (deltaLength * deltaLength) / attraction_constant;
		assert Double.isNaN(force) == false : "Unexpected mathematical result in TreeLayoutTwo:calcPositions [force]";
		double dx = (xDelta / deltaLength) * force;
		double dy = (yDelta / deltaLength) * force;
		if (!isLocked(v1)) {
			FRPoint2D fr = getFRLocation(v1);
			fr.offset(-dx, -dy);
		}
		if (!isLocked(v2)) {
			FRPoint2D fr = getFRLocation(v2);
			fr.offset(dx, dy);
		}
	}

	protected void calcRepulsion(VertexPresentation<T> vertex) {
		T v = vertex.getData();
		FRPoint2D fr = getFRLocation(v);
		fr.setLocation(0, 0);
		Point2D pv = transform(v);
		for(Vertex<T> u : getVertexes()) {
			if (!v.equals(u.getData())) {
				Point2D pu = transform(u.getData());
				double xDelta = pv.getX() - pu.getX();
				double yDelta = pv.getY() - pu.getY();
				double deltaLength = Math.max(EPSILON, pv.distance(pu));
				double force = (repulsion_constant * repulsion_constant) / deltaLength;				
				double dx = (xDelta / deltaLength) * force;
				double dy = (yDelta / deltaLength) * force;
				fr.offset(dx, dy);
			}
		}
	}
	
	/**
	 * reduce the temperature as the layout approaches a better configuration
	 */
	private void cool() {
		this.temperature *= (1.0 - this.currentIteration / (double) this.maxIterations);
	}
	
	/**
	 * Returns true once the current iteration has passed the maximum count, <tt>maxIterations</tt>.
	 */
	public boolean done() {
		return (this.currentIteration > this.maxIterations || this.temperature < 1.0 / (double) this.max_dimension);
	}
	
	/**
	 * Sets the maximum number of iterations.
	 */
	public void setMaxIterations(int maxIterations) {
		this.maxIterations = maxIterations;
	}
	
	private FRPoint2D getFRLocation(T v) {
		return this.frLocations.get(v);
	}
	
	@SuppressWarnings("serial")
	protected static class FRPoint2D extends Point2D.Double {

		protected void offset(double x, double y) {
			this.x += x;
			this.y += y;
		}
	}
}
