package gui.graph.layout;

import util.graph.AbstractGraph;
import util.graph.Vertex;
import gui.graph.presentation.GraphPresentation;
import gui.graph.presentation.VertexPresentation;
import java.awt.Dimension;
import java.awt.geom.Point2D;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * This class implements the Fruchterman-Reingold force-directed algorithm for node layout. 
 * 
 * <p>Behavior is determined by the following settable parameters:
 * <ul>
 * <li/> attraction multiplier: how much edges try to keep their vertexes together
 * <li/> repulsion multiplier: how much vertexes try to push each other apart
 * <li/> maximum iterations: how many iterations this algorithm will use before stopping
 * </ul>
 * Each of the first two defaults to 0.75; the maximum number of iterations defaults to 700.
 * 
 * @see "Fruchterman and Reingold, 'Graph Drawing by Force-directed Placement'"
 * 
 * @author Mohamed Wahbi
 */
public class KKLayout<T extends Comparable<? super T>> extends Layout<T> {

	private final double BORDER = VertexPresentation.WIDTH;
	private double EPSILON = 0.1D;
	private int currentIteration;
	private int maxIterations = 1000;
	private double L; // the ideal length of an edge
	private double K = 1; // arbitrary cost number
	private Map<T, Map<T, Double>> distanceMatrix = null;
	private boolean adjustForGravity = true;
	private boolean exchangeVertices = true;
	private Map<T, Point2D> kkLocations = null;
	/**
	 * The diameter of the visible graph. In other words, the maximum over all pairs
	 * of vertexes of the length of the shortest path between a and bf the visible graph.
	 */

	/**
	 * A multiplicative factor which partly specifies the "preferred" length of an edge (L).
	 */
	private double length_factor = 0.9;

	/**
	 * A multiplicative factor which specifies the fraction of the graph's diameter to be 
	 * used as the inter-vertex distance between disconnected vertexes.
	 */
	private double disconnected_multiplier = 0.5;

	private List<Vertex<T>> vertexes;
	private int calcDeltaXYIteration = 50;

	/**
	 * Creates an instance for the specified graph.
	 */
	public KKLayout(GraphPresentation<T> graph) {
		super(graph);
		this.vertexes = getVertexes();
		this.distanceMatrix = new HashMap<T, Map<T, Double>>();
		getGraph().FloydWarshall();
		for(Vertex<T> u : this.getVertexes()) {
			Map<T, Double> map = new HashMap<T, Double>();
			for(Vertex<T> v : this.getVertexes()) {
				double distance = getGraph().getShorthestDistance(u, v);
				map.put(v.getData(), distance);
			}
			this.distanceMatrix.put(u.getData(), map);
		}
		this.kkLocations = new HashMap<T, Point2D>();
		for(Vertex<T> vertex : getVertexes()) {
			T v = vertex.getData();
			this.kkLocations.put(v, getCoordinates(v));
		}
	}

	/**
	 * Creates an instance of size {@code d} for the specified graph.
	 */
	public KKLayout(GraphPresentation<T> graph, Dimension size) {
		super(graph, size, new RandomLocations<T>(graph.getVertexes(), size));
		this.vertexes = getVertexes();
		// max_dimension = Math.max(size.height, size.width);
		this.kkLocations = new HashMap<T, Point2D>();
		for(Vertex<T> vertex : getVertexes()) {
			T v = vertex.getData();
			this.kkLocations.put(v, getCoordinates(v));
		}
		initialize();
	}

	/**
	 * Sets a multiplicative factor which 
	 * partly specifies the "preferred" length of an edge (L).
	 */
	public void setLengthFactor(double length_factor) {
		this.length_factor = length_factor;
	}

	/**
	 * Sets a multiplicative factor that specifies the fraction of the graph's diameter to be 
	 * used as the inter-vertex distance between disconnected vertexes.
	 */
	public void setDisconnectedDistanceMultiplier(double disconnected_multiplier) {
		this.disconnected_multiplier = disconnected_multiplier;
	}

	/**
	 * Sets the maximum number of iterations.
	 */
	public void setMaxIterations(int maxIterations) {
		this.maxIterations = maxIterations;
	}

	/**
	 * This one is an incremental visualization.
	 */
	public boolean isIncremental() {
		return true;
	}

	/**
	 * Returns true once the current iteration has passed the maximum count.
	 */
	public boolean done() {
		return (currentIteration > maxIterations);
	}

	private void doInit() {
		currentIteration = 0;
		if (graph != null) {
			int diameter = getGraph().getDiameter(true);
			double L0 = Math.min(getHeight(), getWidth());
			this.L = (L0 / diameter) * length_factor; // length_factor used to be hardcoded to 0.9
			// L = 0.75 * Math.sqrt(height * width / n);
			for(int i = 0 ; i < vertexes.size() - 1 ; i++) {
				T vi = vertexes.get(i).getData();
				for(int j = i + 1 ; j < vertexes.size() ; j++) {
					double dist = diameter * disconnected_multiplier;
					T vj = vertexes.get(j).getData();
					double d_ij = vertexes.get(i).getPathDistance(vertexes.get(j));
					dist = Math.min(d_ij, dist);
					// double d_ji = getGraph().getDistance(vj, vi);
					// dist = Math.min(d_ji, dist);

					distanceMatrix.get(vi).put(vj, dist);
					distanceMatrix.get(vj).put(vi, dist);
				}
			}
		}
	}

	public void initialize() {
		this.graph.initialize();
		doInit();
		while (!done()) {
			step();
		}
		updateGraphLocation();
	}

	public void reset() {
		// currentIteration = 0;
		doInit();
		while (!done()) {
			step();
		}
		updateGraphLocation();
	}

	public void step() {
		currentIteration++;
	//	double energy = calcEnergy();

		int n = vertexes.size();
		if (n == 0)
			return;

		double maxDeltaM = 0;
		T pm = null; // the node having max deltaM
		for(Vertex<T> vertex : vertexes) {
			T v = vertex.getData();
			if (!isLocked(v)) {
				double deltam = calcDeltaM(v);
				if (maxDeltaM < deltam) {
					maxDeltaM = deltam;
					pm = vertex.getData();
				}
			}
		}
		if (pm == null) return;

		for(int i = 0 ; i < calcDeltaXYIteration  ; i++) {
			Point2D dxy = calcDeltaXY(pm);
			offsetKKLocation(pm, dxy.getX(), dxy.getY());
			double deltam = calcDeltaM(pm);
			if (deltam < EPSILON) break;
		}

		if (adjustForGravity)
			adjustForGravity();

		if (exchangeVertices && maxDeltaM < EPSILON) {
			double energy = calcEnergy();
			for(int i = 0 ; i < n - 1 ; i++) {
				T vi = vertexes.get(i).getData();
				if (!isLocked(vi)){
					for(int j = i + 1 ; j < n ; j++) {
						T vj = vertexes.get(j).getData();
						if (!isLocked(vj)) {
							double xenergy = calcEnergyIfExchanged(vi, vj);
							if (energy > xenergy) {
								Point2D pi = getKKLocation(vi);
								Point2D pj = getKKLocation(vj);
								double xi = pi.getX();
								double yi = pi.getY();
								pi.setLocation(pj);
								pj.setLocation(xi, yi);
								return;
							}
						}
					}
				}
			}
		}
	}

	/**
	 * Shift all vertexes so that the center of gravity is located at
	 * the center of the screen.
	 */
	public void adjustForGravity() {
		Dimension d = getSize();
		double height = d.getHeight();
		double width = d.getWidth();
		double gx = 0;
		double gy = 0;
		List<T> vertexes = graph.getVertexes();
		for(T v : vertexes) {
			Point2D p = getKKLocation(v);
			if (p == null) {
				System.out.println(this.graph.getGraph().pretty());
				System.out.println(v + " @ " + vertexes);
			}
			gx += p.getX();
			gy += p.getY();
		}
		gx /= vertexes.size();
		gy /= vertexes.size();
		double diffx = width / 2 - gx;
		double diffy = height / 2 - gy;
		for(T v : vertexes) {
			offsetKKLocation(v, diffx, diffy);
		}
	}

	/*
	 * (non-Javadoc)
	 * @see edu.uci.ics.jung.visualization.layout.AbstractLayout#setSize(java.awt.Dimension)
	 */
	@Override
	public void setSize(Dimension size) {
		if (initialized == false)
			setInitializer(new RandomLocations<T>(this.graph.getVertexes(), size));
		super.setSize(size);
	}

	/**
	 * Enable or disable gravity point adjusting.
	 */
	public void setAdjustForGravity(boolean on) {
		adjustForGravity = on;
	}

	/**
	 * Returns true if gravity point adjusting is enabled.
	 */
	public boolean getAdjustForGravity() {
		return adjustForGravity;
	}

	/**
	 * Enable or disable the local minimum escape technique by
	 * exchanging vertexes.
	 */
	public void setExchangeVertices(boolean on) {
		exchangeVertices = on;
	}

	/**
	 * Returns true if the local minimum escape technique by
	 * exchanging vertexes is enabled.
	 */
	public boolean getExchangeVertices() {
		return exchangeVertices;
	}

	/**
	 * Determines a step to new position of the vertex m.
	 */
	private Point2D calcDeltaXY(T v) {
		double dE_dxv = 0;
		double dE_dyv = 0;
		double d2E_d2xv = 0;
		double d2E_dxy = 0;
		double d2E_dyx = 0;
		double d2E_d2yv = 0;
		for(int i = 0 ; i < vertexes.size() ; i++) {
			T u = vertexes.get(i).getData();
			if (!u.equals(v)) {
				double dist = distanceMatrix.get(v).get(u);
				double l = L * dist;
				double k = K / (dist * dist);
				Point2D pv = getKKLocation(v);
				Point2D pu = getKKLocation(u);
				double dx = pv.getX() - pu.getX();
				double dy = pv.getY() - pu.getY();
				double d = pv.distance(pu);
				double ddd = Math.pow(d, 3);

				dE_dxv += k * (1 - l / d) * dx;
				dE_dyv += k * (1 - l / d) * dy;
				d2E_d2xv += k * (1 - l * dy * dy / ddd);
				d2E_dxy += k * l * dx * dy / ddd;
				d2E_d2yv += k * (1 - l * dx * dx / ddd);
			}
		}
		// d2E_dxy equals to d2E_dyx.
		d2E_dyx = d2E_dxy;
		double denomi = d2E_d2xv * d2E_d2yv - d2E_dxy * d2E_dyx;
		double deltaX = (d2E_dxy * dE_dyv - d2E_d2yv * dE_dxv) / denomi;
		double deltaY = (d2E_dyx * dE_dxv - d2E_d2xv * dE_dyv) / denomi;
		return new Point2D.Double(deltaX, deltaY);
	}

	/**
	 * Calculates the gradient of energy function at the vertex v.
	 */
	private double calcDeltaM(T v) {
		double dEdxm = 0;
		double dEdym = 0;
		for(int i = 0 ; i < vertexes.size() ; i++) {
			T u = vertexes.get(i).getData();
			if (!u.equals(v)) {
				double dist = distanceMatrix.get(v).get(u);
				double l = L * dist;
				double k = K / (dist * dist);
				Point2D pv = getKKLocation(v);
				Point2D pu = getKKLocation(u);
				double dx = pv.getX() - pu.getX();
				double dy = pv.getY() - pu.getY();
				double common = k * (1 - l / pu.distance(pv));
				dEdxm += common * dx;
				dEdym += common * dy;
			}
		}
		return Math.sqrt(dEdxm * dEdxm + dEdym * dEdym);
	}

	/**
	 * Calculates the energy function E.
	 */
	private double calcEnergy() {
		double energy = 0;
		for(int i = 0 ; i < vertexes.size() - 1 ; i++) {
			T vi = vertexes.get(i).getData();
			for(int j = i + 1 ; j < vertexes.size() ; j++) {
				T vj = vertexes.get(j).getData();
				double dist = distanceMatrix.get(vi).get(vj);
				double l = L * dist;
				double k = K / (dist * dist);
				Point2D pi = getKKLocation(vi);
				Point2D pj = getKKLocation(vj);
				energy += k / 2 * (pi.distanceSq(pj) + l * l - 2 * l * pi.distance(pj));
			}
		}
		return energy;
	}

	/**
	 * Calculates the energy function E as if positions of the
	 * specified vertexes are exchanged.
	 */
	private double calcEnergyIfExchanged(T p, T q) {
		double energy = 0; 
		for(int i = 0 ; i < vertexes.size() - 1 ; i++) {
			T vi = vertexes.get(i).getData();
			for(int j = i + 1 ; j < vertexes.size() ; j++) {
				T vj = vertexes.get(j).getData();
				Point2D pi = getKKLocation(vi);
				Point2D pj = getKKLocation(vj);
				if (vi.equals(p)) pi = getKKLocation(q);
				if (vj.equals(q)) pj = getKKLocation(p);
				double dist = distanceMatrix.get(vi).get(vj);
				double l = L * dist;
				double k = K / (dist * dist);
				energy += k / 2 * (pi.distanceSq(pj) + l * l - 2 * l * pi.distance(pj));
			}
		}
		return energy;
	}
	
	private Point2D getKKLocation(T v) {
		return this.kkLocations.get(v);
	}

	private void offsetKKLocation(T v, double xOffset, double yOffset) {
		Point2D coord = this.kkLocations.get(v);
		coord.setLocation(coord.getX() + xOffset, coord.getY() + yOffset);
	}
	
	public AbstractGraph<T> getGraph() {
		return getGraphPresentation().getGraph();
	}
	
	protected void updateGraphLocation() {
		for(Vertex<T> vertex: getVertexes()) {
			T v = vertex.getData();
			Point2D location = getKKLocation(v);

			double borderWidth = getWidth() / BORDER;
			double newX = location.getX();
			if (newX < borderWidth) {
				newX = borderWidth + Math.random() * borderWidth * 2.0;
			} else if (newX > (getWidth() - borderWidth)) {
				newX = getWidth() - borderWidth - Math.random() * borderWidth * 2.0;
			}

			double borderHeight = getHeight() / BORDER;
			double newY = location.getY();
			if (newY < borderHeight) {
				newY = borderHeight + Math.random() * borderHeight * 2.0;
			} else if (newY > (getHeight() - borderHeight)) {
				newY = getHeight() - borderHeight - Math.random() * borderHeight * 2.0;
			}
			newX = Math.max(BORDER, Math.min(newX, getWidth()-BORDER));
			newY = Math.max(BORDER, Math.min(newY, getHeight()-BORDER));		
			setLocation(v, newX, newY);
		}
		super.updateGraphLocation();
	}
}
