package gui.graph.layout;

import util.graph.Edge;
import util.graph.Vertex;
import gui.graph.presentation.EdgePresentation;
import gui.graph.presentation.GraphPresentation;
import gui.graph.presentation.VertexPresentation;
import java.awt.Dimension;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

/**
 * Abstract class for Graph Layouts. It handles some of the basic functions: storing coordinates, 
 * maintaining the dimensions, initializing the locations, maintaining locked vertexes. 
 * 
 * @author Mohamed Wahbi
 * @param <T> the vertex type
 * @param <E> the edge type
 */
public abstract class Layout<T extends Comparable<? super T>> {

	/**
	 * a set of vertexes that should not move in relation to the other vertexes
	 */
	private Vector<T> locked = new Vector<T>();

	protected Dimension size = this.size = new Dimension(600, 400);;
	protected GraphPresentation<T> graph = null;
	protected boolean initialized = false;
	protected Map<T, Point2D> locations = null;//= new HashMap<T, Point2D>();

	/**
	 * Creates an instance which does not initialize the vertex locations.
	 * 
	 */
	public Layout() {
		this.locations = new HashMap<T, Point2D>();
	}
	
	/**
	 * Creates an instance which does not initialize the vertex locations.
	 * 
	 * @param graph the graph presentation for which the layout algorithm is to be created.
	 */
	public Layout(GraphPresentation<T> graph) {
		if (graph == null) {
			throw new IllegalArgumentException("Graph must be non-null");
		}
		this.graph = graph;
		this.locations = new RandomLocations<T>(graph.getVertexes(), size);
	}
	
	/**
	 * Creates an instance which does not initialize the vertex locations.
	 * 
	 * @param graph the graph presentation for which the layout algorithm is to be created.
	 * @param size the preferred size of the layout
	 */
	public Layout(GraphPresentation<T> graph, Dimension size) {
		if (graph == null) {
			throw new IllegalArgumentException("Graph must be non-null");
		}
		this.graph = graph;
		this.size = size;
		this.locations = new RandomLocations<T>(graph.getVertexes(), size);
	}

	/**
	 * Creates an instance which initializes the vertex locations.
	 * 
	 * @param graph the graph presentation for which the layout algorithm is to be created.
	 * @param locations the initial locations. 
	 */
	protected Layout(GraphPresentation<T> graph, Dimension size, Map<T, Point2D> locations) {
		this(graph, size);
		this.locations = locations;
		this.initialized = true;
	}
	

	/**
	 * Initializes fields in the node that may not have been set during the constructor. 
	 * Must be called before the iterations begin. 
	 */
	public abstract void initialize();

	/**
	 * 
	 */
	public abstract void reset();
	
	/**
	 * Sets the graph presentation to the given one
	 * @param graph the graph presentation  
	 */
	public void setGraph(GraphPresentation<T> graph) {
		this.graph = graph;
		if (size != null && graph != null) {
			initialize();
		}
	}

	/**
	 * When a visualization is resized, it presumably wants to fix the locations of the vertexes 
	 * and possibly to reinitialize its data. The current method calls <tt>initializeLocations</tt> 
	 * followed by <tt>initialize_local</tt>.
	 */
	public void setSize(Dimension size) {
		if (size != null && this.graph != null) {
			Dimension oldSize = this.size;
			this.size = size;
			initialize();
			if (oldSize != null) {
				adjustLocations(oldSize, size);
			}
		}
	}

	private void adjustLocations(Dimension oldSize, Dimension size) {
		int xOffset = (size.width - oldSize.width) / 2;
		int yOffset = (size.height - oldSize.height) / 2;
		// now, move each vertex to be at the new screen center
		for(T v : this.getGraphPresentation().getVertexes()) {
			offsetVertex(v, xOffset, yOffset);
		}
	}

	public void setInitializer(Map<T, Point2D> locations) {
		if (this.locations.equals(locations)) {
			throw new IllegalArgumentException("Layout cannot be initialized with itself");
		}
		this.locations = locations;
		initialized = true;
	}

	/**
	 * Returns the current size of the visualization space, according to the last call to resize().
	 * 
	 * @return the current size of the screen
	 */
	public Dimension getSize() {
		return size;
	}

	/**
	 * Returns the Coordinates object that stores the vertex'x and y location.
	 * 
	 * @param vertex A Vertex that is a part of the Graph being visualized.
	 * @return A Coordinates object with x and y locations.
	 */
	public Point2D getCoordinates(T vertex) {
		return locations.get(vertex);
	}

	public Point2D transform(T vertex) {
		return getCoordinates(vertex);
	}

	/**
	 * Returns the x coordinate of the vertex from the Coordinates object.
	 * in most cases you will be better off calling transform(vertex).
	 */
	public double getX(T vertex) {
		assert getCoordinates(vertex) != null : "Cannot getX for an unmapped vertex " + vertex;
		return getCoordinates(vertex).getX();
	}

	/**
	 * Returns the y coordinate of the vertex from the Coordinates object.
	 * In most cases you will be better off calling transform(vertex).
	 */
	public double getY(T vertex) {
		assert getCoordinates(vertex) != null : "Cannot getY for an unmapped vertex " + vertex;
		return getCoordinates(vertex).getY();
	}

	/**
	 * @param vertex
	 * @param xOffset
	 * @param yOffset
	 */
	protected void offsetVertex(T vertex, double xOffset, double yOffset) {
		Point2D c = getCoordinates(vertex);
		setLocation(vertex, c.getX() + xOffset, c.getY() + yOffset);
	}

	/**
	 * Returns for the graph that represents all vertexes.
	 * 
	 * @return the graph that contains all vertexes.
	 */
	public GraphPresentation<T> getGraphPresentation() {
		return this.graph;
	}

	/**
	 * Forcibly moves a vertex to the (x,y) location by setting its x and y locations to the given location. 
	 * Does not add the vertex to the "locked" list, and (in the default implementation) does not make any
	 * adjustments to the rest of the graph.
	 * 
	 * @param vertex the vertex to set its location 
	 * @param x the x location 
	 * @param y the y location
	 */
	public void setLocation(T vertex, double x, double y) {
		Point2D coord = getCoordinates(vertex);
		coord.setLocation(x, y);
	}
	
	/**
	 * 
	 * @param vertex the vertex to set its location 
	 * @param p the (x,y) location 
	 */
	public void setLocation(T vertex, Point2D p) {
		Point2D coord = getCoordinates(vertex);
		coord.setLocation(p);
	}

	/**
	 * Locks {@code vertex} in place if {@code state} is {@code true}, otherwise unlocks it.
	 */
	public void lock(T vertex, boolean state) {
		if (state == true){
			this.locked.add(vertex);
		} else{
			this.locked.remove(vertex);
		}
	}

	/**
	 * Locks all vertexes in place if {@code lock} is {@code true}, otherwise unlocks all vertexes.
	 */
	public void lock(boolean lock) {
		for(T vertex : getGraphPresentation().getVertexes()) {
			lock(vertex, lock);
		}
	}
	
    /**
     * Returns {@code true} if the position of vertex {@code vertex} is locked; {@code false} otherwise.
     * @param vertex the vertex
     */
    public boolean isLocked(T vertex) {
    	return this.locked.contains(vertex);
    } 

	/**
	 * Returns set of vertexes presentations.
	 * 
	 * @return set of vertexes presentations
	 */
	public Vector<VertexPresentation<T>> getVertexPresentations() {
		return this.graph.getVertexPresentations();
	}
	
	/**
	 * Returns set of vertexes.
	 * 
	 * @return set of vertexes
	 */
	public List<Vertex<T>> getVertexes() {
		List<Vertex<T>> vertexes = new ArrayList<Vertex<T>>();
		for(VertexPresentation<T> vertex : this.getVertexPresentations()) {
			vertexes.add(vertex.getVertex());
		}
		return vertexes;
	}
	
	/**
	 * Returns set of edges presentations.
	 * 
	 * @return set of edges presentations
	 */
	public Vector<EdgePresentation<T>> getEdgePresentations() {
		return this.graph.getEdgePresentations();
	}
	
	/**
	 * Returns the set of edges.
	 * 
	 * @return set of edges
	 */
	public Set<Edge<T>> getEdges() {
		return this.graph.getEdges();
	}

	/**
	 * @return the locations
	 */
	public Map<T, Point2D> getLocations() {
		return this.locations;
	}
	
	/**
	 * @param locations the locations to set
	 */
	public void setLocations(Map<T, Point2D> locations) {
		this.locations = locations;
	}
	
	protected void updateGraphLocation() {
		for(VertexPresentation<T> vertex : graph.getVertexPresentations()) {
			vertex.setCenter(transform(vertex.getData()));
		}
	}
	
	protected double getWidth() {
		return this.size.getWidth();
	}
	
	protected double getHeight() {
		return this.size.getHeight();
	}
}
