package gui.graph.presentation;

import util.graph.Edge;
import util.graph.Vertex;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Polygon;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.Toolkit;
import java.awt.font.FontRenderContext;
import java.awt.font.TextLayout;
import java.awt.geom.AffineTransform;
import java.awt.geom.Ellipse2D;
import java.awt.geom.GeneralPath;
import java.awt.geom.Line2D;
import java.awt.geom.PathIterator;
import java.awt.geom.Point2D;
import java.awt.geom.QuadCurve2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 * This class provides a presentation for <tt>Edge<T></tt>.
 * 
 * @author Mohamed Wahbi
 * 
 */
public class EdgePresentation<T extends Comparable<? super T>> {

	private final Edge<T> edge;

	protected final float base = 8; // 10
	protected final float height = 8; // 8
	protected final float notch_height = 4; // 4
	protected VertexPresentation<T> source = null;
	protected VertexPresentation<T> target = null;
	protected boolean isDirected;
	protected boolean isWeighted;
	protected boolean isLabeled;
	protected boolean show = true;
	protected static final int SELECTING_DISTANCE = 5;
	protected static final int ARROW_WIDTH = 3;
	protected static final int ARROW_LENGTH = 10;
	protected static final double ARROW_SIDE = Math.sqrt(ARROW_WIDTH * ARROW_WIDTH + ARROW_LENGTH * ARROW_LENGTH);
	protected final BasicStroke DASHED_STROKE = new BasicStroke(2f, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND, 10.0f, new float[]{4.0f}, 0.0f);
	protected final BasicStroke ROUND_STROKE = new BasicStroke(3.5f, // Line width
			BasicStroke.CAP_ROUND,    // End-cap style
            BasicStroke.JOIN_ROUND); // Vertex join style
	
	/**
	 * Weight Shift
	 */
	public static final int WEIGHT_SHIFT = 2;
	protected int weightWidth;
	protected int weightHeight;
	protected final List<ElementStateListener> listeners;
	protected double control_offset = 10.0d;
	protected boolean select = false;
	protected Color color = Color.black;
	protected Color colorLabel = Color.blue;
	protected Color colorMark = Color.magenta;
	protected Color colorSelection = Color.red;
	protected Color colorSelectionLabel = Color.black;
	private boolean marked = false;

	/**
	 * Creates edge presentation with given start and end points.
	 * 
	 * @param edge edge
	 * @param vertexesMap map from vertexes to their presentations
	 * @param isDirected true if you want directed edge; false otherwise
	 * @param isWeighted true if you want weighted edge; false otherwise
	 */
	public EdgePresentation(Edge<T> edge, Map<T, VertexPresentation<T>> vertexesMap, boolean isDirected, boolean isWeighted) {
		this.edge = edge;
		this.isWeighted = isWeighted;
		this.isDirected = isDirected;
		this.source = vertexesMap.get(edge.getSource().getData());
		this.target = vertexesMap.get(edge.getTarget().getData());
		this.listeners = new ArrayList<ElementStateListener>();
	}
	
	/**
	 * Creates edge presentation with given start and end points.
	 * 
	 * @param edge edge
	 * @param source source vertex presentation
	 * @param target source vertex presentation	 
	 * @param isDirected true if you want directed edge; false otherwise
	 * @param isWeighted true if you want weighted edge; false otherwise
	 */
	public EdgePresentation(Edge<T> edge, VertexPresentation<T> source, VertexPresentation<T> target, boolean isDirected, boolean isWeighted) {
		this.edge = edge;
		this.source = source;
		this.target = target;
		this.isWeighted = isWeighted;
		this.isDirected = isDirected;
		this.listeners = new ArrayList<ElementStateListener>();
	}

	/**
	 * Creates edge presentation with given start and end points.
	 * 
	 * @param edge edge
	 * @param source source vertex presentation
	 * @param target source vertex presentation
	 * @param isDirected true if you want directed edge; false otherwise
	 * @param isWeighted true if you want weighted edge; false otherwise
	 */
	public EdgePresentation(Edge<T> edge, VertexPresentation<T> source, VertexPresentation<T> target, boolean isWeighted) {
		this(edge, source, target, edge.isDirected(), isWeighted);
	}

	/**
	 * Creates a new edge presentation and new edge. 
	 * 
	 * @param source source vertex presentation
	 * @param target source vertex presentation
	 * @param isDirected true if you want directed edge; false otherwise
	 * @param isWeighted true if you want weighted edge; false otherwise
	 */
	public EdgePresentation(VertexPresentation<T> source, VertexPresentation<T> target, boolean isDirected, boolean isWeighted) {
		this.edge = new Edge<T>(source.getVertex(), target.getVertex());
		this.source = source;
		this.target = target;
		this.isWeighted = isWeighted;
		this.isDirected = isDirected;
		this.listeners = new ArrayList<ElementStateListener>();
	}

	/**
	 * Adds the specified ElementStateListaner to receive notification about state change.
	 * 
	 * @param listener
	 */
	public void addElementStateListener(ElementStateListener listener) {
		this.listeners.add(listener);
	}

	/**
	 * Removes the specified ElementStateListaner from this presentation.
	 * 
	 * @param listener
	 */
	public void removeElementStateListener(ElementStateListener listener) {
		this.listeners.remove(listener);
	}

	private void fireStructureChanged() {
		for(ElementStateListener listener : this.listeners) {
			listener.elementStructureChanged();
		}
	}

	/**
	 * <pre>
	 * This becoms very simple if we get into some basic geometry. Let's see...
	 * Ellipse ecuation: a*a/x*x+b*b/y*y=1, where a,b are orizontal/vertical radius.
	 * Line ecuation; x=k*y.
	 * 
	 * The solution of this system is the intersection point:
	 * 
	 * y=(1/k)*sqrt(a*a+k*k*b*b);
	 * x=sqrt(a*a+k*k*b*b);
	 * 
	 * You defined the line as (x1,y1)(x2,y2), the line ecuation is: (x2-x1)/x+(y2-y1)/y=0
	 * </pre>
	 * @param start
	 * @param end
	 * @return
	 */
	protected Point2D calculatePoint2D(Point2D start, Point2D end) {
		double a = source.getShape().getBounds2D().getWidth()/2;
		double b = source.getShape().getBounds2D().getHeight()/2;
//		double a = VertexPresentation.WIDTH / 2;
//		double b = VertexPresentation.HEIGHT / 2;
		double alpha = Math.atan(((start.getY() - end.getY()) * a) / ((start.getX() - end.getX()) * b));
		if ((start.getX() - end.getX()) < 0) {
			alpha += Math.PI;
		}
		double x = end.getX() + a * Math.cos(alpha);
		double y = end.getY() + b * Math.sin(alpha);
		return new Point2D.Double(x, y);
	}

	/**
	 * Returns edge start point.
	 * 
	 * @return edge start point
	 */
	public Point2D getStart() {
		Point2D sourceCenter = this.source.getCenter();
		assert sourceCenter != null : "sourceCenter is null for source:" + source;
		Point2D targetCenter = this.target.getCenter();
		assert targetCenter != null : "targetCenter is null for target:" + target;
		return calculatePoint2D(targetCenter, sourceCenter);
	}

	/**
	 * Returns edge end point.
	 * 
	 * @return edge end point
	 */
	public Point2D getEnd() {
		Point2D sourceCenter = this.source.getCenter();
		Point2D targetCenter = this.target.getCenter();
		return calculatePoint2D(sourceCenter, targetCenter);
	}

	/**
	 * Returns model edge.
	 * 
	 * @return edge
	 */
	public Edge<T> getEdge() {
		return this.edge;
	}

	/**
	 * Paints this edge using given graphics
	 * 
	 * @param gr graphics
	 * @param bgColor background color
	 */
	public void paint(Graphics2D gr, Color bgColor) {
		Stroke oldStroke = gr.getStroke();
		AffineTransform at = AffineTransform.getTranslateInstance(source.getX(), source.getY());
		double dx = target.getX() - source.getX();
		double dy = target.getY() - source.getY();
		double thetaRadians = (float) Math.atan2(dy, dx);
		at.rotate(thetaRadians);
		double dist = Math.sqrt(dx * dx + dy * dy);
		at.scale(dist, 1.0);
		double controlY = 2 * control_offset;
		Shape edge = new QuadCurve2D.Double(0.0f, 0.0f, 0.5f, controlY, 1.0f, 0.0f);
		Shape control = new Rectangle2D.Double(0.5f, controlY, 0, 0);
		edge = at.createTransformedShape(edge);
		control = at.createTransformedShape(control);
		if(marked) {
			gr.setColor(colorMark);
			gr.setStroke(ROUND_STROKE);
			gr.draw(edge);
		}
		if(select) {
			gr.setColor(colorSelection);
		} else {
			gr.setColor(color);
		}
		gr.setStroke(new BasicStroke(1.2f));
		gr.draw(edge);
		gr.setStroke(oldStroke);
		//gr.setStroke(actualStroke);
		if (this.isDirected) {
			double a = VertexPresentation.WIDTH / 2;
			double b = VertexPresentation.HEIGHT / 2;
			Shape vertex = new Ellipse2D.Double(target.getX() - a, target.getY() - b, 2 * a, 2 * b);
			drawArrow(gr, edge, vertex, thetaRadians);
		}
		if(isWeighted || isLabeled) {
			drawLabelAndWeight(gr, control, bgColor);
		}
	}

	protected void drawArrow(Graphics2D gr, Shape edge, Shape vertex, double thetaRadians) {
		AffineTransform at = new AffineTransform();
		GeneralPath path = new GeneralPath(edge);
		float[] seg = new float[6];
		Point2D p1 = null;
		Point2D p2 = null;
		for(PathIterator i = path.getPathIterator(null, 1) ; !i.isDone() ; i.next()) {
			int ret = i.currentSegment(seg);
			if (ret == PathIterator.SEG_MOVETO) {
				p2 = new Point2D.Float(seg[0], seg[1]);
			} else if (ret == PathIterator.SEG_LINETO) {
				p1 = p2;
				p2 = new Point2D.Float(seg[0], seg[1]);
				if (vertex.contains(p2)) {
					at = getArrowTransform(new Line2D.Float(p1, p2), vertex);
					break;
				}
			}
		}
		at.rotate(-thetaRadians);
		Shape arrow = getNotchedArrow(base, height, notch_height);
		arrow = at.createTransformedShape(arrow);
		if(select) {
			gr.setColor(colorSelection);
		} else {
			gr.setColor(color);
		}
		gr.fill(arrow);
		gr.draw(arrow);
	}

	public AffineTransform getArrowTransform(Line2D edgeShape, Shape vertexShape) {
		float dx = (float) (edgeShape.getX1() - edgeShape.getX2());
		float dy = (float) (edgeShape.getY1() - edgeShape.getY2());
		// iterate over the line until the edge shape will place the
		// arrowhead closer than 'arrowGap' to the vertex shape boundary
		while ((dx * dx + dy * dy) > 1.0) {
			try {
				edgeShape = getLastOutsideSegment(edgeShape, vertexShape);
			} catch (IllegalArgumentException e) {
				System.err.println(e.toString());
				return null;
			}
			dx = (float) (edgeShape.getX1() - edgeShape.getX2());
			dy = (float) (edgeShape.getY1() - edgeShape.getY2());
		}
		double atheta = Math.atan2(dx, dy) + Math.PI / 2;
		AffineTransform at = AffineTransform.getTranslateInstance(edgeShape.getX1(), edgeShape.getY1());
		at.rotate(-atheta);
		return at;
	}

	protected Line2D getLastOutsideSegment(Line2D line, Shape shape) {
		if (shape.contains(line.getP2()) == false) {
			String errorString = "line end point: " + line.getP2() + " is not contained in shape: " + shape.getBounds2D();
			throw new IllegalArgumentException(errorString);
			// return null;
		}
		Line2D left = new Line2D.Double();
		Line2D right = new Line2D.Double();
		// subdivide the line until its left segment intersects
		// the shape boundary
		do {
			subdivide(line, left, right);
			line = right;
		} while (shape.contains(line.getP1()) == false);
		// now that right is completely inside shape,
		// return left, which must be partially outside
		return left;
	}

	/**
	 * divide a Line2D into 2 new Line2Ds that are returned
	 * in the passed left and right instances, if non-null
	 * @param src the line to divide
	 * @param left the left side, or null
	 * @param right the right side, or null
	 */
	protected void subdivide(Line2D src, Line2D left, Line2D right) {
		double x1 = src.getX1();
		double y1 = src.getY1();
		double x2 = src.getX2();
		double y2 = src.getY2();
		double mx = x1 + (x2 - x1) / 2.0;
		double my = y1 + (y2 - y1) / 2.0;
		if (left != null) {
			left.setLine(x1, y1, mx, my);
		}
		if (right != null) {
			right.setLine(mx, my, x2, y2);
		}
	}

	protected void drawArrow(Graphics gr, Point2D start, Point2D end) {
		double a = VertexPresentation.WIDTH / 2;
		double b = VertexPresentation.HEIGHT / 2;
		double alpha = Math.atan(((start.getY() - end.getY()) * a) / ((start.getX() - end.getX()) * b));
		if ((start.getX() - end.getX()) < 0) {
			alpha += Math.PI;
		}
		if (start.equals(end)) {
			alpha = 0;
		}
		Polygon triangle = new Polygon();
		double x = end.getX() + a * Math.cos(alpha);
		double y = end.getY() + b * Math.sin(alpha);
		triangle.addPoint((int) Math.round(x), (int) Math.round(y));
		double beta = Math.atan((double) ARROW_WIDTH / ARROW_LENGTH);
		double x1 = x + ARROW_SIDE * Math.cos(beta + alpha);
		double y1 = y + ARROW_SIDE * Math.sin(beta + alpha);
		triangle.addPoint((int) Math.round(x1), (int) Math.round(y1));
		double x2 = x + ARROW_SIDE * Math.cos(alpha - beta);
		double y2 = y + ARROW_SIDE * Math.sin(alpha - beta);
		triangle.addPoint((int) Math.round(x2), (int) Math.round(y2));
		gr.setColor(Color.GREEN);
		gr.drawPolygon(triangle);
		gr.fillPolygon(triangle);
	}

	private void drawLabelAndWeight(Graphics2D gr, Shape control, Color backgroundColor) {
		String label = "";
		if (isLabeled && isWeighted) {
			label = getLabel() + "=" + Double.toString(this.getWeight());
		} else if (isLabeled) {
			label = getLabel();
		} else if (isWeighted) {
			label = Double.toString(this.getWeight());
		} else {
			return;
		}
		double control_x = control.getBounds2D().getCenterX();
		double control_y = control.getBounds2D().getCenterY();
		double qx1 = (control_x + source.getX()) / 2;
		double qy1 = (control_y + source.getY()) / 2;
		double qx2 = (control_x + target.getX()) / 2;
		double qy2 = (control_y + target.getY()) / 2;
		double x = (qx1 + qx2) / 2;
		double y = (qy1 + qy2) / 2;
		this.drawLabelAndWeight(gr, label, x, y, backgroundColor);
	}

	protected void drawLabelAndWeight(Graphics2D gr, Point2D start, Point2D end, Color backgroundColor) {
		String label = "";
		if (isLabeled && isWeighted) {
			label = getLabel() + "=" + Double.toString(this.getWeight());
		} else if (isLabeled) {
			label = getLabel();
		} else if (isWeighted) {
			label = Double.toString(this.getWeight());
		} else {
			return;
		}
		double x = (start.getX() + end.getX()) / 2;
		double y = (start.getY() + end.getY()) / 2;
		this.drawLabelAndWeight(gr, label, x, y, backgroundColor);
	}
	
	protected void drawLabelAndWeight(Graphics2D gr, String label, double x, double y, Color backgroundColor) {
//		this.weightWidth = gr.getFontMetrics().stringWidth(label);
//		this.weightHeight = gr.getFontMetrics().getMaxAscent() - gr.getFontMetrics().getMaxDescent();
//		double xx = x - (this.weightWidth - 1) / 2;
//		double yy = y - (this.weightHeight - 1) / 2;
//		gr.setColor(backgroundColor);
//		Shape rectangle = new Rectangle2D.Double(xx - WEIGHT_SHIFT / 2, yy - WEIGHT_SHIFT, this.weightWidth + WEIGHT_SHIFT, this.weightHeight + WEIGHT_SHIFT);
//		gr.fill(rectangle);
//		if(select) {
//			gr.setColor(colorSelectionLabel);
//		} else {
//			gr.setColor(colorLabel);
//		}
		//gr.drawString(label, (int) x, (int) y + this.weightHeight - 1);
		
		Font font = gr.getFont();
		int screenRes = Toolkit.getDefaultToolkit().getScreenResolution();
		int fontSize = (int) Math.round(screenRes / 6.5);
		FontRenderContext context = gr.getFontRenderContext();
		// Create TextLayout
		TextLayout layout = new TextLayout(label, new Font(font.getName(), font.getStyle(), fontSize), context);
		Rectangle2D bounds = layout.getBounds();
		// Position of origin to draw the TextLayout.
		x -= bounds.getWidth() / 2;
		y += bounds.getHeight() / 2;
		Shape rectangle = new Rectangle2D.Double(x + bounds.getX(), y + bounds.getY(), bounds.getWidth(), bounds.getHeight());
		gr.setColor(backgroundColor);
		gr.fill(rectangle);
		if(select) {
			gr.setColor(colorSelectionLabel);
		} else {
			gr.setColor(colorLabel);
		}
		layout.draw(gr, (float) x, (float) y);
	}

	/**
	 * Sets weight of this presentation.
	 * 
	 * @param weight
	 */
	public void setWeight(double weight) {
		this.edge.setWeight(weight);
		fireStructureChanged();
	}

	/**
	 * Returns this presentation weight.
	 * 
	 * @return weight
	 */
	public double getWeight() {
		return this.edge.getWeight();
	}

	/**
	 * Returns this presentation label.
	 * 
	 * @return string the label
	 */
	public String getLabel() {
		return this.edge.getLabel();
	}

	/**
	 * Return upper left point of weight rect.
	 * 
	 * @return upper left point of weight rect
	 */
	public Point2D getUpperLeftWeightPoint() {
		if (!this.isWeighted) {
			return null;
		}
		double x = (getStart().getX() + getEnd().getX()) / 2;
		double y = (getStart().getY() + getEnd().getY()) / 2;
		return new Point2D.Double(x - WEIGHT_SHIFT / 2, y - this.weightHeight / 2 - WEIGHT_SHIFT / 2);
	}

	/**
	 * Checks whether this presentation contains the given point.
	 * 
	 * @param point 
	 * @return true if it does; false otherwise
	 */
	public boolean containsPoint(Point2D point) {
		Point2D start = getStart();
		Point2D end = getEnd();
		double scalarProdWithStart =
				scalarProduct(point.getX() - start.getX(), point.getY() - start.getY(), end.getX() - start.getX(), end.getY() - start.getY());
		double scalarProdWithEnd =
				scalarProduct(point.getX() - end.getX(), point.getY() - end.getY(), end.getX() - start.getX(), end.getY() - start.getY());
		if ((scalarProdWithStart >= 0) && (scalarProdWithEnd <= 0)) {
			return (distanceToEdge(point, start, end) <= SELECTING_DISTANCE);
		}
		return false;
	}

	/**
	 * Verifies whether this presentation lays whithin giver rectangle
	 * 
	 * 
	 * @param x1 left upper x
	 * @param y1 left upper y
	 * @param x2 right lower x
	 * @param y2 right lower y
	 * @return true if it does; false otherwise
	 */
	public boolean isInRect(double x1, double y1, double x2, double y2) {
		return isPointInRect(getStart(), x1, y1, x2, y2) && isPointInRect(getEnd(), x1, y1, x2, y2);
	}

	private boolean isPointInRect(Point2D point, double x1, double y1, double x2, double y2) {
		return (Math.min(x1, x2) <= point.getX()) && (Math.max(x1, x2) >= point.getX()) && (Math.min(y1, y2) <= point.getY())
				&& (Math.max(y1, y2) >= point.getY());
	}

	/**
	 * Checks whether given point is in this presentation weight rectangle.
	 * 
	 * @param point
	 * @return true if it is; false otherwise
	 */
	public boolean isWeightSelected(Point2D point) {
		double x = (getStart().getX() + getEnd().getX()) / 2;
		double y = (getStart().getY() + getEnd().getY()) / 2;
		return (isWeighted && (x - WEIGHT_SHIFT / 2) <= point.getX()) && (point.getX() <= (x + weightWidth + WEIGHT_SHIFT))
				&& ((y - weightHeight - WEIGHT_SHIFT / 2) <= point.getY()) && (point.getY() <= (y + weightHeight + WEIGHT_SHIFT));
	}

	private double scalarProduct(double x1, double y1, double x2, double y2) {
		return x1 * x2 + y1 * y2;
	}

	private double distanceToEdge(Point2D point, Point2D p1, Point2D p2) {
		double result =
				(p1.getY() - p2.getY()) * point.getX() + (p2.getX() - p1.getX()) * point.getY() + (p1.getX() * p2.getY() - p2.getX() * p1.getY());
		result /= Math.sqrt((p1.getX() - p2.getX()) * (p1.getX() - p2.getX()) + (p1.getY() - p2.getY()) * (p1.getY() - p2.getY()));
		return Math.round(Math.abs(result));
	}

	/**
	 * Returns a target vertex of this edge.
	 * 
	 * @return target vertex
	 */
	public VertexPresentation<T> getTargetPresentation() {
		return this.target;
	}

	/**
	 * Returns a source vertex of this edge.
	 * 
	 * @return source vertex
	 */
	public VertexPresentation<T> getSourcePresentation() {
		return this.source;
	}

	/**
	 * Returns a target vertex of this edge.
	 * 
	 * @return target vertex
	 */
	public Vertex<T> getTarget() {
		return this.edge.getTarget();
	}

	/**
	 * Returns a source vertex of this edge.
	 * 
	 * @return source vertex
	 */
	public Vertex<T> getSource() {
		return this.edge.getSource();
	}

	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		// StringBuffer sb = new StringBuffer();
		// sb.append("(");
		return "[" + this.source.toString() + ", " + this.target.toString() + "]";
	}

	/**
	 * Returns an arrowhead in the shape of a simple isosceles triangle
	 * with the specified base and height measurements.  It is placed
	 * with the vertical axis along the negative x-axis, with its base
	 * centered on (0,0).
	 */
	public static GeneralPath getWedgeArrow(float base, float height) {
		GeneralPath arrow = new GeneralPath();
		arrow.moveTo(0, 0);
		arrow.lineTo(-height, base / 2.0f);
		arrow.lineTo(-height, -base / 2.0f);
		arrow.lineTo(0, 0);
		return arrow;
	}

	/**
	 * Returns an arrowhead in the shape of an isosceles triangle
	 * with an isoceles-triangle notch taken out of the base,
	 * with the specified base and height measurements.  It is placed
	 * with the vertical axis along the negative x-axis, with its base
	 * centered on (0,0).
	 */
	public static GeneralPath getNotchedArrow(float base, float height, float notch_height) {
		GeneralPath arrow = new GeneralPath();
		arrow.moveTo(0, 0);
		arrow.lineTo(-height, base / 2.0f);
		arrow.lineTo(-(height - notch_height), 0);
		arrow.lineTo(-height, -base / 2.0f);
		arrow.lineTo(0, 0);
		return arrow;
	}


	/**
	 * Returns an arrowhead in the shape of an isosceles triangle
	 * with an isoceles-triangle notch taken out of the base,
	 * with the specified base and height measurements.  It is placed
	 * with the vertical axis along the negative x-axis, with its base
	 * centered on (0,0).
	 */
	public static GeneralPath getNotchedArrow(int x, int y, float base, float height, float notch_height) {
		GeneralPath arrow = new GeneralPath(x, y);
		arrow.moveTo(x, y);
		arrow.lineTo(-height, base / 2.0f);
		arrow.lineTo(-(height - notch_height), 0);
		arrow.lineTo(-height, -base / 2.0f);
		arrow.lineTo(0, 0);
		return arrow;
	}

	public void setWeighted(boolean weighted) {
		this.isWeighted = weighted;
	}

	public void setLabeled(boolean labeled) {
		this.isLabeled = labeled;
	}

	public void select() {
		this.select  = true;
	}

	public void unselect() {
		this.select = false;
	}

	public boolean isSelected() {
		return select;
	}

	public void show(boolean show) {
		this.show = show;
	}
	
	/**
	 * Marks this presentation if marked is true; otherwise unmark it.
	 * @param markColor 
	 *
	 */
	public void setMarked(boolean marked) {
		this.source.setMarked(marked);
		this.target.setMarked(marked);
		this.marked  = marked;
	}
	
}
