package gui.graph.presentation;

import util.graph.Vertex;
import gui.graph.control.GraphElementSelection;
import gui.graph.control.MouseTool;
import gui.graph.control.PopupMenuTool;
import gui.graph.control.VisualStateListener;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.util.ArrayList;
import javax.swing.JComponent;
import javax.swing.JPopupMenu;

/**
 * A component for graph viewing and editing.
 * 
 * @author Mohamed Wahbi
 * 
 */
@SuppressWarnings("serial")
public class GraphComponent<T extends Comparable<? super T>> extends JComponent {
	
	private GraphPresentation<T> graphPresentation;
	private GraphElementSelection<T> graphElementSelection;
	private final MouseTool<T> mouseTool;
	private final JPopupMenu popupMenu;
	private boolean pickingMode = true;

	private final VisualStateListener selectionStateListener = new VisualStateListener() {
		public void visualChangesCancelled() {
			makeVisualChanges();
		}
		public void visualStateChanged() {
			makeVisualChanges();
		}
	};
	
	/**
	 * Creates component without graph to present.
	 */
	public GraphComponent() {
		this(new Dimension(600, 400));
	}
	
	/**
	 * Creates component without graph to present.
	 *
	 */
	public GraphComponent(Dimension size) {
		super();
		this.setSize(size);
		this.setPreferredSize(size);
		this.setBackground(Color.white);
		this.mouseTool = new MouseTool<T>(this);
		addMouseListener(mouseTool);
		addMouseMotionListener(mouseTool);
		addMouseWheelListener(mouseTool);
		this.popupMenu = new PopupMenuTool<T>(this);
		setComponentPopupMenu(popupMenu);
	}
	
	/**
	 * Creates component with given presentation.
	 * 
	 * @param graph graph presentation
	 */
	public GraphComponent(GraphPresentation<T> graph, Dimension size) {
		this(size);
		this.setGraphPresentation(graph);
	}
	
	/**
	 * Sets new graph presentation to this component
	 * 
	 * @param presentation new graph presentation
	 */
	public void setGraphPresentation(GraphPresentation<T> presentation) {
		this.graphPresentation = presentation;
		graphElementSelection = new GraphElementSelection<T>(presentation);
		graphElementSelection.addVisualStateListener(selectionStateListener);
		graphElementSelection.setGraphPresentation(graphPresentation);
		makeVisualChanges();
	}

	public void makeVisualChanges() {
		setPreferredSize(graphPresentation.getPreferredSize());
		revalidate();
		repaint();
	}
	
	/**
	 * Returns graph presentation.
	 * 
	 * @return graph presentation
	 */
	public GraphPresentation<T> getGraphPresentation() {
		return this.graphPresentation;
	}
	
	/**
	 * Adds temporary component to this component.
	 * 
	 * @param component the component to add.
	 */
	public void addTempComponent(Component component) {
		add(component);
		revalidate();
		repaint();
	}

	/**
	 * Removes temporary component from this component.
	 * 
	 * @param component the component to remove
	 */
	public void removeTempComponent(Component component) {
		remove(component);
		revalidate();
		repaint();
	}

	@Override
	protected void paintComponent(Graphics g) {
		g.setColor(Color.white);
		g.fillRect(0, 0, getWidth(), getHeight());
		Graphics2D gr = (Graphics2D) g;
		gr.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		gr.setRenderingHint(RenderingHints.KEY_COLOR_RENDERING, RenderingHints.VALUE_COLOR_RENDER_QUALITY);
		if (this.graphPresentation != null) {
			this.graphPresentation.paint(gr, getBackground());
			this.graphElementSelection.select();
		}
		this.mouseTool.paint(gr, getBackground());
	}

	public void scale(float scale) {
		zoom(scale);
	}

	public void zoom(double scale) {
		Dimension size = getSize();
		size.width *= scale;
		size.height *= scale;
		this.graphPresentation.scale(scale);
		makeVisualChanges();
		//setPreferredSize(size);
		setSize(size);
		revalidate();
		repaint();
	}

	/**
	 * Returns graph element selection.
	 * 
	 * @return graph element selection
	 */
	public GraphElementSelection<T> getGraphElementSelection() {
		return graphElementSelection;
	}

	public boolean isInPickingMode() {
		return pickingMode;
	}

	public void setPickingMode(boolean pickingMode) {
		this.pickingMode = pickingMode;
	}

	/**
	 * Marks presentations of given vertexes.
	 * 
	 * @param path to mark
	 */
	public void markPath(ArrayList<Vertex<T>> path) {
		graphPresentation.markPath(path);
		makeVisualChanges();
	}

	/**
	 * Unmarks presentations of given vertexes.
	 * 
	 * @param path to mark
	 */
	public void unmark() {
		graphPresentation.unmark();
		makeVisualChanges();
	}
}
