package gui.graph.presentation;

import gui.graph.layout.CircleLayout;
import gui.graph.layout.FRLayout;
import gui.graph.layout.KKLayout;
import gui.graph.layout.Layout;
import gui.tree.presentation.TreePanel.Orientation;
import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GridBagLayout;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.lang.reflect.Constructor;
import java.util.ArrayList;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.DefaultListCellRenderer;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import util.graph.AbstractGraph;
import util.graph.UndirectedGraph;
import util.graph.Vertex;

@SuppressWarnings({"serial", "unchecked"})
public class GraphPanel<T extends Comparable<? super T>> extends JPanel {

	 
	private final int controlHeight = 100;
	private Orientation orientation = Orientation.HORIZONTAL;
	private final static int CONTROL_WIDTH = 185;
	private final static int CONTROL_HEIGHT = 100;
	private GraphComponent<T> graphComponent;
	private JComboBox layoutComboBox = null;
	Class<T>[] layoutClasses = new Class[] {FRLayout.class, KKLayout.class, CircleLayout.class};
	private JComboBox mouseModeBox = null;
	private JPanel controls;

	/**
	 * Creates component with given presentation.
	 * 
	 * @param graph graph presentation
	 */
	public GraphPanel() {
		this(new UndirectedGraph<T>(), new Dimension(600, 500), Orientation.HORIZONTAL);
	}

	/**
	 * Creates component with given presentation.
	 * 
	 * @param graph graph presentation
	 */
	public GraphPanel(AbstractGraph<T> graph, Dimension size, Orientation orientation) {
		this(new GraphPresentation<T>(graph, graph.isDirected(), false), size, orientation);
	}

	/**
	 * Creates component with given presentation.
	 * 
	 * @param graph graph presentation
	 */
	public GraphPanel(GraphPresentation<T> graph, Dimension size, Orientation orientation) {
		super(new BorderLayout());
		this.orientation = orientation;
		setPreferredSize(size);
		Dimension graphComponentSize = new Dimension(size);
		Dimension controlPanelSize = new Dimension(size);
		setPreferredSize(size);
		if (this.orientation == Orientation.VERTICAL) {
			graphComponentSize.setSize(size.getWidth(), size.getHeight() - CONTROL_HEIGHT);
			controlPanelSize.setSize(size.getWidth(), CONTROL_HEIGHT);
		} else {
			graphComponentSize.setSize(size.getWidth() - CONTROL_WIDTH, size.getWidth());
			controlPanelSize.setSize(CONTROL_WIDTH, size.getWidth());
		}
		this.graphComponent = new GraphComponent<T>(graph, graphComponentSize);
		setGraphPresentation(graph);
		createGraphCanvas();
		createControlPanel(controlPanelSize);
	}

	/**
	 * Sets new graph presentation to this component
	 * 
	 * @param presentation new graph presentation
	 */
	public void setGraphPresentation(GraphPresentation<T> presentation) {
		this.graphComponent.setGraphPresentation(presentation);
		makeVisualChanges();
	}

	private void makeVisualChanges() {
		Dimension graphComponentSize = getGraphPresentation().getPreferredSize();
		Dimension size = new Dimension(graphComponentSize.width, graphComponentSize.height + controlHeight);
		setPreferredSize(size);
		revalidate();
		repaint();
	}

	public JComboBox getLayoutsComboBox() {
		if (layoutComboBox == null) {
			layoutComboBox = new JComboBox(layoutClasses);
			layoutComboBox.setRenderer(new DefaultListCellRenderer() {

				public Component getListCellRendererComponent(JList list, Object value, int index, boolean isSelected, boolean cellHasFocus) {
					String valueString = value.toString();
					valueString = valueString.substring(valueString.lastIndexOf('.') + 1);
					return super.getListCellRendererComponent(list, valueString, index, isSelected, cellHasFocus);
				}
			});
			layoutComboBox.addItemListener(new ItemListener() {

				public void itemStateChanged(ItemEvent e) {
					if (e.getStateChange() == ItemEvent.SELECTED) {
						Class<T> clazz = (Class<T>) e.getItem();
						try {
							Layout<T> layout = getLayoutFor(clazz, getGraphPresentation());
							layout.setSize(graphComponent.getSize());
							setGraphLayout(layout);
						} catch (Exception ex) {
							ex.printStackTrace();
						}
					}
				}
			});
			layoutComboBox.setSelectedItem(layoutClasses[0]);
			try {
				setGraphLayout(getLayoutFor(layoutClasses[0], getGraphPresentation()));
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		return layoutComboBox;
	}

	/**
	 * @return Returns the modeBox.
	 */
	public JComboBox getMouseModeComboBox() {
		if (mouseModeBox == null) {
			mouseModeBox = new JComboBox(new String[] {"PICKING", "TRANSFORMING",});
			mouseModeBox.addItemListener(new ItemListener() {

				public void itemStateChanged(ItemEvent paramItemEvent) {
					if (mouseModeBox.getSelectedItem().equals("PICKING")) {
						graphComponent.setPickingMode(true);
						graphComponent.getGraphElementSelection().unselect();
					} else {
						graphComponent.setPickingMode(false);
						Dimension size = graphComponent.getPreferredSize();
						Rectangle rect = new Rectangle(0, 0, size.width, size.height);
						graphComponent.getGraphElementSelection().selectInRect(rect);
					}
				}
			});
		}
		mouseModeBox.setSelectedItem("PICKING");
		return mouseModeBox;
	}

	private void setGraphLayout(Layout<T> layout) {
		layout.initialize();
		revalidate();
		repaint();
	}

	private Layout<T> getLayoutFor(Class<T> layoutClass, GraphPresentation<T> graph) throws Exception {
		Object[] args = new Object[] {graph};
		Constructor<T> constructor = layoutClass.getConstructor(new Class[] {GraphPresentation.class});
		return ((Layout<T>) constructor.newInstance(args));
	}

	private void createGraphCanvas() {
		JScrollPane centerPanel = new JScrollPane(graphComponent);
		centerPanel.setBorder(BorderFactory.createEtchedBorder());
		add(centerPanel, BorderLayout.CENTER);
	}

	private void createControlPanel(Dimension size) {
		controls = new JPanel();
		controls.setPreferredSize(size);
		controls.setBorder(BorderFactory.createEtchedBorder());
		if (orientation == Orientation.VERTICAL) {
			controls.setLayout(new BoxLayout(controls, BoxLayout.X_AXIS));
		} else {
			controls.setLayout(new BoxLayout(controls, BoxLayout.Y_AXIS));
		}
		// controls.add(Box.createHorizontalGlue());
		JButton plus = new JButton("+");
		plus.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				graphComponent.scale(1.1f);
			}
		});
		JButton minus = new JButton("-");
		minus.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				graphComponent.scale(1 / 1.1f);
			}
		});
		JPanel zoomControls = new JPanel();
		//zoomControls.setLayout(new BoxLayout(zoomControls, BoxLayout.X_AXIS));
		zoomControls.setBorder(BorderFactory.createTitledBorder("Zoom"));
		zoomControls.add(plus);
		zoomControls.add(minus);
		controls.add(zoomControls);
		// controls.add(Box.createHorizontalGlue());

		JPanel modeControls = new JPanel();
		// modeControls.setLayout(new BoxLayout(modeControls, BoxLayout.X_AXIS));
		modeControls.setBorder(BorderFactory.createTitledBorder("MouseTool Mode"));
		modeControls.add(getMouseModeComboBox());
		controls.add(modeControls);
		// controls.add(Box.createHorizontalGlue());

		JPanel layoutPanel = new JPanel();
		layoutPanel.setBorder(BorderFactory.createTitledBorder("Graph layout"));
		layoutPanel.add(getLayoutsComboBox());
		controls.add(layoutPanel);
		// controls.add(Box.createRigidArea(new Dimension(10, 0)));
		//controls.add(Box.createHorizontalGlue());

		JPanel edgeConfig = new JPanel(new GridBagLayout());
		//edgeConfig.setLayout(new BoxLayout(edgeConfig, BoxLayout.X_AXIS));
		edgeConfig.setBorder(BorderFactory.createTitledBorder("Constraints"));
		final JRadioButton labeledEdge = new JRadioButton("Labeled", false);
		labeledEdge.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				getGraphPresentation().setLabeledEdges(labeledEdge.isSelected());
				graphComponent.revalidate();
				graphComponent.repaint();
			}
		});
		boolean isWeighted = graphComponent.getGraphPresentation().isWeighted();
		final JRadioButton weightedEdge = new JRadioButton("Weighted", isWeighted );
		edgeConfig.add(weightedEdge);
		weightedEdge.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				getGraphPresentation().setWeightedEdges(weightedEdge.isSelected());
				graphComponent.revalidate();
				graphComponent.repaint();
			}
		});
		edgeConfig.add(labeledEdge);
		controls.add(edgeConfig);
		// controls.add(Box.createHorizontalGlue());

		JPanel vertexConfig = new JPanel(new GridBagLayout());
		//vertexConfig.setLayout(new BoxLayout(vertexConfig, BoxLayout.Y_AXIS));
		vertexConfig.setBorder(BorderFactory.createTitledBorder("Variables"));
		final JRadioButton scalableVertex = new JRadioButton("Scalable", false);
		scalableVertex.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				getGraphPresentation().setScalableVertexes(scalableVertex.isSelected());
				graphComponent.revalidate();
				graphComponent.repaint();
			}
		});
		vertexConfig.add(scalableVertex);
		final JRadioButton labeledVertex = new JRadioButton("Labeled", true);
		labeledVertex.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				getGraphPresentation().setLabeledVertexes(labeledVertex.isSelected());
				graphComponent.revalidate();
				graphComponent.repaint();
			}
		});
		vertexConfig.add(labeledVertex);
		getGraphPresentation().setLabeledVertexes(labeledVertex.isSelected());
		controls.add(vertexConfig);
		
//		JPanel computationsPanel = new JPanel();
//		vertexConfig.setBorder(BorderFactory.createTitledBorder("Computations"));
//		//computationsPanel.setLayout(new BoxLayout(computationsPanel, BoxLayout.Y_AXIS));
//		final JRadioButton diameter = new JRadioButton("Diameter", false);
//		diameter.addActionListener(new ActionListener() {
//
//			public void actionPerformed(ActionEvent e) {
//				boolean marked = diameter.isSelected();
//				ArrayList<Vertex<T>> path = graphComponent.getGraphPresentation().getGraph().getDiameterPath(true);
//				marksPath(path, marked);
//			}
//		});
//		computationsPanel.add(diameter);
//
//		final JRadioButton print = new JRadioButton("Print", false);
//		print.addActionListener(new ActionListener() {
//
//			public void actionPerformed(ActionEvent e) {
//				if (print.isSelected()) {
//					String graph = graphComponent.getGraphPresentation().getGraph().pretty();
//					System.out.println(graph);
//				}
//			}
//		});
//		computationsPanel.add(print);
//		controls.add(computationsPanel);
		
		// controls.add(Box.createHorizontalGlue());
		JPanel southPanel = new JPanel();
		southPanel.setLayout(new BoxLayout(southPanel, BoxLayout.X_AXIS));
		southPanel.add(Box.createVerticalStrut(10));
		southPanel.add(controls);
		southPanel.add(Box.createVerticalStrut(10));
		if (orientation == Orientation.VERTICAL) {
			add(southPanel, BorderLayout.SOUTH);
		} else {
			add(southPanel, BorderLayout.WEST);
		}
	}

	protected void marksPath(ArrayList<Vertex<T>> path, boolean marked) {
		if(marked) {
			graphComponent.markPath(path);
		} else {
			graphComponent.unmark();
		}
	}
	
	/**
	 * Returns graph presentation.
	 * 
	 * @return graph presentation
	 */
	public GraphPresentation<T> getGraphPresentation() {
		return this.graphComponent.getGraphPresentation();
	}

	@Override
	protected void paintComponent(Graphics g) {
		Graphics2D gr = (Graphics2D) g;
		gr.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		gr.setRenderingHint(RenderingHints.KEY_COLOR_RENDERING, RenderingHints.VALUE_COLOR_RENDER_QUALITY);
		if (this.getGraphPresentation() != null) {
			this.graphComponent.paint(g);
		}
	}

}
