package gui.graph.presentation;

import util.graph.AbstractGraph;
import util.graph.DirectedGraph;
import util.graph.Edge;
import util.graph.UndirectedGraph;
import util.graph.Vertex;
import gui.graph.layout.RandomLocations;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;


/**
 * This class provides a presentation for any implementation of <tt>AbstractGraph</tt>.
 * 
 * @author Mohamed Wahbi
 * 
 */
public class GraphPresentation<T extends Comparable<? super T>> {

	private AbstractGraph<T> graph = null;
	private Vector<VertexPresentation<T>> vertexes = null;
	private Vector<EdgePresentation<T>> edges = null;
	private Map<T, VertexPresentation<T>> vertexesMap = null;
	private Map<Edge<T>, EdgePresentation<T>> edgesMap = null;
	private Vector<GraphStateListener> listeners = null;
	private boolean isDirected;
	private boolean isWeighted;
	
	/**
	 * Shift from (0,0).
	 */
	public static final int SHIFT = 30;

	private final ElementStateListener elementStateListener = new ElementStateListener() {
		public void elementPositionChanged() {
			firePositionChanged();
		}
		public void elementStructureChanged() {
			fireStructureChanged();
		}
	};
	
	private boolean isPaused;
	
	private boolean wasStructureChangedRecieved;
	
	private boolean wasPositionChangedRecieved;
	
	/**
	 * Constructs presentation for the given graph
	 * 
	 * @param graph graph
	 * @param isDirected true if you want directed graph; false otherwise
	 * @param isWeighted true if you want weighted graph; false otherwise
	 */
	public GraphPresentation(AbstractGraph<T> graph, boolean isDirected, boolean isWeighted) {
		this.graph = graph;
		this.isDirected = isDirected;
		this.listeners = new Vector<GraphStateListener>();
		this.vertexes = new Vector<VertexPresentation<T>>();
		this.edges = new Vector<EdgePresentation<T>>();
		this.vertexesMap = new HashMap<T, VertexPresentation<T>>();
		this.edgesMap = new HashMap<Edge<T>, EdgePresentation<T>>();
		//int i = 0;
		Map<Vertex<T>, Point2D> locations = new RandomLocations<Vertex<T>>(graph.getVertexes(), new Dimension(500, 400));
		for (Vertex<T> vertex : this.graph.getVertexes()) {
			Point2D location = locations.get(vertex);
			VertexPresentation<T> vp = new VertexPresentation<T>(vertex, location);//getCenterForVertex(i));
			//i++;
			this.vertexes.add(vp);
			this.vertexesMap.put(vertex.getData(), vp);
		}
		for (Vertex<T> vertex : this.graph.getVertexes()) {
			for (Edge<T> edge : vertex.getOutgoingEdges()) {
				if (this.edgesMap.containsKey(edge)) {
					continue;
				}
				VertexPresentation<T> source = this.vertexesMap.get(vertex.getData());
				Vertex<T> opposite = edge.getOppositeVertex(vertex);
				VertexPresentation<T> target = this.vertexesMap.get(opposite.getData());
				EdgePresentation<T> ep = new EdgePresentation<T>(edge, source, target, isDirected, isWeighted);
				this.edges.add(ep);
				//System.out.println(ep);
				this.edgesMap.put(edge, ep);
				//ep.addElementStateListener(this.elementStateListener);
			}
		}
		this.isPaused = false;
		this.wasPositionChangedRecieved = false;
		this.wasStructureChangedRecieved = false;
		this.isWeighted = isWeighted;
		this.setWeightedEdges(isWeighted);
	}
	
	/**
	 * Constructs new graph presentation and new graph
	 * 
	 * @param isDirected true if you want directed graph; false otherwise
	 * @param isWeighted true if you want weighted graph; false otherwise
	 */
	public GraphPresentation(boolean isDirected, boolean isWeighted) {
		if (isDirected) {
			this.graph = new DirectedGraph<T>();
		} else {
			this.graph = new UndirectedGraph<T>();
		}
		this.isWeighted = isWeighted;
		this.isDirected = isDirected;
		this.listeners = new Vector<GraphStateListener>();
		this.vertexes = new Vector<VertexPresentation<T>>();
		this.edges = new Vector<EdgePresentation<T>>();
		this.vertexesMap = new HashMap<T, VertexPresentation<T>>();
		this.edgesMap = new HashMap<Edge<T>, EdgePresentation<T>>();
	}

//	private Point getCenterForVertex(int k) {
//		int vertexesNumber = this.graph.getVertexes().size();
//		int x = (int) Math.round(this.radix * Math.cos(2 * Math.PI * k / vertexesNumber));
//		int y = (int) Math.round(this.radix * Math.sin(2 * Math.PI * k / vertexesNumber));
//		return new Point(this.radix + SHIFT + x, this.radix + SHIFT + y);
//	}

	/**
	 * Adds the specified GraphStateListaner to receive notification about state change.
	 * 
	 * @param listener
	 */
	public void addGraphStateListener(GraphStateListener listener) {
		this.listeners.add(listener);
	}
	
	/**
	 * Removes the specified GraphStateListaner from this presentation.
	 * 
	 * @param listener
	 */
	public void removeGraphStateListener(GraphStateListener listener) {
		this.listeners.remove(listener);
	}
	
	private void firePositionChanged() {
		if (this.isPaused) {
			this.wasPositionChangedRecieved = true;
		} else {
			for (GraphStateListener listener : this.listeners) {
				listener.positionChanged();
			}
		}
	}
	
	private void fireStructureChanged() {
		if (this.isPaused) {
			this.wasStructureChangedRecieved = true;
		} else {
			for (GraphStateListener listener : this.listeners) {
				listener.structureChanged();
			}
		}
	}
	
	/**
	 * Pauses all changes firing until resume is not called.
	 *
	 */
	public void pauseFiringChanges() {
		this.isPaused = true;
	}
	
	/**
	 * Resumes firing changes.
	 *
	 */
	public void resumeFiringChanges() {
		this.isPaused = false;
		if (this.wasPositionChangedRecieved) {
			firePositionChanged();
			this.wasPositionChangedRecieved = false;
		}
		if (this.wasStructureChangedRecieved) {
			fireStructureChanged();
			this.wasStructureChangedRecieved = false;
		}
	}

	/**
	 * Paints this graph using given graphics
	 * 
	 * @param gr graphics
	 * @param bgColor background color
	 */
	public void paint(Graphics2D gr, Color bgColor) {
		for (EdgePresentation<T> ep : this.edges) {
			ep.paint(gr, bgColor);
		}
		for (VertexPresentation<T> vp : this.vertexes) {
			vp.paint(gr);
		}
	}
	
	/**
	 * Adds this vertex presentation to graph presentation.
	 * 
	 * @param vp vertex presentation
	 */
	public void addVertex(VertexPresentation<T> vp) {
		this.graph.addVertex(vp.getVertex());
		this.vertexes.add(vp);
		this.vertexesMap.put(vp.getVertex().getData(), vp);
		vp.addElementStateListener(this.elementStateListener);
		fireStructureChanged();
	}
	
	/**
	 * Removes this vertex presentation from graph presentation.
	 * 
	 * @param vp vertex presentation
	 */
	public void removeVertex(VertexPresentation<T> vp) {
		for (Edge<T> edge : vp.getVertex().getOutgoingEdges()) {
			EdgePresentation<T> ep = this.edgesMap.remove(edge);
			this.edges.remove(ep);
		}
		for (Edge<T> edge : vp.getVertex().getIncomingEdges()) {
			EdgePresentation<T> ep = this.edgesMap.remove(edge);
			this.edges.remove(ep);
		}
		this.graph.removeFromGraph(vp.getVertex());
		this.vertexes.remove(vp);
		this.vertexesMap.remove(vp.getData());
		vp.removeElementStateListener(this.elementStateListener);
		fireStructureChanged();
	}
	
	/**
	 * Adds this edge presentation to graph presentation.
	 * 
	 * @param ep edge presentation
	 */
	public void addEdge(EdgePresentation<T> ep) {
		this.graph.addEdge(ep.getEdge());
		this.edges.add(ep);
		this.edgesMap.put(ep.getEdge(), ep);
		ep.addElementStateListener(this.elementStateListener);
		fireStructureChanged();
	}
	
	/**
	 * Removes this edge presentation from graph presentation.
	 * 
	 * @param ep edge presentation
	 */
	public void removeEdge(EdgePresentation<T> ep) {
		this.graph.disconnect(ep.getEdge().getSource(), ep.getEdge().getTarget());
		this.edges.remove(ep);
		this.edgesMap.remove(ep);
		ep.removeElementStateListener(elementStateListener);
		fireStructureChanged();
	}

	/**
	 * Checks whether given vertexes are connected.
	 * 
	 * @param sourcePr source
	 * @param destinPr target
	 * @return true if they are; false otherwise
	 */
	public boolean areConnected(VertexPresentation<T> sourcePr, VertexPresentation<T> destinPr) {
		return this.graph.areConnected(sourcePr.getVertex(), destinPr.getVertex());
	}
	
	/**
	 * Returns preferred graph view size
	 * 
	 * @return size
	 */
	public Dimension getPreferredSize() {
		double resultX = 0;
		double resultY = 0;
		for (VertexPresentation<T> vp : this.vertexes) {
			resultX = Math.max(resultX, vp.getX() + VertexPresentation.WIDTH + 1);
			resultY = Math.max(resultY, vp.getY() + VertexPresentation.HEIGHT + 1);
		}
		return new Dimension((int) resultX, (int) resultY);
	}

	/**
	 * Returns model graph of this presentation.
	 * 
	 * @return graph
	 */
	public AbstractGraph<T> getGraph() {
		return this.graph;
	}
	
	/**
	 * Returns set of vertexes presentations.
	 * 
	 * @return set of vertexes presentations
	 */
	public Vector<VertexPresentation<T>> getVertexPresentations() {
		return this.vertexes;
	}
	
	/**
	 * Returns set of vertexes.
	 * 
	 * @return set of vertexes
	 */
	public List<T> getVertexes() {
		return new ArrayList<T>(this.vertexesMap.keySet());
//		Set<T> vertexes = new HashSet<T>(); 
//		for(Vertex<T> vertex : this.vertexesMap.keySet()) {
//			vertexes.add(vertex.getData());
//		}
//		return vertexes;
	}
	
	/**
	 * Returns set of edges presentations.
	 * 
	 * @return set of edges presentations
	 */
	public Vector<EdgePresentation<T>> getEdgePresentations() {
		return this.edges;
	}
	
	/**
	 * Returns the set of edges.
	 * 
	 * @return set of edges
	 */
	public Set<Edge<T>> getEdges() {
		return this.edgesMap.keySet();
	}
	
	/**
	 * Returns presentations of edges incident to given vertex presentation.
	 * 
	 * @param vp vertex presentation
	 * @return edges incident to given vertex
	 */
	public Set<EdgePresentation<T>> getIncidendEdgePresentations(VertexPresentation<T> vp) {
		Set<EdgePresentation<T>> result = new HashSet<EdgePresentation<T>>();
		for (Edge<T> edge : vp.getVertex().getOutgoingEdges()) {
			result.add(this.edgesMap.get(edge));
		}
		for (Edge<T> edge : vp.getVertex().getIncomingEdges()) {
			result.add(this.edgesMap.get(edge));
		}
		return result;
	}
	
	/**
	 * Checks whether this presentation is weighted.
	 * 
	 * @return true if it is; false otherwise
	 */
	public boolean isWeighted() {
		return this.isWeighted;
	}
	
	/**
	 * Checks whether this presentation is directed.
	 * 
	 * @return true if it is; false otherwise
	 */
	public boolean isDirected() {
		return this.isDirected;
	}
	
	/**
	 * Returns edge presentation for given edge.
	 * 
	 * @param edge edge
	 * @return edge presentation
	 */
	public EdgePresentation<T> getPresentationForEdge(Edge<T> edge) {
		return this.edgesMap.get(edge);
	}
	
	/**
	 * Returns vertex presentation for given vertex.
	 * 
	 * @param vertex vertex
	 * @return vertex presentation
	 */
	public VertexPresentation<T> getPresentationForVertex(Vertex<T> vertex) {
		return this.vertexesMap.get(vertex);
	}

	/**
	 * Returns the number of vertexes belonging to the this graph presentation.
	 * 
	 * @return  the number of vertexes in the graph
	 */
	public int getVertexCount() {
		return this.getVertexes().size();
	}

	/**
	 * @return the locations
	 */
	public Map<T, Point2D> getLocations() {
		Map<T, Point2D> locations = new HashMap<T, Point2D>();
		for(VertexPresentation<T> vertex : this.vertexes) {
			locations.put(vertex.getData(), vertex.getCenter());
		}
		return locations;
	}

	public void scale(double scale) {
		for(VertexPresentation<T> vertex : vertexes) {
			vertex.scale(scale);
		}
	}

	public void initialize() {
		for(VertexPresentation<T> vertex : vertexes) {
			vertex.initialize();
		}
	}

	public void setWeightedEdges(boolean weighted) {
		this.isWeighted = weighted;
		for(EdgePresentation<T> edge : edgesMap.values()) {
			edge.setWeighted(weighted);
		}
	}

	public void setLabeledEdges(boolean labeled) {
		for(EdgePresentation<T> edge : edgesMap.values()) {
			edge.setLabeled(labeled);
		}
	}

	public void setAssignedVertexes(boolean assigned) {
		for(VertexPresentation<T> vertex : vertexesMap.values()) {
			vertex.setLabeled(assigned);
		}
	}

	public void setLabeledVertexes(boolean labeled) {
		for(VertexPresentation<T> vertex : vertexesMap.values()) {
			vertex.setLabeled(labeled);
		}
	}

	public void setScalableVertexes(boolean scalable) {
		for(VertexPresentation<T> vertex : vertexesMap.values()) {
			vertex.setScalable(scalable);
		}
	}

	public void removeSelectedVertex() {
		Vector<VertexPresentation<T>> selectedVertexes = new Vector<VertexPresentation<T>>();
		for(VertexPresentation<T> vp : this.vertexes) {
			if(vp.isSelected()) {
				selectedVertexes.add(vp);
			}
		}
		for(VertexPresentation<T> vp : selectedVertexes) {
			this.removeVertex(vp);
		}
	}

	/**
	 * Marks presentations of given vertexes.
	 * 
	 * @param path to mark
	 */
	public void markPath(ArrayList<Vertex<T>> path) {
		for(int i = 0 ; i < path.size()-1 ; i++) {
			Vertex<T> source = path.get(i);
			Vertex<T> target = path.get(i+1);
			for(EdgePresentation<T> ep : getEdgePresentations()) {
				if (ep.getEdge().contains(source) && ep.getEdge().contains(target)) {
					ep.setMarked(true);
				}
			}
		}
	}

	/**
	 * Unmarks presentations of all edges.
	 * 
	 */
	public void unmark() {
		for(EdgePresentation<T> ep : getEdgePresentations()) {
			ep.setMarked(false);
		}
	}
	
}
