package gui.graph.presentation;

import java.util.HashMap;
import java.util.Map;

/**
 * This class presents a way to convert different presentation types to each other.
 * 
 * @author Mohamed Wahbi
 *
 */
public class PresentationConverter<T extends Comparable<T>> {
	
	/**
	 * Converts unweighted presentation to weighted.
	 * 
	 * @param presentation
	 * @return converted presentation
	 */
	public static <T extends Comparable<T>> GraphPresentation<T> convertToWeighted(GraphPresentation<T> presentation) {
		GraphPresentation<T> copyPres = new GraphPresentation<T>(presentation.isDirected(), true);
		copyPresentation(presentation, copyPres, presentation.isDirected(), true);
		return copyPres;
		
	}
	
	/**
	 * Converts weighted presentation to unweighted.
	 * 
	 * @param presentation
	 * @return converted presentation
	 */
	public static <T extends Comparable<T>> GraphPresentation<T> convertToUnweighted(GraphPresentation<T> presentation) {
		GraphPresentation<T> copyPres = new GraphPresentation<T>(presentation.isDirected(), false);
		copyPresentation(presentation, copyPres, presentation.isDirected(), false);
		return copyPres;
	}
	
	/**
	 * Converts directed presentation to undirected.
	 * 
	 * @param presentation
	 * @return converted presentation
	 */
	public static <T extends Comparable<T>> GraphPresentation<T> convertToUndirected(GraphPresentation<T> presentation) {
		GraphPresentation<T> copyPres = new GraphPresentation<T>(false, presentation.isWeighted());
		copyPresentation(presentation, copyPres, false, presentation.isWeighted());
		return copyPres;
	}
	
	private static <T extends Comparable<T>> void copyPresentation(GraphPresentation<T> originalPres, GraphPresentation<T> copyPres, boolean isDirected, boolean isWeighted) {
		Map<VertexPresentation<T>, VertexPresentation<T>> copyVertices = new HashMap<VertexPresentation<T>, VertexPresentation<T>>();
		for (VertexPresentation<T> vp : originalPres.getVertexPresentations()) {
			VertexPresentation<T> copyVP = new VertexPresentation<T>(vp.getCenter(), vp.getVertex().getData());
			copyPres.addVertex(copyVP);
			copyVertices.put(vp, copyVP);
		}
		for (EdgePresentation<T> ep : originalPres.getEdgePresentations()) {
			VertexPresentation<T> copySource = copyVertices.get(originalPres.getPresentationForVertex(ep.getEdge().getSource()));
			VertexPresentation<T> copyDestin = copyVertices.get(originalPres.getPresentationForVertex(ep.getEdge().getTarget()));
			if (copyPres.areConnected(copySource, copyDestin)) {
				copyPres.getGraph().getConnectingEdge(copySource.getVertex(), copyDestin.getVertex()).setWeight(ep.getWeight());
			} else {
				EdgePresentation<T> copyEP = new EdgePresentation<T>(copySource, copyDestin, isDirected, isWeighted);
				copyEP.setWeight(ep.getWeight());
				copyPres.addEdge(copyEP);
			}
		}
	}
}
