package gui.graph.presentation;

import util.graph.Vertex;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.Toolkit;
import java.awt.font.FontRenderContext;
import java.awt.font.TextLayout;
import java.awt.geom.AffineTransform;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.List;

/**
 * This class provides a presentation for <tt>Vertex</tt>.
 * 
 * @author Mohamed Wahbi
 * 
 */
public class VertexPresentation<T extends Comparable<? super T>> {

	public static int WIDTH = 50, HEIGHT = 30;
	private final Vertex<T> vertex;
	private boolean labeled = false;
	private boolean scalable = false;
	private boolean selected = false;
	private final Point2D center = new Point2D.Double();
	private Color fillColor = Color.yellow;
	private Color borderColor = Color.blue;
	private Color labelColor = Color.blue;
	private Color fillSelectionColor = Color.red;
	private Color borderSelectionColor = Color.black;
	private Color labelSelectionColor = Color.yellow;
	private final List<ElementStateListener> listeners;
	private Shape shape = null;
	private boolean marked;
	protected final BasicStroke DASHED_STROKE = new BasicStroke(1.5f, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND, 10.0f, new float[] {4.0f}, 0.0f);
	protected Color colorMark = Color.magenta;

	/**
	 * Creates a vertex presentation with given position.
	 * 
	 * @param vertex
	 *            model vertex
	 * 
	 * @param center
	 *            center position
	 */
	public VertexPresentation(Vertex<T> vertex, Point2D center) {
		// HEIGHT = WIDTH = 20;
		this.vertex = vertex;
		this.center.setLocation(center);
		this.listeners = new ArrayList<ElementStateListener>();
		this.shape = new Ellipse2D.Double(getX() - WIDTH / 2, getY() - HEIGHT / 2, WIDTH, HEIGHT);
	}

	/**
	 * Creates new vertex presentation with given position and new vertex.
	 * 
	 * @param center center position
	 */
	public VertexPresentation(Point2D center, T vertex) {
		this.listeners = new ArrayList<ElementStateListener>();
		this.vertex = new Vertex<T>(vertex);
		this.center.setLocation(center);
		this.shape = new Ellipse2D.Double(getX() - WIDTH / 2, getY() - HEIGHT / 2, WIDTH, HEIGHT);
	}

	/**
	 * Adds the specified ElementStateListaner to receive notification about state change.
	 * 
	 * @param listener
	 */
	public void addElementStateListener(ElementStateListener listener) {
		this.listeners.add(listener);
	}

	/**
	 * Removes the specified ElementStateListaner from this presentation.
	 * 
	 * @param listener
	 */
	public void removeElementStateListener(ElementStateListener listener) {
		this.listeners.remove(listener);
	}

	private void firePositionChanged() {
		for(ElementStateListener listener : this.listeners) {
			listener.elementPositionChanged();
		}
	}

	/**
	 * Retrieves the center point of this presentation.
	 * 
	 * @return center point
	 */
	public Point2D getCenter() {
		return (Point2D) this.center.clone();
	}

	/**
	 * Sets the center point of this presentation.
	 * 
	 * @param center center point
	 */
	public void setCenter(Point2D center) {
		double x = center.getX() - shape.getBounds2D().getCenterX();
		double y = center.getY() - shape.getBounds2D().getCenterY();
		AffineTransform at = AffineTransform.getTranslateInstance(x, y);
		this.shape = at.createTransformedShape(this.shape);
		this.center.setLocation(center);
		this.firePositionChanged();
	}

	/**
	 * Paints this vertex using given graphics
	 * 
	 * @param g graphics
	 */
	public void paint(Graphics2D gr) {
		Stroke oldStroke = gr.getStroke();
		if (isMarked()) {
			gr.setColor(colorMark);
			gr.setStroke(DASHED_STROKE);
		} else if (isSelected()) {
			gr.setColor(fillSelectionColor);
			gr.setStroke(new BasicStroke(1.2f));
		} else {
			gr.setColor(fillColor);
		}
		gr.fill(shape);
		if (isSelected()) {
			gr.setColor(borderSelectionColor);
		} else {
			gr.setColor(borderColor);
		}
		gr.draw(shape);
		if (isLabeled()) {
			if (isSelected()) {
				gr.setColor(labelSelectionColor);
			} else {
				gr.setColor(labelColor);
			}
			String label = this.vertex.toString();
			if (label.startsWith("(")) {
				label = label.substring(label.indexOf("(") + 1, label.indexOf(")")); // remove ( and )
			}
			Font font = gr.getFont();
			int screenRes = Toolkit.getDefaultToolkit().getScreenResolution();
			int fontSize = (int) Math.round(screenRes / 7.0);
			FontRenderContext context = gr.getFontRenderContext();
			// Create TextLayout
			TextLayout layout = new TextLayout(label, new Font(font.getName(), font.getStyle(), fontSize), context);
			Rectangle2D bounds = layout.getBounds();
			// Position of origin to draw the TextLayout.
			float x = (float) (getX() - bounds.getWidth() / 2);
			float y = (float) (getY() + bounds.getHeight() / 2);
			layout.draw(gr, x, y);
		}
		gr.setStroke(oldStroke);
	}

	/**
	 * Returns model vertex.
	 * 
	 * @return vertex
	 */
	public Vertex<T> getVertex() {
		return this.vertex;
	}

	/**
	 * checks whether this presentation contains the given point <code>point</code>.
	 * 
	 * @param point the point to checks its belonging to this presentation
	 * @return true if it does; false otherwise
	 */
	public boolean containsPoint(Point2D point) {
		return this.shape.contains(point);
		// return (square(point.getX() - this.getX()) + square(point.getY() - this.getY()) < square(RADIX) );
	}

	/**
	 * Checks whether this presentation lays within giver rectangle
	 * 
	 * 
	 * @param x1 left upper x
	 * @param y1 left upper y
	 * @param x2 right lower x
	 * @param y2 right lower y
	 * @return true if it does; false otherwise
	 */
	public boolean isInRect(double x1, double y1, double x2, double y2) {
		return (Math.min(x1, x2) <= this.getX()) && (Math.max(x1, x2) >= this.getX()) && (Math.min(y1, y2) <= this.getY())
				&& (Math.max(y1, y2) >= this.getY());
	}

	// private double square(double x) {
	// return x * x;
	// }

	public double getX() {
		return this.center.getX();
	}

	public double getY() {
		return this.center.getY();
	}

	/**
	 * Returns the data store in the this vertex model
	 * 
	 * @return the data
	 */
	public T getData() {
		return this.vertex.getData();
	}

	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "" + this.vertex + ":(" + this.getX() + "," + this.getY() + ")";
	}

	public void scale(double scale) {
		if (isScalable()) {
			AffineTransform at = new AffineTransform();
			at.scale(scale, scale);
			shape = at.createTransformedShape(shape);
			double newX = shape.getBounds().getCenterX();
			double newY = shape.getBounds().getCenterY();
			this.setCenter(new Point2D.Double(newX, newY));
		} else {
			AffineTransform at = new AffineTransform();
			double sx = getX() * scale - getX();
			double sy = getY() * scale - getY();
			at.translate(sx, sy);// .scale(scale, scale);
			shape = at.createTransformedShape(shape);
			double newX = shape.getBounds().getCenterX();
			double newY = shape.getBounds().getCenterY();
			this.setCenter(new Point2D.Double(newX, newY));
		}
	}

	public Shape getShape() {
		return this.shape;
	}

	public void setShape(Shape shape) {
		this.shape = shape;
	}

	public Rectangle getBounds() {
		return this.shape.getBounds();
	}

	public void initialize() {
		this.shape = new Ellipse2D.Double(getX() - WIDTH / 2, getY() - HEIGHT / 2, WIDTH, HEIGHT);
	}

	/**
	 * @return the labeled
	 */
	public boolean isLabeled() {
		return labeled;
	}

	/**
	 * @param labeled the labeled to set
	 */
	public void setLabeled(boolean labeled) {
		this.labeled = labeled;
	}

	public void select() {
		this.selected = true;
	}

	public void unselect() {
		this.selected = false;
	}

	public boolean isSelected() {
		return selected;
	}

	/**
	 * @return the scalable
	 */
	public boolean isScalable() {
		return this.scalable;
	}

	/**
	 * @param scalable the scalable to set
	 */
	public void setScalable(boolean scalable) {
		this.scalable = scalable;
	}

	/**
	 * Marks this vertex marked if the given boolean is true; otherwise unmark it.
	 * @param markColor 
	 *
	 */
	public void setMarked(boolean marked) {
		this.marked = marked;
	}

	public boolean isMarked() {
		return this.marked;
	}
}
