package gui.tree.control;

import gui.graph.presentation.VertexPresentation;
import gui.tree.presentation.TreeComponent;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Stroke;
import java.awt.event.InputEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.HashMap;
import java.util.Map;

/**
 * Basic selection state.
 * 
 * @author Mohamed Wahbi
 *
 */
public class MouseTool<T extends Comparable<? super T>> implements MouseListener, MouseMotionListener, MouseWheelListener {

	private final BasicStroke DASHED_STROKE = new BasicStroke(1.25f, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND, 10.0f, new float[] {2.5f}, 0.0f);
	protected final TreeComponent<T> component;
	private Point2D startPoint;
	private Point2D tmpStartPoint;
	private Point2D endPoint;
	private boolean isInSelectedItem;
	private boolean isInMultiSelectedItem;
	private final Map<VertexPresentation<T>, Point2D> beforeDragPositions;
	private boolean isFirstTime = true;

	/**
	 * Constructs new default state.
	 * 
	 * @param component
	 * @param parentTool
	 */
	public MouseTool(TreeComponent<T> component) {
		this.component = component;
		this.isInSelectedItem = false;
		this.beforeDragPositions = new HashMap<VertexPresentation<T>, Point2D>();
	}

	@Override
	public void mousePressed(MouseEvent e) {
		this.startPoint = e.getPoint();
		this.tmpStartPoint = e.getPoint();
		this.isInSelectedItem = component.getTreeElementSelection().isInSelectedItem(e.getPoint());
		if(!component.isInPickingMode()) {
			for(VertexPresentation<T> vp : component.getTreeElementSelection().getSelectedVertices()) {
				this.beforeDragPositions.put(vp, vp.getCenter());
			}
			return;
		}
		if (!isInSelectedItem && !isCtrlPressed(e)) {
			this.component.getTreeElementSelection().unselect();
		}
		if (!component.getTreeElementSelection().select(e.getPoint())) {
			this.isInMultiSelectedItem = true;
		}
		if (!component.getTreeElementSelection().getSelectedVertices().isEmpty()) {
			for(VertexPresentation<T> vp : component.getTreeElementSelection().getSelectedVertices()) {
				this.beforeDragPositions.put(vp, vp.getCenter());
			}
		}
	}

	@Override
	public void mouseClicked(MouseEvent e) {
		if(!component.isInPickingMode()) {
			//startPoint = e.getPoint();
			if (isCtrlPressed(e)) {
				this.component.getTreeElementSelection().select(e.getPoint());
			}
		} else if (isCtrlPressed(e) && isInSelectedItem) {
			this.component.getTreeElementSelection().unselectItem(e.getPoint());
		} else {
			if (!isCtrlPressed(e)) {
				this.component.getTreeElementSelection().unselect();
			}
			this.component.getTreeElementSelection().select(e.getPoint());
		}
	}

	@Override
	public void mouseReleased(MouseEvent e) {
		if(!component.isInPickingMode()) {
			endPoint = null;
			beforeDragPositions.clear();
		} else if (isInMultiSelectedItem) {
			this.endPoint = null;
			this.isInMultiSelectedItem = !component.isInPickingMode();
		} else if(startPoint!=null) {
			if (((startPoint.getX() - e.getPoint().x) != 0) || ((startPoint.getY() - e.getPoint().y) != 0)) {
				if (!component.getTreeElementSelection().getSelectedVertices().isEmpty()) {
					if (!isFirstTime) {
						this.component.getTreeElementSelection().unselect();
						this.component.getTreePresentation().pauseFiringChanges();
						for(VertexPresentation<T> vp : beforeDragPositions.keySet()) {
							component.getTreeElementSelection().selectVertex(vp);
						}
					} else {
						isFirstTime = false;
					}
					component.getTreePresentation().resumeFiringChanges();
					beforeDragPositions.clear();
				}
			}
		}
		component.makeVisualChanges();
		component.repaint();
		e.consume();
	}

	@Override
	public void mouseDragged(MouseEvent e) {
		if(!component.isInPickingMode()) {
			endPoint = e.getPoint();
			double dx = e.getPoint().x - tmpStartPoint.getX();
			double dy = e.getPoint().y - tmpStartPoint.getY();
			component.getTreeElementSelection().moveSelectedVertex(dx, dy);
			tmpStartPoint = e.getPoint();
		} else if (isInMultiSelectedItem) {
			endPoint = e.getPoint();
			component.getTreeElementSelection().unselect();
			component.getTreeElementSelection().selectInRect(startPoint.getX(), startPoint.getY(), endPoint.getX(), endPoint.getY());
		} else {
			double dx = e.getPoint().x - tmpStartPoint.getX();
			double dy = e.getPoint().y - tmpStartPoint.getY();
			component.getTreeElementSelection().moveSelectedVertex(dx, dy);
			tmpStartPoint = e.getPoint();
		}
		component.makeVisualChanges();
		component.repaint();
		e.consume();
	}

	private boolean isCtrlPressed(MouseEvent e) {
		return ((e.getModifiers() & InputEvent.CTRL_MASK) == InputEvent.CTRL_MASK);
	}

	public void paint(Graphics2D gr, Color color) {
		if (endPoint != null && component.isInPickingMode()) {
			Stroke prevStroke = gr.getStroke();
			gr.setStroke(DASHED_STROKE);
			gr.setColor(new Color(10, 10, 205));
			double x = Math.min(startPoint.getX(), endPoint.getX());
			double y = Math.min(startPoint.getY(), endPoint.getY());
			double width = Math.abs(endPoint.getX() - startPoint.getX());
			double height = Math.abs(endPoint.getY() - startPoint.getY());
			Rectangle2D rect = new Rectangle2D.Double(x, y, width, height);
			gr.draw(rect);
			gr.setStroke(prevStroke);
		}
	}

	@Override
	public void mouseMoved(MouseEvent e) {}

	@Override
	public void mouseEntered(MouseEvent e) {}

	@Override
	public void mouseExited(MouseEvent e) {}
	
	@Override
	public void mouseWheelMoved(MouseWheelEvent e) {
		double scale = 1;
		if(e.getWheelRotation() < 0) {
			scale = 1.1d;
		} else {
			scale = 1/1.1d;
		}
		component.zoom(scale);
	}
}
