package gui.tree.control;

import gui.graph.control.VisualStateListener;
import gui.graph.presentation.EdgePresentation;
import gui.graph.presentation.VertexPresentation;
import gui.tree.presentation.TreePresentation;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
 * This class represents a tool for selecting, unselecting, moving etc. tree presentation elements.
 * 
 * 
 * @author Mohamed Wahbi
 *
 */
public class TreeElementSelection<T extends Comparable<? super T>> {

	private TreePresentation<T> treePresentation;
	private final List<VisualStateListener> listeners;

	/**
	 * Creates new selection for given presentation.
	 * 
	 * @param treePresentation
	 */
	public TreeElementSelection(TreePresentation<T> treePresentation) {
		this.treePresentation = treePresentation;
		this.listeners = new ArrayList<VisualStateListener>();
	}

	/**
	 * Adds the specified VisualStateListaner to receive notification about state change.
	 * 
	 * @param listener
	 */
	public void addVisualStateListener(VisualStateListener listener) {
		this.listeners.add(listener);
	}

	/**
	 * Removes the specified VisualStateListaner from this presentation.
	 * 
	 * @param listener
	 */
	public void removeVisualStateListener(VisualStateListener listener) {
		this.listeners.remove(listener);
	}

	/**
	 * Sets new tree presentation.
	 * 
	 * @param treePresentation
	 */
	public void setTreePresentation(TreePresentation<T> treePresentation) {
		this.treePresentation = treePresentation;
		this.unselect();
	}

	private void fireVisualStateChange() {
		for(VisualStateListener listener : listeners) {
			listener.visualStateChanged();
		}
	}

	private void fireVisualChandesCancelled() {
		for(VisualStateListener listener : listeners) {
			listener.visualChangesCancelled();
		}
	}
	
	/**
	 * Returns currently selected vertexes.
	 * 
	 * @return currently selected vertexes
	 */
	public Set<VertexPresentation<T>> getSelectedVertices() {
		Set<VertexPresentation<T>> selectedVertices = new HashSet<VertexPresentation<T>>();
		for(VertexPresentation<T> vp : treePresentation.getVertexPresentations()) {
			if(vp.isSelected()) {
				selectedVertices.add(vp);
			}
		}
		return selectedVertices;
	}

	/**
	 * Returns currently selected edges.
	 * 
	 * @return currently selected edges
	 */
	public Set<EdgePresentation<T>> getSelectedEdges() {
		Set<EdgePresentation<T>> selectedEdges = new HashSet<EdgePresentation<T>>();
		for(EdgePresentation<T> ep : treePresentation.getEdgePresentations()) {
			if(ep.isSelected()) {
				selectedEdges.add(ep);
			}
		}
		return selectedEdges;
	}	

	/**
	 * Selects given vertex presentation.
	 * 
	 * @param vp vertex presentation
	 */
	public void selectVertex(VertexPresentation<T> vp) {
		vp.select();
		this.fireVisualStateChange();
	}

	/**
	 * Selects given edge presentation.
	 * 
	 * @param ep edge presentation
	 */
	public void selectEdge(EdgePresentation<T> ep) {
		ep.select();
		this.fireVisualStateChange();
	}

	/**
	 * Finds and marks element containing given point.
	 * 
	 * @param point point
	 * 
	 * @return true if element was found; false otherwise
	 */
	public boolean select(Point point) {
		boolean selectVertices = selectVerticesOnPoint(point);
		boolean selectEdges = selectEdgesOnPoint(point);
		return selectVertices || selectEdges;
	}

	/**
	 * Selects all item within given rectangle.
	 * 
	 * @param rectangle the rectangle
	 */
	public void selectInRect(Rectangle rectangle) {
		this.selectInRect(rectangle.getMinX(), rectangle.getMinY(), rectangle.getMaxX(), rectangle.getMaxY());
	}

	/**
	 * Selects all item within given rectangle.
	 * 
	 * @param x1 left upper corner x
	 * @param y1 left upper corner y
	 * @param x2 right lower corner x
	 * @param y2 right lower corner y
	 */
	public void selectInRect(double x1, double y1, double x2, double y2) {
		for(VertexPresentation<T> vp : treePresentation.getVertexPresentations()) {
			if (vp.isInRect(x1, y1, x2, y2)) {
				vp.select();
			}
		}
		for(EdgePresentation<T> ep : treePresentation.getEdgePresentations()) {
			if (ep.isInRect(x1, y1, x2, y2)) {
				ep.select();
			}
		}
		fireVisualStateChange();
	}

	/**
	 * Checks whether given point lies within an already selected item.
	 * 
	 * @param point point
	 * @return true if it does; false otherwise
	 */
	public boolean isInSelectedItem(Point point) {
		for(VertexPresentation<T> vp : treePresentation.getVertexPresentations()) {
			if (vp.isSelected() && vp.containsPoint(point)) {
				return true;
			}
		}
		for(EdgePresentation<T> ep : treePresentation.getEdgePresentations()) {
			if (ep.isSelected() && ep.containsPoint(point)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Unselects previously selected elements.
	 * 
	 */
	public void unselect() {
		for(EdgePresentation<T> ep : getSelectedEdges()) {
			ep.unselect();
		}
		for(VertexPresentation<T> vp : getSelectedVertices()) {
			vp.unselect();
		}
		fireVisualChandesCancelled();
	}

	/**
	 * Unselects item on given point
	 * 
	 * @param point point
	 */
	public void unselectItem(Point point) {
		for(VertexPresentation<T> vp : getSelectedVertices()) {
			if (vp.containsPoint(point)) {
				vp.unselect();
			}
		}
		for(EdgePresentation<T> ep : getSelectedEdges()) {
			if (ep.containsPoint(point)) {
				ep.unselect();
			}
		}
		fireVisualStateChange();
	}

	/**
	 * Moves selected vertex to the given point.
	 * 
	 * @param dx horizontal shift
	 * @param dy vertical shift
	 */
	public void moveSelectedVertex(double dx, double dy) {
		for(VertexPresentation<T> vp : getSelectedVertices()) {
			Point2D center = vp.getCenter();
			center.setLocation(center.getX() + dx, center.getY() + dy);
			if (center.getX() < VertexPresentation.WIDTH || center.getY() < VertexPresentation.HEIGHT) {
				return;
			}
		}
		for(VertexPresentation<T> vp : getSelectedVertices()) {
			Point2D center = vp.getCenter();
			center.setLocation(center.getX() + dx, center.getY() + dy);
			vp.setCenter(center);
		}
	}

	private boolean selectVerticesOnPoint(Point point) {
		boolean result = false;
		for(VertexPresentation<T> vp : treePresentation.getVertexPresentations()) {
			if (vp.containsPoint(point)) {
				vp.select();
				result = true;
			}
		}
		if (result) {
			fireVisualStateChange();
		}
		return result;
	}

	private boolean selectEdgesOnPoint(Point point) {
		boolean result = false;
		for(EdgePresentation<T> ep : treePresentation.getEdgePresentations()) {
			if (ep.containsPoint(point)) {
				ep.select();
				result = true;
			}
		}
		if (result) {
			fireVisualStateChange();
		}
		return result;
	}

	public void select() {
		for(VertexPresentation<T> vp : getSelectedVertices()) {
			vp.select();
		}
		for(EdgePresentation<T> ep : getSelectedEdges()) {
			ep.select();
		}
	}

}
