package gui.tree.layout;

import gui.tree.presentation.TreePresentation;
import java.awt.Dimension;
import java.awt.geom.Point2D;
import java.util.HashMap;
import java.util.Map;

/**
 * A radial layout for Tree .
 * 
 * @author Mohamed Wahbi
 */
public class RadialTreeLayout<T extends Comparable<? super T>> extends TreeLayoutTwo<T> {

	protected Map<T, PolarPoint> polarLocations;

	/**
	 * Creates an instance for the specified graph with default X and Y distances.
	 */
	public RadialTreeLayout(TreePresentation<T> tree) {
		this(tree, DEFAULT_DISTX, DEFAULT_DISTY);
	}

	/**
	 * Creates an instance for the specified graph and X distance with
	 * default Y distance.
	 */
	public RadialTreeLayout(TreePresentation<T> tree, int distx) {
		this(tree, distx, DEFAULT_DISTY);
	}

	/**
	 * Creates an instance for the specified graph, X distance, and Y distance.
	 */
	public RadialTreeLayout(TreePresentation<T> tree, int distx, int disty) {
		super(tree, distx, disty);
		int nbLeaves = getTree().getNumberOfLeaves();
//		int depth = getTree().getDepth();
		this.distX = (int) (getWidth() / nbLeaves);
		this.distY = (int) (getHeight() / nbLeaves);
	}

	@Override
	protected void buildTree() {
		super.buildTree();
		this.polarLocations = new HashMap<T, PolarPoint>();
		setRadialLocations();
	}

	@Override
	public void setSize(Dimension size) {
		if (size != null && this.treePresentation != null) {
			Dimension oldSize = this.size;
			this.size = size;
			initialize();
			if (oldSize != null) {
				adjustLocations(oldSize, size);
			}
		}
	}

	@Override
	protected void setCurrentPositionFor(T vertex) {
		locations.get(vertex).setLocation(m_currentPoint);
	}

	@Override
	public void setLocation(T v, Point2D location) {
		Point2D c = getCenter();
		Point2D pv = new Point2D.Double(location.getX() - c.getX(), location.getY() - c.getY());
		PolarPoint newLocation = PolarPoint.cartesianToPolar(pv);
		PolarPoint currentLocation = polarLocations.get(v);
		if (currentLocation == null) {
			polarLocations.put(v, newLocation);
		} else {
			currentLocation.setLocation(newLocation);
		}
	}

	/**
	 * Returns the map from vertices to their locations in polar coordinates.
	 */
	public Map<T, PolarPoint> getPolarLocations() {
		return polarLocations;
	}

	@Override
	public Point2D transform(T vertex) {
		PolarPoint p = polarLocations.get(vertex);
		double centerX = getWidth() / 2;
		double centerY = getHeight() / 2;
		Point2D cartesian = PolarPoint.polarToCartesian(p);
		cartesian.setLocation(cartesian.getX() + centerX, cartesian.getY() + centerY);
		return cartesian;
	}

	private Point2D getMaxXY() {
		double maxx = 0;
		double maxy = 0;
		for(Point2D location : locations.values()) {
			maxx = Math.max(maxx, location.getX());
			maxy = Math.max(maxy, location.getY());
		}
		return new Point2D.Double(maxx, maxy);
	}

	private Point2D getMinXY() {
		double minx = getWidth();
		double miny = getHeight();
		for(Point2D location : locations.values()) {
			minx = Math.min(minx, location.getX());
			miny = Math.min(miny, location.getY());
		}
		return new Point2D.Double(minx, miny);
	}

	private void setRadialLocations() {
		Point2D max = getMaxXY();
		double maxx = Math.max(max.getX(), getWidth());
		double maxy = Math.max(max.getY(), getHeight());
		double theta = 2 * Math.PI / maxx;
		double deltaRadius = getWidth() / 2 / maxy;
		for(Map.Entry<T, Point2D> entry : locations.entrySet()) {
			T v = entry.getKey();
			Point2D p = entry.getValue();
			PolarPoint polarPoint = new PolarPoint(p.getX() * theta, (p.getY() - this.distY) * deltaRadius);
			polarLocations.put(v, polarPoint);
		}
	}

	/**
	 * Returns the center of this layout's area.
	 */
	public Point2D getCenter() {
		Point2D min = getMinXY();
		Point2D max = getMaxXY();
		double x = (min.getX() + max.getX())/2;
		double y = (min.getY() + max.getY())/2;
		//x -= min.getX();
		//y -= min.getY();
		return new Point2D.Double(x, y);
//		return new Point2D.Double(getWidth() / 2, getHeight() / 2);
	}

	static class PolarPoint {
		
		private double theta;
		private double radius;

		/**
		 * Creates a new instance with radius and angle each 0.
		 */
		public PolarPoint() {
			this(0, 0);
		}

		/**
		 * Creates a new instance with radius {@code radius} and angle {@code theta}.
		 */
		public PolarPoint(double theta, double radius) {
			this.theta = theta;
			this.radius = radius;
		}

		/**
		 * Returns the angle for this point.
		 */
		public double getTheta() {
			return theta;
		}

		/**
		 * Returns the radius for this point.
		 */
		public double getRadius() {
			return radius;
		}

		/**
		 * Sets the angle for this point to {@code theta}.
		 */
		public void setTheta(double theta) {
			this.theta = theta;
		}

		/**
		 * Sets the radius for this point to {@code theta}.
		 */
		public void setRadius(double radius) {
			this.radius = radius;
		}

		/**
		 * Returns the result of converting <code>polar</code> to Cartesian coordinates.
		 */
		public static Point2D polarToCartesian(PolarPoint polar) {
			return polarToCartesian(polar.getTheta(), polar.getRadius());
		}

		/**
		 * Returns the result of converting <code>(theta, radius)</code> to Cartesian coordinates.
		 */
		public static Point2D polarToCartesian(double theta, double radius) {
			return new Point2D.Double(radius * Math.cos(theta), radius * Math.sin(theta));
		}

		/**
		 * Returns the result of converting <code>point</code> to polar coordinates.
		 */
		public static PolarPoint cartesianToPolar(Point2D point) {
			return cartesianToPolar(point.getX(), point.getY());
		}

		/**
		 * Returns the result of converting <code>(x, y)</code> to polar coordinates.
		 */
		public static PolarPoint cartesianToPolar(double x, double y) {
			double theta = Math.atan2(y, x);
			double radius = Math.sqrt(x * x + y * y);
			return new PolarPoint(theta, radius);
		}

		/**
		 * Sets the angle and radius of this point to those of {@code p}.
		 */
		public void setLocation(PolarPoint polar) {
			this.theta = polar.getTheta();
			this.radius = polar.getRadius();
		}
	}
}
