package gui.tree.layout;

import util.graph.Edge;
import util.graph.Vertex;
import gui.graph.layout.Layout;
import gui.graph.presentation.EdgePresentation;
import gui.graph.presentation.VertexPresentation;
import gui.tree.presentation.TreePresentation;
import java.awt.Dimension;
import java.awt.geom.Point2D;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.Vector;
import util.tree.AbstractTree;
import util.tree.AbstractTreeNode;

/**
 * This class tree layout algorithm
 * @author Mohamed Wahbi
 */
public class TreeLayout<T extends Comparable<? super T>> extends Layout<T> {

	private double xsteps; 
	private double xoffset; 
	private double ysteps; 
	private double yoffset;
	private Map<T,  Double> positions;
	private Map<T,  Double> maxX;
	private Map<T,  Double> minX;
	private TreePresentation<T> treePresentation;
	
	/**
	 * Creates an instance for the specified tree.
	 */
	public TreeLayout(TreePresentation<T> tree) {
		super();
		this.treePresentation = tree;
		this.positions = new HashMap<T,  Double>();
		this.minX = new HashMap<T,  Double>();
		this.maxX = new HashMap<T,  Double>();
		for(T vertex : tree.getVertexes()) {
			this.positions.put(vertex, 0.0);
			this.minX.put(vertex, 0.0);
			this.maxX.put(vertex, 0.0);
			this.locations.put(vertex, new Point2D.Double());
		}
	}

	/**
	 * Creates an instance of size {@code d} for the specified tree.
	 */
	public TreeLayout(TreePresentation<T> tree, Dimension size) {
		this(tree);
		this.size = size;
		initialize();
	}

	public void reset() {
		doInit();
		computesLocations();
		updateTreeLocation();
	}

	public void initialize() {
		this.treePresentation.initialize();
		doInit();
		computesLocations();
		updateTreeLocation();
	}

	private void doInit() {
		computesPositions();
		T root = getTree().getRootNode().getData();
		this.xsteps = (getWidth() / (getMaxX(root) - getMinX(root) +1));
		this.ysteps = (getHeight() / (getTree().getDepth() + 1));
		this.xoffset = (this.xsteps * (0.5 - getMinX(root)));
		this.yoffset = this.ysteps / 2.0;
	}

	private void computesPositions() {
		position(getTree().getRoot(), 0);// xoffset);
	}

	private void computesLocations() {
		computeLocation(getTree().getRoot(), 0);
	}

	/**
	 * Returns set of vertexes presentations.
	 * 
	 * @return set of vertexes presentations
	 */
	public Vector<VertexPresentation<T>> getVertexPresentations() {
		return this.treePresentation.getVertexPresentations();
	}
	
	/**
	 * Returns set of edges presentations.
	 * 
	 * @return set of edges presentations
	 */
	public Vector<EdgePresentation<T>> getEdgePresentations() {
		return this.treePresentation.getEdgePresentations();
	}
	
	/**
	 * Returns the set of edges.
	 * 
	 * @return set of edges
	 */
	public Set<Edge<T>> getEdges() {
		return this.treePresentation.getEdges();
	}
	
	protected void updateTreeLocation() {
		for(VertexPresentation<T> vertex : treePresentation.getVertexPresentations()) {
			vertex.setCenter(transform(vertex.getData()));
		}
	}
	
	protected AbstractTree<T> getTree() {
		return this.treePresentation.getTree();
	}

	public double getPosition(T vertex) {
		return this.positions.get(vertex);
	}

	public void setPosition(T vertex, double position) {
		this.positions.put(vertex, position);
	}

	public double getMinX(T vertex) {
		return this.minX.get(vertex);
	}

	public void setMinX(T vertex, double minX) {
		this.minX.put(vertex, minX);
	}

	public double getMaxX(T vertex) {
		return this.maxX.get(vertex);
	}

	public void setMaxX(T vertex, double maxX) {
		this.maxX.put(vertex, maxX);
	}
	
	private void position(AbstractTreeNode<T> tree, double nextLeafX) {
		T vertex = tree.getNode().getData();
		this.minX.put(vertex, nextLeafX);
		double firstson = nextLeafX;
		double lastson = nextLeafX;
		if (!tree.isLeaf()) {
			boolean isfirstson = true;
			for (AbstractTreeNode<T> st : tree.getChildren()) {
				T child = st.getNode().getData();
				position(st, nextLeafX);
				nextLeafX = getMaxX(child) + 1;
				lastson = getPosition(child);
				if (isfirstson) {
					firstson = getPosition(child);
					isfirstson = false;
				}
			}
			--nextLeafX;
		}
		this.maxX.put(vertex, nextLeafX);
		double position = (firstson + lastson) / 2;
		this.setPosition(vertex, position);
	}

	public void computeLocation(T vertex, int depth) {
		double x = scaleX(getPosition(vertex));
		double y = scaleY(depth);
		this.setLocation(vertex, x, y);
	}
	
	private void computeLocation(AbstractTreeNode<T> tree, int depth) {
		computeLocation(tree.getNode().getData(), depth);
		if (!tree.isLeaf()) {
			for (AbstractTreeNode<T> st : tree.getChildren()) {
				computeLocation(st, depth+1);
			}
		}
	}
	
	public double scaleX(double position) {
		return (int) (position * xsteps) + xoffset;
	}

	public double scaleY(int depth) {
		return (depth * ysteps) + yoffset;
	}

	
	@Override
	public void setSize(Dimension size) {
		if (size != null && this.treePresentation != null) {
			Dimension oldSize = this.size;
			this.size = size;
			initialize();
			if (oldSize != null) {
				adjustLocations(oldSize, size);
			}
		}
		updateTreeLocation();
	}
	
	protected void adjustLocations(Dimension oldSize, Dimension size) {
		int xOffset = (size.width - oldSize.width) / 2;
		int yOffset = (size.height - oldSize.height) / 2;
		// now, move each vertex to be at the new screen center
		for(Vertex<T> vertex : getVertexes()) {
			offsetVertex(vertex.getData(), xOffset, yOffset);
		}
	}
}
