/*
 * Copyright (c) 2005, the JUNG Project and the Regents of the University of
 * California All rights reserved.
 * This software is open-source under the BSD license; see either "license.txt"
 * or http://jung.sourceforge.net/license.txt for a description.
 * Created on Jul 9, 2005
 */

package gui.tree.layout;

import util.graph.Edge;
import util.graph.Vertex;
import gui.graph.layout.Layout;
import gui.graph.presentation.EdgePresentation;
import gui.graph.presentation.VertexPresentation;
import gui.tree.presentation.TreePresentation;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.geom.Point2D;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.Vector;
import util.tree.AbstractTree;
import util.tree.AbstractTreeNode;

/**
 * 
 * @author Mohamed Wahbi 
 */
public class TreeLayoutTwo<T extends Comparable<? super T>> extends Layout<T> {

	protected TreePresentation<T> treePresentation;
	protected Map<T, Integer> basePositions = new HashMap<T, Integer>();
	protected transient Set<T> alreadyDone = new HashSet<T>();

	/**
	 * The default horizontal vertex spacing.  Initialized to 50.
	 */
	public static int DEFAULT_DISTX = 50;

	/**
	 * The default vertical vertex spacing.  Initialized to 50.
	 */
	public static int DEFAULT_DISTY = 50;

	/**
	 * The horizontal vertex spacing.  Defaults to {@code DEFAULT_XDIST}.
	 */
	protected int distX = 50;

	/**
	 * The vertical vertex spacing.  Defaults to {@code DEFAULT_YDIST}.
	 */
	protected int distY = 50;

	protected transient Point m_currentPoint = new Point();

	/**
	 * Creates an instance for the specified graph with default X and Y distances.
	 */
	public TreeLayoutTwo(TreePresentation<T> tree) {
		this(tree, DEFAULT_DISTX, DEFAULT_DISTY);
	}

	/**
	 * Creates an instance for the specified graph and X distance with
	 * default Y distance.
	 */
	public TreeLayoutTwo(TreePresentation<T> treePresentation, int distx) {
		this(treePresentation, distx, DEFAULT_DISTY);
	}

	/**
	 * Creates an instance for the specified graph, X distance, and Y distance.
	 */
	public TreeLayoutTwo(TreePresentation<T> treePresentation, int distx, int disty) {
		if (treePresentation == null)
			throw new IllegalArgumentException("Graph must be non-null");
		if (distx < 1 || disty < 1)
			throw new IllegalArgumentException("X and Y distances must each be positive");
		this.treePresentation = treePresentation;
		this.distX = (int) (getWidth()/(getTree().getNumberOfLeaves()));
		this.distY = (int) (getHeight()/(getTree().getDepth()+2));
		for(T vertex : treePresentation.getVertexes()) {
			this.locations.put(vertex, new Point2D.Double());
		}
	}

	public void initialize() {
		this.treePresentation.initialize();
		buildTree();
		updateTreeLocation();
	}

	protected void buildTree() {
		this.m_currentPoint = new Point(0, 0);
		if (treePresentation != null) {
			Collection<AbstractTreeNode<T>> roots = getTree().getSubtrees();
			calculateDimensionX(roots);
			AbstractTreeNode<T> root = getTree().getRoot();
			calculateDimensionX(root);
			T v = root.getNode().getData();
			m_currentPoint.x += this.basePositions.get(v) / 2 + this.distX;
			buildTree(root, this.m_currentPoint.x);
		}
	}

	protected void buildTree(AbstractTreeNode<T> root, int x) {
		T v = root.getNode().getData();
		if (!alreadyDone.contains(v)) {
			alreadyDone.add(v);
			// go one level further down
			this.m_currentPoint.y += this.distY;
			this.m_currentPoint.x = x;
			this.setCurrentPositionFor(v);
			int sizeXofCurrent = basePositions.get(v);
			int lastX = x - sizeXofCurrent / 2;
			for(AbstractTreeNode<T> child : root.getChildren()) {
				int sizeXofChild = this.basePositions.get(child.getNode().getData());
				int startXofChild = lastX + sizeXofChild / 2;
				buildTree(child, startXofChild);
				lastX = lastX + sizeXofChild + distX;
			}
			this.m_currentPoint.y -= this.distY;
		}
	}

	private int calculateDimensionX(AbstractTreeNode<T> root) {
		int size = 0;
		int childrenNum = root.getChildren().size();

		if (childrenNum != 0) {
			for(AbstractTreeNode<T> child : root.getChildren()) {
				size += calculateDimensionX(child) + distX;
			}
		}
		size = Math.max(0, size - distX);
		basePositions.put(root.getNode().getData(), size);
		return size;
	}

	private int calculateDimensionX(Collection<AbstractTreeNode<T>> roots) {
		int size = 0;
		for(AbstractTreeNode<T> child : roots) {
			int childrenNum = child.getChildren().size();

			if (childrenNum != 0) {
				for(AbstractTreeNode<T> grandSon : child.getChildren()) {
					size += calculateDimensionX(grandSon) + distX;
				}
			}
			size = Math.max(0, size - distX);
			basePositions.put(child.getNode().getData(), size);
		}
		return size;
	}
	protected void setCurrentPositionFor(T vertex) {
		int x = m_currentPoint.x;
		int y = m_currentPoint.y;
		if (x < 0)
			size.width -= x;
		if (x > size.width - distX)
			size.width = x + distX;

		if (y < 0)
			size.height -= y;
		if (y > size.height - distY)
			size.height = y + distY;
		setLocation(vertex, m_currentPoint);
	}

	public void reset() {
		this.initialize();
	}

	/**
	 * Returns set of vertexes presentations.
	 * 
	 * @return set of vertexes presentations
	 */
	public Vector<VertexPresentation<T>> getVertexPresentations() {
		return this.treePresentation.getVertexPresentations();
	}
	
	/**
	 * Returns set of edges presentations.
	 * 
	 * @return set of edges presentations
	 */
	public Vector<EdgePresentation<T>> getEdgePresentations() {
		return this.treePresentation.getEdgePresentations();
	}
	
	/**
	 * Returns the set of edges.
	 * 
	 * @return set of edges
	 */
	public Set<Edge<T>> getEdges() {
		return this.treePresentation.getEdges();
	}

	protected void updateTreeLocation() {
		for(VertexPresentation<T> vertex : treePresentation.getVertexPresentations()) {
			vertex.setCenter(transform(vertex.getData()));
		}
	}
	
	@Override
	public void setSize(Dimension size) {
		if (size != null && this.treePresentation != null) {
			Dimension oldSize = this.size;
			this.size = size;
			initialize();
			this.distX = (int) (getWidth()/(getTree().getNumberOfLeaves()-1));
			this.distY = (int) (getHeight()/(getTree().getDepth()+2));
			if (oldSize != null) {
				adjustLocations(oldSize, size);
			}
		}
	}
	
	protected void adjustLocations(Dimension oldSize, Dimension size) {
		int xOffset = (size.width - oldSize.width) / 2;
		int yOffset = (size.height - oldSize.height) / 2;
		// now, move each vertex to be at the new screen center
		for(Vertex<T> vertex : getVertexes()) {
			offsetVertex(vertex.getData(), xOffset, yOffset);
		}
	}
	
	protected AbstractTree<T> getTree() {
		return treePresentation.getTree();
	}
}
