package gui.tree.presentation;

import gui.tree.control.TreeElementSelection;
import gui.tree.control.MouseTool;
import gui.tree.control.PopupMenuTool;
import gui.graph.control.VisualStateListener;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import javax.swing.JComponent;

/**
 * A component for tree viewing and editing.
 * 
 * @author Mohamed Wahbi
 * 
 */
@SuppressWarnings("serial")
public class TreeComponent<T extends Comparable<? super T>> extends JComponent {
	
	private TreePresentation<T> treePresentation;
	private TreeElementSelection<T> treeElementSelection;
	private final MouseTool<T> mouseTool;
	private final PopupMenuTool<T> popupMenu;
	private boolean pickingMode = true;

	private final VisualStateListener selectionStateListener = new VisualStateListener() {
		public void visualChangesCancelled() {
			makeVisualChanges();
		}
		public void visualStateChanged() {
			makeVisualChanges();
		}
	};
	
	/**
	 * Creates component without tree to present.
	 */
	public TreeComponent() {
		this(new Dimension(600, 400));
	}
	
	/**
	 * Creates component without tree to present.
	 *
	 */
	public TreeComponent(Dimension size) {
		super();
		this.setSize(size);
		this.setPreferredSize(size);
		this.setBackground(Color.white);
		this.mouseTool = new MouseTool<T>(this);
		addMouseListener(mouseTool);
		addMouseMotionListener(mouseTool);
		addMouseWheelListener(mouseTool);
		this.popupMenu = new PopupMenuTool<T>(this);
		setComponentPopupMenu(popupMenu);
	}
	
	/**
	 * Creates component with given presentation.
	 * 
	 * @param tree tree presentation
	 */
	public TreeComponent(TreePresentation<T> tree, Dimension size) {
		this(size);
		this.setTreePresentation(tree);
	}
	
	/**
	 * Sets new tree presentation to this component
	 * 
	 * @param presentation new tree presentation
	 */
	public void setTreePresentation(TreePresentation<T> presentation) {
		this.treePresentation = presentation;
		treeElementSelection = new TreeElementSelection<T>(presentation);
		treeElementSelection.addVisualStateListener(selectionStateListener);
		treeElementSelection.setTreePresentation(treePresentation);
		makeVisualChanges();
	}

	public void makeVisualChanges() {
		setPreferredSize(treePresentation.getPreferredSize());
		revalidate();
		repaint();
	}
	
	/**
	 * Returns tree presentation.
	 * 
	 * @return tree presentation
	 */
	public TreePresentation<T> getTreePresentation() {
		return this.treePresentation;
	}
	
	/**
	 * Adds temporary component to this component.
	 * 
	 * @param component the component to add.
	 */
	public void addTempComponent(Component component) {
		add(component);
		revalidate();
		repaint();
	}

	/**
	 * Removes temporary component from this component.
	 * 
	 * @param component the component to remove
	 */
	public void removeTempComponent(Component component) {
		remove(component);
		revalidate();
		repaint();
	}

	@Override
	protected void paintComponent(Graphics g) {
		g.setColor(Color.white);
		g.fillRect(0, 0, getWidth(), getHeight());
		Graphics2D gr = (Graphics2D) g;
		gr.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		gr.setRenderingHint(RenderingHints.KEY_COLOR_RENDERING, RenderingHints.VALUE_COLOR_RENDER_QUALITY);
		if (this.treePresentation != null) {
			this.treePresentation.paint(gr, getBackground());
			this.treeElementSelection.select();
		}
		this.mouseTool.paint(gr, getBackground());
	}

	public void scale(float scale) {
		zoom(scale);
	}

	public void zoom(double scale) {
		Dimension size = getSize();
		size.width *= scale;
		size.height *= scale;
		this.treePresentation.scale(scale);
		makeVisualChanges();
		//setPreferredSize(size);
		setSize(size);
		revalidate();
		repaint();
	}

	/**
	 * Returns tree element selection.
	 * 
	 * @return tree element selection
	 */
	public TreeElementSelection<T> getTreeElementSelection() {
		return treeElementSelection;
	}

	public boolean isInPickingMode() {
		return pickingMode;
	}

	public void setPickingMode(boolean pickingMode) {
		this.pickingMode = pickingMode;
	}

}
