package gui.tree.presentation;

import util.graph.Edge;
import gui.graph.presentation.EdgePresentation;
import gui.graph.presentation.VertexPresentation;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.util.Map;

/**
 * This class provides a presentation for <tt>Edge<T></tt>.
 * 
 * @author Mohamed Wahbi
 * 
 */
public class TreeEdgePresentation<T extends Comparable<? super T>> extends EdgePresentation<T> {

	/**
	 * Creates edge presentation with given start and end points.
	 * 
	 * @param edge edge
	 * @param vertexesMap map from vertexes to their presentations
	 * @param isDirected true if you want directed edge; false otherwise
	 * @param isWeighted true if you want weighted edge; false otherwise
	 */
	public TreeEdgePresentation(Edge<T> edge, Map<T, VertexPresentation<T>> vertexesMap, boolean isWeighted) {
		super(edge, vertexesMap, true, isWeighted);
	}
	
	/**
	 * Creates edge presentation with given start and end points.
	 * 
	 * @param edge edge
	 * @param source source vertex presentation
	 * @param target source vertex presentation	 
	 * @param isWeighted true if you want weighted edge; false otherwise
	 */
	public TreeEdgePresentation(Edge<T> edge, VertexPresentation<T> source, VertexPresentation<T> target, boolean isWeighted) {
		super(edge, source, target, true, isWeighted);
	}

	/**
	 * Creates a new edge presentation and new edge. 
	 * 
	 * @param source source vertex presentation
	 * @param target source vertex presentation
	 * @param isDirected true if you want directed edge; false otherwise
	 * @param isWeighted true if you want weighted edge; false otherwise
	 */
	public TreeEdgePresentation(VertexPresentation<T> source, VertexPresentation<T> target, boolean isWeighted) {
		super(source, target, true, isWeighted);
	}

	/**
	 * Paints this edge using given graphics
	 * 
	 * @param gr graphics
	 * @param bgColor background color
	 */
	@Override
	public void paint(Graphics2D gr, Color bgColor) {
		if(isSelected()) {
			gr.setColor(colorSelection);
		} else {
			gr.setColor(color);
		}
		Point2D start = getStart();
		Point2D end = getEnd();
		Line2D edge = new Line2D.Double(start, end);
		gr.draw(edge);
		if (isDirected) {
			drawArrow(gr, end);
		}
		if (isWeighted || isLabeled) {
			drawLabelAndWeight(gr, start, end, bgColor);
		}
	}

	protected void drawLabelAndWeight(Graphics2D gr, Point2D start, Point2D end, Color backgroundColor) {
		String label = "";
		if (isLabeled && isWeighted) {
			label = getLabel() + "=" + Double.toString(this.getWeight());
		} else if (isLabeled) {
			label = getLabel();
		} else if (isWeighted) {
			label = Double.toString(this.getWeight());
		} else {
			return;
		}
		double x = (start.getX() + end.getX()) / 2;
		double y = (start.getY() + end.getY()) / 2;
		if(label.length()==0) {
			label = " ";
		}
		this.drawLabelAndWeight(gr, label, x, y, backgroundColor);
	}
	
	protected void drawArrow(Graphics2D gr, Point2D end) {
		double dx = target.getX() - source.getX();
		double dy = target.getY() - source.getY();
		double thetaRadians = (float) Math.atan2(dy, dx);
		AffineTransform at = AffineTransform.getTranslateInstance(end.getX(), end.getY());
		at.rotate(thetaRadians);
		Shape arrow = getNotchedArrow(base, height, notch_height);
		arrow = at.createTransformedShape(arrow);
		if(select) {
			gr.setColor(colorSelection);
		} else {
			gr.setColor(color);
		}
		//gr.fill(arrow);
		gr.draw(arrow);
	}
}
