package gui.tree.presentation;

import gui.graph.layout.Layout;
import gui.tree.layout.RadialTreeLayout;
import gui.tree.layout.TreeLayout;
import gui.tree.layout.TreeLayoutTwo;
import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.lang.reflect.Constructor;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.DefaultListCellRenderer;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import util.tree.AbstractTree;

@SuppressWarnings({"serial", "unchecked"})
public class TreePanel<T extends Comparable<? super T>> extends JPanel {

	public enum Orientation {
		HORIZONTAL, VERTICAL
	}

	private Orientation orientation = Orientation.HORIZONTAL;
	private final static int CONTROL_WIDTH = 185;
	private final static int CONTROL_HEIGHT = 100;

	private TreeComponent<T> treeComponent;
	private JComboBox layoutComboBox = null;
	Class<T>[] layoutClasses = new Class[] {TreeLayout.class, TreeLayoutTwo.class, RadialTreeLayout.class};
	private JComboBox mouseModeBox = null;

	private JPanel controls;

	public TreePanel() {
		this(Orientation.HORIZONTAL);
	}

	/**
	 * Creates component with given presentation.
	 * 
	 */
	public TreePanel(Orientation orientation) {
		this(new AbstractTree<T>(), new Dimension(600, 500), orientation);
	}

	/**
	 * Creates component with given presentation.
	 * 
	 * @param tree tree presentation
	 */
	public TreePanel(AbstractTree<T> tree, Dimension size, Orientation orientation) {
		this(new TreePresentation<T>(tree, false, false), size, orientation);
	}

	/**
	 * Creates component with given presentation.
	 * 
	 * @param tree tree presentation
	 */
	public TreePanel(TreePresentation<T> tree, Dimension size, Orientation orientation) {
		super(new BorderLayout());
		this.orientation = orientation;
		setPreferredSize(size);
		Dimension treeComponentSize = new Dimension(size);
		Dimension controlPanelSize = new Dimension(size);
		setPreferredSize(size);
		if (this.orientation == Orientation.VERTICAL) {
			treeComponentSize.setSize(size.getWidth(), size.getHeight() - CONTROL_HEIGHT);
			controlPanelSize.setSize(size.getWidth(), CONTROL_HEIGHT);
		} else {
			treeComponentSize.setSize(size.getWidth() - CONTROL_WIDTH, size.getWidth());
			controlPanelSize.setSize(CONTROL_WIDTH, size.getWidth());
		}
		this.treeComponent = new TreeComponent<T>(tree, treeComponentSize);
		setTreePresentation(tree);
		createTreeCanvas();
		createControlPanel(controlPanelSize);
	}

	/**
	 * Sets new tree presentation to this component
	 * 
	 * @param presentation new tree presentation
	 */
	public void setTreePresentation(TreePresentation<T> presentation) {
		this.treeComponent.setTreePresentation(presentation);
		makeVisualChanges();
	}

	private void makeVisualChanges() {
		Dimension treeComponentSize = getTreePresentation().getPreferredSize();
		Dimension size = new Dimension(treeComponentSize);
		if (orientation == Orientation.VERTICAL) {
			size.setSize(size.width, size.height + CONTROL_HEIGHT);
		} else {
			size.setSize(size.width + CONTROL_WIDTH, size.height);
		}
		setPreferredSize(size);
		revalidate();
		repaint();
	}

	public JComboBox getLayoutsComboBox() {
		if (layoutComboBox == null) {
			layoutComboBox = new JComboBox(layoutClasses);
			layoutComboBox.setRenderer(new DefaultListCellRenderer() {

				public Component getListCellRendererComponent(JList list, Object value, int index, boolean isSelected, boolean cellHasFocus) {
					String valueString = value.toString();
					valueString = valueString.substring(valueString.lastIndexOf('.') + 1);
					return super.getListCellRendererComponent(list, valueString, index, isSelected, cellHasFocus);
				}
			});
			layoutComboBox.addItemListener(new ItemListener() {

				public void itemStateChanged(ItemEvent e) {
					if (e.getStateChange() == ItemEvent.SELECTED) {
						Class<T> clazz = (Class<T>) e.getItem();
						try {
							Layout<T> layout = getLayoutFor(clazz, getTreePresentation());
							layout.setSize(treeComponent.getSize());
							setTreeLayout(layout);
						} catch (Exception ex) {
							ex.printStackTrace();
						}
					}
				}
			});
			layoutComboBox.setSelectedItem(layoutClasses[0]);
			try {
				setTreeLayout(getLayoutFor(layoutClasses[0], getTreePresentation()));
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		return layoutComboBox;
	}

	/**
	 * @return Returns the modeBox.
	 */
	public JComboBox getMouseModeComboBox() {
		if (mouseModeBox == null) {
			mouseModeBox = new JComboBox(new String[] {"PICKING", "TRANSFORMING",});
			mouseModeBox.addItemListener(new ItemListener() {

				public void itemStateChanged(ItemEvent paramItemEvent) {
					if (mouseModeBox.getSelectedItem().equals("PICKING")) {
						treeComponent.setPickingMode(true);
						treeComponent.getTreeElementSelection().unselect();
					} else {
						treeComponent.setPickingMode(false);
						Dimension size = treeComponent.getPreferredSize();
						Rectangle rect = new Rectangle(0, 0, size.width, size.height);
						treeComponent.getTreeElementSelection().selectInRect(rect);
					}
				}
			});
		}
		mouseModeBox.setSelectedItem("PICKING");
		return mouseModeBox;
	}

	private void setTreeLayout(Layout<T> layout) {
		layout.initialize();
		revalidate();
		repaint();
	}

	private Layout<T> getLayoutFor(Class<T> layoutClass, TreePresentation<T> tree) throws Exception {
		Object[] args = new Object[] {tree};
		Constructor<T> constructor = layoutClass.getConstructor(new Class[] {TreePresentation.class});
		return ((Layout<T>) constructor.newInstance(args));
	}

	private void createTreeCanvas() {
		JScrollPane centerPanel = new JScrollPane(treeComponent);
		centerPanel.setBorder(BorderFactory.createEtchedBorder());
		add(centerPanel, BorderLayout.CENTER);
	}

	private void createControlPanel(Dimension size) {
		controls = new JPanel();
		controls.setPreferredSize(size);
		controls.setBorder(BorderFactory.createEtchedBorder());
		if (orientation == Orientation.VERTICAL) {
			controls.setLayout(new BoxLayout(controls, BoxLayout.X_AXIS));
		} else {
			controls.setLayout(new BoxLayout(controls, BoxLayout.Y_AXIS));
		}
		// controls.add(Box.createHorizontalGlue());
		JButton plus = new JButton("+");
		plus.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				treeComponent.scale(1.1f);
			}
		});
		JButton minus = new JButton("-");
		minus.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				treeComponent.scale(1 / 1.1f);
			}
		});
		JPanel zoomControls = new JPanel();
//		zoomControls.setLayout(new BoxLayout(zoomControls, BoxLayout.X_AXIS));
		zoomControls.setBorder(BorderFactory.createTitledBorder("Zoom"));
		zoomControls.add(plus);
		zoomControls.add(minus);
		controls.add(zoomControls);
		// controls.add(Box.createHorizontalGlue());

		JPanel modeControls = new JPanel();
		// modeControls.setLayout(new BoxLayout(modeControls, BoxLayout.X_AXIS));
		modeControls.setBorder(BorderFactory.createTitledBorder("MouseTool Mode"));
		modeControls.add(getMouseModeComboBox());
		controls.add(modeControls);
		// controls.add(Box.createHorizontalGlue());

		JPanel layoutPanel = new JPanel();
		layoutPanel.setBorder(BorderFactory.createTitledBorder("Tree layout"));
		layoutPanel.add(getLayoutsComboBox());
		controls.add(layoutPanel);
		// controls.add(Box.createRigidArea(new Dimension(10, 0)));
		//controls.add(Box.createHorizontalGlue());

		JPanel edgeConfig = new JPanel(new GridBagLayout());
		GridBagConstraints c = new GridBagConstraints();
		// edgeConfig.setLayout(new BoxLayout(edgeConfig, BoxLayout.X_AXIS));
		edgeConfig.setBorder(BorderFactory.createTitledBorder("Constraints"));
		boolean isWeighted = treeComponent.getTreePresentation().isWeighted();
		final JRadioButton weightedEdge = new JRadioButton("Weighted", isWeighted);
		weightedEdge.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				getTreePresentation().setWeightedEdges(weightedEdge.isSelected());
				treeComponent.revalidate();
				treeComponent.repaint();
			}
		});
		c.weightx = 0.5;
		c.fill = GridBagConstraints.HORIZONTAL;
		c.gridx = 0;
		c.gridy = 0;
		edgeConfig.add(weightedEdge, c);
		final JRadioButton labeledEdge = new JRadioButton("Labeled", false);
		labeledEdge.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				getTreePresentation().setLabeledEdges(labeledEdge.isSelected());
				treeComponent.revalidate();
				treeComponent.repaint();
			}
		});
		c.gridx = 1;
		edgeConfig.add(labeledEdge, c);
		final JRadioButton backEdges = new JRadioButton("Back Edges", false);
		backEdges.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				getTreePresentation().showBackEdges(backEdges.isSelected());
				treeComponent.revalidate();
				treeComponent.repaint();
			}
		});
		c.gridx = 0;
		c.gridy = 1; 
		c.gridwidth = 2;
		c.ipady = 25;     //This component has more breadth compared to other buttons
        c.weightx = 0.0;
		edgeConfig.add(backEdges, c);
		controls.add(edgeConfig);
		// controls.add(Box.createHorizontalGlue());

		JPanel vertexConfig = new JPanel(new GridBagLayout());
		//vertexConfig.setLayout(new BoxLayout(vertexConfig, BoxLayout.Y_AXIS));
		vertexConfig.setBorder(BorderFactory.createTitledBorder("Variables"));
		final JRadioButton scalableVertex = new JRadioButton("Scalable", false);
		scalableVertex.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				getTreePresentation().setScalableVertexes(scalableVertex.isSelected());
				treeComponent.revalidate();
				treeComponent.repaint();
			}
		});
		vertexConfig.add(scalableVertex);
		final JRadioButton labeledVertex = new JRadioButton("Labeled", true);
		labeledVertex.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				getTreePresentation().setLabeledVertexes(labeledVertex.isSelected());
				treeComponent.revalidate();
				treeComponent.repaint();
			}
		});
		vertexConfig.add(labeledVertex);
		getTreePresentation().setLabeledVertexes(labeledVertex.isSelected());
		controls.add(vertexConfig);
		// controls.add(Box.createHorizontalGlue());
		JPanel southPanel = new JPanel();
		southPanel.setLayout(new BoxLayout(southPanel, BoxLayout.X_AXIS));
		southPanel.add(Box.createVerticalStrut(10));
		southPanel.add(controls);
		southPanel.add(Box.createVerticalStrut(10));
		// southPanel.setMinimumSize(new Dimension(500, controlHeight));
		// this.setMinimumSize(new Dimension(500, 200+controlHeight));
		if (orientation == Orientation.VERTICAL) {
			add(southPanel, BorderLayout.SOUTH);
		} else {
			add(southPanel, BorderLayout.WEST);
		}
	}

	/**
	 * Returns tree presentation.
	 * 
	 * @return tree presentation
	 */
	public TreePresentation<T> getTreePresentation() {
		return this.treeComponent.getTreePresentation();
	}

	@Override
	protected void paintComponent(Graphics g) {
		Graphics2D gr = (Graphics2D) g;
		gr.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		gr.setRenderingHint(RenderingHints.KEY_COLOR_RENDERING, RenderingHints.VALUE_COLOR_RENDER_QUALITY);
		if (this.getTreePresentation() != null) {
			this.treeComponent.paint(g);
		}
	}
}
