package gui.tree.presentation;

import util.graph.Edge;
import util.graph.Vertex;
import gui.graph.layout.RandomLocations;
import gui.graph.presentation.EdgePresentation;
import gui.graph.presentation.ElementStateListener;
import gui.graph.presentation.GraphStateListener;
import gui.graph.presentation.VertexPresentation;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;
import util.tree.AbstractTree;


/**
 * This class provides a presentation for any implementation of <tt>AbstractGraph</tt>.
 * 
 * @author Mohamed Wahbi
 * 
 */
public class TreePresentation<T extends Comparable<? super T>> {

	private AbstractTree<T> tree = null;
	private Vector<VertexPresentation<T>> vertexes = null;
	private Vector<EdgePresentation<T>> edges = null;
	private Map<T, VertexPresentation<T>> vertexesMap = null;
	private Map<Edge<T>, EdgePresentation<T>> edgesMap = null;
	private Vector<GraphStateListener> listeners = null;
	private boolean backEdged;
	private boolean isWeighted;

	private final ElementStateListener elementStateListener = new ElementStateListener() {
		public void elementPositionChanged() {
			firePositionChanged();
		}
		public void elementStructureChanged() {
			fireStructureChanged();
		}
	};
	
	/**
	 * Shift from (0,0).
	 */
	public static final int SHIFT = 30;
	
	private boolean isPaused;
	
	private boolean wasStructureChangedRecieved;
	
	private boolean wasPositionChangedRecieved;
	
	/**
	 * Constructs presentation for the given graph
	 * 
	 * @param graph graph
	 * @param isDirected true if you want directed graph; false otherwise
	 * @param isWeighted true if you want weighted graph; false otherwise
	 */
	public TreePresentation(AbstractTree<T> tree, boolean backEdged, boolean isWeighted) {
		this.tree = tree;
		this.backEdged = backEdged;
		this.listeners = new Vector<GraphStateListener>();
		this.vertexes = new Vector<VertexPresentation<T>>();
		this.edges = new Vector<EdgePresentation<T>>();
		this.vertexesMap = new HashMap<T, VertexPresentation<T>>();
		this.edgesMap = new HashMap<Edge<T>, EdgePresentation<T>>();
		//TODO
		Map<Vertex<T>, Point2D> locations = new RandomLocations<Vertex<T>>(tree.getVertexes(), new Dimension(500, 400));
		for (Vertex<T> vertex : this.tree.getVertexes()) {
			Point2D location = locations.get(vertex);
			VertexPresentation<T> vp = new VertexPresentation<T>(vertex, location);//getCenterForVertex(i));
			this.vertexes.add(vp);
			this.vertexesMap.put(vertex.getData(), vp);
		}
		for (Vertex<T> vertex : this.tree.getVertexes()) {
			for (Edge<T> edge : vertex.getOutgoingEdges()) {
				if (this.edgesMap.containsKey(edge)) {
					continue;
				}
				VertexPresentation<T> source = this.vertexesMap.get(vertex.getData());
				Vertex<T> opposite = edge.getOppositeVertex(vertex);
				VertexPresentation<T> target = this.vertexesMap.get(opposite.getData());
				EdgePresentation<T> ep = null;
				if(edge.isBackEdge()) {
					ep = new BackEdgePresentation<T>(edge, source, target, isWeighted);
				} else {
					ep = new TreeEdgePresentation<T>(edge, source, target, isWeighted);
				}
				this.edges.add(ep);
				this.edgesMap.put(edge, ep);
			}
		}
		this.isPaused = false;
		this.wasPositionChangedRecieved = false;
		this.wasStructureChangedRecieved = false;
		this.isWeighted = isWeighted;
		this.setWeightedEdges(isWeighted);
	}
	
	/**
	 * Constructs new graph presentation and new graph
	 * 
	 * @param isDirected true if you want directed graph; false otherwise
	 * @param isWeighted true if you want weighted graph; false otherwise
	 */
	public TreePresentation(boolean backEdged, boolean isWeighted) {
		this.tree = new AbstractTree<T>();
		this.isWeighted = isWeighted;
		this.backEdged = backEdged;
		this.listeners = new Vector<GraphStateListener>();
		this.vertexes = new Vector<VertexPresentation<T>>();
		this.edges = new Vector<EdgePresentation<T>>();
		this.vertexesMap = new HashMap<T, VertexPresentation<T>>();
		this.edgesMap = new HashMap<Edge<T>, EdgePresentation<T>>();
	}

	/**
	 * Adds the specified GraphStateListaner to receive notification about state change.
	 * 
	 * @param listener
	 */
	public void addGraphStateListener(GraphStateListener listener) {
		this.listeners.add(listener);
	}
	
	/**
	 * Removes the specified GraphStateListaner from this presentation.
	 * 
	 * @param listener
	 */
	public void removeGraphStateListener(GraphStateListener listener) {
		this.listeners.remove(listener);
	}
	
	private void firePositionChanged() {
		if (this.isPaused) {
			this.wasPositionChangedRecieved = true;
		} else {
			for (GraphStateListener listener : this.listeners) {
				listener.positionChanged();
			}
		}
	}
	
	private void fireStructureChanged() {
		if (this.isPaused) {
			this.wasStructureChangedRecieved = true;
		} else {
			for (GraphStateListener listener : this.listeners) {
				listener.structureChanged();
			}
		}
	}
	
	/**
	 * Pauses all changes firing until resume is not called.
	 *
	 */
	public void pauseFiringChanges() {
		this.isPaused = true;
	}
	
	/**
	 * Resumes firing changes.
	 *
	 */
	public void resumeFiringChanges() {
		this.isPaused = false;
		if (this.wasPositionChangedRecieved) {
			firePositionChanged();
			this.wasPositionChangedRecieved = false;
		}
		if (this.wasStructureChangedRecieved) {
			fireStructureChanged();
			this.wasStructureChangedRecieved = false;
		}
	}

	/**
	 * Paints this graph using given graphics
	 * 
	 * @param gr graphics
	 * @param bgColor background color
	 */
	public void paint(Graphics2D gr, Color bgColor) {
		for (EdgePresentation<T> ep : this.edges) {
			ep.paint(gr, bgColor);
		}
		for (VertexPresentation<T> vp : this.vertexes) {
			vp.paint(gr);
		}
	}
//	
//	/**
//	 * Adds this vertex presentation to graph presentation.
//	 * 
//	 * @param vp vertex presentation
//	 */
//	public void addVertex(VertexPresentation<T> vp) {
//		this.vertexes.add(vp);
//		this.vertexesMap.put(vp.getVertex().getData(), vp);
//		vp.addElementStateListener(this.elementStateListener);
//		fireStructureChanged();
//	}
//	
	/**
	 * Removes this vertex presentation from graph presentation.
	 * 
	 * @param vp vertex presentation
	 */
	public void removeVertex(VertexPresentation<T> vp) {
		for (Edge<T> edge : vp.getVertex().getOutgoingEdges()) {
			EdgePresentation<T> ep = this.edgesMap.remove(edge);
			this.edges.remove(ep);
		}
		for (Edge<T> edge : vp.getVertex().getIncomingEdges()) {
			EdgePresentation<T> ep = this.edgesMap.remove(edge);
			this.edges.remove(ep);
		}
		//this.tree.removeVertex(vp.getVertex());
		this.vertexes.remove(vp);
		this.vertexesMap.remove(vp);
		vp.removeElementStateListener(this.elementStateListener);
		fireStructureChanged();
	}
//	
//	/**
//	 * Adds this edge presentation to graph presentation.
//	 * 
//	 * @param ep edge presentation
//	 */
//	public void addEdge(BackEdgePresentation<T> ep) {
//		this.graph.addEdge(ep.getEdge());
//		this.edges.add(ep);
//		this.edgesMap.put(ep.getEdge(), ep);
//		ep.addElementStateListener(this.elementStateListener);
//		fireStructureChanged();
//	}
//	
//	/**
//	 * Removes this edge presentation from graph presentation.
//	 * 
//	 * @param ep edge presentation
//	 */
//	public void removeEdge(BackEdgePresentation<T> ep) {
//		this.graph.disconnect(ep.getEdge().getSource(), ep.getEdge().getTarget());
//		this.edges.remove(ep);
//		this.edgesMap.remove(ep);
//		ep.removeElementStateListener(elementStateListener);
//		fireStructureChanged();
//	}

	/**
	 * Checks whether given vertexes are connected.
	 * 
	 * @param sourcePr source
	 * @param destinPr target
	 * @return true if they are; false otherwise
	 */
	public boolean areConnected(VertexPresentation<T> sp, VertexPresentation<T> tp) {
		return this.tree.areConnected(sp.getVertex(), tp.getVertex());
	}
	
	/**
	 * Returns preferred graph view size
	 * 
	 * @return size
	 */
	public Dimension getPreferredSize() {
		double resultX = 0;
		double resultY = 0;
		for (VertexPresentation<T> vp : this.vertexes) {
			resultX = Math.max(resultX, vp.getX() + VertexPresentation.WIDTH + 1);
			resultY = Math.max(resultY, vp.getY() + VertexPresentation.HEIGHT + 1);
		}
		return new Dimension((int) resultX, (int) resultY);
	}

	/**
	 * Returns model graph of this presentation.
	 * 
	 * @return graph
	 */
	public AbstractTree<T> getTree() {
		return this.tree;
	}
	
	/**
	 * Returns set of vertexes presentations.
	 * 
	 * @return set of vertexes presentations
	 */
	public Vector<VertexPresentation<T>> getVertexPresentations() {
		return this.vertexes;
	}
	
	/**
	 * Returns set of vertexes.
	 * 
	 * @return set of vertexes
	 */
	public List<T> getVertexes() {
		return new ArrayList<T>(this.vertexesMap.keySet());
	}
	
	/**
	 * Returns set of edges presentations.
	 * 
	 * @return set of edges presentations
	 */
	public Vector<EdgePresentation<T>> getEdgePresentations() {
		return this.edges;
	}
	
	/**
	 * Returns the set of edges.
	 * 
	 * @return set of edges
	 */
	public Set<Edge<T>> getEdges() {
		return this.edgesMap.keySet();
	}
	
	/**
	 * Returns presentations of edges incident to given vertex presentation.
	 * 
	 * @param vp vertex presentation
	 * @return edges incident to given vertex
	 */
	public Set<EdgePresentation<T>> getIncidendEdgePresentations(VertexPresentation<T> vp) {
		Set<EdgePresentation<T>> result = new HashSet<EdgePresentation<T>>();
		for (Edge<T> edge : vp.getVertex().getOutgoingEdges()) {
			result.add(this.edgesMap.get(edge));
		}
		for (Edge<T> edge : vp.getVertex().getIncomingEdges()) {
			result.add(this.edgesMap.get(edge));
		}
		return result;
	}
	
	/**
	 * Checks whether this presentation is weighted.
	 * 
	 * @return true if it is; false otherwise
	 */
	public boolean isWeighted() {
		return this.isWeighted;
	}
	
	/**
	 * Checks whether this presentation is directed.
	 * 
	 * @return true if it is; false otherwise
	 */
	public boolean isBackEdged() {
		return this.backEdged;
	}
	
	/**
	 * Returns edge presentation for given edge.
	 * 
	 * @param edge edge
	 * @return edge presentation
	 */
	public EdgePresentation<T> getPresentationForEdge(Edge<T> edge) {
		return this.edgesMap.get(edge);
	}
	
	/**
	 * Returns vertex presentation for given vertex.
	 * 
	 * @param vertex vertex
	 * @return vertex presentation
	 */
	public VertexPresentation<T> getPresentationForVertex(Vertex<T> vertex) {
		return this.vertexesMap.get(vertex);
	}

	/**
	 * Returns the number of vertexes belonging to the this graph presentation.
	 * 
	 * @return  the number of vertexes in the graph
	 */
	public int getVertexCount() {
		return this.getVertexes().size();
	}

	/**
	 * @return the locations
	 */
	public Map<T, Point2D> getLocations() {
		Map<T, Point2D> locations = new HashMap<T, Point2D>();
		for(VertexPresentation<T> vertex : this.vertexes) {
			locations.put(vertex.getData(), vertex.getCenter());
		}
		return locations;
	}

	public void scale(double scale) {
		for(VertexPresentation<T> vertex : vertexes) {
			vertex.scale(scale);
		}
	}

	public void initialize() {
		for(VertexPresentation<T> vertex : vertexes) {
			vertex.initialize();
		}
	}

	public void setWeightedEdges(boolean weighted) {
		this.isWeighted = weighted;
		for(EdgePresentation<T> edge : edgesMap.values()) {
			edge.setWeighted(weighted);
		}
	}

	public void setLabeledEdges(boolean labeled) {
		for(EdgePresentation<T> edge : edgesMap.values()) {
			edge.setLabeled(labeled);
		}
	}

	public void setAssignedVertexes(boolean assigned) {
		for(VertexPresentation<T> vertex : vertexesMap.values()) {
			vertex.setLabeled(assigned);
		}
	}

	public void setLabeledVertexes(boolean labeled) {
		for(VertexPresentation<T> vertex : vertexesMap.values()) {
			vertex.setLabeled(labeled);
		}
	}

	public void setScalableVertexes(boolean scalable) {
		for(VertexPresentation<T> vertex : vertexesMap.values()) {
			vertex.setScalable(scalable);
		}
	}

	public void showBackEdges(boolean show) {
		for(EdgePresentation<T> edge : edgesMap.values()) {
			if(edge.getEdge().isBackEdge()) {
				edge.show(show);
			}
		}
	}

	public void removeSelectedVertex() {
		Vector<VertexPresentation<T>> selectedVertexes = new Vector<VertexPresentation<T>>();
		for(VertexPresentation<T> vp : this.vertexes) {
			if(vp.isSelected()) {
				selectedVertexes.add(vp);
			}
		}
		for(VertexPresentation<T> vp : selectedVertexes) {
			this.removeVertex(vp);
		}
	}
}
