package kernel;

import java.lang.management.ManagementFactory;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import kernel.com.CommunicationInterface;
import kernel.com.Message;
import kernel.com.TerminateMsgHandler;
import kernel.events.AgentChangeStateEvent;
import kernel.events.AgentEvent;
import kernel.events.AgentEventListener;
import kernel.events.AgentFinishedEvent;
import kernel.events.PropagationEngine;
import kernel.order.AgentOrdering;
import kernel.vars.integer.IntExternalVar;
import kernel.vars.integer.IntVar;
import kernel.vars.integer.IntVariable;
import simulation.metric.Metric;
import simulation.metric.MetricListener;
import simulation.metric.StatisticsAgentManager;
import simulation.parser.PVariable;
import util.BooleanLock;
import util.Commands;
import util.HashList;
import util.IntArray;
import util.Logger;
import util.VarAssignment;
import util.VarEventMask;
import util.graph.UndirectedGraph;

/**
 * An abstract class that represents a Distributed Constraint Reasoning agent. 
 * Two kind of agent extend it: {@link simple.SimpleAgent} and complex agent
 * 
 * @author Mohamed Wahbi
 */
public abstract class AbstractAgent extends Thread {

	protected HashMap<VarEventMask, IntArray> eventsConfig = null;

	protected Map<VarIdentifier, IntVariable> variables;

	public Map<String, Metric> auxiliaryStatistics = null;
	/**
	 * the agent identifier, it is a unique id in the problem
	 */
	public int agentId;
	/**
	 * The name of this agent
	 */
	public String name;
	/**
	 * A description of this agent. When we use a simple model (i.e. one agent per variable) two agent have the same description are considered locals
	 * (ie. pseudo agents).
	 */
	public String description;

	/**
	 *The agent communicates via its attribute <code>Communicator</code>
	 */
	public Communicator communicator;
	/**
	 *A transmission constraint. It stores all the external constraints, it is also a listener for these constraints. Treatment of all constraint
	 * events are taken by it.
	 */
	public ConstraintsHandler constraintHandler;

	/**
	 * An agent ordering field. Used to compute and store current order. It is used as comparator for external variables in the <code>AgentView</code>
	 */
	public AgentOrdering agentOrder;
	/**
	 * Each agent has an <code>AgentView</code> on which it stores current external variable assignment (i.e. Context).
	 */
	public AgentView agentView;
	/**
	 * A message handler which process the termination detection messages. It is also an agent event listener <code>TerminateMsgHandler</code>
	 */
	public TerminateMsgHandler terminateMsgHandler;
	/**
	 *A flag that indicates whether the distributed search was started or not yet.
	 */
	protected boolean disSearchStarted = false;
	/**
	 *A flag that indicates whether the agent was finished or not yet.
	 */
	protected boolean terminate = false;
	/**
	 * A set of <code>AgentObserver</code> that is plugged on this agent.
	 */
	protected HashSet<AgentObserver> observers = null;
	/**
	 * A set of <code>AgentEventListener</code> that is plugged on this agent.
	 */
	protected HashSet<AgentEventListener> listeners = null;
	/**
	 * To store the end state of the algorithm. It can be affected to SUCCES, FAILURE, ERROR
	 */
	public String endState;

	/**
	 * A flag indicate to the agent if it is interrupted
	 */
	private boolean interrupted;

	/**
	 * This attribute stores the list of id of agentss that control the set of local variables belonging to the same problem as the local variable
	 * controlled by this agent. It is used when we have a complex local problem and we use the simplified model where each variable is controlled by
	 * a single agent.
	 * 
	 * @see simple.SimpleAgent
	 */
	public HashList<VarIdentifier> localAgents = null;

	public Logger logger = null;

	protected double cpuTime;

	/**
	 * Constructs a new AbstractAgent.
	 * 
	 * @param agentId unique id for the agent
	 * @param agName name for the agent (unique for clarity of debugging)
	 * @param description unique description of the agent
	 */
	public AbstractAgent(int agId, String name, String description) {
		super();
		this.agentId = agId;
		this.name = name;
		this.description = description;
		this.agentOrder = new AgentOrdering(this);
		this.agentView = new AgentView(this);
		this.localAgents = new HashList<VarIdentifier>();
		this.observers = new HashSet<AgentObserver>();
		this.listeners = new HashSet<AgentEventListener>();
		this.listeners.add(new StatisticsAgentManager(this));
		this.communicator = new Communicator(this);
		this.constraintHandler = new ConstraintsHandler(this);
		this.eventsConfig = new HashMap<VarEventMask, IntArray>();
		this.variables = new HashMap<VarIdentifier, IntVariable>();
		this.auxiliaryStatistics = new HashMap<String, Metric>();
		for (VarEventMask vt : VarEventMask.values()) {
			this.eventsConfig.put(vt, new IntArray());
		}
	}

	/**
	 * To add a new agent event listener
	 * 
	 * @param listener
	 */
	public void addAgentEventListener(AgentEventListener listener) {
		this.listeners.add(listener);
	}

	/**
	 * Add an external constraint to the agent. Note that all external agent that are in the scope of the given constraint are added as neighbors
	 * 
	 * @param exC
	 *            An external constraint to add
	 */
	public void addConstraint(Constraint exC) {
		this.constraintHandler.add(exC);
		for (int i = 0 ; i < exC.getNbExternalVars() ; i++) {
			VarIdentifier varId = exC.getExternalVar(i).getVarIdentifier();
			this.communicator.addNeighbor(varId);
		}
	}

	/**
	 * Register given agent as a local pseudo agent in order to ignore counting message sent to this agent
	 * 
	 * @param agId
	 */

	public void addLocalAgent(VarIdentifier varId) {
		if (this.localAgents == null)
			this.localAgents = new HashList<VarIdentifier>();
		this.localAgents.add(varId);
	}

	/**
	 * This method must be called whenever a new agent protocol is created. 
	 * It adds all necessary message handlers. In order to implement this function you
	 * must use {@link Communicator#addMsgHandler(com.MessageHandler)}
	 */
	protected abstract void addMsgHandlers();

	/**
	 * This method must be called whenever a new agent protocol is created. 
	 * It initializes the agentView data structure by some variables of other agents. 
	 * @see kernel.AgentView
	 */
	public abstract void initAgentView();

	/**
	 * Adds a observer to this agent.
	 */
	public void addObserver(AgentObserver observer) {
		this.observers.add(observer);
		observer.setAgent(this);
	}

	protected void addVariable(VarIdentifier varId, IntVariable var) {
		this.variables.put(varId, var);
	}

	/**
	 * This function processes the first message with the given type if it already exists in the mail box and returns true. Otherwise, the agent
	 * remains blocked until it is interrupted or the expected message or a termination message arrives.
	 * {@link #blockingReceiveMessage(String msgType, long msTimeout)}
	 */
	public boolean blockingReceiveMessage(String msgType, long msTimeout) throws InterruptedException, StopException {
		return this.communicator.blockingReceiveMessage(msgType, msTimeout);
	}

	/**
	 * This function loops over all neighbors and send them the given message. It calls {@link #sendMessage(Message)} for each neighbor
	 * 
	 * @param msg : message to send
	 */
	public void broadcastMessageToNeighbors(Message msg) {
		for (VarIdentifier neighborId : this.communicator.getNeighbors()) {
			Message message = msg.clone();
			message.setReceiverId(neighborId);
			this.sendMessage(message);
		}
	}

	/**
	 *By default external variables are active on Instantiate event
	 */
	public abstract void configurePropagationEngine();

	protected boolean containsVariable(VarIdentifier varId) {
		return this.variables.containsKey(varId);
	}

	/**
	 * Notifies all listeners that an event is occurred
	 * 
	 * @param agentEvent
	 *            : the event to fire
	 */
	public void fireAgentEvent(AgentEvent agentEvent) {
		for (AgentEventListener listener : listeners) {
			listener.handleAgentEvent(agentEvent);
		}
	}

	/**
	 * Get the value of agentId
	 * 
	 * @return the value of agentId
	 */
	public int getAgentId() {
		return agentId;
	}

	/**
	 * Get the agent name
	 * 
	 * @return the value of name
	 */
	public String getAgentName() {
		return this.name;
	}

	public UndirectedGraph<VarIdentifier> getAgentsGraph() {
		return this.agentOrder.getAgentsGraph();
	}

	/**
	 * 
	 * @return a reference to the object that represent the communication system.
	 */
	protected CommunicationInterface getCommunicationInterface() {
		return this.communicator.getCommunicationInterface();
	}

	/**
	 * @return the unique description of this agent
	 */
	public String getDescription() {
		return this.description;
	}

	/**
	 * Get the value of mailBoxLock
	 * 
	 * @return the value of mailBoxLock
	 */
	public BooleanLock getMailBoxLock() {
		return this.communicator.mailBoxLock;
	}

	public abstract int getMaxDomainSize();

	public abstract int getNbVars();

	/**
	 * check if distributed search is already started
	 * 
	 * @return true when distributed search is already started, otherwise false
	 */

	public HashList<VarIdentifier> getNeighborsOf(VarIdentifier varId) {
		return this.agentOrder.getNeighborsOf(varId);
	}

	public HashList<VarIdentifier> getOrder() {
		return this.agentOrder.getOrder();
	}

	//
	// public void addLink(VarIdentifier varId){
	// if(this.communicator.isNeighbor(varId.getAgentId())){
	// this.makeExternalVar(new VarAssignment(varId, Integer.MIN_VALUE));
	// this.communicator.addNeighbor(varId.getAgentId());
	// }
	// }

	public PropagationEngine getPropagationEngine() {
		return this.constraintHandler.getPropagationEngine();
	}

	protected Variable getVariable(VarIdentifier varId) {
		return this.variables.get(varId);
	}

	// /**
	// * All implementing classes should call this function once at the end of
	// * each cycle. This is used for implementing a cycle cut-off or for
	// * calculating the Cycle-Based Runtime comparison metric.
	// */
	// protected void incrementCycle() {
	// for (AgentObserver observer : observers)
	// if (observer instanceof MetricListener)
	// ((MetricListener) observer).handleIncrementCycle();
	// }

	/**
	 * This function Called once after each constraint check.
	 * 
	 * @see ConstraintsHandler#incrementNumConstraintChecks()
	 * 
	 */
	protected void incrementNumConstraintChecks(int n) {
		for (AgentObserver observer : observers)
			if (observer instanceof MetricListener)
				((MetricListener) observer).handleIncrementNbConstraintChecks(n);
	}
	
	/**
	 * Called for algorithm initialization. All class modeling distributed algorithms extend this agent and then call this method when they are
	 * started.
	 * 
	 * @throws StopException
	 * @throws InterruptedException
	 */
	protected void initAlgorithm() throws StopException {
		this.communicator.addMsgHandler(this.terminateMsgHandler);
		this.addAgentEventListener(terminateMsgHandler);
		this.communicator.addMsgHandler(this.agentOrder);
		this.addMsgHandlers();
		this.agentOrder.startDisAO();
		this.agentOrder.initialize();
		this.communicator.initialize();
		this.agentView.initialize();
		this.initAgentView();
		this.configurePropagationEngine();
	}

	/**
	 * to initialize all observers.
	 */
	private void initObservers() {
		for (AgentObserver observer : observers)
			observer.initialize();
	}

	/**
	 * Interrupts this agent. It is strictly used when execution is limited by a cut-off (max cycle, max CCCs, ...)
	 * <p>
	 * If this agent is blocked in an invocation of the {@link #waitForArrivalMessages(long)} it is notified to break out from the waiting and the
	 * <i>interrupted status</i> is setting to true;
	 */
	public synchronized void interrupt() {
		this.interrupted = true;
		this.getMailBoxLock().setValue(true);
	}

	public boolean isDisSearchStarted() {
		return this.disSearchStarted;
	}

	/**
	 * Tests whether this agent has been finished.
	 * 
	 * @return <code>true</code> if this agent has been finished; <code>false</code> otherwise.
	 */
	public boolean isFinished() {
		return this.terminate;
	}

	/**
	 * Tests whether this agent has been interrupted.
	 * <p>
	 * The <i>interrupted status</i> of the agent is unaffected by this method.
	 * 
	 * @return <code>true</code> if this agent has been interrupted; <code>false</code> otherwise.
	 * @see #interrupt()
	 */
	public synchronized boolean isInterrupted() {
		return this.interrupted;
	}

	/**
	 * DCOP weather the agent is waiting for message.
	 * 
	 * @return <code> true</code> if the agent wait for arrival messages {@link #waitForArrivalMessages(long)}, Otherwise <code>false</code>.
	 */
	public boolean isWaitingForMessage() {
		return this.communicator.isWaitingForMessage();
	}

	public void log(String s, int level) {
		if (this.logger != null)
			logger.printToLogln(s, level);
	}
//
//	public void logDomains() {
//		if (Logger.LOG_DEBUG) {
//			this.logMyDomain();
//			//TODO
////			Iterator<VarIdentifier> it = this.variables.keySet().iterator();
////			while (it.hasNext()) {
////				Variable v = this.variables.get(it.next());
////				this.log(v + " :: " + v.getDomain().toString(), Logger.FINE_LEVEL);
////			}
//		}
//	}

	// /////////////////////////////////////////////////////////////////////////////////////////////

	public abstract void logDomains();

	public abstract IntExternalVar makeExternalVar(IntVar v);

	public abstract IntExternalVar makeExternalVar(VarAssignment v);

	public abstract IntExternalVar makeExternalVar(VarIdentifier vId);

	public abstract IntExternalVar makeExternalVar(VarIdentifier vId, int min, int max);

	public abstract IntExternalVar makeExternalVar(String name, VarIdentifier varId, Integer[] values); 
	
	public abstract IntExternalVar makeExternalVar(PVariable variable); 


	/**
	 * This function inserts the given message to the agent mailbox, It also notify the agent that a message arrived. Notification is done by setting
	 * the state of the mailbox lock to true.
	 * 
	 * @param msg
	 *            : the message that just arrived
	 */
	public synchronized void notifyReception(Message msg) {
		this.communicator.notifyReception(msg);
	}

	/**
	 * to process all currently received messages. The agent execute a loop processing messages until the mailbox becomes empty.
	 * 
	 * @throws StopException
	 *             thrown when agent terminates or when it receives a terminate message
	 */
	public void processSomeMessages() throws StopException {
		this.communicator.processMessages();
	}

	/**
	 * To remove an agent event listener
	 * 
	 * @param listener
	 */
	public void removeAgentEventListener(AgentEventListener listener) {
		this.listeners.remove(listener);
	}

	/**
	 * Removes a observer from this agent.
	 */
	public void removeObserver(AgentObserver observer) {
		if (this.observers.contains(observer)) {
			this.observers.remove(observer);
		}
	}

	// /////////////////////////////////////////////////////////////////////////////////////////////


	/**
	 * @return <b>long</b> : Returns the total CPU time for the current agent (thread) in nanoseconds.
	 */
	public double getCPUTime() {
		double endCpuTime = ManagementFactory.getThreadMXBean().getCurrentThreadCpuTime();
		this.cpuTime = endCpuTime - this.cpuTime;
		return this.cpuTime;
	}

	/**
	 * <code>run</code> method called when starting the thread that control the execution of this agent
	 */
	public void run() {
		this.cpuTime = ManagementFactory.getThreadMXBean().getCurrentThreadCpuTime();
		boolean old = Logger.LOG_DEBUG;
		if (Logger.LOG_DEBUG) {
			//this.logger = new Logger(this.name + ".txt");
			this.logger = new Logger("Agent"+this.agentId+ "-" + this.name + ".txt");
			this.logger.printToLogln("Agent"+this.agentId +"\t"+ this.name, Logger.LEVEL_TRACE_MESSAGE);
			this.logger.printToLogln("class: "+getClass()+" and protocol: "+toString(), Logger.LEVEL_TRACE_MESSAGE);
		}
		try {
			this.initAlgorithm();
			this.initObservers();
			this.disSearchStarted = true;
			this.fireAgentEvent(new AgentChangeStateEvent(this, AgentChangeStateEvent.STATE_RUNNING));
			this.startDistributedSearch();
			while (!this.isFinished()) {
//				long tID = getId();
//				this.cpuTime = this.mxbean.getThreadCpuTime(tID);
				try {
					if (this.waitForArrivalMessages(0)) {
						this.processSomeMessages();
					}
					this.log("agent ID     : " + this.agentId+"  --> neighbours:"+this.communicator.getNeighbors(), Logger.FINE_LEVEL);
					this.log("number of constraints      : " + this.constraintHandler.getNumConstraints(), Logger.FINE_LEVEL);
					this.log("number of message in queue : " + this.communicator.mailBox.size(), Logger.FINE_LEVEL);
				} catch (java.lang.OutOfMemoryError er) {
					System.err.println("agent ID     : " + this.agentId+"  --> neighbours:"+this.communicator.getNeighbors());
					System.err.println("number of constraints      : " + this.constraintHandler.getNumConstraints());
					System.err.println("number of message in queue : " + this.communicator.mailBox.size());
					Runtime rt = Runtime.getRuntime();
					java.text.DecimalFormat df = new java.text.DecimalFormat("###,###,###,###");
					System.err.println("used = " + df.format(rt.totalMemory() - rt.freeMemory()) + " free = " + rt.freeMemory() + " total = "
							+ rt.totalMemory() + " max = " + rt.maxMemory());
					er.printStackTrace();
				}
			}
		} catch (StopException e) {
			this.stopRunning(e.getMessage());
		} catch (Error e) {
			Logger.LOG_DEBUG = true;
			Logger log = new Logger("err" + "_" + this.agentId);
			e.printStackTrace(log.getPrintWriter());
			e.printStackTrace();
			log.close();
			log.printToLogln(e.getMessage(), Logger.INFO_LEVEL);
			this.stopRunning(Commands.ERROR + ": " + e.getMessage());
		} catch (Exception e) {
			Logger.LOG_DEBUG = true;
			Logger log = new Logger("exept" + "_" + this.agentId);
			e.printStackTrace(log.getPrintWriter());
			e.printStackTrace();
			log.close();
			log.printToLogln(e.getMessage(), Logger.INFO_LEVEL);
			this.stopRunning(Commands.ERROR + ": " + e.getMessage());
		}
		// this.log(this.name + " : I have terminated!\n" + "The end state is : "+ this.endState, Logger.INFO_LEVEL);
		if (this.logger != null) {
			this.logger.close();
		}
		Logger.LOG_DEBUG = old;
	}

	/**
	 * Sends a message to another agent using the communication system. To ensure independence between the communication system and the dischoco
	 * platform, the agent assumes that the communication system is perfect. This function is equivalent to
	 * <code> {@link Communicator#sendMessage(Message)}</code> It is suggested that agents always use this function to send messages (as opposed to
	 * sending them directly through the Communicator) since this function will send calls to any observers in this agent.
	 * 
	 * @param msg
	 *            : message to send
	 */
	public void sendMessage(Message msg) {
		this.communicator.sendMessage(msg);
	}

	/**
	 * Set the name of the agent to the given one.
	 * 
	 * @param agName
	 */

	public void setAgentsName(String agName) {
		this.name = agName;
	}

	/**
	 * 
	 * Set the Communication system the the given one. {@link #setCommunicationInterface(CommunicationInterface)}
	 * 
	 * @param com
	 */
	public void setCommunicationInterface(CommunicationInterface com) {
		this.communicator.com = com;
	}

	public void setEndState(String state) {
		this.endState = state;
	}

	/**
	 * This methods returns a string representing the local solution of the agent
	 * 
	 * @return A string which represents the local solution
	 */
	public abstract String solutionToString();

	/**
	 * This is where the agent should implement all of its code. all agents (i.e. algorithm) that extend this agent must call this method to start
	 * distributed search.
	 * 
	 * @throws StopException
	 *             thrown when agent terminates or when it receives a terminate message
	 */
	protected abstract void startDistributedSearch() throws StopException;

	/**
	 * Called by the agent to stop its running loop; It fires a new <code> AgentFinishedEvent </code> event
	 * 
	 * @param endState
	 *            the state with which the algorithm terminate
	 */
	public void stopRunning(String endState) {
		this.setEndState(endState);
		this.communicator.waitingForMessage = true;
		this.fireAgentEvent(new AgentFinishedEvent(this, endState));
		Thread.yield();
		this.terminate = true;
	}

	/**
	 * Causes the agent to wait until a message arrives or until it was interrupted ({@link #isInterrupted()}). When a message arrives the agent is
	 * notified by changing the state of it {@link Communicator#mailBoxLock} to true. The agent can be also interrupted simply by setting it
	 * mailBoxLock to true.
	 * 
	 * @param msTimeout
	 * @return
	 * @throws StopException
	 */
	public boolean waitForArrivalMessages(long msTimeout) throws StopException {
		boolean res;
		try {
			res = this.communicator.waitForArrivalMessages(msTimeout);
		} catch (InterruptedException e) {
			throw new StopException(Commands.ERROR + " " + e.getMessage());
		}
		if (this.interrupted) { throw new StopException(Commands.ERROR + " " + "Interrupted"); }
		return res;
	}

	public HashSet<AgentObserver> getAgentObservers() {
		return this.observers;
	}

	public abstract Collection<Metric> getAuxiliaryStatistics();

	/**
	 * @return
	 */
	public long getNbConstraintChecks() {
		return 0;
	}
	
//	/**
//	 * @return
//	 */
//	public long getCpuTime() {
////		System.out.println(this.getAgentId()+"\t: start : "+this.cpuTime);
//		long endCPU = this.mxbean.getCurrentThreadCpuTime();
//		//.getThreadCpuTime(getId());
//		//ManagementFactory.getRuntimeMXBean().getUptime();
//		//ManagementFactory.getThreadMXBean().getThreadCpuTime(super.getId());
////		System.out.println(this.getAgentId()+"\t: end   : "+endCPU);
//		long tID = Thread.currentThread().getId();
//		System.out.println("get   tID : "+tID);
//		this.cpuTime = endCPU - this.cpuTime;
//		return this.cpuTime;
////		return ManagementFactory.getThreadMXBean().getCurrentThreadCpuTime();
//	}
	
	/**
	 * @return
	 */
	public abstract VarIdentifier getVarIdentifier();
	
	/**
	 * @param neighbor
	 */
	public void addNeighbor(VarIdentifier neighbor) {
		communicator.addNeighbor(neighbor);
	}

	/**
	 * Makes unary constraint X = cste
	 * 
	 * @param cste A constant
	 */
	public abstract Constraint eqC(int cste);

	/**
	 * Makes unary constraint X != cste
	 * 
	 * @param cste A constant
	 */
	public abstract Constraint neqC(int cste);

	/**
	 * Makes unary constraint X <= cste
	 * 
	 * @param cste A constant
	 */
	public abstract Constraint leqC(int cste);

	/**
	 * Makes unary constraint X >= cste
	 * 
	 * @param cste A constant
	 */
	public abstract Constraint geqC(int cste);

	/**
	 * Makes unary constraint X < cste with the given cost.
	 * 
	 * @param cste A constant
	 */
	public Constraint ltC(int cste) {
		return this.leqC(cste - 1);
	}

	/**
	 * Makes unary constraint X > cste
	 * 
	 * @param cste A constant
	 */
	public Constraint gtC(int cste) {
		return this.geqC(cste + 1);
	}

	// //////////////////////////////////////////////////////////////////////////
	/**
	 * Make the constraint X = Y when X is the internal variable, Y is the given external variable <code>y</code>.
	 */
	public abstract Constraint eqY(IntExternalVar y);

	/**
	 * Make the constraint X <= Y when X is the internal variable, Y is the given external variable <code>y</code>.
	 */
	public abstract Constraint leqY(IntExternalVar y);

	/**
	 * Make the constraint X >= Y when X is the internal variable, Y is the given external variable <code>y</code>.
	 */
	public abstract Constraint geqY(IntExternalVar y);

	/**
	 * Make the constraint X < Y when X is the internal variable, Y is the given external variable <code>y</code>.
	 */
	public abstract Constraint ltY(IntExternalVar y);

	/**
	 * Make the constraint X > Y when X is the internal variable, Y is the given external variable <code>y</code>.
	 */
	public abstract Constraint gtY(IntExternalVar y);

	/**
	 * Make the constraint X = |Y| when X is the internal variable, Y is the given external variable <code>y</code>.
	 */
	public abstract Constraint eqAbsY(IntExternalVar y);

	/**
	 * Make the constraint X = |Y + C| when X is the internal variable, Y is the given external variable <code>y</code> and C is a positive constant.
	 */
	public abstract Constraint eqAbsYplusC(IntExternalVar y, int c);

	/**
	 * Make the constraint X != Y when X is the internal variable, Y is the given external variable <code>y</code>.
	 */
	public abstract Constraint neqY(IntExternalVar y);

	/**
	 * 
	 * @param name
	 * @param v0
	 * @param values
	 * @param weights
	 * @return
	 */
	public abstract Constraint makeSoftRelation(String name, PVariable v0, Integer[] values, Integer[] weights);

	/**
	 * 
	 * @param name
	 * @param feasible
	 * @param v0
	 * @param v1
	 * @param tuples
	 * @param transpose
	 * @return
	 */
	public abstract Constraint makeConstraintTable(String name, boolean feasible, PVariable v0, PVariable v1, int[][] tuples, boolean transpose);

	/**
	 * 
	 * @param name
	 * @param v0
	 * @param v1
	 * @param weights
	 * @param transpose
	 * @return
	 */
	public abstract Constraint makeSoftBinRelation(String name, PVariable v0, PVariable v1, Integer[] weights, boolean transpose);

	/**
	 * @param name
	 * @param v0
	 * @param v1
	 * @param tuples
	 * @param weights
	 * @param defaultCost
	 * @param transpose
	 */
	public abstract Constraint makeWeightedTuplesRelation(String name, PVariable v0, PVariable v1, int[][] tuples, Integer[] weights, int defaultCost, boolean transpose);
}
