package kernel;

import java.util.HashSet;
import java.util.SortedSet;
import java.util.TreeSet;

import kernel.events.ConstraintEvent;
import kernel.vars.integer.IntVar;
import util.Instantiation;

/**
 * An abstract class for all implementations of constraints (listeners)
 */
public abstract class AbstractConstraint implements Constraint {
	/**
	 * the unique name of the constraint, used as reference
	 */
	protected String name;
	/**
	 * The index of the constraint. It is the position of this constraint in the list of all constraints of an agent
	 */
	protected int index;
	/**
	 * a field for storing the type of the constraint.
	 */
	protected String type;
	/**
	 * a field for indicate whether the constraint is private or not
	 */
	protected boolean fprivate = false;
	/**
	 * a field for storing the weight of the constraint. It is used for Max-DisCSP, DCOP as cost of the constraint
	 */
	protected int weight = 0;

	/**
	 * a field for indicate whether the constraint is posted or not
	 */
	protected boolean posted = false;
	/**
	 * A reference to the constraint handler
	 */
	protected ConstraintsHandler constraintHandler;

	protected HashSet<VarIdentifier> varEventSources = new HashSet<VarIdentifier>();
	/**
	 * default constructor
	 */
	protected int priority = 0;

	protected ConstraintEvent constAwakeEvent;

	/**
	 * Constructs a constraint with the priority 0.
	 */

	public AbstractConstraint() {
		this(0);
	}

	public AbstractConstraint(int priority) {
		this.priority = priority;
		this.constAwakeEvent = new ConstraintEvent(this, false, priority);
	}

	public void addVarEventSource(VarIdentifier varId) {
		this.varEventSources.add(varId);
	}

	public int compareTo(Constraint o) {
		int c = this.priority - ((AbstractConstraint) o).priority;
		if (c != 0)
			return c;
		return this.index - ((AbstractConstraint) o).index;
		// return this.getNbVars() - ((AbstractConstraint) o).getNbVars();//TODO
	}

	/**
	 * Get the constraint handler that listen to the constraint
	 * 
	 * @return constraintHandler
	 */
	public ConstraintsHandler getConstraintsHandler() {
		return this.constraintHandler;
	}

	/**
	 * Get the type of the constraint
	 * 
	 * @return type : the type the constraint
	 */
	public String getConstraintType() {
		return this.type;
	}

	@Override
	public ConstraintEvent getEvent() {
		return this.constAwakeEvent;
	}

	/**
	 * Get the value of index
	 * 
	 * @return the value of index
	 */
	public int getIndex() {
		return this.index;
	}

	@Override
	public int getNbVars() {
		return getNbInternalVars() + getNbExternalVars();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see Constraint#getOwnerId()
	 */
	public int getOwnerId() {
		return this.constraintHandler.myAgent.getAgentId();
	}

	/**
	 * return the priority of the constraint
	 */

	public int getPriority() {
		return priority;
	}

	/**
	 * Get the reference of the constraint
	 * 
	 * @return the name of the constraint
	 */

	public String getReference() {
		if (this.name != null)
			return this.name;
		else
			return "C_" + this.index;
	}

	public HashSet<VarIdentifier> getVarEventSources() {
		return this.varEventSources;
	}

	@Override
	public SortedSet<AbstractVar> getVars() {
		SortedSet<AbstractVar> vars;
		vars = new TreeSet<AbstractVar>(this.constraintHandler.myAgent.agentOrder);
		for (int i = 0 ; i < this.getNbVars() ; i++)
			vars.add(this.getVar(i));
		return vars;
	}

	/**
	 * Get the value of weight
	 * 
	 * @return the value of weight
	 */
	public int getWeight() {
		return this.weight;
	}

	/**
	 * Notify that a constraint checks is done. Used for computing NCCCs
	 */
	public void incrementNumConstraintChecks(int n) {
		this.constraintHandler.myAgent.incrementNumConstraintChecks(n);
	}

	public void incrWeight() {
		this.weight++;
	}

	/**
	 * Check whether the constraint is posted or not
	 */
	public boolean isPosted() {
		return this.posted;
	}

	/**
	 * Check whether the constraint is private or not
	 */
	public boolean isPrivate() {
		return this.fprivate;
	}

	public void removeVarEventSource(VarIdentifier varId) {
		if (this.varEventSources.contains(varId))
			this.varEventSources.remove(varId);
	}

	/**
	 * set the constraint handler as listener
	 * 
	 * @param listener
	 *            : the constraint handler
	 */
	public void setConstraintListener(ConstraintsHandler listener) {
		this.constraintHandler = listener;
	}

	/**
	 * Set the value of index
	 * 
	 * @param index
	 *            the index of the constraint
	 */
	public void setIndex(int index) {
		this.index = index;
		for (AbstractVar v : this.getVars())
			v.addConstraintIndex(index);
	}

	/**
	 * Set the constraint to be handled
	 */
	public void setPosted() {
		this.posted = true;
	}

	/**
	 * Set the constraint to be private
	 */
	public void setPrivate() {
		this.fprivate = true;
	}

	/**
	 * set the name or reference of the constraint.
	 * 
	 * @param constrName
	 *            : name/reference of the constraint
	 */
	public void setReference(String constrName) {
		this.name = constrName;
	}

	@Override
	public void setUnPosted() {
		this.posted = false;
		for (AbstractVar v : this.getVars()) {
			v.removeConstraintIndex(this.getIndex());
		}
	}

	/**
	 * Set the value of weight
	 * 
	 * @param w
	 *            the new value of weight
	 */
	public void setWeight(int w) {
		this.weight = w;
	}

	public Instantiation toInstantiation() {
		Instantiation ass = new Instantiation(this.constraintHandler.myAgent.agentOrder);
		for (int j = 0 ; j < this.getNbVars() ; j++) {
			IntVar y = (IntVar) this.getVar(j);
			if (y.isInstantiated())
				ass.add(y.getVarAssignment());
		}
		return ass;
	}

}
