/************************************************************************

 **************************************************************************/

package kernel;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Map;
import kernel.com.CommunicationInterface;
import kernel.com.Message;
import kernel.com.MessageHandler;
import kernel.order.CentralizedAO;
import kernel.order.ChainOrder;
import kernel.order.TreeOrder;
import simulation.metric.MetricListener;
import simulation.metric.StatisticsMasterObserver;
import simulation.parser.InstanceTokens;
import simulation.parser.PVariable;
import util.BooleanLock;
import util.Commands;
import util.HashList;
import util.Logger;

/**
 * Class AbstractMaster : It controls the resolution of the problem. It starts the agents, collaborates for initializing the distributed algorithm,
 * and terminates the resolution.
 */
public abstract class AbstractMaster extends MessageHandler implements Runnable {
	
	/**
	 * Unique ID of the unique Master
	 */
	//public final static int MasterId = 0;
	/**
	 * Unique ID of the unique Master
	 */
	public final static VarIdentifier MasterVarId = new VarIdentifier(0, 0);
	/**
	 * The Communication Interface to send and receive messages
	 */
	protected CommunicationInterface com;
	/**
	 * a field for implementation of a locking mechanism for messages reception
	 */
	protected BooleanLock mailBoxLock;
	/**
	 * an array for storing all unique and different Identifiers of all agents
	 */
	protected HashList<VarIdentifier> agentIds;
	/**
	 * a field storing whether distributed search is started or not.
	 */
	protected boolean disSearchStarted = false;
	/**
	 * a field storing whether the Master is terminate or not.
	 */
	private boolean terminate = false;
	/**
	 * The queue of messages that have been received for the master.
	 */
	protected LinkedList<Message> mailBox;
	/**
	 * a set of observers that observer the running of Master.
	 */
	protected HashSet<MasterObserver> observers;
	/**
	 * An observer that implement centralized agent ordering algorithm. It is used when agent order is determined centrally.
	 */
	protected CentralizedAO centralizedAO = null;
	/**
	 * A counter of messages sent
	 */
	private int msgSentCounter = 0;
	/**
	 * End State of distributed search :SUCCESS, FAILURE or ERROR
	 */
	protected String endState = "?";
	/**
	 * a field storing whether master is waiting for message or not.
	 */
	private boolean waitingForMessage;
	/**
	 * to log message and debugging
	 */
	protected Logger logger;

	protected StatisticsMasterObserver statisticObserver = null;

	private boolean IhaveBroadcastEnd = false;

	protected String model = Commands.SIMPLE;

	protected boolean interrupted = false;

	protected int nbRcvdStatistics = 0;

	protected int nbMetric = 0;

	/**
	 * The time start of the resolution
	 */
	protected long timeStart; // mSecond

	protected double totalCPUTime;
	
	private ArrayList<String> statisticskeys;
	
	/** category of problems we can solve by the agents that can create this master agents */
	private String category = InstanceTokens.SATISFACTION; 

	/**
	 * The Master is responsible to create the agents when simulation.
	 * 
	 * @param agents : store a problem instance in format "XDisCSP 1.0"
	 * @return a list of separate threads/agents
	 */
	public Map<VarIdentifier, AbstractAgent> makeAgents(Map<String, PVariable> variables) {
		this.statisticskeys = new ArrayList<String>();
		this.statisticskeys.add(Commands.MSGs);
		this.statisticskeys.add(Commands.NCCCs);
		this.statisticskeys.add(Commands.GNCCCs);
		this.statisticskeys.add(Commands.NCCPUTime);
		this.statisticskeys.add(Commands.CCs);
		this.statisticskeys.add(Commands.SOLVABLE);
		this.statisticskeys.add(Commands.Time);
		this.statisticskeys.add(Commands.CPU);
		this.statisticskeys.add(Commands.comulatedCpuTime);
		this.statisticskeys.add(Commands.AO);
		this.statisticskeys.addAll(getTypesOfMessages());
//		List<String> auxiliaryStatistics = this.getAuxiliaryStatistics();
//		for(String key : auxiliaryStatistics){
//			this.statisticskeys.add(key);
//		}
		boolean first = true;
		Map<VarIdentifier, AbstractAgent> map = new HashMap<VarIdentifier, AbstractAgent>();
		for(PVariable variable : variables.values()){
			AbstractAgent agent = makeAgent(variable);
			variable.setDisChocoAgent(agent);
			map.put(variable.getVarIdentifier(), agent);
			if(first){
				for(String k : agent.auxiliaryStatistics.keySet()){
					this.statisticskeys.add(k);
				}
				for(AgentObserver observer : agent.observers){
					this.statisticskeys.add(observer.name);
				}
				first = false;
			}
		}
		return map;
	}
	
//	protected abstract List<String> getAuxiliaryStatistics();
	

	/**
	 * @return
	 */
	public abstract Collection<String> getTypesOfMessages();

	public abstract AbstractAgent makeAgent(int agId, String name, String description);
	
	public abstract AbstractAgent makeAgent(PVariable variable);

	/**
	 * Constructs a Master
	 */
	public AbstractMaster() {
		super(null, Commands.MASTER);
		this.observers = new HashSet<MasterObserver>();
		this.mailBox = new LinkedList<Message>();
		this.mailBoxLock = new BooleanLock(false);
		this.timeStart = System.nanoTime();
	}
	/**
	 * Constructs a Master
	 */
	public AbstractMaster(String category) {
		this();
		this.category = category;
	}

	/**
	 * There is no agent that owns the Master, this method must return null;
	 */
	public AbstractAgent getMyAgent() {
		return null;
	}

	/**
	 * called when master is started
	 */
	public void run() {
//		System.out.println("timeStart : "+this.timeStart);
		if (Logger.LOG_DEBUG) {
			this.logger = new Logger(Commands.MASTER + ".txt");
			this.logger.printToLog(""+getClass()+" :: "+toString());
		}
		this.agentIds = new HashList<VarIdentifier>(this.com.getRegisteredAgentIds());
		this.agentIds.remove(MasterVarId);
		while (!this.terminate) {
			this.treatReceivedMsgs();
			Thread.yield();
		}
	}

	/**
	 * Procedure to treat received messages. It blocks the Master in a waiting state until a message arrives. When a message is arrived, only the
	 * first received message is handled.
	 */
	public void treatReceivedMsgs() {
		try {
			this.waitForMessage();
			Message msg = this.getNextPendingMessage();
			if (msg != null && !terminate) {
				this.processMessage(msg);
			}
//			if (this.isFinished()){
//				System.out.println("Total time : "+this.totalCPUTime);
//				System.out.println(""+System.currentTimeMillis()+" - "+this.timeStart+" = "+this.totalCPUTime);
//			}
		} catch (Exception e) {
			this.endState = Commands.ERROR;
			this.broadcastSystemMessage(AbstractMaster.MasterVarId);
			this.terminate();
			this.log(e.getMessage(), Logger.ERROR_LEVEL);
			e.printStackTrace();
		}
	}

	/**
	 * Returns the next pending message in the mailbox of Master
	 */
	protected Message getNextPendingMessage() {
		Message msg = null;
		synchronized (mailBox) {
			if (!this.mailBox.isEmpty())
				msg = this.mailBox.removeFirst();
			if (this.mailBox.isEmpty())
				this.mailBoxLock.setValue(false);
		}
		return msg;
	}

	/**
	 * the Master remains blocked until it is interrupted or a message arrives.
	 * 
	 * @throws InterruptedException
	 */
	private void waitForMessage() {
		synchronized (mailBox) {
			this.waitingForMessage = true;
		}
		try {
			this.mailBoxLock.waitUntilTrue(0L);
		} catch (InterruptedException e) {}
		synchronized (mailBox) {
			this.waitingForMessage = false;
		}
	}

	/**
	 * /** Notify Master that a message is arrived.
	 * 
	 * @param msg
	 *            : the arrived message
	 */
	public synchronized void notifyReception(Message msg) {
		synchronized (mailBox) {
			mailBox.addLast(msg);
			this.mailBoxLock.setValue(true);
		}
	}

	/**
	 * Set the communication interface to the given one.
	 * 
	 * @param com
	 */
	public void setCommunicationInterface(CommunicationInterface com) {
		this.com = com;
	}

	/**
	 * check if distributed search is already started
	 * 
	 * @return true when distributed search is already started, false otherwise
	 */
	public boolean isDisSearchStarted() {
		return this.disSearchStarted;
	}

	/**
	 * Get the value of end state
	 * 
	 */

	public String getEndState() {
		return this.endState;
	}

	/**
	 * Checks whether Master has finished or not
	 */
	public boolean isFinished() {
		return this.terminate;
	}

	/**
	 * Get the number of registered agents
	 * 
	 * @return the number of registered agents
	 */
	public int getNumAgents() {
		return this.agentIds.size();
	}

	/**
	 * Set the centralized agent ordering plugin
	 * 
	 * @param centralizedAO
	 */
	public void setCentralizedAO(CentralizedAO centralizedAO) {
		if (this.centralizedAO == null) {
			this.centralizedAO = centralizedAO;
			this.centralizedAO.setMaster(this);
		} else {
			if (((this.centralizedAO instanceof TreeOrder) && (centralizedAO instanceof ChainOrder))) {
				System.err.println("Warnning : cannot convert tree order to chain order");
			} else {
				this.centralizedAO = centralizedAO;
				this.centralizedAO.setMaster(this);
			}
		}
	}

	/**
	 * Add a Master observer who will observe the execution of the Master
	 * 
	 * @param observer
	 *            : A Master observer
	 */
	public void addObserver(MasterObserver observer) {
		if (observer != null) {
			observer.master = this;
			this.observers.add(observer);
		}
	}

	/**
	 * Remove an observer
	 * 
	 * @param observer
	 *            : the master observer to be removed
	 */
	public void removeObserver(MasterObserver observer) {
		this.observers.remove(observer);
		observer.master = null;
	}

	/**
	 * Get the value of {@link #mailBoxLock}
	 * 
	 * @return {@link #mailBoxLock}
	 */
	public BooleanLock getMailBoxLock() {
		return mailBoxLock;
	}

	/**
	 * Send the given message using the communication interface {@link #com}
	 * 
	 * @param msg
	 *            a message to send
	 */
	public void sendMessage(Message msg) {
		msg.setSenderId(MasterVarId);
		msg.setMsgId(this.msgSentCounter++);
		for (MasterObserver observer : this.observers) {
			observer.handleSendMessage(msg);
		}
		this.log("Master send :" + msg.toString(), Logger.LEVEL_TRACE_MESSAGE);
		this.com.sendMessage(msg);
	}

	/**
	 * Broadcast the given message to all agent except that of the given agentId
	 * 
	 * @param msg
	 * @param agentId
	 *            : Id of the agent that is excepted from broadcast
	 */
	public void broadcast(Message msg, VarIdentifier agentId) {
		msg.setSenderId(MasterVarId);
		for (VarIdentifier agId : this.agentIds) {
			if (!agentId.equals(agId)) {
				Message message = msg.clone();
				message.setReceiverId(agId);
				this.sendMessage(message);
			}
		}
	}

	/**
	 * Handles the given message
	 * 
	 * @param msg
	 *            : a message that must be handled
	 * @throws StopException
	 */
	public void processMessage(Message msg) throws StopException {
		this.log(msg.toString(), Logger.LEVEL_TRACE_MESSAGE);
		boolean consumed = false;
		for (MasterObserver observer : this.observers) {
			if (observer.handleReceivedMessage(msg)) {
				consumed = true;
			}
		}
		if (this.centralizedAO != null) {
			if (msg.getType().equals(Commands.AO)) {
				this.centralizedAO.processMessage(msg);
				consumed = true;
			}
		}
		String endStat = (String) msg.get(Commands.END_STATE);
		if (endStat != null) {
			this.endState = endStat;
			Boolean broadcast = (Boolean) msg.get(Commands.BROADCAST_SYSTEM_MSG);
			if (broadcast != null && broadcast.booleanValue()) {
				this.broadcastSystemMessage(msg.getSenderId());
			}
		} else if (!consumed) { 
			throw new Error("\n" +
					"Unknown message type : " + msg.getType() + 
					" \n\t msg : " + msg.toString()); }
		if (this.checkEnd()) {
			this.terminate();
		}
	}

	public boolean checkEnd() {
		if (this.nbMetric == 0)
			return true;
		return this.nbRcvdStatistics == this.getNumAgents();
	}

	public void log(String s, int level) {
		if (logger != null) {
			this.logger.printToLogln(s+"\n", level);
		}
	}

	/**
	 * Broadcast a system message that holds the end state to all agent except that of the given agentId
	 * 
	 * @param agentId : Id of the agent that is excepted from broadcast
	 */
	public void broadcastSystemMessage(VarIdentifier agentId) {
		if (!this.IhaveBroadcastEnd) {
			Message msg = this.createMessage(Commands.TERMINATION_DETECTION);
			msg.addContents(Commands.TERMINATE, this.endState);
			this.broadcast(msg, agentId);
			if (this.statisticObserver == null) {
				this.terminate();
			}
			this.IhaveBroadcastEnd = true;
		}
	}

	/**
	 * DCOP weather the Master is waiting for message.
	 * 
	 * @return boolean <code> true</code> if the Master wait for arrival messages
	 * @see #waitingForMessage Otherwise <code>false</code>.
	 */

	public boolean isWaitingForMessage() {
		if (this.terminate) { return true; }
		return this.waitingForMessage;
	}

	/**
	 * Force Master to terminate execution.
	 */
	public void terminate() {
		this.waitingForMessage = false;
		this.terminate = true;
		this.mailBoxLock.setValue(true);
		this.totalCPUTime = System.nanoTime();
//		System.out.println("end time  : "+this.totalCPUTime);
		this.totalCPUTime -= this.timeStart;
	}

	/**
	 * Interrupts this agent. It is strictly used when execution is limited by a cut-off (max cycle, max CCCs, ...)
	 * <p>
	 * If this agent is blocked in an invocation of the {@link #waitForArrivalMessages(long)} it is notified to break out from the waiting and the
	 * <i>interrupted status</i> is setting to true;
	 */
	public synchronized void interrupt() {
		this.interrupted = true;
		this.getMailBoxLock().setValue(true);
	}

	public void addMetric(MetricListener metric) {
		if (this.statisticObserver == null) {
			this.statisticObserver = new StatisticsMasterObserver();
			this.addObserver(this.statisticObserver);
		}
		if(this.statisticObserver.addMetric(metric)) {
			this.nbMetric++;
		}
	}

	public HashMap<String, Double> getStatistics() {
		if (this.statisticObserver != null) {
			HashMap<String, Double> statistics = this.statisticObserver.getStatistics();
			double value = this.getNumAgents();
			if(statistics.containsKey(Commands.TERMINATION_DETECTION)){
				value += statistics.get(Commands.TERMINATION_DETECTION);
			}
			statistics.put(Commands.TERMINATION_DETECTION, value);
			if(this.endState.equals(Commands.SUCCESS)) {
				statistics.put(Commands.SOLVABLE, (double) 1);
			} else if(this.endState.equals(Commands.FAILURE)) {
				statistics.put(Commands.SOLVABLE, (double) 0);
			} else if(this.endState.equals(Commands.ERROR)) {
				statistics.put(Commands.SOLVABLE, (double) -1);
			}
			return statistics;
		}
		return new HashMap<String, Double>();
	}

	/**
	 * 
	 */
	public ArrayList<String> getStatisticsKeys() {
		return this.statisticskeys;
	}

	public void setEndState(String state) {
		this.endState = state;

	}

	public String getModel() {
		return model;
	}

	/**
	 * The category of problems that can solve agents created by this master agent
	 * @return
	 */
	public String getCategory() {
		return this.category ;
	}

	public CentralizedAO getCentralizedAO() {
		return this.centralizedAO;
	}

	public HashList<VarIdentifier> getOrder() {
		return this.centralizedAO.getOrder();
	}

	public HashList<VarIdentifier> getNeighbors(VarIdentifier varId) {
		return this.getCentralizedAO().getAgentsComplexity().get(varId).getNeighbors();
	}

	public void incrNumRcvdStatistics() {
		this.nbRcvdStatistics++;
	}
	
	/**
	 * @return the totalCPUTime
	 */
	public double getTotalCPUTime() {
		return this.totalCPUTime;
	}

}
