package kernel;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;

import kernel.events.VariableEvent;
import kernel.events.VariableEventListener;
import util.IVarIdentifier;
import util.IntArray;

/**
 * An abstract class for all implementations of variables.
 * 
 * @author Redouane EZZAHIR
 * 
 */
public abstract class AbstractVar implements Variable, IVarIdentifier {
	/**
	 * 
	 */
	private static final long serialVersionUID = -6874299914019781734L;
	/**
	 * The identifier of the variable
	 */
	protected VarIdentifier varId;
	/**
	 * The agent owning the variable
	 */
	protected AbstractAgent owner;
	/**
	 * The name of the variable
	 */
	protected String name;
	/**
	 * A filed stores the time counter of the assignment of the variable.
	 */
	protected int timeCounter = 0;

	protected HashMap<Integer, IntArray> listeners; // map event to list of constraints

	protected HashMap<Integer, IntArray> constrEventMap; // map constraints to set of events//

	protected VariableEvent event;

	public VariableEvent getEvent() {
		return event;
	}

	/**
	 * Constructs an variables
	 * 
	 * @param owner the agent owning the variable
	 * @param name the name of the variable
	 */
	public AbstractVar(AbstractAgent owner, String name, VarIdentifier varId) {
		this.owner = owner;
		this.name = name;
		this.varId = varId;
		this.listeners = new HashMap<Integer, IntArray>();
		this.constrEventMap = new HashMap<Integer, IntArray>();
	}

	/**
	 * Constructs an variables
	 * 
	 * @param owner the agent owning the variable
	 * @param name the name of the variable
	 */
	public AbstractVar(AbstractAgent owner, String name) {
		this.owner = owner;
		this.name = name;
		this.listeners = new HashMap<Integer, IntArray>();
		this.constrEventMap = new HashMap<Integer, IntArray>();
		this.varId = new VarIdentifier(owner.getAgentId());
	}

	public IntArray getEventsType(int index) {
		if (this.constrEventMap.containsKey(index))
			return this.constrEventMap.get(index).clone();
		return new IntArray();
	}

	public void addConstraintIndex(int idx) {
		this.constrEventMap.put(idx, new IntArray());
	}

	public Set<Integer> getConstraintIndices() {
		return this.constrEventMap.keySet();
	}

	public void removeConstraintIndex(int index) {
		if (this.constrEventMap.containsKey(index)) {
			IntArray evts = this.constrEventMap.get(index);
			for (int e : evts) {
				IntArray l = this.listeners.get(e);
				if (l != null) {
					Iterator<Integer> it = l.iterator();
					while (it.hasNext()) {
						if (it.next() == index) {
							it.remove();
							break;
						}
					}
				}
			}
			this.constrEventMap.remove(index);
		}
	}

	public AbstractAgent getAgentOwner() {
		return this.owner;
	}

	public String getName() {
		return this.name;
	}

	/**
	 * Set the identifier of the variable to be the given identifier
	 */

	public void setVarIdentifier(VarIdentifier varId) {
		this.varId = varId.clone();
	}

	/**
	 * Returns the identifier of the variable
	 */
	public VarIdentifier getVarIdentifier() {
		return this.varId;
	}

	/**
	 * Returns the ID of the agent owning the variable
	 */

	public int getAgentId() {
		return this.varId.getAgentId();
	}

	/**
	 * Returns the index of the variable in its local problem
	 */
	public int getVarIndex() {
		return this.varId.getVarIndex();
	}

	public int hashCode() {
		return this.varId.hashCode();
	}

	/**
	 * Compare lexicographically this variable with the variable of the given identifier
	 */
	public int compareTo(IVarIdentifier v) {
		return this.varId.compareTo(v);
	}

	/**
	 * Returns the time counter of the variable (i.e. the value corresponding to the last assignment)
	 */
	public int getTimeCounter() {
		return this.timeCounter;
	}

	/**
	 * Set the time counter to the given value
	 */
	public void setTimeCounter(int counter) {
		this.timeCounter = Math.max(this.timeCounter, counter);
	}

	/**
	 * Adds a listener (constraint) to the listener list of the variable
	 */
	public void addVariableEventListener(int eventType, VariableEventListener listener) {
		int idx = ((Constraint) listener).getIndex();
		((Constraint) listener).addVarEventSource(this.varId);
		if (!this.listeners.containsKey(eventType)) {
			this.listeners.put(eventType, new IntArray());
		}
		this.listeners.get(eventType).add(idx);
		if (!this.constrEventMap.containsKey(idx)) {
			this.constrEventMap.put(idx, new IntArray());
		}
		this.constrEventMap.get(idx).add(eventType);
	}

	/**
	 * Removes a listener (constraint) from the listener list of the variable
	 */
	public void removeVariableEventListener(int eventType, VariableEventListener listener) {
		if (this.listeners.containsKey(eventType)) {
			int idx = ((Constraint) listener).getIndex();
			this.listeners.get(eventType).remove(idx);
			this.constrEventMap.get(idx).remove(eventType);
			if (this.constrEventMap.get(idx).isEmpty()) {
				((Constraint) listener).removeVarEventSource(varId);
			}
		}
	}

	/**
	 * Removes a listener (constraint) from the listener list of the variable
	 */
	public void removeVariableEventListener(VariableEventListener listener) {
		for (int eventType : this.listeners.keySet())
			this.removeVariableEventListener(eventType, listener);
		this.constrEventMap.remove(((Constraint) listener).getIndex());
		((Constraint) listener).removeVarEventSource(varId);
	}

	/**
	 * Returns the number of constraints that listen for the variable
	 */
	public int getNumListeners() {
		return this.listeners.size();
	}

	/**
	 * Checks whether the given type of event is activated or not for the variable
	 */

	public boolean isEventActivated(int eventType) {
		return this.listeners.containsKey(eventType);
	}

	/**
	 * {@inheritDoc}
	 */
	public void fireChangedDomain(int basicEvt, int idx) {
		if (this.isEventActivated(basicEvt))
			this.owner.getPropagationEngine().postEvent(this, basicEvt, idx);
	}

	/**
	 * {@inheritDoc}
	 */
	public void fireChangedValue(int basicEvt, int idx) {
		if (this.isEventActivated(basicEvt)) {
			this.owner.constraintHandler.postEvent(this, basicEvt, idx);
			this.owner.constraintHandler.propagate();
		}
		// else
		// this.owner.log(" the event " + basicEvt+ " is not active for  "+basicEvt+this.event, 1);
	}

	public IntArray getListeners(int evtType) {
		if (this.listeners.containsKey(evtType))
			return this.listeners.get(evtType);
		return new IntArray();

	}

}
