package kernel;

import kernel.com.Message;
import kernel.events.AgentEventListener;

/**
 * Class AgentObserver : Model an observer that observes the execution of an agent. A class for modularly adding functionality/observability to the
 * agent. It is added to agent as a Plugin to allow access to the protected member functions.
 * 
 * @author Redouane EZZAHIR
 * 
 */

public abstract class AgentObserver implements AgentEventListener, Cloneable {
	/**
	 * The agent for which the observer is assigned
	 */
	public AbstractAgent myAgent = null;
	protected String name;

	public AgentObserver(String name){
		this.name = name;
	}
	/**
	 * This function is executed after the observer has been assigned to an agent by running {@link AbstractAgent#addObserver(AgentObserver)}
	 */
	public void initialize() {
		this.myAgent.addAgentEventListener(this);
	}

	/**
	 * Called when the observer is removed from its agent.
	 */
	public void removed() {
		this.myAgent.removeAgentEventListener(this);
		this.myAgent = null;
	}

	/**
	 * Called when the observer is assigned to the given agent. Sets myAgent field to the given agent
	 * 
	 * @param agent
	 *            The agent for which we add the observer
	 * @throws IllegalArgumentException
	 */

	public void setAgent(AbstractAgent agent) throws IllegalArgumentException {
		if (agent != null && this.myAgent != null && this.myAgent != agent) { throw new IllegalArgumentException(
				"Error: observer is already assigned to a different agent!"); }
		this.myAgent = agent;
		if (this.myAgent != null) {
			this.initialize();
		} else {
			this.removed();
		}
	}

	/**
	 * This function is called after the agent executes <code>receiveMessage()</code>, but before the message is handled.
	 * 
	 * @param msg
	 *            The message the agent is trying to receive.
	 * @return boolean indicate if the message has been consumed by the observer or not.
	 * @throws StopException
	 * 
	 * @see kernel.Communicator
	 */
	abstract public boolean handleReceivedMessage(Message msg) throws StopException;

	/**
	 * This function is called after the agent executes <code>sendMessage()</code>, but before the message is actually sent.
	 * 
	 * @param msg
	 *            The message the agent is trying to send.
	 * 
	 * @see AbstractAgent#sendMessage(kernel.com.Message)
	 */
	abstract public void handleSendMessage(Message msg);

}
