package kernel;

import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.SortedSet;
import java.util.TreeSet;
import kernel.events.VariableEvent;
import kernel.vars.integer.IntExternalVar;
import simple.SimpleAgent;
import util.HashList;
import util.Instantiation;
import util.IntArray;
import util.Logger;
import util.VarAssignment;

/**
 * Class AgentView : Model the agent-view where an agent stores the received assignments from other agents
 * 
 * @author Mohamed Wahbi
 */
public class AgentView implements Iterable<IntExternalVar> {

	/**
	 * A stored set of the external variables
	 */
	private SortedSet<IntExternalVar> vars;
	/**
	 * The agent owner this agentView
	 */
	public AbstractAgent myAgent;

	/**
	 * Constructs an agent-view
	 * 
	 * @param agent
	 *            : the agent owner of this view
	 */
	public AgentView(AbstractAgent agent) {
		this.myAgent = agent;
	}

	protected void initialize() {
		if (Logger.LOG_DEBUG) {
			this.vars = Collections.synchronizedSortedSet(new TreeSet<IntExternalVar>(this.myAgent.agentOrder));
		} else {
			this.vars = new TreeSet<IntExternalVar>(this.myAgent.agentOrder);
		}
	}

	/**
	 *@return VarAssignment if the corresponding variable exists, otherwise null
	 * 
	 */
	public VarAssignment getVarAssignment(VarIdentifier varId) {
		Iterator<IntExternalVar> iterateur = this.iterator();
		while (iterateur.hasNext()) {
			IntExternalVar v = (IntExternalVar) iterateur.next();
			if (v.getVarIdentifier().equals(varId)) {
				return v.getVarAssignment();
			}
		}
		return null;
	}

	/**
	 * Checks whether this view contains an external variable with the given identifier or not
	 * 
	 * @return True if it exists, false otherwise
	 * @param varId
	 *            : a variable identifier
	 */
	public boolean contains(VarIdentifier varId) {
		Iterator<IntExternalVar> iterateur = this.iterator();
		while (iterateur.hasNext()) {
			IntExternalVar v = (IntExternalVar) iterateur.next();
			if (v.getVarIdentifier().equals(varId)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Get the external variable of the given identifier
	 * 
	 * @return IntExternalVar
	 * @param varId
	 *            : a variable identifier
	 */
	public IntExternalVar getVar(VarIdentifier varId) {
		Iterator<IntExternalVar> it = this.iterator();
		while (it.hasNext()) {
			IntExternalVar v = (IntExternalVar) it.next();
			if (varId.equals(v.getVarIdentifier()))
				return v;
		}
		return null;
	}

	/**
	 * Add new external variable if isn't exists
	 * 
	 * @param var
	 */
	public void add(IntExternalVar var) {
		if (!contains(var)) {
			if (!this.vars.add(var))
				throw new Error(this.myAgent.getAgentName() + " fail to add " + var + " to agentview: " + this.toString());
		}
	}

	/**
	 * Remove the given external variable from this agent-view
	 * 
	 * @param var
	 *            : variable to remove
	 */
	public void remove(IntExternalVar var) {
		this.vars.remove(var);
	}

	/**
	 * Checks whether this view contains the given <tt>var</tt> or not
	 * 
	 * @return true if it exists, false otherwise.
	 * @param var
	 *            : the variable to check
	 */
	public boolean contains(IntExternalVar var) {
		return this.vars.contains(var);
	}

	/**
	 * Checks whether all a
	 * 
	 * @return boolean
	 */
	public boolean isCompletlyInstantiated() {
		Iterator<IntExternalVar> it = this.iterator();
		while (it.hasNext()) {
			IntExternalVar v = (IntExternalVar) it.next();
			if (!v.isInstantiated())
				return false;
		}
		return true;
	}

	/**
	 * 
	 * @return int the number of external variables referenced in this agent-view
	 */
	public int size() {
		return this.vars.size();
	}

	/**
	 * Returns a string representation of this agent-view
	 */
	public String toString() {
		String rep = "";
		synchronized (this) {
			if (this.vars == null)
				return "{?}";
			rep = "{" + this.vars.toString() + "}";
		}
		return rep;
	}

	/**
	 * Checks whether this agent-view is compatibles with the given instantiation or not; They are compatible if the common variables are assigned the
	 * same values
	 * 
	 * @param inst
	 *            : an instantiation to check its compatibility with the agent view
	 * @return true if they are compatible, false otherwise.
	 */
	public boolean compatibleWith(Instantiation inst) {
		Iterator<VarAssignment> it = inst.iterator();
		while (it.hasNext()) {
			VarAssignment v = it.next();
			if (!this.compatible(v)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Get the context of the agents of highest priority
	 * 
	 * @return an instantiation of higher agents
	 */
	public Instantiation getContextsOfHigherAgents() {
		if (this.myAgent instanceof SimpleAgent) {
			VarIdentifier varId = ((SimpleAgent) this.myAgent).var.getVarIdentifier();
			return this.getInstantiationOfAgentsHigherThan(varId);
		} else {
			throw new UnsupportedOperationException("Not yet implemented for complex agents");
		}
	}

	/**
	 * Get the instantiation of the agents of lowest priority
	 * 
	 * @return an instantiation of lower agents
	 */
	public Instantiation getInstantiationOfLowerAgents() {
		if (this.myAgent instanceof SimpleAgent) {
			VarIdentifier varId = ((SimpleAgent) this.myAgent).var.getVarIdentifier();
			return this.getInstantiationOfAgentsLowerThan(varId);
		} else {
			throw new UnsupportedOperationException("Not yet implemented for complex agents");
		}
	}

	/**
	 * Get the instantiation of the agents of priority higher than the agent of the given Id
	 * 
	 * @return an instantiation of agents who are higher than the agent of identifier agId
	 */

	public Instantiation getInstantiationOfAgentsHigherThan(VarIdentifier varId) {
		Instantiation inst = new Instantiation(this.myAgent.agentOrder);
		Iterator<IntExternalVar> it = this.iterator();
		while (it.hasNext()) {
			IntExternalVar exV = it.next();
			if (this.myAgent.agentOrder.compareOrder(exV.getVarIdentifier(), varId) > 0) {
				if (exV.isInstantiated()) {
					inst.add(exV.getVarAssignment());
				}
			}
		}
		return inst;
	}

	/**
	 * Get the instantiation of the agents of priority lower than the agent of the given Id
	 * 
	 * @return an instantiation of agents who are lower than the agent of identifier agId
	 */

	public Instantiation getInstantiationOfAgentsLowerThan(VarIdentifier varId) {// TODO varId ou lieu de agId
		Instantiation inst = new Instantiation(this.myAgent.agentOrder);
		Iterator<IntExternalVar> it = this.iterator();
		while (it.hasNext()) {
			IntExternalVar exV = it.next();
			if (this.myAgent.agentOrder.compareOrder(exV.getVarIdentifier(), varId) < 0) {
				if (exV.isInstantiated()) {
					inst.add(exV.getVarAssignment());
				}
			}
		}
		return inst;
	}

	/**
	 * Returns an instantiation that contains assignment of all assigned agents
	 * 
	 */

	public Instantiation toInstantiation() {
		Instantiation inst = new Instantiation(this.myAgent.agentOrder);
		Iterator<IntExternalVar> it = this.iterator();
		while (it.hasNext()) {
			IntExternalVar exV = it.next();
			if (exV.isInstantiated()) {
				inst.add(exV.getVarAssignment());
			}
		}
		return inst;
	}

	/**
	 * @return an instantiation that contains assignment of all assigned neighbors
	 */
	public Instantiation getInstantiationOfNeighbors() {
		Instantiation inst = new Instantiation(this.myAgent.agentOrder);
		Iterator<IntExternalVar> it = this.iterator();
		while (it.hasNext()) {
			IntExternalVar exV = it.next();
			if (exV.isInstantiated() && this.myAgent.communicator.isNeighbor(exV.getVarIdentifier())) {
				inst.add(exV.getVarAssignment());
			}
		}
		return inst;
	}

	/**
	 * 
	 * @param agentIds
	 * @return
	 */
	public Instantiation getInstantiationOf(IntArray agentIds) {
		Instantiation inst = new Instantiation(this.myAgent.agentOrder);
		Iterator<IntExternalVar> it = this.iterator();
		while (it.hasNext()) {
			IntExternalVar exV = it.next();
			if (exV.isInstantiated() && agentIds.contains(exV.getAgentId())) {
				inst.add(exV.getVarAssignment());
			}
		}
		return inst;
	}

	/**
	 * Get an iterator of external variable
	 * 
	 * @return Iterator<IntExternalVar>
	 * 
	 */
	public Iterator<IntExternalVar> iterator() {
		// this.refresh();
		return this.vars.iterator();
	}

	/**
	 *All external variables that constructed the agentView ordered. When the order changed dynamically, this method must be called to rearrange the
	 * variables according to the new order.
	 * 
	 * @param view2
	 */
	public void refresh(HashSet<IntExternalVar> view) {
		this.initialize();
		Iterator<IntExternalVar> it = view.iterator();
		while (it.hasNext()) {
			this.vars.add(it.next());
		}
	}

	public HashSet<IntExternalVar> getSetOfExternalVars() {
		HashSet<IntExternalVar> varsSet = new HashSet<IntExternalVar>();
		varsSet.addAll(this.vars);
		return varsSet;
	}

	/**
	 * Checks whether the given assignment is compatible or not with the agent view
	 * 
	 * @param varAss
	 *            : an agent assignment
	 */

	public boolean compatible(VarAssignment varAss) {
		VarAssignment v2 = this.getVarAssignment(varAss.getVarIdentifier());
		if (v2 == null)
			return true;
		return v2.getValue() == varAss.getValue();
	}

	/**
	 * Instantiate external variable if local time counter is less than the time counter of given variable and having different value if local value
	 * is equals to the value of given variable, only update time counter; Instantiate external variable if local time counter is equal to the time
	 * counter of given variable and the local variable is not assigned
	 * 
	 * update nogoods store if Explanation is Activated
	 * 
	 * @param varAss
	 */
	public boolean priorityMerge(VarAssignment varAss) {
		IntExternalVar exV = this.getVar(varAss.getVarIdentifier());
		if (exV == null) {
			throw new Error("\n" + this.myAgent + " AgentView:" + "\n" + this + "\npriorityMerge [" + varAss + "]\n"
					+ "try to set value of an unknown external variable: " + varAss.getVarIdentifier() + "\n" + "myOrder     : "
					+ this.myAgent.getOrder() + "\n" + "myNeighbors : " + this.myAgent.communicator.getNeighbors());
		}
		int comp = this.compareTimeCounter(varAss);
		if (comp > 0) {
			if (exV.getValue() != varAss.getValue()) {
				exV.instantiate(varAss);
				return true;
			} else {
				exV.setTimeCounter(varAss.getTimeCounter());
				return false;
			}
		} else if (comp == 0 && !exV.isInstantiated()) {
			exV.instantiate(varAss);
			return true;
		}
		return false;
	}

	/**
	 * iterates over all VarAssignment (varAss) including in the given <tt> view </tt> and call {@link #priorityMerge(VarAssignment)}
	 * 
	 * @param view the instantiation to merge
	 * @return true if the agentview was updated otherwise false
	 */
	public boolean priorityMerge(Instantiation view) {
		boolean updated = false;
		Iterator<VarAssignment> it = view.iterator();
		while (it.hasNext()) {
			VarAssignment varAss = it.next();
			updated |= this.priorityMerge(varAss);
		}
		return updated;
	}

	@Deprecated
	public void merge(Instantiation view) {
		Iterator<VarAssignment> it = view.iterator();
		while (it.hasNext()) {
			VarAssignment varAss = it.next();
			IntExternalVar exV = this.getVar(varAss.getVarIdentifier());
			if (exV == null) {
				this.add(this.myAgent.makeExternalVar(varAss));
			} else {
				this.priorityMerge(varAss);
			}
		}
	}

	/**
	 * Iterates over all VarAssignment (varAss) including in the given <tt> view </tt> and calls {@link #clear(VarAssignment)}
	 * 
	 * @param view
	 */
	public void clear(Instantiation view) {
		IntExternalVar[] exVars = this.vars.toArray(new IntExternalVar[this.size()]);
		for(int i = exVars.length - 1 ; i >= 0 ; i--) {
			if (view.contains(exVars[i].getVarIdentifier()) && exVars[i].isInstantiated()) {
				exVars[i].desinstanciate(VariableEvent.NOCAUSE);
			}
		}
	}

	/**
	 * unassigns the given variable, and if the explanation is activated for each external variable clean all nogoods/eliminators holding the
	 * unassigned variable
	 * 
	 * @param varAss
	 *            the variable assignment to remove its corresponding assignment from the agent view
	 */
	public void clear(VarAssignment varAss) {
		IntExternalVar exV = this.getVar(varAss.getVarIdentifier());
		if (exV == null) {
			throw new Error(this.myAgent + " AgentView " + this + " clear(" + varAss + ") try to remove value " + " of an unknown external: "
					+ varAss);
		}
		if (exV.isInstantiated()) {
			exV.desinstanciate(VariableEvent.NOCAUSE);
		}
	}

	/**
	 * unassigns all instantiated variables and clean nogoods store (eliminators) if explanation is activated
	 */
	public void clear() {
		IntExternalVar[] exVars = this.vars.toArray(new IntExternalVar[this.size()]);
		for(int i = exVars.length - 1 ; i >= 0 ; i--) {
			if (exVars[i].isInstantiated()) {
				exVars[i].desinstanciate(VariableEvent.NOCAUSE);
			}
		}
	}

	public void clearExcept(Instantiation view) {
		IntExternalVar[] exVars = this.vars.toArray(new IntExternalVar[this.size()]);
		for(int i = exVars.length - 1 ; i >= 0 ; i--) {
			if (exVars[i].isInstantiated()) {
				VarIdentifier vId = exVars[i].getVarIdentifier();
				if (!view.contains(vId)) {
					exVars[i].desinstanciate(VariableEvent.NOCAUSE);
				} else if (view.getVarAssignment(vId).getTimeCounter() > exVars[i].getTimeCounter()) {
					exVars[i].desinstanciate(VariableEvent.NOCAUSE);
				}
			}
		}
	}

	/**
	 * Unassigns the set of external variables of the sorted set whose elements are greater than or equal to he given variable: <tt>fromVarAssig</tt>.
	 * 
	 * @param fromVarAssig
	 *            A variable assignment
	 */

	public void clearFrom(VarAssignment fromVarAssig) {
		if (fromVarAssig == null)
			return;
		IntExternalVar exV = this.getVar(fromVarAssig.getVarIdentifier());
		SortedSet<IntExternalVar> set = this.vars.tailSet(exV);
		IntExternalVar[] exVars = set.toArray(new IntExternalVar[set.size()]);
		for(int i = exVars.length - 1 ; i >= 0 ; i--) {
			if (exVars[i].isInstantiated()) {
				exVars[i].desinstanciate(VariableEvent.NOCAUSE);
			}
		}
	}

	public void clearExclusiveFrom(VarAssignment fromVarAssig) {
		if (fromVarAssig == null)
			return;
		IntExternalVar exV = this.getVar(fromVarAssig.getVarIdentifier());
		if (exV == null)
			return;
		SortedSet<IntExternalVar> set = this.vars.tailSet(exV);
		IntExternalVar[] exVars = set.toArray(new IntExternalVar[set.size()]);
		for(int i = exVars.length - 1 ; i >= 1 ; i--) {
			if (exVars[i].isInstantiated()) {
				exVars[i].desinstanciate(VariableEvent.NOCAUSE);
			}
		}
	}

	/**
	 * Returns the assignment of the external variable that follows in the sorted set the given variable
	 * 
	 * @param v
	 *            A variable assignment
	 * 
	 */

	public VarAssignment getNextVarAssignment(VarAssignment v) {
		IntExternalVar exV = this.getVar(v.getVarIdentifier());
		if (exV == null)
			return null;
		else {
			Iterator<IntExternalVar> it = this.vars.tailSet(exV).iterator();
			it.next();
			if (it.hasNext())
				return it.next().getVarAssignment();
			return null;
		}

	}

	public boolean isSubSetOf(Instantiation view) {
		return this.toInstantiation().isSubSetOf(view);
	}

	public boolean contains(Instantiation view) {
		return this.toInstantiation().contains(view);
	}

	public void updateTimeCounter(VarAssignment v) {
		IntExternalVar exv = this.getVar(v.getVarIdentifier());
		if (exv != null)
			exv.setTimeCounter(Math.max(exv.getTimeCounter(), v.getTimeCounter()));
	}

	/**
	 * @param varAss
	 * @return a positive number if the counter time of the given assignment is bigger than the counter of its stored image <br/>
	 *         0 if equality <br/>
	 *         negative number otherwise;
	 */
	public int compareTimeCounter(VarAssignment varAss) {
		IntExternalVar var = this.getVar(varAss.getVarIdentifier());
		assert (var != null) : ("\n" + this.myAgent + " " + this.myAgent.getOrder() + "\n" + varAss + " is not in my agentview :" + this.toString());
		// if(var == null){
		// throw new Error("\n" + this.myAgent + " " + this.myAgent.getOrder() + "\n" +
		// "varAss (" + varAss +") is not in my agentview: \n" + this.toString());
		// }
		return varAss.getTimeCounter() - var.getTimeCounter();
	}

	public HashSet<VarAssignment> ensureOnlyHigherAgentsInView() {
		HashSet<VarAssignment> oldHigher = new HashSet<VarAssignment>();
		HashList<VarIdentifier> highAgents = this.myAgent.agentOrder.getHigherAgents();
		for(VarIdentifier varId : highAgents) {
			if (!this.contains(varId)) {
				IntExternalVar extVar = this.myAgent.makeExternalVar(varId);
				this.add(extVar);
			}
		}
		Iterator<IntExternalVar> it = this.iterator();
		while (it.hasNext()) {
			IntExternalVar v = it.next();
			if (!highAgents.contains(v.getAgentId())) {
				if (v.isInstantiated()) {
					oldHigher.add(v.getVarAssignment());
					v.desinstanciate(VariableEvent.NOCAUSE);
				}
				it.remove();
			}
		}
		return oldHigher;
	}

	public ExternalVar getVar(int index) {
		if (index < 0 || index >= this.size()) {
			throw new IndexOutOfBoundsException();
		}
		Iterator<IntExternalVar> it = this.vars.iterator();
		int i = -1;
		while (it.hasNext()) {
			i++;
			ExternalVar v = it.next();
			if (i == index) {
				return v;
			}
		}
		return null;
	}

	public ExternalVar lastVar() {
		return this.vars.last();
	}

	public void remove(ExternalVar o) {
		this.vars.remove(o);
	}

	public int[] getTimeStamp() {
		int[] timeStamp = new int[this.size()];
		int i = 0;
		for(IntExternalVar v : this) {
			timeStamp[i++] = v.getTimeCounter();
		}
		return timeStamp;
	}

}
