/************************************************************************

 **************************************************************************/

package kernel;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import kernel.com.CommunicationInterface;
import kernel.com.Message;
import kernel.com.MessageHandler;
import util.BooleanLock;
import util.Commands;
import util.HashList;
import util.Logger;

/**
 * Class Communicator : Every agent uses a communicator to send and receipt its messages.
 */
public class Communicator {

	/**
	 * A map of message handler by type. Each message handler is responsible to treat a type of messages
	 */
	protected HashMap<String, MessageHandler> msgHandlers;
	/**
	 * Agent controlling the communicator
	 */
	protected AbstractAgent myAgent;
	/**
	 * Communication interface to send and receive messages from other agents
	 */
	protected CommunicationInterface com;
	/**
	 * a field for implementation of a locking mechanism for messages reception
	 */
	protected BooleanLock mailBoxLock;
	/**
	 * The queue of messages that have been received for this agent.
	 */
	protected LinkedList<Message> mailBox;
	/**
	 * An array storing the Ids of neighbors
	 */
	protected HashList<VarIdentifier> neighbors;
	/**
	 * a field storing whether master is waiting for message or not.
	 */
	protected volatile boolean waitingForMessage = false;

	/**
	 * A counter that stores the Id of last sent message
	 */
	private long msgId = 0;

	private long nbMessageSent = 0;

	/**
	 * lower and higher neighbors
	 */
	protected HashList<VarIdentifier> lowerNeighbors0 = null;

	protected HashList<VarIdentifier> higherNeighbors0 = null;

	/**
	 * Constructs the communicator
	 * 
	 * @param agent
	 *            : the agent owner the communicator
	 */
	public Communicator(AbstractAgent agent) {
		this.myAgent = agent;
		this.msgHandlers = new HashMap<String, MessageHandler>();
		this.neighbors = new HashList<VarIdentifier>();
		this.mailBoxLock = new BooleanLock(false);
		this.mailBox = new LinkedList<Message>();
	}

	/**
	 * called to initialize the communicator
	 */
	public void initialize() {
		this.lowerNeighbors0 = this.computeLowerNeighbors();
		this.higherNeighbors0 = this.computeHigherNeighbors();
	}

	/**
	 * Set the communication interface to the given reference
	 * 
	 * @param comm
	 *            : a reference to the Communication interface to be used by the communicator to send and receive messages
	 */
	protected void setCommunicationInterface(CommunicationInterface comm) {
		this.com = comm;
	}

	protected CommunicationInterface getCommunicationInterface() {
		return this.com;
	}

	/**
	 * Adds a message handler
	 * 
	 * @param msgH
	 *            the message handler to add
	 */
	public void addMsgHandler(MessageHandler msgH) {
		this.msgHandlers.put(msgH.getHandlerType(), msgH);
	}

	public String[] getMsgHandlerTypes() {
		String[] h = new String[this.msgHandlers.size()];
		h = this.msgHandlers.keySet().toArray(h);
		return h;
	}

	/**
	 * DCOP weather the agent is waiting for message.
	 * 
	 * @return <code> true</code> if the Master wait for arrival messages {@link #waitForArrivalMessages(long)}, Otherwise <code>false</code>.
	 */
	protected boolean isWaitingForMessage() {
		return waitingForMessage;
	}

	/**
	 * Send the given message using the communication interface {@link #com}
	 * 
	 * @param msg
	 *            a message to send
	 */
	protected void sendMessage(Message msg) {
		VarIdentifier rvId = msg.getReceiverId();
		msg.setSenderId(this.myAgent.getVarIdentifier());
		msg.setMsgId(++msgId);
		if (this.myAgent.disSearchStarted)
			for (AgentObserver observer : this.myAgent.observers) {
				observer.handleSendMessage(msg);
			}
		if (!this.myAgent.isFinished()) {
			this.logMsg(msg, "sendMessage   ");
			this.com.sendMessage(msg);
		}
		if (myAgent.localAgents == null || (myAgent != null && !myAgent.localAgents.contains(rvId))){
			setNbMessageSent(getNbMessageSent() + 1);
		}
	}

	/**
	 * Log message <code>msg</code> with the specified operation on this message
	 * 
	 * @param msg A message to log
	 * @param opperationOnMsg  Operation on message: send message or process message
	 */
	private void logMsg(Message msg, String opperationOnMsg) {
		if (Logger.LOG_DEBUG) {
			this.myAgent.log(opperationOnMsg + " : " + msg.toString(), Logger.LEVEL_TRACE_MESSAGE);
			this.myAgent.log("AgentView      : " + this.myAgent.agentView.toString(), Logger.DEFAULT_LEVEL);
			this.myAgent.log("mySolution     : " + this.myAgent.solutionToString() + "\n", Logger.DEFAULT_LEVEL);

		}
	}

	/**
	 * Causes the agent to wait until a message arrives or until it was interrupted ({@link AbstractAgent#interrupt()} ) When a message arrives the
	 * agent is notified by changing the state of it {@link Communicator#mailBoxLock} to true. The agent can be also interrupted simply by setting it
	 * mailBoxLock to true.
	 * 
	 * @return boolean
	 * @param msTimeout
	 * @throws InterruptedException
	 */
	protected boolean waitForArrivalMessages(long msTimeout) throws InterruptedException {
		this.waitingForMessage = true;
		boolean result = mailBoxLock.waitUntilTrue(msTimeout);
		this.waitingForMessage = false;
		return result;
	}

	/**
	 * This function inserts the given message to the agent mailbox, It also notify the agent that a message arrived. Notification is done by setting
	 * the state of the mailbox lock to true.
	 * 
	 * @param msg
	 *            : the message that just arrived
	 */

	protected void notifyReception(Message msg) {
		synchronized (mailBox) {
			if (msg.getSenderId().equals(kernel.AbstractMaster.MasterVarId)) {
				this.mailBox.addFirst(msg);
			} else {
				this.mailBox.addLast(msg);
			}
			this.mailBoxLock.setValue(true);
		}
	}

	/**
	 * This function processes the first message with the given type if it already exists in the mail box and returns true. Otherwise, the agent
	 * remains blocked until it is interrupted or the expected message or a termination message arrives.
	 * 
	 * @param msgType
	 * @param msTimeout
	 * @throws InterruptedException
	 * @throws StopException
	 */
	protected boolean blockingReceiveMessage(String msgType, long msTimeout) throws InterruptedException, StopException {
		// only wait for the specified amount of time
		long endTime = System.currentTimeMillis() + msTimeout;
		long msRemaining = msTimeout;
		boolean end = false;
		Message msg = null;
		do {
			if (!msgType.equals(Commands.AO)) {
				end = terminationDetection();
			}
			if (!end) {
				synchronized (this.mailBox) {
					msg = this.getNextMessage(msgType);
					while (this.mailBoxLock.isTrue() && msg == null) {
						msg = this.getNextMessage(msgType);
					}
					if (msg != null && !this.mailBox.isEmpty()) {
						this.mailBoxLock.setValue(true);
					}
				}
			} else {
				return false;
			}
			if (msg == null) {
				this.waitForArrivalMessages(msRemaining);
				msRemaining = endTime - System.currentTimeMillis();
			}
		} while (msg == null && (msRemaining > 0L));
		if (msg != null && !end) {
			this.processMessage(msg);
			return true;
		}

		return false;
	}

	/**
	 * Called to handle the given message <code>msg</code>; Firstly, <code>msg</code> is handled by the observers
	 * {@link AgentObserver#handleReceivedMessage(Message)} secondly, <code>msg</code> is passed as argument to the method processMessage of its
	 * corresponding handler {@link kernel.com.MessageHandler#processMessage(Message)}
	 * 
	 * @param msg
	 * @throws StopException
	 */

	private void processMessage(Message msg) throws StopException {
		if (this.myAgent.disSearchStarted) {
			for (AgentObserver observer : this.myAgent.observers) {
				observer.handleReceivedMessage(msg);
			}
		}
		if (this.msgHandlers.containsKey(msg.getType())) {
			// this.myAgent.log("Agent state beffore processing msg "+ msg.getMsgId(),Logger.INFO_LEVEL);
			this.logMsg(msg, "receiveMessage");
			this.myAgent.logDomains();
			this.myAgent.constraintHandler.logConstraints();
			this.msgHandlers.get(msg.getType()).processMessage(msg);
			// this.myAgent.log("Agent state after processing  msg " +msg.getMsgId(),Logger.INFO_LEVEL);
			this.myAgent.logDomains();
			// this.myAgent.constraintHandler.logConstraints();
		}
	}

	/**
	 * Called to check whether a message of termination detection is arrived or not
	 * 
	 * @throws StopException
	 *             when a terminate message is received
	 */

	private boolean terminationDetection() throws StopException {
		Message msg = this.getNextMessage(Commands.TERMINATION_DETECTION);
		if (msg != null) {
			for (AgentObserver observer : this.myAgent.observers) {
				observer.handleReceivedMessage(msg);
			}
			this.msgHandlers.get(Commands.TERMINATION_DETECTION).processMessage(msg);
			return this.myAgent.isFinished();
		}
		return false;
	}

	/**
	 * Handles all messages in the mail box
	 * 
	 * @throws StopException
	 *             when a termination message is received or when agent detects end of algorithm
	 */
	protected void processMessages() throws StopException {
		Message msg = this.getNextMessage();
		while (msg != null) {
			this.processMessage(msg);
			if (this.myAgent.isFinished())
				return;
			msg = this.getNextMessage();
		}
	}

	/**
	 * Returns the next message in the mail box without waiting its arrived. The method is called when we are sure that such message is arrived
	 * {@link #waitForArrivalMessages(long)}
	 * 
	 */
	private Message getNextMessage() {
		Message msg = null;
		synchronized (mailBox) {
			if (!this.mailBox.isEmpty()) {
				msg = this.mailBox.removeFirst();
			}
			if (this.mailBox.isEmpty())
				this.mailBoxLock.setValue(false);
		}
		return msg;
	}
	
	/**
	 * This method returns the state of the mailbox, i.e., true if the mailbox is empty, false otherwise.
	 * 
	 * @return boolean : <b><i>true</i></b> if the mailbox is empty, otherwise, <b><i>false</i></b>.
	 */
	public boolean isMailBoxEmpty() {
		synchronized (mailBox) {
			return this.mailBox.isEmpty();
		}
	}

	/**
	 * Returns the first message of the given type in the mail box without waiting its arrived. The method is called when we are sure that such
	 * message is arrived {@link #waitForArrivalMessages(long)}
	 */
	private Message getNextMessage(String type) {
		Message msg = null;
		synchronized (mailBox) {
			Iterator<Message> itr = mailBox.iterator();
			while (itr.hasNext()) {
				Message m = itr.next();
				if (m.getType().equals(type)) {
					msg = m;
					itr.remove();
					break;
				}
			}
			if (msg == null) {
				this.mailBoxLock.setValue(false);
			}
		}
		return msg;
	}

	/**
	 * @return the number of message sent
	 */
	public long getNumMsgSent() {
		return this.msgId;
	}

	/**
	 * Add an agent Id as neighbor
	 * 
	 * @param agId
	 */
	public void addNeighbor(VarIdentifier varId) {
		this.neighbors.add(varId);
	}

	/**
	 * Add an agent Id as neighbor
	 * 
	 * @param agId
	 */
	protected void removeNeighbor(VarIdentifier varId) {
		this.neighbors.remove(varId);
	}

	/**
	 * Returns the set of agents neighbors
	 * 
	 * @return the value of agentsNeighbors
	 */
	public HashList<VarIdentifier> getNeighbors() {
		return this.neighbors;
	}

	/**
	 * Returns the set of agents neighbors
	 * 
	 * @return the value of agentsNeighbors
	 */
	public HashList<VarIdentifier> getNeighbors0() {
		return HashList.union(this.higherNeighbors0, this.lowerNeighbors0);
	}

	/**
	 * Compute and returns the set of lower neighbors
	 * 
	 * @return IntArray : contains the set of lower neighbors
	 */
	private HashList<VarIdentifier> computeLowerNeighbors() {
		HashList<VarIdentifier> lowerN = new HashList<VarIdentifier>();
		for (VarIdentifier neig : this.neighbors) {
			if (this.myAgent.agentOrder.isLower(neig)) {
				lowerN.add(neig);
			}
		}
		return lowerN;
	}

	/**
	 * Compute and returns the set of higher neighbors
	 * 
	 * @return IntArray : contains the set of higher neighbors
	 */
	private HashList<VarIdentifier> computeHigherNeighbors() {
		HashList<VarIdentifier> higherN = new HashList<VarIdentifier>();
		for (VarIdentifier neig : this.neighbors) {
			if (this.myAgent.agentOrder.isHigher(neig)) {
				higherN.add(neig);
			}
		}
		return higherN;
	}

	/**
	 * Checks if the given id is that of a neighbor or not
	 * 
	 * @return true if it is the case, false otherwise
	 * @param agId : Id of the agent to check if it is neighbor or not
	 */
	public boolean isNeighbor(VarIdentifier varId) {
		return this.neighbors.contains(varId);
	}

	/**
	 * @return the set of agents neighbors who are lower than the agent of the given Id
	 * 
	 * @param agId : agent Id
	 */
	public HashList<VarIdentifier> getNeighborsLowerThan(VarIdentifier varId) {
		HashList<VarIdentifier> lowerAgents = this.myAgent.agentOrder.getAgentsLowerThan(varId);
		return HashList.intersection(this.neighbors, lowerAgents);
	}

	/**
	 * @return the set of agents neighbors who are lower than the agent of the given Id
	 * 
	 * @param agId : agent Id
	 */
	public HashList<VarIdentifier> getNeighborsHigherThan(VarIdentifier varId) {
		HashList<VarIdentifier> higherAgents = this.myAgent.agentOrder.getAgentsHigherThan(varId);
		return HashList.intersection(this.neighbors, higherAgents);
	}

	public HashList<VarIdentifier> getLowerNeighbors() {
		return this.computeLowerNeighbors();
	}

	public HashList<VarIdentifier> getHigherNeighbors() {
		return this.computeHigherNeighbors();
	}

	public boolean isHigherNeighbor(VarIdentifier varId) {
		return this.getHigherNeighbors().contains(varId);
	}

	public boolean isLowerNeighbor(VarIdentifier varId) {
		return this.getLowerNeighbors().contains(varId);
	}

	public HashList<VarIdentifier> getHigherNeighbors0() {
		return this.higherNeighbors0;
	}

	public HashList<VarIdentifier> getLowerNeighbors0() {
		return this.lowerNeighbors0;
	}

	/**
	 * @return the nbMessageSent
	 */
	public long getNbMessageSent() {
		return nbMessageSent;
	}

	/**
	 * @param nbMessageSent the nbMessageSent to set
	 */
	public void setNbMessageSent(long nbMessageSent) {
		this.nbMessageSent = nbMessageSent;
	}

}
