/*****************************************************************************

 ******************************************************************************/

package kernel;

import java.util.Collection;
import java.util.SortedSet;
import kernel.events.VariableEventListener;
import util.Instantiation;

public interface Constraint extends VariableEventListener, Comparable<Constraint> {

	public static final Integer INFINITY = Integer.MAX_VALUE;
	// //////////// Constraint types ///////////////////////
	public static final String eq = "=";
	public static final String lt = "<";
	public static final String gt = ">";
	public static final String leq = "<=";
	public static final String geq = ">=";
	public static final String neq = "#";
	public static final String feasP = "feasible pairs";
	public static final String infeasP = "infeasible pairs";
	public static final String soft = "soft";
	public static final String link = "link";
	public static final String DISTANCE = "Distance";
	public static final String NOGOOD = "Nogood";
	// //////// Operators /////////////////////////////////
	public final static int EQ = 0;
	public final static int LT = -1;
	public final static int GT = 1;
	public final static int LEQ = -2;
	public final static int GEQ = 2;
	public final static int NEQ = 3;
	// ///////// propagation methods ///////////////////
	public static final int FC = 1;
	public static final int AC2001 = 2;
	public static final int FB = 3;

	public void awake();

	/**
	 * Get the constraint handler that listen to the constraint
	 * 
	 * @return constraintHandler
	 */
	public ConstraintsHandler getConstraintsHandler();

	/**
	 * Get the type of the constraint
	 * 
	 * @return type : the type the constraint
	 */
	public String getConstraintType();

	/**
	 * Get the external variable with position <code>idx</code> in the constraint scope
	 * 
	 * @param idx : the index of external variable in the constraint scope
	 * @return An external variable
	 */
	public AbstractVar getExternalVar(int idx);

	/**
	 * Get the external variables in the constraint scope
	 * 
	 * @return the set of external variable
	 */
	public Collection<AbstractVar> getExternalVars();

	/**
	 * Returns the index of the constraint
	 */
	public int getIndex();

	/**
	 * Returns the internal variable with the specified position <code>idx</code> in constraint scope
	 * 
	 * @param inx is position of the returned variable among internal varriable in constraint scope.
	 */
	public AbstractVar getInternalVar(int idx);

	/**
	 * Returns the number of external variables within the constraint scope
	 */
	public int getNbExternalVars();

	/**
	 * Returns the number of internal variables within the constraint scope
	 */
	public int getNbInternalVars();

	/**
	 * Returns the number of variables within the constraint scope
	 */

	public int getNbVars();

	/**
	 * Returns the ID of the agent owning the constraint
	 */
	public int getOwnerId();

	/**
	 * 
	 * @return
	 */

	public int getPriority();

	/**
	 * Returns a reference of the constraint
	 * 
	 * @return the name of the constraint
	 */
	public String getReference();

	/**
	 * 
	 * @param i
	 * @return
	 */
	public AbstractVar getVar(int i);

	/**
	 * 
	 * @return
	 */
	public SortedSet<AbstractVar> getVars();

	/**
	 * Returns the current weight of the constraint
	 */
	public int getWeight();

	/**
	 * Notify that the constraint was checked called when a consistency check is made. It is used for computing NCCCs
	 */
	public void incrementNumConstraintChecks(int n);

	/**
	 * 
	 */
	public void incrWeight();

	/**
	 * Checks whether all variables within the constraint scope are instantiated or not
	 */
	public boolean isCompletlyInstantiated();

	public boolean isConsistent(Variable variable, int value);

	/**
	 * Checks whether the constraint is posted or not
	 */
	public boolean isPosted();

	/**
	 * Checks whether the constraint is private or not
	 * 
	 */
	public boolean isPrivate();

	/**
	 * Checks whether the constraint is satisfied or not
	 */
	public boolean isSatisfied();

	/**
	 * Add the constraint handler as a listener
	 * 
	 * @param listener
	 */
	public void setConstraintListener(ConstraintsHandler listener);

	/**
	 * Set the index of the constraint to the given value
	 * 
	 * @param Index
	 */
	public void setIndex(int index);

	/**
	 * Mark that the constraint is posted, a constraint not posted is not considered in the computing
	 */
	public void setPosted();

	/**
	 * Set the constraint be private
	 */
	public void setPrivate();

	/**
	 * Set the name of the constraint to the given reference
	 * 
	 * @param reference
	 *            is the new name of the constraint
	 */
	public void setReference(String reference);

	/**
	 * 
	 */
	public void setUnPosted();

	/**
	 * Set the weight of the constraint to the given value
	 * 
	 * @param w
	 *            : the newest weight
	 */
	public void setWeight(int weight);

	/**
	 * Returns an instantiation that contains assignment of all assigned variables within the constraint scope
	 */
	public Instantiation toInstantiation();

}
