package kernel;

import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import kernel.constraints.integer.explain.SENoGoodConstraint;
import kernel.events.ConstraintEvent;
import kernel.events.ConstraintListener;
import kernel.events.DisChocoEngine;
import kernel.events.PropagationEngine;
import kernel.events.VariableEvent;
import util.IntArray;
import util.Logger;
import util.VarEventMask;

public class ConstraintsHandler implements ConstraintListener {

	public static final int MAX_INT_COST = Integer.MAX_VALUE;

	public static final int MIN_INT_COST = Integer.MIN_VALUE;

	private PropagationEngine propagationEngine;
	/**
	 * the owner of the constraint handler
	 */
	public AbstractAgent myAgent;

	private int lastIndex = 0;

	protected Hashtable<Integer, Constraint> constraints;
	/**
	 * The list of constraints that will be removed
	 */
	protected IntArray constrWillBeRemoved;

	private boolean initialialPostAwake = true;

	public void processConstraint(Constraint c, Variable v, int evtType) {
		c.awakeOnVariableEvent(v, evtType);
	}

	/**
	 * Constructs the Constraints Handler
	 * 
	 * @param agent
	 */
	public ConstraintsHandler(AbstractAgent agent) {
		this.myAgent = agent;
		// this.indices = new LinkedList<Integer>();
		this.constraints = new Hashtable<Integer, Constraint>();
		this.constrWillBeRemoved = new IntArray();
		this.propagationEngine = new DisChocoEngine();
	}

	// ================================================================================

	/**
	 * Called to process all the constraint events stored in the queue @
	 */
	public void propagate() {
		this.propagationEngine.propagate();
		this.treatConstraintRemoving();
	}

	// ================================================================================

	private void treatConstraintRemoving() {
		for (int cIndex : this.constrWillBeRemoved) {
			if(this.constraints.containsKey(cIndex)){
				this.constraints.get(cIndex).setUnPosted();
				this.constraints.remove(cIndex);
				// this.indices.offer( cIndex);
			}
		}
		this.constrWillBeRemoved = new IntArray();

	}

	// ================================================================================

	public void notifyPropagateVarEvent(Variable touchedVar, int evtType) {
		this.myAgent.log("propagate event (" + touchedVar.toString() + ")[" + (evtType == VariableEvent.DESINSTANCIATE? " DESINSTANCIATE " : "")
				+ (evtType == VariableEvent.INSTANCIATE? "INSTANCIATE " : "") + (evtType == VariableEvent.RESTORE_VAL? " RESTORE_VAL " : "")
				+ (evtType == VariableEvent.RESTORE_INF? " RESTORE_INF " : "") + (evtType == VariableEvent.RESTORE_SUP? " RESTORE_SUP " : "")
				+ (evtType == VariableEvent.INCREMENT_INF? " INCREMENT_INF " : "")
				+ (evtType == VariableEvent.DECREMENT_SUP? ", DECREMENT_SUP " : "") + (evtType == VariableEvent.REMOVE_VAL? " REMOVE_VAL " : "")
				+ "]", Logger.LEVEL_TRACE_FUNCTION_1);
		// TODO add statistics
	}

	// ================================================================================

	/**
	 * Post an external constraint to be dealt with during the search of solution
	 * 
	 * @param exC
	 *            An external constraint to post
	 */
	public void post(Constraint exC) {
		this.registerConstraint(exC, true);
		this.activate(exC);
	}

	// ================================================================================

	public void registerConstraint(Constraint c, boolean initial) {
		c.setPosted();
		ConstraintEvent event = c.getEvent();
		this.propagationEngine.registerEvent(event);
		this.propagationEngine.postConstAwake(c, initial);
	}

	// ================================================================================

	public void postEvent(AbstractVar var, int basicEvt, int cause) {
		this.propagationEngine.postEvent(var, basicEvt, cause);
	}

	// ================================================================================

	public void postConstAwake(Constraint c, boolean initialized) {
		this.propagationEngine.postConstAwake(c, initialized);
		// if(!this.initialialPostAwake)
		// this.propagate();
	}

	// ================================================================================

	/**
	 * Add a new external constraint. It will only be stored. to take into account this constraint in the computation, it must be posted by calling
	 * the {@link #post(Constraint)} method
	 * 
	 * @param exC
	 *            A new external constraint to add
	 */
	public void add(Constraint exC) {
		exC.setConstraintListener(this);
		exC.setIndex(lastIndex);
		this.constraints.put(lastIndex, exC);
		lastIndex++;
	}

	// ================================================================================
	public void dynamicAdd(Constraint exC) {
		this.add(exC);
		this.post(exC);
		// this.myAgent.log("dynamicAdd "+exC +"\n"+this, Logger.INFO_LEVEL);
	}

	// ================================================================================

	/**
	 * Returns the constraint with the given index
	 * 
	 * @return Constraint
	 * @param index
	 */
	public Constraint getConstraint(int index) {
		return this.constraints.get(index);
	}

	// ================================================================================

	/**
	 * Returns the number of stored constraints
	 * 
	 * @return integer
	 */
	public int getNumConstraints() {
		return this.constraints.size();
	}

	// ================================================================================

	/**
	 * Returns an iterator over all stored constraints
	 */
	public Iterator<Constraint> getConstraintsIterator() {
		return this.constraints.values().iterator();
	}

	// ================================================================================

	public void logConstraints() {
		this.myAgent.log(this.toString(), Logger.LEVEL_TRACE_FUNCTION_2);
	}

	// ================================================================================

	/**
	 * Returns a string representing the handler. It prints all stored constraint
	 */

	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append("[ the posted constraints of myAgent ("+this.myAgent.getAgentId()+"), where myOrder is: "+this.myAgent.getOrder()+" are:\n");
		for (int i = 0 ; i <= this.lastIndex ; i++) {
			if (this.constraints.containsKey(i)) {
				Constraint c = this.getConstraint(i);
				if(c.isPosted()){
					sb.append(" \t " + c.toString() + "\n");
				}
			}
		}
		sb.append("]");
		return sb.toString();
	}

	// ================================================================================

	public PropagationEngine getPropagationEngine() {
		return this.propagationEngine;
	}

	// ================================================================================

	public void removeMe(Constraint c) {
		constrWillBeRemoved.add(c.getIndex());
	}

	// ================================================================================
	/**
	 * Removes all variable events (used when we don't propagate events for example when clear agent-view )
	 */
	public void flushEvents() {
		this.propagationEngine.flushEvents();
	}

	// ================================================================================

	@Override
	public void propagateConstraintEvent(ConstraintEvent event) {
		((Constraint) event.getTouchedObject()).propagate();
	}

	// ================================================================================

	@Override
	public void propagateVariableEvent(Variable v, int evtType, int evtCause) {
		IntArray indices = ((AbstractVar) v).getListeners(evtType);
		SortedSet<Constraint> constraints = new TreeSet<Constraint>();
		// Firstly we sort the constraints
		for (int index : indices) {
			constraints.add(this.getConstraint(index));
		}
		// Secondly get the sorted indices
		indices = new IntArray();
		Iterator<Constraint> it = constraints.iterator();
		while (it.hasNext()) {
			indices.add(it.next().getIndex());
		}
		// Propagate following priority order order
		for (int i : indices) {
			Constraint c = this.getConstraint(i);
			if (c != null && c.getIndex() != evtCause) {
				processConstraint(c, v, evtType);
			}
		}
	}

	// =============================================================================
	public void update() {
		for (Constraint c : this.constraints.values()) {
			this.activate(c);
		}
		this.propagate();
		this.initialialPostAwake = false;
	}

	// ============================================================================
	private void activate(Constraint c) {
		for (AbstractVar var : c.getVars()) {
			VarEventMask VarEventMask = this.getVarEventMask(c, var);
			IntArray config = this.myAgent.eventsConfig.get(VarEventMask);
			IntArray varEvents = var.getEventsType(c.getIndex());
			for (int eventType : varEvents) {
				if (!config.contains(eventType)) {
					var.removeVariableEventListener(eventType, c);
					if (c.getVarEventSources().size() == 0) {
						c.setUnPosted();
						// if (c instanceof SENoGoodConstraint) {
						// this.removeMe(c);
						// }
					} else if (c instanceof SENoGoodConstraint) {
						this.removeMe(c);
					}
				}
			}
			varEvents = var.getEventsType(c.getIndex());
			for (int eventType : config) {
				if (!varEvents.contains(eventType)) {
					var.addVariableEventListener(eventType, c);
					if (!c.isPosted()) {
						registerConstraint(c, initialialPostAwake);
					}
				}
			}
		}
	}

	// ================================================================================
	public Set<Constraint> getConstraintsWhereVarIsTheLastInScope(AbstractVar var) {
		Set<Constraint> constraint = new HashSet<Constraint>();
		for (int idx : var.getConstraintIndices()) {
			Constraint exC = this.getConstraint(idx);
			if (isTheLastVarInScope(var, exC)) {
				constraint.add(exC);
			}
		}
		return constraint;
	}

	// =================================================================================
	public Set<Constraint> getConstraintsWhereVarIsTheFirstInScope(AbstractVar var) {
		Set<Constraint> constraint = new HashSet<Constraint>();
		for (int idx : var.getConstraintIndices()) {
			Constraint exC = this.getConstraint(idx);
			if (this.isTheFirstVarInScope(var, exC)) {
				constraint.add(exC);
			}
		}
		return constraint;
	}

	// ===============================================================================
	public Set<Constraint> getConstraintsWhereVarIsTheSecondLastInScope(AbstractVar var) {
		Set<Constraint> constraint = new HashSet<Constraint>();
		for (int idx : var.getConstraintIndices()) {
			Constraint exC = this.getConstraint(idx);
			if (isTheSecondLastVarInScope(var, exC)) {
				constraint.add(exC);
			}
		}
		return constraint;
	}

	// ================================================================================
	public Set<Constraint> getConstraintsWhereVarIsTheSecondInScope(AbstractVar var) {
		Set<Constraint> constraint = new HashSet<Constraint>();
		for (int idx : var.getConstraintIndices()) {
			Constraint exC = this.getConstraint(idx);
			if (isTheSecondVarInScope(var, exC)) {
				constraint.add(exC);
			}
		}
		return constraint;
	}

	// ==============================================================================
	public VarEventMask getVarEventMask(Constraint c, AbstractVar var) {
		assert (c.getVars().contains(var));
		if (var instanceof ExternalVar) {
			boolean isHigher = this.myAgent.agentOrder.isHigher(var.getVarIdentifier());
			return isHigher? VarEventMask.EXTERNAL_VAR_HIGHER_THAN_LOCAL_VAR : VarEventMask.EXTERNAL_VAR_LOWER_THAN_LOCAL_VAR;
		} else {
			if (isTheLastVarInScope(var, c)) { return VarEventMask.LOCAL_VAR_IS_THE_LAST; }
			if (this.isTheFirstVarInScope(var, c)) { return VarEventMask.LOCAL_VAR_IS_FIRST; }
			return VarEventMask.LOCAL_VAR_IS_BETWEEN_L_H;
		}

	}

	// ================================================================================
	// //////////////////// Static Methods ////////////////////////////////////////////
	// ================================================================================

	public static boolean isTheLastVarInScope(AbstractVar var, Constraint exC) {
		SortedSet<AbstractVar> vars = exC.getVars();
		return (!vars.isEmpty() && vars.last().equals(var));
	}

	// =================================================================================
	public boolean isTheFirstVarInScope(AbstractVar var, Constraint exC) {
		SortedSet<AbstractVar> vars = exC.getVars();
		return (!vars.isEmpty() && vars.first().equals(var));
	}

	// ==================================================================================
	public static boolean isTheSecondLastVarInScope(AbstractVar var, Constraint exC) {
		SortedSet<AbstractVar> vars = exC.getVars();
		if (vars.size() >= 2) {
			AbstractVar scondLast = vars.headSet(vars.last()).last();
			if (scondLast.equals(var)) { return true; }
		}
		return false;
	}

	// =================================================================================
	public static boolean isTheSecondVarInScope(AbstractVar var, Constraint exC) {
		SortedSet<AbstractVar> vars = exC.getVars();
		if (vars.size() >= 2) {
			vars.remove(vars.first());
			AbstractVar second = vars.first();
			if (second.equals(var)) { return true; }
		}
		return false;
	}
	// ==================================================================================

	public String getConstraints() {
		StringBuffer sb = new StringBuffer();
		sb.append("{");
		for(Constraint c : this.constraints.values()) {
			sb.append(c.toString());
			sb.append(";");
		}
		sb.deleteCharAt(sb.length()-1);
		sb.append("}");
		return sb.toString();
	}

}
