package kernel;

import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Random;
import java.util.StringTokenizer;
import util.HashList;
import util.IVarIdentifier;
import util.Tools;
import static simulation.parser.InstanceTokens.VARIABLE_IDENTIFIER_SEPARATOR;

/**
 * A class for representing the identifiers of variables.
 * 
 * 
 * @author Redouane Ezzahir
 * 
 */
public final class VarIdentifier implements IVarIdentifier, Cloneable, Comparable<IVarIdentifier> {
	/**
	 * 
	 */
	private static final long serialVersionUID = -4527543357838318955L;
	/**
	 * Identifier of the agent owning the variable with the identifier.
	 */
	private final int agentId;
	/**
	 * Index of this variable in its local problem.
	 */
	private final int varIndex;

	/**
	 * Constructs a variable identifier
	 * 
	 * @param agId the Id of the agent owning the variable with the identifier
	 * @param varIndex the index in the local problem of the variable with the identifier
	 */
	public VarIdentifier(int agId, int varIndex) {
		this.agentId = agId;
		this.varIndex = varIndex;
	}

	public VarIdentifier(int agId) {
		this(agId, 1);
	}

	/**
	 * @param varId
	 */
	public VarIdentifier(VarIdentifier varId) {
		this.agentId = varId.agentId;
		this.varIndex = varId.varIndex;
	}

	public VarIdentifier(String string) {
		this(fromString(string));
	}

	/**
	 * Returns the index in the local problem of the variable with the identifier
	 */

	public int getVarIndex() {
		return this.varIndex;
	}

	/**
	 * Returns the ID of the agent owning the variable with the identifier.
	 */
	public int getAgentId() {
		return this.agentId;
	}

	/**
	 * Returns a cloned object of the variable identifier
	 */
	public VarIdentifier clone() {
		return new VarIdentifier(this.agentId, this.varIndex);
	}

	/**
	 * Checks whether the identifier is equals to the given object
	 */
	public boolean equals(Object o) {
		if (o == null || !(o instanceof VarIdentifier)){
			return false;
		}
		VarIdentifier varId = (VarIdentifier) o;
		return (this.agentId == varId.agentId) && (this.varIndex == varId.varIndex);
	}

	public int hashCode() {
		return this.agentId * 1000 + this.varIndex;
	}

	/**
	 * Compare two variable identifiers lexicographically.
	 * 
	 * @return this.agentId - v.agentId if different from zero this.varIndex - v. varIndex otherwise
	 */
	@Override
	public int compareTo(IVarIdentifier v) {
		if (v != null) {
			int d = this.getAgentId() - v.getAgentId();
			if (d == 0)
				d = this.getVarIndex() - v.getVarIndex();
			return d;
		}
		return Integer.MIN_VALUE;
	}

	/**
	 * Returns a string representing this object
	 */
	public String toString() {
		return this.agentId + VARIABLE_IDENTIFIER_SEPARATOR + this.varIndex;
	}

	/**
	 * Make a VarIdentifier from the given string
	 * 
	 * @param str : the string to construct the identifier from; Example "5.1"
	 * @return A new VarIdentifier
	 */
	public synchronized static VarIdentifier fromString(String string) {
		int agentId = -1;
		int varIndex = 1;
		if(string.contains(VARIABLE_IDENTIFIER_SEPARATOR)) {
			StringTokenizer tok = new StringTokenizer(string, VARIABLE_IDENTIFIER_SEPARATOR);
			agentId = Tools.tryParseInt(tok.nextToken());
			varIndex = Tools.tryParseInt(tok.nextToken());
		} else {
			agentId = Tools.tryParseInt(string);
		}
		return new VarIdentifier(agentId, varIndex);
	}

	public static void main(String args[]) {
		String[] vi = {"1.0", "2.0", "3.0", "4.0", "14.0", "5.0", "6.0", "7.0", "8.0", "9.0", "10.0", "11.0", "12.0", "13.0", "15.0"};

		HashMap<VarIdentifier, Integer> map = new java.util.HashMap<VarIdentifier, Integer>();
		for (int i = 0 ; i < vi.length ; i++) {
			VarIdentifier v = VarIdentifier.fromString(vi[i]);

			map.put(v, v.hashCode());
		}
	//	System.out.println(map + " " + map.containsKey(new VarIdentifier(15)));
		HashList<VarIdentifier> list = new HashList<VarIdentifier>();
		list.add(new VarIdentifier(3));
		list.add(new VarIdentifier(1));
		list.add(new VarIdentifier(2));
		list.add(new VarIdentifier(1));
		System.out.println(list);
		System.out.println(list.indexOf(4));
		System.out.println(list.indexOf(new VarIdentifier(3)));
		
		HashList<VarIdentifier> lexico = new HashList<VarIdentifier>(list);
		Collections.sort(lexico, new Comparator<VarIdentifier>(){
			@Override
			public int compare(VarIdentifier varId0, VarIdentifier varId1) {
				return varId0.compareTo(varId1);
			}
		});
		System.out.println("lexico: "+lexico);
		Random rand = new Random();
		rand.setSeed(System.currentTimeMillis());
		List<VarIdentifier> _list = new HashList<VarIdentifier>(lexico);
		HashList<VarIdentifier> random = new HashList<VarIdentifier>();
		while(!_list.isEmpty()) {
			int choosen = rand.nextInt(_list.size());
			random.add(_list.remove(choosen));
		}
		System.out.println("random: "+random);
	}

}
