package kernel.com;

import java.io.Serializable;
import java.util.Hashtable;
import util.Commands;
import kernel.VarIdentifier;

/**
 * This class models a message that will be exchanged between agents. Message contains a set of attributes (from, to, type, ...) and a "content".
 * 
 * @authors 
 * <a href="mailto:ezzahir@ensa-agadir.ac.ma?subject=DisChoco &cc=wahbi@lirmm.fr &cc=bessiere@lirmm.fr">
 * Redoaune Ezzahir</a> and 
 * <a href="mailto:wahbi@lirmm.fr?subject=DisChoco &cc=ezzahir@ensa-agadir.ac.ma &cc=bessiere@lirmm.fr">
 * Mohamed Wahbi</a>
 *
 */

public class Message implements Cloneable, Serializable {

	private static final long serialVersionUID = 5725211850619445129L;
	/**
	 * The type of the message
	 */
	private final String msgType;
	/**
	 * Denotes the identity of the sender of the message
	 */
	private VarIdentifier from;
	/**
	 * Denotes the identity of the intended recipients of the message
	 */
	private VarIdentifier to;
	/**
	 * The contents of the message
	 */
	private Hashtable<String, Object> contents;
	/**
	 * message Id that allow to give the causal order of messages sent
	 */
	private long msgId = 0;

	private long totalMsg = 0;
	/**
	 * A field to attach the non concurrent computation time
	 */
	private double NCCPUTime = 0;
	/**  A field to attach the number of non-concurrent constraint checks */
	private long NCCCs = 0;

	/**
	 * Constructs a new message within the given type
	 * 
	 * @param type  : a string repressing the type of the message to make
	 */

	public Message(String type) {
		this.msgType = type;
	}

	/**
	 * Returns the type of the message
	 */
	public String getType() {
		return this.msgType;
	}

	/**
	 * Set the recipient ID.
	 * 
	 * @param to
	 *            Id of the agent recipient
	 */

	public void setReceiverId(VarIdentifier to) {
		this.to = to;
	}

	/**
	 * Returns the ID of the recipient
	 */
	public VarIdentifier getReceiverId() {
		return to;
	}

	/**
	 * Set the sender ID
	 * 
	 * @param from
	 *            : ID of the sender
	 */
	public void setSenderId(VarIdentifier from) {
		this.from = from;
	}

	/**
	 * Returns the sender ID
	 */
	public VarIdentifier getSenderId() {
		return this.from;
	}

	/**
	 * Returns the message ID
	 */
	public long getMsgId() {
		return msgId;
	}

	/**
	 * Set the message ID to the given value
	 */
	public void setMsgId(long msgId) {
		this.msgId = msgId;
	}

	/**
	 * Adds a new data to the contents of the message It Maps the specified <code>key</code> to the specified <code>value</code> in contents
	 * hashtable.
	 * 
	 * @param key
	 *            the hashtable key
	 * @param value
	 *            the data
	 */
	public void addContents(String key, Serializable value) {
		if (contents == null)
			contents = new Hashtable<String, Object>();
		contents.put(key, value);
	}

	/**
	 * Returns the number of non concurrent computation time
	 */
	public double getNCCPUTime() {
		return NCCPUTime;
	}

	/**
	 * Set the number of non concurrent computation time to the given value
	 */
	public void setNCCPUTime(double nccputime) {
		this.NCCPUTime = nccputime;
	}


	/**
	 * Returns the number of non-concurrent constraint checks
	 */
	public long getNCCCs() {
		return NCCCs;
	}

	/**
	 * Set the number of non-concurrent constraint checks to the given value
	 */
	public void setNCCCs(long ncccs) {
		this.NCCCs = ncccs;
	}

	/**
	 * Returns the data to which the specified key is mapped, or {@code null} if the contents of this message contains no mapping for the key.
	 * 
	 * @param key
	 *            key the key whose associated data is to be returned
	 * @return the data to which the specified key is mapped, or {@code null} the contents of this message contains no mapping for the key
	 */
	public Object get(String key) {
		if (contents != null)
			return contents.get(key);
		return null;
	}

	/**
	 * Returns <tt>true</tt> if the contents of this message contains a mapping for the specified key. More formally, returns <tt>true</tt> if and
	 * only if the contents of this message contains a mapping for a key <tt>key</tt> such that <tt>(key==null ? k==null : key.equals(k))</tt>.
	 */
	public boolean contains(String key) {
		return this.get(key) != null;
	}

	/**
	 * Returns a clone of the message
	 */
	@SuppressWarnings("unchecked")
	public Message clone() {
		Message msg = new Message(msgType);
		msg.from = this.from;
		msg.to = this.to;
		msg.msgId = this.msgId;
		msg.NCCPUTime = this.NCCPUTime;
		msg.NCCCs = this.NCCCs;
		msg.contents = (Hashtable<String, Object>) this.contents.clone();
		return msg;
	}

	/**
	 * Returns a string representation of this message
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append("message <Id:" + this.msgId + ", Type:" + this.msgType + "> from " + from + " to " + to + " : ");
		sb.append(Commands.NCCPUTime+":"+this.NCCPUTime+" "+Commands.NCCCs+":"+this.NCCCs+" ");
		sb.append("[ ");
		if (this.contents != null) {
			for (String key : this.contents.keySet()) {
				sb.append("" + key + ":" + this.contents.get(key) + ", ");
			}
		}
		sb.append("]");
		return sb.toString();
	}
	
	public long getSize(){
		long size = 0;
		if (this.contents != null) {
			for (String key : this.contents.keySet()) {
				size += this.contents.get(key).toString().length();
			}
		} 
		return size;
	}

	/**
	 * An implementation of messages comparator. it is used to store message in a stored set
	 */
	public static class Comparator implements java.util.Comparator<Message> {
		public int compare(Message msg1, Message msg2) {
			double def1 = (msg1.getNCCPUTime() - msg2.getNCCPUTime());
			if (def1 < 0) return -1;
			else if(def1>0) {
				return 1;
			}
			long def2 = msg1.msgId - msg2.msgId;
			if (def2 != 0)
				return (int) def2;
			return msg2.from.compareTo(msg1.from);
		}
	}

	public long getTotatNumMessages() {
		return totalMsg;
	}

	/**
	 * Set the number of non concurrent computation time to the given value
	 */
	public void setTotalNumMessage(long totalmsg) {
		this.totalMsg = totalmsg;
	}

}
