/************************************************************************

 **************************************************************************/

package kernel.com;

import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;

import kernel.AbstractAgent;
import kernel.StopException;
import kernel.VarIdentifier;

/**
 * An abstract class that defines how to process messages. The user can extend this class and define the body of the abstract method "processMessage"
 * according to the type of message.
 * <p>
 * For example, in the implementation of the ABT, two message handlers are extended : OkMsgHandler and NogoodsMsgHandler
 * <p>
 * .
 * 
 * @author Redouane EZZAHIR
 * 
 */
public abstract class MessageHandler {
	/**
	 * The type of the handler (OK, Info, Nogood, Cost, ...)
	 */
	protected String handlerType;
	/**
	 * The agent owning this handler. It is null for the Master that extends this class.
	 */
	public AbstractAgent myAgent = null;

	ConcurrentHashMap<VarIdentifier, ConcurrentLinkedQueue<Message>> messages = null;

	protected boolean packet = false;

	public void addMessage(Message msg) {
		VarIdentifier senderId = msg.getSenderId();
		if (!this.messages.containsKey(senderId))
			this.messages.put(msg.getSenderId(), new ConcurrentLinkedQueue<Message>());
		this.messages.get(senderId).add(msg);
	}

	public boolean processNextMessage() throws StopException {
		return false;
	}

	/**
	 * Called for process a message. It must be implemented for each type of message
	 * 
	 * @param msg
	 *            Message to be processed
	 * @throws StopException
	 */
	public abstract void processMessage(Message msg) throws StopException;

	/**
	 * Constructs a message handler
	 * 
	 * @param agent
	 *            the owner
	 * @param HandlerType
	 *            the type of message to handle
	 */
	public MessageHandler(AbstractAgent agent, String HandlerType) {
		this.myAgent = agent;
		this.handlerType = HandlerType;
	}

	/**
	 * Get the value of handlerType
	 * 
	 * @return the value of handlerType
	 */
	public String getHandlerType() {
		return this.handlerType;
	}

	/**
	 * Make a message of type <code>handlerType</code>
	 * 
	 * @return Message
	 */
	public Message createMessage() {
		return new Message(this.handlerType);
	}

	/**
	 * Make a message of type the given <code>handlerType</code>
	 * 
	 * @return Message
	 * @param handlerType
	 */
	public Message createMessage(String handlerType) {
		return new Message(handlerType);
	}
}
