package kernel.constraints.integer;

import java.util.BitSet;

public class CouplesTable extends ConsistencyRelation implements BinRelation {

	/** matrix of consistency/inconsistency */
	protected BitSet table;

	/** the offset of the first value of the first variable, i.e., x */
	protected int offset1;

	/** the offset of the first value of the second variable, i.e., y */
	protected int offset2;

	/** the size of the initial domain of the first variable, i.e., x */
	protected int n1;
	
	/** the size of the initial domain of the second variable, i.e., y */
	protected int n2;
	
	/** A filed stores whether the relation is transposed or not. */
	protected boolean transpose = false;
	
	/** the number of support of each value of the first variable in the domain of the second variable. */
	protected int[] nbSupport = null;
	
	protected String name = null;
	
	public CouplesTable() {}
	
	/**
	 * Create a constraint table.
	 * 
	 * @param feas : the feasibility of the table, true if the tuple presented are the allowed tuples,
	 * otherwise, false
	 * @param offset1 : the offset of the first value of the first variable, i.e., x
	 * @param offset2 : the offset of the first value of the second variable, i.e., y
	 * @param n1 : the size of the initial domain of the first variable, i.e., x
	 * @param n2 : the size of the initial domain of the second variable, i.e., y 
	 */
	public CouplesTable(boolean feas, int offset1, int offset2, int n1, int n2, boolean transpose) {
		this.offset1 = offset1;
		this.offset2 = offset2;
		this.n1 = n1;
		this.n2 = n2;
		this.table = new BitSet(n2 * n1);
		this.feasible = feas;
		this.nbSupport = new int[n1];
		this.transpose = transpose;
		for(int i=0 ; i<this.nbSupport.length ; i++){
			if(!this.feasible){
				this.nbSupport[i] = n2;
			} else {
//				this.nbSupport[i] = 0;
			}
		}
	}

	/**
	 * This method computes the opposite relation by "reusing" the table of consistency
	 * 
	 * @return the opposite relation
	 */
	public ConsistencyRelation getOpposite() {
		CouplesTable t = new CouplesTable();
		t.feasible = !feasible;
		t.table = table;
		t.offset1 = offset1;
		t.offset2 = offset2;
		t.n1 = n1;
		t.n2 = n2;
		t.nbSupport = nbSupport;
		return t;
	}

	/**
	 * Adds tuple to the consistency table
	 * 
	 * @param value_x A value of the first variable
	 * @param value_y A value of the second variable
	 */
	public void setCouple(int value_x, int value_y) {
		if(this.feasible){
			if (this.transpose){
				++this.nbSupport[value_y - offset1];
			} else { 
				++this.nbSupport[value_x - offset1];
			}
		} else {
			if (this.transpose){
				--this.nbSupport[value_y - offset1];
			} else { 
				--this.nbSupport[value_x - offset1];
			}
		}
		table.set((value_x - offset1) * n1 + value_y - offset2);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see simple.integer.constraints.BinRelation#isConsistent(int, int)
	 */
	public boolean isConsistent(int value_x, int value_y) {
		if (this.transpose)
			return table.get((value_y - offset1) * n1 + value_x - offset2) == feasible;
		return table.get((value_x - offset1) * n1 + value_y - offset2) == feasible;
	}
	
	/**
	 * This method returns a string representation of the table. 
	 * @return String : a string representing the table
	 */
	public String toString() {
		String s = "";
		for (int i = this.table.nextSetBit(0) ; i > -1 ; i = this.table.nextSetBit(i + 1)) {
			int y = i % n1 + offset2;
			int x = (i - y + offset2) / n1 + this.offset1;
			s += "(" + x + "," + y + ") ";
		}
		s += "\n { ";
		for(int x=0 ; x<this.nbSupport.length ; x++){
			int value = (x + this.offset1);
			s += "(" + value + "," + this.nbSupport[x] + ") ";
		}
		s += "}";
		s += "\nfeasible : "+this.feasible+"";
		s += "\ntranspose: "+this.transpose+"";
		return "[" + s + "]";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see simple.integer.constraints.ConsistencyRelation#transpose()
	 */
	@Override
	public void transpose() {
		this.transpose = true;
	}

	/* (non-Javadoc)
	 * @see kernel.constraints.integer.BinRelation#getNumberOfSupport(int)
	 */
	@Override
	public int getNbOfSupports(int value) {
		if(feasible){
			return this.nbSupport[value - offset1];
		} else {
			return this.nbSupport[value - offset1];
		}
	}

	/* (non-Javadoc)
	 * @see kernel.constraints.integer.BinRelation#getNumberOfSupport(int)
	 */
	@Override
	public int getNbOfConflicts(int value) {
		if(feasible){
			return this.n1 - this.nbSupport[value - offset1];
		} else {
			return this.n1 - this.nbSupport[value - offset1];
		}
	}

}
