/************************************************************************

 **************************************************************************/

package kernel.constraints.integer;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import kernel.AbstractConstraint;
import kernel.AbstractVar;
import kernel.Variable;
import kernel.events.VariableEvent;
import kernel.vars.integer.IntExternalVar;
import kernel.vars.integer.IntVar;
import kernel.vars.integer.IntVariable;
import simple.dcsp.SimpleConstraint;
import util.IntArray;

/**
 * An abstract class for all implementations of (binary) constraints in the simple mode over internal variable and external variable.
 */
public abstract class SIntBinConstraint extends AbstractConstraint implements SimpleConstraint {

	protected IntVariable v0;

	protected IntVariable v1;

	/**
	 * A constant
	 */
	protected int cste = 0;

	/**
	 * Constructs a binary external constraint
	 * 
	 * @param var0 An Internal variable or an external variable
	 * @param var1 An external variable when var0 is internal, otherwise internal variable
	 */
	public SIntBinConstraint(IntVariable var0, IntExternalVar var1) {
		v0 = var0;
		v1 = var1;
	}

	/**
	 * Constructs a binary external constraint.
	 * 
	 * @param var0
	 *            An Internal variable or an external variable
	 * @param var1
	 *            An external variable when var0 is internal, otherwise internal variable
	 * @param cste
	 *            A constant
	 */
	public SIntBinConstraint(IntVariable var0, IntExternalVar var1, int cste) {
		this(var0, var1);
		this.cste = cste;
	}

	public abstract boolean isConsistent(int val1, int val2);

	@Override
	public boolean isConsistent(Variable v, int val) {
		if (v.equals(this.v0) && v1.isInstantiated()) { return this.isConsistent(val, v1.getValue()); }
		if (v.equals(this.v1) && v0.isInstantiated()) { return this.isConsistent(this.v0.getValue(), val); }
		this.incrementNumConstraintChecks(1);
		return true;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see Constraint#getNbVars()
	 */
	public int getNbVars() {
		return 2;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see simple.SimpleConstraint#getInternalVar(int)
	 */
	public IntVariable getInternalVar(int idx) {
		if (idx != 0)
			throw new Error("index must equal to zero idx:" + idx);
		if (v1 instanceof IntExternalVar) {
			return v0;
		} else {
			return v1;
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see Constraint#getNbInternalVars()
	 */
	public int getNbInternalVars() {
		return 1;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see Constraint#getNbExternalVars()
	 */
	public int getNbExternalVars() {
		return 1;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see Constraint#getExternalVar(int)
	 */
	public AbstractVar getExternalVar(int idx) {
		if (idx != 0)
			throw new Error("index must equal to zero idx:" + idx);
		if (v1 instanceof IntExternalVar) {
			return v1;
		} else {
			return v0;
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see simple.SimpleConstraint#isConsistent(int)
	 */
	public boolean isConsistent(int internalVal) {
		IntExternalVar exter = (IntExternalVar) this.getExternalVar(0);
		if (exter.isInstantiated()) {
			return this.isConsistent(internalVal, exter.getValue());
		}
//		this.incrementNumConstraintChecks(1);
		return true;
	}

	public int findSupport(int value, IntVar v) {
		boolean transpose = false;
		if (v.equals(v0))
			transpose = true;
		Iterator<Integer> it = v.getDomain().iterator();
		while (it.hasNext()) {
			int val2 = it.next();
			if (transpose) {
				if (this.isConsistent(val2, value))
					return val2;
			} else if (this.isConsistent(value, val2))
				return val2;

		}
		return Integer.MIN_VALUE;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see Constraint#isSatisfied()
	 */
	public boolean isSatisfied() {
		if (this.isCompletlyInstantiated()) { return this.isConsistent(v0.getValue(), v1.getValue()); }
		return true;
	}

	public boolean isCompletlyInstantiated() {
		return v0.isInstantiated() && v1.isInstantiated();
	}

	/**
	 * A simple Forward Checking
	 */
	@Override
	public void awakeOnVariableEvent(Variable v, int evtType) {
		if (evtType == VariableEvent.INSTANCIATE) {
			Iterator<Integer> it;
			if (v.equals(v0)) {
				it = v1.getDomain().iterator();
				while (it.hasNext()) {
					int val1 = it.next();
					if (!this.isConsistent(v0.getValue(), val1)) {
						v1.removeValue(val1, this.index);
					}
				}
			} else {
				it = v0.getDomain().iterator();
				while (it.hasNext()) {
					int val0 = it.next();
					if (!this.isConsistent(val0, v1.getValue())) {
						v0.removeValue(val0, this.index);
					}
				}
			}
		}
	}

	public AbstractVar getVar(int i) {
		if (i == 0)
			return this.v0;
		return v1;

	}

	@Override
	public void awake() {
		if (this.varEventSources.contains(v0.getVarIdentifier())) {
			IntArray eventTypes = v0.getEventsType(this.index);
			if (eventTypes.contains(VariableEvent.INSTANCIATE) && v0.isInstantiated()) {
				this.awakeOnVariableEvent(v0, VariableEvent.INSTANCIATE);
			}
			if (eventTypes.contains(VariableEvent.REMOVE_VAL)) {
				this.awakeOnVar(v1);
			}
		}
		if (this.varEventSources.contains(v1.getVarIdentifier())) {
			IntArray eventTypes = v1.getEventsType(this.index);
			if (eventTypes.contains(VariableEvent.INSTANCIATE) && v1.isInstantiated()) {
				this.awakeOnVariableEvent(v1, VariableEvent.INSTANCIATE);
			}
			if (eventTypes.contains(VariableEvent.REMOVE_VAL)) {
				this.awakeOnVar(v0);
			}
		}
	}

	public abstract void awakeOnVar(IntVariable v);

	@Override
	public Collection<AbstractVar> getExternalVars() {
		HashSet<AbstractVar> externalVars = new HashSet<AbstractVar>();
		externalVars.add(this.v1);
		return Collections.unmodifiableSet(externalVars);
	}
	
}
