package kernel.constraints.integer.explain;

import java.util.Collection;
import java.util.Iterator;
import kernel.Variable;
import kernel.constraints.integer.SAbsolute;
import kernel.events.VariableEvent;
import kernel.vars.integer.explain.ExplainedIntExternalVar;
import kernel.vars.integer.explain.ExplainedIntVar;
import kernel.vars.integer.explain.ExplainedIntVariable;
import kernel.vars.integer.explain.Explanation;
import util.IntArray;

/**
 * The absolute constraint v0 = |v1 + cste| is a binary constraint with v0 = var0, v1 = var1 and cste is an integer constant
 */
public class SEAbsolute extends SAbsolute implements Propagator {

	public SEAbsolute(ExplainedIntVariable var0, ExplainedIntExternalVar var1) {
		super(var0, var1);
	}

	public SEAbsolute(ExplainedIntVariable var0, ExplainedIntExternalVar var1, int cste) {
		super(var0, var1, cste);
	}

	// ///////////////////////////////////////////////////////////////////////////////
	@Override
	public void setUnPosted() {
		super.setUnPosted();
		if (v1.isInstantiated())
			((ExplainedIntVar) v0).removeExplanation(this.index);
		if (v0.isInstantiated())
			((ExplainedIntVar) v1).removeExplanation(this.index);
	}

	public void awakeOnRestores(ExplainedIntVar var) {
		if (var.equals(v0)) {
			if (v1.isInstantiated()) {
				Iterator<Integer> it = var.getDomain().getRepairIterator();
				while (it.hasNext()) {
					int value = it.next();
					if (!isConsistent(value, v1.getValue()))
						var.removeValue(value, index, index);
				}
			}
		} else if (var.equals(v1)) {
			if (v0.isInstantiated()) {
				Iterator<Integer> it = var.getDomain().getRepairIterator();
				while (it.hasNext()) {
					int value = it.next();
					if (isConsistent(v0.getValue(), value))
						var.removeValue(value, index, index);
				}
			}
		}
	}

	@Override
	public void awakeOnVariableEvent(Variable v, int evtType) {
		this.incrementNumConstraintChecks(1);
		ExplainedIntVar var = (ExplainedIntVar) v;
		switch (evtType) {
			case VariableEvent.INSTANCIATE :
				this.awakeOnInst(var);
			break;
			case VariableEvent.DESINSTANCIATE :
				this.awakeOnDesInst(var);
			break;
			case VariableEvent.REMOVE_VAL :
				this.awakeOnRemovals(var);
			break;
			case VariableEvent.RESTORE_VAL :
				this.awakeOnRestores(var);
			break;
			case VariableEvent.DECREMENT_SUP :
				this.awakeOnDecSup(var);
			break;
			case VariableEvent.INCREMENT_INF :
				this.awakeOnIncrInf(var);
			break;
			default :
				return;
		}

	}

	// //////////////////////////////////////////////////////////////////////////
	/**
	 * v0 = |v1| When v0 is instantiated, v1 is restricted to 2 values : v0.value and -(v0.value) When v1 is instantiated, v0 is instantiated to
	 * abs(v1)
	 * 
	 * 
	 * @param The
	 *            variable that is instantiated. @
	 * 
	 */
	public void awakeOnInst(ExplainedIntVar var) {
		if (var.equals(v0)) {
			int val = v0.getValue();
			if (!v1.canBeInstantiatedTo(val - cste) && v1.canBeInstantiatedTo(-val - cste)) {
				((ExplainedIntVar) v1).restrictDomainTo(-val - cste, index, index);
			} else if (v1.canBeInstantiatedTo(val - cste) && !v1.canBeInstantiatedTo(-val - cste)) {
				((ExplainedIntVar) v1).restrictDomainTo(val - cste, index, index);
			} else {
				((ExplainedIntVar) v1).updateInf(-val - cste, index, index);
				((ExplainedIntVar) v1).updateSup(val - cste, index, index);
				((ExplainedIntVar) v1).removeInterval(-val - cste + 1, val - cste - 1, index);
			}
		} else if (var.equals(v1)) {
			((ExplainedIntVar) v0).restrictDomainTo(Math.abs(v1.getValue() + cste), index, index);
		}

	}

	@Override
	public void awakeOnDesInst(ExplainedIntVar var) {
		if (var.equals(v0)) {
			((ExplainedIntVar) v1).removeExplanation(index);
		} else if (var.equals(v1)) {
			((ExplainedIntVar) v0).removeExplanation(index);
		}
	}

	/**
	 * If v0.inf increases, values from -v0.inf to v0.inf are forbidden for v1 If v1.inf increases, it depends on the situation
	 * 
	 * @param var
	 * @param values
	 *            @
	 */
	public void awakeOnIncrInf(ExplainedIntVar var) {
		Collection<Explanation> expl = var.getInfExplanations();
		if (var.equals(v0)) {
			// absolute variable lower bound is increased
			// values from -v0.inf - cste to v0.inf - cste are forbidden for v1
			if (v1.getInf() + cste >= 0) {
				((ExplainedIntVar) v1).updateInf(v0.getInf() - cste, index, expl);
			} else if (v1.getSup() + cste <= 0) // v1 <= 0 => v1.sup = - v0.inf
				((ExplainedIntVar) v1).updateSup(-v0.getInf() - cste, index, expl);
			else if (v1.getInf() + cste > -v0.getInf()) // v1 > -v0 => v1.inf = v0.inf
				((ExplainedIntVar) v1).updateInf(v0.getInf() - cste, index, expl);
			else if (v1.getSup() + cste < v0.getInf()) //    
				((ExplainedIntVar) v1).updateSup(-v0.getInf() - cste, index, expl);
			else {
				((ExplainedIntVar) v1).removeInterval(-v0.getInf() + cste + 1, v0.getInf() - cste - 1, index, expl); //  
			}
		} else if (var.equals(v1)) {
			// free variable lower bound is increased
			if (v1.getInf() + cste > -v0.getInf() /* v0.getInf() > 0 by definition */
					&& v1.getInf() + cste < v0.getInf()) { /* v0.getInf() > cste by definition */// TODO
				((ExplainedIntVar) v1).updateInf(v0.getInf() - cste, -1, ((ExplainedIntVar) v0).getInfExplanations());
			} else if (v1.getInf() + cste >= 0) {
				((ExplainedIntVar) v0).updateInf(v1.getInf() + cste, index, expl);
				((ExplainedIntVar) v0).updateSup(v1.getSup() + cste, index, expl);
				detectSymetricalHoles(((ExplainedIntVar) v0), (ExplainedIntVar) v1, v1.getInf() + cste, v1.getSup() + cste);
			} else if (v1.getSup() + cste <= 0) {
				((ExplainedIntVar) v0).updateSup(-v1.getInf() - cste, index, expl);
				((ExplainedIntVar) v0).updateInf(-v1.getSup() - cste, index, expl);
			} else {
				((ExplainedIntVar) v0).updateSup(Math.max(-v1.getInf() - cste, v1.getSup() + cste), index, expl);
				detectSymetricalHoles(((ExplainedIntVar) v0), (ExplainedIntVar) v1, -v1.getInf() - cste, v1.getSup() + cste);
			}
		}
	}

	/**
	 * Try to detect symmetrical holes (i.e., value such that both value and -(value) are not in the domain of v1, in which case abs(value) can be
	 * removed from the domain of v0) @
	 */
	private void detectSymetricalHoles(ExplainedIntVar v0, ExplainedIntVar v1, int inf, int sup) {
		for (int valeur = Math.max(inf, v0.getInf()) ; valeur <= Math.min(sup, v0.getSup()) ; valeur = v0.getDomain().getNextValue(valeur)) {
			this.incrementNumConstraintChecks(1);
			if (!v1.canBeInstantiatedTo(valeur - cste) && !v1.canBeInstantiatedTo(-valeur - cste)) {
				IntArray removedSupports = new IntArray(new int[] {-valeur - cste, valeur - cste});
				v0.removeValue(valeur, this.index, v1.getExplanations(removedSupports));

			}
		}
	}

	/**
	 * If v0.sup decreases, v1 is limited to (-(v0.sup) .. v0.sup) If v1.sup decreases, it depends on the situation @
	 */

	public void awakeOnDecSup(ExplainedIntVar var) {
		Collection<Explanation> expl = var.getSupExplanations();
		if (var.equals(v0)) {
			((ExplainedIntVar) v1).updateSup(v0.getSup() - cste, index, expl);
			((ExplainedIntVar) v1).updateInf(-v0.getSup() - cste, index, expl);
		} else if (var.equals(v1)) {
			if (v1.getSup() + cste > -v0.getInf() && v1.getSup() + cste < v0.getInf()) {
				((ExplainedIntVar) v1).updateSup(-v0.getInf() - cste, -1, ((ExplainedIntVar) v0).getInfExplanations());
			} else if (v1.getInf() + cste >= 0) {
				((ExplainedIntVar) v0).updateSup(v1.getSup() + cste, index, expl);
				((ExplainedIntVar) v0).updateInf(v1.getInf() + cste, index, expl);
			} else if (v1.getSup() + cste <= 0) {
				((ExplainedIntVar) v0).updateInf(-v1.getSup() - cste, index, expl);
				((ExplainedIntVar) v0).updateSup(-v1.getInf() - cste, index, expl);
				detectSymetricalHoles(((ExplainedIntVar) v0), (ExplainedIntVar) v1, -v1.getSup() - cste, -v1.getInf() - cste);
			} else { // general case: x.sup = max(abs(y.inf),abs(y.sup))
				((ExplainedIntVar) v0).updateSup(Math.max(-v1.getInf() - cste, v1.getSup()) + cste, index, expl);
				detectSymetricalHoles(((ExplainedIntVar) v0), (ExplainedIntVar) v1, v1.getSup() + cste, -v1.getInf() - cste);
			}

		}
	}

	/**
	 * When a value is removed from the domain of v0: then this value and its opposite are removed from the domain of v1. When a value is removed from
	 * the domain of v1: IF its opposite is not in the domain of v1 either, then its absolute value is removed from the domain of v0.
	 * 
	 * @param var
	 * @param values
	 *            the values that are removed from domain of <code> var </code> @
	 */

	public void awakeOnRemovals(ExplainedIntVar var) {
		Iterator<Integer> it = var.getDomain().getDeltaIterator();
		if (var.equals(v0)) {
			boolean change = false;
			while (it.hasNext()) {
				int val = it.next();
				if (val >= 0) {
					change = true;
					this.incrementNumConstraintChecks(1);
					((ExplainedIntVar) v1).removeValue(val - cste, index, var.getExplanation(val));
				}
			}
			if (change)
				updateMinFromHoles();// TODO
		} else if (var.equals(v1)) {
			while (it.hasNext()) {
				int val = it.next();
				this.incrementNumConstraintChecks(1);
				if (!v1.canBeInstantiatedTo(-val - cste)) {
					((ExplainedIntVar) v0).removeValue(val + cste, index, var.getExplanation(val));
				}
			}
		}
	}

	/**
	 * Deduce the minimum value of v0 from the eventual central gap in the domain of v1 (v0 = |v1+cste|). @
	 */
	private void updateMinFromHoles() {
		this.incrementNumConstraintChecks(1);
		if (v1.getInf() + cste < 0 && v1.getSup() + cste > 0) {
			int minPositiveValue = v1.getDomain().getNextValue(-1) + cste;
			int maxNegativeValue = v1.getDomain().getPrevValue(1) + cste;
			((ExplainedIntVar) v0).updateInf(Math.min(-maxNegativeValue, minPositiveValue), index, ((ExplainedIntVar) v1).getDomExplanations());
		}
	}

}
