package kernel.constraints.integer.explain;

import java.util.Collection;
import java.util.SortedSet;
import java.util.TreeSet;

import kernel.AbstractConstraint;
import kernel.AbstractVar;
import kernel.VarIdentifier;
import kernel.Variable;
import kernel.events.VariableEvent;
import kernel.vars.integer.explain.ExplainedIntVar;
import simple.SingleExplainedVarAgent;
import simple.dcsp.Nogood;
import simple.dcsp.SimpleConstraint;
import util.Instantiation;
import util.IntArray;
import util.VarAssignment;

public class SENoGoodConstraint extends AbstractConstraint implements SimpleConstraint, Propagator {

	protected ExplainedIntVar[] lhsVars;
	/**
	 * The right hand side of the nogood
	 */
	protected ExplainedIntVar rhsVar;
	/**
	 * 
	 */
	protected IntArray rhsVals = new IntArray();

	public SENoGoodConstraint(SingleExplainedVarAgent owner, Nogood nogood) {
		assert (!nogood.isEmpty());
		this.rhsVar = (ExplainedIntVar) owner.makeVariable(nogood.getRhsVarIdentifier());
		this.lhsVars = new ExplainedIntVar[nogood.getLhs().size()];
		int i = 0;
		for (VarAssignment v : nogood.getLhs()) {
			ExplainedIntVar exV = (ExplainedIntVar) owner.makeVariable(v.getVarIdentifier());
			this.lhsVars[i++] = exV;
			assert (exV.getValue() == v.getValue());
		}
		for (int val : nogood.getRhsValues()) {
			this.rhsVals.add(val);
		}
		this.type = NOGOOD;
	}

	public SENoGoodConstraint(ExplainedIntVar var, Nogood nogood) {
		this((SingleExplainedVarAgent) var.getAgentOwner(), nogood);
		assert (nogood.getRhsVarIdentifier().equals(var.getVarIdentifier()));
	}

	public void setActive() {
		for (ExplainedIntVar v : this.lhsVars) {
			v.addVariableEventListener(VariableEvent.DESINSTANCIATE, this);
		}
		this.rhsVar.addVariableEventListener(VariableEvent.RESTORE_VAL, this);
	}

	public void addRhsValue(int val) {
		this.rhsVals.add(val);
	}

	@Override
	public boolean isConsistent(int internalVal) {
		this.incrementNumConstraintChecks(1);
		if (!this.rhsVals.contains(internalVal))
			return true;
		return false;
	}

	@Override
	public AbstractVar getInternalVar(int idx) {
		throw new UnsupportedOperationException();
	}

	@Override
	public int getNbExternalVars() {
		throw new UnsupportedOperationException();
	}

	@Override
	public AbstractVar getExternalVar(int idx) {
		throw new UnsupportedOperationException();
	}


	@Override
	public Collection<AbstractVar> getExternalVars() {
		throw new UnsupportedOperationException();
	}

	@Override
	public int getNbInternalVars() {
		throw new UnsupportedOperationException();
	}

	@Override
	public int getNbVars() {
		throw new UnsupportedOperationException();
	}

	@Override
	public AbstractVar getVar(int i) {
		if (i >= this.lhsVars.length) { return (AbstractVar) this.rhsVar; }
		return (AbstractVar) this.lhsVars[i];
	}

	@Override
	public boolean isCompletlyInstantiated() {
		return this.rhsVar.isInstantiated();
	}

	@Override
	public boolean isConsistent(Variable v, int val) {
		return (v.equals(rhsVar) && this.rhsVals.contains(val));
	}

	@Override
	public boolean isSatisfied() {
		this.incrementNumConstraintChecks(1);
		if (rhsVar.isInstantiated())
			return !this.rhsVals.contains(rhsVar.getValue());
		return true;
	}

	@Override
	public void awakeOnVariableEvent(Variable v, int evtType) {
		this.incrementNumConstraintChecks(1);
		if (this.InLhs(v) && evtType == VariableEvent.DESINSTANCIATE) {
			for (int val : this.rhsVals)
				this.rhsVar.removeExplanation(val, index);
			this.constraintHandler.removeMe(this);
		}
	}

	public boolean hasLhsEquals(Instantiation lhs) {
		if (lhs.size() != this.lhsVars.length)
			return false;
		for (ExplainedIntVar var : this.lhsVars) {
			if (!lhs.contains(var.getVarIdentifier()))
				return false;
			if (lhs.getVarAssignment(var.getVarIdentifier()).getValue() != var.getValue())
				return false;
		}
		return true;
	}

	private boolean InLhs(Variable v) {
		if (this.lhsVars == null)
			return false;
		for (ExplainedIntVar var : this.lhsVars) {
			if (var.getVarIdentifier().equals(v.getVarIdentifier()))
				return true;
		}
		return false;
	}

	@Override
	public void propagate() {
		this.awake();
	}

	@Override
	public void awakeOnRestores(ExplainedIntVar var) {
		this.awake();
	}

	@Override
	public void awake() {
		this.incrementNumConstraintChecks(1);
		for (ExplainedIntVar v : this.lhsVars) {
			if (!v.isInstantiated()) {
				for (int val : this.rhsVals)
					this.rhsVar.removeExplanation(val, index);
				this.constraintHandler.removeMe(this);
				return;
			}
		}
		for (int val : this.rhsVals) {
			if (this.rhsVar.getDomain().contains(val))
				this.rhsVar.removeValue(val, this.index, this.index);
		}
	}

	@Override
	public SortedSet<AbstractVar> getVars() {
		SortedSet<AbstractVar> vars = new TreeSet<AbstractVar>(this.constraintHandler.myAgent.agentOrder);
		for (ExplainedIntVar v : this.lhsVars)
			vars.add((AbstractVar) v);
		vars.add((AbstractVar) this.rhsVar);
		return vars;
	}

	public Instantiation toInstantiation() {
		Instantiation ass = new Instantiation(this.constraintHandler.myAgent.agentOrder);
		for (ExplainedIntVar v : this.lhsVars)
			if (v.isInstantiated())
				ass.add(v.getVarAssignment());

		return ass;
	}

	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append("C_" + this.index + ":");
		for (int i = 0 ; i < this.lhsVars.length ; i++){
			sb.append(this.lhsVars[i].toString());
		}
		sb.append(" -->   ");
		sb.append(this.rhsVar.getVarIdentifier()+ "!=" + this.rhsVals);
		return sb.toString();
	}

	@Override
	public void setUnPosted() {
		this.rhsVar.removeExplanation(this.index);
		super.setUnPosted();
	}

	@Override
	public void awakeOnDecSup(ExplainedIntVar var) {
		this.awake();
	}

	@Override
	public void awakeOnDesInst(ExplainedIntVar var) {
		this.awake();
	}

	@Override
	public void awakeOnIncrInf(ExplainedIntVar var) {
		this.awake();
	}

	@Override
	public void awakeOnInst(ExplainedIntVar var) {}

	@Override
	public void awakeOnRemovals(ExplainedIntVar var) {
		this.awake();
	}

	// ==================================================================================================

	public boolean hasRhsId(VarIdentifier varId) {
		return this.rhsVar.getVarIdentifier().equals(varId);
	}
}
