/*
DisChoco: Distributed Constraint Reasoning Solver 
Copyright (C) 2007-2013 Redouane Ezzahir & Mohamed Wahbi 

DisChoco is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DisChoco is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

To contact the authors: <http://dischoco.sourceforge.net/>
*/ 
package kernel.constraints.integer.soft;

import static simple.dcop.SimpleSoftConstraint.MAX_COST;
import static simple.dcop.SimpleSoftConstraint.MIN_COST;
import kernel.constraints.integer.CouplesTable;
import simulation.parser.InstanceTokens;


/**
 * This class will be used for presenting soft tables where each couple 
 * has a cost in [0, infinity] 
 * 
 * @version DisChoco 2.5.3<br/>
 * kernel.constraints.integer.soft.SoftCouplesTable.java
 * @author Mohamed Wahbi
 */
public class SoftCouplesTable extends CouplesTable {

	private Integer[] weights;
	
	/**
	 * Create a soft constraint table.
	 * 
	 * @param feas : the feasibility of the table, true if the tuple presented are the allowed tuples,
	 * otherwise, false
	 * @param offset1 : the offset of the first value of the first variable, i.e., x
	 * @param offset2 : the offset of the first value of the second variable, i.e., y
	 * @param n1 : the size of the initial domain of the first variable, i.e., x
	 * @param n2 : the size of the initial domain of the second variable, i.e., y 
	 * @param defaultCost : the default cost of tuples.  
	 */
//	public SoftCouplesTable(boolean feas, int offset1, int offset2, int n1, int n2, Integer defaultCost) {
//		super(feas, offset1, offset2, n1, n2);
//		weights = new Integer[n1*n2];
//		for(int i = 0; i<weights.length ; i++){
//			weights[i] = defaultCost; 
//		}
//	}
	
	
	/**
	 * @return the weights
	 */
	public Integer[] getWeights() {
		return this.weights;
	}

	/**
	 * Create a soft constraint table.
	 * 
	 * @param offset1 : the offset of the first value of the first variable, i.e., x
	 * @param offset2 : the offset of the first value of the second variable, i.e., y
	 * @param n1 : the size of the initial domain of the first variable, i.e., x
	 * @param n2 : the size of the initial domain of the second variable, i.e., y 
	 * @param costs : the default cost of tuples.  
	 */
	public SoftCouplesTable(String name, int offset1, int offset2, int n1, int n2, Integer[] weights, boolean transpose) {
		this.name = name;
		this.transpose = transpose;
		this.offset1 = offset1;
		this.offset2 = offset2;
		this.n1 = n1;
		this.n2 = n2;
		this.transpose = transpose;
		assert ((n1*n2) == weights.length) : "Problem with weights length="+weights.length+" <> "+n1+"*"+n2+" = "+(n1*n2);
		this.weights = new Integer[n1*n2];
		System.arraycopy(weights, 0, this.weights, 0, weights.length);
	}
	
	/**
	 * This method adds a tuple to the table. 
	 * 
	 * @param value_x A value of the first variable.
	 * @param value_y A value of the second variable.
	 * @param cost the cost of the couple.
	 */
	public void setCouple(int value_x, int value_y, int cost) {
		if(this.feasible){
			if (this.transpose){
				++this.nbSupport[value_y - offset1];
			} else { 
				++this.nbSupport[value_x - offset1];
			}
		} else {
			if (this.transpose){
				--this.nbSupport[value_y - offset1];
			} else { 
				--this.nbSupport[value_x - offset1];
			}
		}
		int index = (value_x - offset1) * n1 + value_y - offset2;
		table.set(index);
		weights[index] = cost;
	}
	
	/*
	 * (non-Javadoc)
	 * 
	 * @see simple.integer.constraints.BinRelation#getCost(int, int)
	 */
	public Integer getCost(int value_x, int value_y) {
		int index = (value_x - offset1) * n1 + value_y - offset2;
		if (this.transpose)
			index = (value_y - offset1) * n1 + value_x - offset2;
		return weights[index];
	}
	
	/* (non-Javadoc)
	 * @see kernel.constraints.integer.BinRelation#getMinCost(int)
	 */
	public Integer getMinCost(int value) {
		int minCost = MAX_COST;
		for(int i = 0 ; i < n2 ; i++){
			int value_y = i + offset1;
			minCost = Math.min(minCost, this.getCost(value, value_y));
		}
		return minCost;
	}

	/* (non-Javadoc)
	 * @see kernel.constraints.integer.BinRelation#getMaxCost(int)
	 */
	public Integer getMaxCost(int value) {
		int maxCost = MIN_COST;
		for(int i = 0 ; i < n2 ; i++){
			int value_y = i + offset1;
			maxCost = Math.max(maxCost, this.getCost(value, value_y));
		}
		return maxCost;
	}

	/**
	 * @return
	 */
	public int getN2() {
		return this.n2;
	}

	
	/**
	 * This method returns a string representation of the table. 
	 * @return String : a string representing the table
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append(this.name+" ");
		for(int i=0 ; i<this.n1 ; i++) {
			int value_x = i+offset1;
			for(int j=0 ; j<this.n2 ; j++) {
				int value_y = j + offset2; 
				int index = (value_x - offset1) * n1 + value_y - offset2;
				Integer cost = this.weights[index];
				if(!cost.equals(MAX_COST)) {
					sb.append(cost);
					sb.append(InstanceTokens.COST_SEPARATOR);
					sb.append(value_x);
					sb.append(InstanceTokens.VALUE_SEPARATOR);
					sb.append(value_y);
					sb.append(InstanceTokens.TUPLES_SEPARATOR);
				}
			}
		}
		return sb.substring(0, sb.length()-1);
	}
}
