package kernel.constraints.integer.soft;

import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import kernel.AbstractConstraint;
import kernel.AbstractVar;
import kernel.Variable;
import kernel.vars.integer.IntVariable;
import simple.dcop.SimpleSoftConstraint;
import util.IntArray;

public class SoftUnaryIntConstraint extends AbstractConstraint implements SimpleSoftConstraint {

	private final IntVariable x;
	private Integer[] weights;
	private Integer[] values;

	/**
	 * @param name
	 * @param variable
	 * @param weights
	 */
	public SoftUnaryIntConstraint(String name, IntVariable variable, Integer[] values, Integer[] weights) {
		this.x = variable;
		this.type = name;
		this.values = values.clone();
		int n = this.values .length;
		assert (n == weights.length) : ("Problem with weights length="+weights.length+" <> "+n);
		this.weights = new Integer[n];
		System.arraycopy(weights, 0, this.weights, 0, weights.length);
	}
	
	/* (non-Javadoc)
	 * @see simple.SimpleConstraint#evaluate(int)
	 */
	@Override
	public Integer evaluate(int value) {
		this.incrementNumConstraintChecks(1);
		int index = Arrays.binarySearch(this.values, value);
		return this.weights[index];
	}
	
	/* (non-Javadoc) 
	 * @see simple.SimpleSoftConstraint#getCost(int, int)
	 */
	@Override
	public Integer getCost(int value, int y) {
		return this.evaluate(value);
	}
	

	/*
	 * (non-Javadoc)
	 * 
	 * @see Constraint#isSatisfied()
	 */
	public boolean isSatisfied() {
		if (!this.isCompletlyInstantiated()) {
			this.incrementNumConstraintChecks(1);
			return true;
		}
		int value = this.x.getValue();
		return this.isConsistent(value);
	}

	/**
	 * @param value the value of the internal variable
	 */
	public boolean isConsistent(int value) {
		return !this.evaluate(value).equals(MAX_COST);
	}

	public void propagate() {
		Iterator<Integer> it = this.x.getDomain().iterator();
		while (it.hasNext()) {
			int value = it.next();
			if (!this.isConsistent(value)) {
				this.x.removeValue(value, this.index);
			}
		}
	}

	/* (non-Javadoc)
	 * @see kernel.Constraint#awake()
	 */
	@Override
	public void awake() {
		if (this.varEventSources.contains(x.getVarIdentifier())) {
			IntArray evtTypes = x.getEventsType(this.index);
			for (int e : evtTypes) {
				this.awakeOnVariableEvent(x, e);
			}
		}
	}	

	/*
	 * (non-Javadoc)
	 * 
	 * @see Constraint#isCompletlyInstantiated()
	 */
	public boolean isCompletlyInstantiated() {
		return this.x.isInstantiated();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see simple.SimpleConstraint#getInternalVar()
	 */
	public IntVariable getInternalVar(int index) {
		if (index != 0){
			throw new IndexOutOfBoundsException("index : " + index + " must be equals to zero");
		}
		return this.x;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see Constraint#getNbExternalVars()
	 */
	public int getNbExternalVars() {
		return 0;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see Constraint#getNbInternalVars()
	 */
	public int getNbInternalVars() {
		return 1;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see Constraint#getNbVars()
	 */
	public int getNbVars() {
		return 1;
	}

	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append("C_" + this.getIndex());
		sb.append("[ " + type + "@" + this.x.toString()+ "] ");
		sb.append(", posted=" + this.isPosted());
		sb.append(", cost=" + this.getCost());
		return  sb.toString();
	}
	
	private String getCost() {
		if(this.x.isInstantiated()){
			int myValue = this.x.getValue();
			int index = Arrays.binarySearch(this.values, myValue);
			return Integer.toString(this.weights[index]);
		}
		return new String("?");
	}

	/**
	 * Not used for the unary constraint
	 * 
	 * @throws UnsupportedOperationException
	 */
	public AbstractVar getExternalVar(int index) {
		throw new UnsupportedOperationException("There is not external variables in unary constraint");
	}

	/**
	 * Not used for the unary constraint
	 */
	public void setActiveOnInternalVars() {
		/* noting */
	}

	/**
	 * Not used for the unary constraint
	 */
	public void setPassiveOnInternalVars() {
		/* noting */
	}

	/**
	 * Not used for the unary constraint
	 */
	public void setActiveOnExternalVars() { 
		/* noting */
	}

	/**
	 * Not used for the unary constraint
	 */
	public void setPassiveOnExternalVars() { 
		/* noting */
	}

	@Override
	public void awakeOnVariableEvent(Variable var, int evtType) {
		/* noting */
	}

	@Override
	public boolean isConsistent(Variable var, int value) {
		if (var.equals(x)){
			return this.isConsistent(value);
		} else {
			throw new Error("The given var " + var + " is different from " + x);
		}
	}

	@Override
	public AbstractVar getVar(int index) {
		assert (index == 0) : "index : " + index + " must be equals to zero";
		return x;
	}
	
	/* (non-Javadoc)
	 * @see simple.dcop.SimpleSoftConstraint#getMinCost(int)
	 */
	@Override
	public Integer getMinCost(int value) {
		return this.evaluate(value);
	}
	
	/* (non-Javadoc)
	 * @see simple.dcop.SimpleSoftConstraint#getMaxCost(int)
	 */
	@Override
	public Integer getMaxCost(int value) {
		return this.evaluate(value);
	}

	@Override
	public Collection<AbstractVar> getExternalVars() {
		return Collections.emptySet();
	}
	
}
