package kernel.events;

import kernel.ConstraintsHandler;
import kernel.Variable;
import util.BitSet;

/**
 * The Class variable events.
 */
public abstract class VariableEvent implements DisChocoEvent<Variable>, Comparable<VariableEvent> {

	public static final int DESINSTANCIATE = 1;
	// ====================================================================
	/**
	 * Constant value associated to the inf bound restoration prop.
	 */
	public final static int RESTORE_INF = 2;
	/**
	 * Constant value associated to the sup bound restoration prop.
	 */
	public final static int RESTORE_SUP = 3;

	public static final int RESTORE_VAL = 4;
	// ====================================================================
	public static final int INSTANCIATE = 5;
	// ====================================================================
	public static final int DECREMENT_SUP = 6;
	public static final int INCREMENT_INF = 7;
	public static final int REMOVE_VAL = 8;
	// ====================================================================
	/**
	 * Semantic of the cause of an event: -1 means that the event is active without any precise cause.
	 */
	public static final int NOCAUSE = -1;
	/**
	 * empty bitvector for the event type.
	 */
	public final static int NOEVENT = -2;
	/**
	 * empty bitvector for the event type.
	 */
	public final static int EMPTYEVENT = 0;
	// ====================================================================
	/**
	 * stores the type of update performed on the variable
	 */
	protected int eventType = EMPTYEVENT;
	/**
	 * Cause of the varEvent.
	 */
	protected int cause = NOEVENT;

	/**
	 * A field storing the variable on which the event was occurred
	 */
	protected Variable touchedVar;

	protected ConstraintsHandler constraintHandler;

	// ====================================================================

	public VariableEvent(Variable var) {
		this.touchedVar = var;
		this.constraintHandler = var.getAgentOwner().constraintHandler;
	}

	// ====================================================================

	@Override
	public Variable getTouchedObject() {
		return this.touchedVar;
	}

	// ====================================================================

	public int getEventType() {
		return this.eventType;
	}

	// ====================================================================
	/**
	 * Returns the cause of this varEvent.
	 */
	public int getCause() {
		return cause;
	}

	// ====================================================================

	@Override
	public boolean isActive() {
		return this.eventType != EMPTYEVENT;
	}

	// ====================================================================
	@Override
	public int compareTo(VariableEvent ve) {
		if (this == ve)
			return 0;
		return BitSet.lowestBit(this.eventType) - BitSet.lowestBit(ve.eventType);
	}

	public boolean equals(VariableEvent ve) {
		return (this == ve);
	}

	// =====================================================================

	public void recordEventTypeAndCause(int eventType, int idx) {

		// if no such event was active on the same variable
		if ((cause == NOEVENT) || (this.eventType == EMPTYEVENT)) { // note: these two tests should be equivalent
			// the variable event is reduced to event, and the cause is recorded
			this.eventType = (1 << eventType);
			this.cause = idx;
		} else {
			// otherwise, this basic event is added to all previous updates that
			// are possibly mending on the same variable
			this.eventType = (this.eventType | (1 << eventType));
			// in case the cause of this update is different from the previous
			// cause,
			// all causes are forgotten (so that the constraints that caused the
			// event will be reawaken)
			if (cause != idx) {
				cause = NOCAUSE;
			}
		}
		// this.touchedVar.getAgentOwner().log(this+" recored event " + eventType, 1);

	}

	// ======================================================================================
	/**
	 * useful for debugging
	 */
	public String toString() {
		return ("VarEvt(" + touchedVar.toString() + ")[" + eventType + ":" + (BitSet.getBit(eventType, DESINSTANCIATE)? "D." : "")
				+ (BitSet.getBit(eventType, INSTANCIATE)? "I." : "") + (BitSet.getBit(eventType, RESTORE_VAL)? "Rv." : "")
				+ (BitSet.getBit(eventType, RESTORE_INF)? "Ri." : "") + (BitSet.getBit(eventType, RESTORE_SUP)? "Rs." : "")
				+ (BitSet.getBit(eventType, INCREMENT_INF)? "iI." : "") + (BitSet.getBit(eventType, DECREMENT_SUP)? "dS." : "")
				+ (BitSet.getBit(eventType, REMOVE_VAL)? "Rm." : "") + "]");
	}
	// =======================================================================================

}
