/*****************************************************************************************

 *****************************************************************************************/
package kernel.order;

import java.util.Comparator;
import java.util.HashSet;
import java.util.Vector; 
import kernel.AbstractAgent;
import kernel.AbstractMaster;
import kernel.AgentObserver;
import kernel.StopException;
import kernel.VarIdentifier;
import kernel.com.Message;
import kernel.com.MessageHandler;
import kernel.vars.integer.IntExternalVar;
import simple.SimpleAgent;
import simulation.metric.MetricListener;
import util.Commands;
import util.HashList;
import util.IVarIdentifier;
import util.Logger;
import util.graph.UndirectedGraph;

/**
 * A Class to store and manage the order of variables/agents. It is used as comparator of variables. The guarantee that the order of variables in the
 * {@link kernel.AgentView} and the instantiations
 * 
 * @see util.Instantiation is the same as the current order.
 */

public class AgentOrdering extends MessageHandler implements Comparator<IVarIdentifier> {
	/**
	 * structure containing all available agents orderings in the plate-form
	 */
	public static final String[] ORDERINGS = {Commands.MinDomOverDegAO, Commands.MaxDegPlusMinDomAO, Commands.TreeAO, Commands.LexicoAO, Commands.RandomAO};

	/**
	 * A data structure representing the constraint graph 
	 * @see UndirectedGraph
	 */
	protected UndirectedGraph<VarIdentifier> graph = null;
	
	/**
	 * An array stores the current order. The current order is given by the order of IDs in the array. ID that is in position 0 is the ID of the agent
	 * with the highest priority.
	 */
	protected HashList<VarIdentifier> agentsOrder = new HashList<VarIdentifier>();
	/**
	 * A field indicates whether the order is directed or not. ABT uses a directed order but DBA do not uses a directed order
	 */
	protected boolean directed = false;
	/**
	 * A field indicates whether the order is received from the Master or not; NB. Agents can delegate to the Master to construct agents ordering in a
	 * centralized way.
	 */

	protected boolean rcvOrder = false;

	/**
	 * Constructs an agent ordering object
	 * 
	 * @param agent the owner of this object
	 */
	public AgentOrdering(AbstractAgent agent) {
		super(agent, Commands.AO);
	}
	
	public void initialize() {}

	/**
	 * Set the agent order to the given order
	 * 
	 * @param newOrder
	 *            the newest agent order to set
	 */
	public boolean setNewOrder(HashList<VarIdentifier> newOrder) {
		if (this.agentsOrder.equals(newOrder)){
			this.myAgent.log("The current order "+this.agentsOrder+" equals the new order "+newOrder, Logger.FINE_LEVEL);
			return false;
		}
		if (this.myAgent.agentView != null) {
			HashSet<IntExternalVar> view = this.myAgent.agentView.getSetOfExternalVars();
			this.agentsOrder = newOrder;
			this.myAgent.agentView.refresh(view);
			this.myAgent.log("The order is updated to "+this.agentsOrder+" where and agent view is "+this.myAgent.agentView, Logger.FINE_LEVEL);
		} else {
			this.agentsOrder = newOrder;
			this.myAgent.log("The order is updated to "+this.agentsOrder, Logger.FINE_LEVEL);
		}
		return true;
	}

	/**
	 * Returns the current order
	 * 
	 * @return the value of agentsOrder
	 */
	public HashList<VarIdentifier> getOrder() {
		return this.agentsOrder.clone();
	}

	/**
	 * Checks whether the order is directed or not.
	 */
	public boolean isDirectedOrder() {
		return this.directed;
	}

	/**
	 * Marks that the order is directed
	 */
	public void setDirectedOrder() {
		this.directed = true;
	}

	/**
	 * Returns the rank of the agent with the given ID in the current order
	 */
	public int getRankOf(VarIdentifier varId) {
		return this.agentsOrder.indexOf(varId);
	}

	/**
	 * Returns the varId of the variable/agent of the given rank.
	 */
	public VarIdentifier getAgentOfRank(int rank) {
		// assert(rank < this.agentsOrder.size()) ;
		return this.agentsOrder.get(rank);
	}

	/**
	 *Returns an array that represents agents lower than the agent of the given ID
	 */
	public HashList<VarIdentifier> getAgentsLowerThan(VarIdentifier agId) {
		int rank = this.agentsOrder.indexOf(agId);
		int toIndex = this.agentsOrder.size();
		return this.agentsOrder.subList(rank+1, toIndex);
	}

	/**
	 *Returns an array that represents agents higher than the agent of the given ID
	 */
	public HashList<VarIdentifier> getAgentsHigherThan(VarIdentifier agId) {
		int rank = this.agentsOrder.indexOf(agId);
		return this.agentsOrder.subList(0, rank);
	}

	/**
	 *Returns an array that represents agents who are between the agents of the given IDs
	 */
	public HashList<VarIdentifier> getAgentsBetween(VarIdentifier agId1, VarIdentifier agId2) {
		int rank1 = this.agentsOrder.indexOf(agId1);
		int rank2 = this.agentsOrder.indexOf(agId2);
		assert (rank1==rank2) : "we can not get variables between the same variable agId1="+agId1+" agId2="+agId2;
		int minRank = Math.min(rank1, rank2);
		int maxRank = Math.max(rank1, rank2);
		return this.agentsOrder.subList(minRank+1, maxRank);
	}

	/**
	 * Checks whether the agent with the given ID is lower than the agent owning this object or not.
	 * 
	 */

	public boolean isLower(VarIdentifier agId) {
		return this.getLowerAgents().contains(agId);
	}

	/**
	 * Checks whether the agent with the given ID is higher than the agent owning this object or not.
	 * 
	 */
	public boolean isHigher(VarIdentifier varId) {
		return this.getHigherAgents().contains(varId);
	}

	/**
	 * Returns an array that represents agents lower than the agent owning this object.
	 * 
	 * @see #getAgentsLowerThan(int)
	 */
	public HashList<VarIdentifier> getLowerAgents() {
		if(this.myAgent instanceof SimpleAgent) {
			VarIdentifier varId = ((SimpleAgent) this.myAgent).var.getVarIdentifier();
			return this.getAgentsLowerThan(varId);
		} else {
			throw new UnsupportedOperationException("not yet implemented for complex agent");
		}
	}

	/**
	 * Returns an array that represents agents higher than the agent owning this object.
	 * 
	 * @see #getAgentsHigherThan(int)
	 */

	public HashList<VarIdentifier> getHigherAgents() {
		if(this.myAgent instanceof SimpleAgent) {
			VarIdentifier varId = ((SimpleAgent) this.myAgent).var.getVarIdentifier();
			return this.getAgentsHigherThan(varId);
		} else {
			throw new UnsupportedOperationException("Not yet implemented for complex agent");
		}
	}

	/**
	 * Returns the ID of the agent that follows the agent with the given ID in the agents order.
	 */
	public VarIdentifier getNextOf(VarIdentifier varId) {
		int rank = this.getRankOf(varId) + 1;
		if (rank < this.agentsOrder.size()){
			return this.agentsOrder.get(rank);
		} else{
			return null;
		}
	}

	public VarIdentifier getPreviousOf(VarIdentifier varId) {
		int rank = this.getRankOf(varId) - 1;
		if (rank > -1){
			return this.agentsOrder.get(rank);
		} else {
			return null;
		}
	}

	/**
	 * Compares the order of two agents Returns : getRankOf(agId1) - getRankOf(agId0)
	 * 
	 * @return >0 if agId0 is higher (of high priority) than agId1 <0 if agId0 is lower (of low priority) than agId1 0 if agId0==agId1
	 * @see #getRankOf(int)
	 */
	public int compareOrder(VarIdentifier varId0, VarIdentifier varId1) {
		return (this.getRankOf(varId1) - this.getRankOf(varId0));
	}

	/**
	 * Chooses and returns the lower agent between the given agents with the given IDs
	 */
	public VarIdentifier chooseLower(VarIdentifier varId0, VarIdentifier varId1) {
		if (this.compareOrder(varId0, varId1) < 0){
			return varId0;
		}
		return varId1;
	}

	/**
	 * Chooses and returns the higher agent between the given agents with the given IDs
	 */
	public VarIdentifier chooseHigher(VarIdentifier varId0, VarIdentifier varId1) {
		if (this.compareOrder(varId0, varId1) > 0)
			return varId0;
		return varId1;
	}

	/**
	 * The default agent ordering uses a centralized algorithms that is running by the master. This method is called to send agent complexity to the
	 * master and block waiting for order.
	 * 
	 * @throws StopException
	 */
	public void startDisAO() throws StopException {
		this.sendComplexity();
		this.waitOrder();
	}

	/**
	 * Blocking the agent owning this object in waiting until the order is received.
	 * 
	 * @throws StopException
	 * @throws
	 */
	public void waitOrder() throws StopException {
		while (!rcvOrder) {
			try {
				this.myAgent.blockingReceiveMessage(Commands.AO, 0);
			} catch (InterruptedException e) {
				e.printStackTrace();
				System.exit(-1);
			}
		}
	}

	/**
	 * {@inheritDoc} handles a message ORDER
	 */
	@SuppressWarnings("unchecked")
	public void processMessage(Message msg) throws StopException {
		if (msg.getType().equals(Commands.AO)) {
			processOrder(msg);
		}
		if (msg.contains(Commands.METRIC)) {
			Vector<MetricListener> metrics = (Vector<MetricListener>) msg.get(Commands.METRIC);
			for (MetricListener metric : metrics) {
				this.myAgent.addObserver((AgentObserver) metric);
			}
		}

	}

	public void processOrder(Message msg) {
		this.agentsOrder = HashList.fromString((String) msg.get(Commands.ORDER), VarIdentifier.class);
		this.graph = UndirectedGraph.fromString((String) msg.get(Commands.GRAPH), VarIdentifier.class);
		this.rcvOrder = true;
		// this.myAgent.log("process Order :" + msg.toString(), Logger.LEVEL_TRACE_MESSAGE);
	}

	/**
	 * Called to send complexity to the master when a centralized agent ordering is used.
	 * 
	 * @see AgentComplexity
	 */
	public void sendComplexity() {
		AgentComplexity cmpx = new AgentComplexity();
		cmpx.setMaxDomain(this.myAgent.getMaxDomainSize());
		cmpx.setNumberConstraints(this.myAgent.constraintHandler.getNumConstraints());
		cmpx.setNumberLocalVars(this.myAgent.getNbVars());
		cmpx.setNeighbors(this.myAgent.communicator.getNeighbors());
		cmpx.setLocalAgents(this.myAgent.localAgents);
		Message msg = this.createMessage();
		msg.addContents(Commands.AGENT_COMPLEXITY, cmpx);
		msg.setReceiverId(AbstractMaster.MasterVarId);
		this.myAgent.sendMessage(msg);
	}

	/**
	 * Returns the rank of the agent owning this object.
	 */
	public int getMyRank() {
		if(this.myAgent instanceof SimpleAgent) {
			VarIdentifier varId = ((SimpleAgent) this.myAgent).var.getVarIdentifier();
			return this.getRankOf(varId);
		} else {
			throw new UnsupportedOperationException("not yet implemented for complex agent");
		}
	}

	/**
	 * Return the number of agents (i.e. the size of {@link #agentsOrder}
	 */
	public int getNumAgents() {
		return this.agentsOrder.size();
	}

	/**
	 * Compares the order of two variables with the given identifiers. when they are belong to different agents it returns the result of the
	 * comparison of the order of their agents. When the two variables are belong to the same agent the returned value is difference between the
	 * indices of two variables in the local problem.
	 * 
	 * @see #compareOrder(int, int)
	 */
	public int compare(IVarIdentifier var0, IVarIdentifier var1) {
		VarIdentifier varId0 = new VarIdentifier(var0.getAgentId(), var0.getVarIndex());
		VarIdentifier varId1 = new VarIdentifier(var1.getAgentId(), var1.getVarIndex());
		int comp = this.compareOrder(varId1, varId0);
		return (comp != 0)? comp : var0.compareTo(var1);
	}

	public static CentralizedAO getCentralizedAO(String orderType) {
		if (orderType.equals(Commands.LexicoAO)) { return new LexicographicAO(); }
		if (orderType.equals(Commands.RandomAO)) { return new RandomAO(); }
		if (orderType.equals(Commands.MinDomOverDegAO)) { return new MinDomOverDegAO(); }
		if (orderType.equals(Commands.MaxDegPlusMinDomAO)) { return new MaxDegPlusMinDomAO(); }
		if (orderType.equals(Commands.TreeAO)) { return new TreeAO(); }
		throw new Error("Unknown type of order : " + orderType);

	}

	public HashList<VarIdentifier> getNeighborsOf(VarIdentifier varId) {
		return graph.neighbors(varId);
	}

	/**
	 * Return names of available agent ordering
	 */
	public static String[] getOrderTypes() {
		return ORDERINGS;
	}

	public UndirectedGraph<VarIdentifier> getAgentsGraph() {
		return this.graph;
	}

	/**
	 * 
	 * @param agentIds
	 * @return
	 */
	public VarIdentifier getLowerAgent(HashList<VarIdentifier> varIds) {
		for(VarIdentifier varId : varIds) {
			HashList<VarIdentifier> lowerAgents = this.getAgentsLowerThan(varId);
			if(HashList.intersection(varIds, lowerAgents).size() == 0) { 
				return varId; 
			}
		}
		return null;
	}

	/**
	 * 
	 * @param agentIds
	 * @return
	 */
	public VarIdentifier getHigherAgent(HashList<VarIdentifier> varIds) {
		for(VarIdentifier varId : varIds) {
			HashList<VarIdentifier> lowerAgents = this.getAgentsHigherThan(varId);
			if(HashList.intersection(varIds, lowerAgents).size() == 0) { 
				return varId; 
			}
		}
		return null;
	}

	/**
	 * @param varIds
	 */
	public void updateOrder(HashList<VarIdentifier> varIds) {
		for(VarIdentifier varId : this.agentsOrder) {
			if (!varIds.contains(varId))
				varIds.add(varId);
		}
		this.setNewOrder(varIds);
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return this.agentsOrder.toString();
	}

	
}
