/************************************************************************

 **************************************************************************/

package kernel.order;

import java.util.HashMap;
import kernel.AbstractMaster;
import kernel.StopException;
import kernel.VarIdentifier;
import kernel.com.Message;
import kernel.com.MessageHandler;
import util.Commands;
import util.HashList;
import util.graph.UndirectedGraph;

/**
 * An abstract class for all implementation of centralized agent ordering
 */
abstract public class CentralizedAO extends MessageHandler {

	/**
	 * the master to monitor its execution by the observer
	 */
	protected AbstractMaster master = null;

	protected UndirectedGraph<VarIdentifier> graph = null;

	/**
	 * Set the master to observe
	 * 
	 * @param master
	 */
	public void setMaster(AbstractMaster master) {
		this.master = master;
	}

	/**
	 * A map to store all received agent complexity (complexity by ID)
	 */
	protected HashMap<VarIdentifier, AgentComplexity> agentsComplexity = null;
	/**
	 * A field stores the current number of received complexity. When this number becomes equal to the total number of agents construction of the
	 * agents order begins.
	 */
	protected int tmpRcvdMsgCounter = 0;

	protected HashList<VarIdentifier> order = new HashList<VarIdentifier>();

	/**
	 * Constructs a centralized agent ordering
	 */
	public CentralizedAO() {
		super(null, Commands.AO);
		this.graph = new UndirectedGraph<VarIdentifier>();
		this.agentsComplexity = new HashMap<VarIdentifier, AgentComplexity>();
	}

	/**
	 * Called to build agents order using the received complexities
	 * 
	 * @return IntArray an array representing the resulting order
	 */
	public abstract HashList<VarIdentifier> buildAgentsOrder();

	public HashList<VarIdentifier> getOrder() {
		return new HashList<VarIdentifier>(this.order);
	}

	/**
	 * Adds an agent complexity
	 * 
	 * @param agId the ID of the agent with the given complexity
	 * @param complexity the complexity of the agent with the given ID
	 */
	public void addAgentComplexity(VarIdentifier varId, AgentComplexity complexity) {
		this.agentsComplexity.put(varId, complexity);
	}

	/**
	 * Returns the map of all received complexity
	 */
	public HashMap<VarIdentifier, AgentComplexity> getAgentsComplexity() {
		return this.agentsComplexity;
	}

	/**
	 * {@inheritDoc} Handles a message ORDER. It adds the attached complexity to the map and increment the counter {@link #tmpRcvdMsgCounter}. When
	 * the counter becomes equals to the number of known agents it start the construction of the order and broadcast the resulting order to all
	 * agents.
	 * 
	 */
	@Override
	public void processMessage(Message msg) throws StopException {
		if (msg.getType().equals(Commands.AO)) {
			AgentComplexity cmpx = (AgentComplexity) msg.get(Commands.AGENT_COMPLEXITY);
			VarIdentifier senderId = msg.getSenderId();
			this.addAgentComplexity(senderId, cmpx);
			this.tmpRcvdMsgCounter++;
			// if(Logger.LOG_DEBUG)
			// this.master.log(" number of rcvd msg " + this.tmpRcvdMsgCounter
			// + "  the number of agent is " + this.master.getNumAgents(), Logger.LEVEL_TRACE_MESSAGE);
			if (this.tmpRcvdMsgCounter == this.master.getNumAgents()) {
				for (VarIdentifier v : agentsComplexity.keySet()) {
					//graph.addVertex(v);
					AgentComplexity complexity = agentsComplexity.get(v);
					for (VarIdentifier w : complexity.getNeighbors()) {
						String label = complexity.getConstraintName(w);
						//graph.addVertex(w);
						graph.addEdge(v, w, label);
					}
				}
				this.order = this.buildAgentsOrder();
				this.broadcastOrder();
			}
		}
	}

	/**
	 * Called to broadcast the order.
	 * 
	 * @param order
	 *            an array representing the agent ordering
	 */
	public void broadcastOrder() {
		Message msg = this.master.createMessage(Commands.AO);
		//msg.addContents(Commands.VARID, this.myAgent.getVarIdentifier());
		msg.addContents(Commands.ORDER, order.toString());
		msg.addContents(Commands.GRAPH, graph.toString());
		this.master.broadcast(msg, AbstractMaster.MasterVarId);
	}

}
