package kernel.order;

import java.util.Comparator;
import java.util.HashMap;
import kernel.VarIdentifier;
import kernel.com.Message;
import util.Commands;
import util.HashList;
import util.graph.Vertex;
import util.tree.AbstractTree;
import util.tree.AbstractTreeNode;

public class TreeAO extends CentralizedAO implements TreeOrder {

	public static int DFSCHAIN = 1;
	public static int BFSCHAIN = 2;

	public static short MinDomOverDeg 	= 0;
	public static short MaxDeg 			= 1;
	public static short MaxDynDeg 		= 2;
	public static short MinID 			= 3;
	public static short Random 			= 4;
	public static short MaxDegMinDom 	= 5;
	public static short MLSP 			= 6;
	public static short MaximumCardinalitySet = 7;
	
	public static short Heuristic 	 = MaxDeg;

	/**
	 * Array will keep track of already chosen agents
	 **/
	public int orderSwitch = DFSCHAIN;
	protected AbstractTree<VarIdentifier> constraintTree = null;
	protected HashList<VarIdentifier> order = new HashList<VarIdentifier>();
	protected Comparator<Integer> comparator = null;

	public TreeAO() {
		super();
	}
	
	public TreeAO(short heuristic) {
		super();
		Heuristic = heuristic;
	}
	
	/**
	 * 
	 */
	@Override
	public HashList<VarIdentifier> buildAgentsOrder() {
		if(Heuristic == MaxDegMinDom){
			constraintTree = graph.DFS(new MaxDegMinDomComparator());
		} else if(Heuristic == MinDomOverDeg){
			constraintTree = graph.DFS(new MinDomOverDegComparator());
		} else if(Heuristic == MaxDeg){
			constraintTree = graph.DFS(new MaxDegreeComparator());
		} else if(Heuristic == MaxDynDeg){
			constraintTree = graph.DynamicMaxDegreeDFS();
		} else if(Heuristic == MLSP) {
			constraintTree = graph.MLSPTree();
		} else if(Heuristic == MaximumCardinalitySet) {
			constraintTree = graph.MCSDFS();
		} else {
			constraintTree = graph.DFS();
		}
		try {
			this.saveOrder();
		} catch (Exception e) {
			e.printStackTrace();
		}
		return this.order;
	}

	private void saveOrder() throws Exception {
		if (this.orderSwitch == BFSCHAIN){
			this.order = HashList.fromString(constraintTree.toBFSString(), VarIdentifier.class);
		} else if (this.orderSwitch == DFSCHAIN){
			this.order = HashList.fromString(constraintTree.toDFSString(), VarIdentifier.class);
		}
	}
	
	public HashList<VarIdentifier> getChildren(VarIdentifier varId) {
		AbstractTreeNode<VarIdentifier> subtree = this.constraintTree.getSubTree(varId);
		return subtree.getChildrensData();
	}
	
	public VarIdentifier getParentOf(VarIdentifier varId) {
		return this.constraintTree.getParentData(varId);
	}
	
	public int orderDepth() {
		return constraintTree.getDepth();
	}

	public AbstractTree<VarIdentifier> getPriorityTree() {
		if (orderSwitch < 4) {
			return this.constraintTree;
		} else {
			throw new Error(" getPriorityTree() ");
		}
	}
	
	public HashMap<VarIdentifier, HashList<VarIdentifier>> getNeighborsIds() {
		HashMap<VarIdentifier, HashList<VarIdentifier>> neighborsIds = new HashMap<VarIdentifier, HashList<VarIdentifier>>();
		for (VarIdentifier varId : this.agentsComplexity.keySet()) {
			neighborsIds.put(varId, this.agentsComplexity.get(varId).getNeighbors());
		}
		return neighborsIds;
	}

	/**
	 * 
	 * @param order
	 */
	@Override
	public void broadcastOrder() {
		AbstractTree<VarIdentifier> tree = this.getPriorityTree();
		for (VarIdentifier varId : this.getAgentsComplexity().keySet()) {
			VarIdentifier parent = this.getParentOf(varId);
			HashList<VarIdentifier> children = this.getChildren(varId);
//			HashMap<VarIdentifier, HashList<VarIdentifier>> neighbors = this.getNeighborsIds();
			Message msg = createMessage(Commands.AO);
			msg.addContents(Commands.ORDER, order.toString());
			msg.addContents(Commands.GRAPH, graph.toString());
			msg.addContents("Parent-Id", (parent==null)?"null":parent.toString());
			msg.addContents("Children-Ids", children.toString());
			msg.addContents(Commands.TREE, tree.toString());
			msg.setReceiverId(varId);
			this.master.sendMessage(msg);
		}
	}

	public String pretty() {
		return this.getPriorityTree().pretty();
	}

	public String toString() {
		return  this.getPriorityTree().toString();
	}
	
	/**
	 * MaxDegree: Maximum Connected Node Comparator: Max-degree
	 * 
	 * At each step, for a given vertex, select a not yet visited neighbor having the maximum neighbors. 
	 */
	class MaxDegreeComparator implements Comparator<Vertex<VarIdentifier>> {

		@Override
		public int compare(Vertex<VarIdentifier> o1, Vertex<VarIdentifier> o2) {
			VarIdentifier varId1 = o1.getData();
			AgentComplexity cmpx1 = agentsComplexity.get(varId1);
			VarIdentifier varId2 = o2.getData();
			AgentComplexity cmpx2 = agentsComplexity.get(varId2);
			int deg1 = cmpx1.getNbNeighbors();
			int deg2 = cmpx2.getNbNeighbors();
			int cmp = deg2 - deg1; // max degree
			if (cmp == 0) {
				return o1.compareTo(o2);
			}
			return cmp;
		}
	}
	/**
	 * MaxDegree: Maximum Connected Node Comparator: Max-degree
	 * 
	 * At each step, for a given vertex, select a not yet visited neighbor having the maximum neighbors. 
	 */
	private class MaxDegMinDomComparator implements Comparator<Vertex<VarIdentifier>> {

		@Override
		public int compare(Vertex<VarIdentifier> o1, Vertex<VarIdentifier> o2) {
			VarIdentifier varId1 = o1.getData();
			AgentComplexity cmpx1 = agentsComplexity.get(varId1);
			VarIdentifier varId2 = o2.getData();
			AgentComplexity cmpx2 = agentsComplexity.get(varId2);
			int deg1 = cmpx1.getNbNeighbors();
			int deg2 = cmpx2.getNbNeighbors();
			int cmp = deg2 - deg1; // max degree
			if (cmp == 0) {
				int dom1 = cmpx1.getMaxDomain();
				int dom2 = cmpx2.getMaxDomain();
				cmp = dom1 - dom2;
				if (cmp == 0) {
					return o1.compareTo(o2);
				}
			}
			return cmp;
		}
	}
	/**
	 * MaxDegree: Maximum Connected Node Comparator: Max-degree
	 * 
	 * At each step, for a given vertex, select a not yet visited neighbor having the maximum neighbors. 
	 */
	private class MinDomOverDegComparator implements Comparator<Vertex<VarIdentifier>> {

		@Override
		public int compare(Vertex<VarIdentifier> o1, Vertex<VarIdentifier> o2) {
			VarIdentifier varId1 = o1.getData();
			AgentComplexity cmpx1 = agentsComplexity.get(varId1);
			VarIdentifier varId2 = o2.getData();
			AgentComplexity cmpx2 = agentsComplexity.get(varId2);
			int dom1 = cmpx1.getMaxDomain();
			int deg1 = cmpx1.getNbNeighbors();
			double domOverDeg1 = dom1/deg1;
			assert !Double.isNaN(domOverDeg1) : "Unexpected mathematical result in TreeLayoutTwo:calcPositions";
			int dom2 = cmpx2.getMaxDomain();
			int deg2 = cmpx2.getNbNeighbors();
			double domOverDeg2 = dom2/deg2;
			assert !Double.isNaN(domOverDeg2) : "Unexpected mathematical result in TreeLayoutTwo:calcPositions";
						
			double domOverDeg = (int)(domOverDeg2 - domOverDeg1); // max degree
			if(domOverDeg > 0) {
				return (int) Math.floor(domOverDeg)+1;
			} else if(domOverDeg < 0) {
				return (int) Math.ceil(domOverDeg)-1;
			} else {
				return o1.compareTo(o2);
			}
		}
	}
	
}
