package kernel.order;

import kernel.AbstractAgent;
import kernel.VarIdentifier;
import kernel.com.Message;
import simple.SimpleAgent;
import util.Commands;
import util.HashList;
import util.graph.Vertex;
import util.tree.AbstractTree;

public class TreeDisAO extends AgentOrdering {

	/** my parent */
	protected VarIdentifier parentId;
	/** my children */
	protected HashList<VarIdentifier> childrenIds;
	/** agents graph on tree structure */
	protected AbstractTree<VarIdentifier> orderTree;
	/**
	 * ancestors and descendants : neighbors union children
	 */
	protected HashList<VarIdentifier> descendants; // my sons, their sons, ...
	protected HashList<VarIdentifier> ancestors; // my father, his father, ...

	private HashList<VarIdentifier> pseudoDecendants = null;
	private HashList<VarIdentifier> pseudoAncestors = null;

	public TreeDisAO(AbstractAgent agent) {
		super(agent);
	}

	public void processOrder(Message msg) {
		super.processOrder(msg);
		String parent = (String) msg.get("Parent-Id");
		this.parentId = parent.equals("null")?null:VarIdentifier.fromString(parent);
		this.childrenIds = HashList.fromString((String) msg.get("Children-Ids"), VarIdentifier.class);
		this.orderTree = AbstractTree.fromString((String) msg.get(Commands.TREE), VarIdentifier.class);
		this.rcvOrder = true;
	}

	public void initialize() {
		for (VarIdentifier chId : this.childrenIds) {
			if (!this.myAgent.communicator.isNeighbor(chId)) {
				this.addLink(chId);
			}
		}
		if (parentId!=null && !this.myAgent.communicator.isNeighbor(parentId)) {
			this.addLink(new VarIdentifier(parentId));
		}
		if(this.myAgent instanceof SimpleAgent) {
			VarIdentifier varId = ((SimpleAgent) this.myAgent).var.getVarIdentifier();
			// ----------------------------------------------------------------------------//
			this.descendants = this.orderTree.getDescendantsData(varId);
			// ----------------------------------------------------------------------------//
			this.ancestors = this.orderTree.getAncestorsData(varId);
			this.pseudoAncestors = this.getPseudoAncestors(varId);
			this.pseudoDecendants = this.getPseudoDescendants(varId);
		} else {
			throw new UnsupportedOperationException("not yet implemented for complex agents");
		}
		
	}

	private void addLink(VarIdentifier varId) {
		this.myAgent.makeExternalVar(varId);
		this.myAgent.communicator.addNeighbor(varId);
	}

	public boolean isMyChild(VarIdentifier varId) {
		return this.childrenIds.contains(varId);
	}

	public boolean isMyParent(VarIdentifier varId) {
		return varId.equals(this.parentId);
	}

	/**
	 * @return IntArray
	 */
	public HashList<VarIdentifier> getDescendants() {
		return new HashList<VarIdentifier>(this.descendants);
	}

	/**
	 * @return IntArray
	 */
	public HashList<VarIdentifier> getAncestors() {
		return this.ancestors;
	}

	public HashList<VarIdentifier> getChildrenIds() {
		return new HashList<VarIdentifier>(this.childrenIds).clone();
	}

	public VarIdentifier getParentId() {
		return this.parentId;
	}

	public HashList<VarIdentifier> getPseudoAncestors() {
		return  new HashList<VarIdentifier>(this.pseudoAncestors);
	}

	/**
	 * return the sub-set of ancestors of the given agent (including its parent) that it or one of its descendants is involved in constraints with
	 * 
	 * @param agId
	 * @return
	 */
	public HashList<VarIdentifier> getPseudoAncestors(VarIdentifier varId) {
		HashList<VarIdentifier> pseudoAncestors = new HashList<VarIdentifier>();
		if (this.orderTree.getParent(varId) == null) {
			return pseudoAncestors;
		}
		pseudoAncestors.add(this.orderTree.getParent(varId).getData());
		HashList<VarIdentifier> ancestors = this.orderTree.getAncestorsData(varId);
		for(VarIdentifier ancestor : ancestors) {
			if (this.getNeighborsOf(varId).contains(ancestor)) {
				pseudoAncestors.add(ancestor);
			} else {
				HashList<VarIdentifier> decendants = this.orderTree.getDescendantsData(varId);
				for(VarIdentifier decendant : decendants) {
					if (this.getNeighborsOf(ancestor).contains(decendant)) {
						pseudoAncestors.add(ancestor);
					}
				}
			}
		}
		return pseudoAncestors;
	}

	/**
	 * return the sub-set of descendants of the given agent (including its children) that it or one of its ancestors is involved in constraints with
	 * 
	 * @param varId
	 * @return
	 */
	public HashList<VarIdentifier> getPseudoDescendants(VarIdentifier varId) {
		HashList<VarIdentifier> pseudoDescendants = new HashList<VarIdentifier>();
		if (this.orderTree.getDescendants(varId).size() == 0) {
			return pseudoDescendants;
		}
		HashList<VarIdentifier> decendants = this.orderTree.getDescendantsData(varId);
		for(VarIdentifier decendant : decendants) {
			if (this.getNeighborsOf(varId).contains(decendant)) {
				pseudoDescendants.add(decendant);
			} else {
				HashList<VarIdentifier> ancestors = this.orderTree.getAncestorsData(varId);
				for(VarIdentifier ancestor : ancestors) {
					if (this.getNeighborsOf(ancestor).contains(decendant)) {
						pseudoDescendants.add(decendant);
					}
				}
			}
		}
		pseudoDescendants.addAll(this.getChildrenIds(varId));
		return pseudoDescendants;
	}

	public HashList<VarIdentifier> getChildrenIds(VarIdentifier varId) {
		if (this.orderTree.getSubTree(varId) != null) {
			return this.orderTree.getSubTree(varId).getChildrensData();
		}
		return new HashList<VarIdentifier>();
	}

	/**
	 * 
	 * @param agId
	 * @return
	 */
	public HashList<VarIdentifier> getAncestorsBeforFirstNeighbor(VarIdentifier varId) {
		HashList<VarIdentifier> res = new HashList<VarIdentifier>();
		if (this.orderTree.getParent(varId) == null)
			return res;
		HashList<VarIdentifier> ancestors = this.orderTree.getAncestorsData(varId);
		HashList<VarIdentifier> order = this.myAgent.agentOrder.getAgentsHigherThan(varId);
		for(int i = 0 ; i < order.size() && i < ancestors.size() ; i++) {
			VarIdentifier ancestor = order.get(i);
			if (ancestors.contains(ancestor)) {
				if (!this.getNeighborsOf(varId).contains(ancestor)) {
					res.add(ancestor);
				}
			}
		}
		return res;
	}

	public HashList<Vertex<VarIdentifier>> getDescendants(VarIdentifier varId) {
		return this.orderTree.getDescendants(varId);
	}

	public AbstractTree<VarIdentifier> getPriorityOrderTree() {
		return this.orderTree;
	}

	public HashList<VarIdentifier> getPseudoDescendants() {
		return this.pseudoDecendants;
	}

	public int getMyLevel() {
		return this.getAncestors().size();
	}

	public String toString(){
		StringBuffer sb = new StringBuffer();
		sb.append("parent="+this.getParentId()+", ");
		sb.append("ancestors="+this.ancestors+", ");
		sb.append("children="+this.getChildrenIds()+"] ");
//		sb.append("[linkedD="+IntArray.intersection(this.getDescendants(), this.getNeighborsOf(this.myAgent.a))+"] ");
		return sb.toString();
	}
}
