package kernel.vars.integer;

import java.util.BitSet;
import java.util.Iterator;
import java.util.Random;

/**
 * A class models enumerated integer domains
 */
public class BitSetIntDomain implements IntDomain {

	/**
	 * The name of the domain
	 */
	protected String name;
	/**
	 * The minimal value in the domain
	 **/
	protected int offset;
	/**
	 * The original domain size (constant )
	 **/
	protected int originalSize;
	/**
	 * BitSet of values that are in the domain
	 */
	protected BitSet contents;
	/**
	 * A chained list implementing two subsets of values: - the removed values waiting to be propagated - the removed values being propagated (each
	 * element points to the index of the next element) -1 for the last element
	 */
	protected int[] chain;

	protected int capacity;

	/**
	 * start of the chain for the values waiting to be propagated -1 for empty chains
	 */
	protected int firstIndexToBePropagated = -1;

	/**
	 * start of the chain for the values being propagated -1 for empty chains
	 */
	protected int firstIndexBeingPropagated = -1;

	protected BitSet removedVals;

	/**
	 * Constructs an enumerated domain
	 */
	public BitSetIntDomain(String name, int min, int max) {
		this.name = name;
		this.offset = min;
		this.capacity = max - min + 1;
		this.originalSize = this.capacity;
		this.chain = new int[this.capacity];
		this.contents = new BitSet(this.capacity);
		this.removedVals = new BitSet(this.capacity);
		this.contents.set(0, this.capacity, true);
	}

	/**
	 * Constructs an enumerated domain
	 */
	public BitSetIntDomain(String name, Integer[] sortedValues) {
		this.name = name;
		this.offset = sortedValues[0];
		int heiver = sortedValues[sortedValues.length - 1];
		this.capacity = heiver - this.offset + 1; // number of entries
		this.chain = new int[capacity];
		this.originalSize = sortedValues.length;
		this.contents = new BitSet(this.capacity);
		this.removedVals = new BitSet(this.capacity);
		for (int i = 0 ; i < sortedValues.length ; i++) {
			int index = sortedValues[i] - this.offset;
			this.contents.set(index);
		}

	}

	/**
	 * @param index of the value to return
	 * @return int the value with index in parameter
	 */
	public int getValueWithIndex(int index) {
		if (index < 0 || index > this.originalSize - 1)
			throw new IndexOutOfBoundsException(index + "");
		return this.offset + index;
	}

	/**
	 * Removes temporarily the given value from the domain, this value remains in the original domain
	 * 
	 * @param value
	 *            the value to remove
	 */
	public boolean remove(int value) {
		int i = this.indexOf(value);
		if (i > -1)
			return this.removeIndex(i);
		return false;
	}

	/**
	 * {@inheritDoc}
	 */
	public boolean updateInf(int inf) {
		boolean removed = false;
		int newInfIndex = inf - offset;
		for (int i = contents.nextSetBit(0) ; i < newInfIndex && i > -1 ; i = contents.nextSetBit(i + 1))
			removed |= this.removeIndex(i);
		return removed;
	}

	/**
	 * {@inheritDoc}
	 */
	public boolean updateSup(int sup) {
		boolean removed = false;
		int newsupIndex = sup - offset;
		int i = contents.length() - 1;
		while (i > newsupIndex && i > -1) {
			removed |= this.removeIndex(i--);
		}
		return removed;
	}

	/**
	 * remove the value with the given index
	 * 
	 * @param i
	 *            the index of the value to remove in the domain
	 * @return true if the value with index i was removed and false otherwise
	 */
	public boolean removeIndex(int i) {
		if (this.contents.get(i)) {
			this.contents.set(i, false);
			this.removedVals.set(i);
			chain[i] = firstIndexToBePropagated;
			firstIndexToBePropagated = i;
			return true;
		}
		return false;
	}

	/**
	 * 
	 * Adds the given value in the original domain. When the value is not already in original domain the original domain size is incremented by 1.
	 */
	public boolean addValue(int value) {
		if (!this.originalDomainContains(value)) {
			this.originalSize++;
			this.capacity++;
			IntDomain d = cloneOriginalDomain();
			int inf = d.getInf();
			if (value < inf)
				this.offset = value;
			BitSet newContents = new BitSet();
			newContents.set(value - this.offset);
			Iterator<Integer> it = d.iterator();
			while (it.hasNext())
				newContents.set(it.next() - this.offset);
			this.contents = newContents;
			int[] newChain = new int[this.capacity];
			System.arraycopy(this.chain, 0, newChain, 0, chain.length);
			this.chain = newChain;
			return true;
		}
		return false;
	}

	/**
	 * Checks whether current domain contains the given value or not.
	 * 
	 * @return <i><b>true</b></i> if current domain contains <b>value</b> passed in parameter and return <i><b>false</b></i> otherwise
	 */
	public boolean contains(int value) {
		int idx = this.indexOf(value);
		if (idx > -1)
			return this.contents.get(idx);
		return false;
	}

	/**
	 * Checks whether original domain contains the given value or not.
	 * 
	 * @return <i><b>true</b></i> if original domain contains <b>value</b> passed in parameter and return <i><b>false</b></i> otherwise
	 */
	public boolean originalDomainContains(int value) {
		return (this.contains(value) || this.IsRemoved(value));
	}

	public boolean IsRemoved(int value) {
		int i = this.indexOf(value);
		if (i < 0)
			return false;
		return this.removedVals.get(i);
	}

	/**
	 * @return the index of <b>value</b> passed in parameters
	 */
	public int indexOf(int value) {
		int i = value - offset;
		if (i > -1 && !this.contents.get(i) && !this.removedVals.get(i)) { return -1; }
		return i;
	}

	/**
	 * @return the size of the current domain
	 */
	public int size() {
		return this.contents.cardinality();
	}

	/**
	 * Returns a domain iterator over values in current domain
	 */
	public Iterator<Integer> iterator() {
		return new IntDomIterator(this);
	}

	/**
	 * A class for iterate over values of the current domain
	 */

	protected class IntDomIterator implements Iterator<Integer> {
		protected BitSetIntDomain domain;
		protected int currentIndex = -1;

		private IntDomIterator(BitSetIntDomain dom) {
			domain = dom;
		}

		public boolean hasNext() {
			return (domain.contents.nextSetBit(currentIndex + 1) != -1);
		}

		public Integer next() {
			currentIndex = domain.contents.nextSetBit(currentIndex + 1);
			return currentIndex + offset;
		}

		/* (non-Javadoc)
		 * @see java.util.Iterator#remove()
		 */
		@Override
		public void remove() {
			removeIndex(size()-1);
		}
		
	}

	public Iterator<Integer> getDeltaIterator() {
		return new getDeltaIterator(this);
	}

	protected class getDeltaIterator implements Iterator<Integer> {
		protected BitSetIntDomain domain;
		protected int currentIndex = -1;

		private getDeltaIterator(BitSetIntDomain dom) {
			domain = dom;
			currentIndex = -1;
		}

		public boolean hasNext() {
			if (currentIndex == -1) {
				return (firstIndexBeingPropagated != -1);
			} else {
				return (chain[currentIndex] != -1);
			}
		}

		public Integer next() {
			if (currentIndex == -1) {
				currentIndex = firstIndexBeingPropagated;
			} else {
				currentIndex = chain[currentIndex];
			}
			return currentIndex + offset;
		}

		/* (non-Javadoc)
		 * @see java.util.Iterator#remove()
		 */
		@Override
		public void remove() {
			// TODO Auto-generated method stub	
		}
		
	}

	@Override
	public Iterable<Integer> removedValsIterator() {
		return new removedValuesIterator(this);
	}

	/**
	 * A class for iterate over values of the current domain
	 */

	protected class removedValuesIterator implements Iterable<Integer>, Iterator<Integer>{
		protected BitSetIntDomain domain;
		protected int currentIndex = -1;

		private removedValuesIterator(BitSetIntDomain dom) {
			domain = dom;
		}

		public boolean hasNext() {
			return (domain.removedVals.nextSetBit(currentIndex + 1) != -1);
		}

		public Integer next() {
			currentIndex = domain.removedVals.nextSetBit(currentIndex + 1);
			return currentIndex + offset;
		}

		/* (non-Javadoc)
		 * @see java.util.Iterator#remove()
		 */
		@Override
		public void remove() {
			removeIndex(size()-1);
		}

		/* (non-Javadoc)
		 * @see java.lang.Iterable#iterator()
		 */
		@Override
		public Iterator<Integer> iterator() {
			return this;
		}
		
	}

	/**
	 * Returns a domain iterator over values in original domain
	 */
	public Iterator<Integer> originalDomainIterator() {
		return new OriginalIntDomainIterator(this);
	}

	/**
	 * A class for iterate over values in original domain
	 * 
	 */

	protected class OriginalIntDomainIterator implements Iterator<Integer> {
		protected int offset;
		protected BitSet original;
		protected int currentIndex = -1;

		public OriginalIntDomainIterator(BitSetIntDomain dom) {
			this.offset = dom.offset;
			this.original = new BitSet();
			this.original.or(dom.contents);
			this.original.or(dom.removedVals);
		}

		public boolean hasNext() {
			return (original.nextSetBit(currentIndex + 1) != -1);
		}

		public Integer next() {
			currentIndex = original.nextSetBit(currentIndex + 1);
			return currentIndex + this.offset;
		}

		/* (non-Javadoc)
		 * @see java.util.Iterator#remove()
		 */
		@Override
		public void remove() {
			removeIndex(size()-1);
		}
	}

	/**
	 * Returns the current domain in an integer array
	 */
	public Integer[] toArray() {
		Iterator<Integer> it = this.iterator();
		Integer[] values = new Integer[this.size()];
		int i = 0;
		while (it.hasNext()) {
			values[i] = it.next();
			i++;
		}
		return values;
	}

	/**
	 * @return the value of the inferior bound of the domain
	 */
	public int getInf() {
		if (this.size() == 0)
			return Integer.MIN_VALUE;
		return this.contents.nextSetBit(0) + offset;
	}

	/**
	 * @return the value of the superior bound of the domain
	 */
	public int getSup() {
		if (this.size() == 0)
			return Integer.MIN_VALUE;
		return this.contents.length() - 1 + offset;
	}
	/**
	 * Accessing the smallest value stored in the domain and strictly greater than <i>x</i>. Does not require <i>x</i> to be in the domain.
	 */
	public int getNextValue(int value) {//TODO corrected by Mohamed 2011-04-17
		if (this.size() == 0)
			return Integer.MAX_VALUE;
		if (value < this.getInf())
			return getInf();
		else if(value >= this.getSup()){
			return Integer.MAX_VALUE;
		}
		int next = contents.nextSetBit(value +1 - this.offset);//TODO
		return next + this.offset;
	}

	/**
	 * Return a random value from the current domain
	 */
	public int getRandomValue() {
		int r = new Random().nextInt(this.size());
		int j = 0;
		for (int i = contents.nextSetBit(0) ; i >= 0 ; i = contents.nextSetBit(i + 1)) {
			if (j == r) { return this.getValueWithIndex(i); }
			j++;
		}
		return this.getInf();
	}

	/**
	 * Returns the value preceding <code>value</code>
	 */
	public int getPrevValue(int x) {
		if (this.size() == 0)
			return Integer.MIN_VALUE;
		int i = this.indexOf(x);
		if (i < 0) {
			if (x > this.getSup())
				return this.getSup();
			else
				return Integer.MIN_VALUE;
		}
		int prvi = this.prevSetBit(this.contents, i);
		if (prvi < 0)
			return Integer.MIN_VALUE;
		else {
			return prvi + this.offset;
		}
	}

	private int prevSetBit(BitSet B, int i) {
		if (i >= B.length())
			return B.length() - 1;
		int j = B.nextSetBit(0);
		int jj = -1;
		while (j >= 0 && j < i) {
			jj = j;
			j = B.nextSetBit(j + 1);
		}
		return jj;
	}

	@Override
	public IntDomain cloneOriginalDomain() {
		BitSetIntDomain d = new BitSetIntDomain(name, offset, offset + this.originalSize);
		d.contents = (BitSet) this.contents.clone();
		d.contents.or(removedVals);
		return d;
	}

	/**
	 * The delta domain container is "frozen" (it can no longer accept new value removals) so that this set of values can be iterated as such
	 */
	public void freezeDeltaDomain() {
		// freeze all data associated to bounds for the the event
		// super.freezeDeltaDomain();
		// if the delta domain is already being iterated, it cannot be frozen
		if (firstIndexBeingPropagated != -1) {}// throw new IllegalStateException();
		else {
			// the set of values waiting to be propagated is now "frozen" as such,
			// so that those value removals can be iterated and propagated
			firstIndexBeingPropagated = firstIndexToBePropagated;
			// the container (link list) for values waiting to be propagated is reinitialized to an empty set
			firstIndexToBePropagated = -1;
		}
	}

	/**
	 * after an iteration over the delta domain, the delta domain is reopened again.
	 * 
	 * @return true iff the delta domain is reopened empty (no updates have been made to the domain while it was frozen, false iff the delta domain is
	 *         reopened with pending value removals (updates were made to the domain, while the delta domain was frozen).
	 */
	public boolean releaseDeltaDomain() {
		// release all data associated to bounds for the the event
		// super.releaseDeltaDomain();
		// special case: the set of removals was not being iterated (because the variable was instantiated, or a bound was updated)
		if (firstIndexBeingPropagated == -1) {
			// remove all values that are waiting to be iterated
			firstIndexToBePropagated = -1;
			// return true because the event has been "flushed" (nothing more is awaiting)
			return true;
		} else { // standard case: the set of removals was being iterated
			// empty the set of values that were being propagated
			firstIndexBeingPropagated = -1;
			// if more values are waiting to be propagated, return true
			return (firstIndexToBePropagated == -1);
		}
	}

	/**
	 * cleans the data structure implementing the delta domain
	 */
	public void clearDeltaDomain() {
		firstIndexBeingPropagated = -1;
		firstIndexToBePropagated = -1;
	}

	/**
	 * Returns a string representation of this domain
	 */
	public String toString() {
		StringBuffer buf = new StringBuffer(" Domain      : {");
		int maxDisplay = 10;
		int count = 0;
		for (Iterator<Integer> it = this.iterator() ; (it.hasNext() && count < maxDisplay) ;) {
			int value = it.next();
			count++;
			if (count > 1) {
				buf.append(", ");
			}
			buf.append(value);
		}
		if (this.size() > maxDisplay) {
			buf.append("..., ");
			buf.append(this.getSup());
		}
		buf.append("}");
		return buf.toString();
	}

	@Override
	public void reset() {
		this.contents.or(this.removedVals);
		this.removedVals.clear();
		this.clearDeltaDomain();
	}

	@Override
	public int getOriginalDomainSize() {
		return this.originalSize;
	}

	/* (non-Javadoc)
	 * @see kernel.vars.integer.IntDomain#getMaxConflictValue()
	 */
	@Override
	public int getMaxConflictValue() {
		// TODO Auto-generated method stub
		return 0;
	}

	/* (non-Javadoc)
	 * @see kernel.vars.integer.IntDomain#getMinConflictValue()
	 */
	@Override
	public int getMinConflictValue() {
		// TODO Auto-generated method stub
		return 0;
	}
}
