package kernel.vars.integer;

import java.util.Iterator;


/**
 * Interface for integer domain implementation
 * 
 * @version dischoco 2.5.3<br/>
 * kernel.vars.integer.IntDomain.java
 * @author Mohamed Wahbi
 */
public interface IntDomain extends Iterable<Integer>{
	/**
	 * Updates the inferior bound to the given value and explains it with the given index <code> expIdx </code>
	 */
	public boolean updateInf(int inf);

	/**
	 * Updates the superior bound to the given value and explains it with the given index <code> expIdx </code>
	 */
	public boolean updateSup(int x);

	/**
	 * Deletes temporarily the given value from the domain, this value remains in the original domain when explanation mechanism is activated,
	 * otherwise set any integer
	 * 
	 * @param value
	 *            the value to remove
	 * @return whether the value is removed from the domain or not
	 */
	public boolean remove(int value);

	/**
	 * @return the index of <b>value</b> passed in parameters
	 */
	public int indexOf(int value);

	/**
	 * Accessing the smallest value stored in the domain and strictly greater than <i>x</i>. Does not require <i>x</i> to be in the domain.
	 */
	public int getNextValue(int x);

	/**
	 * Returns the value preceding <code>value</code>
	 */
	public int getPrevValue(int value);

	/**
	 * Returns the current domain in an integer array
	 */
	public Integer[] toArray();

	/**
	 * Adds or restores the given value in the current domain. When the value is not in original domain the original domain size is incremented by 1.
	 */
	public boolean addValue(int value);

	/**
	 * Checks whether current domain contains the given value or not.
	 * 
	 * @return <i><b>true</b></i> if current domain contains <b>value</b> passed in parameter and return <i><b>false</b></i> otherwise
	 */
	public boolean contains(int value);

	public boolean IsRemoved(int i);

	/**
	 * @param index
	 *            of the value to return
	 * @return the value with index in parameter
	 */
	public int getValueWithIndex(int index);

	/**
	 * @return the value of the inferior bound of the domain
	 */
	public int getInf();

	/**
	 * @return the value of the superior bound of the domain
	 */
	public int getSup();

	/**
	 * Return a random value from the current domain
	 */
	public int getRandomValue();

	/**
	 * Return a value with minimal conflict counter from the current domain
	 */
	public int getMinConflictValue();

	/**
	 * Return a value with maximal conflict counter from the current domain
	 */
	public int getMaxConflictValue();

//	/**
//	 * Returns a domain iterator over values in current domain
//	 */
//	public IntDomainIterator iterator();

	/**
	 * Returns a domain iterator over values in original domain
	 */
	public Iterator<Integer> originalDomainIterator();

	/**
	 * Returns an iterator over current removed values
	 */
	public Iterator<Integer> getDeltaIterator();

	public Iterable<Integer> removedValsIterator();

	public IntDomain cloneOriginalDomain();

	/**
	 * Checks whether original domain contains the given value or not.
	 * 
	 * @return <i><b>true</b></i> if original domain contains <b>value</b> passed in parameter and return <i><b>false</b></i> otherwise
	 */
	public boolean originalDomainContains(int value);

	public int getOriginalDomainSize();

	/**
	 * @return the size of the current domain
	 */
	public int size();

	public boolean releaseDeltaDomain();

	public void freezeDeltaDomain();

	public void clearDeltaDomain();

	public void reset();
}
