package kernel.vars.integer;

import kernel.Variable;
import util.IVarIdentifier;
import util.VarAssignment;

/**
 * An interface for all implementations of integer variable in simple mode.
 * 
 * @author Redouane EZZAHIR
 * 
 */
public interface IntVar extends Variable, IVarIdentifier {

	/**
	 * <b>Public user API:</b> <i>Domains :</i> returns the object responsible for storing the enumeration of values in the domain
	 * 
	 * @return the objects responsible for storing the enumeration of values in the domain
	 */
	public IntDomain getDomain();

	/**
	 * <b>Public user API:</b> <i>Domains :</i> retrieves the value immediately (but strictly) after <i>i</i> in the domain
	 * 
	 * @param val
	 *            the pivot value. May or may not be in the domain
	 * @return the value immediately after the domain
	 */

	public int getNextDomainValue(int val);

	/**
	 * <b>Public user API:</b> <i>Domains :</i> retrieves the value immediately (but strictly) before <i>i</i> in the domain.
	 * 
	 * @param val
	 *            the pivot value. May or may not be in the domain
	 * @return the value immediately before the domain
	 */
	public int getPrevDomainValue(int val);

	/**
	 * Returns the lower bound of the variable domain (e.g. the smallest value that the variable can be assigned).
	 * 
	 * @return the domain lower bound
	 */

	public int getInf();

	/**
	 * Returns the upper bound of the variable domain (e.g. the greatest value that the variable can be assigned).
	 * 
	 * @return the domain upper bound
	 */

	public int getSup();

	/**
	 * Returns the current value assigned to variable
	 * 
	 * @return the value of the variable
	 */
	public int getValue();

	public int getPreviousValue();

	public VarAssignment getVarAssignment();

	/**
	 * Assigns the value of this variable
	 * 
	 * @param val
	 *            the value of the variable
	 */
	public void instantiate(int val);

	public void desinstanciate(int index);

	/**
	 * <i>Propagation events</i> updating the lower bound of a variable (ie: removing all value strictly below the new lower bound from the domain).
	 * 
	 * @param val
	 *            a lower bound of the domain (the new one, if better than the one currently stored)
	 * @param idx
	 *            the index of the external constraint
	 */

	public boolean updateInf(int val, int idx);

	/**
	 * <i>Propagation events</i> updating the upper bound of a variable (ie: removing all value strictly above the new upper bound from the domain).
	 * 
	 * @param val
	 *            an upper bound of the domain (the new one, if better than the one currently stored)
	 * @param idx
	 *            the index of the external constraint
	 */
	public boolean updateSup(int val, int idx);

	/**
	 * <i>Propagation events</i> updating the domain of a variable (by removing a value)
	 * 
	 * @param val
	 *            the value that is not in the domain
	 * @param idx
	 *            the index of the external constraint
	 */

	public boolean removeValue(int val, int idx);

	/**
	 * <i>Propagation events</i> updating the domain of a variable (by removing an interval, ie, a sequence of consecutive values)
	 * 
	 * @param min
	 *            the lower bound of the forbidden interval
	 * @param max
	 *            the upper bound of the forbidden interval
	 * @param index
	 *            the index of the external constraint that filters the domain
	 */
	public boolean removeInterval(int min, int max, int index);

	/**
	 * <i>Propagation events</i> restricting a variable domain (ie: removing all other values from the domain)
	 * 
	 * @param val
	 *            the value of the variable which can remain in its field
	 * @param index
	 *            the index of the constraint that cause domain restricting
	 */
	public boolean restrictDomainTo(int val, int index);

	/**
	 * <b>Public user API:</b> <i>Domains :</i> retrieves the number of values in the domain.
	 * 
	 * @return the number of values in the domain
	 */

	public int getDomainSize();

	/**
	 * <b>Public user API:</b> <i>Domains :</i> testing whether a value is in the domain.
	 * 
	 * @param val
	 *            the tested value
	 * @return whether a value is in the domain
	 */

	public boolean canBeInstantiatedTo(int val);

	/**
	 * <b>Public user API:</b> <i>Domains :</i> testing whether one value among a list is present in the domain.
	 * 
	 * @param valList
	 *            the list of values.
	 * @return whether one value among a list is present in the domain
	 */
	public boolean canBeInstantiatedIn(int[] valList);

	public IntDomain cloneOriginalDomain();

	public boolean originalDomainContains(int val);

}
