package kernel.vars.integer;

import kernel.AbstractAgent;
import kernel.AbstractVar;
import kernel.VarIdentifier;
import kernel.events.IntVariableEvent;
import kernel.events.VariableEvent;
import simulation.parser.InstanceTokens;
import util.VarAssignment;

/**
 * A class implements the integer variable for simple mode. It will be the single variable controlled by a simple agent
 */
public class IntVariable extends AbstractVar implements IntVar {
	/**
	 * 
	 */
	private static final long serialVersionUID = -6874299914019781734L;
	/**
	 * A field stores the current assigned value
	 */
	protected int currentValue;
	/**
	 * A field stores previous assigned value
	 */
	protected int previousValue;

	protected IntDomain domain;

	/**
	 * Constructs a new integer variable
	 * 
	 * @param owner
	 *            the agent that controls this variable
	 * @param name
	 *            the name of the variable
	 * @param min
	 *            the min value of its domain
	 * @param max
	 *            the max value of its domain
	 */
	public IntVariable(AbstractAgent owner, String name, int min, int max) {
		super(owner, name);
		this.domain = new BitSetIntDomain("D" + owner.name, min, max);
		this.currentValue = Integer.MIN_VALUE;
		this.previousValue = Integer.MIN_VALUE;
		this.event = new IntVariableEvent(this);

	}

	/**
	 * Constructs a new integer variable
	 * 
	 * @param owner the agent that controls this variable
	 * @param name the name of the variable
	 * @param values The domain of the variable
	 */
	public IntVariable(AbstractAgent owner, String name, Integer[] values) {
		super(owner, name);
		this.domain = new BitSetIntDomain(InstanceTokens.DOMAIN_PREFIX + name, values);
		this.currentValue = Integer.MIN_VALUE;
		this.previousValue = Integer.MIN_VALUE;
		this.event = new IntVariableEvent(this);

	}

	/**
	 * Constructs a new integer variable
	 * 
	 * @param owner the agent that controls this variable
	 * @param name the name of the variable
	 * @param values The domain of the variable
	 */
	public IntVariable(AbstractAgent owner, String name, VarIdentifier varId, Integer[] values) {
		super(owner, name, varId);
		this.domain = new BitSetIntDomain(InstanceTokens.DOMAIN_PREFIX + name, values);
		this.currentValue = Integer.MIN_VALUE;
		this.previousValue = Integer.MIN_VALUE;
		this.event = new IntVariableEvent(this);
	}

	/**
	 * Constructs a new integer variable
	 * 
	 * @param owner the agent that controls this variable
	 * @param min
	 * @param max
	 */
	public IntVariable(AbstractAgent owner, int min, int max) {
		super(owner, owner.name);
		this.domain = new BitSetIntDomain("D" + owner.name, min, max);
		this.currentValue = Integer.MIN_VALUE;
		this.previousValue = Integer.MIN_VALUE;
		this.event = new IntVariableEvent(this);

	}

	/**
	 * Constructs a new integer variable
	 * 
	 * @param owner
	 *            the agent that controls this variable
	 * @param values
	 *            The domain of the variable
	 */
	public IntVariable(AbstractAgent owner, Integer[] values) {
		super(owner, owner.name);
		this.domain = new BitSetIntDomain("D" + owner.name, values);
		this.currentValue = Integer.MIN_VALUE;
		this.previousValue = Integer.MIN_VALUE;
		this.event = new IntVariableEvent(this);
	}

	/**
	 * Get the value of previous value
	 * 
	 * @return the value of previousValue
	 */
	public int getPreviousValue() {
		return this.previousValue;
	}

	/**
	 * {@inheritDoc}
	 */

	public void instantiate(int value) {
		this.timeCounter++;
		if (this.currentValue != value) {
			this.desinstanciate(VariableEvent.NOCAUSE);
			this.currentValue = value;
			this.fireChangedValue(VariableEvent.INSTANCIATE, VariableEvent.NOCAUSE);
		}
	}

	/**
	 * remove the current assigned value of the variable It unassigns the variable
	 */

	public void desinstanciate(int index) {
		if (this.isInstantiated()) {
			this.previousValue = this.currentValue;
			this.currentValue = Integer.MIN_VALUE;
			if (this.previousValue != Integer.MIN_VALUE) {
				// this.owner.log(this+"="+this.previousValue + "" +this.event.toString(),1);
				this.fireChangedValue(VariableEvent.DESINSTANCIATE, index);
			}
		}
	}

	/**
	 * {@inheritDoc}
	 */

	public boolean removeValue(int value, int index) {
		if (this.getDomain().remove(value)) {
			if (this.isInstantiated()) {
				if (value == this.getValue())
					this.desinstanciate(index);
			}
			this.fireChangedDomain(VariableEvent.REMOVE_VAL, index);
			return true;
		}
		return false;

	}

	/**
	 * {@inheritDoc}
	 * 
	 * @param index
	 */

	// public boolean delete(int value, int index) {
	// if (this.getDomain().delete(value)) {
	// if (this.isInstantiated()) {
	// if (value == this.getValue())
	// this.desinstanciate(index);
	// }
	//			 
	// this.fireChangedDomain(VariableEvent.DELETE_VAL, index);
	// return true;
	// }
	// return false;
	// }

	/**
	 * {@inheritDoc}
	 */
	public boolean updateInf(int inf, int index) {
		boolean update = this.getDomain().updateInf(inf);
		if (update) {
			if (this.isInstantiated()) {
				if (this.getValue() < inf)
					this.desinstanciate(index);
			}
			this.fireChangedDomain(VariableEvent.INCREMENT_INF, index);
		}
		return update;
	}

	/**
	 * {@inheritDoc}
	 */
	public boolean updateSup(int sup, int index) {
		boolean update = this.getDomain().updateSup(sup);
		if (update) {
			if (this.isInstantiated()) {
				if (this.getValue() > sup)
					this.desinstanciate(index);
			}

			this.fireChangedDomain(VariableEvent.DECREMENT_SUP, index);
		}
		return update;
	}

	/**
	 * {@inheritDoc}
	 */

	@Override
	public boolean removeInterval(int min, int max, int index) {
		boolean update = false;
		while (min <= max && this.domain.size() > 0) {
			update |= this.domain.remove(min);
			min = this.getNextDomainValue(min);
		}
		if (update) {

			this.fireChangedDomain(VariableEvent.RESTORE_VAL, index);
		}
		return update;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean restrictDomainTo(int val, int index) {
		boolean update = false;
		update |= this.updateInf(val, index);
		update |= this.updateSup(val, index);

		return update;
	}

	/**
	 * {@inheritDoc}
	 */

	public boolean canBeInstantiatedIn(int[] valList) {
		for (int i = 0 ; i < valList.length ; i++) {
			if (this.canBeInstantiatedTo(valList[i]))
				return true;
		}
		return false;
	}

	/**
	 * {@inheritDoc}
	 */
	public boolean canBeInstantiatedTo(int value) {
		return this.getDomain().contains(value);
	}

	/**
	 * {@inheritDoc}
	 */
	public int getValue() {
		return this.currentValue;
	}

	/**
	 * {@inheritDoc}
	 */

	public IntDomain getDomain() {
		return (IntDomain) this.domain;
	}

	/**
	 * {@inheritDoc}
	 */

	public int getInf() {
		return this.getDomain().getInf();
	}

	/**
	 * {@inheritDoc}
	 */

	public int getNextDomainValue(int value) {
		int next = this.getDomain().getNextValue(value);
		return next;
	}

	/**
	 * {@inheritDoc}
	 */

	public int getPrevDomainValue(int value) {
		return this.getDomain().getPrevValue(value);
	}

	/**
	 * {@inheritDoc}
	 */

	public int getSup() {
		return this.getDomain().getSup();
	}

	/**
	 * {@inheritDoc}
	 */
	public boolean isInstantiated() {
		return this.currentValue != Integer.MIN_VALUE;
	}

	/**
	 * {@inheritDoc}
	 */

	public String toString() {
		if (this.currentValue != Integer.MIN_VALUE){
			return "(" + this.varId + "=" + this.currentValue + "/" + this.timeCounter + ")";
		} else {
			return "(" + this.varId + "=" + "?/" + this.timeCounter + ")";
		}
	}

	/**
	 * Returns a variable assignment that represents the assignment of the Variable
	 */
	public VarAssignment getVarAssignment() {
		return new VarAssignment(varId, currentValue, timeCounter);
	}

	/**
	 * @return a clone object of original domain
	 */

	public IntDomain cloneOriginalDomain() {
		return ((IntDomain) this.domain).cloneOriginalDomain();
	}

	@Override
	public int getDomainSize() {
		return this.domain.size();
	}

	public boolean originalDomainContains(int val) {
		return this.domain.originalDomainContains(val);
	}

	public int getOriginalDomainSize() {
		return ((IntDomain) this.domain).getOriginalDomainSize();
	}

}
