package kernel.vars.integer.explain;

import java.util.Collection;
import static simulation.parser.InstanceTokens.DOMAIN_PREFIX;
import kernel.AbstractAgent;
import kernel.AbstractVar;
import kernel.Constraint;
import kernel.VarIdentifier;
import kernel.events.ExplainedIntVarEvent;
import kernel.events.VariableEvent;
import kernel.events.VariableEventListener;
import kernel.vars.integer.IntExternalVar;
import kernel.vars.integer.IntVar;
import simple.SingleExplainedVarAgent;
import util.IntArray;
import util.VarAssignment;

@SuppressWarnings("serial")
public class ExplainedIntExternalVar extends IntExternalVar implements ExplainedIntVar {

	public ExplainedIntExternalVar(AbstractAgent owner, IntVar v, String name) {
		this(owner, v.getVarIdentifier(), name, v.getDomain().toArray());
		this.name = v.getName();
	}

	public ExplainedIntExternalVar(AbstractAgent owner, VarIdentifier varId, int min, int max) {
		super(owner, varId, min, max);
		this.domain = new ExplainedBitSetIntDomain(DOMAIN_PREFIX + owner.name, min, max);
		this.event = new ExplainedIntVarEvent(this);

	}

	public ExplainedIntExternalVar(AbstractAgent owner, VarIdentifier varId, String name, Integer[] values) {
		super(owner, varId, name, values);
		this.domain = new ExplainedBitSetIntDomain(DOMAIN_PREFIX + owner.name, values);
		this.event = new ExplainedIntVarEvent(this);
	}

	public ExplainedIntExternalVar(AbstractAgent owner, VarAssignment varAss) {
		this(owner, varAss.getVarIdentifier(), varAss.getValue(), varAss.getValue());
		this.timeCounter = varAss.getTimeCounter();
		this.currentValue = varAss.getValue();
	}

	/**
	 * Removes a listener (constraint) from the listener list of the variable
	 */
	public void removeVariableEventListener(int eventType, VariableEventListener listener) {
		if (eventType == VariableEvent.DESINSTANCIATE) {
			for (AbstractVar v : ((Constraint) listener).getVars())
				if (this.isInstantiated())
					((ExplainedIntVar) v).removeExplanation(((Constraint) listener).getIndex());
		}
		super.removeVariableEventListener(eventType, listener);
	}

	@Override
	public boolean removeExplanation(int value, int index) {
		boolean restored = false;
		if (this.getDomain().removeExplanation(value, index)) {
			// this.owner.log(this + "  removeExplanation "+ index + "  val:"+ value + "       ", 1);

			this.fireChangedDomain(VariableEvent.RESTORE_VAL, index);
			restored = true;
		}

		return restored;
	}

	@Override
	public boolean removeExplanation(int index) {
		boolean restored = this.getDomain().removeExplanation(index);
		if (restored) {
			// this.owner.log(this + "  removeExplanation " + index + "       ", 1);
			this.fireChangedDomain(VariableEvent.RESTORE_VAL, index);
		}
		return restored;
	}

	@Override
	public boolean removeExplanationForInterval(int min, int max, int index) {
		// this.owner.log(this+"  removeExplanation "+ index, 1);
		boolean restored = false;
		for (int val = min ; val <= max ; val++) {
			if (this.getDomain().originalDomainContains(val)) {
				restored |= this.getDomain().removeExplanation(val, index);
			}
		}
		if (restored) {
			this.fireChangedDomain(VariableEvent.RESTORE_VAL, index);
		}
		return restored;
	}

	@Override
	public ExplainedIntDomain getDomain() {
		return (ExplainedIntDomain) this.domain;
	}

	public boolean removeValue(int val, int index, int expl) {
		if (this.getDomain().remove(val, expl)) {
			// this.owner.log(this + "  removeValue " + val+ "  because " + expl + " ", 1);

			if (this.isInstantiated()) {
				if (val == this.getValue())
					this.desinstanciate(index);
			}
			this.fireChangedDomain(VariableEvent.REMOVE_VAL, index);
			return true;
		}
		return false;

	}

	@Override
	public boolean updateInf(int inf, int index, int expl) {
		boolean update = this.getDomain().updateInf(inf, expl);
		if (update) {
			if (this.isInstantiated()) {
				if (this.getValue() < inf)
					this.desinstanciate(index);
			}
			this.fireChangedDomain(VariableEvent.INCREMENT_INF, index);
			// this.fireChangedDomain(VariableEvent.REMOVE_VAL, index);
		}
		return update;
	}

	@Override
	public boolean updateSup(int sup, int index, int expl) {
		boolean update = this.getDomain().updateSup(sup, expl);
		if (update) {
			if (this.isInstantiated()) {
				if (this.getValue() > sup)
					this.desinstanciate(index);
			}
			this.fireChangedDomain(VariableEvent.DECREMENT_SUP, index);
			// this.fireChangedDomain(VariableEvent.REMOVE_VAL, index);
		}
		return update;
	}

	@Override
	public boolean removeInterval(int min, int max, int index, int nogood) {
		boolean update = false;
		for (int val = this.getNextDomainValue(min - 1) ; val <= max ; val = this.getNextDomainValue(val))
			update |= this.getDomain().remove(val, nogood);
		if (this.isInstantiated()) {
			if (!this.domain.contains(this.getValue()))
				this.desinstanciate(index);
		}
		if (update)
			this.fireChangedDomain(VariableEvent.REMOVE_VAL, index);
		return update;

	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean restrictDomainTo(int value, int index, int expl) {
		return this.updateInf(value, index, expl) || this.updateSup(value, index, expl);
	}

	/**
	 * @return a clone object of original domain
	 */
	@Override
	public ExplainedIntDomain cloneOriginalDomain() {
		return this.getDomain().cloneOriginalDomain();
	}

	@Override
	public Collection<Explanation> getExplanations(IntArray vals) {
		return this.getDomain().getExplanations(vals);
	}

	@Override
	public Collection<Explanation> getExplanation(int val) {
		return this.getDomain().getExplanation(val);
	}

	@Override
	public boolean updateInf(int val, int idx) {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean updateSup(int val, int idx) {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean removeValue(int val, int idx) {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean removeInterval(int min, int max, int index) {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean restrictDomainTo(int val, int index) {
		throw new UnsupportedOperationException();
	}

	@Override
	public int makeConstraintNogood(int val, Collection<Explanation> expl) {
		SingleExplainedVarAgent agent = (SingleExplainedVarAgent) this.owner;
		int idx = agent.buildNogoodConstraint(this, val, expl);
		return idx;
	}

	@Override
	public int makeConstraintNogood(IntArray values, Collection<Explanation> expl) {
		SingleExplainedVarAgent agent = (SingleExplainedVarAgent) this.owner;
		int idx = agent.buildNogoodConstraint(this, values, expl);
		return idx;
	}

	@Override
	public Collection<Explanation> getDomExplanations() {
		return this.getDomain().getDomExplanations();
	}

	@Override
	public Collection<Explanation> getInfExplanations() {
		return this.getDomain().getInfExplanations();
	}

	@Override
	public Collection<Explanation> getSupExplanations() {
		return this.getDomain().getSupExplanations();
	}

	@Override
	public boolean removeValue(int value, int index, Collection<Explanation> expl) {
		int exp = makeConstraintNogood(value, expl);
		return this.removeValue(value, index, exp);
	}

	@Override
	public boolean restrictDomainTo(int value, int index, Collection<Explanation> expl) {
		return this.updateInf(value, index, expl) || this.updateSup(value, index, expl);
	}

	@Override
	public boolean removeInterval(int min, int max, int index, Collection<Explanation> expl) {
		IntArray values = new IntArray();
		for (int value = this.getNextDomainValue(min - 1) ; value <= max ; value = this.getNextDomainValue(value)){
			values.add(value);
		}
		if(values.size()>0){
			int nogood = this.makeConstraintNogood(values, expl);
			return this.removeInterval(min, max, index, nogood);
		}
		return false;
	}

	@Override
	public boolean updateInf(int inf, int index, Collection<Explanation> expl) {
		IntArray vals = new IntArray();
		for (int value = this.getInf() ; value < inf ; value = this.getNextDomainValue(value)){
			vals.add(value);
		}
		int exp = this.makeConstraintNogood(vals, expl);
		return this.updateInf(inf, index, exp);
	}
	
	@Override
	public boolean updateSup(int sup, int index, Collection<Explanation> expl) {
		IntArray vals = new IntArray();
		for (int value = this.getSup() ; value > sup ; value = this.getPrevDomainValue(value)){
			vals.add(value);
		}
		int exp = this.makeConstraintNogood(vals, expl);
		return this.updateSup(sup, index, exp);
	}

}
