/************************************************************************

 **************************************************************************/

package simple;

import java.util.Collection;
import kernel.AbstractAgent;
import kernel.Constraint;
import kernel.ExternalVar;
import kernel.VarIdentifier;
import kernel.constraints.integer.ExtensionBinConstraint;
import kernel.constraints.integer.SAbsolute;
import kernel.constraints.integer.SDistanceXYC;
import kernel.constraints.integer.SEqualXYC;
import kernel.constraints.integer.SGreaterThanEqualXYC;
import kernel.constraints.integer.SGreaterThanXYC;
import kernel.constraints.integer.SLessThanEqualXYC;
import kernel.constraints.integer.SLessThanXYC;
import kernel.constraints.integer.SNotEqualXYC;
import kernel.constraints.integer.SUnaryIntConstraint;
import kernel.constraints.integer.soft.SoftExtensionBinConstraint;
import kernel.constraints.integer.soft.SoftTuplesConstraint;
import kernel.constraints.integer.soft.SoftUnaryIntConstraint;
import kernel.vars.integer.IntExternalVar;
import kernel.vars.integer.IntValSelector;
import kernel.vars.integer.IntVar;
import kernel.vars.integer.IntVariable;
import kernel.vars.integer.MinIntValSelector;
import simple.dcsp.SimpleConstraint;
import simulation.metric.Metric;
import simulation.parser.PVariable;
import util.Logger;
import util.VarAssignment;

/**
 * An abstract agent for the simple mode. This class defines methods to make internal variable and all types of external constraints for a simple
 * model (One variable per agent).
 */
public abstract class SimpleAgent extends AbstractAgent implements SimpleModel {

	/**
	 * The single internal variable controlled by the agent
	 */
	public IntVariable var;
	/**
	 * A value selector for the controlled variable
	 */
	protected IntValSelector valSelector;
	public final Constraint TRUE = null;
	public final Constraint FALSE = null;

	/**
	 * Constructs a simple agent
	 * 
	 * @param agentId the ID of the agent
	 * @param agName the name of the agent
	 * @param description A string given a description the agent
	 */
	public SimpleAgent(int agId, String name, String description) {
		super(agId, name, description);
		this.valSelector = new MinIntValSelector(this);
	}


	/**
	 * Constructs a simple agent
	 * 
	 * @param variable 
	 */
	public SimpleAgent(PVariable variable) {
		super(variable.getAgentId(), variable.getName(), variable.getDescription());
		this.makeInternalVar(variable);
		this.valSelector = new MinIntValSelector(this);
	}

	public IntVariable makeVariable(VarIdentifier varId) {
		if (this.agentId == varId.getAgentId()) return this.var;
		else return this.makeExternalVar(varId);
	}

	public IntVariable makeVariable(String name, VarIdentifier varId, Integer...values) {
		if (this.agentId == varId.getAgentId()) return this.var;
		else return this.makeExternalVar(name, varId, values);
	}

	/**
	 * {@inheritDoc}
	 */

	public int getNbVars() {
		return 1;
	}

	public int getMaxDomainSize() {
		return this.var.getDomainSize();
	}

	/* (non-Javadoc)
	 * @see kernel.AbstractAgent#solutionToString()
	 */
	@Override
	public String solutionToString() {
		return this.var.toString();
	}

	public String toString() {
		return "Agent " + this.name + " : variable=" + this.var + ", neighbors="+this.communicator.getNeighbors()+", constraints="+this.constraintHandler.getConstraints();
	}

	/**
	 * Constructs the single controlled internal variable
	 */
	public IntVariable makeInternalVar(String name, int id, Integer[] values) {
		this.var = new IntVariable(this, name, values);
		this.var.setVarIdentifier(new VarIdentifier(this.agentId, id));
		this.addVariable(var.getVarIdentifier(), var);
		return var;
	}

	/**
	 * Constructs the single controlled internal variable
	 */
	public IntVariable makeInternalVar(PVariable variable) {
		return this.makeInternalVar(variable.getName(), variable.getVarIdentifier(), variable.getDomain().getValues());
	}

	/**
	 * Constructs the single controlled internal variable
	 */
	public IntVariable makeInternalVar(String name, VarIdentifier varId, Integer[] values) {
		this.var = new IntVariable(this, name, varId, values);
		this.addVariable(varId, var);
		return var;
	}
	
	/**
	 * Constructs the single controlled internal variable
	 */
	public IntVariable makeInternalVar(Integer[] domain) {
		this.var = new IntVariable(this, domain);
		//this.var.setVarIdentifier(new VarIdentifier(this.agentId));
		this.addVariable(var.getVarIdentifier(), var);
		return var;
	}

	/**
	 * Constructs the single controlled internal variable
	 */
	public IntVariable makeInternalVar(int min, int max) {
		this.var = new IntVariable(this, min, max);
		//this.var.setVarIdentifier(new VarIdentifier(this.agentId));
		this.addVariable(var.getVarIdentifier(), var);
		return this.var;
	}

	@Override
	public IntExternalVar makeExternalVar(VarIdentifier varId) {
		return this.makeExternalVar(new VarAssignment(varId, Integer.MIN_VALUE));
	}

	public IntExternalVar makeExternalVar(IntVar v) {
		if (!this.containsVariable(v.getVarIdentifier())) {
			IntExternalVar exV = new IntExternalVar(this, v);
			this.addVariable(exV.getVarIdentifier(), exV);
			return exV;
		}
		return (IntExternalVar) this.getVariable(v.getVarIdentifier());
	}

	public IntExternalVar makeExternalVar(VarAssignment v) {
		if (!this.containsVariable(v.getVarIdentifier())) {
			IntExternalVar exV = new IntExternalVar(this, v);
			this.addVariable(exV.getVarIdentifier(), exV);
			return exV;
		}
		return (IntExternalVar) this.getVariable(v.getVarIdentifier());

	}

	public IntExternalVar makeExternalVar(VarIdentifier vId, int min, int max) {
		if (!this.containsVariable(vId)) {
			IntExternalVar exV = new IntExternalVar(this, vId, min, max);
			this.addVariable(exV.getVarIdentifier(), exV);
			return exV;
		}
		return (IntExternalVar) this.getVariable(vId);
	}

	public IntExternalVar makeExternalVar(String name, int agId, int id, Integer[] values) {
		return makeExternalVar(name, new VarIdentifier(agId, id), values);
	}

	public IntExternalVar makeExternalVar(PVariable variable) {
		return makeExternalVar(variable.getName(), variable.getVarIdentifier(), variable.getDomain().getValues());
	}

	public IntExternalVar makeExternalVar(String name, VarIdentifier varId, Integer[] values) {
		if (!this.containsVariable(varId)) {
			IntExternalVar exV = new IntExternalVar(this, varId, name, values);
			this.addVariable(exV.getVarIdentifier(), exV);
			return exV;
		}
		return (IntExternalVar) this.getVariable(varId);
	}

	public ExternalVar getExternalVar(int agId) {
		return this.getExternalVar(new VarIdentifier(agId));
	}

	public ExternalVar getExternalVar(VarIdentifier varId) {
		if (this.containsVariable(varId)){
			return (ExternalVar) this.getVariable(varId);
		}
		return null;
	}

	// //////////////////////////////////////////////////////////////////////////
	// ///////////// Unary constraints /////////////////////////////////////
	/**
	 * Makes an unary constraint that maps each value to a cost
	 */	
	@Override
	public Constraint makeSoftRelation(String name, PVariable variable, Integer[] values, Integer[] weights){
		if(this.var ==null) {
			this.makeInternalVar(variable);
		}
		Constraint constraint = new SoftUnaryIntConstraint(name, this.var, values, weights);
		this.addConstraint(constraint);
		return constraint;
	}

	/**
	 * Makes unary constraint X = cste
	 * 
	 * @param cste A constant
	 */
	public Constraint eqC(int cste) {
		Constraint exC = new SUnaryIntConstraint(Constraint.eq, var, cste);
		this.addConstraint(exC);
		return exC;
	}

	/**
	 * Makes unary constraint X != cste
	 * 
	 * @param cste
	 *            A constant
	 */

	public Constraint neqC(int cste) {
		Constraint exC = new SUnaryIntConstraint(Constraint.neq, var, cste);
		this.addConstraint(exC);
		return exC;
	}

	/**
	 * Makes unary constraint X <= cste
	 * 
	 * @param cste A constant
	 */
	public Constraint leqC(int cste) {
		Constraint exC = new SUnaryIntConstraint(Constraint.lt, var, cste + 1);
		this.addConstraint(exC);
		return exC;
	}

	/**
	 * Makes unary constraint X >= cste
	 * 
	 * @param cste A constant
	 */
	public Constraint geqC(int cste) {
		Constraint exC = new SUnaryIntConstraint(Constraint.gt, var, (cste + 1));
		this.addConstraint(exC);
		return exC;
	}

	/**
	 * Makes unary constraint X < cste with the given cost.
	 * 
	 * @param cste
	 *            A constant
	 */
	public Constraint ltC(int cste) {
		return this.leqC(cste - 1);
	}

	/**
	 * Makes unary constraint X > cste
	 * 
	 * @param cste
	 *            A constant
	 */
	public Constraint gtC(int cste) {
		return this.geqC(cste + 1);
	}

	// //////////////////////////////////////////////////////////////////////////
	/**
	 * Make the constraint X = Y when X is the internal variable, Y is the given external variable <code>y</code>.
	 */
	public Constraint eqY(IntExternalVar y) {
		Constraint exC = this.eqYplusC(y, 0);
		return exC;
	}

	/**
	 * Make the constraint X <= Y when X is the internal variable, Y is the given external variable <code>y</code>.
	 */
	public Constraint leqY(IntExternalVar y) {
		Constraint exC = this.leqYplusC(y, 0);
		return exC;
	}

	/**
	 * Make the constraint X >= Y when X is the internal variable, Y is the given external variable <code>y</code>.
	 */
	public Constraint geqY(IntExternalVar y) {
		Constraint exC = this.geqYplusC(y, 0);
		return exC;
	}

	/**
	 * Make the constraint X < Y when X is the internal variable, Y is the given external variable <code>y</code>.
	 */
	public Constraint ltY(IntExternalVar y) {
		Constraint exC = this.leqYplusC(y, -1);
		return exC;
	}

	/**
	 * Make the constraint X > Y when X is the internal variable, Y is the given external variable <code>y</code>.
	 */
	public Constraint gtY(IntExternalVar y) {
		Constraint exC = this.geqYplusC(y, +1);
		return exC;
	}

	/**
	 * Make the constraint X = |Y| when X is the internal variable, Y is the given external variable <code>y</code>.
	 */
	public Constraint eqAbsY(IntExternalVar y) {
		SimpleConstraint exC = new SAbsolute(this.var, y);
		this.addConstraint(exC);
		return exC;
	}

	/**
	 * Make the constraint X = |Y + C| when X is the internal variable, Y is the given external variable <code>y</code> and C is a positive constant.
	 */
	public Constraint eqAbsYplusC(IntExternalVar y, int c) {
		SimpleConstraint exC = new SAbsolute(this.var, y, c);
		this.addConstraint(exC);
		return exC;
	}

	/**
	 * Make the constraint X != Y when X is the internal variable, Y is the given external variable <code>y</code>.
	 */
	public Constraint neqY(IntExternalVar y) {
		Constraint exC = this.neqYplusC(y, 0);
		return exC;
	}

	/**
	 * Make the constraint X = Y + C when X is the internal variable, Y is the given external variable <code>y</code> , and C is a positive constant.
	 */
	public Constraint eqYplusC(IntExternalVar y, int c) {
		SimpleConstraint exC = new SEqualXYC(this.var, y, c);
		this.addConstraint(exC);
		return exC;
	}

	/**
	 * Make the constraint X != Y + C when X is the internal variable, Y is the given external variable <code>y</code> , and C is a positive constant.
	 */
	public Constraint neqYplusC(IntExternalVar y, int c) {
		SimpleConstraint exC = new SNotEqualXYC(this.var, y, c);
		this.addConstraint(exC);
		return exC;
	}

	/**
	 * Make the constraint X <= Y + C when X is the internal variable, Y is the given external variable <code>y</code> , and C is a positive constant.
	 */
	public Constraint leqYplusC(IntExternalVar y, int c) {
		SimpleConstraint exC = new SLessThanEqualXYC(this.var, y, c);
		this.addConstraint(exC);
		return exC;
	}

	/**
	 * Make the constraint X < Y + C when X is the internal variable, Y is the given external variable <code>y</code> , and C is a positive constant.
	 */
	public Constraint ltYplusC(IntExternalVar y, int c) {
		SimpleConstraint exC = new SLessThanXYC(this.var, y, c);
		this.addConstraint(exC);
		return exC;
	}

	/**
	 * Make the constraint X >= Y + C when X is the internal variable, Y is the given external variable <code>y</code>, and C is a constant
	 */
	public Constraint geqYplusC(IntExternalVar y, int c) {
		SimpleConstraint exC = new SGreaterThanEqualXYC(this.var, y, c);
		this.addConstraint(exC);
		return exC;
	}

	/**
	 * Make the constraint X > Y + C when X is the internal variable, Y is the given external variable <code>y</code>, and C is a positive constant.
	 */
	public Constraint gtYplusC(IntExternalVar y, int c) {
		SimpleConstraint exC = new SGreaterThanXYC(this.var, y, c + 1);
		this.addConstraint(exC);
		return exC;
	}

	/**
	 * Make the constraint |X - Y| Operator C when X is the internal variable, Y is the given external variable <code>y</code>, and C is a positive
	 * constant.
	 */
	public Constraint distanceXYC(IntExternalVar y, int c, int operator) {
		if (operator == Constraint.GEQ) {
			c--;
			operator = Constraint.GT;
		}
		if (operator == Constraint.LEQ) {
			operator = Constraint.LT;
			c++;
		}
		SimpleConstraint exC = new SDistanceXYC(this.var, y, c, operator);
		this.addConstraint(exC);
		return exC;
	}

//
//	/**
//	 * Create an extension binary constraints with non allowed tuples. 
//	 * 
//	 * @param var : the external variable
//	 * @param tuples : list of forbidden tuples.
//	 * @return Constraint the binary extension constraint.
//	 */
//	public Constraint feasiblePair(IntExternalVar var, List<int[]> tuples) {
//		Constraint exC = new SoftExtensionBinConstraint(Constraint.feasP, this.var, var, tuples);
//		this.addConstraint(exC);
//		return exC;
//	}
//
//	/**
//	 * Create an extension binary constraints with non allowed tuples.
//	 * 
//	 * @param var : the external variable
//	 * @param tuples : list of allowed tuples.
//	 * @return Constraint the binary extension constraint.
//	 */
//	public Constraint infeasiblePair(IntExternalVar var, List<int[]> tuples) {
//		Constraint exC = new SoftExtensionBinConstraint(Constraint.infeasP, this.var, var, tuples);
//		this.addConstraint(exC);
//		return exC;
//	}

	/**
	 * Create an extension binary constraints with non allowed tuples. 
	 * 
	 * @param var : the external variable
	 * @param tuples : list of forbidden tuples.
	 * @return Constraint the binary extension constraint.
	 */
//	public Constraint softfeasiblePair(IntExternalVar var, List<int[]> tuples, List<Integer> costs, Integer defaultCost) {
//		//TODO 
//		Constraint exC = new SoftExtensionBinConstraint(Constraint.soft, this.var, var, tuples, costs, defaultCost);
//		this.addConstraint(exC);
//		return exC;
//	}

	@Override
	public void logDomains() {
		this.log("My" + this.var.getDomain().toString(), Logger.LEVEL_TRACE_FUNCTION_1);
	}

//	public Constraint makePairAC(IntExternalVar exv, List<int[]> list, boolean feas) {
//		if (feas) { return this.feasiblePair(exv, list); }
//		return this.infeasiblePair(exv, list);
//	}

	@Override
	public Collection<Metric> getAuxiliaryStatistics() {
		this.auxiliaryStatistics.put("nbAssignments", new Metric("nbAssignments", this.var.getTimeCounter()));
		return this.auxiliaryStatistics.values();
	}
	
	/**
	 * Create an extension binary constraints with non allowed tuples. 
	 * 
	 * @par feasible : boolean allowed or forbidden tuples  
	 * @param v1 : the external variable
	 * @param tuples : the list of forbidden tuples.
	 * @param name : the name of the constraint 
	 * @return Constraint the binary extension constraint.
	 */
	@Override
	public Constraint makeConstraintTable(String name, boolean feasible, PVariable v0, PVariable v1, int[][] tuples, boolean transpose) {
		IntExternalVar exVar = makeExternalVar(v1);
		Constraint exC = new ExtensionBinConstraint(name, feasible, this.var, exVar, tuples, transpose);
		this.addConstraint(exC);
		return exC;
	}
	
	@Override
	public Constraint makeSoftBinRelation(String name, PVariable v0, PVariable v1, Integer[] weights, boolean transpose){
		IntExternalVar exVar = makeExternalVar(v1);
		Constraint exC = new SoftExtensionBinConstraint(name, this.var, exVar, weights, transpose);
		this.addConstraint(exC);
		return exC;
	}
	
	@Override
	public Constraint makeWeightedTuplesRelation(String name, PVariable v0, PVariable v1, int[][] tuples, Integer[] weights, int defaultCost, boolean transpose){
		IntExternalVar exVar = makeExternalVar(v1);
		Constraint exC = new SoftTuplesConstraint(name, this.var, exVar, tuples, weights, defaultCost, transpose);
		this.addConstraint(exC);
		return exC;
	}

	@Override
	public VarIdentifier getVarIdentifier() {
		return this.var.getVarIdentifier();
	}

	public IntVariable getMyVar() {
		return this.var;
	}
}
