/*
DisChoco: Distributed Constraint Reasoning Solver 
Copyright (C) 2007-2013 Redouane Ezzahir & Mohamed Wahbi 

DisChoco is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DisChoco is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

To contact the authors: <http://dischoco.sourceforge.net/>
*/
package simple.dcop.adopt;

import static simple.dcop.DCOPConstraintHandler.sumCosts;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import kernel.AbstractAgent;
import kernel.Constraint;
import kernel.StopException;
import kernel.VarIdentifier;
import kernel.com.Message;
import kernel.com.MessageHandler;
import kernel.events.VariableEvent;
import kernel.order.TreeDisAO;
import kernel.vars.integer.IntExternalVar;
import simple.SimpleAgent;
import simple.dcop.DCOPConstraintHandler;
import simple.dcsp.afcfamily.TermMsgHandler;
import simulation.parser.PVariable;
import util.Commands;
import util.HashList;
import util.Instantiation;
import util.Logger;
import util.VarAssignment;
import util.VarEventMask;

/**
 * This class implements the Adopt algorithm as described in: 
 * Saving Messages in ADOPT-based Algorithms [Gutierrez and Meseguer, 2010]
 * 
 * @author Mohamed Wahbi
 */
public class AdoptAgent extends SimpleAgent {
	
//    final static BigInteger BIG_INFINITY = BigInteger.valueOf(Long.MAX_VALUE);
	final static Integer INFINITY = Constraint.INFINITY;
	
	public static final String COST = Commands.COST;
	public static final String VALUE = Commands.VALUE;
	public static final String THRESHOLD = Commands.THRESHOLD;
	public static final String CONTEXT = Commands.CONTEXT;
	public static final String INFO = Commands.INFO;

	/** Bounded-error approximation */
	private static Integer BOUNDED_ERROR = 0;
	/** upper bound */
	private Integer UB = INFINITY;
	/** lower bound **/
	private Integer LB = 0;
	/** threshold **/
	private Integer TH = 0;
	
	/**
	 * A map of costs, there is a {@link Cost} for each value in the domain of current variable 
	 * and for every child : map[value, map[child, cost]] .
	 **/
	protected Map<Integer, HashMap<VarIdentifier, Cost>> childrenCosts = null;
	
	/** The handler of VALUE messages */
	protected ValueMsgHandler valueMsgHandler = null;

	/** The handler of COST messages */
	protected CostMsgHandler costMsgHandler = null;

	/** A flag that indicates if parent has terminated */
	protected boolean parentHasTerminated = false;

	/** A map to store the computed cost from constraint functions*/
	protected Map<Integer, Integer> localCost = null;

	private Map<Integer, Integer> initialCosts;

	/**
	 * @param agentId
	 * @param agName
	 * @param description
	 */
	public AdoptAgent(int agentId, String agName, String description) {
		super(agentId, agName, description);
		this.terminateMsgHandler = new TerminateMsgHandler(this);
		this.valueMsgHandler 	 = new ValueMsgHandler(this);
		this.costMsgHandler 	 = new CostMsgHandler(this);
		this.constraintHandler 	 = new DCOPConstraintHandler(this);
		this.childrenCosts 		 = new HashMap<Integer, HashMap<VarIdentifier, Cost>>();
		this.localCost 			 = new HashMap<Integer, Integer>();
		this.initialCosts 		 = new HashMap<Integer, Integer>();
		this.agentOrder 		 = new TreeDisAO(this);
	}
	
	/**
	 * @param agentId
	 * @param agName
	 * @param description
	 */
	public AdoptAgent(PVariable variable) {
		super(variable);
		this.terminateMsgHandler = new TerminateMsgHandler(this);
		this.valueMsgHandler 	 = new ValueMsgHandler(this);
		this.costMsgHandler 	 = new CostMsgHandler(this);
		this.constraintHandler 	 = new DCOPConstraintHandler(this);
		this.childrenCosts 		 = new HashMap<Integer, HashMap<VarIdentifier, Cost>>();
		this.localCost 			 = new HashMap<Integer, Integer>();
		this.initialCosts 		 = new HashMap<Integer, Integer>();
		this.agentOrder 		 = new TreeDisAO(this);
	}

	@Override
	public void configurePropagationEngine() {
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_HIGHER_THAN_LOCAL_VAR).add(VariableEvent.INSTANCIATE);
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_HIGHER_THAN_LOCAL_VAR).add(VariableEvent.DESINSTANCIATE);
		this.eventsConfig.get(VarEventMask.LOCAL_VAR_IS_THE_LAST).add(VariableEvent.RESTORE_VAL);
		this.eventsConfig.get(VarEventMask.LOCAL_VAR_IS_BETWEEN_L_H).add(VariableEvent.RESTORE_VAL);
		this.constraintHandler.update();
	}

	/*
	 * (non-Javadoc)
	 * @see kernel.AbstractAgent#initAgentView()
	 */
	@Override
	public void initAgentView() { 
		HashList<VarIdentifier> varIds = this.getAgentOrder().getAncestors();
		for(VarIdentifier varId : varIds) {
			VarAssignment var = new VarAssignment(varId, Integer.MIN_VALUE);
			IntExternalVar extVar = this.makeExternalVar(var);
			this.agentView.add(extVar);
		}
	}

	@Override
	protected void addMsgHandlers() {
		this.communicator.addMsgHandler(this.valueMsgHandler);
		this.communicator.addMsgHandler(this.costMsgHandler);
	}
	
	@Override
	public String toString() {
		return super.toString()+" solution= "+this.solutionToString();
	}

	protected TreeDisAO getAgentOrder(){
		return ((TreeDisAO) this.agentOrder);
	}
	
	protected VarIdentifier getParentId() {
		return getAgentOrder().getParentId();
	}

	/**
	 * @return if the current agent is the root of the DFS tree. 
	 */
	protected boolean amIRoot() {
		return getParentId()==null;
	}

	/**
	 * This method returns <i>true</i> if the given agent (<i>agId</i>) is the parent of the 
	 * current agent. Otherwise, it returns <i>false</i>. 
	 * 
	 * @param agId the given agent identifier
	 * @return if the given agent is the parent of the current agent on the DFS tree. 
	 */
	protected boolean isMyParent(VarIdentifier varId) {
		if(amIRoot()) {
			return false;
		}
		return getParentId().equals(varId);
	}

	/**
	 * @return if the current agent is the leaf of the DFS tree. 
	 */
	protected boolean isLeaf() {
		return getChildrenIds().isEmpty();
	}

	/**
	 * 
	 * @return
	 */
	private HashList<VarIdentifier> getChildrenIds() {
		return getAgentOrder().getChildrenIds();
	}

	/**
	 * 
	 * @return
	 */
	private boolean isMyParentAssigned() {
		return this.agentView.getVarAssignment(new VarIdentifier(getParentId())).isInstantiated();
	}
	
	/**
	 * This method initializes the costs of each value value in the current domain for every child
	 */
	public void initializesCost() {
		for(int value : this.var.getDomain()) {
			HashMap<VarIdentifier, Cost> costs = new HashMap<VarIdentifier, Cost>();
			for(VarIdentifier child : this.getChildrenIds()) {
				costs.put(child, new Cost());
			}
			this.childrenCosts.put(value, costs);
		}
	}

	/**
	 * This method from {@link kernel.AbstractAgent} starts the distributed search
	 * 
	 * @throws StopExeption
	 */
	@Override
	public void startDistributedSearch() throws StopException {
		this.initializesCost();
		this.computeLocalCost();
		this.initialCosts.putAll(this.localCost);
		int myValue = this.var.getInf();
		this.assignValue(myValue);
		int minimalCost = Collections.min(new ArrayList<Integer>(this.initialCosts.values()));
		if (amIRoot()) {
			this.parentHasTerminated = true;
			/* set initial threshold to given error bounded */
			this.setThreshold(minimalCost + BOUNDED_ERROR);
//			this.threshold = BOUNDED_ERROR;
			//maintainAllocationInvariant();
		} else {
			this.setThreshold(minimalCost);
		}
		this.sendValueMessages();
	}
	
	/**
	 * @param minimalCost
	 */
	protected void setThreshold(int threshold) {
		this.TH = threshold;
	}

	public void backtrack() throws StopException {
		this.computeLocalCost();
		this.log("\n## backtrack call : "+getBoundString()+", myLocalCost: "+localCost, Logger.DEFAULT_LEVEL);
		int valLB = this.computeLowerBound();
		int valUB = this.computeUpperBound();
		
		if (amIRoot() && this.LB.compareTo(INFINITY)==0) {
			throw new StopException(Commands.FAILURE);
		}
		
		this.log("-# Maintain Threshold Invariant : "+getBoundString(), Logger.FINE_LEVEL);	
		// Maintain Threshold Invariant
		this.maintainThresholdInvariant();
		this.log("+# Maintain Threshold Invariant : "+getBoundString(), Logger.FINE_LEVEL);	

		int myValue = this.var.getValue();
		if (this.UB.compareTo(this.TH)==0) {
			this.assignValue(valUB);
			this.log("assignValue that minimizes the UB, UB=" + UB + " equals TH=" + this.TH + ", new value=" + valUB, Logger.DEFAULT_LEVEL);
		} else if (this.LB(myValue).compareTo(this.TH) > 0) {
			this.assignValue(valLB);
			this.log("assignValue that minimizes the LB, LB(" + myValue + ")="+this.LB(myValue)+" > TH=" + this.TH + ", new value=" + valLB+" and the new LB(" + valLB + ")="+LB, Logger.DEFAULT_LEVEL);
		} else {
			this.log("$$ I like my value ("+myValue+") since its LB("+myValue+")="+this.LB(myValue)+" ,"+getBoundString(), Logger.FINE_LEVEL);
		}
		
//		this.log("## Maintain Allocation Invariant:"+getBoundString(), Logger.FINE_LEVEL);
//		this.log(this.localCost+"\n"+getChildrenCostsString(), Logger.FINE_LEVEL);
		this.maintainAllocationInvariant(); // to send threshold with VALUE message

		this.log("## send Value Messages\n", Logger.FINE_LEVEL);
		this.sendValueMessages();

		if (this.UB.compareTo(this.TH)==0 && this.parentHasTerminated) {
			this.log("## send Terminate Messages\n", Logger.FINE_LEVEL);
			((TerminateMsgHandler) this.terminateMsgHandler).sendTerminateMessages();
		}
		
		// send cost if I'm not the root and my parent not terminated
		if (!parentHasTerminated && isMyParentAssigned()) {
			this.sendCostMessage();
		}
		
	}

	/**
	 * @param value
	 */
	protected void assignValue(int value) {
		this.var.instantiate(value); 
	}

	/**
	 * 
	 */
	protected void sendValueMessages() {
		this.valueMsgHandler.sendValueMessages();
	}

	/**
	 * 
	 */
	protected void sendCostMessage() {
		Cost cost = getMyCost();
		this.costMsgHandler.sendCost(cost);
	}

	/**
	 * 
	 * @param context
	 */
	protected void updateAgentView(Instantiation context) {
		if(agentView.priorityMerge(context)) {
			clearIncompatibleCosts(); 
			this.log("I updated my AgentView with: "+context, Logger.FINE_LEVEL);
		}
	}

	/**
	 * 
	 * @param context
	 * @param varAss
	 * @param threshold
	 */
	protected void updateAgentView(Instantiation context, VarAssignment varAss, Integer threshold) {
		this.updateAgentView(context);
		if(agentView.priorityMerge(varAss)) {
			this.log("I updated my AgentView with: "+varAss, Logger.FINE_LEVEL);
			clearIncompatibleCosts();
//			if(isMyParent(varAss.getVarIdentifier())){
//				this.setThreshold(threshold);
//				this.log("I updated my threshold to the received one: "+threshold, Logger.FINE_LEVEL);
//			}
		}
		context.add(varAss);
		if(agentView.compatibleWith(context) && isMyParent(varAss.getVarIdentifier())) {
			this.setThreshold(threshold);
			this.log("==>I updated my threshold to the received one: "+threshold, Logger.FINE_LEVEL);
		}
	}
	
	public void computeLocalCost(){
		for(int value : this.var.getDomain()){
			Integer cost = ((DCOPConstraintHandler) this.constraintHandler).localCost(value);
			this.localCost.put(value, cost);
//			if(cost.equals(INFINITY)) {
//				this.var.removeValue(value, VariableEvent.NOCAUSE);
//			}
		}
//		for(int value : this.var.getDomain().removedValsIterator()){
//			this.localCost.put(value, INFINITY);
//		}
	}

	/**
	 * This method computes the Lower Bound <b><i>LB = min{LB(v), forall v in D_i}</i></b>, 
	 * and returns the value that minimizes LB(v). 
	 * 
	 * @return Integer : the value <i>v</i> from the domain with the minimum lower bound LB(v)
	 */
    public int computeLowerBound() {
    	int myValue = this.var.getValue();
    	Integer lb = LB(myValue);
        for(int value : this.var.getDomain()){
            Integer candidate_lb = LB(value);
            if (lb.compareTo(candidate_lb) > 0) {
            	lb = candidate_lb;
				myValue = value;
            }
        }
		this.log("ComputeLB: [LB="+LB+", TH="+this.TH+", context: "+this.agentView+
				"] --> value minimizing the LB is: "+myValue+" and the new LB is "+lb+"", Logger.FINE_LEVEL);
		this.LB = lb; // update Lower Bound 
        return myValue;
    }
	
	/**
	 * This method computes the Upper Bound <b><i>UB = min{UB(v), forall v in D_i}</i></b>, 
	 * and returns the value that minimizes UB(v). 
	 * 
	 * @return Integer : the value <i>v</i> from the domain with the minimum upper bound UB(v)
	 */
    public int computeUpperBound() {
    	int myValue = this.var.getValue();
    	Integer ub = UB(myValue);
    	for(int value : this.var.getDomain()){
            Integer candidate_ub = UB(value);
            if (ub.compareTo(candidate_ub) > 0) {
            	ub = candidate_ub;
				myValue = value;
            }
        }
		this.log("ComputeUB: [UB="+UB+", TH="+this.TH+", context: "+this.agentView+
				"] --> value minimizing the UB is: "+myValue+" and the new UB is "+ub+"\n", Logger.FINE_LEVEL);
		this.UB = ub; // update Upper Bound ;
        return myValue;
    }

	/**
	 * This methods returns the Cost of the given child associated to the given value
	 * @param child the identifier of the child
	 * @param value 
	 * @return  Cost
	 */
	public Cost getCost(VarIdentifier child, int value) {
		if (this.childrenCosts.containsKey(value)) return this.childrenCosts.get(value).get(child);
		return null;
	}

	/**
	 * 
	 * @param value
	 * @return
	 */
	private Collection<Cost> getCost(int value) {
		return this.childrenCosts.get(value).values();
	}

	/**
	 * This method checks the validity of costs received from the children with the current context. 
	 * Thus, it clears every child cost with context incompatible with the current context
	 * if contexts are incompatible, zero out costs.
	 * 
	 */
	public void clearIncompatibleCosts() {
		for(HashMap<VarIdentifier, Cost> costs : this.childrenCosts.values()){
			for(Cost cost : costs.values()){
				if (!this.agentView.compatibleWith(cost.getContext())) {
					cost.clear();
				}
			}
		}
	}

	/**
	 * This method maintains the threshold invariant. Thus, LB <= threshold <= UB
	 * 
	 */
	public void maintainThresholdInvariant() {
		/* if we are the root, then perform the bounded-error approximation */
		if(amIRoot()) {
			this.LB += BOUNDED_ERROR;
		}
		if (this.TH.compareTo(this.LB) < 0) {
			this.log("maintainThresholdInvariant: threshold was " + this.TH + " and is now set to LB=" + LB, Logger.FINE_LEVEL);
			this.setThreshold(this.LB);
		}
		if (this.TH.compareTo(this.UB) > 0) {
			this.log("maintainThresholdInvariant: threshold was " + this.TH + " and is now set to UB=" + UB, Logger.FINE_LEVEL);
			this.setThreshold(this.UB);
		}
	}
	

	/**
	 * maintains: for each value <b><i>v</i></b> of this agent, the summation of the local cost 
	 * of <b><i>v</i></b> with all children thresholds of <b><i>v</i></b> must equals the local 
	 * threshold. <br>
	 * 
	 * <b><i>threshold = localCost(v) + sum{for each child in children, threshold(v, child)}</i></b>.
	 * <br>
	 * 
	 * In other words, that divide threshold among self and sons.
	 */
	protected void maintainAllocationInvariant() {
		int myValue = this.var.getValue();
		HashMap<VarIdentifier, Cost> currentCosts = childrenCosts.get(myValue);
		Integer myLocalCost = this.localCost.get(myValue);
		this.log("\n-## Maintain Allocation Invariant: "+getBoundString()+"\nmyLocalCosts="+this.localCost
				+", myValue="+myValue+", myLocalCost="+myLocalCost+"\n"+getChildrenCostsString(myValue), 
				Logger.FINE_LEVEL);
		while (this.TH.compareTo(sumCosts(myLocalCost, sumThresholds(myValue))) > 0 ) {
//			for(int child : getChildrenIds()) {
//				Cost cost = costs.get(child);
//				Integer ub = cost.getUb();
//				Integer th = cost.getThreshold();
//				if(ub.compareTo(th) > 0) {
//					cost.setThreshold(th + 1);
//					break;
//				}
//			}
			for(Map.Entry<VarIdentifier, Cost> entry : currentCosts.entrySet()) {
				Cost cost = entry.getValue();
				Integer ub = cost.getUb();
				Integer th = cost.getThreshold();
				if(ub.compareTo(th) > 0) {
					Integer amount = this.TH - sumCosts(myLocalCost, sumThresholds(myValue));
					th = Math.min(th + amount, ub);
					cost.setThreshold(th);
					this.log("I set the threshold of "+entry.getKey()+" to "+th+" --> "+cost, Logger.FINE_LEVEL);
					//cost.setThreshold(th + 1);
					break;
				}
			}
		}
		//maintainChildThresholdInvariant(); //TODO check it is necessarily 
		while (this.TH.compareTo(sumCosts(myLocalCost, sumThresholds(myValue))) < 0 ) {
//			for(int child : getChildrenIds()) {
//				Cost cost = currentCosts.get(child);
//				Integer lb = cost.getLb();
//				Integer th = cost.getThreshold();
//				if (lb.compareTo(th) < 0) {
//					cost.setThreshold(th - 1);
//					break;
//				}
//			}
			for(Map.Entry<VarIdentifier, Cost> entry : currentCosts.entrySet()) {
				Cost cost = entry.getValue();
				Integer lb = cost.getLb();
				Integer th = cost.getThreshold();
				if(lb.compareTo(th) < 0) {
					Integer amount = sumCosts(myLocalCost, sumThresholds(myValue)) - this.TH;
					th = Math.max(th - amount, lb);
					cost.setThreshold(th);
					this.log("I set the threshold of "+entry.getKey()+" to "+th+" --> "+cost, Logger.FINE_LEVEL);
//					cost.setThreshold(th - 1);
					break;
				}
			}
		}
		this.log("+## Maintain Allocation Invariant: "+getBoundString()+"\n"+getChildrenCostsString(myValue), Logger.FINE_LEVEL);
	}

	private Integer sumThresholds(int value) {
		Integer sum = new Integer(0);
		for(Cost cost : this.childrenCosts.get(value).values()){
			sum = sumCosts(sum, cost.getThreshold());
			if(sum.compareTo(INFINITY) >= 0) return INFINITY;
		}
		this.log("The sum of Thresholds from children for value="+value+ " is "+sum, Logger.FINE_LEVEL);
		return sum;
	}
	
	/**
	 * This method maintains for each value <i>v</i>, and child <i>ch</i>: 
	 * lb[v, ch] <= th[v, ch] <= ub[v, ch]
	 * 
	 */
	protected void maintainChildThresholdInvariant(int logValue) {
		this.log("-** maintain Child Threshold Invariant: \n"+getChildrenCostsString(logValue), Logger.FINE_LEVEL);
		for(int value : var.getDomain()) {
			HashMap<VarIdentifier, Cost> currentCosts = childrenCosts.get(value);
			for(Cost cost : currentCosts.values()) {
				Integer lb = cost.getLb();
				Integer th = cost.getThreshold();
				if(lb.compareTo(th) > 0) {
					cost.setThreshold(lb);
				}
			}
		}
		for(int value : var.getDomain()) {
			HashMap<VarIdentifier, Cost> currentCosts = childrenCosts.get(value);
			for(Cost cost : currentCosts.values()) {
				Integer ub = cost.getUb();
				Integer th = cost.getThreshold();
				if(ub.compareTo(th) < 0) {
					cost.setThreshold(ub);
				}
			}
		}
		this.log("+** maintain Child Threshold Invariant: \n"+getChildrenCostsString(logValue), Logger.FINE_LEVEL);
	}

	/**
	 * This method computes the lower bound <b><i>LB(value)</i></b> of the given value <b><i>value</i></b>. 
	 * The lower bound equals the sum of the local cost of the given value (cost with assignments 
	 * of agents in the context) with the lower bound stored for all children. 
	 * 
	 * @param value : the value to which we want to compute the Lower Bound
	 * @return Integer: LB(value) = localCost(value) + \sum{child in children} lb(value, child);
	 */
	public Integer LB(int value) {
		BigInteger sum = BigInteger.valueOf(this.localCost.get(value));
		assert sum.compareTo(BigInteger.valueOf(0)) >= 0 : "\nProblem the local cost="+sum+" of value="+value+" is < 0";
		// make sure we don't overflow when dealing with hard constraints 
        if(sum.compareTo(BigInteger.valueOf(INFINITY)) >= 0) return INFINITY;
		for(Cost cost : this.getCost(value)) {
	        Integer lb = cost.getLb();
			// make sure we don't overflow when dealing with hard constraints 
	        if(lb.compareTo(INFINITY) >= 0) return INFINITY;
	        sum = sum.add(BigInteger.valueOf(lb));
			// make sure we don't overflow when dealing with hard constraints 
	        if(sum.compareTo(BigInteger.valueOf(INFINITY)) >= 0) return INFINITY;
		}
		this.log("LB("+value+")="+sum.intValue(), Logger.FINE_LEVEL);
		return sum.intValue();
	}

	/**
	 * This method computes the upper bound <b><i>UB(value)</i></b> of the given value <b><i>value</i></b>. 
	 * The upper bound equals the sum of the local cost of the given value (cost with assignments 
	 * of agents in the context) with the upper bound stored for all children. 
	 * 
	 * @param value : the value to which we want to compute the Upper Bound
	 * @return Integer: UB(value) = localCost(value) + \sum{child in children} ub(value, child);
	 */
	public Integer UB(int value) {
		BigInteger sum = BigInteger.valueOf(this.localCost.get(value));
		assert sum.compareTo(BigInteger.valueOf(0)) >= 0 : "\nProblem the local cost="+sum+" of value="+value+" is < 0";
		// make sure we don't overflow when dealing with hard constraints 
        if(sum.compareTo(BigInteger.valueOf(INFINITY)) >= 0) return INFINITY;
		for(Cost cost : this.getCost(value)) {
	        Integer ub = cost.getUb();
			// make sure we don't overflow when dealing with hard constraints 
	        if(ub.compareTo(INFINITY) >= 0) return INFINITY;
	        sum = sum.add(BigInteger.valueOf(ub));
			// make sure we don't overflow when dealing with hard constraints 
	        if(sum.compareTo(BigInteger.valueOf(INFINITY)) >= 0) return INFINITY;
		}
		this.log("UB("+value+")="+sum.intValue(), Logger.FINE_LEVEL);
		return sum.intValue();
	}

	/**
	 * @return
	 */
	protected Cost getMyCost() {
		int myValue = this.var.getValue();
		Instantiation context = agentView.getInstantiationOfNeighbors();
		for(Cost cost : childrenCosts.get(myValue).values()){
			context.priorityMerge(cost.getContext());
		}
		return new Cost(context, LB, UB);
	}
	
	public class ValueMsgHandler extends MessageHandler { 
		public ValueMsgHandler(AbstractAgent agent) {
			super(agent, VALUE);
		}

		public void processMessage(Message msg) throws StopException {
			if(!parentHasTerminated){
				VarAssignment varAss = VarAssignment.fromString((String) msg.get(INFO));
				Instantiation context = Instantiation.fromString((String) msg.get(CONTEXT), agentOrder);
				Integer threshold = (Integer) msg.get(THRESHOLD);
				updateAgentView(context, varAss, threshold);
				if(communicator.isMailBoxEmpty()){
					backtrack();
				}
				/*
				boolean thresholdUpdated = false, agentViewUpdated = false;
				//agentView.priorityMerge(context);
				for(VarAssignment assignment : context) {
					if(agentView.priorityMerge(assignment)) {
						agentViewUpdated = true;
					}
				}
				if(isMyParent(varAss.getVarIdentifier()) && agentView.compatibleWith(context)){
					threshold = (Integer) msg.get(THRESHOLD);
					thresholdUpdated = true;
					//maintainThresholdInvariant();
				}
				if(agentView.priorityMerge(varAss)) {
					agentViewUpdated = true;
				}
				if(agentViewUpdated ){
					clearIncompatibleCosts();
					backtrack();
				} else if(thresholdUpdated) {
					backtrack();
				}
				*/
			}
		}
		
		/** This method sends a VALUE message to each lower neighbor */
		public void sendValueMessages() {
			VarAssignment varAss = var.getVarAssignment();
			HashMap<VarIdentifier, Cost> costs = childrenCosts.get(varAss.getValue());
			HashList<VarIdentifier> lowerNeighbors = communicator.getLowerNeighbors();
			HashList<VarIdentifier> children = getChildrenIds();
			Instantiation context = agentView.toInstantiation();
			for (VarIdentifier child : getChildrenIds()) {
				sendValue(child, varAss, costs.get(child).getThreshold(), context);
			}
			HashList<VarIdentifier> linkedDescendants = new HashList<VarIdentifier>(lowerNeighbors);
			linkedDescendants.removeAll(children);
			for (VarIdentifier neighbor : linkedDescendants) {
				sendValue(neighbor, varAss, 0, context);
			}
		}
		
		public void sendValue(VarIdentifier neighbor, VarAssignment varAss, Integer threshold, Instantiation context) {
			Message msg = createMessage();
			msg.addContents(INFO, varAss.toString());
			msg.addContents(THRESHOLD, threshold); 
			msg.addContents(CONTEXT, context.toString()); 
			msg.setReceiverId(neighbor);
			sendMessage(msg);
		}
	}
	
	public class CostMsgHandler extends MessageHandler { 
		
		public CostMsgHandler(AdoptAgent agent) {
			super(agent, COST);
		}

		@Override
		public void processMessage(Message msg) throws StopException {
//			if (IwasTerminated) return;
			Cost cost = (Cost) msg.get(COST);
			Instantiation context = cost.getContext();
			VarAssignment myAss = context.remove(var.getVarIdentifier());
			if(!parentHasTerminated){
				updateAgentView(context);
			}
			if(agentView.compatibleWith(context)){
				VarIdentifier child = msg.getSenderId();
				Integer lb = cost.getLb();
				Integer ub = cost.getUb();
				updateCost(myAss.getValue(), child, lb, ub, context);
				maintainChildThresholdInvariant(myAss.getValue());
			}
			if(communicator.isMailBoxEmpty()){
				backtrack();
			}
		}
		
		public void sendCost(Cost cost) {
			Message msg = createMessage();
			msg.addContents(COST, cost);
			msg.setReceiverId(getParentId());
			sendMessage(msg);
		}
	}
	
	public class TerminateMsgHandler extends TermMsgHandler {
		int myChildrenTerminated = 0;
		public TerminateMsgHandler(AbstractAgent agent) {
			super(agent);
		}

		@Override
		public boolean processTerminationDetection(Message msg) throws StopException {
			VarIdentifier senderId = msg.getSenderId();
			if(isMyParent(senderId)){
				parentHasTerminated = true;
				Instantiation context = (Instantiation) msg.get(Commands.CONTEXT);
				updateAgentView(context);
				backtrack();
			} else {
				this.myChildrenTerminated++;
				if(this.myChildrenTerminated==getChildrenIds().size()) {
					if(amIRoot()){
						throw new StopException(Commands.SUCCESS);
					} else {
						this.sendTerminate(getParentId());
					}
				}
			}
			return true;
		}
		
		public void sendTerminateMessages() {
			HashList<VarIdentifier> children = getChildrenIds();
			if(children.isEmpty()){
				this.sendTerminate(getParentId());
			} else {
				Instantiation currentContext = this.myAgent.agentView.toInstantiation();
				VarAssignment myAss = var.getVarAssignment();
				currentContext.add(myAss);
				for (VarIdentifier child : children) {
					this.sendTerminate(child, currentContext);
				}
			}
		}

		/**
		 * @param child
		 * @param currentContext
		 */
		private void sendTerminate(VarIdentifier child, Instantiation currentContext) {
			Message msg = this.createMessage();
			msg.addContents(Commands.CONTEXT, currentContext);
			msg.setReceiverId(child);
			sendMessage(msg);
		}

		/**
		 * @param child
		 */
		private void sendTerminate(VarIdentifier parent) {
			Message msg = this.createMessage();
			msg.setReceiverId(parent);
			sendMessage(msg);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public String solutionToString() {
		String solution = this.var.toString();
		if(amIRoot()){
			return "[LB="+this.LB+", UB="+(this.UB.compareTo(INFINITY)==0?"+inf":this.UB)+"] "+solution;
		}
		return solution;
	}

	/**
	 * @param value
	 * @param child
	 * @param newCost
	 */
	public void updateCost(int value, VarIdentifier child, Integer lb, Integer ub, Instantiation context) {
		Cost cost = childrenCosts.get(value).get(child);
		cost.setLb(lb);
		cost.setUb(ub);
		cost.setContext(context);
	}

	/**
	 * @return
	 */
	protected String getBoundString() {
		return "[LB="+LB+", UB="+UB+", TH="+this.TH+"]";
	}
	
	protected String getChildrenCostsString() {
		if(isLeaf()) {
			return "";
		}
		StringBuffer sb = new StringBuffer();
		for(Map.Entry<Integer, HashMap<VarIdentifier, Cost>> entry : this.childrenCosts.entrySet()) {
			sb.append(entry.toString()+"\n");
		}
		return sb.toString();
	}
	
	protected String getChildrenCostsString(int value) {
		if(isLeaf()) {
			return "";
		}
		StringBuffer sb = new StringBuffer(value+"=");
		HashMap<VarIdentifier, Cost> map = this.childrenCosts.get(value);
		sb.append(map.toString()+"\n");
		return sb.toString();
	}
}
