package simple.dcop.adopt;

import java.math.BigInteger;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import kernel.AbstractAgent;
import kernel.Constraint;
import kernel.StopException;
import kernel.VarIdentifier;
import kernel.com.Message;
import kernel.com.MessageHandler;
import kernel.events.VariableEvent;
import kernel.order.TreeDisAO;
import kernel.vars.integer.IntExternalVar;
import simple.SimpleAgent;
import simple.dcop.DCOPConstraintHandler;
import simple.dcsp.afcfamily.TermMsgHandler;
import simulation.parser.PVariable;
import util.Commands;
import util.HashList;
import util.Instantiation;
import util.Logger;
import util.VarAssignment;
import util.VarEventMask;

/**
 * This class implements the original version of Adopt algorithm for a single variable as described in: 
 * "An Asynchronous Complete Method for Distributed Constraint Optimization" [Modi, P.J. et al. 2005]
 * 
 * @version dischoco 2.5.3<br/>
 * simple.dcop.adopt.Adopt_ModiAgent.java
 * @author Mohamed Wahbi
 */
public class Adopt_ModiAgent extends SimpleAgent {

	private final static Integer INFINITY = Constraint.INFINITY;
	public static final String COST = Commands.COST;
	public static final String VALUE = Commands.VALUE;
	public static final String THRESHOLD = Commands.THRESHOLD;
	public static final String CONTEXT = Commands.CONTEXT;
	public static final String INFO = Commands.INFO;

	/** Bounded-error approximation */
	private static Integer BOUNDED_ERROR = new Integer(0);
	/** upper bound */
	protected Integer UB = INFINITY;
	/** lower bound **/
	protected Integer LB = new Integer(0);
	
	/** threshold **/
	protected Integer threshold = new Integer(0);
	
	/**
	 * A map of costs, there is a {@link Cost} for each value in the domain of current variable 
	 * and for every child : map[child, map[value, cost]] .
	 **/	//TODO [value, [child, COST]]  // childrenCosts
	protected HashMap<Integer, HashMap<VarIdentifier, Cost>> childrenCosts = null;
	
	/** Adopt Value Message Handler */
	protected ValueMsgHandler valueMsgHandler = null;

	/** Adopt Cost Message Handler */
	protected CostMsgHandler costMsgHandler = null;

	/** Adopt Threshold Message Handler */
	protected ThresholdMsgHandler thresholdMsgHandler = null;

	/** A flag that indicates if parent has terminated */
	protected boolean parentHasTerminated = false;
	
	/** a flag that indicates if this agent terminated */
	protected boolean ITerminate = false;

	protected Map<Integer, Integer> localCost = null;

	/**
	 * @param agentId
	 * @param agName
	 * @param description
	 */
	public Adopt_ModiAgent(int agentId, String agName, String description) {
		super(agentId, agName, description);
		this.valueMsgHandler 	 = new ValueMsgHandler(this);
		this.costMsgHandler 	 = new CostMsgHandler(this);
		this.thresholdMsgHandler = new ThresholdMsgHandler(this);
		this.constraintHandler 	 = new DCOPConstraintHandler(this);
		this.childrenCosts 		 = new HashMap<Integer, HashMap<VarIdentifier, Cost>>();
		this.localCost 			 = new HashMap<Integer, Integer>();
		this.terminateMsgHandler = new TerminateMsgHandler(this);
		this.agentOrder 		 = new TreeDisAO(this);
	}
	
	/**
	 * @param agentId
	 * @param agName
	 * @param description
	 */
	public Adopt_ModiAgent(PVariable variable) {
		super(variable);
		this.valueMsgHandler 	 = new ValueMsgHandler(this);
		this.costMsgHandler 	 = new CostMsgHandler(this);
		this.thresholdMsgHandler = new ThresholdMsgHandler(this);
		this.constraintHandler 	 = new DCOPConstraintHandler(this);
		this.childrenCosts 		 = new HashMap<Integer, HashMap<VarIdentifier, Cost>>();
		this.localCost 			 = new HashMap<Integer, Integer>();
		this.terminateMsgHandler = new TerminateMsgHandler(this);
		this.agentOrder 		 = new TreeDisAO(this);
	}

	/*
	 * (non-Javadoc)
	 * @see kernel.AbstractAgent#configurePropagationEngine()
	 */
	@Override
	public void configurePropagationEngine() {
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_HIGHER_THAN_LOCAL_VAR).add(VariableEvent.INSTANCIATE);
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_HIGHER_THAN_LOCAL_VAR).add(VariableEvent.DESINSTANCIATE);
		this.eventsConfig.get(VarEventMask.LOCAL_VAR_IS_THE_LAST).add(VariableEvent.RESTORE_VAL);
		this.eventsConfig.get(VarEventMask.LOCAL_VAR_IS_BETWEEN_L_H).add(VariableEvent.RESTORE_VAL);
		this.constraintHandler.update();	//TODO
	}

	/*
	 * (non-Javadoc)
	 * @see kernel.AbstractAgent#initAgentView()
	 */
	@Override
	public void initAgentView() {//TODO higher neighbors instead of ancestors 
		HashList<VarIdentifier> varIds = this.getAgentOrder().getAncestors();
		for(VarIdentifier varId : varIds) {
			VarAssignment var = new VarAssignment(varId, Integer.MIN_VALUE);
			IntExternalVar extVar = this.makeExternalVar(var);
			this.agentView.add(extVar);
		}
	}

	@Override
	public void addMsgHandlers() {
		this.communicator.addMsgHandler(this.valueMsgHandler);
		this.communicator.addMsgHandler(this.costMsgHandler);
		this.communicator.addMsgHandler(this.thresholdMsgHandler);
	}
	
	@Override
	public String toString() {
		return super.toString()+" sol= "+this.solutionToString();
	}

	protected TreeDisAO getAgentOrder(){
		return ((TreeDisAO) this.agentOrder);
	}
	
	protected VarIdentifier getParentId() {
		return getAgentOrder().getParentId();
	}

	/**
	 * @return
	 */
	protected boolean amIRoot() {
		return getParentId()==null;
	}

	/**
	 * @return
	 */
	protected boolean isMyParent(VarIdentifier varId) {
		if(amIRoot()) {
			return false;
		}
		return getParentId().equals(varId);
	}

	/**
	 * @return
	 */
	protected boolean isLeaf() {
		return getChildrenIds().isEmpty();
	}

	private HashList<VarIdentifier> getChildrenIds() {
		return getAgentOrder().getChildrenIds();
	}
	
	/**
	 * This method initializes the costs of each value value in the current domain for every child
	 */
	public void initializesCost() {
		for(int value : this.var.getDomain()) {
			HashMap<VarIdentifier, Cost> costs = new HashMap<VarIdentifier, Cost>();
			for(VarIdentifier child : this.getChildrenIds()) {
				costs.put(child, new Cost());
			}
			this.childrenCosts.put(value, costs);
		}
	}

	/**
	 * This method from {@link kernel.AbstractAgent} starts the distributed search
	 * 
	 * @throws StopExeption
	 */
	@Override
	public void startDistributedSearch() throws StopException {
		this.initializesCost();
		this.computeLocalCost();
		int myValue = this.var.getInf();
		this.assignValue(myValue);
		if (amIRoot()) {
			this.parentHasTerminated = true;
			/* set initial threshold to given error bounded */
			this.threshold = BOUNDED_ERROR;
			maintainAllocationInvariant();
		} else {
			this.threshold = new Integer(0);
		}
		this.sendValueMessages();
	}

	/**
	 * 
	 * @param value
	 */
	protected void assignValue(int value) {
		this.var.instantiate(value);
	}
	
	/**
	 * 
	 */
	protected void sendValueMessages() {
		this.valueMsgHandler.sendValueMessages(); 
		this.thresholdMsgHandler.sendThresholds();
	}
	
	public void backtrack() throws StopException {
		this.computeLocalCost();
		this.log("## backtrack call : [LB:"+LB+", UB:"+UB+", TH:"+threshold+"], myLocalCost: "+localCost, Logger.DEFAULT_LEVEL);
		
		int valLB = this.computeLowerBound();
		int valUB = this.computeUpperBound();
		// Maintain Threshold Invariant 
		this.maintainThresholdInvariant();

		if (this.UB.equals(this.threshold)) {
			this.assignValue(valUB); 
		} else if (this.LB(this.var.getValue()).compareTo(this.threshold) > 0) {
			this.assignValue(valLB); 
		}
		
		this.maintainAllocationInvariant(); // to send threshold with VALUE message
		this.sendValueMessages();

		if (this.UB.equals(this.threshold) && this.parentHasTerminated) {
			((TerminateMsgHandler) this.terminateMsgHandler).sendTerminateMessages();
		}
		// send cost if I'm not the root and my parent not terminated
		if(!parentHasTerminated && isMyParentAssigned() ){
			this.sendCostMessage();
		}
	}
	
	/**
	 * 
	 */
	protected void sendCostMessage() {
		this.costMsgHandler.sendCost();
	}

	/**
	 * 
	 * @param context
	 */
	protected void updateAgentView(Instantiation context) {
		if(agentView.priorityMerge(context)) {
			clearIncompatibleCosts(); 
		}
	}

	/**
	 * @param value
	 * @param child
	 * @param newCost
	 */
	public void updateCost(int value, VarIdentifier child, Integer lb, Integer ub, Instantiation context) {
		Cost cost = childrenCosts.get(value).get(child);
		cost.setLb(lb);
		cost.setUb(ub);
		cost.setContext(context);
	}
	
	/**
	 * @return
	 */
	private boolean isMyParentAssigned() {
		return this.agentView.getVarAssignment(new VarIdentifier(getParentId())).isInstantiated();
	}

	public void computeLocalCost(){
		for(int value : this.var.getDomain()){
			Integer cost = ((DCOPConstraintHandler) this.constraintHandler).localCost(value);
			this.localCost.put(value, cost);
		}
	}

	/**
	 * This method computes the Lower Bound <b><i>LB = min{LB(v), forall v in D_i}</i></b>, 
	 * and returns the value that minimizes LB(v). 
	 * 
	 * @return Integer : the value <i>v</i> from the domain with the minimum lower bound LB(v)
	 */
    public int computeLowerBound() {
    	int myValue = this.var.getValue();
    	Integer lb = LB(myValue);
        for(int value : this.var.getDomain()){
            Integer candidate_lb = LB(value);
            if (lb.compareTo(candidate_lb) > 0) {
            	lb = candidate_lb;
				myValue = value;
            }
        }
		this.log("ComputeLB: [LB="+LB+", TH="+threshold+", context: "+this.agentView+
				"] --> value minimizing the LB is: "+myValue+" and the new LB is "+lb+"\n", Logger.FINE_LEVEL);
		this.LB = lb; // update Lower Bound 
        return myValue;
    }
	
	/**
	 * This method computes the Upper Bound <b><i>UB = min{UB(v), forall v in D_i}</i></b>, 
	 * and returns the value that minimizes UB(v). 
	 * 
	 * @return Integer : the value <i>v</i> from the domain with the minimum upper bound UB(v)
	 */
    public int computeUpperBound() {
    	int myValue = this.var.getValue();
    	Integer ub = UB(myValue);
    	for(int value : this.var.getDomain()){
            Integer candidate_ub = UB(value);
            if (ub.compareTo(candidate_ub) > 0) {
            	ub = candidate_ub;
				myValue = value;
            }
        }
		this.log("ComputeUB: [UB="+UB+", TH="+threshold+", context: "+this.agentView+
				"] --> value minimizing the UB is: "+myValue+" and the new UB is "+ub+"\n", Logger.FINE_LEVEL);
		this.UB = ub; // update Upper Bound ;
        return myValue;
    }

	/**
	 * This methods returns the Cost of the given child associated to the given value
	 * @param child the identifier of the child
	 * @param value 
	 * @return  Cost
	 */
	public Cost getCost(int child, int value) {
		if (this.childrenCosts.containsKey(value)) return this.childrenCosts.get(value).get(child);
		return null;
	}

	/**
	 * 
	 * @param value
	 * @return
	 */
	private Collection<Cost> getCost(int value) {
		return this.childrenCosts.get(value).values();
	}

	/**
	 * This method checks the validity of costs received from the children with the current context. 
	 * Thus, it clears every child cost with context incompatible with the current context
	 * if contexts are incompatible, zero out costs.
	 * 
	 */
	public void clearIncompatibleCosts() {
		for(HashMap<VarIdentifier, Cost> costs : this.childrenCosts.values()){
			for(Cost cost : costs.values()){
				if (!this.agentView.compatibleWith(cost.getContext())) {
					cost.clear();
				}
			}
		}
	}

	/**
	 * This method maintains the threshold invariant. Thus, LB <= threshold <= UB
	 * 
	 */
	public void maintainThresholdInvariant() {
		/* if we are the root, then perform the bounded-error approximation */
		if(amIRoot()) {
			this.LB += BOUNDED_ERROR;
		}
		if (this.threshold.compareTo(this.LB) < 0) {
			this.log("maintainThresholdInvariant: threshold was " + threshold + " and is now " + LB+" (LB)", Logger.FINE_LEVEL);
			this.threshold = this.LB;
		}
		if (this.threshold.compareTo(this.UB) > 0) {
			this.log("maintainThresholdInvariant: threshold was " + threshold + " and is now " + UB+" (UB)", Logger.FINE_LEVEL);
			this.threshold = this.UB;
		}
	}
	

	/**
	 * maintains: for each value <b><i>v</i></b> of this agent, the summation of the local cost 
	 * of <b><i>v</i></b> with all children thresholds of <b><i>v</i></b> must equals the local 
	 * threshold. <br>
	 * 
	 * <b><i>threshold = localCost(v) + sum{for each child in children, threshold(v, child)}</i></b>.
	 * <br>
	 * 
	 * In other words, that divide threshold among self and sons.
	 */
	protected void maintainAllocationInvariant() {
		int myValue = this.var.getValue();
		HashMap<VarIdentifier, Cost> currentCosts = childrenCosts.get(myValue);
		Integer myLocalCost = this.localCost.get(myValue);
		while (this.threshold.compareTo(myLocalCost + sumThresholds(myValue)) > 0 ) {
//			for(int child : getChildrenIds()) {
//				Cost cost = costs.get(child);
//				Integer ub = cost.getUb();
//				Integer th = cost.getThreshold();
//				if(ub.compareTo(th) > 0) {
//					cost.setThreshold(th + 1);
//					break;
//				}
//			}
			for(Cost cost : currentCosts.values()) {
				Integer ub = cost.getUb();
				Integer th = cost.getThreshold();
				if(ub.compareTo(th) > 0) {
					cost.setThreshold(th + 1);
					break;
				}
			}
		}
		//maintainChildThresholdInvariant(); //TODO check it is necessarily 
		while (this.threshold.compareTo(myLocalCost + sumThresholds(myValue)) < 0 ) {
//			for(int child : getChildrenIds()) {
//				Cost cost = currentCosts.get(child);
//				Integer lb = cost.getLb();
//				Integer th = cost.getThreshold();
//				if (lb.compareTo(th) < 0) {
//					cost.setThreshold(th - 1);
//					break;
//				}
//			}
			for(Cost cost : currentCosts.values()) {
				Integer lb = cost.getLb();
				Integer th = cost.getThreshold();
				if(lb.compareTo(th) < 0) {
					cost.setThreshold(th - 1);
					break;
				}
			}
		}
	}

	private Integer sumThresholds(int value) {
		Integer sum = new Integer(0);
		for(Cost cost : this.childrenCosts.get(value).values()){
			Integer th = cost.getThreshold(); 
			if(th.compareTo(INFINITY) >= 0) return INFINITY;
			sum += th;					
		}
		return sum;
	}
	
	/**
	 * This method maintains for each value <i>v</i>, and child <i>ch</i>: 
	 * lb[v, ch] <= th[v, ch] <= ub[v, ch]
	 * 
	 */
	protected void maintainChildThresholdInvariant() {
		for(int value : var.getDomain()) {
			HashMap<VarIdentifier, Cost> currentCosts = childrenCosts.get(value);
			for(Cost cost : currentCosts.values()) {
				Integer lb = cost.getLb();
				Integer th = cost.getThreshold();
				if(lb.compareTo(th) > 0) {
					cost.setThreshold(lb);
				}
			}
		}
		for(int value : var.getDomain()) {
			HashMap<VarIdentifier, Cost> currentCosts = childrenCosts.get(value);
			for(Cost cost : currentCosts.values()) {
				Integer ub = cost.getUb();
				Integer th = cost.getThreshold();
				if(ub.compareTo(th) < 0) {
					cost.setThreshold(ub);
				}
			}
		}
	}

	/**
	 * This method computes the lower bound <b><i>LB(value)</i></b> of the given value <b><i>value</i></b>. 
	 * The lower bound equals the sum of the local cost of the given value (cost with assignments 
	 * of agents in the context) with the lower bound stored for all children. 
	 * 
	 * @param value : the value to which we want to compute the Lower Bound
	 * @return Integer: LB(value) = localCost(value) + \sum{child in children} lb(value, child);
	 */
	public Integer LB(int value) {
		BigInteger sum = BigInteger.valueOf(this.localCost.get(value));
		assert sum.compareTo(BigInteger.valueOf(0)) >= 0 : "\nProblem the local cost="+sum+" of value="+value+" is < 0";
		// make sure we don't overflow when dealing with hard constraints 
        if(sum.compareTo(BigInteger.valueOf(INFINITY)) >= 0) return INFINITY;
		for(Cost cost : this.getCost(value)) {
	        Integer lb = cost.getLb();
			// make sure we don't overflow when dealing with hard constraints 
	        if(lb.compareTo(INFINITY) >= 0) return INFINITY;
	        sum = sum.add(BigInteger.valueOf(lb));
			// make sure we don't overflow when dealing with hard constraints 
	        if(sum.compareTo(BigInteger.valueOf(INFINITY)) >= 0) return INFINITY;
		}
		return sum.intValue();
	}

	/**
	 * This method computes the upper bound <b><i>UB(value)</i></b> of the given value <b><i>value</i></b>. 
	 * The upper bound equals the sum of the local cost of the given value (cost with assignments 
	 * of agents in the context) with the upper bound stored for all children. 
	 * 
	 * @param value : the value to which we want to compute the Upper Bound
	 * @return Integer: UB(value) = localCost(value) + \sum{child in children} ub(value, child);
	 */
	public Integer UB(int value) {
		BigInteger sum = BigInteger.valueOf(this.localCost.get(value));
		assert sum.compareTo(BigInteger.valueOf(0)) >= 0 : "\nProblem the local cost="+sum+" of value="+value+" is < 0";
		// make sure we don't overflow when dealing with hard constraints 
        if(sum.compareTo(BigInteger.valueOf(INFINITY)) >= 0) return INFINITY;
		for(Cost cost : this.getCost(value)) {
	        Integer ub = cost.getUb();
			// make sure we don't overflow when dealing with hard constraints 
	        if(ub.compareTo(INFINITY) >= 0) return INFINITY;
	        sum = sum.add(BigInteger.valueOf(ub));
			// make sure we don't overflow when dealing with hard constraints 
	        if(sum.compareTo(BigInteger.valueOf(INFINITY)) >= 0) return INFINITY;
		}
		return sum.intValue();
	}

	/**
	 * @return
	 */
	private Cost getMyCost() {
		int myValue = this.var.getValue();
		Instantiation context = agentView.getInstantiationOfNeighbors();
		for(Cost cost : childrenCosts.get(myValue).values()){
			context.priorityMerge(cost.getContext());
		}
		if(context.isEmpty()){
			System.out.println(this+" "+new Cost(context, LB, UB));
			throw new Error("");
		}
		return new Cost(context, LB, UB);
	}
	
	protected class ThresholdMsgHandler extends MessageHandler {
		public ThresholdMsgHandler(AbstractAgent agent) {
			super(agent, THRESHOLD);
		}

		public void processMessage(Message msg) throws StopException {
			Instantiation context = (Instantiation) msg.get(CONTEXT);
			if(agentView.compatibleWith(context)){
				threshold = (Integer) msg.get(THRESHOLD);
//				maintainThresholdInvariant();
				backtrack();
			}
		}

		public void sendThresholds() {
			Instantiation context = agentView.toInstantiation();
			int myValue = var.getValue();
			HashMap<VarIdentifier, Cost> costs = childrenCosts.get(myValue);
			for(VarIdentifier child : getChildrenIds()){
				this.sendThreshold(child, costs.get(child).getThreshold(), context);
			}
		}

		public void sendThreshold(VarIdentifier child, Integer threshold, Instantiation context) {
			Message msg = this.createMessage();
			msg.addContents(THRESHOLD, threshold); 
			msg.addContents(CONTEXT, context); 
			msg.setReceiverId(child);
			sendMessage(msg);
		}
	}
	
	public class ValueMsgHandler extends MessageHandler { 
		public ValueMsgHandler(AbstractAgent agent) {
			super(agent, VALUE);
		}

		public void processMessage(Message msg) throws StopException {
			if(!parentHasTerminated){ 
				VarAssignment varAss = (VarAssignment) msg.get(INFO);
				if(agentView.priorityMerge(varAss)) {
					clearIncompatibleCosts(); 
//					maintainThresholdInvariant();
					backtrack();
				}
			}
		}
		
		/** This method sends a VALUE message to each lower neighbor */
		public void sendValueMessages() {
			VarAssignment varAss = var.getVarAssignment();
			HashList<VarIdentifier> lowerNeighbors = communicator.getLowerNeighbors();
			for (VarIdentifier neighbor : lowerNeighbors) {
				sendValue(neighbor, varAss);
			}
		}
		
		public void sendValue(VarIdentifier neighbor, VarAssignment varAss) {
			Message msg = createMessage();
			msg.addContents(INFO, varAss);
			msg.setReceiverId(neighbor);
			sendMessage(msg);
		}
	}
	
	public class CostMsgHandler extends MessageHandler { 
		
		public CostMsgHandler(Adopt_ModiAgent agent) {
			super(agent, COST);
		}

		@Override
		public void processMessage(Message msg) throws StopException {
			Cost cost = (Cost) msg.get(COST);
			Instantiation context = cost.getContext();
			VarAssignment myAss = context.remove(var.getVarIdentifier());
			if(!parentHasTerminated){
				updateAgentView(context);
			}
			if(agentView.compatibleWith(context)){
				VarIdentifier child = msg.getSenderId();
				Integer lb = cost.getLb();
				Integer ub = cost.getUb();
				updateCost(myAss.getValue(), child, lb, ub, context);
				maintainChildThresholdInvariant();
			}
			backtrack();
		}
		
		public void sendCost() {
			Message msg = createMessage();
			msg.addContents(COST, getMyCost());
			msg.setReceiverId(getParentId());
			sendMessage(msg);
		}
	}
	
	public class TerminateMsgHandler extends TermMsgHandler {
		int myChildrenTerminated = 0;
		public TerminateMsgHandler(AbstractAgent agent) {
			super(agent);
		}

		@Override
		public boolean processTerminationDetection(Message msg) throws StopException {
			VarIdentifier senderId = msg.getSenderId();
			if(isMyParent(senderId)){
				parentHasTerminated = true;
				Instantiation context = (Instantiation) msg.get(Commands.CONTEXT);
				agentView.priorityMerge(context);
				backtrack();
			} else {
				if(amIRoot()){
					throw new StopException(Commands.SUCCESS);
				} else {
					myChildrenTerminated++;
					if(myChildrenTerminated==getChildrenIds().size()) {
						this.sendTerminate(getParentId());
					}
					ITerminate = true;
				}
			}
			return true;
		}
		
		public void sendTerminateMessages() {
			HashList<VarIdentifier> children = getChildrenIds();
			if(children.isEmpty()){
				this.sendTerminate(getParentId());
			} else {
				Instantiation currentContext = this.myAgent.agentView.toInstantiation();
				VarAssignment myAss = var.getVarAssignment();
				currentContext.add(myAss);
				for (VarIdentifier child : children) {
					this.sendTerminate(child, currentContext);
				}
			}
		}

		/**
		 * @param child
		 * @param currentContext
		 */
		private void sendTerminate(VarIdentifier child, Instantiation currentContext) {
			Message msg = this.createMessage();
			msg.addContents(Commands.CONTEXT, currentContext);
			msg.setReceiverId(child);
			sendMessage(msg);
		}

		/**
		 * @param child
		 */
		private void sendTerminate(VarIdentifier parent) {
			Message msg = this.createMessage();
			msg.setReceiverId(parent);
			sendMessage(msg);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public String solutionToString() {
		String solution = this.var.toString();
		if(amIRoot()){
			return "[LB="+this.LB+", UB="+this.UB+"] "+solution;
		}
		return solution;
	}
}
