/*
DisChoco: Distributed Constraint Reasoning Solver 
Copyright (C) 2007-2013 Redouane Ezzahir & Mohamed Wahbi 

DisChoco is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DisChoco is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A SimplePARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

To contact the authors: <http://dischoco.sourceforge.net/>
 */
package simple.dcop.afb;

import java.io.Serializable;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import kernel.AbstractAgent;
import kernel.AbstractVar;
import kernel.Constraint;
import kernel.StopException;
import kernel.VarIdentifier;
import kernel.com.Message;
import kernel.com.MessageHandler;
import kernel.events.VariableEvent;
import kernel.vars.integer.IntExternalVar;
import simple.SimpleAgent;
import simple.dcop.DCOPConstraintHandler;
import simple.dcop.SimpleSoftConstraint;
import simulation.metric.Metric;
import simulation.parser.PVariable;
import util.Commands;
import util.HashList;
import util.Instantiation;
import util.IntArray;
import util.VarAssignment;
import util.VarEventMask;

/**
 * Implementation of the Asynchronous Branch and Bound algorithm (AFB) for DCOPs. 
 * AFB have been proposed by 
 * <b>Gershman, A., Meisels, A., Zivan, R.</b>: 
 * <i>Asynchronous Forward Bounding for Distributed Constraints Optimization</i>. 
 * In: Proceedings of ECAI'06. pp. 103107 (2006). 
 * 
 * @since January 1, 2013<br/>
 * @authors Redouane Ezzahir & Mohamed Wahbi
 */
public class AFBAgent extends SimpleAgent {

	public static final Integer INFINITY = Constraint.INFINITY;

	public static final String CONTEXT = Commands.CONTEXT;
	/** The label of <b>OK</b> message sent from an agent to its successor (aka CPA message in the original paper above) */
	public static final String OK = "OK";
	/** The label of <b>BACK</b> message sent from an agent to its predecessor. This message does not appear in the original 
	 * paper (it is called CPA message). Here we differentiate between OK and BACK to facilitate the understanding of the algorithm.*/
	public static final String BACK = "BACK";
	/** The label of <b>FB</b> message sent from an agent to all unassigned agents to compute their lower bound on the partial assignment (aka FB_CPA message). */
	public static final String FB = "FB";
	/** The label of <b>LB</b> message sent from an agent to a higher agent as response to its request of computing a lower bound on a partial assignement (aka FB_ESTIMATE message) */
	public static final String LB = "LB";
	/** The label of <b>SOLUTION</b> message sent by the last agent to all other agents to inform them about a new solution.*/
	public static final String SOLUTION = "SOL";

	public static final String COST = "COST";

	/** The <b>OK</b> message sent from an agent to its successor (aka CPA message in the original paper above), 
	 * i.e., from <i>A<sub>j</sub></i> to <i>A<sub>j+1</sub></i>. */
	protected OKMessageHandler okMsgHandler = null; //
	/** The <b>BACK</b> message sent from an agent to its predecessor. This message does not appear in the original 
	 * paper (it is called CPA message). Here we differentiate between OK and BACK to facilitate the understanding 
	 * of the algorithm, i.e., from <i>A<sub>j</sub></i> to <i>A<sub>j-1</sub></i>.*/
	protected BackMessageHandler backMsgHandler = null;
	/** The <b>FB</b> message sent from an agent to all unassigned agents to compute their lower bound on the partial 
	 * assignment (aka FB_CPA message), i.e., from <i>A<sub>j</sub></i> to <i>A<sub>k</sub></i> s.t. k &gt; j. */
	protected FBMessageHandler fbMsgHandler = null;
	/** The <b>LB</b> message sent from an agent to a higher agent as response to its request of computing a lower 
	 * bound on a partial assignment (aka FB_ESTIMATE message), i.e., from <i>A<sub>j</sub></i> to <i>A<sub>i</sub></i> s.t. i&lt;j. */
	protected LBMessageHandler lbMsgHandler = null;
	/** The <b>SOLUTION</b> message sent by the last agent to all other agents to inform them about a new solution, 
	 * i.e., from <i>A<sub>j</sub></i> to <i>all agents</i>.*/
	protected SolMessageHandler solMsgHandler = null;

	/**
	 * The position of the current agent (A<sub>j</sub>) in order (ranging in 1..n)
	 */
	protected int myLevel;

	/**
	 * The cost best full assignment found so far. 
	 */
	protected Integer UB;

	/**
	 * A heuristic that stores for each possible value in the domain a lower bound on the constraints with lower agents. 
	 * <p>
	 * H : each element of this vector is the minimal lower bound on the cost of constraints involving the assignment
	 * <i>A<sub>j</sub> = v<sub>j</sub></i> and all agents <i>A<sub>k</sub></i> such that <i>k &gt; j</i>.
	 * </br>
	 * For each agent <i>A<sub>j</sub></i> and each value in its domain <i>v<sub>j</sub> &isin; D<sub>j</sub></i>, 
	 * we denote the minimal cost of
	 * the assignment (<i>x<sub>j</sub>=v<sub>j</sub></i>) incurred by an agent <i>A<sub>k</sub></i> by 
	 * <i>h<sub>k</sub>(x<sub>j</sub>=v<sub>j</sub>) = 
	 * min<sub>v<sub>k</sub> &isin; D<sub>k</sub></sub>{c<sub>jk</sub>(v<sub>j</sub>,v<sub>k</sub>)}</i>. 
	 * </br>
	 * We define <i>h(x<sub>j</sub>=v<sub>j</sub></i>) to be the sum of <i>h<sub>k</sub>(x<sub>j</sub>=v<sub>j</sub>)</i> 
	 * over all <i>k &gt; j</i>.  
	 * Intuitively, <i>h(x<sub>j</sub>=v<sub>j</sub>)</i> is a lower bound on the cost of constraints involving 
	 * the assignment (<i>x<sub>j</sub>=v<sub>j</sub></i>) and all agents <i>A<sub>k</sub></i> such that <i>k &gt; j</i>. 
	 * </p>
	 */
	protected Integer[] H;

	/**
	 * The local cost of each value v<sub>j</sub> &isin; D<sub>j</sub>
	 */
	protected Integer[] LC;

	protected HashMap<VarIdentifier, Integer> collectedLowerBounds;

	protected TimeStamp timeStamp;

	/** The initial domain, <i>D<sup>0</sup><sub>j</sub></i>, of the variable of current agent (<i>x<sub>j</sub></i>). */
	protected IntArray initialDomain;

	/** The current domain, <i>D<sub>j</sub></i>, of the variable of current agent (<i>x<sub>j</sub></i>). */
	protected IntArray currentDomain;

	/** The best solution found so far. */
	protected Instantiation bestSolution;

	protected PartialAssignment partialAssignment;

	public AFBAgent(int agId, String name, String description) {
		super(agId, name, description);
		this.constructorHelper();
	}

	public AFBAgent(PVariable variable) {
		super(variable);
		this.constructorHelper();
	}

	/**
	 * The constructor helper
	 */
	private void constructorHelper() {
		this.terminateMsgHandler = new TermMsgHandler(this);
		this.constraintHandler = new DCOPConstraintHandler(this);
		this.auxiliaryStatistics.put(Commands.SOLUTION_COST, new Metric(Commands.SOLUTION_COST, INFINITY, Metric.min));
		this.okMsgHandler = new OKMessageHandler(this);
		this.backMsgHandler = new BackMessageHandler(this);
		this.fbMsgHandler = new FBMessageHandler(this);
		this.lbMsgHandler = new LBMessageHandler(this);
		this.solMsgHandler = new SolMessageHandler(this);
		this.collectedLowerBounds = new HashMap<VarIdentifier, Integer>();
	}

	@Override
	public void configurePropagationEngine() {
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_HIGHER_THAN_LOCAL_VAR).add(VariableEvent.INSTANCIATE);
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_LOWER_THAN_LOCAL_VAR).add(VariableEvent.INSTANCIATE);
		this.eventsConfig.get(VarEventMask.LOCAL_VAR_IS_BETWEEN_L_H).add(VariableEvent.INSTANCIATE);
		this.constraintHandler.update();
	}

	@Override
	protected void addMsgHandlers() {
		this.communicator.addMsgHandler(this.okMsgHandler);
		this.communicator.addMsgHandler(this.backMsgHandler);
		this.communicator.addMsgHandler(this.fbMsgHandler);
		this.communicator.addMsgHandler(this.lbMsgHandler);
		this.communicator.addMsgHandler(this.solMsgHandler);
	}

	/*
	 * (non-Javadoc)
	 * @see kernel.AbstractAgent#initAgentView()
	 */
	@Override
	public void initAgentView() {
		HashList<VarIdentifier> varIds = this.agentOrder.getHigherAgents();
		for(VarIdentifier varId : varIds) {
			if (!this.communicator.isNeighbor(varId)) {
				VarAssignment varAss = new VarAssignment(varId, Integer.MIN_VALUE);
				IntExternalVar extVar = this.makeExternalVar(varAss);
				this.agentView.add(extVar);
			} else {
				this.agentView.add((IntExternalVar) this.getExternalVar(varId));
			}
		}
	}

	@Override
	public void startDistributedSearch() throws StopException {
		this.initialization();
		if (this.isInitializerAgent()) {
			this.partialAssignment = new PartialAssignment(this.agentView.toInstantiation(), new Integer(0));
			this.assign();
		}
	}

	/**
	 * This method checks if the current agent is the first agent in the agent ordering. 
	 * @return
	 */
	protected boolean isInitializerAgent() {
		return this.agentOrder.getRankOf(this.getVarIdentifier()) == 0;
	}

	/**
	 * This method checks if the current agent is the last agent in the agent ordering. 
	 * @return
	 */
	protected boolean isLastAgent() {
		return this.getNextAgent() == null;
	}

	/**
	 * This method returns the next agent in the ordering. 
	 * @return
	 */
	protected VarIdentifier getNextAgent() {
		return this.agentOrder.getNextOf(this.getVarIdentifier());
	}

	/**
	 * This method returns the previous agent in the ordering. 
	 * @return
	 */
	protected VarIdentifier getPreviousAgent() {
		return this.agentOrder.getPreviousOf(getVarIdentifier());
	}

	/**
	 * This method checks if all external variables in the scope of the given constraint are lower than the current 
	 * agent with regard to the agent ordering. 
	 * @param constraint
	 * @return true if all external variables are lower than the current agent; false otherwise. 
	 */
	private boolean areExternalVariablesLower(Constraint constraint) {
		for(AbstractVar xi : constraint.getExternalVars()) {
			if (this.agentOrder.isHigher(xi.getVarIdentifier())) {
				return false;
			}
		}
		return true;
	}

	/**
	 * This method computes {@link #H} for each value in the initial domain ({@link #initialDomain}). 
	 */
	protected void computeH() {
		for(int idx : this.var.getConstraintIndices()) {
			Constraint constraint = this.constraintHandler.getConstraint(idx);
			if (constraint instanceof SimpleSoftConstraint) {
				SimpleSoftConstraint softConstraint = (SimpleSoftConstraint) constraint;
				boolean allExternalVarsAreLower = areExternalVariablesLower(softConstraint);
				if (allExternalVarsAreLower) {
					for(int index = 0 ; index < this.initialDomain.size() ; index++) {
						int value = this.initialDomain.get(index);
						Integer cost = softConstraint.getMinCost(value);
						H[index] = addition(H[index], cost);
					}
				}
			}
		}
	}

	/**
	 * This method initializes the structure and bounds used by the algorithm. 
	 */
	private void initialization() {
		this.UB = INFINITY;
		this.currentDomain = new IntArray(var.getDomain().toArray());
		this.initialDomain = currentDomain.clone();
		this.myLevel = this.agentOrder.getMyRank() + 1;
		this.H = new Integer[currentDomain.size()];
		this.LC = new Integer[currentDomain.size()];
		Arrays.fill(this.H, new Integer(0));
		Arrays.fill(this.LC, new Integer(0));
		this.computeH();
		this.timeStamp = new TimeStamp(this.myLevel);
	}

	/**
	 * This method unassigns the current current variable
	 */
	protected void unassignCurrentVariable() {
		this.var.desinstanciate(VariableEvent.NOCAUSE);
	}

	/**
	 * This method is called when a value should be assigned to the current variable.
	 * @param value
	 * @throws StopException 
	 */
	protected void assignCurrentVariable(int value) throws StopException {
		this.var.instantiate(value);
		PartialAssignment cpa = this.partialAssignment.clone();
		VarAssignment myAssi = this.var.getVarAssignment().clone();
		cpa.add(myAssi);
		Integer cost = addition(cpa.getCost(), LC(myAssi.getValue()));
		cpa.setCost(cost);
		if (this.isLastAgent()) {
			this.reportSolution(cpa);
			this.assign();
		} else {
			this.sendForward(cpa);
		}
	}

	/**
	 * This method is called when the current succeeds in assigning its variable. It sends {@link #FB} 
	 * messages to all unassigned agents and one {@link #OK} message to the next agent in the ordering. 
	 * @param cpa
	 * @throws StopException
	 */
	protected void sendForward(PartialAssignment cpa) throws StopException {
		this.sendFB(cpa.getAssignments());
		this.sendOk(getNextAgent(), cpa);
	}

	/**
	 * This method returns the index of a value on the initial domain. 
	 * @param value : a value from the domain of the current variable. 
	 * @return int : the index of the given value on the initial domain.
	 */
	private int initialIndexOf(int value) {
		return this.initialDomain.indexOf(value);
	}

	/**
	 * This method returns the index of a value on the current domain. 
	 * @param value : a value from the domain of the current variable. 
	 * @return int : the index of the given value on the current domain.
	 */
	protected int currentIndexOf(int value) {
		return this.currentDomain.indexOf(value);
	}

	/**
	 * This method returns the Local Cost of the giving value <tt>value</tt>.
	 * @param value : a value from the domain of the current variable. 
	 * @return Integer : the Local Cost of the value. 
	 */
	protected Integer LC(int value) {
		return this.LC[initialIndexOf(value)];
	}

	/**
	 * This method computes the lower bound of the given value <tt>value</tt>. 
	 * <pre>
	 * LB(value) = LC[value] + H[v];
	 * </pre>
	 * @param value : a value from the domain of the current variable. 
	 * @return Integer : the lower bound of the given value. 
	 */
	protected Integer LowerBound(int value) {
		return addition(this.LC[initialIndexOf(value)], H[initialIndexOf(value)]);
	}

	/**
	 * This method computes the lower bound of the given value <tt>value</tt> with the actual context of higher agents. 
	 * <pre>
	 * LB(context U value) = cost[context] + LB[value];
	 * </pre>
	 * @param value : a value from the domain of the current variable. 
	 * @return Integer : the lower bound of the given value with the current context. 
	 */
	protected Integer evalute(int value) {
		return addition(this.partialAssignment.getCost(), this.LowerBound(value));
	}

	/**
	 * This method computes a local cost for the given value <tt>value</tt>.
	 * @param value : a value from the domain of the current variable. 
	 */
	protected void computeLocalCost(int value) {
		this.LC[initialIndexOf(value)] = ((DCOPConstraintHandler) this.constraintHandler).localCost(value);
	}

	protected void assign() throws StopException {
		clearCollectedLowerBounds();
		int lastIndex = -1;
		if (this.var.isInstantiated()) {
			lastIndex = this.currentDomain.indexOf(var.getValue());
			unassignCurrentVariable();
		}
		int index = lastIndex + 1;
		boolean found = false;
		while (!found && index < currentDomain.size()) {
			int myValue = this.currentDomain.get(index);
			this.computeLocalCost(myValue);
			if (evalute(myValue).compareTo(UB) < 0) {
				assignCurrentVariable(myValue);
				found = true;
			}
			index++;
		}
		if (!found) {
			backtrack();
		}
	}

	protected void clearCollectedLowerBounds() {
		this.collectedLowerBounds.clear();
	}

	/**
	 * <pre>
	 * procedure backtrack:
	 * clear estimates 
	 * if (A<sub>j</sub>=A<sub>1</sub>) then 
	 * 	broadcast(TERMINATE) 
	 * else 
	 * 	sendMSG:BACKCPA(cpa) to A<sub>j-1</sub>
	 * </pre>
	 * @throws StopException
	 */
	protected void backtrack() throws StopException {
		clearCollectedLowerBounds();
		if (isInitializerAgent()) {
			if (this.UB.equals(INFINITY)) {
				this.auxiliaryStatistics.get(Commands.SOLUTION_COST).value = this.UB;
				throw new StopException(Commands.FAILURE);
			} else {
				throw new StopException(Commands.SUCCESS);
			}
		} else {
			this.backtrackToPreviousAgent();
			this.agentView.clearFrom(this.agentView.lastVar().getVarAssignment());
			if (this.var.isInstantiated()) {
				this.unassignCurrentVariable();
			}
		}
	}

	protected void sendOk(VarIdentifier nextId, PartialAssignment pa) {
		this.okMsgHandler.sendOKMessage(nextId, pa);
	}

	protected void sendFB(Instantiation pa) {
		this.fbMsgHandler.sendFBMessages(pa);
	}

	protected void backtrackToPreviousAgent() {
		this.backMsgHandler.sendBackMessage(getPreviousAgent(), this.partialAssignment);
	}

	/**
	 * This method updates the agent view of the current agent by the given view <tt>view</tt>. 
	 * @param view
	 */
	public void updateAgentView(Instantiation view) {
		this.agentView.clearExcept(view);
		this.agentView.priorityMerge(view);
	}

	protected void reportSolution(PartialAssignment solution) {
		this.bestSolution = solution.getAssignments();
		this.UB = solution.getCost();
		this.auxiliaryStatistics.get(Commands.SOLUTION_COST).value = this.UB;
		this.solMsgHandler.reportSolution(solution.getAssignments(), solution.getCost());
	}

	/**
	 * This method re-initializes the local costs of all values to 0. 
	 */
	protected void clearLocalCosts() {
		Arrays.fill(this.LC, new Integer(0));
	}

	protected void processOK(PartialAssignment pa) throws StopException {
		this.partialAssignment = pa;
		unassignCurrentVariable();
		clearLocalCosts();
		updateAgentView(pa.getAssignments());
		clearCollectedLowerBounds();
		if (pa.getCost().compareTo(this.UB) >= 0) {
			backtrack();
		} else {
			assign();
		}
	}

	protected void processBack(PartialAssignment pa) throws StopException {
		if (pa.getCost().compareTo(this.UB) >= 0) {
			backtrack();
		} else {
			assign();
		}
	}

	/**
	 * <pre>
	 * processFB(A<sub>i</sub>, pa) { 
	 * 	lb <- estimation based on pa;
	 *	sendMsg:LB(lb, pa) to A<sub>i</sub>;
	 * }
	 * </pre>
	 * @param senderId
	 * @param cpa
	 * @throws StopException
	 */
	protected void processFB(VarIdentifier senderId, Instantiation pa) throws StopException {
		Integer lb = Integer.MAX_VALUE;
		for(int value : currentDomain) {
			this.computeLocalCost(value);
			Integer lb_candidate = this.LowerBound(value);
			if (lb_candidate.compareTo(lb) < 0) {
				lb = lb_candidate;
			}
		}
		this.lbMsgHandler.sendLBMessage(pa, senderId, lb);
	}

	protected int sumLowerBounds() {
		int LB = 0;
		for(Integer lb : collectedLowerBounds.values()) {
			LB = addition(LB, lb);
		}
		return LB;
	}

	/**
	 * <pre>
	 * processLB(lb, pa, A<sub>k</sub>)
	 * lb<sub>k</sub> <- lb
	 * if ( CPA.cost + all saved estimates) >= UB )
	 * 	assign()
	 * </pre>
	 * @param lb
	 * @param pa
	 * @param senderId
	 * @throws StopException
	 */
	protected void processLB(Integer lb, Instantiation pa, VarIdentifier senderId) throws StopException {
		this.collectedLowerBounds.put(senderId, lb);
		Integer bound = addition(partialAssignment.getCost(), LC(var.getValue()));
		bound = addition(bound, sumLowerBounds());
		if (bound.compareTo(this.UB) >= 0) {
			assign();
		}
	}

	public void processSolution(Instantiation solution, Integer cost) {
		if (cost.compareTo(UB) < 0) {
			this.bestSolution = solution;
			this.UB = cost;
			this.auxiliaryStatistics.get(Commands.SOLUTION_COST).value = this.UB;
		}
	}

	public class OKMessageHandler extends MessageHandler {

		public OKMessageHandler(AbstractAgent agent) {
			super(agent, OK);
		}

		@Override
		public void processMessage(Message msg) throws StopException {
			PartialAssignment pa = (PartialAssignment) msg.get(CONTEXT);
			int updatedLevel = compareTimeStamp(pa.assignments, pa.size());
			if (updatedLevel >= 0) {
				processOK(pa);
			}
		}

		public void sendOKMessage(VarIdentifier nextId, PartialAssignment pa) {
			Message msg = createMessage();
			msg.setReceiverId(nextId);
			msg.addContents(CONTEXT, pa.clone());
			sendMessage(msg);
		}

	}

	public class FBMessageHandler extends MessageHandler {

		public FBMessageHandler(AbstractAgent agent) {
			super(agent, FB);
		}

		@Override
		public void processMessage(Message msg) throws StopException {
			Instantiation pa = (Instantiation) msg.get(CONTEXT);
			int updatedLevel = compareTimeStamp(pa, pa.size());
			if (updatedLevel >= 0) {
				if (updatedLevel > 0) {
					updateAgentView(pa);
				}
				processFB(msg.getSenderId(), pa);

			}
		}

		public void sendFBMessages(Instantiation pa) {
			for(VarIdentifier lower : agentOrder.getLowerAgents()) {
				Message msg = createMessage();
				msg.setReceiverId(lower);
				msg.addContents(CONTEXT, pa.clone());
				sendMessage(msg);
			}
		}
	}

	public class LBMessageHandler extends MessageHandler {

		public LBMessageHandler(AbstractAgent agent) {
			super(agent, LB);
		}

		@Override
		public void processMessage(Message msg) throws StopException {
			Instantiation pa = (Instantiation) msg.get(CONTEXT);
			Integer lb = (Integer) msg.get(LB);
			int updatedLevel = compareTimeStamp(pa, pa.size());
			if (updatedLevel == 0 && var.isInstantiated()) {
				processLB(lb, pa, msg.getSenderId());
			}
		}

		public void sendLBMessage(Instantiation pa, VarIdentifier receiverId, Integer lb) {
			Message msg = createMessage();
			msg.setReceiverId(receiverId);
			msg.addContents(CONTEXT, pa.clone());
			msg.addContents(LB, lb);
			sendMessage(msg);
		}
	}

	public class BackMessageHandler extends MessageHandler {

		public BackMessageHandler(AbstractAgent agent) {
			super(agent, BACK);
		}

		@Override
		public void processMessage(Message msg) throws StopException {
			PartialAssignment pa = (PartialAssignment) msg.get(CONTEXT);
			int updatedLevel = compareTimeStamp(pa.assignments, pa.size());
			if (updatedLevel == 0 && var.isInstantiated()) {
				pa.removeLast();
				processBack(pa);
			}
		}

		public void sendBackMessage(VarIdentifier previousId, PartialAssignment pa) {
			Message msg = this.createMessage();
			msg.addContents(CONTEXT, pa.clone());
			msg.setReceiverId(previousId);
			sendMessage(msg);
		}
	}

	public class SolMessageHandler extends MessageHandler {

		public SolMessageHandler(AbstractAgent agent) {
			super(agent, SOLUTION);
		}

		@Override
		public void processMessage(Message msg) throws StopException {
			processSolution((Instantiation) msg.get(CONTEXT), (Integer) msg.get(COST));
		}

		public void reportSolution(Instantiation pa, Integer ub) {
			HashList<VarIdentifier> agentsIDs = getOrder();
			agentsIDs.remove(getVarIdentifier());
			for(VarIdentifier agentId : agentsIDs) {
				Message msg = createMessage();
				msg.addContents(CONTEXT, pa.clone());
				msg.addContents(COST, ub);
				msg.setReceiverId(agentId);
				sendMessage(msg);
			}
		}
	}

	public int compareTimeStamp(Instantiation pa, int upto) {
		int[] timeStamp = pa.getTimeStamp();
		int[] myTimeStamp = new int[this.myLevel];
		int i = 0;
		for(IntExternalVar v : agentView) {
			myTimeStamp[i++] = v.getTimeCounter();
		}
		myTimeStamp[i] = var.getTimeCounter();

		int j = 0;
		for( ; j < myTimeStamp.length && j < timeStamp.length && j <= upto ; j++) {
			if (myTimeStamp[j] > timeStamp[j]) {
				return -(j + 1);
			} else if (myTimeStamp[j] < timeStamp[j]) {
				return j + 1;
			}
		}
		if (j <= upto && myTimeStamp.length < timeStamp.length) {
			return j + 1;
		}
		return 0;
	}

	/**
	 * This method returns the sum of two costs. 
	 * @param c1 : the first cost.
	 * @param c2 : the second cost. 
	 * @return Integer : the sum of two given costs <tt>c1</tt> and <tt>c2</tt>. 
	 */
	public static Integer addition(Integer c1, Integer c2) {
		return DCOPConstraintHandler.sumCosts(c1, c2);
	}

	@Override
	public String solutionToString() {
		String solution = "";
		if (this.isInitializerAgent()) {
			if (this.isFinished()) {
				solution = "Cost=" + this.UB;
			}
			solution += " sol=" + (this.isFinished()? this.bestSolution : this.var.getVarAssignment());
		} else {
			solution = "" + (this.isFinished()? this.bestSolution : this.var.getVarAssignment());
		}
		return solution;
	}

	class PartialAssignment implements Cloneable, Serializable {

		protected static final long serialVersionUID = 1L;

		protected Instantiation assignments;

		private Integer cost = 0;

		public PartialAssignment(Comparator<? super VarAssignment> comparator) {
			this.assignments = new Instantiation(comparator);
			this.cost = 0;
		}

		public PartialAssignment(Instantiation assignments, Comparator<? super VarAssignment> comparator) {
			this.assignments = new Instantiation(assignments.getInstantiation(), comparator);
		}

		public PartialAssignment(Instantiation assignments, Integer cost) {
			this(assignments, assignments.comparator());
			this.cost = cost;
		}

		public void removeLast() {
			this.assignments.remove(last());
		}

		public PartialAssignment clone() {
			PartialAssignment clone = new PartialAssignment(this.assignments.comparator());
			clone.assignments = this.assignments.clone();
			clone.cost = this.cost;
			return clone;
		}

		public int size() {
			return this.assignments.size();
		}

		public VarAssignment last() {
			return this.assignments.last();
		}

		public void add(VarAssignment varAssign) {
			this.assignments.add(varAssign);
		}

		public String toString() {
			return "cost: " + getCost() + ", PA=" + assignments.toString();
		}

		public PartialAssignment subCPA(int level) {
			return new PartialAssignment(this.getAssignments(level), this.assignments.comparator());
		}

		public Instantiation getAssignments(int level) {
			if (level <= 0) {
				return new Instantiation(this.assignments.comparator());
			}
			if (level >= this.assignments.size()) {
				return this.assignments;
			}
			VarAssignment v = this.assignments.get(level);
			return this.assignments.headSet(v);
		}

		public Instantiation getAssignments(VarAssignment varAss) {
			return this.assignments.headSet(varAss);
		}

		public void update(PartialAssignment pa) {
			this.setAssignments(pa.assignments);
			this.cost = pa.cost;
		}

		/**
		 * @return the assignments
		 */
		public Instantiation getAssignments() {
			return this.assignments;
		}

		/**
		 * @param assignments the assignments to set
		 */
		public void setAssignments(Instantiation assignments) {
			this.assignments = assignments;
		}

		/**
		 * @return the cost
		 */
		public Integer getCost() {
			return this.cost;
		}

		/**
		 * @param cost the cost to set
		 */
		public void setCost(Integer cost) {
			this.cost = cost;
		}
	}

}
