package simple.dcop.afb;

import java.io.Serializable;
import java.util.Arrays;
import kernel.AbstractAgent;
import kernel.AbstractVar;
import kernel.Constraint;
import kernel.StopException;
import kernel.VarIdentifier;
import kernel.com.Message;
import kernel.com.MessageHandler;
import kernel.events.VariableEvent;
import kernel.vars.integer.IntExternalVar;
import simple.SimpleAgent;
import simple.dcop.DCOPConstraintHandler;
import simple.dcop.SimpleSoftConstraint;
import simulation.metric.Metric;
import simulation.parser.PVariable;
import util.Commands;
import util.HashList;
import util.Instantiation;
import util.IntArray;
import util.VarAssignment;
import util.VarEventMask;

/**
 * Implementation of the Asynchronous Branch and Bound with Backjumping algorithm (AFB_BJ) for DCOPs
 * AFB_BJ have been proposed by 
 * <b>Gershman, A., Meisels, A., Zivan, R.</b>: 
 * <i>Asynchronous Forward Bounding for Distributed COPs</i>. 
 * JAIR 34, 6188 (2009). 
 * 
 * @since January 1, 2013<br/>
 * @authors Redouane Ezzahir & Mohamed Wahbi
 */
public class AFB_BJAgent extends SimpleAgent {

	public static final Integer INFINITY = Constraint.INFINITY;
	public static final String CONTEXT = Commands.CONTEXT;
	/** The label of <b>OK</b> message sent from an agent to its successor (aka CPA message in the original paper above) */
	public static final String OK = "OK";
	/** The label of <b>BACK</b> message sent from an agent to its predecessor. This message does not appear in the original 
	 * paper (it is called CPA message). Here we differentiate between OK and BACK to facilitate the understanding of the algorithm.*/
	public static final String BACK = "BACK";
	/** The label of <b>FB</b> message sent from an agent to all unassigned agents to compute their lower bound on the partial assignment (aka FB_CPA message). */
	public static final String FB = "FB";
	/** The label of <b>LB</b> message sent from an agent to a higher agent as response to its request of computing a lower bound on a partial assignement (aka FB_ESTIMATE message) */
	public static final String LB = "LB";
	/** The label of <b>SOLUTION</b> message sent by the last agent to all other agents to inform them about a new solution.*/
	public static final String SOLUTION = "SOL";

	public static final String COST = "COST";

	/** The <b>OK</b> message sent from an agent to its successor (aka CPA message in the original paper above), 
	 * i.e., from <i>A<sub>j</sub></i> to <i>A<sub>j+1</sub></i>. */
	protected OKMessageHandler okMsgHandler = null; //
	/** The <b>BACK</b> message sent from an agent to its predecessor. This message does not appear in the original 
	 * paper (it is called CPA message). Here we differentiate between OK and BACK to facilitate the understanding 
	 * of the algorithm, i.e., from <i>A<sub>j</sub></i> to <i>A<sub>j-1</sub></i>.*/
	protected BackMessageHandler backMsgHandler = null;
	/** The <b>FB</b> message sent from an agent to all unassigned agents to compute their lower bound on the partial 
	 * assignment (aka FB_CPA message), i.e., from <i>A<sub>j</sub></i> to <i>A<sub>k</sub></i> s.t. k>j. */
	protected FBMessageHandler fbMsgHandler = null;
	/** The <b>LB</b> message sent from an agent to a higher agent as response to its request of computing a lower 
	 * bound on a partial assignment (aka FB_ESTIMATE message), i.e., from <i>A<sub>j</sub></i> to <i>A<sub>i</sub></i> s.t. i&lt;j. */
	protected LBMessageHandler lbMsgHandler = null;
	/** The <b>SOLUTION</b> message sent by the last agent to all other agents to inform them about a new solution, 
	 * i.e., from <i>A<sub>j</sub></i> to <i>all agents</i>.*/
	protected SolMessageHandler solMsgHandler = null;

	/**
	 * The position of the current agent (A<sub>j</sub>) in order (ranging in 1..n)
	 */
	protected int myLevel;

	/**
	 * The cost best full assignment found so far. 
	 */
	protected Integer UB;

	/**
	 * A heuristic that stores for each possible value in the domain a lower bound on the constraints with lower agents. 
	 * <p>
	 * H : each element of this vector is the minimal lower bound on the cost of constraints involving the assignment
	 * <i>A<sub>j</sub> = v<sub>j</sub></i> and all agents <i>A<sub>k</sub></i> such that <i>k &gt; j</i>.
	 * </br>
	 * For each agent <i>A<sub>j</sub></i> and each value in its domain <i>v<sub>j</sub> &isin; D<sub>j</sub></i>, 
	 * we denote the minimal cost of
	 * the assignment (<i>x<sub>j</sub>=v<sub>j</sub></i>) incurred by an agent <i>A<sub>k</sub></i> by 
	 * <i>h<sub>k</sub>(x<sub>j</sub>=v<sub>j</sub>) = 
	 * min<sub>v<sub>k</sub> &isin; D<sub>k</sub></sub>{c<sub>jk</sub>(v<sub>j</sub>,v<sub>k</sub>)}</i>. 
	 * </br>
	 * We define <i>h(x<sub>j</sub>=v<sub>j</sub></i>) to be the sum of <i>h<sub>k</sub>(x<sub>j</sub>=v<sub>j</sub>)</i> 
	 * over all <i>k &gt; j</i>.  
	 * Intuitively, <i>h(x<sub>j</sub>=v<sub>j</sub>)</i> is a lower bound on the cost of constraints involving 
	 * the assignment (<i>x<sub>j</sub>=v<sub>j</sub></i>) and all agents <i>A<sub>k</sub></i> such that <i>k &gt; j</i>. 
	 * </p>
	 */
	protected Integer[] H;

	/**
	 * The local cost of each value <i>v<sub>j</sub> &isin; D<sub>j</sub></i> at each level i (LC[i][v<sub>j</sub>]), i&lt;j. 
	 */
	protected Integer[][] LC;

	protected Integer[][] sumLowerBounds;

	protected TimeStamp timeStamp;

	/** The initial domain, <i>D<sup>0</sup><sub>j</sub></i>, of the variable of current agent (<i>x<sub>j</sub></i>). */
	protected IntArray initialDomain;

	/** The current domain, <i>D<sub>j</sub></i>, of the variable of current agent (<i>x<sub>j</sub></i>). */
	protected IntArray currentDomain;

	protected Instantiation bestSolution;

	protected PartialAssignement partialAssignment;

	public AFB_BJAgent(int agId, String name, String description) {
		super(agId, name, description);
		this.constructorHelper();
	}

	public AFB_BJAgent(PVariable variable) {
		super(variable);
		this.constructorHelper();
	}

	/**
	 * The constructor helper
	 */
	private void constructorHelper() {
		this.terminateMsgHandler = new TermMsgHandler(this);
		this.constraintHandler = new DCOPConstraintHandler(this);
		this.auxiliaryStatistics.put(Commands.SOLUTION_COST, new Metric(Commands.SOLUTION_COST, INFINITY, Metric.min));
		this.okMsgHandler = new OKMessageHandler(this);
		this.backMsgHandler = new BackMessageHandler(this);
		this.fbMsgHandler = new FBMessageHandler(this);
		this.lbMsgHandler = new LBMessageHandler(this);
		this.solMsgHandler = new SolMessageHandler(this);
	}

	@Override
	public void configurePropagationEngine() {
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_HIGHER_THAN_LOCAL_VAR).add(VariableEvent.INSTANCIATE);
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_LOWER_THAN_LOCAL_VAR).add(VariableEvent.INSTANCIATE);
		this.eventsConfig.get(VarEventMask.LOCAL_VAR_IS_BETWEEN_L_H).add(VariableEvent.INSTANCIATE);
		this.constraintHandler.update();
	}

	@Override
	protected void addMsgHandlers() {
		this.communicator.addMsgHandler(this.okMsgHandler);
		this.communicator.addMsgHandler(this.backMsgHandler);
		this.communicator.addMsgHandler(this.fbMsgHandler);
		this.communicator.addMsgHandler(this.lbMsgHandler);
		this.communicator.addMsgHandler(this.solMsgHandler);
	}

	/*
	 * (non-Javadoc)
	 * @see kernel.AbstractAgent#initAgentView()
	 */
	@Override
	public void initAgentView() {
		HashList<VarIdentifier> varIds = this.agentOrder.getHigherAgents();
		for(VarIdentifier varId : varIds) {
			if (!this.communicator.isNeighbor(varId)) {
				VarAssignment varAss = new VarAssignment(varId, Integer.MIN_VALUE);
				IntExternalVar extVar = this.makeExternalVar(varAss);
				this.agentView.add(extVar);
			} else {
				this.agentView.add((IntExternalVar) this.getExternalVar(varId));
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * @see kernel.AbstractAgent#startDistributedSearch()
	 */
	@Override
	public void startDistributedSearch() throws StopException {
		this.initialization();
		if (this.isInitializerAgent()) {
			this.assign();
		}
	}

	/**
	 * This method checks if the current agent is the first agent in the agent ordering. 
	 * @return
	 */
	protected boolean isInitializerAgent() {
		return this.agentOrder.getRankOf(this.getVarIdentifier()) == 0;
	}

	/**
	 * This method checks if the current agent is the last agent in the agent ordering. 
	 * @return
	 */
	protected boolean isLastAgent() {
		return this.getNextAgent() == null;
	}

	/**
	 * This method returns the next agent in the ordering. 
	 * @return
	 */
	protected VarIdentifier getNextAgent() {
		return this.agentOrder.getNextOf(this.getVarIdentifier());
	}

	/**
	 * This method checks if all external variables in the scope of the given constraint are lower than the current 
	 * agent with regard to the agent ordering. 
	 * @param constraint
	 * @return true if all external variables are lower than the current agent; false otherwise. 
	 */
	private boolean areExternalVariablesLower(Constraint constraint) {
		for(AbstractVar xi : constraint.getExternalVars()) {
			if (this.agentOrder.isHigher(xi.getVarIdentifier())) {
				return false;
			}
		}
		return true;
	}

	/**
	 * This method computes {@link #H} for each value in the initial domain ({@link #initialDomain}). 
	 */
	protected void computeH() {
		for(int idx : this.var.getConstraintIndices()) {
			Constraint constraint = this.constraintHandler.getConstraint(idx);
			if (constraint instanceof SimpleSoftConstraint) {
				SimpleSoftConstraint softConstraint = (SimpleSoftConstraint) constraint;
				boolean allExternalVarsAreLower = areExternalVariablesLower(softConstraint);
				if (allExternalVarsAreLower) {
					for(int index = 0 ; index < this.initialDomain.size() ; index++) {
						int value = this.initialDomain.get(index);
						Integer cost = softConstraint.getMinCost(value);
						this.H[index] = addition(this.H[index], cost);
					}
				}
			}
		}
	}

	/**
	 * This method initializes the structure and bounds used by the algorithm. 
	 */
	private void initialization() {
		this.UB = INFINITY;
		this.currentDomain = new IntArray(var.getDomain().toArray());
		this.initialDomain = currentDomain.clone();
		this.myLevel = this.agentOrder.getMyRank() + 1;
		this.H = new Integer[currentDomain.size()];
		Arrays.fill(this.H, new Integer(0));
		this.computeH();
		this.LC = new Integer[this.myLevel][currentDomain.size()];
		for(int i = 0 ; i < this.LC.length ; i++){
			Arrays.fill(this.LC[i], new Integer(0));
		}
		this.sumLowerBounds = new Integer[this.myLevel + 1][currentDomain.size()];
		this.clearCollectedLowerBounds();
		this.partialAssignment = new PartialAssignement(this.myLevel);
		this.timeStamp = new TimeStamp(this.myLevel);
	}

	/**
	 * This method unassigns the current current variable
	 */
	protected void unassignMyVariable() {
		this.var.desinstanciate(VariableEvent.NOCAUSE);
	}

	/**
	 * This method is called when a value should be assigned to the current variable.
	 * @param value
	 * @throws StopException 
	 */
	protected void assignCurrentVariable(int value) throws StopException {
		this.var.instantiate(value);
		this.updateTimeStamp(this.partialAssignment.size());
		PartialAssignement pa = this.partialAssignment.clone();
		VarAssignment myAssi = this.var.getVarAssignment().clone();
		pa.add(myAssi, addition(pa.getCost(), LC(myAssi.getValue())));
		if (this.isLastAgent()) {
			this.reportSolution(pa);
			this.assign();
		} else {
			this.sendForward(pa);
		}
	}

	/**
	 * This method is called when the current succeeds in assigning its variable. It sends {@link #FB} 
	 * messages to all unassigned agents and one {@link #OK} message to the next agent in the ordering. 
	 * @param cpa
	 * @throws StopException
	 */
	protected void sendForward(PartialAssignement pa) throws StopException {
		this.sendFB(pa);
		this.sendOk(getNextAgent(), pa.clone());
	}

	/**
	 * This method returns the index of a value on the initial domain. 
	 * @param value : a value from the domain of the current variable. 
	 * @return Integer : the index of the given value on the initial domain.
	 */
	protected int initialIndexOf(int value) {
		return this.initialDomain.indexOf(value);
	}
	
	/**
	 * This method returns the index of a value on the current domain. 
	 * @param value : a value from the domain of the current variable. 
	 * @return Integer : the index of the given value on the current domain.
	 */
	protected int currentIndexOf(int value) {
		return this.currentDomain.indexOf(value);
	}
	
	/**
	 * This method returns the Local Cost of the giving value <tt>value</tt>.
	 * @param value : a value from the domain of the current variable. 
	 * @return Integer : the Local Cost of the value. 
	 */
	protected Integer LC(int value) {
		return this.LC[this.myLevel - 1][initialIndexOf(value)];
	}

	/**
	 * This method computes the lower bound of the given value. 
	 * <pre>
	 * LB(value) = LC[value] + H[v];
	 * </pre>
	 * 
	 * @param value
	 * @param value2 
	 * @return
	 */
	protected int LowerBound(int k, int value) {
		return addition(this.LC[k][initialIndexOf(value)], H[initialIndexOf(value)]);
	}

	protected Integer evalute(int value) {
		return addition(this.partialAssignment.getCost(), this.LowerBound(this.myLevel - 1, value));
	}

	/**
	 * This method removes all lower bounds collected from lower agents. 
	 */
	protected void clearCollectedLowerBounds() {
		for(int i = 0 ; i < this.sumLowerBounds.length ; i++){
			Arrays.fill(this.sumLowerBounds[i], new Integer(0));
		}
	}

	protected void sendOk(VarIdentifier nextId, PartialAssignement pa) {
		this.okMsgHandler.sendOKMessage(nextId, pa);
	}

	protected void sendFB(PartialAssignement pa) {
		this.fbMsgHandler.sendFBMessages(pa);
	}

	public void updateAgentView(Instantiation view) {
		this.agentView.clearExcept(view);
		this.agentView.priorityMerge(view);
	}

	protected void reportSolution(PartialAssignement solution) {
		this.bestSolution = solution.getAssignments();
		this.UB = solution.getCost();
		this.auxiliaryStatistics.get(Commands.SOLUTION_COST).value = this.UB;
		this.solMsgHandler.reportSolution(solution.getAssignments(), solution.getCost());
	}

	protected void assign() throws StopException {
		int lastIndex = -1;
		if (this.var.isInstantiated()) {
			lastIndex = this.currentDomain.indexOf(var.getValue());
			unassignMyVariable();
		}
		int index = lastIndex + 1;
		boolean found = false;
		while (!found && index < this.currentDomain.size()) {
			int myValue = this.currentDomain.get(index);
			if (evalute(myValue).compareTo(this.UB) < 0) {
				assignCurrentVariable(myValue);
				found = true;
			}
			index++;
		}
		if (!found) {
			backtrack();
		}
	}

	protected void reorderDomain() {
		int length = this.currentDomain.size();
		for(int i = 0 ; i < length - 1 ; i++) {
			int value_i = this.currentDomain.get(i);
			for(int j = i + 1 ; j < length ; j++) {
				int value_j = this.currentDomain.get(j);
				if (this.LC(value_j).compareTo(this.LC(value_i)) < 0) {
					this.currentDomain.swapByIndexes(i, j);
					value_i = value_j;
				}
			}
		}
	}

	protected void processOK(PartialAssignement pa, int updatedLevel) throws StopException {
		unassignMyVariable();
		clearCollectedLowerBounds();
		if (updatedLevel > 0) {
			this.partialAssignment.update(pa);
			this.updateAgentView(pa.getAssignments());
			cumputeLocalCosts(updatedLevel, this.myLevel - 1);
		}
		reorderDomain();
		if (pa.getCost().compareTo(this.UB) >= 0) {
			backtrack();
		} else {
			assign();
		}
	}

	private void cumputeLocalCosts(int fromLevel, int toLevel) {
		for(int i = fromLevel ; i <= toLevel ; i++) {
			this.LC[i] = this.computeLocalCost(i);
		}
	}

	protected Integer[] computeLocalCost(int i) {
		if (i > this.agentView.size()) {
			return null;
		}
		Integer[] lc = new Integer[this.initialDomain.size()];
		Arrays.fill(lc, new Integer(0));
		for(int idx : this.var.getConstraintIndices()) {
			Constraint constraint = this.constraintHandler.getConstraint(idx);
			if (constraint instanceof SimpleSoftConstraint) {
				SimpleSoftConstraint sc = (SimpleSoftConstraint) constraint;
				if (this.isScopeHigherThan(sc, i)) {
					for(int index = 0 ; index < this.initialDomain.size() ; index++) {
						int value = this.initialDomain.get(index);
						lc[index] = addition(lc[index], sc.evaluate(value));
					}
				}
			} else {
				throw new UnsupportedOperationException("not yet implemented");
			}
		}
		return lc;
	}

	/**
	 * This method checks if the constraint can be used when computing local cost of the given level <tt>level</tt>.  
	 * That is, if the last variable in the scope of the constraint is higher than level in the ordering of agents. 
	 * @param constraint : the constraint
	 * @param level: the level
	 * @return
	 */
	private boolean isScopeHigherThan(Constraint constraint, int level) {
		for(AbstractVar xi : constraint.getExternalVars()) {
			if (!xi.isInstantiated() || this.agentOrder.getRankOf(xi.getVarIdentifier()) >= level) {
				return false;
			}
		}
		return true;
	}

	protected void processBack(Integer cost) throws StopException {
		if (cost.compareTo(this.UB) >= 0) {
			backtrack();
		} else {
			assign();
		}
	}

	/**
	 * <pre>
	 * processFB(A<sub>i</sub>, pa) { 
	 * 	lb <- estimation based on pa;
	 *	sendMsg:LB(lb, pa) to A<sub>i</sub>;
	 * }
	 * </pre>
	 * @param senderId
	 * @param updatedLevel 
	 * @param pa
	 * @throws StopException
	 */
	protected void processFB(VarIdentifier senderId, PartialAssignement pa, int updatedLevel) throws StopException {
		if (updatedLevel > 0) {
			this.partialAssignment.update(pa);
			this.updateAgentView(pa.getAssignments());
			this.cumputeLocalCosts(updatedLevel, pa.size());
		}
		Integer[] lb = this.computeLowerBoundsVector(pa.size());
		this.lbMsgHandler.sendLBMessage(pa.getAssignments(), senderId, lb);
	}

	private Integer[] computeLowerBoundsVector(int untilLevel) {
		Integer[] lb = new Integer[untilLevel + 1];
		for(int i = 0 ; i < lb.length ; i++) {
			lb[i] = computeLowerBound(i);
		}
		return lb;
	}

	private Integer computeLowerBound(int untilLevel) {
		Integer lb = Integer.MAX_VALUE;
		for(int value : currentDomain) {
			Integer lb_candidate = this.LowerBound(untilLevel, value);
			if (lb_candidate.compareTo(lb) < 0) {
				lb = lb_candidate;
			}
		}
		return lb;
	}

	/**
	 * <pre>
	 * processLB(lb, pa, A<sub>k</sub>)
	 * lb<sub>k</sub> <- lb
	 * if ( CPA.cost + all saved estimates) >= UB )
	 * 	assign()
	 * </pre>
	 * @param lb
	 * @param pa
	 * @param senderId
	 * @throws StopException
	 */
	protected void processLB(Integer[] lb, Instantiation pa, VarIdentifier senderId) throws StopException {
		int value = pa.last().getValue();
		int idx = this.initialIndexOf(value);

		for(int i = 0 ; i < this.sumLowerBounds.length ; i++) {
			this.sumLowerBounds[i][idx] = addition(lb[i], this.sumLowerBounds[i][idx]);
		}
		if (value == var.getValue()) {
			Integer bound = addition(partialAssignment.getCost(), LC(value));
			bound = addition(bound, this.sumLowerBounds[this.myLevel][idx]);
			if (bound.compareTo(this.UB) >= 0) {
				assign();
			}
		}

	}

	public void processSolution(Instantiation solution, Integer cost) {
		if (cost.compareTo(this.UB) < 0) {
			this.bestSolution = solution;
			this.UB = cost;
			this.auxiliaryStatistics.get(Commands.SOLUTION_COST).value = this.UB;
		}
	}

	public class OKMessageHandler extends MessageHandler {

		public OKMessageHandler(AbstractAgent agent) {
			super(agent, OK);
		}

		@Override
		public void processMessage(Message msg) throws StopException {
			PartialAssignement pa = (PartialAssignement) msg.get(CONTEXT);
			int updatedLevel = compareTimeStamp(pa.getAssignments(), pa.size());
			if (updatedLevel >= 0) {
				processOK((PartialAssignement) msg.get(CONTEXT), updatedLevel);
			}
		}

		public void sendOKMessage(VarIdentifier nextId, PartialAssignement pa) {
			Message msg = createMessage();
			msg.setReceiverId(nextId);
			msg.addContents(CONTEXT, pa.clone());
			sendMessage(msg);
		}

	}

	public class FBMessageHandler extends MessageHandler {

		public FBMessageHandler(AbstractAgent agent) {
			super(agent, FB);
		}

		@Override
		public void processMessage(Message msg) throws StopException {
			PartialAssignement pa = (PartialAssignement) msg.get(CONTEXT);
			int updatedLevel = compareTimeStamp(pa.getAssignments(), pa.size());
			if (updatedLevel >= 0) {
				if (updatedLevel > 0 || agentView.getVar(msg.getSenderId()).isInstantiated()){
					processFB(msg.getSenderId(), pa, updatedLevel);
				}
			}
		}

		public void sendFBMessages(PartialAssignement pa) {
			for(VarIdentifier lower : agentOrder.getLowerAgents()) {
				Message msg = createMessage();
				msg.setReceiverId(lower);
				msg.addContents(CONTEXT, pa.clone());
				sendMessage(msg);
			}
		}
	}

	public class LBMessageHandler extends MessageHandler {

		public LBMessageHandler(AbstractAgent agent) {
			super(agent, LB);
		}

		@Override
		public void processMessage(Message msg) throws StopException {
			Instantiation pa = (Instantiation) msg.get(CONTEXT);
			int updatedLevel = compareTimeStamp(pa, pa.size());
			if (updatedLevel == 0 && var.isInstantiated()) {
				Integer[] lb = (Integer[]) msg.get(LB);
				processLB(lb, pa, msg.getSenderId());
			}
		}

		public void sendLBMessage(Instantiation pa, VarIdentifier receiverId, Integer[] lb) {
			Message msg = createMessage();
			msg.setReceiverId(receiverId);
			msg.addContents(CONTEXT, pa.clone());
			msg.addContents(LB, lb);
			sendMessage(msg);
		}
	}

	public class BackMessageHandler extends MessageHandler {

		public BackMessageHandler(AbstractAgent agent) {
			super(agent, BACK);
		}

		@Override
		public void processMessage(Message msg) throws StopException {
			PartialAssignement pa = (PartialAssignement) msg.get(CONTEXT); // TODO Why?
			int updatedLevel = compareTimeStamp(pa.getAssignments(), pa.size());
			if (updatedLevel == 0 && var.isInstantiated()) {
				pa.removeLast();
				processBack(pa.getCost()); 
			}
		}

		public void sendBackMessage(VarIdentifier toId, PartialAssignement pa) {
			Message msg = this.createMessage();
			msg.addContents(CONTEXT, pa.clone());
			msg.setReceiverId(toId);
			sendMessage(msg);
		}
	}

	public class SolMessageHandler extends MessageHandler {

		public SolMessageHandler(AbstractAgent agent) {
			super(agent, SOLUTION);
		}

		@Override
		public void processMessage(Message msg) throws StopException {
			processSolution((Instantiation) msg.get(CONTEXT), (Integer) msg.get(COST));
		}

		public void reportSolution(Instantiation pa, Integer ub) {
			HashList<VarIdentifier> agentsIDs = getOrder();
			agentsIDs.remove(getVarIdentifier());
			for(VarIdentifier agentId : agentsIDs) {
				Message msg = createMessage();
				msg.addContents(CONTEXT, pa.clone());
				msg.addContents(COST, ub);
				msg.setReceiverId(agentId);
				sendMessage(msg);
			}
		}
	}

	protected void updateTimeStamp(int index) {
		this.timeStamp.update(index, this.var.getTimeCounter());
	}

	protected void backtrack() throws StopException {
		int j = backtrackTo();
		if (j == -1) {
			if (this.UB == INFINITY) {
				throw new StopException(Commands.FAILURE);
			} else {
				throw new StopException(Commands.SUCCESS);
			}
		}
		VarIdentifier toId = this.agentOrder.getAgentOfRank(j - 1);
		this.backMsgHandler.sendBackMessage(toId, this.partialAssignment.subCPA(j));
		this.agentView.clearFrom(this.agentView.getVarAssignment(toId));
		if (this.var.isInstantiated()){
			this.unassignMyVariable();
		}
	}

	protected int backtrackTo() {
		for(int j = this.myLevel - 1 ; j > 0 ; j--) {
			for(int value : this.currentDomain) {
				int bound = addition(LB(j - 1, value), subtraction(this.PC(j), this.PC(j - 1)));
				if (bound < this.UB) {
					return j;
				}
			}
		}
		return -1;
	}

	private Integer PC(int level) {
		return this.partialAssignment.getCost(level);
	}

	private Integer LB(int level, int value) {
		int idx = this.initialIndexOf(value);
		Integer lb = addition(this.LC[level][idx], this.PC(level));
		return addition(lb, this.sumLowerBounds[level][idx]);
	}

	public int compareTimeStamp(Instantiation pa, int upto) {
		int[] timeStamp = pa.getTimeStamp();
		int[] myTimeStamp = new int[this.myLevel];

		int i = 0;
		for(IntExternalVar v : agentView) {
			myTimeStamp[i++] = v.getTimeCounter();
		}
		myTimeStamp[i] = var.getTimeCounter();

		int j = 0;
		for( ; j < myTimeStamp.length && j < timeStamp.length && j <= upto ; j++) {
			if (myTimeStamp[j] > timeStamp[j]) {
				return -(j + 1);
			} else if (myTimeStamp[j] < timeStamp[j]) {
				return j + 1;
			}
		}
		if (j <= upto && myTimeStamp.length < timeStamp.length) {
			return j + 1;
		}
		return 0;
	}

	/**
	 * This method returns the sum of two costs. 
	 * @param c1 : the first cost.
	 * @param c2 : the second cost. 
	 * @return Integer : the sum of two given costs <tt>c1</tt> and <tt>c2</tt>. 
	 */
	public static Integer addition(Integer c1, Integer c2) {
		return DCOPConstraintHandler.sumCosts(c1, c2);
	}

	/**
	 * This method returns the subtraction of two costs. 
	 * @param c1 : the first cost.
	 * @param c2 : the second cost. 
	 * @return Integer : the subtraction of cost <tt>c2</tt> from cost <tt>c1</tt>. 
	 */
	public static Integer subtraction(Integer c1, Integer c2) {
		return DCOPConstraintHandler.subCosts(c1, c2);
	}

	@Override
	public String solutionToString() {
		String solution = "";
		if (this.isInitializerAgent()) {
			if (this.isFinished()) {
				solution = "Cost=" + this.UB;
			}
			solution += " sol=" + (this.isFinished()? this.bestSolution : this.var.getVarAssignment());
		} else {
			solution = "" + (this.isFinished()? this.bestSolution : this.var.getVarAssignment());
		}
		return solution;
	}

	class PartialAssignement implements Cloneable, Serializable {

		private static final long serialVersionUID = 1L;

		private Instantiation assignments;

		private int[] costs;

		public PartialAssignement(int size) {
			this.setAssignments(new Instantiation());
			this.costs = new int[size];
		}

		/**
		 * @return the assignments
		 */
		public Instantiation getAssignments() {
			return assignments;
		}

		/**
		 * @param assignments the assignments to set
		 */
		public void setAssignments(Instantiation assignments) {
			this.assignments = assignments;
		}

		public Integer getCost(int i) {
			if (i <= 0)
				return 0;
			return costs[i - 1];
		}

		public Integer getCost() {
			if (getAssignments().isEmpty())
				return 0;
			return this.costs[getAssignments().size() - 1];
		}

		public void removeLast() {
			this.costs[getAssignments().size()] = 0;
			this.getAssignments().remove(getAssignments().last());
		}

		public PartialAssignement clone() {
			PartialAssignement clone = new PartialAssignement(this.size());
			clone.setAssignments(this.getAssignments().clone());
			clone.costs = Arrays.copyOf(this.costs, costs.length);
			return clone;

		}

		public Integer size() {
			return getAssignments().size();
		}

		public void add(VarAssignment varAssign, Integer cost) {
			this.getAssignments().add(varAssign);
			this.costs[getAssignments().size() - 1] = cost;
		}

		public String toString() {
			return "Cost: " + this.getCost() + " Assignments:" + getAssignments() + " " + Arrays.toString(this.costs);
		}

		public PartialAssignement subCPA(int k) {
			PartialAssignement cpa = new PartialAssignement(k + 1);
			cpa.setAssignments(this.getAssignments(k));
			cpa.costs = Arrays.copyOf(this.costs, k + 1);
			return cpa;
		}

		public Instantiation getAssignments(Integer k) {
			if (k <= 0) {
				return new Instantiation();
			}
			if (k >= this.assignments.size()) {
				return this.assignments;
			}
			VarAssignment v = this.assignments.get(k);
			return this.assignments.headSet(v);
		}

		public void update(PartialAssignement pa) {
			this.setAssignments(pa.getAssignments());
			for(int i = 0 ; i < pa.costs.length ; i++) {
				this.costs[i] = pa.costs[i];
			}
			for(int i = pa.costs.length ; i < this.costs.length ; i++) {
				this.costs[i] = 0;
			}
		}
	}
}
