/*
DisChoco: Distributed Constraint Reasoning Solver 
Copyright (C) 2007-2013 Redouane Ezzahir & Mohamed Wahbi 

DisChoco is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DisChoco is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

To contact the authors: <http://dischoco.sourceforge.net/>
 */
package simple.dcop.bnbadopt;

import java.math.BigInteger;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import kernel.AbstractAgent;
import kernel.AgentView;
import kernel.StopException;
import kernel.VarIdentifier;
import kernel.com.Message;
import kernel.com.MessageHandler;
import kernel.events.VariableEvent;
import kernel.order.TreeDisAO;
import kernel.vars.integer.IntExternalVar;
import simple.SimpleAgent;
import simple.dcop.DCOPConstraintHandler;
import simple.dcop.adopt.Cost;
import simple.dcsp.afcfamily.TermMsgHandler;
import simulation.metric.Metric;
import simulation.parser.PVariable;
import util.Commands;
import util.HashList;
import util.Instantiation;
import util.Logger;
import util.VarAssignment;
import util.VarEventMask;

/**
 * This class implements the BnB-Adopt algorithm as described in: 
 * BnB-ADOPT: An Asynchronous Branch-and-Bound DCOP Algorithm [Yeoh, et al., 2008]
 * 
 * @author Mohamed Wahbi
 */
public class BnBAdoptPlusAgent extends SimpleAgent {

	protected HashMap<VarIdentifier, ValueMsg> lastValSent = new HashMap<VarIdentifier, ValueMsg>();

	protected HashMap<VarIdentifier, Boolean> lastRcvdThReq = new HashMap<VarIdentifier, Boolean>();

	protected boolean contextChange = false;

	protected Cost lastCostSent = new Cost();

	class ValueMsg {

		int th;
		int value;

		public ValueMsg(VarAssignment varAss, Integer threshold) {
			this.th = threshold;
			this.value = (Integer) varAss.getValue();
		}

		boolean equelsVal(VarAssignment varAss) {
			return this.value == (Integer) varAss.getValue();
		}
	}

	// /////////////////////////////////////////////////////////////////////////BnBAdopt+

	// final static BigInteger BIG_INFINITY = BigInteger.valueOf(Long.MAX_VALUE);
	private static final Integer INFINITY = Integer.MAX_VALUE;

	public static final String COST = Commands.COST;
	public static final String VALUE = Commands.VALUE;
	public static final String THRESHOLD = Commands.THRESHOLD;
	public static final String CONTEXT = Commands.CONTEXT;
	public static final String INFO = Commands.INFO;

	/** upper bound */
	protected Integer UB = INFINITY;
	/** lower bound **/
	protected Integer LB = new Integer(0);

	/** threshold **/
	protected Integer TH = INFINITY;

	/**
	 * A map of costs, there is a {@link Cost} for each value in the domain of current variable 
	 * and for every child : map[value, map[child, cost]] .
	 **/
	// protected Map<Integer, HashMap<Integer, Cost>> childrenCosts = null;

	protected Map<Integer, HashMap<VarIdentifier, Integer>> lowerBounds;
	protected Map<Integer, HashMap<VarIdentifier, Integer>> upperBounds;
	protected Map<Integer, HashMap<VarIdentifier, Instantiation>> contexts;

	/** The handler of VALUE messages */
	protected ValueMsgHandler valueMsgHandler = null;

	/** The handler of COST messages */
	protected CostMsgHandler costMsgHandler = null;

	/** A flag that indicates if parent has terminated */
	protected boolean parentHasTerminated = false;

	protected Map<Integer, Integer> localCost = null;

	public boolean thresholdReq = false;

	/**
	 * @param agentId
	 * @param agName
	 * @param description
	 */
	public BnBAdoptPlusAgent(int agentId, String agName, String description) {
		super(agentId, agName, description);
		this.constructorHelper();
	}

	/**
	 * @param variable
	 */
	public BnBAdoptPlusAgent(PVariable variable) {
		super(variable);
		this.constructorHelper();
	}

	/**
	 * 
	 */
	private void constructorHelper() {
		this.terminateMsgHandler = new TerminateMsgHandler(this);
		this.valueMsgHandler = new ValueMsgHandler(this);
		this.costMsgHandler = new CostMsgHandler(this);
		this.constraintHandler = new DCOPConstraintHandler(this);
		// this.childrenCosts = new HashMap<Integer, HashMap<Integer, Cost>>();
		this.lowerBounds = new HashMap<Integer, HashMap<VarIdentifier, Integer>>();
		this.upperBounds = new HashMap<Integer, HashMap<VarIdentifier, Integer>>();
		this.contexts = new HashMap<Integer, HashMap<VarIdentifier, Instantiation>>();
		this.localCost = new HashMap<Integer, Integer>();
		this.agentOrder = new TreeDisAO(this);
		this.auxiliaryStatistics.put(Commands.SOLUTION_COST, new Metric(Commands.SOLUTION_COST, INFINITY, Metric.min));
	}

	/*
	 * (non-Javadoc)
	 * @see kernel.AbstractAgent#configurePropagationEngine()
	 */
	@Override
	public void configurePropagationEngine() {
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_HIGHER_THAN_LOCAL_VAR).add(VariableEvent.INSTANCIATE);
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_HIGHER_THAN_LOCAL_VAR).add(VariableEvent.DESINSTANCIATE);
		this.eventsConfig.get(VarEventMask.LOCAL_VAR_IS_THE_LAST).add(VariableEvent.RESTORE_VAL);
		this.eventsConfig.get(VarEventMask.LOCAL_VAR_IS_BETWEEN_L_H).add(VariableEvent.RESTORE_VAL);
		this.constraintHandler.update();
	}

	/*
	 * (non-Javadoc)
	 * @see kernel.AbstractAgent#initAgentView()
	 */
	@Override
	public void initAgentView() {
		HashList<VarIdentifier> varIds = this.getAgentOrder().getAncestors();
		for(VarIdentifier varId : varIds) {
			VarAssignment varAss = new VarAssignment(varId, Integer.MIN_VALUE);
			IntExternalVar extVar = this.makeExternalVar(varAss);
			this.agentView.add(extVar);
		}
	}

	@Override
	protected void addMsgHandlers() {
		this.communicator.addMsgHandler(this.valueMsgHandler);
		this.communicator.addMsgHandler(this.costMsgHandler);
	}

	@Override
	public String toString() {
		return super.toString() + " solution= " + this.solutionToString();
	}

	/**
	 * This method returns the agent ordering. 
	 * @return the agent ordering in form of a tree. 
	 */
	protected TreeDisAO getAgentOrder() {
		return ((TreeDisAO) this.agentOrder);
	}

	/**
	 * This method returns the Identifier of the parent of the current agent
	 * @return the ID parent of the current agent
	 */
	protected VarIdentifier getParentId() {
		return getAgentOrder().getParentId();
	}

	/**
	 * This method returns <i>true</i> if the given agent (<i>agId</i>) is the parent of the 
	 * current agent. Otherwise, it returns <i>false</i>. 
	 * 
	 * @param agId the given agent identifier
	 * @return if the given agent is the parent of the current agent on the DFS tree. 
	 */
	protected boolean isMyParent(VarIdentifier agId) {
		return this.getAgentOrder().isMyParent(agId);
	}

	/**
	 * This method checks if the current agent is the root of the tree ordering. 
	 * @return true if the current agent is the root of the tree, otherwise, false.
	 */
	protected boolean amIRoot() {
		return getParentId() == null;
	}

	/**
	 * This method checks if the current agent is the leaf of the tree ordering 
	 * @return true if the current agent is the a leaf of the tree, otherwise, false.
	 */
	protected boolean isLeaf() {
		return this.getChildrenIds().isEmpty();
	}

	/**
	 * This method returns the identifiers of the children of the current agent on the tree ordering. 
	 * 
	 * @return IntArray : the IDs of the children of the this agent. 
	 */
	private HashList<VarIdentifier> getChildrenIds() {
		return getAgentOrder().getChildrenIds();
	}

	/**
	 * This method checks if the parent of the current agent in the tree ordering is assigned in the {@link AgentView}.  
	 * @return true if the parent is assigned, otherwise, false
	 */
	private boolean isMyParentAssigned() {
		VarIdentifier varId = new VarIdentifier(getParentId());
		VarAssignment varAss = this.agentView.getVarAssignment(varId);
		return varAss.isInstantiated();
	}

	/**
	 * @param value
	 * @return
	 */
	private Integer h(int value) {
		return new Integer(0);
	}

	/**
	 * 
	 */
	protected void initializesStructures() {
		Iterator<Integer> it = this.var.getDomain().iterator();
		while (it.hasNext()) {
			int value = it.next();
			this.lowerBounds.put(value, new HashMap<VarIdentifier, Integer>());
			this.upperBounds.put(value, new HashMap<VarIdentifier, Integer>());
			this.contexts.put(value, new HashMap<VarIdentifier, Instantiation>());
			for(VarIdentifier child : this.getChildrenIds()) {
				this.initializesChild(value, child);
				this.contexts.get(value).put(child, agentView.toInstantiation());
			}
		}
		for(VarIdentifier id : this.getAgentOrder().getPseudoDescendants()) {
			lastRcvdThReq.put(id, new Boolean(true));
		}
	}

	/**
	 * This method initializes the lower and upper bounds of the given child for the given value. 
	 * each value value in the current domain for every child
	 */
	protected void initializesChild(int value, VarIdentifier child) {
		this.lowerBounds.get(value).put(child, h(value));
		this.upperBounds.get(value).put(child, INFINITY);
	}

	/**
	 * @param value
	 * @param exChild
	 * @return
	 */
	private Integer sumChildrenLB(int value, VarIdentifier exChild) {
		HashList<VarIdentifier> children = this.getChildrenIds();
		children.remove(exChild);
		Integer sum = new Integer(0);
		for(VarIdentifier child : children) {
			Integer lb = this.lowerBounds.get(value).get(child);
			if (lb.compareTo(INFINITY) >= 0)
				return INFINITY;
			sum = DCOPConstraintHandler.sumCosts(sum, lb);
		}
		return sum;
	}

	/**
	 * 
	 */
	protected void initializesSelf() {
		int myValue = this.computeLowerBound();;
		this.assignValue(myValue);
		this.TH = INFINITY;
		this.thresholdReq = true;
	}

	/**
	 * This method from {@link kernel.AbstractAgent} starts the distributed search
	 * 
	 * @throws StopExeption
	 */
	@Override
	public void startDistributedSearch() throws StopException {
		this.initAssignementOfSCA();
		this.initializesStructures(); // It includes initializesChild(int value, int child)
		this.computeLocalCost();
		this.assignValue(this.var.getInf());
		this.initializesSelf();
		this.backtrack();
	}

	/**
	 * This method initializes the assignments of all pseudo-ancestors agents
	 * SCA is the set of ancestors of self (including its parent) that it or one 
	 * of its descendants is involved in constraint with. 
	 */
	protected void initAssignementOfSCA() {
		for(VarIdentifier varId : this.getAgentOrder().getPseudoAncestors()) {
			if (this.communicator.isNeighbor(varId)) {
				IntExternalVar extVar = this.agentView.getVar(varId);
				VarAssignment varAss = this.agentView.getVarAssignment(varId);
				int value = extVar.getInf();
				if (value == Integer.MAX_VALUE) {
					continue; // Really it is not a neighbors but it is my parent in pseudo tree
				}
				varAss.setValue(value);
				extVar.instantiate(varAss);
				extVar.setTimeCounter(0);
			}
		}
	}

	public void backtrack() throws StopException {

		this.computeLocalCost();
		this.log("## backtrack call : [LB:" + LB + ", UB:" + UB + ", TH:" + TH + "], myLocalCost: " + localCost, Logger.DEFAULT_LEVEL);

		int value = this.computeLowerBound();
		this.computeUpperBound();

		int myValue = this.var.getValue();
		if (this.LB(myValue).compareTo(Math.min(this.TH, UB)) >= 0) {
			this.assignValue(value);
			this.log("changeValue LB(" + myValue + ") > min(TH,UB)=" + Math.min(this.TH, UB) + ", new value=" + value, Logger.DEFAULT_LEVEL);
		}

		if ((this.amIRoot() && this.UB.equals(this.LB)) || this.parentHasTerminated) {
			this.parentHasTerminated = true;
			((TerminateMsgHandler) this.terminateMsgHandler).sendTerminateMessages();
		} else {
			if (!this.isLeaf()) {
				this.sendValueMessages();
			}

			// send cost if I'm not the root and my parent not terminated
			if (!this.amIRoot() && isMyParentAssigned()) {
				this.sendCostMessage();
			}
		}
	}

	/**
	 * @param value
	 */
	protected void assignValue(int value) {
		this.var.instantiate(value);
	}

	/**
	 * 
	 */
	protected void sendValueMessages() {
		this.valueMsgHandler.sendValueMessages();
	}

	/**
	 * 
	 */
	protected void sendCostMessage() {
		this.costMsgHandler.sendCost();
	}

	/**
	 * @param value
	 * @param child
	 * @param lb
	 * @param ub
	 * @param context
	 */
	private void updateBounds(int value, VarIdentifier child, Integer lb, Integer ub, Instantiation context) {
		lb = Math.max(lb, this.lowerBounds.get(value).get(child));
		this.lowerBounds.get(value).put(child, lb);
		ub = Math.min(ub, this.upperBounds.get(value).get(child));
		this.upperBounds.get(value).put(child, ub);
		this.contexts.get(value).put(child, context);
	}

	public void computeLocalCost() {
		for(int value : this.var.getDomain()) {
			Integer cost = ((DCOPConstraintHandler) this.constraintHandler).localCost(value);
			this.localCost.put(value, cost);
		}
	}

	/**
	 * This method computes the Lower Bound <b><i>LB = min{LB(v), forall v in D_i}</i></b>, 
	 * and returns the value that minimizes LB(v). 
	 * 
	 * @return Integer : the value <i>v</i> from the domain with the minimum lower bound LB(v)
	 */
	public int computeLowerBound() {
		int myValue = this.var.getValue(); // Integer.MIN_VALUE; //Empty
		Integer lb = LB(myValue);
		for(int value : this.var.getDomain()) {
			Integer candidate_lb = LB(value);
			if (lb.compareTo(candidate_lb) > 0) {
				lb = candidate_lb;
				myValue = value;
			}
		}
		this.log("ComputeLB: [LB=" + LB + ", TH=" + TH + ", context: " + this.agentView + "] --> value minimizing the LB is: " + myValue
				+ " and the new LB is " + lb + "\n", Logger.FINE_LEVEL);
		this.LB = lb; // update Lower Bound
		return myValue;
	}

	/**
	 * This method computes the Upper Bound <b><i>UB = min{UB(v), forall v in D_i}</i></b>, 
	 * and returns the value that minimizes UB(v). 
	 * 
	 * @return Integer : the value <i>v</i> from the domain with the minimum upper bound UB(v)
	 */
	public int computeUpperBound() {
		int myValue = this.var.getValue(); // Integer.MIN_VALUE; //Empty
		Integer ub = UB(myValue);
		for(int value : this.var.getDomain()) {
			Integer candidate_ub = UB(value);
			if (ub.compareTo(candidate_ub) > 0) {
				ub = candidate_ub;
				myValue = value;
			}
		}
		this.log("ComputeUB: [UB=" + UB + ", TH=" + TH + ", context: " + this.agentView + "] --> value minimizing the UB is: " + myValue
				+ " and the new UB is " + ub + "\n", Logger.FINE_LEVEL);
		this.UB = ub; // update Upper Bound ;
		return myValue;
	}

	/**
	 * This method computes the lower bound <b><i>LB(value)</i></b> of the given value <b><i>value</i></b>. 
	 * The lower bound equals the sum of the local cost of the given value (cost with assignments 
	 * of agents in the context) with the lower bound stored for all children. 
	 * 
	 * @param value : the value to which we want to compute the Lower Bound
	 * @return Integer: LB(value) = localCost(value) + \sum{child in children} lb(value, child);
	 */
	public Integer LB(int value) {
		BigInteger sum = BigInteger.valueOf(this.localCost.get(value));
		assert sum.compareTo(BigInteger.valueOf(0)) >= 0 : "\nProblem the local cost=" + sum + " of value=" + value + " is < 0";
		// make sure we don't overflow when dealing with hard constraints
		if (sum.compareTo(BigInteger.valueOf(INFINITY)) >= 0)
			return INFINITY;
		for(Integer lb : this.lowerBounds.get(value).values()) {
			// make sure we don't overflow when dealing with hard constraints
			if (lb.compareTo(INFINITY) >= 0)
				return INFINITY;
			sum = sum.add(BigInteger.valueOf(lb));
			// make sure we don't overflow when dealing with hard constraints
			if (sum.compareTo(BigInteger.valueOf(INFINITY)) >= 0)
				return INFINITY;
		}
		return sum.intValue();
	}

	/**
	 * This method computes the upper bound <b><i>UB(value)</i></b> of the given value <b><i>value</i></b>. 
	 * The upper bound equals the sum of the local cost of the given value (cost with assignments 
	 * of agents in the context) with the upper bound stored for all children. 
	 * 
	 * @param value : the value to which we want to compute the Upper Bound
	 * @return Integer: UB(value) = localCost(value) + \sum{child in children} ub(value, child);
	 */
	public Integer UB(int value) {
		BigInteger sum = BigInteger.valueOf(this.localCost.get(value));
		assert sum.compareTo(BigInteger.valueOf(0)) >= 0 : "\nProblem the local cost=" + sum + " of value=" + value + " is < 0";
		// make sure we don't overflow when dealing with hard constraints
		if (sum.compareTo(BigInteger.valueOf(INFINITY)) >= 0)
			return INFINITY;
		for(Integer ub : this.upperBounds.get(value).values()) {
			// make sure we don't overflow when dealing with hard constraints
			if (ub.compareTo(INFINITY) >= 0)
				return INFINITY;
			sum = sum.add(BigInteger.valueOf(ub));
			// make sure we don't overflow when dealing with hard constraints
			if (sum.compareTo(BigInteger.valueOf(INFINITY)) >= 0)
				return INFINITY;
		}
		return sum.intValue();
	}

	public class ValueMsgHandler extends MessageHandler {

		public ValueMsgHandler(AbstractAgent agent) {
			super(agent, VALUE);
		}

		public void processMessage(Message msg) throws StopException {
			if (!parentHasTerminated) {
				VarAssignment varAss = (VarAssignment) msg.get(INFO);
				Integer threshold = (Integer) msg.get(THRESHOLD);
				Instantiation view = agentView.toInstantiation();
				contextChange |= agentView.priorityMerge(varAss);
				if (!agentView.compatibleWith(view)) {
					Iterator<Integer> it = var.getDomain().iterator();
					while (it.hasNext()) {
						int value = it.next();
						for(VarIdentifier child : getChildrenIds()) {
							HashList<VarIdentifier> SCPc = getAgentOrder().getPseudoAncestors(child);
							if (SCPc.contains(msg.getSenderId())) {
								initializesChild(value, child);
							}
						}
					}
					initializesSelf();
				}
				if (isMyParent(varAss.getVarIdentifier())) {
					TH = threshold;
				}
				if (communicator.isMailBoxEmpty()) {
					backtrack();
				}
			}
		}

		/** This method sends a VALUE message to each lower neighbor */
		public void sendValueMessages() {
			VarAssignment varAss = var.getVarAssignment();
			HashList<VarIdentifier> lowerNeighbors = ((TreeDisAO) this.myAgent.agentOrder).getPseudoDescendants();
			HashList<VarIdentifier> children = getChildrenIds();
			for(VarIdentifier child : children) {
				int myValue = (Integer) varAss.getValue();
				Integer th = Math.min(TH, UB);
				th = DCOPConstraintHandler.subCosts(th, localCost.get(myValue));
				th = DCOPConstraintHandler.subCosts(th, sumChildrenLB(myValue, child));
				sendValue(child, varAss, th);
			}
			lowerNeighbors.removeAll(children);
			for(VarIdentifier neighbor : lowerNeighbors) {
				sendValue(neighbor, varAss, INFINITY);
			}
		}

		public void sendValue(VarIdentifier neighbor, VarAssignment varAss, Integer threshold) {
			if (lastValSent.containsKey(neighbor) && lastValSent.get(neighbor).equelsVal(varAss) && !lastRcvdThReq.get(neighbor)) {
				return;
			}
			lastValSent.put(neighbor, new ValueMsg(varAss, threshold));
			lastRcvdThReq.put(neighbor, false);
			// ///////////////////////////////////////////////////////////////////BnBAdopt+
			Message msg = createMessage();
			msg.addContents(INFO, varAss);
			msg.addContents(THRESHOLD, threshold);
			msg.setReceiverId(neighbor);
			sendMessage(msg);
		}
	}

	public class CostMsgHandler extends MessageHandler {

		public CostMsgHandler(AbstractAgent agent) {
			super(agent, COST);
		}

		@Override
		public void processMessage(Message msg) throws StopException {
			if (!parentHasTerminated) {
				Instantiation context = (Instantiation) msg.get(Commands.CONTEXT);
				VarAssignment myAss = context.remove(var.getVarIdentifier());
				Instantiation view = agentView.toInstantiation();
				Boolean threReq = (Boolean) msg.get(Commands.THRESHOLD);
				lastRcvdThReq.put(msg.getSenderId(), threReq);
				boolean agentViewUpdated = agentView.priorityMerge(context);
				contextChange |= agentViewUpdated;
				if (!agentView.compatibleWith(view)) {
					Iterator<Integer> it = var.getDomain().iterator();
					while (it.hasNext()) {
						int value = it.next();
						for(VarIdentifier child : getChildrenIds()) {
							HashList<VarIdentifier> SCPc = getAgentOrder().getPseudoAncestors(child);
							for(VarIdentifier id : SCPc) {
								VarIdentifier varId = new VarIdentifier(id);
								if (view.contains(varId) && !agentView.compatible(view.getVarAssignment(varId))) {
									initializesChild(value, child);
								}
							}
						}
					}
				}
				if (agentView.compatibleWith(context)) {
					int myValue = (Integer) myAss.getValue();
					VarIdentifier child = msg.getSenderId();
					Integer lb = (Integer) msg.get(Commands.LB);
					Integer ub = (Integer) msg.get(Commands.UB);
					updateBounds(myValue, child, lb, ub, context);
				}
				if (agentViewUpdated) {// !agentView.compatibleWith(view)) {
					initializesSelf();
				}
				if (communicator.isMailBoxEmpty()) {
					backtrack();
				}
			}
		}

		public void sendCost() {
			Instantiation context = agentView.toInstantiation();
			if (// lastCostSent.getContext().contains(context) &&
				// lastCostSent.getContext().size() == context.size() &&
			lastCostSent.getLb().equals(LB) && lastCostSent.getUb().equals(UB) && !contextChange) {
				return;
			} else {
				lastCostSent.setLb(LB);
				lastCostSent.setUb(UB);
				// lastCostSent.setContext(context.clone());
			}
			Message msg = createMessage();
			msg.addContents(Commands.CONTEXT, context);
			msg.addContents(Commands.LB, LB);
			msg.addContents(Commands.UB, UB);
			msg.addContents(Commands.THRESHOLD, new Boolean(thresholdReq));
			msg.setReceiverId(getParentId());
			sendMessage(msg);
			thresholdReq = false;
			contextChange = false;
		}
	}

	public class TerminateMsgHandler extends TermMsgHandler {

		int myChildrenTerminated = 0;

		public TerminateMsgHandler(AbstractAgent agent) {
			super(agent);
		}

		@Override
		public boolean processTerminationDetection(Message msg) throws StopException {
			if (isMyParent(msg.getSenderId())) {
				parentHasTerminated = true;
				sendTerminateMessages();
			} else {
				if (amIRoot()) {
					auxiliaryStatistics.get(Commands.SOLUTION_COST).value = UB;
					throw new StopException(Commands.SUCCESS);
				} else {
					this.myChildrenTerminated++;
					if (parentHasTerminated && this.myChildrenTerminated == getChildrenIds().size()) {
						this.sendTerminate(getParentId());
					}
				}
			}
			return true;
		}

		public void sendTerminateMessages() {
			HashList<VarIdentifier> children = getChildrenIds();
			if (children.isEmpty()) {
				this.sendTerminate(getParentId());
			} else {
				for(VarIdentifier child : children) {
					this.sendTerminate(child);
				}
			}
		}

		/**
		 * @param receiver
		 */
		private void sendTerminate(VarIdentifier receiver) {
			Message msg = this.createMessage();
			msg.setReceiverId(receiver);
			sendMessage(msg);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public String solutionToString() {
		String solution = this.var.toString();
		if (amIRoot()) {
			return "[LB=" + this.LB + ", UB=" + (this.UB.equals(INFINITY)? "+inf" : this.UB) + "] " + solution;
		}
		return solution;
	}

}
