package simple.dcop.synchbb;

import kernel.AbstractAgent;
import kernel.Constraint;
import kernel.StopException;
import kernel.VarIdentifier;
import kernel.com.Message;
import kernel.com.MessageHandler;
import kernel.events.VariableEvent;
import kernel.vars.integer.IntExternalVar;
import simple.SimpleAgent;
import simple.dcop.DCOPConstraintHandler;
import simple.dcsp.afcfamily.TermMsgHandler;
import simulation.parser.PVariable;
import util.Commands;
import util.HashList;
import util.Instantiation;
import util.VarAssignment;
import util.VarEventMask;

/**
 * Implementation of the improved version of Synchronous Branch & Bound algorithm (SBB) for DCOPs
 * 
 * This is the original SBB algorithm by [Hirayama and Yokoo, 1997], and adapted later to solve DCOPs 
 * as proposed by Meisels (Springer Verlag, 2008), with two additional performance improvements: 
 * - BACKTRACK messages do not include the path
 * - PATH messages strip the first part of the path that has not changed
 * 
 * @version dischoco 2.5.3<br/>
 * simple.dcop.synchbb.SynBBAgent.java
 * @author Mohamed Wahbi
 */
public class SBBAgent extends SimpleAgent {
	public static final String CONTEXT = Commands.CONTEXT;
	/** The type of the backtrack messages */
	public static final String BACKTRACK = Commands.BACKTRACK;
	/** The type of the upper bound (SOLUTION) messages */
	public static final String SOLUTION = Commands.SOLUTION;
	
	/** The type of the messages containing the current partial assignment */
	public static final String CPA = Commands.CPA;
	
	/** The type of the messages broadcast by the last variable containing the current upper bound */
	public static final String COST = Commands.COST; /// @todo Do not include the solution?
	
	protected CPAMsgHandler pathMessageHandler = null;
	protected BackMsgHandler backMessageHandler = null;
	protected SolutionMsgHandler solutionMessageHandler = null;
	
	protected Integer UpperBound = Constraint.INFINITY;
	protected Integer currentCost = new Integer(0);
	protected Instantiation solution = new Instantiation();

	/**
	 * Constructors
	 * 
	 * @param description
	 * @param agentName
	 */
	public SBBAgent(int agentId, String agName, String description) {
		super(agentId, agName, description);
		this.terminateMsgHandler = new TermMsgHandler(this);
		this.constraintHandler = new DCOPConstraintHandler(this);
		// -------------- Messages Handler ------------------//
		this.pathMessageHandler = new CPAMsgHandler(this);
		this.backMessageHandler = new BackMsgHandler(this);
		this.solutionMessageHandler = new SolutionMsgHandler(this);
	}
	
	/**
	 * Constructors
	 * 
	 * @param variable 
	 */
	public SBBAgent(PVariable variable) {
		super(variable);
		this.terminateMsgHandler = new TermMsgHandler(this);
		this.constraintHandler = new DCOPConstraintHandler(this);
		// -------------- Messages Handler ------------------//
		this.pathMessageHandler = new CPAMsgHandler(this);
		this.backMessageHandler = new BackMsgHandler(this);
		this.solutionMessageHandler = new SolutionMsgHandler(this);
	}


	@Override
	public void configurePropagationEngine() {
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_HIGHER_THAN_LOCAL_VAR).add(VariableEvent.INSTANCIATE);
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_HIGHER_THAN_LOCAL_VAR).add(VariableEvent.DESINSTANCIATE);
		this.constraintHandler.update();
	}

	/* (non-Javadoc)
	 * @see kernel.AbstractAgent#addMsgHandlers()
	 */
	@Override
	protected void addMsgHandlers() {
		this.communicator.addMsgHandler(this.pathMessageHandler);
		this.communicator.addMsgHandler(this.backMessageHandler);
		this.communicator.addMsgHandler(this.solutionMessageHandler);
	}

	/* (non-Javadoc)
	 * @see kernel.AbstractAgent#initAgentView()
	 */
	@Override
	public void initAgentView() {
		HashList<VarIdentifier> varIds = this.agentOrder.getHigherAgents();
		for (VarIdentifier varId : varIds) {
			if (!this.communicator.isNeighbor(varId)) {
				VarAssignment varAss = new VarAssignment(varId, Integer.MIN_VALUE);
				IntExternalVar extVar = this.makeExternalVar(varAss);
				this.agentView.add(extVar);
			} else {
				this.agentView.add((IntExternalVar) this.getExternalVar(varId));
			}
		}
	}

	/**
	 * Start distributed search from the initializing agent
	 * 
	 * @throws StopExeption
	 */
	public void startDistributedSearch() throws StopException { 
		if (isInitializerAgent()) {
			this.assignCPA();
		}
	}

	/**
	 * @return
	 */
	private boolean isInitializerAgent() {
		return this.agentOrder.getRankOf(this.getVarIdentifier()) == 0;
	}

	// /////////////////////////////////////////////////////////////////////////////////////////////////
	/**
	 * this function is called when ever we need to instantiate the current partial assignment, if doesn't succeed it launches a backtrack event
	 */
	public void assignCPA() throws StopException {
		if (this.valSelector.hasNextValue()) {
			int myValue = this.valSelector.getNextValue();
			Integer localCost = ((DCOPConstraintHandler) this.constraintHandler).localCost(myValue);
			Integer accumulatedCost = DCOPConstraintHandler.sumCosts(this.currentCost, localCost);
			if (accumulatedCost.compareTo(this.UpperBound) < 0) {
				this.var.instantiate(myValue);
				Instantiation context = this.agentView.toInstantiation();
				context.add(this.var.getVarAssignment());
				VarIdentifier nextAgent = this.getNextAgent();
				if (nextAgent == null) { // A solution is found
					reportSolution(context, accumulatedCost);
				} else { // send forward a PATH message to the next agent
					this.pathMessageHandler.sendPath(nextAgent, context, accumulatedCost);
				}
			} else { //remove the value from the domain
				this.var.removeValue(myValue, VariableEvent.NOCAUSE);
				assignCPA();
			}
		} else {
			this.backtrack();
		}
	}

	/**
	 * @param context
	 * @param accumulatedCost
	 * @throws StopException 
	 */
	private void reportSolution(Instantiation solution, Integer accumulatedCost) throws StopException {
		this.solutionMessageHandler.reportSolution(solution, accumulatedCost);
		this.UpperBound = accumulatedCost;
		this.solution = solution;//remove the value from the domain
		//System.out.println("solution: "+this.UpperBound+" "+solution);
		this.var.removeValue(this.var.getValue(), VariableEvent.NOCAUSE);
		assignCPA();
	}

	private VarIdentifier getNextAgent() {
		VarIdentifier nextAgent = this.agentOrder.getNextOf(this.getVarIdentifier());
		return nextAgent;
	}

	/**
	 * called when contradiction is generated.
	 * 
	 * @throws StopException
	 */
	public void backtrack() throws StopException {	
		VarIdentifier predecessor = this.agentOrder.getPreviousOf(this.getVarIdentifier());
		if (predecessor == null) {
			if(this.UpperBound.equals(Constraint.INFINITY)) {
				throw new StopException(Commands.FAILURE);
			} else {
				throw new StopException(Commands.SUCCESS);
			}
		} else {
			this.backMessageHandler.sendBack(predecessor, agentView.toInstantiation(), this.UpperBound);
			this.agentView.clear();
			this.var.getDomain().reset();
		}
	}

	public void processBack(Instantiation context, Integer cost) throws StopException {
		int myValue = context.getVarAssignment(this.var.getVarIdentifier()).getValue();
		this.var.removeValue(myValue, VariableEvent.NOCAUSE);
		this.assignCPA();
	}

	public void processCPA(Instantiation context, Integer cost) throws StopException {
		this.currentCost = cost;
		this.agentView.priorityMerge(context);
		assignCPA();
	}

	public void processSolution(Instantiation solution, Integer ub) {
		this.UpperBound = ub;
		this.solution  = solution;
	}
	
	public class CPAMsgHandler extends MessageHandler {
		public CPAMsgHandler(AbstractAgent agent) {
			super(agent, CPA);
		}

		@Override
		public void processMessage(Message msg) throws StopException {
			Instantiation context = (Instantiation) msg.get(CPA);
			Integer cost = (Integer) msg.get(COST);
			processCPA(context, cost);
		}
		
		public void sendPath(VarIdentifier receiver, Instantiation cpa, Integer cost) {
			Message msg = this.createMessage();
			msg.addContents(CPA, cpa);
			msg.addContents(COST, cost);
			msg.setReceiverId(receiver);
			sendMessage(msg);
		}

	}

	public class BackMsgHandler extends MessageHandler {
		public BackMsgHandler(AbstractAgent agent) {
			super(agent, BACKTRACK);
		}

		@Override
		public void processMessage(Message msg) throws StopException {
			Instantiation context = (Instantiation) msg.get(CONTEXT);
			Integer cost = (Integer) msg.get(COST);
			processBack(context, cost);
		}
		
		public void sendBack(VarIdentifier receiver, Instantiation context, Integer cost) {
			Message msg = this.createMessage();
			msg.addContents(CONTEXT, context);
			msg.addContents(COST, cost);
			msg.setReceiverId(receiver);
			sendMessage(msg);
		}
	}

	public class SolutionMsgHandler extends MessageHandler {
		public SolutionMsgHandler(AbstractAgent agent) {
			super(agent, SOLUTION);
		}

		@Override
		public void processMessage(Message msg) throws StopException {
			Instantiation solution = (Instantiation) msg.get(CONTEXT);
			Integer ub = (Integer) msg.get(COST);
			processSolution(solution, ub);
		}
		
		public void reportSolution(Instantiation solution, Integer cost) {
			HashList<VarIdentifier> varIds = getOrder();
			varIds.remove(getVarIdentifier());
			for (VarIdentifier varId : varIds) {
				Message msg = this.createMessage();
				msg.addContents(CONTEXT, solution);
				msg.addContents(COST, cost);
				msg.setReceiverId(varId);
				sendMessage(msg);
			}
		}
	}

	/* (non-Javadoc)
	 * @see kernel.AbstractAgent#solutionToString()
	 */
	@Override
	public String solutionToString() {
		String solution = "";
		if (this.isInitializerAgent()) {
			solution = "Cost="+this.UpperBound+" sol="+this.solution.toString();
		}
		return solution;
	}

}
