package simple.dcsp.abtfamily.abt;

import java.util.HashMap;
import kernel.AbstractAgent;
import kernel.AbstractMaster;
import kernel.StopException;
import kernel.VarIdentifier;
import kernel.com.Message;
import kernel.com.TerminateMsgHandler;
import kernel.events.AgentEvent;
import simple.SimpleAgent;
import simple.dcsp.abtfamily.IdleEvent;
import util.BooleanLock;
import util.Commands;
import util.HashList;
import util.Instantiation;
import util.Logger;
import util.VarAssignment;

/**
 * termination detection (Silaghi et al. 2000g)
 * 
 * @author redouane
 * 
 */
public class ABTTermDetection extends TerminateMsgHandler {

	protected boolean amIAnInitialEndAgent = false;
	protected HashMap<VarIdentifier, Instantiation> acceptedMsg;
	protected BooleanLock whenIdleLock = new BooleanLock(false);
	protected Instantiation prevSentAccepted = null;

	public ABTTermDetection(AbstractAgent agent) {
		super(agent);
		this.acceptedMsg = new HashMap<VarIdentifier, Instantiation>();
	}

	/*
	 * (non-Javadoc)
	 * @see com.TerminateMsgHandler#processTerminationDetection(.com.Message)
	 */
	public boolean processTerminationDetection(Message msg) throws StopException {
		Instantiation accepted = (Instantiation) msg.get(Commands.ACCEPTED);
		VarIdentifier senderId = msg.getSenderId();
		if (accepted != null) {
			if (!this.acceptedMsg.containsKey(senderId)) {
				this.acceptedMsg.put(senderId, accepted);
				this.myAgent.log("****** processTerminationDetection: accepted is added: "+accepted+"\n", Logger.DEFAULT_LEVEL);
			} else {
				Instantiation stored = this.acceptedMsg.get(senderId);
				if (accepted.compareTimeCounters(stored) > 0) {
					this.acceptedMsg.put(senderId, accepted);
					this.myAgent.log("****** processTerminationDetection: accepted is added: "+accepted+"\n", Logger.DEFAULT_LEVEL);
				} else {
					this.myAgent.log("****** processTerminationDetection: accepted is obsolete: "+accepted+"\n", Logger.DEFAULT_LEVEL);
				}
			}
		} else {
			this.myAgent.log("****** processTerminationDetection: accepted is null\n", Logger.DEFAULT_LEVEL);
		}
		return true;
	}

	/*
	 * (non-Javadoc)
	 * @see com.TerminateMsgHandler#handleAgentEvent(events.AgentEvent )
	 */
	@Override
	public void handleAgentEvent(AgentEvent agentEvent) {
		if (agentEvent instanceof IdleEvent && !this.myAgent.isFinished()) {
			this.tryToSendAccepted();
		}
		if (agentEvent instanceof kernel.events.AgentFinishedEvent) {
			this.askMasterToTerminate(true);
		}
	}

	/**
	 * 
	 */
	public void tryToSendAccepted() {
		this.myAgent.log("****** processTerminationDetection: tryToSendAccepted!\n", Logger.DEFAULT_LEVEL);
		if (this.allAcceptedMsgHaveBeenReceived()) {
			Instantiation accepted = this.BuildAcceptedIfAllCompatible();
			if (accepted != null) {
				if (this.prevSentAccepted != null && this.prevSentAccepted.isSubSetOf(accepted)) {
					return;
				}
				this.prevSentAccepted = accepted;
				HashList<VarIdentifier> highNeighbors = this.myAgent.communicator.getHigherNeighbors();
				for(VarIdentifier varId : highNeighbors) {
					this.sendAccepted(accepted, varId);
				}
				if (this.amIAnInitialEndAgent()) {
					this.sendAccepted(accepted, AbstractMaster.MasterVarId);
				}
			}
		} else {
			this.myAgent.log("****** processTerminationDetection: all accepted msgs have not been received!\n" + 
					"LowerNeighbors: " + this.myAgent.communicator.getLowerNeighbors() + "\n" + 
					"Received: " + acceptedMsg.keySet() + "\n", Logger.DEFAULT_LEVEL);
		}
	}

	protected boolean amIAnInitialEndAgent() {
		return this.amIAnInitialEndAgent;
	}

	protected Instantiation BuildAcceptedIfAllCompatible() {
		Instantiation union = this.myAgent.agentView.getContextsOfHigherAgents();
		VarAssignment mySol = ((SimpleAgent) this.myAgent).var.getVarAssignment();
		union.add(mySol);
		this.myAgent.log("mySol:" + mySol, Logger.INFO_LEVEL);
		for(Instantiation accepted : this.acceptedMsg.values()) {
			if (union.compatible(accepted)) {
				union.priorityMerge(accepted);
			} else {
				this.myAgent.log("****** processTerminationDetection: BuildAcceptedIfAllCompatible (not compatible)" + union + "\nwith" + accepted+"\n", Logger.INFO_LEVEL);
				return null;
			}
		}
		return union;
	}

	protected void sendAccepted(Instantiation accepted, VarIdentifier receiverID) {
		Message msg = this.createMessage();
		msg.addContents(Commands.ACCEPTED, accepted.clone());
		msg.setReceiverId(receiverID);
		this.myAgent.sendMessage(msg);
	}

	protected synchronized boolean allAcceptedMsgHaveBeenReceived() {
		HashList<VarIdentifier> lowerNeighbors = this.myAgent.communicator.getLowerNeighbors();
		return acceptedMsg.keySet().containsAll(lowerNeighbors);
	}

	public void sendInitialEndAgentMessage() {
		HashList<VarIdentifier> highNeighbors = this.myAgent.communicator.getHigherNeighbors();
		this.amIAnInitialEndAgent = (highNeighbors.size() == 0);
		Message msg = this.createMessage();
		msg.addContents(Commands.INITIAL_END_AGENT, new Boolean(this.amIAnInitialEndAgent));
		msg.setReceiverId(AbstractMaster.MasterVarId);
		this.myAgent.sendMessage(msg);
	}

}
