package simple.dcsp.abtfamily.abtdo;

import java.util.HashMap;
import kernel.AbstractMaster;
import kernel.MasterObserver;
import kernel.VarIdentifier;
import kernel.com.Message;
import util.Commands;
import util.HashList;
import util.Instantiation;
import util.Logger;

/**
 * termination detection (Silaghi et al. 2000)
 * 
 * @authors 
 * <a href="mailto:ezzahir@ensa-agadir.ac.ma?subject=DisChoco &cc=wahbi@lirmm.fr &cc=bessiere@lirmm.fr">
 * Redoaune Ezzahir</a> and 
 * <a href="mailto:wahbi@lirmm.fr?subject=DisChoco &cc=ezzahir@ensa-agadir.ac.ma &cc=bessiere@lirmm.fr">
 * Mohamed Wahbi</a>
 *
 */
public class TerminationDetector extends MasterObserver {

	protected HashMap<VarIdentifier, Instantiation> acceptedMsg;
	protected HashMap<VarIdentifier, CurrentOrder> currentOrders = null;
	protected HashList<VarIdentifier> InitEndAgents = null;
	protected CurrentOrder bestOrder = null;

	public TerminationDetector() {
		super();
		this.acceptedMsg = new HashMap<VarIdentifier, Instantiation>();
		this.currentOrders = new HashMap<VarIdentifier, CurrentOrder>();
		this.InitEndAgents = new HashList<VarIdentifier>();
	}

	@Override
	public boolean handleReceivedMessage(Message msg) {
		if (msg.getType().equals(Commands.ORDER)) {
			this.ProcessOrder(msg);
			this.checkEnd();
		}
		if (msg.getType().equals(Commands.TERMINATION_DETECTION)) {
			Instantiation accepted = (Instantiation) msg.get(Commands.ACCEPTED);
			CurrentOrder rcvOrder = (CurrentOrder) msg.get(Commands.ORDER);
			VarIdentifier senderId = (VarIdentifier) msg.get(Commands.ID);
			this.setBestOrder(rcvOrder);
			int cmp = this.bestOrder.compare(rcvOrder);
			if (cmp <= 0) {
				this.currentOrders.put(senderId, rcvOrder);
				Instantiation stored = this.acceptedMsg.get(senderId);
				if (accepted.compareTimeCounters(stored) > 0) {
					this.acceptedMsg.put(senderId, accepted);
				}
			} else {
				if (cmp > this.bestOrder.getOrder().indexOf(senderId) + 1) {
					if (this.currentOrders.containsKey(senderId)) {
						CurrentOrder senderOrder = this.currentOrders.get(senderId);
						if (senderOrder.compare(rcvOrder) < 0) {
							this.currentOrders.put(senderId, rcvOrder);
						}
						Instantiation stored = this.acceptedMsg.get(senderId);
						if (accepted.compareTimeCounters(stored) > 0) {
							this.acceptedMsg.put(senderId, accepted);
						}
					} else {
						this.currentOrders.put(senderId, rcvOrder);
						this.acceptedMsg.put(senderId, accepted);
					}
				}
			}
			this.checkEnd();
		}
		return true;
	}

	private void ProcessOrder(Message msg) {
		this.setBestOrder((CurrentOrder) msg.get(Commands.ORDER));
	}

	private HashList<VarIdentifier> computeInitialEndAgents() {
		this.InitEndAgents = new HashList<VarIdentifier>();
		for (VarIdentifier agId : this.bestOrder.getOrder()) {
			if (this.GammaPlus(agId).size() == 0) {
				this.InitEndAgents.add(agId);
			}
		}
		return this.InitEndAgents;
	}

	private HashList<VarIdentifier> GammaPlus(VarIdentifier varId) {
		HashList<VarIdentifier> higherAgents = this.bestOrder.getOrder().getElementsBefor(varId);
		HashList<VarIdentifier> neighbours = this.master.getNeighbors(varId);
		return HashList.intersection(higherAgents, neighbours);
	}

	private boolean checkEnd() {
		this.computeInitialEndAgents();
		if (!(currentOrders.keySet().containsAll(InitEndAgents) && this.acceptedMsg.keySet().containsAll(InitEndAgents))) {
			this.master.log("Not yet receive fromall IEAgents\n"+
						"IEAgendts     : " + this.InitEndAgents + "\n"+
						"bestOrder     : " + this.bestOrder + "\n"+
						"currentOrders : " + this.currentOrders + "\n", Logger.INFO_LEVEL);
			return false;
		}
		this.master.log("\nI check the end : " + "\n"+
					"IEAgendts 	   : " + this.InitEndAgents + "\n"+
					"bestOrder 	   : " + this.bestOrder + "\n"+
					"acceptedMsg   : " + this.acceptedMsg + "\n"+
					"currentOrders : " + this.currentOrders + "\n", Logger.INFO_LEVEL);
		Instantiation solution = new Instantiation();
		for (VarIdentifier varId : this.InitEndAgents) {
			if (!this.currentOrders.containsKey(varId) || !this.acceptedMsg.containsKey(varId)) {
				if (Logger.LOG_DEBUG) {
					this.master.log("==>  orders or acceptedMsg do not contains " + varId, 1);
				}
				return false;
			}
			int cmp = this.bestOrder.compare(this.currentOrders.get(varId));
			if (cmp != 0 && Math.abs(cmp) <= this.bestOrder.getOrder().indexOf(varId) + 1) {
				this.master.log("==>  problem in orders \n" + varId + " => " + this.currentOrders.get(varId) + " and \n" + this.bestOrder, 1);
				return false;
			}
			Instantiation accepted = this.acceptedMsg.get(varId);
			if (solution.compatible(accepted)) {
				solution.priorityMerge(accepted);
				
			} else {
				this.master.log("non compatible accepted " + solution + " \n" + accepted, Logger.INFO_LEVEL);
				return false;
			}
		}
		for (VarIdentifier varId : this.InitEndAgents) {
			Instantiation accepted = this.acceptedMsg.get(varId);
			if(solution.compareTimeCounters(accepted) != 0){
				this.master.log("different counters: " + solution + " \n" + accepted, Logger.INFO_LEVEL);
				return false;
			}
		}
		this.master.log("terminate with SUCCESS : " + solution, Logger.INFO_LEVEL);
		this.master.setEndState(Commands.SUCCESS);
		this.master.broadcastSystemMessage(AbstractMaster.MasterVarId);
		return true;
		// System.err.println(this.master.getEndState());
		// System.exit(0);

	}

	@Override
	public void handleSendMessage(Message msg) {}

	public void setBestOrder(CurrentOrder currentOrder) {
		if (this.bestOrder == null) {
			this.bestOrder = currentOrder;
		} else {
			if (this.bestOrder.compare(currentOrder) < 0) {
				this.bestOrder = currentOrder;
			}
		}
	}

	public String toString() {
		String term = "";
		term = "Termination detection  of :" + this.master.getModel();
		return term;
	}
}
