package simple.dcsp.afcfamily.afc;

import kernel.StopException;
import kernel.VarIdentifier;
import kernel.com.Message;
import kernel.com.MessageHandler;
import kernel.events.VariableEvent;
import kernel.vars.integer.IntExternalVar;
import simple.SimpleAgent;
import simple.SingleExplainedVarAgent;
import simple.dcsp.Nogood;
import simple.dcsp.afcfamily.TermMsgHandler;
import simulation.metric.Metric;
import simulation.parser.PVariable;
import util.Commands;
import util.HashList;
import util.Instantiation;
import util.VarAssignment;
import util.VarEventMask;

public class AFCAgent extends SingleExplainedVarAgent {

	public static final String FC_CPA = Commands.FC_CPA;
	public static final String BACK_CPA = Commands.BACKTRACK;
	public static final String OBSOLETE_CPA = "obsoleteCPA";
	public static final String SC = "SC";
	public static final String CPA = Commands.CPA;
	public static final String NOT_OK = Commands.NOT_OK;

	protected long stepCounter = 0L;
	protected CPAMsgHandler cpaMsgH;
	protected FCCPAMsgHandler fcCPAmsgH;
	protected BackCPAMsgHandler backCPAMsgH;
	protected NotOKMsgHandler notOkMsgH;

	protected boolean agentViewConsistent = true;

	public AFCAgent(int agId, String name, String description) {
		super(agId, name, description);
		this.terminateMsgHandler = new TermMsgHandler(this);
		this.cpaMsgH = new CPAMsgHandler(this);
		this.fcCPAmsgH = new FCCPAMsgHandler(this);
		this.backCPAMsgH = new BackCPAMsgHandler(this);
		this.notOkMsgH = new NotOKMsgHandler(this);
		this.initAuxStatistics();
	}
	
	public AFCAgent(PVariable variable) {
		super(variable);
		this.terminateMsgHandler = new TermMsgHandler(this);
		this.cpaMsgH = new CPAMsgHandler(this);
		this.fcCPAmsgH = new FCCPAMsgHandler(this);
		this.backCPAMsgH = new BackCPAMsgHandler(this);
		this.notOkMsgH = new NotOKMsgHandler(this);
		this.initAuxStatistics();
	}

	private void initAuxStatistics() {
		this.auxiliaryStatistics.put(Commands.nbObsoleteNogood, new Metric(Commands.nbObsoleteNogood));
		this.auxiliaryStatistics.put(Commands.nbAssignments, new Metric(Commands.nbAssignments));
		this.auxiliaryStatistics.put(OBSOLETE_CPA, new Metric(OBSOLETE_CPA));
	}

	// ===========================================================================
	// ///////////////// Initialization ////////////////////////////
	// ===========================================================================

	@Override
	public void configurePropagationEngine() {
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_HIGHER_THAN_LOCAL_VAR).add(VariableEvent.INSTANCIATE);
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_HIGHER_THAN_LOCAL_VAR).add(VariableEvent.DESINSTANCIATE);
		this.eventsConfig.get(VarEventMask.LOCAL_VAR_IS_THE_LAST).add(VariableEvent.RESTORE_VAL);
		this.eventsConfig.get(VarEventMask.LOCAL_VAR_IS_BETWEEN_L_H).add(VariableEvent.RESTORE_VAL);
		this.constraintHandler.update();
	}

	// ==========================================================================

	@Override
	public void addMsgHandlers() {
		this.communicator.addMsgHandler(this.cpaMsgH);
		this.communicator.addMsgHandler(this.fcCPAmsgH);
		this.communicator.addMsgHandler(this.backCPAMsgH);
		this.communicator.addMsgHandler(this.notOkMsgH);
	}

	// ==========================================================================

	public void initAgentView() {
		HashList<VarIdentifier> varIds = this.agentOrder.getHigherAgents();
		for (VarIdentifier varId : varIds) {
			if (!this.communicator.isNeighbor(varId)) {
				VarAssignment var = new VarAssignment(varId, Integer.MIN_VALUE);
				IntExternalVar extVar = this.makeExternalVar(var);
				this.agentView.add(extVar);
			} else {
				this.agentView.add((IntExternalVar) this.getExternalVar(varId));
			}
		}
	}

	// ==========================================================================
	// /////////////////////////// AFC Procedure
	// /////////////////////////////////
	// ==========================================================================

	/**
	 * Start distributed search from the initializing agent
	 * 
	 * @throws StopExeption
	 */
	public void startDistributedSearch() throws StopException {
		if (this.agentOrder.getMyRank() == 0) {
			this.AssignCPA();
		}
	}

	// ==========================================================================

	/**
	 * this function is called when ever we need to instantiate the current partial assignment, if doesn't succeed it launches a this.Backtrack
	 * procedure
	 */
	public void AssignCPA() throws StopException {
		if (this.valSelector.hasNextValue()) {
			int value = this.valSelector.getNextValue();
			this.var.instantiate(value);
			Instantiation CPA = this.agentView.toInstantiation();
			CPA.add(this.var.getVarAssignment());
			this.stepCounter++;
			this.sendCPA(CPA);
		} else {
			this.agentViewConsistent = false;
			this.agentView.clearExclusiveFrom(this.solveMyNogoods().getRhs());
			this.Backtrack();
		}
	}

	// ==========================================================================

	public void sendCPA(Instantiation cpa) throws StopException {
		VarIdentifier nextAgent = this.getNextAgent();
		if (nextAgent == null) { throw new StopException(Commands.SUCCESS); }
		this.cpaMsgH.sendCPAMessage(cpa, nextAgent, this.stepCounter);
		this.fcCPAmsgH.sendFC_CPAmessages(cpa, nextAgent, this.stepCounter);
	}

	public VarIdentifier getNextAgent() {
		return this.agentOrder.getNextOf(this.getVarIdentifier());
	}

	public void Backtrack() throws StopException {
		if (this.agentOrder.getMyRank() == 0) { throw new StopException(Commands.FAILURE); }
		this.agentViewConsistent = false;
		Instantiation cpa = this.agentView.toInstantiation();
		VarIdentifier backTo = cpa.last().getVarIdentifier();
		this.backCPAMsgH.sendBack(cpa, backTo, this.stepCounter);
		if (this.var.isInstantiated()) {
			this.var.desinstanciate(VariableEvent.NOCAUSE);
		}

	}

	// ==========================================================================

	protected void receiveCPA(Instantiation cpa, long sc, boolean fromBackMsg) throws StopException {
		if (sc > this.stepCounter) {
			this.stepCounter = sc;
		}
		if (!agentViewConsistent) {
			Instantiation view = agentView.toInstantiation();
			if (cpa.contains(view)) {
				this.Backtrack();
			} else {
				this.agentViewConsistent = true;
			}
		}
		if (agentViewConsistent) {
			if (fromBackMsg) {
				this.makeNoGoodConstraint(new Nogood(cpa));
				this.AssignCPA();
			} else {
				if (this.updateAgentView(cpa)) {
					this.AssignCPA();
				} else {
					this.Backtrack();
				}
			}
		}
	}

	// ==========================================================================
	public void forwardCheck(Instantiation cpa, long sc) {
		if (sc > this.stepCounter) {
			this.stepCounter = sc;
			if (!this.agentViewConsistent) {
				Instantiation view = agentView.toInstantiation();
				if (!cpa.contains(view)) {
					this.agentViewConsistent = true;
				}
			}
			if (this.agentViewConsistent) {
				if (!this.updateAgentView(cpa)) {
					this.sendNotOk();
				}
			}
		} else {
			this.auxiliaryStatistics.get(OBSOLETE_CPA).value++;
		}

	}

	// ==========================================================================

	public void sendNotOk() {
		Instantiation view = this.agentView.toInstantiation();
		this.notOkMsgH.sendNotOks(view, this.stepCounter);
	}

	public void processNotOk(Instantiation notOk, long sc) {
		Instantiation view = this.agentView.toInstantiation();
		if (view.contains(notOk)) {
			this.agentView.clearExcept(notOk);
			this.agentView.priorityMerge(notOk);
			this.agentViewConsistent = false;
		} else if (!notOk.contains(view)) {
			if (sc > this.stepCounter) {
				this.stepCounter = sc;
				this.agentView.clearExcept(notOk);
				this.agentView.priorityMerge(notOk);
				this.agentViewConsistent = false;
			}
		}
	}

	// ===========================================================================
	public boolean updateAgentView(Instantiation cpa) {
		this.agentView.clearExcept(cpa);
		for (VarAssignment v : cpa) {
			this.agentView.priorityMerge(v);
			if (var.getDomainSize() == 0) { return false; }
		}
		return true;
	}

	// ===========================================================================
	public class CPAMsgHandler extends MessageHandler {

		public CPAMsgHandler(SimpleAgent algo) {
			super(algo, CPA);
		}

		public void processMessage(Message msg) throws StopException {
			Instantiation cpa = (Instantiation) msg.get(CPA);
			long sc = (Long) msg.get(SC);
			AFCAgent.this.receiveCPA(cpa, sc, false);
		}

		public void sendCPAMessage(Instantiation cpa, VarIdentifier receiverId, long sc) {
			Message msg = createMessage();
			msg.addContents(CPA, cpa.clone());
			msg.addContents(SC, sc);
			msg.setReceiverId(receiverId);
			sendMessage(msg);
		}

	}

	// ========================================================================
	public class FCCPAMsgHandler extends MessageHandler {
		
		public FCCPAMsgHandler(AFCAgent agent) {
			super(agent, FC_CPA);
		}

		@Override
		public void processMessage(Message msg) throws StopException {
			Instantiation cpa = (Instantiation) msg.get(FC_CPA);
			long sc = (Long) msg.get(SC);
			AFCAgent.this.forwardCheck(cpa, sc);
		}
		/**
		 * send the FC_CPA to unassigned neighbors 
		 * @param cpa
		 * @param nextAgent
		 * @param sc
		 */
		public void sendFC_CPAmessages(Instantiation cpa, VarIdentifier nextAgent, long sc) {
			HashList<VarIdentifier> unassignedNeighbors = communicator.getNeighbors();
			unassignedNeighbors.removeAll(cpa.getVarIds());
			if(unassignedNeighbors.contains(nextAgent)){
				unassignedNeighbors.remove(nextAgent);
			}
			for (VarIdentifier varId : unassignedNeighbors) {
				Message msg = createMessage();
				msg.addContents(FC_CPA, cpa.clone());
				msg.addContents(SC, sc);
				msg.setReceiverId(varId);
				sendMessage(msg);
			}
		}
	}

	// ==========================================================================
	public class NotOKMsgHandler extends MessageHandler {

		public NotOKMsgHandler(AFCAgent agent) {
			super(agent, NOT_OK);
		}

		@Override
		public void processMessage(Message msg) throws StopException {
			Instantiation notOk = (Instantiation) msg.get(NOT_OK);
			long sc = (Long) msg.get(SC);
			AFCAgent.this.processNotOk(notOk, sc);
		}

		public void sendNotOks(Instantiation notOk, long sc) { //TODO
			HashList<VarIdentifier> allUnassigned = agentOrder.getAgentsLowerThan(notOk.last().getVarIdentifier());
			allUnassigned.remove(getVarIdentifier());
			for (VarIdentifier varId : allUnassigned) {
				this.sendNotOk(notOk.clone(), varId, sc);
			}
		}

		private void sendNotOk(Instantiation notOk, VarIdentifier varId, long sc) {
			Message msg = this.createMessage();
			msg.addContents(NOT_OK, notOk);
			msg.addContents(SC, sc);
			msg.setReceiverId(varId);
			sendMessage(msg);
		}
	}

	// ==============================================================================
	public class BackCPAMsgHandler extends MessageHandler {

		public BackCPAMsgHandler(AFCAgent agent) {
			super(agent, BACK_CPA);
		}

		@Override
		public void processMessage(Message msg) throws StopException {
			Instantiation cpa = (Instantiation) msg.get(CPA);
			long sc = (Long) msg.get(SC);
			AFCAgent.this.receiveCPA(cpa, sc, true);
		}

		public void sendBack(Instantiation cpa, VarIdentifier backTo, long sc) {
			Message msg = this.createMessage();
			msg.addContents(CPA, cpa);
			msg.addContents(SC, sc);
			msg.setReceiverId(backTo);
			sendMessage(msg);
		}
	}
	// ===============================================================================

}
