package simple.dcsp.afcfamily.afc;

import kernel.AbstractAgent;
import kernel.StopException;
import kernel.VarIdentifier;
import kernel.com.Message;
import kernel.com.MessageHandler;
import kernel.events.VariableEvent;
import kernel.vars.integer.IntExternalVar;
import simple.SimpleAgent;
import simple.SingleExplainedVarAgent;
import simple.dcsp.Nogood;
import simple.dcsp.afcfamily.TermMsgHandler;
import simulation.metric.Metric;
import simulation.parser.PVariable;
import util.Commands;
import util.HashList;
import util.Instantiation;
import util.VarAssignment;
import util.VarEventMask;

public class AFCNRAgent extends SingleExplainedVarAgent {

	public static final String OBSOLETE_CPA = "obsoleteCPA";
	public static final String SC = "SC";

	protected long stepCounter = 0L;
	protected CPAMsgHandler cpaMsgH;
	protected FCCPAMsgHandler fcCPAmsgH;
	protected BackCPAMsgHandler backCPAMsgH;
	protected NotOKMsgHandler notOkMsgH;

	protected boolean agentViewConsistent = true;

	Nogood nokNogood = null;

	public AFCNRAgent(int agentId, String agName, String description) {
		super(agentId, agName, description);
		this.terminateMsgHandler = new TermMsgHandler(this);
		this.cpaMsgH = new CPAMsgHandler(this);
		this.fcCPAmsgH = new FCCPAMsgHandler(this);
		this.backCPAMsgH = new BackCPAMsgHandler(this);
		this.notOkMsgH = new NotOKMsgHandler(this);
		this.initAuxStatistics();
	}

	public AFCNRAgent(PVariable variable) {
		super(variable);
		this.terminateMsgHandler = new TermMsgHandler(this);
		this.cpaMsgH = new CPAMsgHandler(this);
		this.fcCPAmsgH = new FCCPAMsgHandler(this);
		this.backCPAMsgH = new BackCPAMsgHandler(this);
		this.notOkMsgH = new NotOKMsgHandler(this);
		this.initAuxStatistics();
	}

	private void initAuxStatistics() {
		this.auxiliaryStatistics.put(Commands.nbObsoleteNogood, new Metric(Commands.nbObsoleteNogood));
		this.auxiliaryStatistics.put(Commands.nbAssignments, new Metric(Commands.nbAssignments));
		this.auxiliaryStatistics.put(OBSOLETE_CPA, new Metric(OBSOLETE_CPA));
	}

	// ===========================================================================
	// ///////////////// Initialization
	// //////////////////////////////////////////
	// ==========================================================================
	@Override
	public void configurePropagationEngine() {
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_HIGHER_THAN_LOCAL_VAR).add(VariableEvent.INSTANCIATE);
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_HIGHER_THAN_LOCAL_VAR).add(VariableEvent.DESINSTANCIATE);
		this.eventsConfig.get(VarEventMask.LOCAL_VAR_IS_THE_LAST).add(VariableEvent.RESTORE_VAL);
		this.eventsConfig.get(VarEventMask.LOCAL_VAR_IS_BETWEEN_L_H).add(VariableEvent.RESTORE_VAL);
		this.constraintHandler.update();
	}

	// ==========================================================================

	@Override
	public void addMsgHandlers() {
		this.communicator.addMsgHandler(this.cpaMsgH);
		this.communicator.addMsgHandler(this.fcCPAmsgH);
		this.communicator.addMsgHandler(this.backCPAMsgH);
		this.communicator.addMsgHandler(this.notOkMsgH);
	}

	// ==========================================================================
	public void initAgentView() {
		HashList<VarIdentifier> varIds = this.agentOrder.getHigherAgents();
		for (VarIdentifier varId : varIds) {
			if (!this.communicator.isNeighbor(varId)) {
				VarAssignment var = new VarAssignment(varId, Integer.MIN_VALUE);
				IntExternalVar extVar = this.makeExternalVar(var);
				this.agentView.add(extVar);
			} else {
				this.agentView.add((IntExternalVar) this.getExternalVar(varId));
			}
		}
	}

	// ==========================================================================
	// //////////////// AFC Procedure ////////////////
	// ==========================================================================

	/**
	 * Start distributed search from the initializing agent
	 * 
	 * @throws StopExeption
	 */
	public void startDistributedSearch() throws StopException {
		if (this.agentOrder.getMyRank() == 0) {
			this.AssignCPA();
		}
	}

	// ==========================================================================

	/**
	 * this function is called when ever we need to instantiate the current partial assignment, if doesn't succeed it launches a this.Backtrack
	 * procedure
	 */
	public void AssignCPA() throws StopException {
		if (this.valSelector.hasNextValue()) {
			int value = this.valSelector.getNextValue();
			this.var.instantiate(value);
			Instantiation CPA = this.agentView.toInstantiation();
			CPA.add(this.var.getVarAssignment());
			this.stepCounter++;
			this.sendCPA(CPA);
		} else {
			this.agentViewConsistent = false;
			this.Backtrack();
		}
	}

	// ==========================================================================

	public void sendCPA(Instantiation cpa) throws StopException {
		VarIdentifier nextVarId = this.getNextAgent();
		if (nextVarId == null) { throw new StopException(Commands.SUCCESS); }
		this.cpaMsgH.sendCPAMessage(cpa, nextVarId, this.stepCounter);
		this.fcCPAmsgH.sendFC_CPAmessages(cpa, nextVarId, this.stepCounter);
	}

	public VarIdentifier getNextAgent() {
		return this.agentOrder.getNextOf(this.getVarIdentifier());
	}

	public void Backtrack() throws StopException {
		if (this.agentOrder.getMyRank() == 0) { throw new StopException(Commands.FAILURE); }
		Nogood nogood = null;
		if (this.nokNogood != null) {
			nogood = this.nokNogood;
		} else {
			nogood = this.solveMyNogoods();
		}
		if (nogood.isEmpty()) { throw new StopException(Commands.FAILURE); }
		this.agentView.clearExclusiveFrom(nogood.getRhs());
		Instantiation cpa = this.agentView.toInstantiation();
		VarIdentifier backTo = cpa.last().getVarIdentifier();
		assert (nogood.getRhsVarIdentifier().equals(backTo));
		this.backCPAMsgH.sendBack(cpa, backTo, nogood, this.stepCounter);
		if (this.var.isInstantiated())
			this.var.desinstanciate(VariableEvent.NOCAUSE);
		this.agentViewConsistent = false;
	}

	// ===========================================================================
	public boolean updateAgentView(Instantiation cpa) {
		this.agentView.clearExcept(cpa);
		for (VarAssignment v : cpa) {
			this.agentView.priorityMerge(v);
			if (var.getDomainSize() == 0) { return false; }
		}
		return true;
	}

	// ===========================================================================
	// ===========================================================================

	protected void receiveCPA(Instantiation cpa, long sc, boolean fromBackMsg) throws StopException {
		if (sc > this.stepCounter) {
			this.stepCounter = sc;
		}
		if (!agentViewConsistent) {
			Instantiation view = agentView.toInstantiation();
			if (cpa.contains(view)) {
				this.Backtrack();
			} else {
				this.nokNogood = null;
				this.agentViewConsistent = true;
			}
		}
		if (agentViewConsistent) {
			if (this.updateAgentView(cpa)) {
				this.AssignCPA();
			} else {
				this.Backtrack();
			}
		}
	}

	// ==========================================================================
	public void forwardCheck(Instantiation cpa, long sc) throws StopException {
		if (sc > this.stepCounter) {
			this.stepCounter = sc;
			if (!this.agentViewConsistent) {
				Instantiation view = agentView.toInstantiation();
				if (!cpa.contains(view)) {
					this.nokNogood = null;
					this.agentViewConsistent = true;
				}
			}
			if (this.agentViewConsistent) {
				if (!this.updateAgentView(cpa)) {
					this.sendNotOk();
					this.agentViewConsistent = false;
				}
			}
		} else {
			this.auxiliaryStatistics.get(OBSOLETE_CPA).value++;
		}

	}

	// ==========================================================================
	public void sendNotOk() throws StopException {
		Nogood nogood = this.solveMyNogoods();
		if (nogood.isEmpty()) { throw new StopException(Commands.FAILURE); }
		this.notOkMsgH.sendNotOks(this.agentView.toInstantiation(), nogood, this.stepCounter);
	}

	public void processNotOk(Instantiation notOk, Nogood nogood, long sc) {
		Instantiation view = this.agentView.toInstantiation();
		if (view.contains(notOk)) {
			this.nokNogood = nogood;
			this.agentView.clearExcept(notOk);
			this.agentView.priorityMerge(notOk);
			this.var.desinstanciate(VariableEvent.NOCAUSE);
			this.agentViewConsistent = false;
		} else if (!notOk.contains(view)) {
			if (sc > this.stepCounter) {
				this.stepCounter = sc;
				this.nokNogood = nogood;
				this.agentView.clearExcept(notOk);
				this.agentView.priorityMerge(notOk);
				this.agentViewConsistent = false;
				this.var.desinstanciate(VariableEvent.NOCAUSE);
			}
		}
	}

	public void receiveBackCPA(Instantiation cpa, Nogood nogood, long sc, boolean b) throws StopException {
		assert (agentViewConsistent && nogood.getRhsVarIdentifier().equals(getVarIdentifier()));
		if (sc > this.stepCounter) {
			this.stepCounter = sc;
		}
		this.makeNoGoodConstraint(nogood);
		this.AssignCPA();
	}

	// ===========================================================================
	public class CPAMsgHandler extends MessageHandler {

		public CPAMsgHandler(SimpleAgent algo) {
			super(algo, Commands.CPA);
		}

		public void processMessage(Message msg) throws StopException {
			Instantiation cpa = (Instantiation) msg.get(Commands.CPA);
			long sc = (Long) msg.get(SC);
			receiveCPA(cpa, sc, false);
		}

		public void sendCPAMessage(Instantiation cpa, VarIdentifier receiverId, long sc) {
			Message msg = createMessage();
			msg.addContents(Commands.CPA, cpa.clone());
			msg.addContents(SC, sc);
			msg.setReceiverId(receiverId);
			sendMessage(msg);
		}

	}

	// ========================================================================
	public class FCCPAMsgHandler extends MessageHandler {

		public FCCPAMsgHandler(AbstractAgent agent) {
			super(agent, Commands.FC_CPA);
		}

		@Override
		public void processMessage(Message msg) throws StopException {
			Instantiation cpa = (Instantiation) msg.get(Commands.FC_CPA);
			long sc = (Long) msg.get(SC);
			forwardCheck(cpa, sc);
		}

		public void sendFC_CPAmessages(Instantiation cpa, VarIdentifier nextVarId, long sc) {
			HashList<VarIdentifier> unassigned = agentOrder.getOrder();
			unassigned.removeAll(cpa.getVarIds());
			unassigned.remove(nextVarId);
			for (VarIdentifier varId : unassigned) {
				Message msg = createMessage();
				msg.addContents(Commands.FC_CPA, cpa.clone());
				msg.addContents(SC, sc);
				msg.setReceiverId(varId);
				sendMessage(msg);
			}
		}
	}

	// ==========================================================================
	public class NotOKMsgHandler extends MessageHandler {

		public NotOKMsgHandler(AbstractAgent agent) {
			super(agent, Commands.NOT_OK);
		}

		@Override
		public void processMessage(Message msg) throws StopException {
			Instantiation notOk = (Instantiation) msg.get(Commands.NOT_OK);
			Nogood nogood = (Nogood) msg.get(Commands.NOGOOD);
			long sc = (Long) msg.get(SC);
			processNotOk(notOk, nogood, sc);
		}

		public void sendNotOks(Instantiation notOk, Nogood nogood, long sc) {
			HashList<VarIdentifier> unassigned = agentOrder.getOrder();
			unassigned.removeAll(notOk.getVarIds());
			unassigned.remove(getVarIdentifier());
			for (VarIdentifier varId : unassigned) {
				this.sendNotOk(notOk.clone(), nogood.clone(), varId, sc);
			}
		}

		private void sendNotOk(Instantiation notOk, Nogood nogood, VarIdentifier varId, long sc) {
			Message msg = this.createMessage();
			msg.addContents(Commands.NOT_OK, notOk);
			msg.addContents(Commands.NOGOOD, nogood);
			msg.addContents(SC, sc);
			msg.setReceiverId(varId);
			sendMessage(msg);
		}
	}

	// ==============================================================================
	public class BackCPAMsgHandler extends MessageHandler {

		public BackCPAMsgHandler(AbstractAgent agent) {
			super(agent, Commands.BACKTRACK);
		}

		@Override
		public void processMessage(Message msg) throws StopException {
			Instantiation cpa = (Instantiation) msg.get(Commands.CPA);
			long sc = (Long) msg.get(SC);
			Nogood nogood = (Nogood) msg.get(Commands.NOGOOD);
			receiveBackCPA(cpa, nogood, sc, true);
		}

		public void sendBack(Instantiation view, VarIdentifier backTo, Nogood nogood, long sc) {
			Message msg = this.createMessage();
			msg.addContents(Commands.CPA, view);
			msg.addContents(Commands.NOGOOD, nogood);
			msg.addContents(SC, sc);
			msg.setReceiverId(backTo);
			sendMessage(msg);
		}
	}
	// ===============================================================================

}
