package simple.dcsp.afcfamily.afc_dbj;

import kernel.StopException;
import kernel.VarIdentifier;
import kernel.com.Message;
import kernel.com.MessageHandler;
import kernel.events.VariableEvent;
import kernel.vars.integer.IntExternalVar;
import simple.SimpleAgent;
import simple.SingleExplainedVarAgent;
import simple.dcsp.Nogood;
import simple.dcsp.afcfamily.TermMsgHandler;
import simple.dcsp.afcfamily.afcng.TimeStamp;
import simulation.metric.Metric;
import simulation.parser.PVariable;
import util.Commands;
import util.HashList;
import util.Instantiation;
import util.VarAssignment;
import util.VarEventMask;

/**
 * The Asynchronous Forward Checking with direct Backjumping (AFC-dbj) published in [Nguyen et al. 2004] 
 * the static version of their improvement of the basic AFC
 * 
 * @authors 
 * <a href="mailto:ezzahir@ensa-agadir.ac.ma?subject=DisChoco &cc=wahbi@lirmm.fr &cc=bessiere@lirmm.fr">
 * Redoaune Ezzahir</a> and 
 * <a href="mailto:wahbi@lirmm.fr?subject=DisChoco &cc=ezzahir@ensa-agadir.ac.ma &cc=bessiere@lirmm.fr">
 * Mohamed Wahbi</a>
 *
 */
public class AFCDBJAgent extends SingleExplainedVarAgent {

	public static final String TIMESTAMP = "TimeStamp";
	public static final String OBSOLETE_CPA = "obsoleteCPA";
	public static final String OBSOLETE_FC = "obsoleteFC";
	public static final String FC = Commands.FC_CPA;
	public static final String CPA = Commands.CPA;
	public static final String NOT_OK = Commands.NOT_OK;
	public static final String BACKTRACK = Commands.BACKTRACK;

	protected TimeStamp timeStamp;
	/** The <tt>CPA</tt> message handler */
	protected CPAMsgHandler cpaMsgH;
	/** The <tt>FC</tt> message handler */
	protected FCMsgHandler fcmsgH;
	/** The <tt>BackCPA</tt> message handler */
	protected BackCPAMsgHandler backCPAMsgH;
	/** The <tt>NOT_OK</tt> message handler */
	protected NotOKMsgHandler notOkMsgH;

	protected boolean agentViewConsistent = true;
	
	/*
	Problem dans l'algorithme dans les lignes 12 et 13 de receiveNG() si un agent reoit 
	un Not_ok il se dsinstancie (pb: dans le cas ou il est suivi d'un BT)
	*/

	public AFCDBJAgent(int agId, String name, String description) {
		super(agId, name, description);
		this.terminateMsgHandler = new TermMsgHandler(this);
		this.cpaMsgH = new CPAMsgHandler(this);
		this.fcmsgH = new FCMsgHandler(this);
		this.backCPAMsgH = new BackCPAMsgHandler(this);
		this.notOkMsgH = new NotOKMsgHandler(this);
		this.timeStamp = new TimeStamp();
		this.initAuxStatistics();
	}
	
	public AFCDBJAgent(PVariable variable) {
		super(variable);
		this.terminateMsgHandler = new TermMsgHandler(this);
		this.cpaMsgH = new CPAMsgHandler(this);
		this.fcmsgH = new FCMsgHandler(this);
		this.backCPAMsgH = new BackCPAMsgHandler(this);
		this.notOkMsgH = new NotOKMsgHandler(this);
		this.timeStamp = new TimeStamp();
		this.initAuxStatistics();
	}
	
	/**
	 * initialize the auxiliary statistics. 
	 */
	private void initAuxStatistics() {
		this.auxiliaryStatistics.put(Commands.nbObsoleteNogood, new Metric(Commands.nbObsoleteNogood));
		this.auxiliaryStatistics.put(Commands.nbAssignments, new Metric(Commands.nbAssignments));
		this.auxiliaryStatistics.put(OBSOLETE_CPA, new Metric(OBSOLETE_CPA));
		this.auxiliaryStatistics.put(OBSOLETE_FC, new Metric(OBSOLETE_FC));
	}
	
	/**
	 * The initialization phase of the constraint handler. In this method we activate the 
	 * constraints with higher agents. 
	 */
	@Override
	public void configurePropagationEngine() {
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_HIGHER_THAN_LOCAL_VAR).add(VariableEvent.INSTANCIATE);
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_HIGHER_THAN_LOCAL_VAR).add(VariableEvent.DESINSTANCIATE);
		this.eventsConfig.get(VarEventMask.LOCAL_VAR_IS_THE_LAST).add(VariableEvent.RESTORE_VAL);
		this.eventsConfig.get(VarEventMask.LOCAL_VAR_IS_BETWEEN_L_H).add(VariableEvent.RESTORE_VAL);
		this.constraintHandler.update();
	}
	
	/**
	 * Add the handlers of messages to the communicator. 
	 */
	@Override
	public void addMsgHandlers() {
		this.communicator.addMsgHandler(this.cpaMsgH);
		this.communicator.addMsgHandler(this.fcmsgH);
		this.communicator.addMsgHandler(this.backCPAMsgH);
		this.communicator.addMsgHandler(this.notOkMsgH);
	}
	
	/**
	 * Initialize the agent view of the agent in order to include all agents higher than the 
	 * current agent. 
	 */
	public void initAgentView() {
		HashList<VarIdentifier> agents = this.agentOrder.getHigherAgents();
		for (VarIdentifier agId : agents) {
			if (!this.communicator.isNeighbor(agId)) {
				VarAssignment var = new VarAssignment(new VarIdentifier(agId), Integer.MIN_VALUE);
				IntExternalVar extVar = this.makeExternalVar(var);
				this.agentView.add(extVar);
			} else {
				this.agentView.add((IntExternalVar) this.getExternalVar(agId));
			}
		}
	}

//==============================================================================================
//		                       AFC-dbj procedures
//==============================================================================================

	/**
	 * This function starts distributed search from the initializing agent, that is the first agent in the ordering.
	 * 
	 * @throws StopExeption
	 */
	public void startDistributedSearch() throws StopException {
		if (this.agentOrder.getMyRank() == 0) {
			this.AssignCPA();
		}
	}

	// ==========================================================================

	/**
	 * This function is called when ever we need to instantiate the variable of the current agent and to 
	 * extend the current partial assignment, if doesn't succeed (no value is possible for the current variable) 
	 * it launches a backtrack by calling procedure {@link #Backtrack()} 
	 */
	public void AssignCPA() throws StopException {
		if (this.valSelector.hasNextValue()) {
			int value = this.valSelector.getNextValue();
			this.var.instantiate(value);
			this.incrementMyStamp();
			Instantiation cpa = this.agentView.toInstantiation();
			cpa.add(this.var.getVarAssignment());
			this.sendCPA(cpa);
		} else {
			this.agentViewConsistent = false;
			this.agentView.clearExclusiveFrom(this.solveMyNogoods().getRhs());
			this.Backtrack();
		}
	}

	/**
	 * This function is called after an instantiation is performed in order to send the extended 
	 * current partial assignments to the next agent in a CPA message 
	 * {@link CPAMsgHandler#sendCPAMessage(Instantiation, int)} and copies in FC_CPA messages 
	 * to all connected lower agents (lower neighbors) {@link FCMsgHandler#sendFCmessages(Instantiation, int)}
	 */
	public void sendCPA(Instantiation cpa) throws StopException {
		VarIdentifier nextVarId = this.getNextAgent();
		if (nextVarId == null) { throw new StopException(Commands.SUCCESS); }
		this.cpaMsgH.sendCPAMessage(cpa, nextVarId);
		this.fcmsgH.sendFCmessages(cpa, nextVarId);
	}
	/** 
	 * 
	 * @return the next agent in the ordering
	 */
	public VarIdentifier getNextAgent() {
		return this.agentOrder.getNextOf(getVarIdentifier());
	}

	public void Backtrack() throws StopException {
		if (this.agentOrder.getMyRank() == 0) {
			throw new StopException(Commands.FAILURE); 
		}
		this.agentViewConsistent = false;
		Instantiation cpa = this.agentView.toInstantiation();
		VarIdentifier target = cpa.last().getVarIdentifier();
		this.backCPAMsgH.sendBack(cpa, target);
		this.notOkMsgH.sendNotOks(cpa, target);
		if (this.var.isInstantiated()) {
			this.var.desinstanciate(VariableEvent.NOCAUSE);
		}
	}

	// ===========================================================================
	public boolean updateAgentView(Instantiation cpa) {
		this.agentView.clearExcept(cpa);
		for (VarAssignment v : cpa) {
			this.agentView.priorityMerge(v);
			if (var.getDomainSize() == 0) { return false; }
		}
		return true;
	}
	
	// ==========================================================================
	protected void processCPA(Instantiation cpa) throws StopException {
		Instantiation view = agentView.toInstantiation();
		if (!agentViewConsistent && cpa.contains(view)) {
			return;
		}
		this.agentViewConsistent = true;
		if (this.updateAgentView(cpa)) {
			this.AssignCPA();
		} else {
			this.Backtrack();
		}
	}

	// ==========================================================================
	public void processFC(Instantiation cpa) throws StopException {
		if (!this.agentViewConsistent) {
			Instantiation view = agentView.toInstantiation();
			if (!cpa.contains(view)) {
				this.agentViewConsistent = true;
			}
		}
		if (this.agentViewConsistent) {
			if (!this.updateAgentView(cpa)) {
				this.Backtrack();
			}
		}
	}
	
	// ==========================================================================
	protected void processBackCPA(Instantiation cpa, TimeStamp ts) throws StopException {
		if (this.var.isInstantiated()) {
			if (timeStamp.compareTimeStamp(ts) >= 0) {
				this.makeNoGoodConstraint(new Nogood(cpa));
				this.AssignCPA();
			}
		}
	}
	/**
	 * This function processes the received NotOk message. It checks if the agent view contains the NotOk. 
	 * If it is the case, it sets the agentview to the received NotOk and the consistent flag to false.
	 * Otherwise, the received NotOk is more up-to-date than the agentview. The agent then sets the time-stamp 
	 * to the received one, the consistent flag to false and the agent view to the received received NotOk. 
	 * If the current variable is instantiated, it must be unassigned. However, we found a small bug in that 
	 * case. the variable must be unassigned just when the received NotOk is not contained in the agent-view. 
	 *  
	 * 
	 * @param not_ok the received inconsistent current partial assignment  
	 * @param ts the time-stamp associated to the received not_ok
	 * @throws StopException
	 */
	public void processNotOk(Instantiation not_ok, TimeStamp ts) throws StopException {
		Instantiation view = this.agentView.toInstantiation();
		if (view.contains(not_ok)) {
			this.agentView.clearExcept(not_ok);
			this.agentView.priorityMerge(not_ok);
			this.agentViewConsistent = false;
		} else if (timeStamp.compareTimeStamp(ts) > 0) {
			this.agentViewConsistent = false;
			timeStamp.setTimeStamp(ts);
			if (!this.updateAgentView(not_ok)) {
				this.Backtrack();
			}
			if (this.var.isInstantiated()) {
				this.var.desinstanciate(VariableEvent.NOCAUSE);
			}
		}
//		if (this.var.isInstantiated()) {
//			this.var.desinstanciate(VariableEvent.NOCAUSE);
//		}
	}
	
	/**
	 * increment the counter of the current variable on the time-stamp 
	 */
	public void incrementMyStamp() {
		this.timeStamp.update(this.agentOrder.getMyRank());
	}
	
	/** The <tt>CPA</tt> message handler */
	public class CPAMsgHandler extends MessageHandler {

		public CPAMsgHandler(SimpleAgent algo) {
			super(algo, CPA);
		}

		public void processMessage(Message msg) throws StopException {
			Instantiation cpa = (Instantiation) msg.get(CPA);
			TimeStamp ts = (TimeStamp) msg.get(TIMESTAMP);		
			if (timeStamp.compareTimeStamp(ts) > 0) {
				timeStamp.setTimeStamp(ts);
				processCPA(cpa);
			} else {
				auxiliaryStatistics.get(OBSOLETE_CPA).value++;
			}
		}
		/**
		 * Send the extended current partial assignment to the next agent in a <tt>CPA</tt> message 
		 * 
		 * @param cpa the current partial assignment
		 * @param nextAgent is the receiver of the <tt>CPA</tt> message
		 */
		public void sendCPAMessage(Instantiation cpa, VarIdentifier nextAgent) {
			Message msg = createMessage();
			msg.addContents(CPA, cpa.clone());
			msg.addContents(TIMESTAMP, timeStamp.clone());
			msg.setReceiverId(nextAgent);
			sendMessage(msg);
		}

	}
	
	/** The <tt>FC</tt> message handler */
	public class FCMsgHandler extends MessageHandler {

		public FCMsgHandler(AFCDBJAgent agent) {
			super(agent, FC);
		}

		@Override
		public void processMessage(Message msg) throws StopException {
			Instantiation cpa = (Instantiation) msg.get(FC);
			TimeStamp ts = (TimeStamp) msg.get(TIMESTAMP);		
			if (timeStamp.compareTimeStamp(ts) > 0) {
				timeStamp.setTimeStamp(ts);
				processFC(cpa);
			} else {
				auxiliaryStatistics.get(OBSOLETE_FC).value++;
			}
		}
		/**
		 * This function by {@link AFCDBJAgent#AssignCPA()} in order to send copies of the extended 
		 * current partial assignment <tt>cpa</tt> in <tt>FC_CPA</tt> messages to all connected lower agents 
		 * (lower neighbors) except the next agent.  
		 * 
		 * @param cpa the current partial assignment
		 * @param nextVarId 
		 */
		public void sendFCmessages(Instantiation cpa, VarIdentifier nextVarId) {
			HashList<VarIdentifier> lowerneighbors = communicator.getLowerNeighbors0();
			if(lowerneighbors.contains(nextVarId)){
				lowerneighbors.remove(nextVarId);
			}
			for (VarIdentifier lower : lowerneighbors) {
				Message msg = createMessage();
				msg.addContents(FC, cpa.clone());
				msg.addContents(TIMESTAMP, timeStamp.clone());
				msg.setReceiverId(lower);
				sendMessage(msg);
			}
		}
	}
	
	/** The <tt>NOT_OK</tt> message handler */	
	public class NotOKMsgHandler extends MessageHandler {
		
		public NotOKMsgHandler(AFCDBJAgent agent) {
			super(agent, NOT_OK);
		}

		@Override
		public void processMessage(Message msg) throws StopException {
			Instantiation notOk = (Instantiation) msg.get(NOT_OK);
			TimeStamp ts = (TimeStamp) msg.get(TIMESTAMP);
			processNotOk(notOk, ts);
		}

		public void sendNotOks(Instantiation notOk, VarIdentifier target) {
			//communicator.getNeighborsLowerThan(target);
			HashList<VarIdentifier> lowerneighbors = agentOrder.getAgentsLowerThan(target); 
			for (VarIdentifier lower : lowerneighbors) {
				sendNotOk(notOk.clone(), lower);
			}
		}

		private void sendNotOk(Instantiation notOk, VarIdentifier agId) {
			Message msg = this.createMessage();
			msg.addContents(NOT_OK, notOk);
			msg.addContents(TIMESTAMP, timeStamp.clone());
			msg.setReceiverId(agId);
			sendMessage(msg);
		}
	}
	
	/** The <tt>BackCPA</tt> message handler */
	public class BackCPAMsgHandler extends MessageHandler {

		public BackCPAMsgHandler(AFCDBJAgent agent) {
			super(agent, BACKTRACK);
		}

		@Override
		public void processMessage(Message msg) throws StopException {
			Instantiation cpa = (Instantiation) msg.get(CPA);
			TimeStamp ts = (TimeStamp) msg.get(TIMESTAMP);
			processBackCPA(cpa, ts);
		}

		public void sendBack(Instantiation cpa, VarIdentifier target) {
			Message msg = this.createMessage();
			msg.addContents(CPA, cpa);
			msg.addContents(TIMESTAMP, timeStamp.clone());
			msg.setReceiverId(target);
			sendMessage(msg);
		}
	}
}
