package simple.dcsp.afcfamily.afcng;

import kernel.AbstractAgent;
import kernel.StopException;
import kernel.VarIdentifier;
import kernel.com.Message;
import kernel.com.MessageHandler;
import kernel.events.VariableEvent;
import kernel.vars.integer.IntExternalVar;
import simple.SingleExplainedVarAgent;
import simple.dcsp.Nogood;
import simple.dcsp.afcfamily.TermMsgHandler;
import simulation.metric.Metric;
import simulation.parser.PVariable;
import util.Commands;
import util.HashList;
import util.Instantiation;
import util.VarAssignment;
import util.VarEventMask;

public class AFCngAgent extends SingleExplainedVarAgent {

	public static final String NEXT_ID = "NEXT-ID";
	public static final String TIMESTAMP = "TimeStamp";
	public static final String OBSOLETE_CPA = "ObsoleteCPA";
	public static final String CPA = Commands.CPA;
	public static final String NOGOOD = Commands.NOGOOD;

	protected boolean agentViewConsistent = true;
	protected CPAMsgHandler cpaMsgH;
	protected NogoodMsgHandler ngdMsgH;
	protected TimeStamp timeStamp;

	public AFCngAgent(int agId, String name, String description) {
		super(agId, name, description);
		this.terminateMsgHandler = new TermMsgHandler(this);
		this.cpaMsgH = new CPAMsgHandler(this);
		this.ngdMsgH = new NogoodMsgHandler(this);
		this.timeStamp = new TimeStamp();
		this.initAuxStatistics();
	}
	
	public AFCngAgent(PVariable variable) {
		super(variable);
		this.terminateMsgHandler = new TermMsgHandler(this);
		this.cpaMsgH = new CPAMsgHandler(this);
		this.ngdMsgH = new NogoodMsgHandler(this);
		this.timeStamp = new TimeStamp();
		this.initAuxStatistics();
	}

	private void initAuxStatistics() {
		this.auxiliaryStatistics.put(Commands.nbObsoleteNogood, new Metric(Commands.nbObsoleteNogood));
		this.auxiliaryStatistics.put(Commands.nbAssignments, new Metric(Commands.nbAssignments));
		this.auxiliaryStatistics.put(OBSOLETE_CPA, new Metric(OBSOLETE_CPA));
	}

	@Override
	public void configurePropagationEngine() {
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_HIGHER_THAN_LOCAL_VAR).add(VariableEvent.INSTANCIATE);
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_HIGHER_THAN_LOCAL_VAR).add(VariableEvent.DESINSTANCIATE);
		this.eventsConfig.get(VarEventMask.LOCAL_VAR_IS_THE_LAST).add(VariableEvent.RESTORE_VAL);
		this.eventsConfig.get(VarEventMask.LOCAL_VAR_IS_BETWEEN_L_H).add(VariableEvent.RESTORE_VAL);
		this.constraintHandler.update();
	}

	public void addMsgHandlers() {
		this.communicator.addMsgHandler(this.cpaMsgH);
		this.communicator.addMsgHandler(this.ngdMsgH);
	}

	// ==========================================================================

	public void initAgentView() {
		HashList<VarIdentifier> agents = this.agentOrder.getHigherAgents();
		for (VarIdentifier agId : agents) {
			if (!this.communicator.isNeighbor(agId)) {
				VarAssignment var = new VarAssignment(new VarIdentifier(agId), Integer.MIN_VALUE);
				IntExternalVar extVar = this.makeExternalVar(var);
				this.agentView.add(extVar);
			} else{
				this.agentView.add((IntExternalVar) this.getExternalVar(agId));
			}
		}
	}

	// ==========================================================================
	/**
	 * Start distributed search from the initializing agent
	 * 
	 * @throws StopExeption
	 */
	public void startDistributedSearch() throws StopException {
		if (this.agentOrder.getMyRank() == 0) {
			this.assignCPA();
		}
	}

	// ==========================================================================

	/**
	 * this function is called when ever we need to instantiate the current partial assignment, if doesn't succeed it launches a backtrack event
	 */
	public void assignCPA() throws StopException {
		if (this.valSelector.hasNextValue()) {
			int val = this.valSelector.getNextValue();
			this.var.instantiate(val);
//			this.incrementMyStamp();
			Instantiation cpa = this.agentView.toInstantiation();
			cpa.add(this.var.getVarAssignment());
			this.updateTimeStamp(cpa.size()-1);
			this.agentViewConsistent = true;
			this.sendCPA(cpa);
		} else {
			this.backtrack();
		}
	}

	// ==========================================================================
	/**
	 * 
	 */
	public void sendCPA(Instantiation cpa) throws StopException {
		VarIdentifier nextVarId = this.getNextAgent();
		if (nextVarId == null) { throw new StopException(Commands.SUCCESS); }
		this.cpaMsgH.sendCPAMessages(cpa, nextVarId);
	}

	// ==========================================================================

	public VarIdentifier getNextAgent() {
		return this.agentOrder.getNextOf(this.getVarIdentifier());
	}

	// ==========================================================================
	public void receiveCPA(TimeStamp ts, Instantiation cpa, VarIdentifier nextId) throws StopException {
		if (this.timeStamp.compareTimeStamp(ts) > 0) {
			this.timeStamp.setTimeStamp(ts);
			this.processCPA(cpa, nextId);
		} else {
			this.auxiliaryStatistics.get(OBSOLETE_CPA).value++;
		}
	}

	// ==========================================================================

	public void processCPA(Instantiation cpa, VarIdentifier nextId) throws StopException {
		this.agentViewConsistent = true;
		if (!this.updateAgentView(cpa)) {
			this.backtrack();
		} else {
			this.checkAssign(nextId);
		}
	}

	public boolean updateAgentView(Instantiation cpa) {
		this.agentView.clearExcept(cpa);
		for (VarAssignment v : cpa) {
			this.agentView.priorityMerge(v);
			if (this.var.getDomainSize() == 0) {
				this.constraintHandler.flushEvents();
				return false;
			}
		}
		return true;
	}

	public void checkAssign(VarIdentifier nextId) throws StopException {
		if (this.getVarIdentifier().equals(nextId)) {
			this.assignCPA();
		}
	}

	public void backtrack() throws StopException {
		Nogood nogood = this.solveMyNogoods();
		if (nogood.isEmpty()) {
			throw new StopException(Commands.FAILURE);
		}
		this.agentViewConsistent = false;
		this.agentView.clearExclusiveFrom(nogood.getRhs());
		this.ngdMsgH.sendNogood(nogood);
		if (this.var.isInstantiated())
			this.var.desinstanciate(VariableEvent.NOCAUSE);
	}

	// ==========================================================================
	public void processNogood(TimeStamp timeStamp, Nogood nogood) throws StopException {
		int comp = this.timeStamp.compareTimeStamp(timeStamp);
		if (comp > 0)
			this.timeStamp.setTimeStamp(timeStamp);
		if (this.agentViewConsistent) {
			if (comp >= 0 && this.isCoherent(nogood)) { // && this.agentView.contains(cpa)) {
				this.makeNoGoodConstraint(nogood);
				this.assignCPA();
			} else {
				this.auxiliaryStatistics.get(Commands.nbObsoleteNogood).value++;
			}
		} else {
			this.auxiliaryStatistics.get(Commands.nbObsoleteNogood).value++;
		}
	}

	// ===========================================================================
	protected boolean isCoherent(Nogood nogood) {
		return this.agentView.compatibleWith(nogood.getLhs()) && nogood.getRhsValue() == this.var.getValue();
	}

	// ===========================================================================
	// ///////////////////////// message handlers Classes //////////////////////
	// ===========================================================================
	public class CPAMsgHandler extends MessageHandler {
		public CPAMsgHandler(AFCngAgent algo) {
			super(algo, CPA);
		}

		public void processMessage(Message msg) throws StopException {
			Instantiation cpa = (Instantiation) msg.get(CPA);
			VarIdentifier nextId = (VarIdentifier) msg.get(NEXT_ID);
			TimeStamp timeStamp = (TimeStamp) msg.get(TIMESTAMP);
			receiveCPA(timeStamp, cpa, nextId);
		}

		public void sendCPAMessages(Instantiation cpa, VarIdentifier nextVarId) {
			this.sendCPAMessage(cpa.clone(), nextVarId, nextVarId);
			HashList<VarIdentifier> lowerNeighbors = communicator.getLowerNeighbors();
			if(lowerNeighbors.contains(nextVarId)){
				lowerNeighbors.remove(nextVarId);
			}
			for (VarIdentifier receiverId : lowerNeighbors) {
				this.sendCPAMessage(cpa.clone(), receiverId, nextVarId);
			}
		}

		public void sendCPAMessage(Instantiation cpa, VarIdentifier receiverId, VarIdentifier nextId) {
			Message msg = createMessage();
			msg.addContents(CPA, cpa);
			msg.addContents(NEXT_ID, nextId);
			msg.addContents(TIMESTAMP, timeStamp.clone());
			msg.setReceiverId(receiverId);
			sendMessage(msg);
		}
	}

	// ==============================================================================
	protected class NogoodMsgHandler extends MessageHandler {
		public NogoodMsgHandler(AbstractAgent agent) {
			super(agent, NOGOOD);
		}

		public void processMessage(Message msg) throws StopException {
			Nogood ng = (Nogood) msg.get(Commands.NOGOOD);
			TimeStamp timeStamp = (TimeStamp) msg.get(TIMESTAMP);
			processNogood(timeStamp, ng);
		}

		public void sendNogood(Nogood nogood) {
			Message msg = this.createMessage();
			msg.addContents(Commands.NOGOOD, nogood);
			msg.addContents(TIMESTAMP, timeStamp.clone());
			msg.setReceiverId(nogood.getRhsVarIdentifier());
			sendMessage(msg);
		}
	}

//	public void incrementMyStamp() {
//		this.timeStamp.update(this.agentOrder.getMyRank());
//	}
	// ==============================================================================
	private void updateTimeStamp(int index) {
		this.timeStamp.update(index, this.var.getTimeCounter());
	}
}