package simple.dcsp.afcfamily.afcng.domDeg;

import java.util.HashMap;
import kernel.AbstractAgent;
import kernel.StopException;
import kernel.VarIdentifier;
import kernel.com.Message;
import kernel.com.MessageHandler;
import simple.dcsp.afcfamily.afcng.DynamicAFCngAgent;
import util.Commands;
import util.HashList;
import util.Instantiation;

public class AFCngDomDegAgent extends DynamicAFCngAgent {

	protected HashMap<VarIdentifier, Long> counters;
	protected HashMap<VarIdentifier, Integer> domainSizes;
	protected DomMsgHandler domSizeH;

	public static final int MIN_DOM = 0;
	public static final int DOM_DEG = 1;
	public static final int DOM_FUTUR_DEG = 2;
	protected int heuristic = DOM_FUTUR_DEG; // static or dynamic heuristic

	/**
	 * 
	 * @param agentId
	 * @param agName
	 * @param description
	 * @param heuristic
	 */
	public AFCngDomDegAgent(int agentId, String agName, String description, int heuristic) {
		super(agentId, agName, description);
		this.domSizeH = new DomMsgHandler(this);
		this.domainSizes = new HashMap<VarIdentifier, Integer>();
		this.counters = new HashMap<VarIdentifier, Long>();
		this.heuristic = heuristic;
		for(VarIdentifier agId : this.getOrder()) {
			domainSizes.put(agId, Integer.MAX_VALUE);
			counters.put(agId, 0L);
		}
	}

	/**
	 * 
	 * @param agentId
	 * @param agName
	 * @param description
	 */
	public AFCngDomDegAgent(int agentId, String agName, String description) {
		super(agentId, agName, description);
		this.domainSizes = new HashMap<VarIdentifier, Integer>();
		this.counters = new HashMap<VarIdentifier, Long>();
	}

	/**
	 * 
	 */
	@Override
	public void addMsgHandlers() {
		super.addMsgHandlers();
		this.communicator.addMsgHandler(this.domSizeH);
	}

	/**
	 * 
	 */
	@Override
	public void checkAssign(VarIdentifier nextId) throws StopException {
		if (this.getVarIdentifier().equals(nextId)) {
			this.assignCPA();
		} else {
			this.sendDomMessages();
		}
	}

	/**
	 * 
	 */
	private void sendDomMessages() {
		HashList<VarIdentifier> unassigned = this.getOrder();
		unassigned.removeAll(this.agentView.toInstantiation().getVarIds());
		unassigned.remove(this.agentId);
		for(VarIdentifier agId : unassigned) {
			this.domSizeH.sendDomSizeMessage(agId);
		}
	}

	/**
	 * 
	 * @param cpa
	 * @return
	 */
	public VarIdentifier getNextAgent(Instantiation cpa) {
		HashList<VarIdentifier> unassigned = this.agentOrder.getLowerAgents();
		if (unassigned.size() == 0)
			return null;
		switch (this.heuristic) {
			case MIN_DOM :
				return getMinDom(unassigned);
			default :
				return getMinDomDeg(unassigned);
		}
	}

	/**
	 * 
	 * @param futureAgents
	 * @return
	 */
	public VarIdentifier getMinDom(HashList<VarIdentifier> futureAgents) {
		int minSize = Integer.MAX_VALUE;
		VarIdentifier nextAgent = null;
		for(VarIdentifier agId : futureAgents) {
			int dom = domainSizes.get(agId);
			if (minSize > dom) {
				minSize = dom;
				nextAgent = agId;
			}
		}
		return nextAgent;
	}

	/**
	 * 
	 * @param futureAgents
	 * @return
	 */
	public VarIdentifier getMinDomDeg(HashList<VarIdentifier> futureAgents) {
		double minDomDeg = Double.MAX_VALUE, domDeg;
		VarIdentifier nextAgent = null;
		for(VarIdentifier agId : futureAgents) {
			int dom = domainSizes.get(agId);
			int deg = this.getDeg(agId, futureAgents);
			if (deg == 0) {
				domDeg = Double.POSITIVE_INFINITY;
			} else {
				domDeg = dom / deg;
			}
			if (minDomDeg > domDeg) {
				minDomDeg = domDeg;
				nextAgent = agId;
			}
		}
		return nextAgent;
	}

	/**
	 * 
	 * @param agId
	 * @param unassigned
	 * @return
	 */
	private int getDeg(VarIdentifier agId, HashList<VarIdentifier> unassigned) {
		if (heuristic == DOM_DEG) {
			return this.agentOrder.getAgentsGraph().getNeighbors(agId).size();
		}
		int future_deg = 0;
		for(VarIdentifier neigh : this.agentOrder.getNeighborsOf(agId)) {
			if (unassigned.contains(neigh))
				future_deg++;
		}
		return future_deg;
	}

	// =============================================================================================
	// ////////////////////// DomMsgHandler //////////////////////////////////////////////////
	// =============================================================================================
	class DomMsgHandler extends MessageHandler {

		/**
		 * @param agent
		 * @param HandlerType
		 */
		public DomMsgHandler(AbstractAgent agent) {
			super(agent, Commands.DomainSize);
		}

		/*
		 * (non-Javadoc)
		 * @see kernel.com.MessageHandler#processMessage(kernel.com.Message)
		 */
		@Override
		public void processMessage(Message msg) throws kernel.StopException {
			VarIdentifier senderId = msg.getSenderId();
			int dom = (Integer) msg.get(Commands.DomainSize + "");
			domainSizes.put(senderId, dom);
		}

		public void sendDomSizeMessage(VarIdentifier receiverId) {
			Message msg = this.createMessage();
			msg.setReceiverId(receiverId);
			msg.addContents(Commands.DomainSize, var.getDomainSize());
			sendMessage(msg);
		}

	}

}
