/************************************************************
Agile-ABT with termination value as heuristic for changing the 
total order among agents, and using the explanations defined 
as a conjunction of the stored nogoods. 
 *************************************************************/

package simple.dcsp.agile.exp;

import static simple.dcsp.agile.exp.AgileMaster.All;
import static simple.dcsp.agile.exp.AgileMaster.MinDom;
import static simple.dcsp.agile.exp.AgileMaster.Neighbors;
import static simple.dcsp.agile.exp.AgileMaster.OrderNeigh;
import static simple.dcsp.agile.exp.AgileMaster.OrderOnly;
import static simple.dcsp.agile.exp.AgileMaster.Topological_ID;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import kernel.AbstractAgent;
import kernel.AgentObserver;
import kernel.Communicator;
import kernel.StopException;
import kernel.VarIdentifier;
import kernel.com.Message;
import kernel.events.AgentEventListener;
import kernel.events.VariableEvent;
import kernel.order.AgentOrdering;
import kernel.vars.integer.IntExternalVar;
import kernel.vars.integer.explain.ExplainedIntVar;
import simple.SingleExplainedVarAgent;
import simple.dcsp.Nogood;
import simple.dcsp.abtfamily.IdleEvent;
import simulation.metric.Metric;
import simulation.metric.MetricListener;
import simulation.parser.PVariable;
import util.Commands;
import util.HashList;
import util.Instantiation;
import util.Tools;
import util.VarAssignment;
import util.VarEventMask;

public class AgileAgent extends SingleExplainedVarAgent {

	public static final String Explanation = "exp";
	public static final String Order = "order";
	public static final String TV = "TV";
	public static final String LINK = "link";
	public static final String State = "state";
	public static final String EXP_CHECKS = "ExpChecks";
	public static final String NCEXP_CHECKS = "NCExpChecks";
	public static final String CHANGE_TARGET = "ChangeTarget";
	public static final String CHANGE_ORDER = "ChangeOrder";
	public static final String KEEP_TARGET = "KeepTarget";
	public static final String IMPROVE_TV = "ImproveTV";
	public static final String UPDATES_FROM_LHS = "UpdatesFromLhs";
	public static final String IMPROVE_TOPOLOGICAL_TV = "ImproveTopologicalTV";

	private short topologicalType = Topological_ID;

	private short heuristic = MinDom;

	public OkMsgHandler okMsgH = null;
	public NogoodMsgHandler ngdMsgH = null;
	public OrderMsgHandler orderMsgH = null;
	public AddLinkMsgHandler adlMsgH = null;

	public Explanations E;
	public Double[] myTV = null;
	private Map<VarIdentifier, Integer> domains;
	private Map<VarIdentifier, Integer> degrees;
	private Map<VarIdentifier, Integer> weights; 

	public boolean onlyOrder = false;
	public boolean onlyToNeighbors = false;
	public boolean IchangeMyOrder = false;
	public boolean topologicalOrder = false;
	private ExplanationObserver expChecksObserver;

	/**
	 * @param agentId
	 * @param agName
	 * @param description
	 * @param dom
	 * @param type
	 * @param topological
	 * @param topologicalType
	 */
	public AgileAgent(int agId, String name, String desc, short heuristic, short type, boolean topological, short topologicalType) {
		super(agId, name, desc);
		this.setType(type);
		this.heuristic = heuristic;
		this.topologicalOrder = topological;
		this.topologicalType = topologicalType;
		this.domains = new HashMap<VarIdentifier, Integer>();
		this.degrees = new HashMap<VarIdentifier, Integer>();
		this.weights = new HashMap<VarIdentifier, Integer>();
		this.agentOrder = new AgileABTAO(this);
		this.constructCommunicator();
		// -------------- Messages Handler ------------------//
		this.okMsgH = new OkMsgHandler(this);
		this.ngdMsgH = new NogoodMsgHandler(this);
		this.orderMsgH = new OrderMsgHandler(this);
		this.adlMsgH = new AddLinkMsgHandler(this);
		this.terminateMsgHandler = new AgileTermDetection(this);
		this.initAuxStatistics();
		this.expChecksObserver = new ExplanationObserver(NCEXP_CHECKS);
		this.addObserver(this.expChecksObserver);
	}
	
	/**
	 * @param agentId
	 * @param agName
	 * @param description
	 * @param dom
	 * @param type
	 * @param topological
	 * @param topologicalType
	 */
	public AgileAgent(PVariable variable, short heuristic, short type, boolean topological, short topologicalType) {
		super(variable);
		this.setType(type);
		this.heuristic = heuristic;
		this.topologicalOrder = topological;
		this.topologicalType = topologicalType;
		this.domains = new HashMap<VarIdentifier, Integer>();
		this.degrees = new HashMap<VarIdentifier, Integer>();
		this.weights = new HashMap<VarIdentifier, Integer>();
		this.agentOrder = new AgileABTAO(this);
		this.constructCommunicator();
		// -------------- Messages Handler ------------------//
		this.okMsgH = new OkMsgHandler(this);
		this.ngdMsgH = new NogoodMsgHandler(this);
		this.orderMsgH = new OrderMsgHandler(this);
		this.adlMsgH = new AddLinkMsgHandler(this);
		this.terminateMsgHandler = new AgileTermDetection(this);
		this.initAuxStatistics();
		this.expChecksObserver = new ExplanationObserver(NCEXP_CHECKS);
		this.addObserver(this.expChecksObserver);
	}

	private void initAuxStatistics() {
		this.auxiliaryStatistics.put(Commands.nbObsoleteNogood, new Metric(Commands.nbObsoleteNogood));
		this.auxiliaryStatistics.put(Commands.nbAssignments, new Metric(Commands.nbAssignments));
		this.auxiliaryStatistics.put(Commands.ObsoleteOk, new Metric(Commands.ObsoleteOk));
		this.auxiliaryStatistics.put(CHANGE_ORDER, new Metric(CHANGE_ORDER, 0L, Metric.max));
		this.auxiliaryStatistics.put(CHANGE_TARGET, new Metric(CHANGE_TARGET));
		this.auxiliaryStatistics.put(KEEP_TARGET, new Metric(KEEP_TARGET));
		this.auxiliaryStatistics.put(IMPROVE_TV, new Metric(IMPROVE_TV));
		this.auxiliaryStatistics.put(UPDATES_FROM_LHS, new Metric(UPDATES_FROM_LHS));
		this.auxiliaryStatistics.put(IMPROVE_TOPOLOGICAL_TV, new Metric(IMPROVE_TOPOLOGICAL_TV));
		this.auxiliaryStatistics.put(EXP_CHECKS, new Metric(EXP_CHECKS));
		this.auxiliaryStatistics.put(Commands.MSGsSIZE, new Metric(Commands.MSGsSIZE));
		this.auxiliaryStatistics.put(Commands.MaxMSGsSIZE, new Metric(Commands.MaxMSGsSIZE, 0L, Metric.max));
	}

	/**
	 * @param msg
	 */
	public void storeStatisticsFor(Message msg) {
		long size = msg.getSize();
		this.auxiliaryStatistics.get(Commands.MSGsSIZE).value += size;
		if(this.auxiliaryStatistics.get(Commands.MaxMSGsSIZE).value < size){
			this.auxiliaryStatistics.get(Commands.MaxMSGsSIZE).value = size;
		}
	}

	private void setType(short type) {
		switch (type) {
			case All :
				this.onlyToNeighbors = false;
				this.onlyOrder = false;
			break;
			case OrderOnly :
				this.onlyToNeighbors = false;
				this.onlyOrder = true;
			break;
			case Neighbors :
				this.onlyToNeighbors = true;
				this.onlyOrder = false;
			break;
			case OrderNeigh :
				this.onlyToNeighbors = true;
				this.onlyOrder = true;
			break;
		}
	}

	/**
	 * 
	 */
	private void constructCommunicator() {
		this.communicator = new Communicator(this) {
			protected boolean waitForArrivalMessages(long msTimeout) throws InterruptedException {
				logDomains();
				this.waitingForMessage = true;
				boolean arrived = false;
				if (((AgileAgent) this.myAgent).isIdle()) {
					arrived = mailBoxLock.waitUntilTrue(5L);
					if (!arrived) {
						this.myAgent.fireAgentEvent(new IdleEvent(this.myAgent));
					}
				}
				if (!arrived) {
					arrived = mailBoxLock.waitUntilTrue(msTimeout);
				}
				this.waitingForMessage = false;
				logDomains();
				return arrived;
			}
		};
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see AbstractAgent#addMsgHandlers()
	 */
	@Override
	public void addMsgHandlers() {
		this.communicator.addMsgHandler(this.okMsgH);
		this.communicator.addMsgHandler(this.ngdMsgH);
		this.communicator.addMsgHandler(this.orderMsgH);
		this.communicator.addMsgHandler(this.adlMsgH);
	}

	@Override
	public void configurePropagationEngine() {
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_HIGHER_THAN_LOCAL_VAR).add(VariableEvent.INSTANCIATE);
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_HIGHER_THAN_LOCAL_VAR).add(VariableEvent.DESINSTANCIATE);
		this.eventsConfig.get(VarEventMask.LOCAL_VAR_IS_THE_LAST).add(VariableEvent.RESTORE_VAL);
		this.eventsConfig.get(VarEventMask.LOCAL_VAR_IS_BETWEEN_L_H).add(VariableEvent.RESTORE_VAL);
		this.constraintHandler.update();
	}

	public void initAgentView() {
		HashList<VarIdentifier> agents = this.getOrder();
		agents.remove(this.getVarIdentifier());
		for (VarIdentifier varId : agents) {
			if (!this.communicator.isNeighbor(varId)) {
				VarAssignment var = new VarAssignment(varId, Integer.MIN_VALUE);
				IntExternalVar extVar = this.makeExternalVar(var);
				this.agentView.add(extVar);
			} else{
				this.agentView.add((IntExternalVar) this.getExternalVar(varId));
			}
		}
	}

	public boolean isIdle() {
		if (this.disSearchStarted) {
			HashList<VarIdentifier> higherN = this.communicator.getHigherNeighbors();
			Instantiation view = this.agentView.getContextsOfHigherAgents();
			boolean idle = view.getVarIds().contains(higherN) && this.var.isInstantiated();
			if (idle) {
				view.add(this.var.getVarAssignment());
				((AgileTermDetection) this.terminateMsgHandler).SetMyState(this.getOrder(), this.getMyTV(), view);
			}
			return idle;
		}
		return false;
	}

	@Override
	protected void startDistributedSearch() throws StopException {
		this.explanationsConstructor();
		this.addAgentEventListener((AgentEventListener) this.terminateMsgHandler);
		((AgileTermDetection) this.terminateMsgHandler).initialize(this.getOrder(), this.getMyTV());
		this.CheckAgentView();
	}

	// ================================================================================================//
	private void explanationsConstructor() {
		this.E = new Explanations(this.getOrder(), this.domains, this.degrees, this.weights, this.heuristic);
		this.myTV = E.computeTV(this.getOrder());
	}

	// ================================================================================================//
	/**
	 * @return
	 */
	protected boolean isMyExplanationChanged() {
		if (this.onlyOrder) { return !this.onlyOrder; }
		HashList<VarIdentifier> newLhs = this.getExplanationLhs().getVarIds();
		int newRhs = this.getExplanationRhs();
		HashList<VarIdentifier> oldLhs = this.E.getLhs(this.getVarIdentifier());
		int oldRhs = this.E.getRhs(this.getVarIdentifier());
		return (oldRhs == newRhs && (oldLhs.contains(newLhs) && (newLhs.contains(oldLhs))));
	}

	// ================================================================================================//
	/**
	 * choose an instantiation (i.e a value) compatible with the agentView
	 * 
	 * @throws StopException
	 */
	public void CheckAgentView() throws StopException {
		this.constraintHandler.propagate();
		int value = this.var.getValue();
		if (!this.var.getDomain().contains(value)) {
			if (this.valSelector.hasNextValue()) {
				int myValue = this.valSelector.getNextValue();
				this.var.instantiate(myValue);
				this.okMsgH.sendOkMessages();
			} else {
				this.Backtrack();
			}
		} else {
			if (this.IchangeMyOrder || this.isMyExplanationChanged()) {
				this.okMsgH.sendOkMessages();
			}
		}
	}

	// ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	public void Backtrack() throws StopException {
		VarIdentifier myVarId = this.getVarIdentifier();
		this.weights.put(myVarId, this.weights.get(myVarId)+1);
		Instantiation nogood = new Instantiation(this.agentOrder);
		HashMap<VarIdentifier, Integer> domainSize = this.solve(nogood);
		if (nogood.isEmpty()) { throw new StopException(Commands.FAILURE); }
		VarIdentifier target = nogood.last().getVarIdentifier();
		this.E.removeExplanationOf(this.getVarIdentifier());
		Explanations E = this.chooseVariableOrder(this.E, nogood, domainSize);	
		if (Tools.compare(this.myTV, E.getNewTV()) < 0) {
			this.E = E;
			this.setNewOrder(E.getNewOrder(), E.getNewTV());
			Nogood ng = new Nogood(E.getNewNG().clone(this.agentOrder));
			this.ngdMsgH.sendNogood(ng);
			this.removeAssignment(ng.getRhs());
			this.orderMsgH.broadcastOrder(ng.getRhsVarIdentifier());
			if (ng.getRhsVarIdentifier().equals(target)) {
				this.auxiliaryStatistics.get(KEEP_TARGET).value++;
			} else {
				this.auxiliaryStatistics.get(CHANGE_TARGET).value++;
			}
			this.auxiliaryStatistics.get(IMPROVE_TV).value++;
		} else {
			Nogood ng = new Nogood(E.getNewNG().clone(this.agentOrder));
			this.ngdMsgH.sendNogood(ng);
			this.auxiliaryStatistics.get(KEEP_TARGET).value++;
			this.removeAssignment(ng.getRhs());
		}
		this.E.clear();
		this.CheckAgentView();
	}


	// ================================================================================================//
	/**
	 * Procedure to check if a new link should be added.
	 * 
	 * @param nogood
	 */
	public void CheckAddLink(Instantiation nogood) {
		if (this.onlyToNeighbors) {
			Iterator<VarAssignment> it = nogood.iterator();
			while (it.hasNext()) {
				VarAssignment ass = it.next();
				if (!this.communicator.isNeighbor(ass.getVarIdentifier())) {
					this.communicator.addNeighbor(ass.getVarIdentifier());
					this.agentView.add(this.makeExternalVar(ass));
					this.adlMsgH.sendAddLink(ass);
				}
			}
		}
	}

	// ================================================================================================//
	/**
	 * 
	 */
	private boolean isCompatibleWithMyOrder(VarIdentifier agId, Instantiation instantiation) {
		HashList<VarIdentifier> predecessors = instantiation.getVarIds();
		HashList<VarIdentifier> higherAgents = this.agentOrder.getAgentsHigherThan(agId);
		return higherAgents.contains(predecessors);
	}

	// ================================================================================================//
	/**
	 * If the received explanation is newest take it
	 */
	public boolean setExplanation(VarIdentifier agId, Instantiation lhs, int rhs) {
		boolean viewCompatible = this.agentView.compatibleWith(lhs);
		boolean orderCompatible = this.isCompatibleWithMyOrder(agId, lhs);
		if (viewCompatible && orderCompatible) {// && this.agentOrder.isHigher(agId)){
			this.E.setExplanation(agId, lhs.getVarIds(), rhs);
			if (lhs.isEmpty()) {
				this.E.updateMaxDomainSize(agId, rhs);
			}
			return true;
		} else {
			this.E.removeExplanationOf(agId);
		}
		return false;
	}

	// ================================================================================================//
	public void updateAgentView(Instantiation lhs) {
		Iterator<VarAssignment> it = lhs.iterator();
		while (it.hasNext()) {
			VarAssignment ass = it.next();
			int cmp = this.agentView.compareTimeCounter(ass);
			if(cmp>0 && !this.agentView.compatible(ass)){
				this.E.removeExplanationsContainning(ass.getVarIdentifier());
			}
			if (cmp > 0 && this.agentOrder.isHigher(ass.getVarIdentifier())) {
				this.agentView.priorityMerge(ass);
				this.auxiliaryStatistics.get(UPDATES_FROM_LHS).value++;
			}
		}
	}

	// ================================================================================================//
	public void updateAgentView(Instantiation lhs, VarAssignment ass) {
		int cmp = this.agentView.compareTimeCounter(ass);
		if (cmp > 0 && !this.agentView.compatible(ass)) {
			this.E.removeExplanationsContainning(ass.getVarIdentifier());
		}
		if (cmp >= 0) {
			this.agentView.priorityMerge(ass);
			this.E.removeExplanationOf(ass.getVarIdentifier());
		} else {
			this.auxiliaryStatistics.get(Commands.ObsoleteOk).value++;
		}
		this.updateAgentView(lhs);
	}

	// ================================================================================================//
	private boolean setNewOrder(HashList<VarIdentifier> order, Double[] tv) {
		System.arraycopy(tv, 0, this.myTV, 0, tv.length);
		if (this.agentOrder.setNewOrder(order)) {
			this.auxiliaryStatistics.get(CHANGE_ORDER).value++;
			this.IchangeMyOrder = true;
			this.E.removeIncompatibleLhsWithOrder(this.getOrder());
			this.E.removeExplanationOfAgents(this.agentOrder.getLowerAgents());
			this.constraintHandler.update();
			return true;
		}
		return false;
	}

	// ================================================================================================//
	/**
	 * @param
	 * @return <!--
	 * 
	 *         <pre>
	 * 		<b><i>true</i></b> if <i>{order,tv}</i> are better than <i>{myOrder,myTV}</i>
	 * 		<b><i>false</i></b> otherwise 
	 * <!--
	 * </pre>
	 */
	private boolean isItBestOrder(HashList<VarIdentifier> myOrder, Double myTV[], HashList<VarIdentifier> order, Double[] tv) {
		int cmp = Tools.compare(myTV, tv);
		if (cmp > 0)
			return false;
		if (cmp < 0)
			return true;
		if (myOrder.compare(order) < 0) {
			return true;
		} else {
			return false;
		}
	}

	// ================================================================================================//
	public boolean checkOrder(HashList<VarIdentifier> order, Double tv[]) {
		if (this.isItBestOrder(this.getOrder(), this.myTV, order, tv)) {
			return this.setNewOrder(order, tv);
		}
		return false;
	}

	// ================================================================================================//
	private void removeAssignment(VarAssignment ass) {
		this.agentView.clear(ass);
		this.E.removeExplanationsContainning(ass.getVarIdentifier());
	}

	protected void incrementNumConstraintChecks(int CCs) {
		for (AgentObserver observer : observers) {
			if (observer instanceof MetricListener) {
				((MetricListener) observer).handleIncrementNbConstraintChecks(CCs);
			}
		}
		this.expChecksObserver.handleIncrementNbConstraintChecks(CCs);
	}

	// ================================================================================================//
	public Explanations updateExplanations(Explanations E, Instantiation ngLhs, VarIdentifier agId, HashMap<VarIdentifier, Integer> domainSize, Instantiation ng) {
		Explanations E_i = E.clone();
		int nbChecks = E_i.removeExplanationsContainning(agId);
		this.expChecksObserver.handleIncrementNbConstraintChecks(nbChecks / 3);
		this.auxiliaryStatistics.get(EXP_CHECKS).value += nbChecks;
		HashList<VarIdentifier> lhs_i = HashList.union(ngLhs.getVarIds(), E_i.getLhs(agId));
		int rhs_i = E_i.getRhs(agId) - 1;
		E_i.setExplanation(agId, lhs_i, rhs_i);
		HashList<VarIdentifier> myLhs = ngLhs.getVarIds();
		int myRhs = domainSize.get(agId);
		E_i.setExplanation(this.getVarIdentifier(), myLhs, myRhs);
		return E_i;
	}

	// ================================================================================================//
	private Explanations chooseVariableOrder(Explanations E, Instantiation nogood, HashMap<VarIdentifier, Integer> domainSize) throws StopException {
		if (nogood.isEmpty()) { throw new StopException(Commands.FAILURE); }
		Explanations selE = null, E_i = null;
		HashList<VarIdentifier> order_i = null;
		Double[] TV_i = this.getMaximumTV(), selTV = this.getMaximumTV();
		VarAssignment x_i = null;
		Instantiation lhs = null, ng_i = null;
		for (int i = 0 ; i < nogood.size() ; i++) {
			x_i = nogood.get(i);
			VarIdentifier agId = x_i.getVarIdentifier();
			lhs = nogood.clone();
			lhs.remove(x_i);
			E_i = this.updateExplanations(E, lhs, agId, domainSize, nogood);
//			if (E_i.getRhs(agId) == 0) { // It not necessary to backtrack to agentId
//				ng_i = this.agentView.getInstantiationOf(E_i.getLhs(agId));
//				E_i = this.chooseVariableOrder(E_i, ng_i, domainSize);
//				order_i = E_i.getNewOrder();
//				TV_i = E_i.getNewTV();
//				ng_i = E_i.getNewNG();
//			} else {
				if(this.topologicalOrder){
					order_i = E_i.getTopologicalOrder(this.topologicalType);
					TV_i = E_i.computeTV(order_i);
				} else {
//					double[] topologicalTV = TV_i.clone();
					order_i = E_i.getBestOrder();
					TV_i = E_i.computeTV(order_i);
//					if (Tools.compare(topologicalTV, TV_i) < 0) {
//						this.auxiliaryStatistics.get(IMPROVE_TOPOLOGICAL_TV).value++;
//					}
				}
				ng_i = nogood.clone();
//			}
			if (Tools.compare(selTV, TV_i) < 0) {
				selE = E_i;
				selTV = TV_i;
				selE.setNewOrder(order_i);
				selE.setNewTV(selTV);
				selE.setNewNG(ng_i);
			}
		}
		return selE;
	}
	
	protected HashMap<VarIdentifier, Integer> solve(Instantiation nogood) {
		HashMap<VarIdentifier, Integer> domainSize = new HashMap<VarIdentifier, Integer>();
		Iterator<Integer> it = this.var.getDomain().originalDomainIterator();
		while (it.hasNext()) {
			int value = it.next();
			Instantiation ng = this.buildBetterNogood((ExplainedIntVar) var, value);
			if (ng != null) {
				for(VarAssignment ass : ng){
					VarIdentifier varId = ass.getVarIdentifier();
					if(!domainSize.containsKey(varId)){
						domainSize.put(varId, this.domains.get(this.getVarIdentifier()));
					}
					int size = domainSize.get(varId)-1;
					domainSize.put(varId, size);
				}
				nogood.priorityMerge(ng);
			}
		}
		return domainSize;
	}

	public Double[] getMyTV() {
		return this.myTV.clone();
	}

	private Double[] getMaximumTV() {
		Double[] tv = new Double[this.getOrder().size()];
		for (int i = 0 ; i < tv.length ; i++) {
			tv[i] = Double.MAX_VALUE;
		}
		return tv;
	}

	// =========================================================================================================================/
	public class AgileABTAO extends AgentOrdering {
		public AgileABTAO(AbstractAgent agent) {
			super(agent);
		}

		public void processOrder(Message msg) {
			super.processOrder(msg);
			HashMap<VarIdentifier, Integer> doms = Tools.mapfromString((String) msg.get(Commands.DomainSize), VarIdentifier.class, Integer.class);
			domains.putAll(doms);
			for (VarIdentifier agId : this.agentsOrder) {
				int degree = this.graph.getNeighbors(agId).size();
				degrees.put(agId, degree);
				weights.put(agId, 1);
			}
		}
	}

	/**
	 * @return
	 */
	public Instantiation getExplanationLhs() {
		return this.solveMyNogoodStore();
	}

	/**
	 * @return
	 */
	public int getExplanationRhs() {
		return this.var.getDomainSize();
	}

	public void setWeights(VarIdentifier agId, Integer weight) {
		this.weights.put(agId, weight);
		this.E.setWeights(agId, weight);
	}

	/**
	 * @param senderId
	 * @return
	 */
	public int getMyWeight() {
		return this.weights.get(getVarIdentifier());
	}
	
}
