/**
 * 
 */
package simple.dcsp.agileabt.explanation;

import static simple.dcsp.agileabt.explanation.AgileABTMaster.DomDeg;
import static simple.dcsp.agileabt.explanation.AgileABTMaster.MinDom;
import static simple.dcsp.agileabt.explanation.AgileABTMaster.Topological_Deg;
import static simple.dcsp.agileabt.explanation.AgileABTMaster.Topological_ID;
import static simple.dcsp.agileabt.explanation.AgileABTMaster.Topological_MaxSucc;

import java.util.HashMap;
import java.util.Iterator;

import kernel.VarIdentifier;
import kernel.order.AgentOrdering;
import util.HashList;
import util.Instantiation;
import util.VarAssignment;

public class Explanations {

	private HashMap<VarIdentifier, Explanation> explanations = null;
	private HashMap<VarIdentifier, Integer> maxDomSize = null;
	private HashMap<VarIdentifier, Integer> degrees = null;
	private final HashList<VarIdentifier> agentIds;

	private HashList<VarIdentifier> newOrder = null;
	private Double[] newTV = null;
	private Instantiation newNG = null;
	private short heuristic = MinDom;

	@SuppressWarnings("unchecked")
	public Explanations(HashList<VarIdentifier> agentIds, HashMap<VarIdentifier, Integer> maxDomSize, HashMap<VarIdentifier, Integer> degrees, short heuristic) {
		this.heuristic = heuristic;
		this.agentIds = new HashList<VarIdentifier>(agentIds);
		this.maxDomSize = (HashMap<VarIdentifier, Integer>) maxDomSize.clone();
		this.degrees = degrees;
		this.explanations = new HashMap<VarIdentifier, Explanation>();
		for (VarIdentifier agId : agentIds) {
			this.removeExplanationOf(agId);
		}
		this.newTV = new Double[agentIds.size()];
	}

	public void updateMaxDomainSize(VarIdentifier agId, int d) {
		int value = Math.min(this.getMaxDomainSize(agId), d);
		this.maxDomSize.put(agId, value);
	}

	public int getMaxDomainSize(VarIdentifier agId) {
		return this.maxDomSize.get(agId);
	}

	public void removeExplanationOf(VarIdentifier agId) {
		Instantiation lhs = new Instantiation();
		Integer rhs = this.maxDomSize.get(agId);
		this.setExplanation(agId, new Explanation(lhs, rhs));
	}

	public Instantiation getLhs(VarIdentifier agId) {
		return this.explanations.get(agId).getLhs();
	}

	public int getRhs(VarIdentifier agId) {
		return this.explanations.get(agId).getRhs();
	}

	public Explanation getExplanation(VarIdentifier agId) {
		return this.explanations.get(agId);
	}

	public void setRhs(VarIdentifier agId, int rhs) {
		this.explanations.get(agId).setRhs(rhs);
	}

	public HashList<VarIdentifier> getNewOrder() {
		return this.newOrder;
	}

	public void setNewOrder(HashList<VarIdentifier> newOrder) {
		this.newOrder = newOrder;
	}

	public Double[] getNewTV() {
		return this.newTV;
	}

	public void setNewTV(Double[] tv) {
		System.arraycopy(tv, 0, this.newTV, 0, tv.length);
	}

	public Instantiation getNewNG() {
		return this.newNG;
	}

	public void setNewNG(Instantiation newNG) {
		this.newNG = newNG;
	}

	public void clear() {
		this.newOrder = null;
		this.newTV = null;
		this.newNG = null;
	}

	@SuppressWarnings("unchecked")
	private HashMap<VarIdentifier, Explanation> cloneExplanations() {
		return (HashMap<VarIdentifier, Explanation>) this.explanations.clone();
	}

	public void setExplanation(VarIdentifier agId, Explanation exp) {
		this.explanations.put(agId, exp);
	}

	public void setExplanation(VarIdentifier agId, Instantiation lhs, int rhs) {
		this.setExplanation(agId, new Explanation(lhs, rhs));
	}

	/**
	 * remove all explanations incompatible with order
	 */
	public int removeIncompatibleLhsWithOrder(HashList<VarIdentifier> order) {
		int nbChecks = 0;
		Iterator<VarIdentifier> it = this.explanations.keySet().iterator();
		while (it.hasNext()) {
			VarIdentifier agId = it.next();
			HashList<VarIdentifier> higherAgents = order.getElementsBefor(agId);// this.getHigherAgents(order, agId);
			HashList<VarIdentifier> predecessors = this.getPredecessors(agId);
			if (!higherAgents.contains(predecessors)) {
				this.removeExplanationOf(agId);
			}
			nbChecks++;
		}
		return nbChecks;
	}

	private HashList<VarIdentifier> getPredecessors(VarIdentifier agId) {
		return this.getLhs(agId).getVarIds();
	}

	/**
	 * remove all explanations incompatible with order
	 */
	public void removeExplanationOfAgents(HashList<VarIdentifier> agents) {
		for (VarIdentifier agId : agents) {
			this.removeExplanationOf(agId);
		}
	}

	/**
	 * 
	 */
	public int refresh(AgentOrdering agentOrder, VarIdentifier myId) {
		return this.removeIncompatibleLhsWithOrder(agentOrder.getOrder());
	}

	public int remove(VarAssignment varAss) {
		int nbCheck = 0;
		Iterator<VarIdentifier> it = this.explanations.keySet().iterator();
		while (it.hasNext()) {
			VarIdentifier agId = it.next();
			Explanation exp = this.explanations.get(agId);
			if (exp.getLhs().contains(varAss.getVarIdentifier())) {
				this.removeExplanationOf(agId);
				nbCheck++;
			}
		}
		return nbCheck;
	}

	public HashList<VarIdentifier> getTopologicalOrder(short topologicalType) {
		HashList<VarIdentifier> order = new HashList<VarIdentifier>();
		HashMap<VarIdentifier, HashList<VarIdentifier>> successors = this.getGraphSuccessor();
		HashMap<VarIdentifier, HashList<VarIdentifier>> graph = this.getGraphPredecessor();
		while (graph.keySet().size() > 0) {
			HashList<VarIdentifier> roots = this.getRoots(graph);
			VarIdentifier selected = null;
			double min = Integer.MAX_VALUE;
			double rhs = 0;
			switch (topologicalType) {
				case Topological_ID : {
					for (VarIdentifier agId : roots) {
						if (this.heuristic == MinDom) {
							rhs = this.getRhs(agId);
						} else if (this.heuristic == DomDeg) {
							rhs = ((double) this.getRhs(agId)) / this.degrees.get(agId);
						}
						if ((min > rhs) || ((min == rhs) && (agId.compareTo(selected) < 0))) {
							min = this.getRhs(agId);
							selected = agId;
						}
					}
					break;
				}
				case Topological_Deg : {
					int deg = Integer.MIN_VALUE;
					for (VarIdentifier agId : roots) {
						if (this.heuristic == MinDom) {
							rhs = this.getRhs(agId);
						} else if (this.heuristic == DomDeg) {
							rhs = ((double) this.getRhs(agId)) / this.degrees.get(agId);
						}
						if (min > rhs || (min == rhs && (this.degrees.get(agId) > deg))) {
							min = this.getRhs(agId);
							deg = this.degrees.get(agId);
							selected = agId;
						}
					}
					break;
				}
				case Topological_MaxSucc : {
					for (VarIdentifier agId : roots) {
						if (this.heuristic == MinDom) {
							rhs = this.getRhs(agId);
						} else if (this.heuristic == DomDeg) {
							rhs = ((double) this.getRhs(agId)) / this.degrees.get(agId);
						}
						if (min > rhs || (min == rhs && successors.get(agId).size() > successors.get(selected).size())) {
							min = this.getRhs(agId);
							selected = agId;
						}
					}
					break;
				}
			}
			order.add(selected);
			this.remove(graph, selected);
		}
		if (order.size() != this.maxDomSize.keySet().size()) { throw new Error("\nacyclic graph : \n" + this); }
		return order;
	}

	/**
	 * 
	 * @return
	 */
	public HashList<VarIdentifier> getBestOrder() {
		HashList<VarIdentifier> order = new HashList<VarIdentifier>();
		HashMap<VarIdentifier, HashList<VarIdentifier>> graph = this.getGraphPredecessor();
		while (order.size() < this.agentIds.size()) {
			HashList<VarIdentifier> allRoots = this.getRoots(graph);
			HashList<VarIdentifier> S1 = this.getMinRoots(allRoots);
			while (S1.size() != 0) {
				if (S1.size() > 1) {
					int d = this.getRhs(S1.get(0));
					HashList<VarIdentifier> S2 = this.getAgentsWithMinRhs(graph, d);
					if (S2.size() == 0) {
						VarIdentifier selected = S1.get(0);
						order.add(selected);
						this.remove(graph, selected);
						break;
					}
					HashList<VarIdentifier> S3 = this.getRootsWithMinPredecessor(graph, S2);
					HashList<VarIdentifier> S4 = this.getMinRoots(S3);
					if (S4.size() == 1) {
						VarIdentifier selected = S4.get(0);
						HashList<VarIdentifier> predecessors = graph.get(selected);
						for (VarIdentifier pred : graph.get(selected)) {
							predecessors.addAll(graph.get(pred));
						}
						while (predecessors.size() > 0) {
							VarIdentifier pred = predecessors.get(0);
							predecessors.remove(pred);
							if (order.contains(graph.get(pred))) {
								order.add(pred);
								this.remove(graph, pred);
							} else {
								predecessors.add(pred);
							}
						}
						order.add(selected);
						this.remove(graph, selected);
						break;
					} else {
						S1 = S4;
					}
					// S1 = S4;
				} else {
					VarIdentifier selected = S1.get(0);
					order.add(selected);
					this.remove(graph, selected);
					break;
				}
			}
		}
		return order;
	}

	/**
	 * 
	 * @param graph
	 * @param agents
	 * @return
	 */
	private HashList<VarIdentifier> getRootsWithMinPredecessor(HashMap<VarIdentifier, HashList<VarIdentifier>> graph, HashList<VarIdentifier> agents) {
		int min = Integer.MAX_VALUE;
		for (VarIdentifier agId : agents) {
			if (graph.get(agId).size() < min) {
				min = graph.get(agId).size();
			}
		}
		HashList<VarIdentifier> ans = new HashList<VarIdentifier>();
		for (VarIdentifier agId : agents) {
			if (graph.get(agId).size() == min) {
				ans.add(agId);
			}
		}
		return ans;
	}

	/**
	 * 
	 * @param graph
	 * @param d
	 * @return
	 */
	private HashList<VarIdentifier> getAgentsWithMinRhs(HashMap<VarIdentifier, HashList<VarIdentifier>> graph, int d) {
		HashList<VarIdentifier> agents = new HashList<VarIdentifier>();
		for (VarIdentifier agId : graph.keySet()) {
			if (this.getRhs(agId) < d) {
				agents.add(agId);
				for (VarIdentifier pre : graph.get(agId)) {
					if (this.getRhs(pre) != d) {
						agents.remove(agId);
						break;
					}
				}
			}
		}
		return agents;
	}

	/**
	 * 
	 * @param roots
	 * @return
	 */
	private HashList<VarIdentifier> getMinRoots(HashList<VarIdentifier> allRoots) {
		int min = Integer.MAX_VALUE;
		for (VarIdentifier agId : allRoots) {
			if (this.getRhs(agId) < min) {
				min = this.getRhs(agId);
			}
		}
		HashList<VarIdentifier> roots = new HashList<VarIdentifier>();
		for (VarIdentifier agId : allRoots) {
			if (this.getRhs(agId) == min) {
				roots.add(agId);
			}
		}
		return roots;
	}

	/**
	 * 
	 * @param graph
	 * @param selected
	 */
	private void remove(HashMap<VarIdentifier, HashList<VarIdentifier>> graph, VarIdentifier selected) {
		for (VarIdentifier agId : graph.keySet()) {
			if (graph.get(agId).contains(selected)) {
				graph.get(agId).remove(selected);
			}
		}
		graph.remove(selected);
	}

	/**
	 * 
	 * @param graph
	 *            is the graph of predecessors
	 * @return
	 */
	private HashList<VarIdentifier> getRoots(HashMap<VarIdentifier, HashList<VarIdentifier>> graph) {
		HashList<VarIdentifier> roots = new HashList<VarIdentifier>();
		for (VarIdentifier agId : graph.keySet()) {
			if ((graph.get(agId).size() == 0)) {
				roots.add(agId);
			}
		}
		return roots;
	}

	/**
	 * 
	 * @return
	 */
	private HashMap<VarIdentifier, HashList<VarIdentifier>> getGraphSuccessor() {
		HashMap<VarIdentifier, HashList<VarIdentifier>> graph = new HashMap<VarIdentifier, HashList<VarIdentifier>>();
		for (VarIdentifier agId : this.agentIds) {
			graph.put(agId, new HashList<VarIdentifier>());
		}
		for (VarIdentifier agId : this.agentIds) {
			HashList<VarIdentifier> predecessors = this.getPredecessors(agId);
			for (VarIdentifier pre : predecessors) {
				graph.get(pre).add(agId);
			}
		}
		return graph;
	}

	/**
	 * 
	 * @return
	 */
	private HashMap<VarIdentifier, HashList<VarIdentifier>> getGraphPredecessor() {
		HashMap<VarIdentifier, HashList<VarIdentifier>> graph = new HashMap<VarIdentifier, HashList<VarIdentifier>>();
		for (VarIdentifier agId : this.agentIds) {
			HashList<VarIdentifier> predecessors = this.getPredecessors(agId);
			graph.put(agId, predecessors);
		}
		return graph;
	}

	/**
	 * 
	 */
	public Explanations clone() {
		HashList<VarIdentifier> agentIds = new HashList<VarIdentifier>(this.agentIds);
		Explanations E = new Explanations(agentIds, cloneMaxSizes(), cloneDegrees(), this.heuristic);
		E.explanations = this.cloneExplanations();
		return E;
	}

	public boolean isAcyclic() {
		for (VarIdentifier agId : this.explanations.keySet()) {
			HashList<VarIdentifier> predecessors = this.getPredecessors(agId);
			for (VarIdentifier pred : predecessors) {
				if (this.getExplanation(pred).getPredecessors().contains(agId)) {
					System.err.println("*******************************************\n" + agId + " :: predecessors" + predecessors + " :: pred: "
							+ pred + "\n" + this);
					return true;
				}
			}
		}
		return false;
	}

	@SuppressWarnings("unchecked")
	public HashMap<VarIdentifier, Integer> cloneMaxSizes() {
		return (HashMap<VarIdentifier, Integer>) this.maxDomSize.clone();
	}

	@SuppressWarnings("unchecked")
	public HashMap<VarIdentifier, Integer> cloneDegrees() {
		return (HashMap<VarIdentifier, Integer>) this.degrees.clone();
	}

	public String toString(VarIdentifier agId) {
		Explanation exp = this.getExplanation(agId);
		return agId + " :: " + exp.toString();
	}

	public String toString() {
		String ans = "";
		for (VarIdentifier agId : this.agentIds) {
			ans += this.toString(agId) + "\n";
		}
		return ans;
	}

	public void removeExplanationsIncompatibleWith(Instantiation inst) {
		Iterator<VarIdentifier> it = this.explanations.keySet().iterator();
		while (it.hasNext()) {
			VarIdentifier agId = it.next();
			Instantiation lhs = this.getLhs(agId);
			if (!inst.compatible(lhs)) {
				this.removeExplanationOf(agId);
			}
		}
	}

	public Double[] computeTV(HashList<VarIdentifier> order) {
		Double[] tv = new Double[order.size()];
		for (int i = 0 ; i < order.size() ; i++) {
			VarIdentifier agId = order.get(i);
			if (this.heuristic == MinDom) {
				tv[i] = new Double(this.getRhs(agId));
			} else {
				tv[i] = ((double) this.getRhs(agId)) / this.degrees.get(agId);
			}
		}
		return tv;
	}

}
