package simple.dcsp.agileabt.explanation;

import java.util.HashMap;
import java.util.Iterator;

import kernel.AbstractMaster;
import kernel.MasterObserver;
import kernel.VarIdentifier;
import kernel.com.Message;
import simple.dcsp.agileabt.StateOfAgent;
import util.Commands;
import util.HashList;
import util.Instantiation;
import util.Logger;
import util.Tools;

/**
 * termination detection (Silaghi et al. 2000)
 * 
 * @author redouane
 * 
 */
public class TerminationDetector extends MasterObserver {
	
	public static final String ORDER = AgileABTAgent.ORDER;
	public static final String TERMINATION_VALUE = AgileABTAgent.TERMINATION_VALUE;
	public static final String STATE = AgileABTAgent.State;

	protected HashMap<VarIdentifier, Instantiation> acceptedMsg;
	protected HashMap<VarIdentifier, StateOfAgent> stateOfAgents = null;

	protected HashList<VarIdentifier> initialEndAgents = new HashList<VarIdentifier>();
	protected HashList<VarIdentifier> order = null;
	protected Double[] tv = null;

	public TerminationDetector() {
		super();
		this.acceptedMsg = new HashMap<VarIdentifier, Instantiation>();
		this.stateOfAgents = new HashMap<VarIdentifier, StateOfAgent>();
	}

	/**
	 * 
	 * @param order
	 * @return
	 */
	private Double[] getMaximumTV(HashList<VarIdentifier> order) {
		Double[] tv = new Double[order.size()];
		for (int i = 0 ; i < tv.length ; i++) {
			tv[i] = Double.MAX_VALUE;
		}
		return tv;
	}

	/**
	 * 
	 * @param order
	 */
	public void setInitialOrder(HashList<VarIdentifier> order) {
		if (this.order == null) {
			this.order = order;
			this.tv = new Double[order.size()];
			Double[] maxTV = this.getMaximumTV(order);
			System.arraycopy(maxTV, 0, this.tv, 0, maxTV.length);
		}
	}

	private void ProcessOrder(Message msg) {
		HashList<VarIdentifier> order = HashList.fromString((String) msg.get(ORDER), VarIdentifier.class);
		Double[] newtv = Tools.arrayFromString((String) msg.get(TERMINATION_VALUE), Double.class);
		this.setInitialOrder(order);
		if (StateOfAgent.areAcceptable(this.order, this.tv, order, newtv) < 0) {
			this.order = order;
			System.arraycopy(newtv, 0, this.tv, 0, newtv.length);
			this.computeInitialEndAgents();
			this.checkEnd();
		}
		this.master.log("IEA: " + this.initialEndAgents + " ==>\n  " + this.order + "@" + Tools.arrayToString(this.tv) + "\n"
					+ this.stateOfAgents + "\n receiveMsg : " + msg + "\n" + this.toString(), Logger.INFO_LEVEL);
	}

	@Override
	public boolean handleReceivedMessage(Message msg) {
		if (msg.getType().equals(ORDER)) {
			this.ProcessOrder(msg);
		}
		if (msg.getType().equals(Commands.TERMINATION_DETECTION)) {
			VarIdentifier senderId = msg.getSenderId();
			StateOfAgent receivedState = (StateOfAgent) msg.get(STATE);
			HashList<VarIdentifier> order = receivedState.getOrder();
			this.setInitialOrder(order);
			Double[] tv = receivedState.getTV();
			int acceptable = StateOfAgent.areAcceptable(this.order, this.tv, order, tv);
			if (acceptable < 0) {
				this.order = order;
				System.arraycopy(tv, 0, this.tv, 0, tv.length);
				this.stateOfAgents.put(senderId, receivedState);
			} else {
				if (acceptable == 0) {
					if (this.stateOfAgents.keySet().contains(senderId)) {
						StateOfAgent stateOfSender = this.stateOfAgents.get(senderId);
						stateOfSender.setAgentView(order, tv, receivedState.getAgentView());
					} else {
						this.stateOfAgents.put(senderId, receivedState);
					}
				}
			}
			this.checkEnd();
		}
		return true;
	}

	private HashList<VarIdentifier> GammaPlus(VarIdentifier agId) {
		HashList<VarIdentifier> higherAgents = this.order.getElementsBefor(agId);
		HashList<VarIdentifier> neighbours = this.master.getNeighbors(agId);
		return HashList.intersection(higherAgents, neighbours);
	}

	private HashList<VarIdentifier> computeInitialEndAgents() {
		this.initialEndAgents = new HashList<VarIdentifier>();
		for (VarIdentifier agId : this.order) {
			if (this.GammaPlus(agId).size() == 0) {
				this.initialEndAgents.add(agId);
			}
		}
		return this.initialEndAgents;
	}

	private StateOfAgent allReceivedAndAllCompatible() {
		StateOfAgent globalState = new StateOfAgent(this.getOrder(), this.getTV());
		for (VarIdentifier agId : this.initialEndAgents) {
			if (!this.stateOfAgents.keySet().contains(agId)) {
				if (Logger.LOG_DEBUG) {
					this.master.log("accepted not yet received from : " + agId, Logger.INFO_LEVEL);
				}
				return null;
			}
			StateOfAgent state = this.stateOfAgents.get(agId);
			if (!globalState.isCompatible(state)) {
				if (Logger.LOG_DEBUG) {
					this.master.log("Incompatible state of : " + agId + "=" + state + " with the globale one:\n" + globalState, Logger.INFO_LEVEL);
				}
				return null;
			}
			globalState.priorityMerge(state.getAgentView());
		}
		return globalState;
	}

	public void removeIncompatibleState() {
		Iterator<VarIdentifier> it = this.stateOfAgents.keySet().iterator();
		while (it.hasNext()) {
			VarIdentifier agId = it.next();
			StateOfAgent state = this.stateOfAgents.get(agId);
			int acceptable = StateOfAgent.areAcceptable(this.order, this.tv, state.getOrder(), state.getTV());
			if (acceptable > 0) {
				it.remove();
			}
		}
	}

	private boolean checkEnd() {
		this.removeIncompatibleState();
		this.computeInitialEndAgents();
		if (this.stateOfAgents.size() < this.initialEndAgents.size()) {
			this.master.log("\nStateOfAgents size : " + this.stateOfAgents.size() + " initialEndAgents size " + this.initialEndAgents.size()+" ="+this.initialEndAgents, Logger.INFO_LEVEL);
			return false;
		}
		StateOfAgent solution = this.allReceivedAndAllCompatible();
		if (solution == null) {
			this.master.log("non compatible accepted " + solution + " \n", Logger.INFO_LEVEL);
			return false;
		}
		if (Logger.LOG_DEBUG) {
			this.master.log("terminate with SUCCESS : " + solution, Logger.INFO_LEVEL);
		}
		this.master.setEndState(Commands.SUCCESS);
		this.master.broadcastSystemMessage(AbstractMaster.MasterVarId);
		return true;
	}

	@Override
	public void handleSendMessage(Message msg) {}

	public String toString() {
		String ans = "\nIEAgents : " + this.initialEndAgents + " ==> " + this.order + "@" + Tools.arrayToString(this.tv) + "\n";
		Iterator<VarIdentifier> it = this.stateOfAgents.keySet().iterator();
		while (it.hasNext()) {
			VarIdentifier agId = it.next();
			StateOfAgent state = this.stateOfAgents.get(agId);
			ans += "" + agId + "=" + state.toString() + "\n";
		}
		return ans;
	}

	/**
	 * @return
	 */
	public HashList<VarIdentifier> getOrder() {
		return this.order;
	}

	/**
	 * @return
	 */
	private Double[] getTV() {
		return this.tv.clone();
	}
}
