package simple.dcsp.dba;

import kernel.StopException;
import kernel.VarIdentifier;
import kernel.events.VariableEvent;
import kernel.vars.integer.IntExternalVar;
import simple.SimpleAgent;
import simulation.parser.PVariable;
import util.Commands;
import util.VarAssignment;
import util.VarEventMask;

public class DBAagent extends SimpleAgent {

	public static int MODE_STARTUP = 0;
	public static int MODE_STOP = 0;
	public static int MODE_WAIT_OK = 1;
	public static int MODE_WAIT_IMPROVE = 2;
	protected int currMode = MODE_STARTUP;
	protected int myTerminationCounter = 0;
	protected boolean quasiLocalMinimum = false;
	protected boolean canMove = false;
	protected boolean consistent = false;
	protected int newValue;
	protected int currentEval;
	protected int counter = 0;
	protected int myImprove = 0;
	protected int maxCycle;
	protected int maxDistance = 0;
	protected int nbAgents = 1;
	protected int maxCycleCutOff = 1000;

	protected OkMsgHandler okMsgH;
	protected ImproveMsgHandler improveMsgH;
	protected int cycle = 0;

	public DBAagent(int agentId, String agName, String description) {
		super(agentId, agName, description);
		this.agentOrder = new DBATDisAO(this);
		this.communicator = new DBACommunicator(this);
		this.terminateMsgHandler = new DBATerminateMsgHandler(this);
		this.constraintHandler = new DBAConstraintHandler(this);
		this.okMsgH = new OkMsgHandler(this);
		this.improveMsgH = new ImproveMsgHandler(this);
	}
	
	public DBAagent(PVariable variable) {
		super(variable);
		this.agentOrder = new DBATDisAO(this);
		this.communicator = new DBACommunicator(this);
		this.terminateMsgHandler = new DBATerminateMsgHandler(this);
		this.constraintHandler = new DBAConstraintHandler(this);
		this.okMsgH = new OkMsgHandler(this);
		this.improveMsgH = new ImproveMsgHandler(this);
	}

	@Override
	public void configurePropagationEngine() {
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_HIGHER_THAN_LOCAL_VAR).add(VariableEvent.INSTANCIATE);
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_LOWER_THAN_LOCAL_VAR).add(VariableEvent.INSTANCIATE);
		this.constraintHandler.update();
	}

	public void addMsgHandlers() {
		this.communicator.addMsgHandler(this.okMsgH);
		this.communicator.addMsgHandler(this.improveMsgH);
	}

	// ==========================================================================

	public void initAgentView() {
		for (VarIdentifier varId : this.communicator.getNeighbors()) {
			this.agentView.add((IntExternalVar) this.getExternalVar(varId));
		}
	}

	@Override
	public void startDistributedSearch() throws StopException {
		if (this.currMode == MODE_STARTUP) {
			this.newValue = this.var.getDomain().getRandomValue();
			canMove = true;
			this.sendOk();
			this.currMode = MODE_WAIT_OK;
		}
		while (!this.isFinished()) {
			try {
				if (this.waitForArrivalMessages(0)) {
					while (this.currMode == MODE_WAIT_OK && !this.isFinished()) {
						this.blockingReceiveMessage(this.okMsgH.getHandlerType(), 100);
					}
					while (this.currMode == MODE_WAIT_IMPROVE && !this.isFinished()) {
						this.blockingReceiveMessage(this.improveMsgH.getHandlerType(), 100);
					}
				}
			} catch (InterruptedException e) {
				throw new StopException(e.getMessage());
			}
		}
	}

	// ////////////////////////////////////////////////////////////////////////////////////////////////////
	public void sendOk() {
		if (consistent) {
			myTerminationCounter++;
			if (myTerminationCounter == maxDistance) {
				// solution found send messages
				((DBATerminateMsgHandler) this.terminateMsgHandler).sendSolution();
				this.currMode = MODE_STOP;
				this.stopRunning(Commands.SUCCESS);
				return;
			}
		}
		if (this.cycle >= this.maxCycle) {
			((DBATerminateMsgHandler) this.terminateMsgHandler).sendNoSolution();
			this.currMode = MODE_STOP;
			this.stopRunning(Commands.FAILURE);
			return;
		}
		if (quasiLocalMinimum) {
			// increase weights of constraint violations
			((DBAConstraintHandler) this.constraintHandler).increaseWeightsOfConstraintViolations();
		}
		if (canMove) {
			// change my current value to the new one
			this.var.instantiate(newValue);
			canMove = false;
		}
		// this.agentView.clear();
		((DBAConstraintHandler) this.constraintHandler).violations.clear();
		this.currMode = DBAagent.MODE_WAIT_OK;
		// send an ok message to all neighbors
		this.okMsgH.sendOkMessages();

	}

	// //////////////////////////////////////////////////////////////////////////////////////////////
	public void processOk(VarAssignment varAss) {
		this.counter++;
		this.agentView.priorityMerge(varAss);
		int nbNeighbors = this.communicator.getNeighbors().size();
		if (this.counter == nbNeighbors) {
			this.counter = 0;
			this.cycle++;
			this.currMode = DBAagent.MODE_WAIT_IMPROVE;
			this.sendImprove();
		}

	}

	// //////////////////////////////////////////////////////////////////////////////////////////////
	public void sendImprove() {
		this.currentEval = ((DBAConstraintHandler) this.constraintHandler).computeEval(this.var.getValue());
		this.computePossibleMaxImprovement();
		if (currentEval == 0) {
			consistent = true;
		} else {
			consistent = false;
			myTerminationCounter = 0;
		}
		if (myImprove > 0) {
			canMove = true;
			quasiLocalMinimum = false;
		} else {
			canMove = false;
			quasiLocalMinimum = true;
		}
		// send an improve message to all neighbors
		this.improveMsgH.sendImproveMessages();

	}

	// ///////////////////////////////////////////////////////////////////////////////////////////

	private void computePossibleMaxImprovement() {
		this.myImprove = 0;
		for(int value : this.var.getDomain()){
			if (value != this.var.getValue()) {
				int eval = ((DBAConstraintHandler) this.constraintHandler).computeEval(value);
				int improve = this.currentEval - eval;
				if (improve > this.myImprove) {
					this.myImprove = improve;
					this.newValue = value;
				}
			}
		}
	}

	// ///////////////////////////////////////////////////////////////////////////////////////////
	public void processImprove(int improve, int eval, int terminateCounter, VarIdentifier senderId) {
		this.counter++;
		this.myTerminationCounter = terminateCounter;
		if (improve > this.myImprove) {
			this.canMove = false;
			this.quasiLocalMinimum = false;
		}
		if (improve == this.myImprove && this.agentOrder.isHigher(senderId)) {
			this.canMove = false;
		}
		if (eval > 0)
			this.consistent = false;

		int nbNeighbors = this.communicator.getNeighbors().size();
		if (this.counter == nbNeighbors) {
			this.counter = 0;
			this.cycle++;
			this.sendOk();
		}
	}

	// ///////////////////////////////////////////////////////////////////////////////////
	public void setGraphProprieties(int nbAgents, int maxDis, boolean graphIsCyclic) {
		this.nbAgents = nbAgents;
		this.maxDistance = maxDis;

		if (graphIsCyclic) {
			this.maxCycle = this.maxCycleCutOff;
		} else {
			this.maxCycle = (this.nbAgents * this.nbAgents);
		}
	}

	public void setMaxCycleCutOff(int maxCycleCutOff) {
		this.maxCycleCutOff = maxCycleCutOff;
	}

}
