package simple.dcsp.sbt;

import kernel.AbstractAgent;
import kernel.StopException;
import kernel.VarIdentifier;
import kernel.com.Message;
import kernel.com.MessageHandler;
import kernel.com.TerminateMsgHandler;
import kernel.events.AgentEvent;
import kernel.events.VariableEvent;
import kernel.vars.integer.IntExternalVar;
import simple.SimpleAgent;
import simulation.parser.PVariable;
import util.Commands;
import util.HashList;
import util.Instantiation;
import util.VarAssignment;
import util.VarEventMask;

/**
 * Implementation of the improved version of SBT;
 * 
 * @author Wahbi Mohamed
 * 
 */
public class SBTAgent extends SimpleAgent {

	protected CPAMsgHandler cpaMsgH = null;
	protected BackCPAMsgHandler backCPAMsgH = null;

	/**
	 * Constructors
	 * 
	 * @param description
	 * @param agentName
	 */
	public SBTAgent(int agentId, String agName, String description) {
		super(agentId, agName, description);
		// -------------- Messages Handler ------------------//
		this.cpaMsgH = new CPAMsgHandler(this);
		this.backCPAMsgH = new BackCPAMsgHandler(this);
		this.terminateMsgHandler = new TermMsgHandler(this);
	}
	
	/**
	 * Constructors
	 * 
	 * @param description
	 * @param agentName
	 */
	public SBTAgent(PVariable variable) {
		super(variable);
		// -------------- Messages Handler ------------------//
		this.cpaMsgH = new CPAMsgHandler(this);
		this.backCPAMsgH = new BackCPAMsgHandler(this);
		this.terminateMsgHandler = new TermMsgHandler(this);
	}

	// /////////////////////////////////////////////////////////////////////////////////////////////////

	@Override
	public void configurePropagationEngine() {
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_HIGHER_THAN_LOCAL_VAR).add(VariableEvent.INSTANCIATE);
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_HIGHER_THAN_LOCAL_VAR).add(VariableEvent.DESINSTANCIATE);
		this.constraintHandler.update();
	}

	/**
	 * 
	 */
	@Override
	public void initAgentView() {
		HashList<VarIdentifier> agents = this.agentOrder.getHigherAgents();
		for(VarIdentifier varId : agents) {
			if (!this.communicator.isNeighbor(varId)) {
				VarAssignment var = new VarAssignment(varId, Integer.MIN_VALUE);
				IntExternalVar extVar = this.makeExternalVar(var);
				this.agentView.add(extVar);
			} else this.agentView.add((IntExternalVar) this.getExternalVar(varId));
		}
	}

	/**
	 * 
	 */
	public void addMsgHandlers() {
		this.communicator.addMsgHandler(this.cpaMsgH);
		this.communicator.addMsgHandler(this.backCPAMsgH);
	}

	/**
	 * Start distributed search from the initializing agent
	 * 
	 * @throws StopExeption
	 */
	public void startDistributedSearch() throws StopException {
		// if(emptyDomain)
		// throw new StopException(Commands.FAILURE);
		if (this.agentOrder.getMyRank() == 0) {
			this.assignCPA();
		}
	}

	// /////////////////////////////////////////////////////////////////////////////////////////////////
	/**
	 * this function is called when ever we need to instantiate the current partial assignment, if doesn't succeed it launches a backtrack event
	 */
	public void assignCPA() throws StopException {
		if (this.valSelector.hasNextValue()) {
			int val = this.valSelector.getNextValue();
			this.var.instantiate(val);
			Instantiation cpa = this.agentView.toInstantiation();
			VarIdentifier nextVarId = this.getNextAgent();
			if (nextVarId == null) {
				throw new StopException(Commands.SUCCESS);
			} else {
				cpa.add(this.var.getVarAssignment());
				this.cpaMsgH.sendCPAMessage(cpa, nextVarId);
			}
		} else {
			this.backtrack();
		}
	}

	private VarIdentifier getNextAgent() {
		return this.agentOrder.getNextOf(this.getVarIdentifier());
	}

	// /////////////////////////////////////////////////////////////////////////////////////////////////
	/**
	 * called when contradiction is generated.
	 * 
	 * @throws StopException
	 */
	public void backtrack() throws StopException {
		VarIdentifier varId = this.agentOrder.getPreviousOf(this.getVarIdentifier());
		if (varId == null) {
			throw new StopException(Commands.FAILURE);
		} else {
			this.backCPAMsgH.sendBack(varId);
			this.agentView.clear();
			this.getPropagationEngine().flushEvents();
			this.var.getDomain().reset();

		}
	}

	public void processCPA(Instantiation cpa) throws StopException {
		this.agentView.priorityMerge(cpa);
		if (var.getDomainSize() == 0) {
			this.backtrack();
		} // FC
		else this.assignCPA();
	}

	public void processBack() throws StopException {
		int myVal = var.getValue();
		var.removeValue(myVal, VariableEvent.NOCAUSE);
		assignCPA();
	}

	public class CPAMsgHandler extends MessageHandler {

		public CPAMsgHandler(SimpleAgent agent) {
			super(agent, Commands.CPA);
		}

		public void processMessage(Message msg) throws StopException {
			Instantiation cpa = (Instantiation) msg.get(Commands.CPA);
			processCPA(cpa);
		}

		public void sendCPAMessage(Instantiation cpa, VarIdentifier receiverId) {
			Message msg = createMessage();
			msg.addContents(Commands.CPA, cpa);
			msg.setReceiverId(receiverId);
			sendMessage(msg);
		}
	}

	public class BackCPAMsgHandler extends MessageHandler {

		public BackCPAMsgHandler(AbstractAgent agent) {
			super(agent, Commands.BACKTRACK);
		}

		@Override
		public void processMessage(Message msg) throws StopException {
			processBack();
		}

		/**
		 * @param instantiation
		 * @param ng
		 */
		public void sendBack(VarIdentifier agId) {
			Message msg = this.createMessage();
			msg.setReceiverId(agId);
			sendMessage(msg);
		}
	}
	
	public class TermMsgHandler extends TerminateMsgHandler {

		public TermMsgHandler(AbstractAgent agent) {
			super(agent);
		}

		@Override
		public boolean processTerminationDetection(Message msg) throws StopException {
			return false;
		}

		@Override
		public void handleAgentEvent(AgentEvent agentEvent) {
			if (agentEvent instanceof kernel.events.AgentFinishedEvent) {
				this.askMasterToTerminate(true);
			}
		}
	}
}
