package simulation;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import kernel.AbstractAgent;
import kernel.AbstractMaster;
import kernel.VarIdentifier;
import simulation.parser.InstanceParser;
import simulation.parser.PConstraint;
import util.Commands;
import util.HashList;
import dischoco.Protocols;

/**
 * A class that models a distributed problem. When a simulation is performed, this class encapsulates all informations for resolving distributed
 * problem: Master, Agents, variables and constraints. Also it allows to load problem from XML file. This class defines methods to create any type of
 * constraint.
 * 
 * @authors 
 * <a href="mailto:ezzahir@ensa-agadir.ac.ma?subject=DisChoco &cc=wahbi@lirmm.fr &cc=bessiere@lirmm.fr">
 * Redoaune Ezzahir</a> and 
 * <a href="mailto:wahbi@lirmm.fr?subject=DisChoco &cc=ezzahir@ensa-agadir.ac.ma &cc=bessiere@lirmm.fr">
 * Mohamed Wahbi</a>
 *
 */
public class DisProblem {
	
	/**
	 * The name of the problem to solve
	 */
	private String pbName;
	/**
	 * A map of the agents that are given by their IDs
	 */

	public HashMap<VarIdentifier, AbstractAgent> agents;
	/**
	 * The Master controlling the resolution of the problem
	 */
	public AbstractMaster master = null;
	
	InstanceParser parser;
	private String protocol;
	private String problemFile;

	/**
	 * Constructs a new problem instance
	 * 
	 * @param name
	 *            the problem name
	 * @param master
	 *            the Master controlling the resolution
	 */
	public DisProblem(String name, AbstractMaster master) {
		this.agents = new HashMap<VarIdentifier, AbstractAgent>();
		this.pbName = name;
		this.master = master;
	}

	/**
	 * Constructs a new problem instance
	 * 
	 * @param the
	 *            algorithm name to solve the instance;
	 * @param master
	 *            the Master controlling the resolution
	 */
	public DisProblem(InstanceParser parser, String protocol) {
		this.parser = parser;
		this.agents = new HashMap<VarIdentifier, AbstractAgent>();
		this.pbName = parser.getProblemName();
		this.master = Protocols.getMaster(protocol);
		this.loadProblem(parser);
	}

	public DisProblem(String problemFile, String protocol) {
		parser = new InstanceParser(problemFile);
		parser.loadInstance(problemFile);
		parser.parse(false);
		this.agents = new HashMap<VarIdentifier, AbstractAgent>();
		this.pbName = parser.getProblemName();
		this.master = Protocols.getMaster(protocol);
		this.setProtocol(protocol);
		this.loadProblem(parser);
		this.problemFile = problemFile;
	}

	/**
	 * make and add a new agent to the problem
	 */
	public AbstractAgent makeAgent(VarIdentifier varId, String description) {
		if (this.getAgent(varId) != null)
			throw new Error("An agent with the same name was already created");
		AbstractAgent agent = this.master.makeAgent(this.getNumAgents() + 1, varId.toString(), description);
		this.agents.put(agent.getVarIdentifier(), agent);
		return agent;
	}

	/**
	 * make and add a new agent to the problem
	 */
	public AbstractAgent makeAgent(String name, String description) {
		VarIdentifier varId = new VarIdentifier(name);
		if (this.getAgent(varId) != null)
			throw new Error("An agent with the same name was already created");
		AbstractAgent agent = this.master.makeAgent(this.getNumAgents() + 1, name, description);
		this.agents.put(varId, agent);
		return agent;
	}
//
//	/**
//	 * Returns the agent with the given name
//	 * 
//	 * @param agentName
//	 *            the name of the agent to return
//	 * @return
//	 */
//	public AbstractAgent getAgent(VarIdentifier varId) {
//		Iterator<VarIdentifier> it = agents.keySet().iterator();
//		while (it.hasNext()) {
//			AbstractAgent ag = agents.get(it.next());
//			if (ag.getVarIdentifier().equals(varId))
//				return ag;
//		}
//		return null;
//	}

	/**
	 * Returns the agent with the given ID
	 * 
	 * @param agId
	 * 
	 */
	public AbstractAgent getAgent(VarIdentifier agId) {
		if (this.agents.containsKey(agId))
			return this.agents.get(agId);
		return null;
	}

	/**
	 * Returns the total number of Agents
	 */
	public int getNumAgents() {
		return this.agents.size();
	}

	/**
	 * Returns the map of agents by by their IDs
	 */
	public HashMap<VarIdentifier, AbstractAgent> getAgents() {
		return this.agents;
	}

	/**
	 * Returns the map of agents by by their IDs
	 */
	public Set<String> getAgentsNames() {
		Set<String> names = new TreeSet<String>();
		for(VarIdentifier id : this.agents.keySet()){
			names.add(this.agents.get(id).getAgentName());
		}
		return names;
	}

	/**
	 * Adds an new agent to the list of agents that participates in the resolution
	 */
	public void addAgent(VarIdentifier varId, AbstractAgent agent) {
		this.agents.put(varId, agent);
	}

	/**
	 * Returns the agent with the given ID
	 * 
	 * @param agId
	 * 
	 */
	public String getName(VarIdentifier agId) {
		if (this.agents.containsKey(agId))
			return this.agents.get(agId).getAgentName();
		return "A"+agId;
	}

	/**
	 * Set the list of the agents to the given one.
	 * 
	 * @param agents
	 */
	public void setAgents(Map<VarIdentifier, ? extends AbstractAgent> agents) {
		if (agents == null && this.agents != null) {
			this.agents.clear();
			this.agents = null;
		} else {
			this.agents.putAll(agents);
//			for(AbstractAgent agent : agents) {
//				this.addAgent(agent);
//			}
		}
	}

	/**
	 * Solve the problem
	 */
	public void solve() {
		DisSolverSimulator disSolver = new DisSolverSimulator(this);
		disSolver.solve();
	}

	/**
	 * Returns a string representation of the problem
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer(this.pbName + ": \n");
		Iterator<VarIdentifier> it = agents.keySet().iterator();
		while (it.hasNext()) {
			AbstractAgent ag = agents.get(it.next());
			sb.append(ag.toString() + "\n");
		}
		return sb.toString();
	}

	/**
	 * Loads Distributed Problem Instance from the given XML parser
	 * 
	 * @param parser
	 */

	public void loadProblem(InstanceParser parser) {
//		String category = this.master.getCategory();
//		if (!category.equals(HYBRID) && !parser.getCategory().equals(category)) {
//			parser.convertToCategory(category);
//		}
		if (parser.isComplexModel() && this.master.getModel().equals(Commands.SIMPLE)){
			parser.convertToSimpleModel();
		}
		if (parser.getModel().equals(Commands.SIMPLE)) {
			this.buildSimpleProblem(parser);
			this.groupLocalAgents();
		} else{
			this.buildComplexProblem(parser);
		}
	}

	/**
	 * A distributed problem with complex local problem (agents with multiple variables) can be decomposed by combining for each local variable a
	 * pseudo-agent or an agent. In such a case, for not counting the messages sent between the pseudo-agents, agents must know its local neighbors.
	 * So, a call to this method allows agents to know its local neighbors
	 */
	private void groupLocalAgents() {
		Map<String, HashList<VarIdentifier>> mapOfLocalAgents = this.parser.getMapOfLocalAgents();
		for(HashList<VarIdentifier> localAgents: mapOfLocalAgents.values()) {
			for(int i=0; i<localAgents.size()-1 ; i++) {
				VarIdentifier varI = localAgents.get(i);
				AbstractAgent Ai = agents.get(varI);
				for(int j=i+1; j<localAgents.size() ; j++) {
					VarIdentifier varJ = localAgents.get(j);
					AbstractAgent Aj = agents.get(varJ);
					Aj.addLocalAgent(varI);
					Ai.addLocalAgent(varJ);
				}
			}
		}
	}

	/**
	 * to build a simple problem instance (simple: one variable per agent)
	 */
	private void buildSimpleProblem(InstanceParser parser) {
		this.setAgents(this.master.makeAgents(parser.getMapOfVariables()));
//		for (PVariable variable : parser.getMapOfVariables().values()){
//			VarIdentifier owner = variable.getVarIdentifier();
//			SimpleAgent agent = (SimpleAgent) this.getAgent(owner);
//			variable.makeIntVariable(agent);
//		}
		for (PConstraint constraint : parser.getMapOfConstraints().values()) {
			constraint.makeConstraint();
		}
	}

	/**
	 * to build a complex problem (Complex: Agent with multiple variable)
	 * 
	 * @param parser
	 */

	private void buildComplexProblem(InstanceParser parser) {
	// TODO
	}

	public String getProblemName() {
		return this.pbName;
	}

	public String getProblemFile() {
		return this.problemFile;
	}

	public void clear() {
		parser.clear();
		this.parser = null;
	}

	public void setProtocol(String protocol) {
		this.protocol = protocol;
	}

	public String getProtocol() {
		return protocol;
	}
}
