package simulation;

import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Random;
import java.util.SortedSet;
import java.util.TreeSet;

import kernel.AbstractAgent;
import kernel.AbstractMaster;
import kernel.VarIdentifier;
import kernel.com.CommunicationInterface;
import kernel.com.Message;

/**
 * 
 * When using DisChoco for simulation the communication system is modeled by MailerAMDS class. The MailerAMDS class implements ChocoCommunication
 * interface. It serves as a message relay in the system and implements an Asynchronous Message Delay Simulator. The Mailer algorithm as presented in
 * the <a href="http://www.cs.bgu.ac.il/~am/My_Papers/Msg_Delay_DCR.pdf"> paper of Roie Zivan and Amnon Meisels</a>
 * 
 * 
 * @authors 
 * <a href="mailto:ezzahir@ensa-agadir.ac.ma?subject=DisChoco &cc=wahbi@lirmm.fr &cc=bessiere@lirmm.fr">
 * Redoaune Ezzahir</a> and 
 * <a href="mailto:wahbi@lirmm.fr?subject=DisChoco &cc=ezzahir@ensa-agadir.ac.ma &cc=bessiere@lirmm.fr">
 * Mohamed Wahbi</a>
 *
 */
public abstract class MailerAMDS implements DisCPSolver, CommunicationInterface {
	/**
	 * The distributed problem instance to be solve
	 */
	protected DisProblem problem;

	/**
	 * Message delay cost coefficient to simulate slow communication network (for example with TCP/IP packets transfer) . Its default value is 200 ms
	 */
	protected int msgDelayCostCoefficient = 100; // 100 ms;

	protected int delayCost = 100;
	/**
	 * A reference to the Master
	 */
	protected AbstractMaster master;
//	/** Non Concurrent Computation Time */
//	protected long NCCT = 0;
	/** Non-Concurrent Constraint Checks */
	protected long NCCCs = 0;

	/** A field stores the total number of loss message in the system */
	protected int numLossMsgs = 0;
	/**
	 * A field stores the total number of loss message in the system
	 */
	protected long numMsgs = 0;
	/**
	 * A field indicates whether evaluation with message delay is used or not.
	 */
	protected boolean evaluateMessageDelay = false;
	/**
	 * A field indicates whether evaluation with message loss is used or not.
	 */
	protected boolean evaluateMessageLoss = false;
	/**
	 * Default probability of message loss when it is activated (0.005).
	 */
	protected double msgLossProbability = 0.005;
	/**
	 * A flag indicates that a message is lost
	 */
	protected boolean lossMsg = false;
	/**
	 * A field to generate random delay
	 */
	protected Random rand;
	/**
	 * A queue of delayed massages used when simulating messages delay
	 */
	protected SortedSet<Message> delayedMsgsQueue;
	/**
	 * Logical time matrix for each connected pair of agents (delay update to ensure FIFO channel)
	 */
	protected volatile long[][] LTM;

	private int minDelayCost = 1;

	private int maxDelayCost = 1;
	/**
	 * Constraints Checks for each agent 
	 */
	protected HashMap<String, Long> CCs = null;

	/**
	 * Constructs the Mailer to perform simulation
	 * 
	 * @param disProb
	 *            the problem instance to be solved
	 */
	public MailerAMDS(DisProblem disProb) {
		this.problem = disProb;
		this.rand = new Random();
		this.delayedMsgsQueue = Collections.synchronizedSortedSet(new TreeSet<Message>(new Message.Comparator()));
		this.CCs = new HashMap<String, Long>();
	}

	/**
	 * Set the Mailer as communication system for each agent
	 */
	public void setMailerAsCommunicationInterface() {
		HashMap<VarIdentifier, AbstractAgent> agents = this.problem.getAgents();
		Iterator<VarIdentifier> it = agents.keySet().iterator();
		while (it.hasNext()) {
			AbstractAgent agent = agents.get(it.next());
			agent.setCommunicationInterface(this);
			this.register(agent.getVarIdentifier());
		}
		this.register(AbstractMaster.MasterVarId);
		this.master.setCommunicationInterface(this);
		this.LTM = new long[this.getNumAgents() + 1][this.getNumAgents() + 1];
	}

	/**
	 * Returns the total number of agent
	 */
	public int getNumAgents() {
		return this.problem.getNumAgents();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.CommunicationInterface#getNetworkLatency()
	 */
	@Override
	public int getNetworkLatency() {
		return this.msgDelayCostCoefficient;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.CommunicationInterface#setNetworkLatency(int)
	 */

	@Override
	public void setNetworkLatency(int msgDelayCostCoefficient) {
		this.msgDelayCostCoefficient = msgDelayCostCoefficient;
	}

	/**
	 * Return the total number of message lost
	 */
	public int getNbLossMsgs() {
		return numLossMsgs;
	}

	/**
	 * to perform a simulation of fast communication environment (Multiprocessors)
	 */
	public void setFastCommunication() {
		this.evaluateMessageDelay = false;
	}

	/**
	 * to perform a simulation of slow communication (Internet, LAN, ..)
	 */
	public void setSlowCommunication() {
		this.minDelayCost = 50;
		this.maxDelayCost = 100;
		this.evaluateMessageDelay = true;
	}

	/**
	 * Broadcast the given message to all registered agents
	 */
	public synchronized void broadcast(Message msg) {
		for (VarIdentifier rcv : this.getRegisteredAgentIds()) {
			if (rcv != msg.getSenderId() && !rcv.equals(AbstractMaster.MasterVarId)) {
				Message m = msg.clone();
				m.setReceiverId(rcv);
				this.sendMessage(m);
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.CommunicationInterface#sendMessage(com.Message)
	 */
	public synchronized void sendMessage(Message msg) {
		if (this.evaluateMessageLoss) {
			this.rand.setSeed(System.currentTimeMillis());
			if (this.rand.nextFloat() < this.msgLossProbability) {
				this.lossMsg = true;
				this.numLossMsgs++;
			}
		}
		if (!this.lossMsg) {
			if (this.evaluateMessageDelay) {
				this.uponReceiveMessage(msg);
			} else {
				this.directSend(msg);
			}
		}
		this.lossMsg = false;
	}

	/**
	 * When the Mailer receives a message it choose a delay for the message and add it to Delayed message queue.
	 */
	private synchronized void uponReceiveMessage(Message msg) {
		synchronized (this.delayedMsgsQueue) {
			this.chooseDelay(msg);
			if (!this.delayedMsgsQueue.add(msg))
				throw new Error("can't add a message to delayed queue " + "\nmessage : " + msg.toString() + "\nQueue : "
						+ this.delayedMsgsQueue.toString());
			this.deliverMessages();
		}
	}

	/**
	 * Send directly the message without adding a delay for it.
	 */

	private synchronized void directSend(Message msg) {
		this.numMsgs++;
		msg.setTotalNumMessage(numMsgs);
		this.updateAgentsLoad(msg);
		this.NCCCs = Math.max(this.NCCCs, msg.getNCCCs());
		VarIdentifier agentId = msg.getReceiverId();
		if (!this.isRegistered(agentId))
			throw new Error("send message to an unknown agent: agentId=" + agentId + ")");
		this.notifyReception(agentId, msg);
	}

	/**
	 * @param msg
	 */
	public abstract void updateAgentsLoad(Message msg);

	/**
	 * Deliver messages from delayed message queue to its recipients
	 */
	private synchronized int deliverMessages() {
		int delivered = 0;
		for (Iterator<Message> it = this.delayedMsgsQueue.iterator() ; it.hasNext() ; ++delivered) {
			Message msg = it.next();
			if (msg.getNCCCs() <= this.NCCCs) {
				msg.setNCCCs(this.NCCCs);
				it.remove();
				this.directSend(msg);
			} else {
				break;
			}
		}
		return delivered;
	}

	/**
	 * Notify recipient that a message is arrived
	 * 
	 * @param agentId
	 *            the ID of recipient of the message
	 * @param msg an arrived message
	 */
	private synchronized void notifyReception(VarIdentifier agentId, Message msg) {
		if (!agentId.equals(AbstractMaster.MasterVarId)){
			this.problem.getAgent(agentId).notifyReception(msg);
		} else {
			this.master.notifyReception(msg);
		}
	}

	/**
	 * Checks whether the agents are in stillness phase or not.
	 */
	protected boolean checkStillness() {
		for(VarIdentifier agentId : this.getRegisteredAgentIds()) {
			if (agentId.equals(AbstractMaster.MasterVarId)) {
				if (!this.master.isWaitingForMessage())
					return false;
			} else {
				if (!this.problem.getAgent(agentId).isWaitingForMessage()) {
					return false;
				}
			}
		}
		return !this.delayedMsgsQueue.isEmpty();
	}

	/**
	 * Called when stillness is detected
	 */
	protected int whenStillness() {
		if (this.delayedMsgsQueue.isEmpty()){
			return 0;
		} else {
			this.NCCCs = this.delayedMsgsQueue.first().getNCCCs();
			return this.deliverMessages();
		}
	}

	@Override
	public void setMessageDelay(int minCCs, int maxCCs) {
		this.minDelayCost = minCCs;
		this.maxDelayCost = maxCCs;
	}

	/**
	 * Choose a delay for the given message
	 */
	public synchronized void chooseDelay(Message msg) {
		this.updateAgentsLoad(msg);
		this.NCCCs = Math.max(this.NCCCs, msg.getNCCCs());
		VarIdentifier from = msg.getSenderId();
		VarIdentifier to = msg.getReceiverId();
		// int bytes;
		// try {
		// bytes = estimateObjectSize(msg);
		// } catch (java.io.NotSerializableException nse) {
		// bytes = 512;
		// }
		// int packet = bytes / 1500 + (bytes % 1500 == 0 ? 0 : 1);
		// int L = getNetworkLatency();
		// int delay = delayCost * (packet * L + (this.rand.nextBoolean() ? 1 : -1) * this.rand.nextInt(L / 10));
		int delay = (minDelayCost + this.rand.nextInt(maxDelayCost - minDelayCost));

		int i = from.getAgentId();
		int j = to.getAgentId();
		this.LTM[i][j] = Math.max(this.NCCCs + delay, this.LTM[i][j]);
		msg.setNCCCs(this.LTM[i][j]);
	}

	// /**
	// * Estimates the size in bytes of an object.
	// */
	// private static int estimateObjectSize(Object object)
	// throws NotSerializableException {
	// int size = 0;
	// try {
	// ByteArrayOutputStream baos = new ByteArrayOutputStream();
	// ObjectOutputStream oos = new ObjectOutputStream(baos);
	// oos.writeObject(object);
	// oos.flush();
	// size = baos.toByteArray().length;
	// oos.close();
	// baos.close();
	// } catch (NotSerializableException nse) {
	// throw (nse);
	// } catch (IOException ioe) {
	// ioe.printStackTrace();
	// }
	// return size;
	// }

	public DisProblem getDisProblem() {
		return this.problem;
	}

}
