package simulation.metric;

import java.util.ArrayList;
import java.util.Collection;

import kernel.AgentObserver;
import kernel.StopException;
import kernel.com.Message;
import kernel.events.AgentEvent;
import kernel.events.AgentFinishedEvent;
import util.Commands;
import util.Logger;

/**
 * A class for measuring the number of non concurrent constraint checks
 * 
 * @author Redouane EZZAHIR
 * 
 */
public class NCCCObserver extends AgentObserver implements MetricListener {

	/**
	 * The default cut-off value (1e8) of non-concurrent constraint checks
	 */
	protected long cutOffNCCCs = (long) 1e8;
	/**
	 * Current AgentC_Cost (NCCC) number of NCCCs of the observed agent
	 */
	protected long NCCCs = 0;
	/**
	 * The number of generic non-concurrent constraints checks
	 */
	public long nbGNCCCs = 0;
	public long nbCCs = 0;

	/**
	 * Constructs an observer for NCCCs
	 */
	public NCCCObserver() {
		super(Commands.NCCCs);
	}
	
	/**
	 * Constructs an observer for NCCCs
	 * @param cutOffNCCCs 
	 */
	public NCCCObserver(long cutOffNCCCs) {
		super(Commands.NCCCs);
		this.cutOffNCCCs = cutOffNCCCs;
	}

	/**
	 * Updates the current Non-Concurrent constraint checks to be the maximum between its value and the value attached in the message <code>msg</code>
	 * 
	 * @throws StopException
	 */
	@Override
	public boolean handleReceivedMessage(Message msg) throws StopException {
		long senderCost = msg.getNCCCs();
		this.NCCCs = Math.max(this.NCCCs, senderCost);
		if (this.NCCCs > this.cutOffNCCCs) {
			throw new StopException(Commands.FAILURE);// + " NCCCs Cut-off");// + " NCCCs Cut-off");
		}
		return true;
	}

	/**
	 * Attach the Current number of NCCCs of the observed agent to the message <code>msg</code>
	 */
	@Override
	public void handleSendMessage(Message msg) {
		if (this.myAgent.isDisSearchStarted()) {
//			msg.addContents(Commands.NCCCs, NCCCs);
			msg.addContents(Commands.GNCCCs, nbGNCCCs);
			msg.addContents(Commands.CCs, nbCCs);
			msg.setNCCCs(NCCCs);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void handleIncrementNbConstraintChecks(int n) {
		this.NCCCs += n;
		this.nbGNCCCs += n;
		this.nbCCs += n;
	}

	@Override
	public MetricListener clone() {
		MetricListener clone = null;
		try {
			clone = (NCCCObserver) super.clone();
		} catch (CloneNotSupportedException e) {
			e.printStackTrace();
		}
		return clone;
	}

	@Override
	public Collection<Metric> getEndStatistics() {
		ArrayList<Metric> m = new ArrayList<Metric>();
		m.add(new Metric(Commands.NCCCs, this.NCCCs, Metric.max));
		m.add(new Metric(Commands.GNCCCs, this.nbGNCCCs, Metric.max));
		m.add(new Metric(Commands.CCs, this.nbCCs, Metric.plus));
		return m;
	}

	/**
	 * This function Handles the event <code> {@link kernel.events.AgentFinishedEvent} </code> It print in the logger the end value of
	 * AgentC_Cost (NCCC)
	 */
	@Override
	public void handleAgentEvent(AgentEvent event) {
		if (event instanceof AgentFinishedEvent) {
			this.myAgent.log("AgentC_Cost (#ncccs) : " + this.NCCCs, Logger.INFO_LEVEL);
			this.myAgent.log("AgentC_Cost (#gncccs) : " + this.nbGNCCCs, Logger.INFO_LEVEL);
			this.myAgent.log("AgentC_Cost (#CCs) : " + this.nbCCs, Logger.INFO_LEVEL);
		}
	}

	@Override
	public void handleIncrementCycle() {}

}
