package simulation.metric;

import java.lang.management.ManagementFactory;
import java.lang.management.ThreadMXBean;
import java.util.ArrayList;
import java.util.Collection;

import kernel.AgentObserver;
import kernel.StopException;
import kernel.com.Message;
import kernel.events.AgentEvent;
import kernel.events.AgentFinishedEvent;
import util.Commands;
import util.Logger;

/**
 * A class for measuring the Non-Concurrent Computation-Time
 * 
 * @authors 
 * <a href="mailto:ezzahir@ensa-agadir.ac.ma?subject=DisChoco &cc=wahbi@users.sourceforge.net &cc=bessiere@lirmm.fr">
 * Redoaune Ezzahir</a> and 
 * <a href="mailto:wahbi@users.sourceforge.net?subject=DisChoco &cc=ezzahir@ensa-agadir.ac.ma &cc=bessiere@lirmm.fr">
 * Mohamed Wahbi</a>
 *
 */
public class NCCPUTimeObserver extends AgentObserver implements MetricListener {

	/**
	 * The default cut-off value (infinity) of non-concurrent constraint checks
	 */
	private double cutOffNCCPUTime = Long.MAX_VALUE;
	/**
	 * The current Non-Concurrent Computation-Time
	 */
	private double NCCPUTime;
	
	/**
	 * The moment when the agent moves from a pending state message to the message processing state 
	 * (i.e when the agent becomes active)
	 */
	private long currentBeginTime;
	/**
	 * Used to returns the total CPU time for the current thread in nanoseconds.
	 */
	private ThreadMXBean tmxb = null;

	/**
	 * constructs an observer
	 */
	public NCCPUTimeObserver() {
		super(Commands.NCCPUTime);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void initialize() {
		super.initialize();
		this.NCCPUTime = 0;
		this.tmxb = ManagementFactory.getThreadMXBean();
		this.myAgent.addAgentEventListener(this);
		this.currentBeginTime = this.tmxb.getThreadCpuTime(this.myAgent.getId());
	}

	/**
	 * Updates the current Non-Concurrent Computation-Time to be the maximum between its value and the value 
	 * attached in the message <code>msg</code>
	 */
	@Override
	public boolean handleReceivedMessage(Message msg) throws StopException {
		double senderCT = msg.getNCCPUTime();
		this.NCCPUTime = Math.max(this.NCCPUTime, senderCT);
		if (this.NCCPUTime > this.cutOffNCCPUTime) {
			throw new StopException(Commands.FAILURE);// + " NCCPUTime Cut-off");
		}
		return true;
	}

	/**
	  * 
	  */
	@Override
	public void handleSendMessage(Message msg) {
		if (this.myAgent.isDisSearchStarted()) {
			if (this.currentBeginTime > -1) {
				long tid = this.myAgent.getId();
//				this.NCCPUTime += (tmxb.getCurrentThreadCpuTime() - this.currentBeginTime) / 10e6;
//				this.currentBeginTime = tmxb.getCurrentThreadCpuTime();
				this.NCCPUTime += (this.tmxb.getThreadCpuTime(tid) - this.currentBeginTime) / 10e6;
				this.currentBeginTime = this.tmxb.getThreadCpuTime(tid);
			} else {
				throw new Error("In Non-Concurrent Computation-Time !");
			}
			msg.setNCCPUTime(NCCPUTime);
		}		
	}

	@Override
	public void handleAgentEvent(AgentEvent event) {
		if (event instanceof AgentFinishedEvent) {
			this.myAgent.log("AgentC_Cost (#ncct) : " + this.NCCPUTime, Logger.INFO_LEVEL);
		}
	}

	public void setCutOffNCCPUTime(long cutOffNCCPUTime) {
		this.cutOffNCCPUTime = cutOffNCCPUTime;
	}

	/**
	 * not used for this metric
	 */
	@Override
	public void handleIncrementNbConstraintChecks(int n) {}

	/**
	 * not used for this metric
	 */
	@Override
	public void handleIncrementCycle() {}

	@Override
	public MetricListener clone() {
		MetricListener clone = null;
		try {
			clone = (NCCPUTimeObserver) super.clone();
		} catch (CloneNotSupportedException e) {
			e.printStackTrace();
		}
		return clone;
	}

	@Override
	public Collection<Metric> getEndStatistics() {
		ArrayList<Metric> m = new ArrayList<Metric>();
		m.add(new Metric(Commands.NCCPUTime, this.NCCPUTime, Metric.max));
		return m;
	}

}
