package simulation.model;

import java.io.File;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.List;
import kernel.AbstractMaster;
import simulation.DisCPSolver;
import simulation.DisProblem;
import simulation.DisSolverSimulator;
import simulation.model.expressions.ExpressionFactory;
import simulation.model.expressions.IntConstantExpression;
import simulation.model.expressions.IntVarExpression;
import simulation.model.expressions.IntegerExpression;
import simulation.model.expressions.VariableExpression;
import simulation.parser.InstanceParser;
import simulation.parser.PAgent;
import simulation.parser.PConstraint;
import simulation.parser.PDomain;
import simulation.parser.PIntensionConstraint;
import simulation.parser.PPredicate;
import simulation.parser.PVariable;
import simulation.parser.PredicateTokens;
import util.Commands;

public class SModel {

	/**
	 * @return the disProblem
	 */
	public DisProblem getDisProblem() {
		return this.disProblem;
	}

	/**
	 * @param disProblem the disProblem to set
	 */
	public void setDisProblem(DisProblem disProblem) {
		this.disProblem = disProblem;
	}

	private static int instanceI = 0;

	public static InstanceParser parser;

	private DisProblem disProblem;

	public SModel(String problemName) {
		instanceI++;
		parser = new InstanceParser(problemName);
//		parser.loadInstance(problemName);
//		parser.parse(false);
	}

	public void addConstraint(PConstraint c) {
		parser.addConstraint(c);
	}

	public DisCPSolver makeDisSolver(String algoName) {
		disProblem = new DisProblem(parser, algoName);
		return new DisSolverSimulator(disProblem);
	}

	public PAgent makeAgent(String name, String description) {
		int id = parser.getMapOfAgents().size() + 1;
		PAgent agent = new PAgent(name, id, description);
		parser.addAgent(agent);
		return agent;
	}

	public PVariable makeVariable(PAgent ag, int min, int max) {
		Integer values[] = new Integer[max - min + 1];
		for (int i = 0 ; i < values.length ; i++){ 
			values[i] = min + i;
		}
		PDomain domain = new PDomain("D" + ag.getId(), values);
		parser.addDomain(domain);
		PVariable var = new PVariable(ag, ag.getName(), 0, ag.getDescription(), domain);
		parser.getMapOfVariables().put(var.getName(), var);
		return var;
	}

	public static PVariable makeVariable(PAgent ag, Integer[] values) {
		int i = parser.getMapOfVariables().size();
		String name = "X" + i;
		PDomain domain = new PDomain("D" + i, values);
		parser.addDomain(domain);
		PVariable var = new PVariable(ag, name, 0, ag.getDescription(), domain);
		parser.addVariable(var);
		return var;
	}

	public void saveXML(String path) {
		StringBuffer sb = new StringBuffer();
		sb.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
		sb.append("<instance>\n");
		sb.append(parser.presentationXML());
		sb.append(parser.agentsXML());
		sb.append(parser.domainsXML());
		sb.append(parser.variablesXML());
		sb.append(parser.relationsXML());
		sb.append(parser.functionsXML());
		sb.append(parser.predicatesXML());
		sb.append(parser.constraintsXML());
		sb.append(getGuiPresentation());
		sb.append("</instance>\n\n");
		try {
			File directory = new File(path + File.separator + parser.getProblemName());
			if (!directory.exists()) {
				directory.mkdir();
			}
			String file = directory.getPath() + File.separator + parser.getProblemName() + "_" + instanceI + ".xml";
			FileWriter writer = new FileWriter(file);
			writer.write(sb.toString());
			writer.close();
		} catch (java.io.IOException e) {
			e.printStackTrace();
		}

	}

	/**
	 * @return
	 */
	private String getGuiPresentation() {
		return 	
		"\t<GuiPresentation " +
		"type=\"Random DisCSP\" " +
		"name=\""+ parser.getProblemName() +"\" " +
		"model=\""+ Commands.SIMPLE +"\" " +
		"nbAgents=\""+parser.getNbAgents()+"\" " +
		"domainSize=\""+parser.getDomainSize()+"\" " +
		"graphDensity=\"0\" " +
		"tightness=\"0\" " +
		"nbConstraints=\""+parser.getNbConstraints()+"\" " +
		"nbTuples=\"0\" />\n";
	}

	// ///////////////////////// ABS EXPRESSIONS
	// ///////////////////////////////////////////////////////
	public static IntegerExpression abs(int cste) {
		return ExpressionFactory.abs(new IntConstantExpression(cste));
	}

	public static IntegerExpression abs(PVariable x) {
		return ExpressionFactory.abs(new IntVarExpression(x));
	}

	public static IntegerExpression abs(IntegerExpression x) {
		return ExpressionFactory.abs(x);
	}

	// ///////////////////////// PLUS EXPRESSIONS
	// ///////////////////////////////////////////////////////
	public static IntegerExpression plus(PVariable x, int cste) {
		return ExpressionFactory.plus(new IntVarExpression(x), new IntConstantExpression(cste));
	}

	public static IntegerExpression plus(PVariable x, PVariable y) {
		return ExpressionFactory.plus(new IntVarExpression(x), new IntVarExpression(y));
	}

	public static IntegerExpression plus(IntegerExpression a, IntegerExpression b) {
		return ExpressionFactory.plus(a, b);
	}

	// ///////////////////////// MOINS EXPRESSIONS
	// ////////////////////////////////////////////////////////

	public static IntegerExpression minus(PVariable x, int cste) {
		return ExpressionFactory.minus(new IntVarExpression(x), new IntConstantExpression(cste));
	}

	public static IntegerExpression minus(PVariable x, PVariable y) {
		return ExpressionFactory.minus(new IntVarExpression(x), new IntVarExpression(y));
	}

	public static IntegerExpression minus(IntegerExpression a, IntegerExpression b) {
		return ExpressionFactory.minus(a, b);
	}

	// //////////////////////////////////////////////////////////////////////////
	// ///////////////////////////////////
	// //////////////////////// Equality constraints
	// //////////////////////////////////////////////////////////////
	// //////////////////////////////////////////////////////////////////////////
	// ////////////////////////////////////
	public static PConstraint eq(PVariable x, PVariable y) {
		return creatEqConstraint(PredicateTokens.EQ, new IntVarExpression(x), new IntVarExpression(y));
	}

	public static PConstraint eq(PVariable x, int a) {
		return eq(new IntVarExpression(x), new IntConstantExpression(a));
	}

	public static PConstraint eq(PVariable x, IntegerExpression a) {
		return eq(new IntVarExpression(x), a);
	}

	public static PConstraint eq(IntegerExpression x, IntegerExpression y) {
		return creatEqConstraint(PredicateTokens.EQ, x, y);
	}

	public static PConstraint neq(PVariable x, PVariable y) {
		return creatEqConstraint(PredicateTokens.NE, new IntVarExpression(x), new IntVarExpression(y));
	}

	public static PConstraint neq(PVariable x, IntegerExpression a) {
		return neq(new IntVarExpression(x), a);
	}

	public static PConstraint neq(IntVarExpression x, IntegerExpression y) {
		return creatEqConstraint(PredicateTokens.NE, x, y);
	}

	public static PConstraint leq(PVariable x, PVariable y) {
		return creatEqConstraint(PredicateTokens.LE, new IntVarExpression(x), new IntVarExpression(y));
	}

	public static PConstraint leq(PVariable x, IntegerExpression a) {
		return leq(new IntVarExpression(x), a);
	}

	public static PConstraint leq(IntegerExpression x, IntegerExpression y) {
		return creatEqConstraint(PredicateTokens.LE, x, y);
	}

	public static PConstraint geq(PVariable x, PVariable y) {
		return creatEqConstraint(PredicateTokens.GE, new IntVarExpression(x), new IntVarExpression(y));
	}

	public static PConstraint geq(PVariable x, IntegerExpression a) {
		return geq(new IntVarExpression(x), a);
	}

	public static PConstraint geq(IntegerExpression x, int cste) {
		return creatEqConstraint(PredicateTokens.GE, x, new IntConstantExpression(cste));
	}

	public static PConstraint geq(IntegerExpression x, IntegerExpression y) {
		return creatEqConstraint(PredicateTokens.GE, x, y);
	}

	public static PConstraint gt(PVariable x, PVariable y) {
		return creatEqConstraint(PredicateTokens.GT, new IntVarExpression(x), new IntVarExpression(y));
	}

	public static PConstraint gt(PVariable x, IntegerExpression a) {
		return gt(new IntVarExpression(x), a);
	}

	public static PConstraint gt(IntegerExpression x, IntegerExpression y) {
		return creatEqConstraint(PredicateTokens.GT, x, y);
	}

	public static PConstraint lt(PVariable x, PVariable y) {
		return creatEqConstraint(PredicateTokens.LT, new IntVarExpression(x), new IntVarExpression(y));
	}

	public static PConstraint lt(PVariable x, IntegerExpression a) {
		return lt(new IntVarExpression(x), a);
	}

	public static PConstraint lt(IntegerExpression x, IntegerExpression y) {
		return creatEqConstraint(PredicateTokens.LT, x, y);
	}

	private static PConstraint creatEqConstraint(String eqtype, IntegerExpression x, IntegerExpression y) {
		String name = "C" + parser.getMapOfConstraints().size();
		PVariable[] scope = buildPVariable(x, y);
		PPredicate predicate = null;
		int i = parser.getMapOfPredicates().size();
		String[] formalAndEffectiveParametersExpression = makeFormalAndEffectiveParametersExpression(x, y);
		String functionalExpression = eqtype + "(" + mekeFunctionalExpression(x, y) + ")";
		predicate = new PPredicate("P" + i, formalAndEffectiveParametersExpression[0], functionalExpression);
		parser.addPredicate(predicate);
		return new PIntensionConstraint(name, scope, predicate, formalAndEffectiveParametersExpression[1]);
	}

	private static PVariable[] buildPVariable(IntegerExpression... list) {
		ArrayList<PVariable> all = new ArrayList<PVariable>();
		for (IntegerExpression exp : list) {
			List<? extends VariableExpression> exprs = exp.getVariableExpressions();
			for (VariableExpression v : exprs) {
				if (v != null) {
					PVariable var = ((IntVarExpression) v).getPVariable();
					if (var != null)
						all.add(var);
				}
			}
		}
		PVariable[] a = new PVariable[all.size()];
		return all.toArray(a);
	}

	private static String[] makeFormalAndEffectiveParametersExpression(IntegerExpression... list) {
		ArrayList<VariableExpression> all = new ArrayList<VariableExpression>();
		for (IntegerExpression exp : list) {
			all.addAll(exp.getVariableExpressions());
		}
		String param[] = new String[] {"", ""};
		for (int i = 0 ; i < all.size() ; i++) {
			param[0] = param[0] + "int X" + i + " ";
			param[1] = param[1] + ((IntVarExpression) all.get(i)).getSymbol() + " ";
		}
		return param;
	}

	private static String mekeFunctionalExpression(IntegerExpression... list) {
		String functionExp = "";
		int fromIndex = 0;
		int i = 0;
		for (IntegerExpression v : list) {
			i++;
			functionExp += v.getFunctionalExpression(fromIndex) + (i < list.length? "," : "");
			fromIndex += v.getVariableExpressions().size();
		}
		return functionExp;
	}

	/**
	 * 
	 */
	public void convertToCSPModel() {
		parser.convertToCSPModel();
	}

	public static void setParser(InstanceParser parser) {
		SModel.parser = parser;
	}

	public InstanceParser getParser() {
		return parser;
	}

	/**
	 * @return
	 */
	public String getProblemName() {
		return this.disProblem.getProblemName();
	}

	/**
	 * @return
	 */
	public AbstractMaster getMaster() {
		return this.disProblem.master;
	}

}
