package simulation.parser;

import static simulation.parser.InstanceTokens.HYBRID;
import static simulation.parser.InstanceTokens.OPTIMIZATION;
import static simulation.parser.InstanceTokens.SATISFACTION;
import java.io.File;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.TreeMap;
import javax.swing.JInternalFrame;
import javax.swing.JPanel;
import javax.swing.JTable;
import kernel.Constraint;
import kernel.VarIdentifier;
import kernel.order.AgentComplexity;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import simulation.parser.PredicateTokens.RelationalOperator;
import util.Commands;
import util.HashList;
import util.VarAssignment;
import util.graph.UndirectedGraph;
import util.tree.AbstractTree;

/**
 * This class corresponds to a Java parser that uses DOM (Document Object Model) to parse DisCSP and WDisCSP instances in format "XDisCSP 1.0". <br>
 * Here, we assume that the instance is well-formed (valid). This class is given for illustration purpose. Feel free to adapt it ! Inspired from model
 * XCSP 2.1 proposed by christophe lecoutre
 * 
 * @version 1.0
 */
public class InstanceParser {

	public static final String VERSION = "version 1.0 (October 18, 2012)";;
	private Document document = null;
	private String problemName = null;
	private String format = InstanceTokens.XDisCSP_1_0;
	private String type = InstanceTokens.DisCSP;
	private String benchmark = Commands.RandomTKC;
	/** Simple or complex */
	private String model = InstanceTokens.SIMPLE_MODEL;
	/** Constraint model : Totally (TKC) or Partial known (PKC) Constraints*/
	private String constraintModel = InstanceTokens.TKC;
	private int maxConstraintArity = 0;
	private Map<String, PDomain> mapOfDomains = null;
	private Map<String, PAgent> mapOfAgents = null;
	private Map<String, PVariable> mapOfVariables = null;
	private Map<String, PRelation> mapOfRelations = null;
	private Map<String, PFunction> mapOfFunctions = null;
	private Map<String, PPredicate> mapOfPredicates = null;
	private Map<String, PConstraint> mapOfConstraints = null;
	private Map<String, HashList<VarIdentifier>> mapOfLocalAgents = null;
	private int nbExtensionConstraints = 0;
	private int nbSoftIntensionConstraints = 0;
	private int nbIntensionConstraints = 0;
	private int nbGlobalConstraints = 0;

	/**
	 * Used to determine if elements of the instance must be displayed when parsing.
	 */
	private boolean displayInstance = false;

	/**
	 * Used for DCOP
	 */
	private Integer maximalCost = Constraint.INFINITY;
	/**
	 * Used for DCOP
	 */
	private Integer initialCost = 0;

	private String satisfiable = "?";

	private String minViolatedConstraints = "?";
	
	/** The category of the problem defined by its constraints: satisfaction, optimization or hybrid */
	private String category;

	public InstanceParser(String problemName) {
		this.mapOfAgents = new TreeMap<String, PAgent>();
		this.mapOfVariables = new TreeMap<String, PVariable>();
		this.mapOfDomains = new TreeMap<String, PDomain>();
		this.mapOfConstraints = new TreeMap<String, PConstraint>();
		this.mapOfPredicates = new TreeMap<String, PPredicate>();
		this.mapOfFunctions = new TreeMap<String, PFunction>();
		this.mapOfRelations = new TreeMap<String, PRelation>();
		this.problemName = problemName;
	}

	public void setPKCModel() {
		this.constraintModel = InstanceTokens.PKC;
	}

	public void setComplexModel() {
		this.model = InstanceTokens.COMPLEX_MODEL;
	}

	public void setSimpleModel() {
		this.model = InstanceTokens.SIMPLE_MODEL;
	}
	
	public String getType() {
		return this.type;
	}

	public String getModel() {
		return this.model;
	}

	public String getCategory() {
		return this.category;
	}

	public String getBenchmark() {
		return this.benchmark;
	}

	public String getProblemName() {
		return this.problemName;
	}

	public void setProblemName(String name) {
		this.problemName = name;
	}

	public int getMaxConstraintArity() {
		return this.maxConstraintArity;
	}

	public String getFormat() {
		return this.format;
	}

	public Map<String, PConstraint> getMapOfConstraints() {
		return this.mapOfConstraints;
	}

	public Map<String, PRelation> getMapOfRelations() {
		return this.mapOfRelations;
	}

	public Map<String, PAgent> getMapOfAgents() {
		return this.mapOfAgents;
	}

	public Map<String, PVariable> getMapOfVariables() {
		return this.mapOfVariables;
	}

	public Map<String, PDomain> getMapOfDomains() {
		return this.mapOfDomains;
	}

	public Map<String, PPredicate> getMapOfPredicates() {
		return this.mapOfPredicates;
	}

	public Map<String, PFunction> getMapOfPFunctions() {
		return this.mapOfFunctions;
	}

	/**
	 * @return 
	 * 
	 */
	public Map<String, HashList<VarIdentifier>> getMapOfLocalAgents() {
		return this.mapOfLocalAgents;
	}

	public int getNbExtensionConstraints() {
		return this.nbExtensionConstraints;
	}

	public int getNbIntensionConstraints() {
		return this.nbIntensionConstraints;
	}

	public int getNbGlobalConstraints() {
		return this.nbGlobalConstraints;
	}

	public String getConstraintsCategory() {
		return (nbExtensionConstraints > 0? "E" : "") + (nbIntensionConstraints > 0? "I" : "") + (nbGlobalConstraints > 0? "G" : "");
	}

	public String getSatisfiable() {
		return this.satisfiable;
	}

	public String getMinViolatedConstraints() {
		return this.minViolatedConstraints;
	}

	/**
	 * Build a DOM object that corresponds to the file whose name is given. <br>
	 * The file must represent a CSP instance according to format XCSP 2.1
	 * 
	 * @param fileName : the name of a file representing a CSP instance.
	 */
	public void loadInstance(String fileName) {
		this.document = XMLManager.load(fileName);
	}

	private void parsePresentation(Element presentation) {
		//================================================================================================
		// name = instance ...
		this.problemName = presentation.getAttribute(InstanceTokens.NAME.trim());
		// type in {DisCSP, DCOP, MaxDisCSP} 
		this.type = presentation.getAttribute(InstanceTokens.TYPE.trim());
		this.type = (this.type.length() == 0 || this.type.equals("?"))? InstanceTokens.DisCSP : this.type;
		// benchmark = instance ...
		this.benchmark = presentation.getAttribute(InstanceTokens.BENCHMARK.trim());
		// maxConstraintArity = 2..n
		String s = presentation.getAttribute(InstanceTokens.MAX_CONSTRAINT_ARITY.trim());
		this.maxConstraintArity = (s.length() == 0 || s.equals("?"))? -1 : Integer.parseInt(s);
		// Simple or Complex
		this.model = presentation.getAttribute(InstanceTokens.MODEL.trim()); 
		// SATISFACTION, OPTIMIZATION, HYBRID
		this.category = presentation.getAttribute(InstanceTokens.CATEGORY.trim()); 
		// constraintModel in {TKC, PKC}
		this.constraintModel = presentation.getAttribute(InstanceTokens.CONSTRAINT_MODEL.trim());
		// XDisCSP 1.0
		this.format = presentation.getAttribute(InstanceTokens.FORMAT.trim());
		//================================================================================================
		if (displayInstance) System.out.println(this.presentationXML());
		s = presentation.getAttribute(InstanceTokens.NB_SOLUTIONS).trim();
		this.satisfiable = s.length() == 0 || s.equals("?")? "unknown" : s.equals("0")? "false" : "true";
		s = presentation.getAttribute(InstanceTokens.MIN_VIOLATED_CONSTRAINTS).trim();
		this.minViolatedConstraints = satisfiable.equals("true")? "0" : s.length() == 0 || s.equals("?")? "unknown" : s;
	}

	public String presentationXML() {
		return "\t<presentation " +
				InstanceTokens.NAME+"=\"" + problemName + "\" " + 
				InstanceTokens.TYPE+"=\"" + type + "\" " + 
				InstanceTokens.BENCHMARK+"=\"" + benchmark + "\" " + 
				InstanceTokens.MODEL + "=\"" + model + "\" " + 
				InstanceTokens.CONSTRAINT_MODEL + "=\"" + constraintModel + "\" "+
				InstanceTokens.FORMAT+" =\"" + format + "\" " + 
				"\"/>\n";
	}

	private Integer[] parseDomainValues(int nbValues, String domValues) {
		int cnt = 0;
		Integer[] values = new Integer[nbValues];
		StringTokenizer st = new StringTokenizer(domValues);
		while (st.hasMoreTokens()) {
			String token = st.nextToken();
			int position = token.indexOf(InstanceTokens.DISCRETE_INTERVAL_SEPARATOR);
			if (position == -1) {
				values[cnt++] = Integer.parseInt(token);
			} else {
				int min = Integer.parseInt(token.substring(0, position));
				int max = Integer.parseInt(token.substring(position + InstanceTokens.DISCRETE_INTERVAL_SEPARATOR.length()));
				for (int j = min ; j <= max ; j++){
					values[cnt++] = j;
				}
			}
		}
		return values;
	}

	private PDomain parseDomain(Element domain) {
		String name = domain.getAttribute(InstanceTokens.NAME);
		int nbValues = Integer.parseInt(domain.getAttribute(InstanceTokens.NB_VALUES));
		Integer[] values = parseDomainValues(nbValues, domain.getTextContent());
		Arrays.sort(values);
		if (nbValues != values.length) throw new RuntimeException();
		return new PDomain(name, values);
	}

	private void parseDomains(Element domainsXML) {
		int nbDomains = Integer.parseInt(domainsXML.getAttribute(InstanceTokens.NB_DOMAINS));
		NodeList domainList = domainsXML.getElementsByTagName(InstanceTokens.DOMAIN);
		if (nbDomains != domainList.getLength()) throw new RuntimeException();
		for (int i = 0 ; i < nbDomains ; i++) {
			PDomain domain = parseDomain((Element) domainList.item(i));
			this.mapOfDomains.put(domain.getName(), domain);
		}
		if (displayInstance) System.out.println(domainsXML());
	}

	public String domainsXML() {
		StringBuffer sb = new StringBuffer();
		sb.append("\t").append(InstanceTokens.DOMAINS).append(' ');
		sb.append(InstanceTokens.NB_DOMAINS).append("=\"" + this.mapOfDomains.size() + "\">\n");
		for (PDomain domain : this.mapOfDomains.values()) {
			sb.append("\t\t" + domain);
		}
		sb.append("\t</").append(InstanceTokens.DOMAINS).append(">\n");
		return sb.toString();
	}

	private PAgent parseAgent(Element agent) {
		String name = agent.getAttribute(InstanceTokens.NAME);
		int agId = Integer.parseInt(agent.getAttribute(InstanceTokens.ID));
		String description = agent.getAttribute(InstanceTokens.DESCRIPTION);
		return new PAgent(name, agId, description);
	}

	private void parseAgents(Element agentsXML) {
		int nbAgents = Integer.parseInt(agentsXML.getAttribute(InstanceTokens.NB_AGENTS));
		NodeList agentsList = agentsXML.getElementsByTagName(InstanceTokens.AGENT);
		if (nbAgents != agentsList.getLength()) throw new RuntimeException();
		for (int i = 0 ; i < nbAgents ; i++) {
			PAgent agent = parseAgent((Element) agentsList.item(i));
			this.mapOfAgents.put(agent.getName(), agent);
		}
		if (displayInstance) System.out.println(agentsXML());
	}

	public String agentsXML() {
		StringBuffer sb = new StringBuffer();
		sb.append("\t<agents " + InstanceTokens.NB_AGENTS + "=\"" + this.mapOfAgents.size() + "\">\n");
		for (PAgent agent : mapOfAgents.values()) {
			sb.append("\t\t" + agent+"\n");
		}
		sb.append("\t</agents>\n");
		return sb.toString();
	}

	private PVariable parseVariable(Element variableElement) {
		String name = variableElement.getAttribute(InstanceTokens.NAME);
		String agentName = variableElement.getAttribute(InstanceTokens.AGENT);
		int id = Integer.parseInt(variableElement.getAttribute(InstanceTokens.ID));
		String domainName = variableElement.getAttribute(InstanceTokens.DOMAIN);
		String description = variableElement.getAttribute(InstanceTokens.DESCRIPTION);
		return new PVariable(mapOfAgents.get(agentName), name, id, description, mapOfDomains.get(domainName));
	}

	private void parseVariables(Element variablesXML) {
		int nbVariables = Integer.parseInt(variablesXML.getAttribute(InstanceTokens.NB_VARIABLES));
		NodeList variablesList = variablesXML.getElementsByTagName(InstanceTokens.VARIABLE);
		if (nbVariables != variablesList.getLength()) throw new RuntimeException();
		for (int i = 0 ; i < variablesList.getLength() ; i++) {
			PVariable variable = parseVariable((Element) variablesList.item(i));
			mapOfVariables.put(variable.getName(), variable);
		}
		if (displayInstance) System.out.println(variablesXML());
	}

	public String variablesXML() {
		StringBuffer buf = new StringBuffer();
		buf.append("\t<variables nbVariables=\"" + this.mapOfVariables.size() + "\">\n");
		for (PVariable variable : mapOfVariables.values()) {
			buf.append("\t\t" + variable+"\n");
		}
		buf.append("\t</variables>\n");
		return buf.toString();
	}

	private PRelation parseRelationTuples(String name, int arity, int nbTuples, String semantics, String textContent) {
		int[][] tuples = new int[nbTuples][arity];
		StringTokenizer st = new StringTokenizer(textContent, InstanceTokens.WHITE_SPACE + InstanceTokens.TUPLES_SEPARATOR);
		for (int i = 0 ; i < tuples.length ; i++){
			for (int j = 0 ; j < arity ; j++){
				tuples[i][j] = Integer.parseInt(st.nextToken());
			}
		}
		return new PRelation(name, arity, nbTuples, semantics, tuples);
	}

	private PRelation parseSoftRelationTuples(String name, int arity, int nbTuples, String semantics, String textContent, String textDefaultCost) {
		int[][] tuples = new int[nbTuples][arity];
		Integer[] weights = new Integer[nbTuples];
		StringTokenizer st = new StringTokenizer(textContent, InstanceTokens.WHITE_SPACE + InstanceTokens.TUPLES_SEPARATOR);
		Integer currentCost = new Integer(-2);
		for (int i = 0 ; i < nbTuples ; i++) {
			String token = st.nextToken();
			int costFlagPosition = token.lastIndexOf(InstanceTokens.COST_SEPARATOR);
			if (costFlagPosition != -1) {
				String cost = token.substring(0, costFlagPosition);
				currentCost = cost.equals(InstanceTokens.INFINITY)? Constraint.INFINITY : Integer.parseInt(cost);
				token = token.substring(costFlagPosition + 1);
			}
			weights[i] = currentCost;
			tuples[i][0] = Integer.parseInt(token);
			for (int j = 1 ; j < arity ; j++)
				tuples[i][j] = Integer.parseInt(st.nextToken());
		}
		int defaultCost = textDefaultCost.equals(InstanceTokens.INFINITY)? Constraint.INFINITY : Integer.parseInt(textDefaultCost);
		return new PSoftRelation(name, arity, nbTuples, semantics, tuples, weights, defaultCost);
	}

	private PRelation parseRelation(Element relation) {
		String name = relation.getAttribute(InstanceTokens.NAME);
		int arity = Integer.parseInt(relation.getAttribute(InstanceTokens.ARITY));
		int nbTuples = Integer.parseInt(relation.getAttribute(InstanceTokens.NB_TUPLES));
		String semantics = relation.getAttribute(InstanceTokens.SEMANTICS);
		if (semantics.equals(InstanceTokens.SOFT)) { 
			String defaultCost = relation.getAttribute(InstanceTokens.DEFAULT_COST);
			return parseSoftRelationTuples(name, arity, nbTuples, semantics, relation.getTextContent(), defaultCost);
		} else {
			return parseRelationTuples(name, arity, nbTuples, semantics, relation.getTextContent());
		}
	}

	private void parseRelations(Element relations) {
		if (relations == null) return;
		int nbRelations = Integer.parseInt(relations.getAttribute(InstanceTokens.NB_RELATIONS));
		NodeList relationsList = relations.getElementsByTagName(InstanceTokens.RELATION);
		if (nbRelations != relationsList.getLength()) throw new RuntimeException();
		for (int i = 0 ; i < relationsList.getLength() ; i++) {
			PRelation relation = parseRelation((Element) relationsList.item(i));
			this.mapOfRelations.put(relation.getName(), relation);
		}
		if (displayInstance) System.out.println(relationsXML());
	}

	public String relationsXML() {
		StringBuffer sb = new StringBuffer();
		if (this.mapOfRelations.size() > 0) {
			sb.append("\t<relations nbRelations=\"" + this.mapOfRelations.size() + "\">\n");
			for (PRelation relation : mapOfRelations.values()) {
				sb.append("\t\t" + relation);
			}
			sb.append("\t</relations>\n");
		}
		return sb.toString();
	}

	private PPredicate parsePredicate(Element predicateElement) {
		String name = predicateElement.getAttribute(InstanceTokens.NAME);
		Element parameters = (Element) predicateElement.getElementsByTagName(InstanceTokens.PARAMETERS).item(0);
		Element expression = (Element) predicateElement.getElementsByTagName(InstanceTokens.EXPRESSION).item(0);
		Element functional = (Element) expression.getElementsByTagName(InstanceTokens.FUNCTIONAL).item(0);
		return new PPredicate(name, parameters.getTextContent().trim(), functional.getTextContent().trim());
	}

	private void parsePredicates(Element predicates) {
		if (predicates == null) return;
		int nbPredicates = Integer.parseInt(predicates.getAttribute(InstanceTokens.NB_PREDICATES));
		NodeList predicatesList = predicates.getElementsByTagName(InstanceTokens.PREDICATE);
		if (nbPredicates != predicatesList.getLength()) throw new RuntimeException("\nwhen parsing predicates="+predicates);
		for (int i = 0 ; i < nbPredicates ; i++) {
			PPredicate predicate = parsePredicate((Element) predicatesList.item(i));
			this.mapOfPredicates.put(predicate.getName(), predicate);
		}
		if (displayInstance) System.out.println(predicatesXML());
	}

	public String predicatesXML() {
		StringBuffer sb = new StringBuffer();
		if (this.mapOfPredicates.size() > 0) {
			sb.append("\t<predicates nbPredicates=\"" + this.mapOfPredicates.size() + "\">\n");
			for(PPredicate predicate : mapOfPredicates.values()){
				sb.append("\t\t" + predicate);
			}
			sb.append("\t</predicates>\n");
		}
		return sb.toString();
	}

	private PFunction parseFunction(Element function) {
		String name = function.getAttribute(InstanceTokens.NAME);
		Element parameters = (Element) function.getElementsByTagName(InstanceTokens.PARAMETERS).item(0);
		Element expression = (Element) function.getElementsByTagName(InstanceTokens.EXPRESSION).item(0);
		Element functional = (Element) expression.getElementsByTagName(InstanceTokens.FUNCTIONAL).item(0);
		return new PFunction(name, parameters.getTextContent(), functional.getTextContent());
	}

	private void parseFunctions(Element functions) {
		if (functions == null) return;
		int nbFunctions = Integer.parseInt(functions.getAttribute(InstanceTokens.NB_FUNCTIONS));
		NodeList functionsList = functions.getElementsByTagName(InstanceTokens.FUNCTION);
		if (nbFunctions != functionsList.getLength()) throw new RuntimeException();
		for (int i = 0 ; i < nbFunctions ; i++) {
			PFunction function = parseFunction((Element) functionsList.item(i));
			this.mapOfFunctions.put(function.getName(), function);
		}
		if (displayInstance) System.out.println(functionsXML());
	}

	public String functionsXML() {
		StringBuffer sb = new StringBuffer();
		if (this.mapOfFunctions.size() > 0) {
			sb.append("\t<functions nbFunctions=\"" + this.mapOfFunctions.size() + "\">\n");
			for (PFunction function :  this.mapOfFunctions.values()) {
				sb.append("\t\t" + function);
			}
			sb.append("\t</functions>\n");
		}
		return sb.toString();
	}

	private PVariable[] parseScope(String scope) {
		StringTokenizer st = new StringTokenizer(scope, " ");
		PVariable[] involvedVariables = new PVariable[st.countTokens()];
		for (int i = 0 ; i < involvedVariables.length ; i++)
			involvedVariables[i] = mapOfVariables.get(st.nextToken());
		return involvedVariables;
	}

	private int searchIn(String str, PVariable[] variables) {
		for (int i = 0 ; i < variables.length ; i++)
			if (variables[i].getName().equals(str)) return i;
		return -1;
	}

	private PConstraint parseElementConstraint(String name, PVariable[] scope, Element parameters) {
		StringTokenizer st = new StringTokenizer(
				Toolkit.insertWhitespaceAround(parameters.getTextContent().trim(), InstanceTokens.BRACKETS), 
				InstanceTokens.WHITE_SPACE);
		PVariable index = this.mapOfVariables.get(st.nextToken()); // index is necessarily a variable
		st.nextToken(); // token [ skipped
		List<Object> table = new ArrayList<Object>();
		String token = st.nextToken();
		while (!token.equals("]")) {
			Object object = this.mapOfVariables.get(token);
			if (object == null) object = Integer.parseInt(token);
			table.add(object);
			token = st.nextToken();
		}
		token = st.nextToken();
		Object value = this.mapOfVariables.get(token);
		if (value == null)
			value = Integer.parseInt(token);
		return new PElement(name, scope, index, table.toArray(new Object[table.size()]), value);
	}

	private PConstraint parseWeightedSumConstraint(String name, PVariable[] scope, Element parameters) {
		NodeList nodeList = parameters.getChildNodes();
		StringTokenizer st = new StringTokenizer(nodeList.item(0).getTextContent(), InstanceTokens.WHITE_SPACE + "[{}]");
		int[] coeffs = new int[scope.length];
		while (st.hasMoreTokens()) {
			int coeff = Integer.parseInt(st.nextToken());
			int position = searchIn(st.nextToken(), scope);
			coeffs[position] += coeff;
		}
		RelationalOperator operator = RelationalOperator.getRelationalOperatorFor(nodeList.item(1).getNodeName());
		int limit = Integer.parseInt(nodeList.item(2).getTextContent().trim());
		return new PWeightedSum(name, scope, coeffs, operator, limit);
	}

	private String buildStringRepresentationOf(Element parameters) {
		NodeList nodeList = parameters.getChildNodes();
		StringBuffer sb = new StringBuffer();
		for (int i = 0 ; i < nodeList.getLength() ; i++) {
			Node node = nodeList.item(i);
			if (node.getNodeName().equals(InstanceTokens.NIL)) {
				sb.append(" ").append(InstanceTokens.NIL).append(" ");
			} else {
				sb.append(Toolkit.insertWhitespaceAround(node.getTextContent(), InstanceTokens.BRACKETS));
			}
		}
		return sb.toString();
	}

	private PConstraint parseCumulativeConstraint(String name, PVariable[] scope, Element parameters) {
		StringTokenizer st = new StringTokenizer(buildStringRepresentationOf(parameters), InstanceTokens.WHITE_SPACE + "{}");
		st.nextToken(); // token '[' skipped
		List<Task> tasks = new ArrayList<Task>();
		String token = st.nextToken();
		while (!token.equals("]")) {
			Object origin = mapOfVariables.get(token);
			if (origin == null)
				origin = token.equals(InstanceTokens.NIL)? null : Integer.parseInt(token);
			token = st.nextToken();
			Object duration = mapOfVariables.get(token);
			if (duration == null)
				duration = token.equals(InstanceTokens.NIL)? null : Integer.parseInt(token);
			token = st.nextToken();
			Object end = mapOfVariables.get(token);
			if (end == null)
				end = token.equals(InstanceTokens.NIL)? null : Integer.parseInt(token);
			token = st.nextToken();
			Object height = mapOfVariables.get(token);
			if (height == null)
				height = Integer.parseInt(token);
			tasks.add(new Task(origin, duration, end, height));
			token = st.nextToken();
		}
		int limit = Integer.parseInt(st.nextToken());
		return new PCumulative(name, scope, tasks.toArray(new Task[tasks.size()]), limit);
	}

	private PConstraint parseConstraint(Element constraintXMl) {
		PConstraint constraint = null;
		String name = constraintXMl.getAttribute(InstanceTokens.NAME);
		int arity = Integer.parseInt(constraintXMl.getAttribute(InstanceTokens.ARITY));
		if (arity > maxConstraintArity) maxConstraintArity = arity;
		PVariable[] scope = parseScope(constraintXMl.getAttribute(InstanceTokens.SCOPE));
		String reference = constraintXMl.getAttribute(InstanceTokens.REFERENCE);
		if (mapOfRelations.containsKey(reference)) {
			nbExtensionConstraints++;
			constraint = new PExtensionConstraint(name, scope, mapOfRelations.get(reference));
		} else if (mapOfFunctions.containsKey(reference)) {
			setNbSoftIntensionConstraints(getNbSoftIntensionConstraints() + 1);
			Element parameters = (Element) constraintXMl.getElementsByTagName(InstanceTokens.PARAMETERS).item(0);
			constraint = new PIntensionConstraint(name, scope, mapOfFunctions.get(reference), parameters.getTextContent());
		} else if (mapOfPredicates.containsKey(reference)) {
			nbIntensionConstraints++;
			Element parameters = (Element) constraintXMl.getElementsByTagName(InstanceTokens.PARAMETERS).item(0);
			constraint = new PIntensionConstraint(name, scope, mapOfPredicates.get(reference), parameters.getTextContent());
		} else {
			nbGlobalConstraints++;
			String lreference = reference.toLowerCase();
			Element parameters = (Element) constraintXMl.getElementsByTagName(InstanceTokens.PARAMETERS).item(0);

			if (lreference.equals(InstanceTokens.getLowerCaseGlobalNameOf(InstanceTokens.ALL_DIFFERENT))) {
				constraint = new PAllDifferent(name, scope);
			} else if (lreference.equals(InstanceTokens.getLowerCaseGlobalNameOf(InstanceTokens.ELEMENT))) {
				constraint = parseElementConstraint(name, scope, parameters);
			} else if (lreference.equals(InstanceTokens.getLowerCaseGlobalNameOf(InstanceTokens.WEIGHTED_SUM))) {
				constraint = parseWeightedSumConstraint(name, scope, parameters);
			} else if (lreference.equals(InstanceTokens.getLowerCaseGlobalNameOf(InstanceTokens.CUMULATIVE))) {
				constraint = parseCumulativeConstraint(name, scope, parameters);
			} else {
				throw new UnsupportedOperationException("Problem with the reference " + reference);
			}
		}
		if (constraint != null && constraintModel.equals(Commands.PKC)) {
			int ownerId = Integer.parseInt(constraintXMl.getAttribute(InstanceTokens.OWNER_ID));
			constraint.setOwnerId(ownerId);
		}
		return constraint;
	}

	private void parseConstraints(Element constraints) {
		if (type.equals(InstanceTokens.DCOP)) {
			String maxCost = constraints.getAttribute(InstanceTokens.MAXIMAL_COST);
			setMaximalCost(maxCost.equals(InstanceTokens.INFINITY)? Constraint.INFINITY : Integer.parseInt(maxCost));
			String initCost = constraints.getAttribute(InstanceTokens.INITIAL_COST);
			setInitialCost(initCost.equals("")? 0 : Integer.parseInt(initCost));
		}
		int nbConstraints = Integer.parseInt(constraints.getAttribute(InstanceTokens.NB_CONSTRAINTS));
		NodeList constraintsList = constraints.getElementsByTagName(InstanceTokens.CONSTRAINT);
		if (nbConstraints != constraintsList.getLength()) throw new RuntimeException();
		for (int i = 0 ; i < nbConstraints ; i++) {
			PConstraint constraint = parseConstraint((Element) constraintsList.item(i));
			mapOfConstraints.put(constraint.getName(), constraint);
		}
		if (displayInstance) System.out.println(constraintsXML());
	}

	public String constraintsXML() {
		StringBuffer sb = new StringBuffer();
		if (this.mapOfConstraints.size() > 0) {
			sb.append("\t<constraints nbConstraints=\"" + this.mapOfConstraints.size() + "\">\n");
			for (PConstraint constraint : mapOfConstraints.values()) {
				sb.append("\t\t" + constraint+"\n");
			}
			sb.append("\t</constraints>\n");
		}
		return sb.toString();
	}

	/**
	 * Parse the DOM object that has been loaded.
	 * 
	 * @param displayInstance : if <code> true </code>, elements of the instance will be displayed.
	 */
	public void parse(boolean displayInstance) {
		this.displayInstance = displayInstance;
		Element instance = document.getDocumentElement();
		parsePresentation((Element) instance.getElementsByTagName(InstanceTokens.PRESENTATION).item(0));
		parseAgents((Element) instance.getElementsByTagName(InstanceTokens.AGENTS).item(0));
		parseDomains((Element) instance.getElementsByTagName(InstanceTokens.DOMAINS).item(0));
		parseVariables((Element) instance.getElementsByTagName(InstanceTokens.VARIABLES).item(0));

		parseRelations((Element) instance.getElementsByTagName(InstanceTokens.RELATIONS).item(0));
		parseFunctions((Element) instance.getElementsByTagName(InstanceTokens.FUNCTIONS).item(0));
		parsePredicates((Element) instance.getElementsByTagName(InstanceTokens.PREDICATES).item(0));
		parseConstraints((Element) instance.getElementsByTagName(InstanceTokens.CONSTRAINTS).item(0));
		
//		try {
//			parseRelations((Element) instance.getElementsByTagName(InstanceTokens.RELATIONS).item(0));
//		} catch (NullPointerException e) {
//			e.printStackTrace();
//		}
//		try {
//			parseFunctions((Element) instance.getElementsByTagName(InstanceTokens.FUNCTIONS).item(0));
//		} catch (NullPointerException e) {
//			e.printStackTrace();
//		}
//		try {
//			parsePredicates((Element) instance.getElementsByTagName(InstanceTokens.PREDICATES).item(0));
//		} catch (NullPointerException e) {
//			e.printStackTrace();
//		}
//		try {
//			parseConstraints((Element) instance.getElementsByTagName(InstanceTokens.CONSTRAINTS).item(0));
//		} catch (NullPointerException e) {
//			e.printStackTrace();
//		}

		this.mapOfLocalAgents = new HashMap<String, HashList<VarIdentifier>>();
		for(PAgent agent : this.mapOfAgents.values()) {
			this.mapOfLocalAgents.put(agent.getName(), new HashList<VarIdentifier>());
		}
	}

	public void addPredicate(PPredicate predicate) {
//		if(mapOfPredicates.values().contains(predicate)) return ;
		for (PPredicate prd : mapOfPredicates.values()) {
			if (predicate.equals(prd)) {
				predicate.name = prd.name; 
				return;
			}
		}
		this.mapOfPredicates.put(predicate.getName(), predicate);
	}

	public void addPredicate(PRelation relation) {
		for (PRelation rlt : mapOfRelations.values()) {
			if (relation.equals(rlt)) {
				relation.name = rlt.name;
				return;
			}
		}
		this.mapOfRelations.put(relation.getName(), relation);
	}

	public void addFunction(PFunction function) {
		for (PFunction fct : mapOfFunctions.values()) {
			if (function.equals(fct)) {
				function.name = fct.name;
				return;
			}
		}
		this.mapOfFunctions.put(function.getName(), function);
	}

	public void addDomain(PDomain domain) {
		for (PDomain dom : mapOfDomains.values()) {
			if (domain.equals(dom)) {
				domain.name = dom.name;
				return;
			}
		}
		this.mapOfDomains.put(domain.getName(), domain);
	}

	public void addAgent(PAgent agent) {
//		for(PAgent ag : mapOfAgents.values()){
//			if(agent.getId() == ag.getId()){
//				agent.name = ag.name;
//				agent.description = ag.description;
//				return;
//			}
//		}
		if (mapOfAgents.containsKey(agent.getName())) {
			System.err.println("An agent with the same name already exist");
			System.exit(-1);
		}
		this.mapOfAgents.put(agent.getName(), agent);
	}

	public void addVariable(PVariable var) {
		if (mapOfVariables.containsKey(var.getName())) {
			System.err.println("An other variable  with the same name  already exist");
			System.exit(-1);
		}
		mapOfVariables.put(var.getName(), var);
	}

	public void addConstraint(PConstraint c) {
		if (!this.mapOfConstraints.containsKey(c.getName())) {
			this.mapOfConstraints.put(c.getName(), c);
		}
	}

	public void convertToSimpleModel() {
		if (this.model.equals(Commands.COMPLEX)) {
			this.model = Commands.SIMPLE;
			Map<String, PAgent> newAgents = new TreeMap<String, PAgent>();
			Iterator<String> it = this.mapOfVariables.keySet().iterator();
			int i = 1;
			while (it.hasNext()) {
				PVariable variable = this.mapOfVariables.get(it.next());
				String agentName = variable.getAgent().getName();
				PAgent agent = new PAgent(variable.getName(), i++ , variable.getAgent().getName());
				variable.setAgent(agent);
				variable.setId(1);
				newAgents.put(agent.getName(), agent);
				this.mapOfLocalAgents.get(agentName).add(variable.getVarIdentifier());
			}
			this.mapOfAgents.clear();
			this.mapOfAgents = newAgents;
		} else {
			throw new UnsupportedOperationException();
		}
	}

	public void convertToCSPModel() {
		Iterator<String> it = this.mapOfVariables.keySet().iterator();
		HashMap<String, String> map = new HashMap<String, String>();
		int i = 1;
		while (it.hasNext()) {
			String key = it.next();
			PVariable var = this.mapOfVariables.get(key);
			String name = var.getName();
			String newName = name.substring(0, 1)+i;
			map.put(name, newName);
			var.setName(newName);
			var.setId(0);
			i++;
		}
		it = this.mapOfConstraints.keySet().iterator();
		while (it.hasNext()) {  
			String key = it.next();
			PConstraint c = this.mapOfConstraints.get(key);
			if(c instanceof PIntensionConstraint){
				String effective = ((PIntensionConstraint)c).effectiveParametersExpression;
				String[] names = effective.split(" ");
				((PIntensionConstraint)c).effectiveParametersExpression = "";
				for(String name : names){
					if(map.containsKey(name)){
						((PIntensionConstraint)c).effectiveParametersExpression += map.get(name)+" ";
					} else {
						((PIntensionConstraint)c).effectiveParametersExpression += name+" ";
					}
				}
			}
		}
	}

	public void initInstanceParser(String problemName) {
		this.mapOfAgents = null;
		this.mapOfVariables = null;
		this.mapOfDomains = null;
		this.mapOfConstraints = null;
		this.mapOfPredicates = null;
		this.mapOfFunctions = null;
		this.mapOfRelations = null;
		this.problemName = null;
		this.mapOfAgents = new TreeMap<String, PAgent>();
		this.mapOfVariables = new TreeMap<String, PVariable>();
		this.mapOfDomains = new TreeMap<String, PDomain>();
		this.mapOfConstraints = new TreeMap<String, PConstraint>();
		this.mapOfPredicates = new TreeMap<String, PPredicate>();
		this.mapOfFunctions = new TreeMap<String, PFunction>();
		this.mapOfRelations = new TreeMap<String, PRelation>();
		this.problemName = problemName;
	}

	public int getNbAgents() {
		return this.mapOfAgents.size();
	}

	public int getDomainSize() {
		Iterator<String> it = this.mapOfDomains.keySet().iterator();
		return this.mapOfDomains.get(it.next()).getNbValues();
	}

	public int getNbConstraints() {
		int nbConstraints = 0;
		nbConstraints += this.getNbExtensionConstraints();
		nbConstraints += this.getNbGlobalConstraints();
		nbConstraints += this.getNbIntensionConstraints();
		return nbConstraints;
	}

	public Document getDocument() {
		return this.document;
	}

	public int getNbVariables() {
		return this.mapOfVariables.size();
	}

	public void clear() {
		this.mapOfAgents.clear();
		this.mapOfConstraints.clear();
		this.mapOfDomains.clear();
		this.mapOfFunctions.clear();
		this.mapOfRelations.clear();
		this.mapOfRelations.clear();
		this.mapOfVariables.clear();
	}

	public HashMap<VarIdentifier, AgentComplexity> getAgentsComplexity() {
		HashMap<VarIdentifier, AgentComplexity> complexities = new HashMap<VarIdentifier, AgentComplexity>();
		Iterator<String> it = this.mapOfVariables.keySet().iterator();
		while (it.hasNext()) {
			String key = it.next();
			PVariable variable = this.mapOfVariables.get(key);
			VarIdentifier varId = variable.getVarIdentifier();
			if (!complexities.containsKey(varId)) {
				complexities.put(varId, new AgentComplexity());
			}
			complexities.get(varId).setLabel(variable.getName());
			complexities.get(varId).setMaxDomain(variable.getDomain().getNbValues());
			complexities.get(varId).IncrementNbLocalVars();
		}
		it = this.mapOfConstraints.keySet().iterator();
		while (it.hasNext()) {
			String key = it.next();
			PConstraint constraint = this.mapOfConstraints.get(key);
			PVariable[] scope = constraint.getScope();
			for (int v = 0 ; v < scope.length ; v++) {
				VarIdentifier varId = scope[v].getVarIdentifier();
				complexities.get(varId).IncrementNbConstraints();
				for (int n = 0 ; n < scope.length ; n++) {
					if (v != n) {
						VarIdentifier neighbor = scope[n].getVarIdentifier();
						complexities.get(varId).addNeighbor(neighbor, constraint.getName());
					}
				}
			}
		}
		return complexities;
	}

	public StringBuffer toStringBuffer(){
		StringBuffer sb = new StringBuffer();
		sb.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
		sb.append("<instance>\n");
		sb.append(this.presentationXML());
		sb.append(this.agentsXML());
		sb.append(this.domainsXML());
		sb.append(this.variablesXML());
		sb.append(this.predicatesXML());
		sb.append(this.functionsXML());
		sb.append(this.relationsXML());
		sb.append(this.constraintsXML());
		sb.append("</instance>\n\n");
		return sb;
	}

	public void saveXML(String path, String name) {
		StringBuffer buf = this.toStringBuffer();
		try {
			File directory = new File(path + File.separator);
			if (!directory.exists()) { directory.mkdir(); }
			String file = directory.getPath() + File.separator + name;
			FileWriter writer = new FileWriter(file);
			writer.write(buf.toString());
			writer.close();
		} catch (java.io.IOException e) {
			e.printStackTrace();
		}

	}

	/**
	 * @param agId
	 * @return
	 */
	public String getAgentName(int agId) {
		for(String key : this.mapOfAgents.keySet()){
			PAgent agent = this.mapOfAgents.get(key);
			if(agent.getId() == agId){
				return agent.getName();
			}
		}
		return null;
	}
	

	/** 
	 * Returns the constraint graph in DOT format
	 * @return a String representation of the constraint graph in DOT format
	 */
	public String toDOT () {
		return getConstraintGraph().toDOTString();
	}

	/**
	 * @return
	 */
	public boolean isSimpleModel() {
		return this.model.equals(InstanceTokens.SIMPLE_MODEL);
	}

	/**
	 * @return
	 */
	public boolean isComplexModel() {
		return this.model.equals(InstanceTokens.COMPLEX_MODEL);
	}

//	/**
//	 * @param var
//	 * @return
//	 */
//	private ArrayList<PVariable> getNeighborVars(PVariable var) {
//		ArrayList<PVariable> neighbors = new ArrayList<PVariable>();
//		for(String key : this.mapOfConstraints.keySet()){
//			PConstraint constraint = this.mapOfConstraints.get(key);
//			PVariable[] scope = constraint.getScope();
//			if(scope[0].getAgentId()==var.getAgentId() && scope[0].getId()==var.getId()){
//				neighbors.add(scope[1]);
//			} else if(scope[1].getAgentId()==var.getAgentId() && scope[1].getId()==var.getId()){
//				neighbors.add(scope[0]);
//			}
//		}
//		return neighbors;
//	}

	/**
	 * @return
	 */
	public UndirectedGraph<VarIdentifier> getConstraintGraph() {
		UndirectedGraph<VarIdentifier> graph = new UndirectedGraph<VarIdentifier>();
		for(PVariable variable : mapOfVariables.values()){
			graph.addVertex(variable.getVarIdentifier());
		}
		if(this.isComplexModel()) {
			for(PAgent agent : mapOfAgents.values()){
				graph.createCluster(agent.getId());
			}
			for(PVariable variable : mapOfVariables.values()){
				graph.addToCluster(variable.getAgentId(), variable.getVarIdentifier());
			}
		}
		for(PConstraint constraint : this.mapOfConstraints.values()){
			PVariable[] scope = constraint.getScope();
			String name = constraint.getName();
			if(scope.length == 1) { //FIXME
				//graph.addEdge(scope[0].getVarIdentifier(), scope[0].getVarIdentifier(), constraint.getName());
			} else if(scope.length == 2) {
				graph.addEdge(scope[0].getVarIdentifier(), scope[1].getVarIdentifier(), name);
			} else {
				for(int i = 0 ; i < scope.length - 1 ; i++) {
					VarIdentifier varId_i = scope[i].getVarIdentifier();
					for(int j = i + 1 ; j < scope.length ; j++) {
						graph.addEdge(varId_i, scope[j].getVarIdentifier(), name);
					}
				}
			}
		}
		return graph;
	}

	/**
	 * @return
	 */
	public UndirectedGraph<VarAssignment> getAssignedConstraintGraph() {
		UndirectedGraph<VarAssignment> graph = new UndirectedGraph<VarAssignment>();
		for(PVariable variable : mapOfVariables.values()){
			int agentId = variable.getAgentId();
			int varIndex = variable.getId();
			VarIdentifier varId = new VarIdentifier(agentId, varIndex);
			//String varName = variable.getName();//.replace(".", "_");
			VarAssignment varAss = new VarAssignment(varId, Integer.MIN_VALUE);
			graph.addVertex(varAss);
		}
		for(PConstraint constraint : this.mapOfConstraints.values()){
			PVariable[] scope = constraint.getScope();
			String name = constraint.getName();
			if(constraint.getArity() == 1) { //FIXME
				//graph.addEdge(scope[0].getVarIdentifier(), scope[0].getVarIdentifier(), constraint.getName());
			} else if(constraint.getArity() == 2) {
				VarIdentifier varId1 = scope[0].getVarIdentifier();
				VarAssignment varAss1 = new VarAssignment(varId1, Integer.MIN_VALUE);
				VarIdentifier varId2 = scope[1].getVarIdentifier();
				VarAssignment varAss2 = new VarAssignment(varId2, Integer.MIN_VALUE);
				graph.addEdge(varAss1, varAss2, name);
			} else {
				for(int i = 0 ; i < scope.length - 1 ; i++) {
					VarIdentifier varId_i = scope[i].getVarIdentifier();
					VarAssignment varAssi = new VarAssignment(varId_i, Integer.MIN_VALUE);
					for(int j = i + 1 ; j < scope.length ; j++) {
						VarAssignment varAssj = new VarAssignment(scope[j].getVarIdentifier(), Integer.MIN_VALUE);
						graph.addEdge(varAssi, varAssj, name);
					}
				}
			}
		}
		return graph;
	}
	
	/**
	 * @return the nbSoftIntensionConstraints
	 */
	public int getNbSoftIntensionConstraints() {
		return nbSoftIntensionConstraints;
	}

	/**
	 * @param nbSoftIntensionConstraints the nbSoftIntensionConstraints to set
	 */
	public void setNbSoftIntensionConstraints(int nbSoftIntensionConstraints) {
		this.nbSoftIntensionConstraints = nbSoftIntensionConstraints;
	}

	/**
	 * @return the maximalCost
	 */
	public int getMaximalCost() {
		return maximalCost;
	}

	/**
	 * @param maximalCost the maximalCost to set
	 */
	public void setMaximalCost(Integer maximalCost) {
		this.maximalCost = maximalCost;
	}

	/**
	 * @return the initialCost
	 */
	public int getInitialCost() {
		return initialCost;
	}

	/**
	 * @param initialCost the initialCost to set
	 */
	public void setInitialCost(int initialCost) {
		this.initialCost = initialCost;
	}

	/**
	 * 
	 */
	public JPanel[] makeConstraintTables() {
		JPanel[] constraints = new JPanel[this.mapOfConstraints.size()];
		int i = 0;
		Iterator<PConstraint> it = this.mapOfConstraints.values().iterator();
		while(it.hasNext()) {
			PConstraint constraint = it.next();
			PRelation relation = ((PExtensionConstraint) constraint).getRelation();
			constraints[i++] = relation.makeConstraintTable(constraint.getScope(), constraint.getName());
		}
		return constraints;
	}

	/**
	 * 
	 */
	public Map<String, JTable> getConstraintTables() {
		Map<String, JTable> constraints = new HashMap<String, JTable>();
		Iterator<PConstraint> it = this.mapOfConstraints.values().iterator();
		while(it.hasNext()) {
			PConstraint constraint = it.next();
			PRelation relation = ((PExtensionConstraint) constraint).getRelation();
			constraints.put(constraint.getName(), relation.getConstraintTable(constraint.getScope(), constraint.getName()));
		}
		return constraints;
	}
	
	/**
	 * 
	 * @return
	 */
	public JInternalFrame[] makeConstraintJInternalFrame() {
		JInternalFrame[] constraints = new JInternalFrame[this.mapOfConstraints.size()];
		int i = 0;
		Iterator<PConstraint> it = this.mapOfConstraints.values().iterator();
		while(it.hasNext()) {
			PConstraint constraint = it.next();
			PRelation relation = ((PExtensionConstraint) constraint).getRelation();
			constraints[i] = relation.makeConstraintInternalFrame(constraint.getScope(), constraint.getName(), i);
			i++;
		}
		return constraints;
	}

	public static void main(String[] args) {
		// The GNU GPL copyright notice
		System.err.println(Commands.copyrightNotice);
		String problem = "./logs/MSP_1.xml";
		problem = "./samples/DiMES/M20_A9/3/MSP_1.xml";
		args = new String[]{problem}; 
		if (args.length != 1) {
			System.out.println("InstanceParser " + VERSION);
			System.out.println("Usage : java ... InstanceParser <instanceName>");
			System.exit(1);
		}
		InstanceParser parser = new InstanceParser(args[0]);
		parser.loadInstance(args[0]);
		parser.parse(false);
		UndirectedGraph<VarIdentifier> graph = parser.getConstraintGraph();
		String fileName = "./logs/graph.dot";
		UndirectedGraph.toDotFile(graph, fileName);//.createFile(fileName, graph.toDOTString());
		UndirectedGraph.fromDOTtoPDF(fileName);
		
		fileName = "./logs/mcstree.dot";
		AbstractTree<VarIdentifier> tree = graph.MCSDFS();
		AbstractTree.toDotFile(tree, fileName);
		UndirectedGraph.fromDOTtoPDF(fileName);
		System.out.println("mcstree depth is="+tree.getDepth());

		fileName = "./logs/maxdegtree.dot";
		tree = graph.MaxDegreeDFS();
		AbstractTree.toDotFile(tree, fileName);
		UndirectedGraph.fromDOTtoPDF(fileName);
		System.out.println("maxdegtree depth is="+tree.getDepth());

		fileName = "./logs/MLSPTree.dot";
		tree = graph.MLSPTree();
		AbstractTree.toDotFile(tree, fileName);
		UndirectedGraph.fromDOTtoPDF(fileName);
		System.out.println("MLSPTree depth is="+tree.getDepth());
	}

	/**
	 * This methods converts all constraints to the given category 
	 * @param category
	 */
	public void convertToCategory(String category) {
		if(category.equals(SATISFACTION)) {
			converToSatisfactionProblem();
		} else if (category.equals(OPTIMIZATION)) {
			converToOptimizationtionProblem();
		} else if (category.equals(HYBRID)) {
			throw new IllegalArgumentException("All problems are "+HYBRID);
		} else {
			throw new IllegalArgumentException(category+" is not in the known categories. That is, "+SATISFACTION+", "+OPTIMIZATION+", "+HYBRID);
		}
	}

	/**
	 * 
	 */
	private void converToOptimizationtionProblem() {
		//FIXME
//		for(Entry<String, PConstraint> entry : this.mapOfConstraints.entrySet()) {
//			String name = entry.getKey();
//			PConstraint constraint = entry.getValue();
//		}
	}

	/**
	 * 
	 */
	private void converToSatisfactionProblem() {
		//FIXME
//		for(Entry<String, PConstraint> entry : this.mapOfConstraints.entrySet()) {
//			String name = entry.getKey();
//			PConstraint constraint = entry.getValue();
//		}
	}
	
}
