package simulation.parser;

import java.util.Arrays;
import java.util.Map;
import java.util.TreeMap;
import kernel.AbstractAgent;

public class PSoftRelation extends PRelation {

	private Integer[] weights;

	/**
	 * defaultCost = Integer.MAX_VALUE if defaultCost is infinity
	 */
	private int defaultCost;

	/**
	 * The max of all weights values and defaultCost.
	 */
	private int maximalCost;

	public PSoftRelation(String name, int arity, int nbTuples, String semantics, int[][] tuples, Integer[] weights, int defaultCost) {
		super(name, arity, nbTuples, semantics);
		
		this.tuples = new int[tuples.length][];
		this.weights = new Integer[weights.length];

	    Map<int[], Integer> map = new TreeMap<int[], Integer>(Toolkit.lexicographicComparator);
	    for (int i = 0; i < weights.length; i++) {
	        map.put(tuples[i], weights[i]);
	    }
	    int i = 0;
	    for (int[] tuple: map.keySet()) {
	    	this.tuples[i++] = tuple;
	    }
		this.weights = map.values().toArray(new Integer[0]);
		
		this.defaultCost = defaultCost;
		this.maximalCost = defaultCost;
		for (int w : weights){
			if (w > maximalCost){
				maximalCost = w;
			}
		}
	}
	
	public Integer[] getWeights() {
		return weights;
	}

	public int getDefaultCost() {
		return defaultCost;
	}

	public int getMaximalCost() {
		return maximalCost;
	}
	
	public int computeCostOf(int[] tuple) {
		int position = Arrays.binarySearch(tuples, tuple, Toolkit.lexicographicComparator);
		return position >= 0? weights[position] : defaultCost;
	}

	public boolean isSimilarTo(int arity, int nbTuples, String semantics, int[][] tuples, int[] weights, int defaultCost) {
		if (!super.isSimilarTo(arity, nbTuples, semantics, tuples))
			return false;
		if (this.defaultCost != defaultCost)
			return false;
		for (int i = 0 ; i < weights.length ; i++)
			if (this.weights[i] != weights[i])
				return false;
		return true;
	}

	public String getTuplesToString() {
		StringBuffer sb = new StringBuffer();
		int currentWeigth = -1;
		for (int i = 0 ; i < nbTuples ; i++) {
			if (i != 0) sb.append(InstanceTokens.TUPLES_SEPARATOR);
			if (weights[i] != currentWeigth) {
				currentWeigth = weights[i];
				sb.append(currentWeigth + InstanceTokens.COST_SEPARATOR);
			}
			for (int j = 0 ; j < arity ; j++) {
				sb.append(tuples[i][j]);
				if (j < arity - 1) sb.append(' ');
			}
		}
		return sb.toString();
	}

	public String toString(){
		StringBuffer sb = new StringBuffer();
		sb.append("<").append(InstanceTokens.RELATION).append(" ");// "<relation "
		sb.append(InstanceTokens.NAME).append("=\"").append(this.name).append("\" "); // name=this.name
		sb.append(InstanceTokens.ARITY).append("=\"").append(this.arity).append("\" "); // arity=this.arity
		sb.append(InstanceTokens.SEMANTICS).append("=\"").append(this.semantics).append("\" "); // semantics=this.semantics
		sb.append(InstanceTokens.NB_TUPLES).append("=\"").append(this.nbTuples).append("\" "); // nbTuples=this.nbTuples
		sb.append(InstanceTokens.DEFAULT_COST).append("=\"").append(this.defaultCost).append("\""); // defaultCost=this.defaultCost
		sb.append(">");
		sb.append(getTuplesToString());
		sb.append("</").append(InstanceTokens.RELATION).append(">\n");// "</relation>"
		return sb.toString();
	}
	
	public String pretty() {
		int displayLimit = 5;
		String s =
				"  relation " + name + " with arity=" + arity + ", semantics=" + semantics + ", nbTuples=" + nbTuples + ", defaultCost="
						+ defaultCost + " : ";
		for (int i = 0 ; i < Math.min(nbTuples, displayLimit) ; i++) {
			s += "(";
			for (int j = 0 ; j < arity ; j++)
				s += (tuples[i][j] + (j < arity - 1? "," : ""));
			s += ") ";
			if (weights != null)
				s += " with cost=" + weights[i] + ", ";
		}
		return s + (nbTuples > displayLimit? "..." : "");
	}
	
	//FIXME
//	/**
//	 * @param owner
//	 * @param v0
//	 * @param v1
//	 * @param name
//	 */
//	@Override
//	protected void makeBinaryConstraint(String name, AbstractAgent owner, PVariable v0, PVariable v1, boolean transpose) {
//		Integer[] values0 = v0.getDomain().getValues();
//		Integer[] values1 = v1.getDomain().getValues();
//		Integer[] costs = new Integer[values0.length*values1.length];
//		for(int i=0 ; i<values0.length ; i++){
//			int val0 = values0[i];
//			for(int j=0 ; j<values1.length ; j++){
//				int val1 = values1[j];
//				costs[i * values0.length + j] = computeCostOf(new int[]{val0, val1});
//			}
//		}
//		owner.makeSoftBinRelation(name, v0, v1, costs, transpose);
//	}

	
	/**
	 * @param owner
	 * @param v0
	 * @param v1
	 * @param name
	 */
	@Override
	protected void makeBinaryConstraint(String name, AbstractAgent owner, PVariable v0, PVariable v1, boolean transpose) {
		owner.makeWeightedTuplesRelation(name, v0, v1, this.tuples, this.weights, this.defaultCost, transpose);
	}
	
	/**
	 * @param name
	 * @param owner
	 * @param variable
	 */
	@Override
	public void makeUnaryConstraint(String name, AbstractAgent owner, PVariable variable) {
		owner.makeSoftRelation(name, variable, variable.getDomain().getValues(), this.weights);
	}
}
