/*
DisChoco: Distributed Constraint Reasoning Solver 
Copyright (C) 2007-2013 Redouane Ezzahir & Mohamed Wahbi 

DisChoco is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DisChoco is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

To contact the authors: <http://dischoco.sourceforge.net/>
*/ 
package util;

import java.lang.reflect.Constructor;
import java.util.Collection;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.StringTokenizer;
import kernel.VarIdentifier;

/**
 * @version dischoco 2.5.3<br/>
 * util.HashList.java
 * @author Mohamed Wahbi
 */

@SuppressWarnings("serial")
public class HashList<E extends Comparable<? super E>> extends LinkedList<E> implements Comparator<E> {
	
	/**
	 * 
	 */
	public HashList() {
		super();
	}
	/**
	 * @param list
	 */
	public HashList(Collection<? extends E> list) {
		super(list);
	}
	
	public HashList(String string, Class<E> elementClass) throws Exception  {
		this();
		if (string.startsWith("[")) {
			string = string.substring(string.indexOf("[") + 1, string.indexOf("]")); // remove [ and ]
		}
		StringTokenizer tok = new StringTokenizer(string, " ");
		while (tok.hasMoreTokens()) {
			Object[] args = new Object[] {tok.nextToken()};
			Constructor<E> constructor = elementClass.getConstructor(new Class[] {String.class});
			this.add(constructor.newInstance(args));
		}
	}

	@Override
	public boolean add(E e) {
		if (this.contains(e)) {
			return false;
		} else {
			return super.add(e);
		}
	}

	@Override
	public boolean addAll(Collection<? extends E> collection) {
		Collection<E> copy = new LinkedList<E>(collection);
		copy.removeAll(this);
		return super.addAll(copy);
	}

	@Override
	public boolean addAll(int index, Collection<? extends E> collection) {
		Collection<E> copy = new LinkedList<E>(collection);
		copy.removeAll(this);
		return super.addAll(index, copy);
	}

	@Override
	public void add(int index, E element) {
		if (this.contains(element)) {
			return;
		} else {
			super.add(index, element);
		}
	}
	
	/**
	 * Returns a list that contains the union of two given lists.
	 */
	public static <E extends Comparable<? super E>> HashList<E> union(HashList<E> list1, HashList<E> list2) {
		HashList<E> union = new HashList<E>(list1);
		union.addAll(list2);
		return union;
	}
	
	/**
	 * Returns a list that contains the union of two given lists.
	 */
	public static <E extends Comparable<? super E>> HashList<E> intersection(HashList<E> list1, HashList<E> list2) {
		HashList<E> union = new HashList<E>(list1);
		union.retainAll(list2);
		return union;
	}
	
	/**
	 * Returns the list with the inverse order of this list. 
	 */
	public HashList<E> inverse() {
		HashList<E> inverve = new HashList<E>();
		for (int i = this.size()-1 ; i > -1 ; i--){
			inverve.add(this.get(i));
		}
		return inverve;
	}

	/**
	 * Checks whether all elements of the given list are in this list or not.
	 */
	public boolean contains(HashList<E> list) {
		return this.containsAll(list);
	}
	
	@Override
	public HashList<E> subList(int fromIndex, int toIndex) {
		return new HashList<E>(super.subList(fromIndex, toIndex));
	}
	
	public static <T extends Comparable<? super T>> HashList<T> fromString(String string, Class<T> elementClass) {
		HashList<T> list = new HashList<T>();
		if (string.startsWith("[")) {
			string = string.substring(string.indexOf("[") + 1, string.indexOf("]")); // remove [ and ]
		}
		StringTokenizer tok = new StringTokenizer(string, " ");
		while (tok.hasMoreTokens()) {
			Object[] args = new Object[] {tok.nextToken()};
			try {
				Constructor<T> constructor = elementClass.getConstructor(new Class[] {String.class});
				T e  = constructor.newInstance(args);
				list.add(e);
			} catch (Exception e1) {
				e1.printStackTrace();
			}
		}
		return list;
	}

	/**
	 * @param myId
	 * @return 
	 * @return
	 */
	public HashList<E> getElementsAfter(E e) {
		int rank = this.indexOf(e);
		int toIndex = this.size();
		return subList(rank + 1, toIndex);
	}

	/**
	 * @param myId
	 * @return 
	 * @return
	 */
	public HashList<E> getElementsBefor(E e) {
		int rank = this.indexOf(e);
		return subList(0, rank);
	}
	
	public int compare(HashList<E> list) {
		int cmp = this.size() - list.size();
		if(cmp != 0) {
			return cmp;
		} else {
			for(int i = 0 ; i < this.size() ; i++) {
				cmp = list.get(i).compareTo(get(i));
				if (cmp != 0) {
					return cmp;
				}
			}
			return 0;
		}
	}

	/* (non-Javadoc)
	 * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
	 */
	@Override
	public int compare(E o1, E o2) {
		int cmp = indexOf(o2) - indexOf(o1);
		if (cmp == 0) {
			return o2.compareTo(o1);
		}
		return cmp;
	}

    /**
     * Returns a shallow copy of this <tt>HashList</tt>. (The elements themselves are not cloned.)
     *
     * @return a shallow copy of this <tt>HashList</tt> instance
     */
	@SuppressWarnings("unchecked")
	@Override
	public HashList<E> clone() {
		return (HashList<E>) super.clone();
	}
	
	public static void main(String[] args) throws Exception {
		HashList<VarIdentifier> list = new HashList<VarIdentifier>();
		VarIdentifier v1 = new VarIdentifier(1);
		VarIdentifier v2 = new VarIdentifier(2);
		VarIdentifier v3 = new VarIdentifier(3);
		VarIdentifier v4 = new VarIdentifier(4);
		VarIdentifier v5 = new VarIdentifier(5);
		list.add(v3);
		list.add(v1);
		list.add(v2);
		list.add(v1);
		list.add(v4);
		list.add(v5);
		String string = list.toString();
		System.out.println(string);
		HashList<VarIdentifier> l = fromString(string, VarIdentifier.class);
		System.out.println(l);
		HashList<VarIdentifier> ll = new HashList<VarIdentifier>(string, VarIdentifier.class);
		System.out.println(ll);
		VarIdentifier x = v1;
		int rank = ll.indexOf(x);
		int toIndex = ll.size();
		System.out.println("lowerThan: "+x+" "+ll.subList(rank+1, toIndex));
		System.out.println("higherThan: "+x+" "+ll.subList(0, rank));

		VarIdentifier y = v5;
		int rank1 = ll.indexOf(x);
		int rank2 = ll.indexOf(y);
		int minRank = Math.min(rank1, rank2);
		int maxRank = Math.max(rank1, rank2);
		HashList<VarIdentifier> between = ll.subList(minRank+1, maxRank);
		System.out.println("between: "+x+" and "+y+" "+between);
		HashList<VarIdentifier> higher = ll.subList(0, ll.indexOf(v4));
		System.out.println(higher+" intersect "+between+" = "+HashList.intersection(higher, between));	
		
		HashList<VarIdentifier> hlist = HashList.fromString("[]", VarIdentifier.class);
		System.out.println(hlist);
	}
}
