/*
DisChoco: Distributed Constraint Reasoning Solver 
Copyright (C) 2007-2012 Redouane Ezzahir & Mohamed Wahbi 

DisChoco is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DisChoco is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

To contact the authors: <http://dischoco.sourceforge.net/>
 */
package util;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.Iterator;
import java.util.StringTokenizer;
import java.util.TreeSet;
import kernel.VarIdentifier;

/**
 * A class models the context or the agents assignment. It is structured to ensure that its iterator will traverse all assignments in descending
 * order, sorted according to <i>current agents ordering</i> or an order given by a comparator. In general the comparator is the <i>
 * {@link kernel.order.AgentOrdering}</i>
 * 
 * @authors 
 * <a href="mailto:ezzahir@ensa-agadir.ac.ma?subject=DisChoco &cc=wahbi@lirmm.fr &cc=bessiere@lirmm.fr">
 * Redoaune Ezzahir</a> and 
 * <a href="mailto:wahbi@lirmm.fr?subject=DisChoco &cc=ezzahir@ensa-agadir.ac.ma &cc=bessiere@lirmm.fr">
 * Mohamed Wahbi</a>
 *
 */
public class Instantiation implements Iterable<VarAssignment>, Assignment, Cloneable, Serializable {

	/**
	 * 
	 */
	private static final long serialVersionUID = 2447271062973464046L;
	/**
	 * A sorted set of variable assignment, sorted according to the <i>current agents ordering</i>
	 */
	private TreeSet<VarAssignment> instantiation = null;

	/**
	 * Constructs a new instantiation with the given comparator
	 */
	public Instantiation(Comparator<? super VarAssignment> comparator) {
		this.instantiation = new TreeSet<VarAssignment>(comparator);
	}

	/**
	 * Default constructor
	 */
	public Instantiation() {
		this.instantiation = new TreeSet<VarAssignment>(new DefaultComparator("[]"));
	}

	/**
	 * Constructs a new instantiation with the given comparator and instantations
	 */
	public Instantiation(Collection<VarAssignment> instantiation, Comparator<? super VarAssignment> comparator) {
		this(comparator);
		this.instantiation.addAll(instantiation);
	}

	/**
	 * Adds a variable assignment to this context. it makes a priority adding : if the given element exist the value of the bigger time counter is
	 * stored
	 * 
	 * @param varAssign
	 */
	public void add(VarAssignment varAssign) {
		if (varAssign == null)
			return;
		VarIdentifier varId = varAssign.getVarIdentifier();
		VarAssignment v = this.getVarAssignment(varId);
		if (v != null) {
			if (varAssign.compareTo(v) > 0) {
				v.setValue(varAssign.getValue());
				v.setTimeCounter(varAssign.getTimeCounter());
			}
		} else {
			if (this.instantiation.comparator() instanceof DefaultComparator)
				((DefaultComparator) this.instantiation.comparator()).update(varId);
			if (!this.instantiation.add(varAssign)) {
				throw new Error("\nfail to add " + varAssign + " to " + this.toString() + " " + this.instantiation.contains(varAssign));
			}
		}
	}

	/**
	 * Returns a view of the portion of this sorted set whose elements are less than or equal to <tt>fromElement</tt>. The returned sorted set is
	 * backed by this sorted set, so changes in the returned sorted set are reflected in this sorted set, and vice-versa. The returned sorted set
	 * supports all optional set operations.
	 * <p>
	 * 
	 * The sorted set returned by this method will throw an <tt>IllegalArgumentException</tt> if the user attempts to insert a element outside the
	 * specified range.
	 * <p>
	 * 
	 * Note: this method always returns a view that contains its (high) endpoint. If you need a view that does not contain this endpoint, and the
	 * element type allows for calculation of the successor a given value, merely request a tailSet bounded by <tt>successor(lowEndpoint)</tt>. For
	 * example, suppose that <tt>s</tt> is a sorted set of strings. The following idiom obtains a view containing all of the strings in <tt>s</tt>
	 * that are strictly less <tt>high</tt>:
	 * 
	 * <pre>
	 * SortedSet tail = s.tailSet(low + &quot;\0&quot;);
	 * </pre>
	 */
	public Instantiation tailSet(VarAssignment varAssign) {
		return new Instantiation(this.instantiation.tailSet(varAssign), this.instantiation.comparator());
	}

	/**
	 * Returns a view of the portion of this sorted set whose elements are strictly greater than <tt>toElement</tt>. The returned sorted set is backed
	 * by this sorted set, so changes in the returned sorted set are reflected in this sorted set, and vice-versa. The returned sorted set supports
	 * all optional set operations.
	 * <p>
	 * 
	 * The sorted set returned by this method will throw an <tt>IllegalArgumentException</tt> if the user attempts to insert a element outside the
	 * specified range.
	 * <p>
	 * 
	 * Note: this method always returns a view that does not contain its (high) endpoint. If you need a view that does contain this end point, and the
	 * element type allows for calculation of the successor a given value, merely request a headSet bounded by <tt>successor(highEndpoint)</tt>. For
	 * example, suppose that <tt>s</tt> is a sorted set of strings. The following idiom obtains a view containing all of the strings in <tt>s</tt>
	 * that are less than or equal to <tt>high</tt>:
	 * 
	 * <pre>
	 * SortedSet head = s.headSet(high + &quot;\0&quot;);
	 * </pre>
	 */
	public Instantiation headSet(VarAssignment varAssign) {
		Instantiation head = new Instantiation(this.instantiation.comparator());
		head.instantiation.addAll(this.instantiation.headSet(varAssign));
		return head.clone();
	}

	/**
	 * Returns the variable assignment of the variable with the given identifier
	 */
	public VarAssignment getVarAssignment(VarIdentifier varId) {
		Iterator<VarAssignment> it = this.iterator();
		while (it.hasNext()) {
			VarAssignment v = it.next();
			if (varId.equals(v.getVarIdentifier()))
				return v;
		}
		return null;
	}

	public boolean equals(Object o) {
		if (o == null)
			return false;
		Instantiation inst = (Instantiation) o;
		if (this.size() != inst.size()) {
			return false;
		}
		for(VarAssignment var : this) {
			VarAssignment ass = inst.getVarAssignment(var.getVarIdentifier());
			if (ass == null || !var.equals(ass)) {
				return false;
			}
		}
		return true;

	}

	/**
	 * Returns <tt>true</tt> if this Instantiation contains an element with the given varId.
	 * 
	 * @param varId
	 *            the id of variable to be checked for containment in this Instantiation.
	 * @return <tt>true</tt> if this Instantiation contains an element with the given varId.
	 */
	public boolean contains(VarIdentifier varId) {
		Iterator<VarAssignment> it = this.iterator();
		while (it.hasNext()) {
			VarAssignment v = it.next();
			if (varId.equals(v.getVarIdentifier()))
				return true;
		}
		return false;
	}

	/**
	 * Returns the IDs of all agents assigned in this context
	 * 
	 */
	public IntArray getAgentIds() {
		IntArray agIds = new IntArray();
		Iterator<VarAssignment> it = this.iterator();
		while (it.hasNext()) {
			agIds.add(it.next().getAgentId());
		}
		return agIds;
	}

	/**
	 * @return
	 */
	public HashList<VarIdentifier> getVarIds() {
		HashList<VarIdentifier> varIds = new HashList<VarIdentifier>();
		Iterator<VarAssignment> it = this.iterator();
		while (it.hasNext()) {
			varIds.add(it.next().getVarIdentifier());
		}
		return varIds;
	}

	public int hashCode() {
		return this.toString().hashCode();
	}

	/**
	 * Checks whether this context is compatible with the given context or not. Two context are compatible if only if they assign the same value for
	 * the same variables
	 */
	public boolean compatible(Instantiation inst) {
		Iterator<VarAssignment> it = this.iterator();
		while (it.hasNext()) {
			VarAssignment v1 = it.next();
			VarAssignment v2 = inst.getVarAssignment(v1.getVarIdentifier());
			if (v2 != null && v1.getValue() != v2.getValue()) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Checks whether this context is compatible with the given assignment or not. Two assignment are compatible if only if they assign the same value
	 * for the same variables
	 */
	public boolean compatible(VarAssignment varAssign) {
		VarAssignment var = this.getVarAssignment(varAssign.getVarIdentifier());
		return (var == null)? true : (var.getValue() == varAssign.getValue());
	}

	/**
	 * 
	 * @deprecated by {@link #priorityMerge(Instantiation)}
	 * @param inst
	 * 
	 */
	public void merge(Instantiation inst) {
		if (this.compatible(inst)) {
			Iterator<VarAssignment> it = inst.iterator();
			while (it.hasNext()) {
				VarAssignment v = it.next();
				this.add(v);
			}
		} else {
			throw new Error("union of two incompatible instantiations :\n" + "Intantiation_1: " + this + "\n" + "Intantiation_2: " + inst);
		}
	}

	/**
	 * Priority merge iterates over variable assignment and ad them by calling add(VarAssignment)
	 * 
	 * @param inst
	 * @see #add(VarAssignment)
	 */
	public void priorityMerge(Instantiation inst) {
		Iterator<VarAssignment> it = inst.iterator();
		while (it.hasNext()) {
			this.add(it.next());
		}
	}

	/**
	 * Returns a clone of this object
	 */

	public Instantiation cloneUseingSelfComparator() {
		Instantiation clone = new Instantiation(this.comparator());
		Iterator<VarAssignment> it = this.iterator();
		while (it.hasNext()) {
			VarAssignment v = it.next();
			clone.add(v.clone());
		}
		return clone;
	}

	/**
	 * Returns a clone of this object
	 */

	public Instantiation clone() {
		Instantiation clone = new Instantiation(new DefaultComparator(this.toString()));
		Iterator<VarAssignment> it = this.iterator();
		while (it.hasNext()) {
			VarAssignment v = it.next();
			clone.add(v.clone());
		}
		return clone;
	}

	/**
	 * Returns the comparator used to sort the variables assignments
	 */

	public Comparator<? super VarAssignment> comparator() {
		return this.instantiation.comparator();
	}

	/**
	 * Returns a string representation of this instantiation
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append("[");
		for(VarAssignment var : this) {
			sb.append(var.toString());
			sb.append(",");
		}
		int newLength = sb.lastIndexOf(",");
		if (newLength > 0) {
			sb.setLength(newLength);
		}
		sb.append("]");
		return sb.toString();
	}

	/**
	* Create an new Instantiation from the given string using comparator
	*/
	public static Instantiation fromString(String str, Comparator<? super VarAssignment> comparator) {
		Instantiation inst;
		if (comparator != null) {
			inst = new Instantiation(comparator);
		} else {
			inst = new Instantiation();
		}
		str = str.substring(1, str.indexOf("]"));
		StringTokenizer tok = new StringTokenizer(str, ",");
		while (tok.hasMoreTokens()) {
			inst.add(VarAssignment.fromString(tok.nextToken()));
		}
		return inst;
	}

	/**
	 * 
	 * @param str
	 * @return
	 */
	public static Instantiation fromString(String str) {
		return fromString(str, new DefaultComparator(str));
	}

	/**
	 * Returns an iterator over the elements in this set in descending order. from hi
	 */
	public synchronized Iterator<VarAssignment> iterator() {
		// this.refresh();
		return this.instantiation.iterator();
	}

	/**
	 * to refresh the order must be call when the agents order change
	 */
	public synchronized Instantiation refresh() {
		Instantiation inst = new Instantiation(this.instantiation.comparator());
		Iterator<VarAssignment> it = this.instantiation.iterator();
		while (it.hasNext()) {
			VarAssignment v = it.next();
			inst.add(v.clone());
		}
		this.instantiation.clear();
		this.instantiation.addAll(inst.instantiation);
		return this;
	}

	/**
	 * Tests is the instantiation is empty or not.
	 */
	public boolean isEmpty() {
		return this.instantiation.isEmpty();
	}

	/**
	 * Performs the Highest possible lowest variable heuristic
	 * 
	 * @param I2
	 * @return boolean return positive number if this Instantiation is better than I2 using HPLV heuristic. negative otherwise
	 */
	public int compareHPLV(Instantiation I2) {
		if (I2 == null || this.isEmpty())
			return 1;
		if (I2.isEmpty())
			return -1;
		VarAssignment last1 = this.last();
		VarAssignment last2 = I2.last();
		// return rank(last2) - rank(last1)
		return this.instantiation.comparator().compare(last2, last1);
		// TODO you can improve this function if last1==last2
		// compare the size or the order of other agents.
	}

	/**
	 * Removes the given variable assignment from the instantiation
	 * 
	 * @param varId
	 *            the identifier of the variable to remove its assignment from this context
	 * @return VarAssignment the removed element
	 */

	public VarAssignment remove(VarAssignment var) {
		VarAssignment v = this.getVarAssignment(var.getVarIdentifier());
		if (v != null) {
			this.instantiation.remove(v);
		}
		return v;
	}

	public VarAssignment remove(VarIdentifier varId) {
		VarAssignment var = this.getVarAssignment(varId);
		assert (var != null) : "The assignment of the variable with the identifier " + varId + " does not exist in:\n" + this;
		return this.remove(var);
	}

	/**
	 * Return the first (according to the order given by the comparator) variable assignment in the instantiation
	 */
	public VarAssignment first() {
		// this.refresh();
		return this.instantiation.first();
	}

	/**
	 * Return the last (according to the order given by the comparator) variable assignment in the instantiation
	 */
	public VarAssignment last() {
		// this.refresh();
		return this.instantiation.last();

	}

	/**
	 * Retuns the number of element in the instantiation
	 * 
	 * @return
	 */

	public int size() {
		return this.instantiation.size();
	}

	/*
	 * (non-Javadoc)
	 * @see util.Assignment#compatible(util.Assignment)
	 */
	public boolean compatible(Assignment o) {
		if (o instanceof VarAssignment) {
			return this.compatible((VarAssignment) o);
		} else if (o instanceof Instantiation) {
			return this.compatible((Instantiation) o);
		}
		throw new Error("Compatibilty check error, Instantiation::compatible(" + o + ")");
	}

	/**
	 * Returns a whether this Instantiation is a subset of the given view
	 */

	public boolean isSubSetOf(Instantiation view) {
		return view.contains(this);
	}

	/**
	 * Returns an array containing all of the elements in this instantiation.
	 */
	public VarAssignment[] toArray() {
		VarAssignment[] vars = new VarAssignment[this.size()];
		return this.instantiation.toArray(vars);
	}

	/**
	 * Returns an array containing all time counter of the elements in this instantiation. 
	 * This array is used as time stamp because the elements are sorted. 
	 */
	public int[] getTimeStamp() {
		int[] ts = new int[this.size()];
		int i = 0;
		for(VarAssignment v : this.instantiation) {
			ts[i++] = v.getTimeCounter();
		}
		return ts;
	}

	/**
	 * Returns the element at the specified position in this Instantiation.
	 * 
	 * @param index
	 *            index of the element to return
	 * @return VarAssignment with the specified index
	 * @throws IndexOutOfBoundsException
	 */
	public VarAssignment get(int index) {
		if (index < this.size() && index > -1)
			return (VarAssignment) this.toArray()[index];
		return null;
	}

	/**
	 * Removes from this Instantiation all of its elements that are contained in the specified instantiation.
	 * 
	 * @param inst
	 *            An instantiation that contains variables assignment to remove
	 * @return <tt>true</tt> if this set changed as a result of the call
	 */
	public boolean removeAll(Instantiation inst) {
		return this.instantiation.removeAll(inst.instantiation);
	}

	/**
	 * @return 1 if the time counter of the given variable is bigger than its stored image in the instantiation 0 if equality -1 otherwise;
	 */
	public int checkTimeCounter(VarAssignment varAss) {
		VarAssignment v = this.getVarAssignment(varAss.getVarIdentifier());
		if (v == null)
			return 1;
		else if (varAss.getValue() != Integer.MIN_VALUE) {
			return varAss.getTimeCounter() - v.getTimeCounter();
		}
		return -1;
	}

	/**
	 * Creates a clone of this object with the given comparator. It copies all elements and sort them using comparator
	 */

	public Instantiation clone(Comparator<? super VarAssignment> comparator) {
		if (comparator == null)
			return this.clone();
		Instantiation clone = new Instantiation(comparator);
		Iterator<VarAssignment> it = this.iterator();
		while (it.hasNext()) {
			VarAssignment v = it.next();
			clone.add(v.clone());
		}
		return clone;
	}

	/**
	 * Checks whether the instantiation contains all elements of the given one or not.
	 * 
	 * @param view
	 *            An Instantiation to checks is included in this instantiation
	 */
	public boolean contains(Instantiation view) {
		Iterator<VarAssignment> it = view.iterator();
		while (it.hasNext()) {
			VarAssignment vAss = it.next();
			VarAssignment v = this.getVarAssignment(vAss.getVarIdentifier());
			if (v == null)
				return false;
			else if (v.getValue() != vAss.getValue())
				return false;
		}
		return true;
	}

	// /**
	// *@param view
	// * A context to compare its counters with their corresponding in
	// * this instantiation
	// *@param upto
	// * Index of last counter to compare with its corresponding in the
	// * instantiation
	// *@return (j+1) : where j is the index of the first newest variable
	// * assignment (i.e first counter in <code>view</code> that is bigger
	// * than its corresponding counter in the instantiation -(j+1) where
	// * j is the index of the first oldest variable assignment (i.e first
	// * counter in <code>view</code> that is smaller than its
	// * corresponding counter in the instantiation; 0 when all counter in
	// * <code>view</code> are equals to its corresponding counters in the
	// * instantiation ;
	// */
	// public int compareTimeStamp(Instantiation view, int upto) {
	// int[] myTimeStamp = this.getTimeStamp();
	// int[] viewTimeStamp = view.getTimeStamp();
	// int j = 0;
	// for (; j < myTimeStamp.length && j < viewTimeStamp.length && j < upto; j++) {
	// if (myTimeStamp[j] > viewTimeStamp[j]) {
	// return -(j + 1);
	// }
	// if (myTimeStamp[j] < viewTimeStamp[j]) {
	// return j + 1;
	// }
	// }
	// if (j != 0 && myTimeStamp.length < viewTimeStamp.length)
	// return myTimeStamp.length;
	// return 0;
	// }
	/**
	 * Compare the time counters of this instantiation with those of the given instantiation. It iterates over variables assignment in view and for
	 * each variable that belongs to both instantiation compares the counters in both instantiation.
	 * 
	 * @return 
	 * <code><b>1</b></code> if the first biggest counter is that of a VarAssignment belonging to this instantiation <br/>
	 * <code><b>-1</b></code> if the first biggest counter is that of a  VarAssignment belonging to the instantiation given in the argument  <br/>
	 * <code><b>0</b></code> if all counter of both instantiation are equal. <br/>
	 */
	public int compareTimeCounters(Instantiation view) {
		if (view == null || view.isEmpty())
			return 1;
		Iterator<VarAssignment> it = view.iterator();
		while (it.hasNext()) {
			VarAssignment vView = it.next();
			VarAssignment v = this.getVarAssignment(vView.getVarIdentifier());
			if (v != null) {
				int comp = v.compareTimeCounter(vView);
				if (comp > 0)
					return 1;
				if (comp < 0)
					return -1;
			}
		}
		return 0;
	}

	// public static void main(String args[]) {
	// final IntArray order = new IntArray(Utility.generatePermutation(
	// new java.util.Random(), 10));
	// @SuppressWarnings("unused")
	// Comparator<VarAssignment> comp = new Comparator<VarAssignment>() {
	// public int compare(VarAssignment v1, VarAssignment v2) {
	// return order.indexOf(v1.getAgentId())
	// - order.indexOf(v2.getAgentId());
	// }
	// };
	// Instantiation inst = new Instantiation();
	// for (int i = 0; i < 10; i++) {
	// VarAssignment v = new VarAssignment(i, 3, i);
	// inst.add(v);
	// }
	// System.out.println(inst.toString() + " \n"
	// + Instantiation.fromString(inst.toString(), null));
	// final IntArray order2 = new IntArray(Utility.generatePermutation(
	// new java.util.Random(), 10));
	// System.out.println("new order " + order2);
	// Instantiation cloneWithNewOrder = inst.clone(new Comparator<VarAssignment>() {
	// public int compare(VarAssignment v1, VarAssignment v2) {
	// return order2.indexOf(v1.getAgentId())
	// - order2.indexOf(v2.getAgentId());
	// }
	// });
	//
	//
	// Instantiation ist = Instantiation.fromString(cloneWithNewOrder.toString());
	// System.out.println(cloneWithNewOrder.toString() + " \n" + ist + " "+ ist.getAgentIds());
	// }

	public static Instantiation union(ArrayList<Instantiation> instantiations) {
		if (instantiations == null) {
			return new Instantiation();
		}
		Instantiation ans = new Instantiation();
		Iterator<Instantiation> it = instantiations.iterator();
		while (it.hasNext()) {
			ans.priorityMerge(it.next());
		}
		return ans;
	}

	public static Instantiation union(ArrayList<Instantiation> instantiations, Comparator<? super VarAssignment> comparator) {
		if (instantiations == null) {
			return new Instantiation(comparator);
		}
		Instantiation ans = new Instantiation(comparator);
		Iterator<Instantiation> it = instantiations.iterator();
		while (it.hasNext()) {
			ans.priorityMerge(it.next());
		}
		return ans;
	}

	public int indexOf(VarIdentifier v) {
		Iterator<VarAssignment> it = this.iterator();
		int i = 0;
		while (it.hasNext()) {
			if (v.equals(it.next().getVarIdentifier()))
				return i;
			i++;
		}
		return -1;
	}
	
	/**
	 * @return the instantiation
	 */
	public TreeSet<VarAssignment> getInstantiation() {
		return this.instantiation;
	}

	// public static void main(String args[]) {
	// Instantiation inst = new Instantiation();
	// for(int i = 0 ; i < 10 ; i++) {
	// VarAssignment v = new VarAssignment(i+1, 1, 2*i, 0);
	// inst.add(v);
	// }
	// System.out.println(inst);
	// VarIdentifier varId = new VarIdentifier(3, 1);
	// VarAssignment ass = new VarAssignment(varId, 2, 0);
	// System.out.println(ass);
	// System.out.println(inst.contains(varId)+" "+varId);
	// System.out.println(inst.compatible(ass));
	// Instantiation sub = inst.headSet(ass);
	// System.out.println(sub);
	// }

}
