/*
DisChoco: Distributed Constraint Reasoning Solver 
Copyright (C) 2007-2013 Redouane Ezzahir & Mohamed Wahbi 

DisChoco is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DisChoco is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

To contact the authors: <http://dischoco.sourceforge.net/>
*/ 
package util;

import java.io.Serializable;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Iterator;
import java.util.StringTokenizer;

/**
 * A class to model Integer array with different values. It is used to store agents IDs, type of events, or index
 */

public class IntArray implements Cloneable, Serializable, Iterable<Integer> { 

	private static final long serialVersionUID = -8695769236841701170L;
	/**
	 * The stored array of different elements
	 */
	protected Integer[] values;
	/**
	 * the number of elements in the stored array
	 */
	protected int size = 0;
	/**
	 * The amount by which the capacity of the array is automatically incremented when its size becomes greater than its capacity. If the capacity
	 * increment is less than or equal to zero, the capacity of the array multiply 2/3 each time it needs to grow.
	 */
	static int capacity = 5;

	/**
	 * Constructs a new integer array
	 */

	public IntArray() {
		this(capacity);
	}

	/**
	 * Constructs a new integer array with the given values
	 */
	public IntArray(Integer[] integers) {
		this(integers.length);
		for (int i = 0 ; i < integers.length ; i++) {
			add(integers[i]);
		}
	}

	/**
	 * Constructs a new integer array with the given values
	 */
	public IntArray(int[] integers) {
		this(integers.length);
		for (int i = 0 ; i < integers.length ; i++) {
			add(integers[i]);
		}
	}

	public IntArray(int initialCapacity) {
		this.values = new Integer[initialCapacity];
	}

	/**
	 * adds a value to the array and increments size if not already exist.
	 * 
	 * @param val
	 *            the value to add.
	 */

	public void add(int val) {
		if (indexOf(val) == -1) {
			if (size == values.length)
				ensureCapacity(size + 1);
			values[size] = val;
			size++;
		}
	}

	/**
	 * Adds value <i>val</i> to the array in the specified position. It replace the old value with the given value
	 * 
	 * @param val
	 * @param index
	 */
	public void add(int val, int index) {
		if (index < 0)
			throw new ArrayIndexOutOfBoundsException(index + " < " + 0);
		else if (index >= size) {
			add(val);
		} else {
			values[index] = val;
		}
	}

	/**
	 * Deletes the component at the specified index. Each component in this array with an index greater or equal to the specified {@code index} is
	 * shifted downward to have an index one smaller than the value it had previously. The size of this array is decreased by {@code 1}.
	 * 
	 * <p>
	 * The index must be a value greater than or equal to {@code 0} and less than the current size of the vector.
	 */
	public int removeAt(int idx) {
		if (idx != -1) {
			int element = values[idx];
			for (int i = idx ; i < size - 1 ; i++){
				values[i] = values[i + 1];
			}
			size--;
			return element;
		} else
			throw new ArrayIndexOutOfBoundsException(idx);
	}

	/**
	 * Removes the given value from this array
	 * 
	 * @param value
	 */
	public void remove(int value) {
		int idx = indexOf(value);
		this.removeAt(idx);
	}

	/**
	 * Checks whether the given integer is in this array or not.
	 */

	public boolean contains(int intobj) {
		return indexOf(intobj) != -1;
	}

	/**
	 * Returns the index of the specified element in this array; or returns -1 if the element is not found.
	 */
	public int indexOf(int intobj) {
		for (int i = 0 ; i < this.size ; i++)
			if (intobj == this.values[i])
				return i;
		return -1;
	}

	/**
	 * Returns the element at the specified position in this array.
	 * 
	 * @param index
	 *            index of the element to return
	 * @return object at the specified index
	 * @throws ArrayIndexOutOfBoundsException
	 *             if the index is out of range ({@code index < 0 || index >= size()})
	 */
	public int get(int index) {
		if (index < 0 || index >= size())
			throw new ArrayIndexOutOfBoundsException("element with index: " + index + " not found " + this.toString());
		return values[index];
	}

	/**
	 * This implements the unsynchronized semantics of ensureCapacity. Synchronized methods in this class can internally call this method for ensuring
	 * capacity without incurring the cost of an extra synchronization.
	 * 
	 * @param minCapacity
	 *            the new of the capacity
	 */

	private void ensureCapacity(int minCapacity) {
		int newSize = values.length + 1;
		while (minCapacity >= newSize) {
			newSize = (int) (3 * newSize) / 2 + 1;
		}
		Integer[] newvalues = new Integer[newSize];
		System.arraycopy(values, 0, newvalues, 0, values.length);
		this.values = newvalues;
		capacity = newSize;
	}

	/**
	 * Returns a string representation of this object
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append("[ ");
		for (int i = 0 ; i < size ; i++) {
			sb.append(values[i] + " ");
		}
		sb.append("]");
		return sb.toString();
	}

	/**
	 * Returns the number of elements in this array
	 * 
	 * @return
	 */

	public int size() {
		return size;
	}

	/**
	 * Creates and returns a copy of this array.
	 */
	public IntArray clone() {
		IntArray clone = new IntArray();
		clone.size = this.size;
		Integer[] newValues = new Integer[values.length];
		System.arraycopy(this.values, 0, newValues, 0, this.values.length);
		clone.values = newValues;
		return clone;
	}

	/**
	 * Returns the stored array
	 * 
	 * @return
	 */
	public int[] getArray() {
		int[] arrayInt = new int[size];
		for (int i = 0 ; i < arrayInt.length ; i++)
			arrayInt[i] = values[i];
		return arrayInt;
	}

	/**
	 * Indicates whether some other object is "equal to" this one. Returns <cod>False</code> when they are different size or when they differ at an
	 * element (two array with the same elements but have different order are not equals : [1 0 2]!=[2 0 1] ) Otherwise <code>True</code>
	 * 
	 */
	public boolean equals(Object o) {
		IntArray arr = (IntArray) o;
		if (arr.size != this.size)
			return false;
		int i = 0;
		for (int e : arr) {
			if (this.get(i++) != e)
				return false;
		}
		return true;
	}

	/**
	 * Returns a array that contains the union of this array with the given one.
	 */
	public static IntArray union(IntArray arrayInt1, IntArray arrayInt2) {
		if (arrayInt1 == null && arrayInt2 == null)
			return null;
		if (arrayInt1 == null)
			return arrayInt2.clone();
		if (arrayInt2 == null)
			return arrayInt1.clone();
		IntArray res = arrayInt1.clone();
		if (arrayInt2 != null)
			for (int i = 0 ; i < arrayInt2.size ; i++)
				res.add(arrayInt2.get(i));
		return res;

	}

	/**
	 * inverse the order of the array
	 */
	public IntArray inverse() {
		IntArray inv = new IntArray();
		for (int i = this.size - 1 ; i > -1 ; i--)
			inv.add(this.get(i));
		return inv;
	}

	/**
	 * Checks whether all element of given array is in this array or not.
	 */
	public boolean contains(IntArray array) {
		for (int i = 0 ; i < array.size() ; i++)
			if (!this.contains(array.get(i)))
				return false;
		return true;
	}

	/**
	 * Returns a new IntArray that is a subArray of this array. The subArray begins with the element at the specified index and extends to the end of
	 * this array.
	 * <p>
	 * Examples: <blockquote>
	 * 
	 * <pre>
	 * "1 2 4 5 6". subArray(2) returns "4 5 6"
	 * </pre>
	 * 
	 * </blockquote>
	 * 
	 * @param fromIndex
	 *            the beginning index, inclusive.
	 * @return the specified subArray.
	 */
	public IntArray subArray(int fromIndex) {
		IntArray array = new IntArray();
		for (int i = fromIndex ; i < this.size ; i++)
			array.add(this.get(i));
		return array;
	}

	/**
	 * Returns a new IntArray that is a subArray of this array. The subArray begins at the specified <code>beginIndex</code> and extends to the
	 * element at index <code>endIndex</code>. Thus the length of the subArray is <code>endIndex-beginIndex + 1</code>.
	 * <p>
	 * Examples: <blockquote>
	 * 
	 * <pre>
	 * "1 2 3 4 5 6 7".substring(2, 4) returns "3 4 5"
	 * </pre>
	 * 
	 * </blockquote>
	 * 
	 * @param beginIndex
	 *            the beginning index, inclusive.
	 * @param endIndex
	 *            the ending index, inclusive.
	 * @return the specified subIntArray.
	 * @return
	 */
	public IntArray subIntArray(int beginIndex, int endIndex) {
		IntArray subArray = new IntArray();
		for (int i = beginIndex ; ((i <= endIndex) && (i < this.size)) ; i++) {
			subArray.add(this.get(i));
		}
		return subArray;
	}

	/**
	 * Returns new array that contains this unless the elements of given array
	 * 
	 * @param array
	 *            An array to hide these elements
	 */

	public IntArray minus(IntArray array) {// TODO static ... minus(arr1, arr2)
		IntArray minus = new IntArray();
		for (int i : this) {
			if (!array.contains(i)) {
				minus.add(i);
			}
		}
		return minus;
	}

	/**
	 * Creates a new IntArray from its representing string
	 */

	public static IntArray fromString(String string) {
		IntArray arrayInt = new IntArray();
		if (string.startsWith("[")) {
			string = string.substring(string.indexOf("[") + 1, string.indexOf("]")); // remove [ and ]
		}
		StringTokenizer tok = new StringTokenizer(string);
		while (tok.hasMoreTokens()) {
			arrayInt.add(Integer.parseInt(tok.nextToken()));
		}
		return arrayInt;
	}

	/**
	 * Splits this array into the given number of sub-array.
	 * 
	 * <p>
	 * The array returned by this method contains <code>nbSubArray</code> subArray of this array that with the same size.
	 * <p>
	 * For example if we split an array of 9 elements {1..9} in 3 nbSubArray IntArray.split(array, 3) returns {{1 2 3}, {4, 5, 6}, {7, 8, 9}}
	 * 
	 * @param array
	 *            An array to split
	 * @param nbSubArray
	 *            the number of sub-array to return.
	 */

	public static IntArray[] split(IntArray array, int nbSubArray) {
		if (array.size() < nbSubArray)
			nbSubArray = array.size();

		IntArray[] res = new IntArray[nbSubArray];

		int resSize = (int) array.size() / nbSubArray;
		for (int j = 0 ; j < res.length ; j++) {
			res[j] = new IntArray();
			for (int i = 0 ; i < resSize ; i++) {
				if (((j * resSize) + i) < array.size) {
					res[j].add(array.get((j * resSize) + i));
				}
			}
		}
		int rest = array.size() % nbSubArray;
		if (rest != 0)
			for (int i = array.size - rest ; i < array.size ; i++) {
				res[res.length - 1].add(array.get(i));
			}

		return res;
	}

	/**
	 * Returns an array that contains the intersection between two arrays.
	 */

	public static IntArray intersection(IntArray a1, IntArray a2) {
		IntArray array = new IntArray();
		for (int i = 0 ; i < a1.size() ; i++) {
			if (a2.contains(a1.get(i)))
				array.add(a1.get(i));
		}
		return array;
	}

	/**
	 * Returns an iterator over elements of this array
	 */
	public Iterator<Integer> iterator() {
		return new Iter(this);
	}

	private class Iter implements Iterator<Integer> {
		IntArray arr;
		int cur = 0;
		int lastRet = -1;

		Iter(IntArray array) {
			this.arr = array;
		}

		public boolean hasNext() {
			return (cur != arr.size);
		}

		public Integer next() {
			int res = arr.get(cur);
			lastRet = cur;
			cur++;
			return res;
		}

		public void remove() {
			if (lastRet == -1)
				throw new IllegalStateException();
			arr.removeAt(lastRet);
			cur--;
			lastRet = -1;
		}
	}

	public void addAll(IntArray array) {
		for (int x : array) {
			this.add(x);
		}

	}

	public boolean isEmpty() {
		return this.size == 0;
	}

	public IntArray getElementsAfter(int element) {
		return this.subIntArray(this.indexOf(element) + 1, this.size() - 1);
	}

	public IntArray getElementsBefor(int element) {
		return this.subIntArray(0, this.indexOf(element) - 1);
	}

	/**
	 * 
	 * @param array
	 * @return <b> 1</b> if array is > than this <br/>
	 *         <b> 0</b> if array is equal to this <br/>
	 *         <b>-1</b> if array is < than this <br/>
	 */
	public int compare(IntArray array) {
		if (this.size != array.size) { return (this.size - array.size); }
		for (int element : array) {
			int rank = array.indexOf(element);
			if (element > this.get(rank)) { return 1; }
			if (element < this.get(rank)) { return -1; }
		}
		return 0;
	}

	/**
	 * 
	 * @param args
	 */
	public static void main(String[] args) {
		// int[] x = { 1 };
		// IntArray a = new IntArray(x);
		// Iterator<Integer> it = a.iterator();
		// while (it.hasNext()) {
		// if (it.next() % 2 != 0)
		// it.remove();
		// }
		// System.out.println(a);
		String str = "[ 1 3 2 4 5 6 7 8 9 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 ]";
		IntArray array = IntArray.fromString(str);
		System.err.println(array);
		System.err.println(array.getElementsBefor(4));
		System.err.println(array.getElementsAfter(4));
	}

	/**
	 * @param comparator
	 */
	public void sort(Comparator<Integer> comparator) {
		Integer[] array = toArray();
		Arrays.sort(array, comparator);
		this.values = array;
	}

	public static IntArray sort(IntArray array, Comparator<Integer> comparator) {
		Integer[] tmp = array.toArray();
		Arrays.sort(tmp, comparator);
		IntArray sorted = new IntArray(array.size);
		for(int i=0 ; i<array.size ; i++){
			sorted.add(tmp[i]);
		}
		return sorted;
	}
	
	public Integer[] toArray() {
		Integer[] arrayInt = new Integer[size];
		for (int i = 0 ; i < arrayInt.length ; i++)
			arrayInt[i] = values[i];
		return arrayInt;
	}

	public void swapByIndexes(int i, int j) {
		if (i < 0 || i>=this.size){
			throw new ArrayIndexOutOfBoundsException("i="+i + " should be betwenn 0 and " +(size-1));
		}
		if (j < 0 || j>=this.size){
			throw new ArrayIndexOutOfBoundsException("j="+j + " should be betwenn 0 and " +(size-1));
		}
		int value_i = values[i];
		values[i] = values[j];
		values[j] = value_i;
	}
}
