/*
DisChoco: Distributed Constraint Reasoning Solver 
Copyright (C) 2007-2013 Redouane Ezzahir & Mohamed Wahbi 

DisChoco is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DisChoco is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

To contact the authors: <http://dischoco.sourceforge.net/>
*/ 
package util;

import java.io.Serializable;
import java.util.StringTokenizer;

import kernel.VarIdentifier;

/**
 * A class models the variable assignment. A variable assignment is a data structure that assigns a value to the variable. It contains also the time
 * counter of the assignment
 * 
 * @version dischoco 2.5.3<br/>
 * util.VarAssignment.java
 * @author Mohamed Wahbi
 */
public class VarAssignment implements Assignment, IVarIdentifier, Serializable {
	
	private static final long serialVersionUID = -7206853296364453798L;
	private static final String UNASSIGNED = "?";
	private static final String EQUALS = "=";
	private static final String COUNTER_SEPARATOR = "/";
	
	/**
	 * The assigned variable identifier
	 */
	protected VarIdentifier varId;
	/**
	 * The assigned value for the variable
	 */
	private int value = Integer.MIN_VALUE;
	/**
	 *A time counter that store the last time (Integer) of the assignment
	 */
	private int timeCounter = 0;

	/**
	 * Constructs a variable assignment without time counter
	 * 
	 * @param varId
	 *            the variable identifier
	 * @param value
	 *            the assigned value
	 */
	public VarAssignment(VarIdentifier varId, int value) {
		this.varId = varId;
		this.value = value;
	}

	/**
	 * Constructs a variable assignment with a time counter
	 * 
	 * @param varId the variable identifier
	 * @param value the assigned value
	 * @param timeCounter the time counter of the assignment
	 */
	public VarAssignment(VarIdentifier varId, int value, int timeCounter) {
		this.varId = varId;
		this.value = value;
		this.timeCounter = timeCounter;
		this.hashCode();
	}

	/**
	 * Constructs a variable assignment for simple mode (Single variable per agent, thus index of variable is 0)
	 * 
	 * @param agId
	 *            the ID of the simple agent owning the variable
	 * @param value
	 *            the assigned value
	 * @param timeCounter
	 *            the time counter of the assignment
	 */
	public VarAssignment(int agId, int value, int timeCounter) {
		this(new VarIdentifier(agId, 1), value, timeCounter);
	}

	/**
	 * Constructs a variable assignment
	 * 
	 * @param agId the ID of the simple agent owning the variable
	 * @param varIndex the index of the variable inside the agent
	 * @param value the assigned value
	 * @param timeCounter the time counter of the assignment
	 */
	public VarAssignment(int agId, int varIndex, int value, int timeCounter) {
		this(new VarIdentifier(agId, varIndex), value);
		this.value = value;
		this.timeCounter = timeCounter;
	}

	/**
	 * Set the assigned value to the given one
	 * 
	 * @param value the value assigned to the variable
	 */
	public void setValue(int value) {
		this.value = value;
	}

	/**
	 * Returns the value assigned to the variable
	 */
	public int getValue() {
		return this.value;
	}

	/**
	 * Return the value of the time counter
	 */
	public int getTimeCounter() {
		return this.timeCounter;
	}

	/**
	 * Set the value of the time counter to the given value
	 */
	public void setTimeCounter(int counter) {
		this.timeCounter = Math.max(this.timeCounter, counter);
	}

	/**
	 * Gets the ID of the agent owning (that had assign) the variable
	 */
	public int getAgentId() {
		return this.varId.getAgentId();
	}

	/**
	 * Returns the index of the variable
	 */
	public int getVarIndex() {
		return this.varId.getVarIndex();
	}

	/**
	 * Returns the identifier of the variable
	 */

	public VarIdentifier getVarIdentifier() {
		return varId;
	}

	/**
	 * Checks whether this variable assignment is equals to the given object or not.
	 */
	public boolean equals(Object o) {
		VarAssignment v = (VarAssignment) o;
		return this.varId.equals(v.varId) && this.value == v.value && this.timeCounter == v.timeCounter;
	}

	public int hashCode() {
		return this.varId.hashCode() + this.value + this.timeCounter;
	}

	/**
	 * Returns a clone of this object.
	 */
	public VarAssignment clone() {
		return new VarAssignment(this.varId.clone(), this.value, this.timeCounter);
	}

	public void setVarIndex(int index) {
		this.varId = new VarIdentifier(this.varId.getAgentId(), index);
	}

	/**
	 * 
	 * {@inheritDoc}
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append("(");
		sb.append(this.varId.toString());
		sb.append(EQUALS);
		sb.append((this.value==Integer.MIN_VALUE)?UNASSIGNED:this.value);
		sb.append(COUNTER_SEPARATOR);
		sb.append(this.timeCounter);
		sb.append(")");
		return sb.toString();
	}

	/**
	 * Create a variable assignment from the given string
	 * 
	 * @param str a string that represents the variable assignment.
	 */
	public static VarAssignment fromString(String string) {
		if (string.startsWith("(")) {
			string = string.substring(string.indexOf("(") + 1, string.indexOf(")")); // remove ( and )
		}
		StringTokenizer st = new StringTokenizer(string, EQUALS);
		VarIdentifier varId = VarIdentifier.fromString(st.nextToken());
		st = new StringTokenizer(st.nextToken(), COUNTER_SEPARATOR);
		String str = st.nextToken();
		int value = str.equals(UNASSIGNED)?Integer.MIN_VALUE:Integer.parseInt(str);
		int counter = Integer.parseInt(st.nextToken());
		return new VarAssignment(varId, value, counter);
	}

	/**
	 * Compare the time counter of this variable assignment with these of the given one. Returns this.timeCounter - v.timeCounter
	 */
	public int compareTimeCounter(VarAssignment v) {
		if (v == null)
			return Integer.MAX_VALUE;
		if (this.varId.equals(v.varId)) { return this.timeCounter - v.timeCounter; }
		throw new Error("Try to compare time counter of two different variables : " + this + " and " + v);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see util.Assignment#compatible(util.Assignment)
	 */
	public boolean compatible(Assignment o) {
		if (o instanceof VarAssignment){
			return this.value==((VarAssignment) o).getValue();
		} else if (o instanceof Instantiation) {
			return ((Instantiation) o).compatible(this);
		}
		throw new IllegalArgumentException("Compatibilty check error VarAssignment::compatible(" + o + ")");
	}

	/**
	 * Unassigns variable. It removes the value assigned to this variable.
	 */
	public void clear() {
		this.value = Integer.MIN_VALUE;
		this.timeCounter = 0;
	}

	/**
	 * Set this variable assignment to be private. It simply assigns the value of the timeCounter to the value this.value = this.timeCounter;
	 */
	public void setPrivate() {
		this.value = this.timeCounter;
	}

	/**
	 * Checks whether a value is assigned or not
	 * 
	 * @return
	 */
	public boolean isInstantiated() {
		return this.value != Integer.MIN_VALUE;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
	public int compareTo(IVarIdentifier v) {
		return this.varId.compareTo(v);
	}
	
}
