/*
 * DisChoco: Distributed Constraint Reasoning Solver
 * Copyright (C) 2007-2013 Redouane Ezzahir & Mohamed Wahbi
 * DisChoco is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * DisChoco is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 * To contact the authors: <http://dischoco.sourceforge.net/>
 */
package util.graph;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Deque;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Scanner;
import java.util.Set;
import java.util.Stack;
import java.util.TreeSet;
import java.util.Vector;
import util.HashList;
import util.tree.AbstractTree;
import util.tree.AbstractTreeNode;

/**
 * This class provides an abstract implementation of old.
 * 
 * @author Mohamed Wahbi
 * 
 */
public abstract class AbstractGraph<T extends Comparable<? super T>> {

	// <T extends ? & Comparable<T>>

	public static final String DOT = "/usr/local/bin/dot";
	protected static final String EDGE_SEPARATOR = ";";
	protected final Vector<Vertex<T>> vertexes;
	/** A list of vertex clusters */
	protected Map<Integer, HashList<T>> clusters = null;

	/** The index of the cluster each vertex belongs to */
	protected Map<T, Integer> clusterOf = null;

	/** Floyd-Warshall all-pairs shortest path algorithm */
	private Map<Vertex<T>, Map<Vertex<T>, Integer>> allShortestPathsDistance;
	private Map<Vertex<T>, Map<Vertex<T>, Deque<Vertex<T>>>> allShortestPaths;
	private int[][] predecessor; // storage for the predecessors
	private int[][] distances; // storage for the d-values
	private ArrayList<Vertex<T>> diameter;

	/**
	 * Creates an empty old.
	 * 
	 */
	public AbstractGraph() {
		this.vertexes = new Vector<Vertex<T>>();
		this.clusters = new HashMap<Integer, HashList<T>>();
		this.clusterOf = new HashMap<T, Integer>();
	}

	public AbstractGraph(AbstractGraph<T> graph) {
		this();
		// Copy the vertices (which copies the edges)
		for(Vertex<T> v : graph.getVertexes()) {
			this.addVertex(new Vertex<T>(v.getData()));
		}
		// Update the object references
		Collection<Edge<T>> edges = graph.getEdges();
		for(Edge<T> e : edges) {
			Vertex<T> source = getVertexInGraph(e.getSource());
			Vertex<T> target = getVertexInGraph(e.getTarget());
			String label = e.getLabel();
			double weight = e.getWeight();
			boolean directed = e.isDirected();
			Edge<T> edge = new Edge<T>(source, target, weight, directed, label);
			this.addEdge(edge);
		}
	}

	public AbstractGraph(List<Edge<T>> edges) {
		this();
		for(Edge<T> edge : edges) {
			this.addEdge(edge);
		}
	}

	/**
	 * This method returns a set vertexes of this old.
	 * 
	 * @return an unmodifiable set of vertexes.
	 */
	public List<Vertex<T>> getVertexes() {
		return Collections.unmodifiableList(this.vertexes);
	}

	public HashList<Edge<T>> getEdges() {
		HashList<Edge<T>> edges = new HashList<Edge<T>>();
		for(Vertex<T> vertex : getVertexes()) {
			edges.addAll(vertex.getOutgoingEdges());
		}
		return edges;
	}

	public Collection<Edge<T>> getEdges(T data) {
		return this.getEdges(new Vertex<T>(data));
	}

	public Collection<Edge<T>> getEdges(Vertex<T> vertex) {
		return Collections.unmodifiableCollection(getVertexInGraph(vertex).getOutgoingEdges());
	}

	/**
	 * Adds a new isolated vertex to this old and returns it.
	 * 
	 * @return new vertex
	 */
	public Vertex<T> addVertex(T data) {
		return this.addVertex(new Vertex<T>(data));
	}

	/**
	 * Adds given vertex to old. Vertex shouldn't be in old and shouldn't have any edges.
	 * 
	 * @param vertex vertex to be added
	 */
	public Vertex<T> addVertex(Vertex<T> vertex) {
		Vertex<T> u = getVertexInGraph(vertex);
		if (u == null) {
			this.vertexes.add(vertex);
			return vertex;
		}
		return u;
	}

	/**
	 * Removes the given vertex from the graph.
	 * 
	 * @param data the data contained by the vertex to remove
	 * @return true if given vertex was in old; false otherwise
	 * 
	 */
	public boolean removeVertex(T data) {
		return this.removeVertex(new Vertex<T>(data));
	}

	/**
	 * Removes the given vertex from the graph.
	 * 
	 * @param vertex vertex to remove
	 * @return true if given vertex was in old; false otherwise
	 * 
	 */
	public abstract boolean removeVertex(Vertex<T> vertex);

	/**
	 * Checks whether the current old is directed or not.
	 * 
	 * @return true if it is; false otherwise
	 */
	public abstract boolean isDirected();

	/**
	 * Checks whether two given vertexes are connected with an edge.
	 * 
	 * @param source source vertex
	 * @param target target vertex
	 * @return true if an edge exists; false otherwise
	 */
	public boolean areConnected(T source, T target) {
		return (getConnectingEdge(source, target) != null);
	}

	/**
	 * Checks whether two given vertexes are connected with an edge.
	 * 
	 * @param source source vertex
	 * @param target target vertex
	 * @return true if an edge exists; false otherwise
	 */
	public boolean areConnected(Vertex<T> source, Vertex<T> target) {
		return (getConnectingEdge(source, target) != null);
	}

	/**
	 * Returns the edge connecting the given vertexes.
	 * 
	 * @param source source vertex
	 * @param target target vertex
	 * @return edge connecting vertexes if they are connected; null otherwise
	 */
	public Edge<T> getConnectingEdge(T source, T target) {
		return this.getConnectingEdge(new Vertex<T>(source), new Vertex<T>(target));
	}

	/**
	 * Returns the edge connecting the given vertexes.
	 * 
	 * @param source source vertex
	 * @param target target vertex
	 * @return edge connecting vertexes if they are connected; null otherwise
	 */
	public abstract Edge<T> getConnectingEdge(Vertex<T> source, Vertex<T> target);

	/**
	 * Connects given vertexes with an edge and returns it. Multiple edges are prohibited.
	 * 
	 * @param source source vertex
	 * @param target target vertex
	 * @return new edge
	 */
	public Edge<T> addEdge(T source, T target) {
		return this.addEdge(new Vertex<T>(source), new Vertex<T>(target));
	}

	/**
	 * Connects given vertexes with an edge and returns it. Multiple edges are prohibited.
	 * 
	 * @param source source vertex
	 * @param target target vertex
	 * @return new edge
	 */
	public abstract Edge<T> addEdge(Vertex<T> source, Vertex<T> target);

	/**
	 * Connects given vertexes with an edge with the given weight and returns it. Multiple edges are prohibited.
	 * 
	 * @param source source vertex
	 * @param target target vertex
	 * @param weight edge weight
	 * @return new weighted edge
	 */
	public Edge<T> addEdge(T source, T target, double weight) {
		return this.addEdge(new Vertex<T>(source), new Vertex<T>(target), weight);
	}

	/**
	 * Connects given vertexes with an edge with the given weight and returns it. Multiple edges are prohibited.
	 * 
	 * @param source source vertex
	 * @param target target vertex
	 * @param weight edge weight
	 * @return new weighted edge
	 */
	public abstract Edge<T> addEdge(Vertex<T> source, Vertex<T> target, double weight);

	/**
	 * Connects given vertexes with an edge with the given weight and returns it. Multiple edges are prohibited.
	 * 
	 * @param source source vertex
	 * @param target target vertex
	 * @param weight edge weight
	 * @param label edge label
	 * @return new weighted edge
	 */
	public Edge<T> addEdge(T source, T target, String label) {
		return this.addEdge(new Vertex<T>(source), new Vertex<T>(target), label);
	}

	/**
	 * Connects given vertexes with an edge with the given weight and returns it. Multiple edges are prohibited.
	 * 
	 * @param source source vertex
	 * @param target target vertex
	 * @param weight edge weight
	 * @param label edge label
	 * @return new weighted edge
	 */
	public abstract Edge<T> addEdge(Vertex<T> source, Vertex<T> target, String label);

	/**
	 * Connects given vertexes with an edge with the given weight and returns it. Multiple edges are prohibited.
	 * 
	 * @param source source vertex
	 * @param target target vertex
	 * @param weight edge weight
	 * @param label edge label
	 * @return new weighted edge
	 */
	public Edge<T> addEdge(T source, T target, String label, double weight) {
		return this.addEdge(new Vertex<T>(source), new Vertex<T>(target), label, weight);
	}

	/**
	 * Connects given vertexes with an edge with the given weight and returns it. Multiple edges are prohibited.
	 * 
	 * @param source source vertex
	 * @param target target vertex
	 * @param weight edge weight
	 * @param label edge label
	 * @return new weighted edge
	 */
	public abstract Edge<T> addEdge(Vertex<T> source, Vertex<T> target, String label, double weight);

	/**
	 * Adds the given edge to the current old. Both edge ends should already be in old and be not connected.
	 * 
	 * @param edge edge to be added
	 */
	public abstract void addEdge(Edge<T> edge);

	/**
	 * Disconnects given vertexes.
	 * 
	 * @param source source vertex
	 * @param target target vertex
	 * @return true if given vertexes were connected; false otherwise
	 */
	public boolean disconnect(T source, T target) {
		return this.disconnect(new Vertex<T>(source), new Vertex<T>(target));
	}

	/**
	 * Disconnects given vertexes.
	 * 
	 * @param source source vertex
	 * @param target target vertex
	 * @return true if given vertexes were connected; false otherwise
	 */
	public abstract boolean disconnect(Vertex<T> source, Vertex<T> target);

	/**
	 * Enables subclasses to add vertexes to old.
	 * 
	 * @param vertex vertex to add
	 */
	protected void addVertexToGraph(Vertex<T> vertex) {
		this.vertexes.add(vertex);
	}

	/**
	 * Enables subclasses to remove vertexes from old.
	 * 
	 * @param vertex vertex to remove.
	 */
	protected void removeVertexFromGraph(Vertex<T> vertex) {
		vertex.removeEdges();
		this.vertexes.remove(vertex);
	}

	/**
	 * Checks whether the given vertex is in old.
	 * 
	 * @param vertex vertex
	 * @return true if vertex is in old; false otherwise
	 */
	protected Vertex<T> getVertexInGraph(Vertex<T> vertex) {
		int index = this.vertexes.indexOf(vertex);
		if (index < 0) {
			return null;
		}
		return vertexes.get(index);
	}

	/**
	 * Checks whether the given vertex is in old.
	 * 
	 * @param vertex vertex
	 * @return true if vertex is in old; false otherwise
	 */
	public boolean isVertexInGraph(T source) {
		return this.isVertexInGraph(new Vertex<T>(source));
	}

	/**
	 * Checks whether the given vertex is in old.
	 * 
	 * @param vertex vertex
	 * @return true if vertex is in old; false otherwise
	 */
	public boolean isVertexInGraph(Vertex<T> source) {
		return this.vertexes.contains(source);
	}

	/**
	 * Checks whether the given vertexes are in old. If one is not
	 * <tt>IllegalArgumentException</tt> is thrown.
	 * 
	 * @param source source vertex
	 * @param target
	 *            target vertex
	 */
	protected void testVertices(Vertex<T> source, Vertex<T> target) {
		if (!isVertexInGraph(source)) {
			throw new IllegalArgumentException("Source vertex not in this old.");
		}
		if (!isVertexInGraph(target)) {
			throw new IllegalArgumentException("Target vertex " + target + " not in this: " + this.vertexes);
		}
	}

	/**
	 * Returns a string representation of the graph
	 * 
	 * @return String 
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append("[");
		for(Edge<T> e : getEdges()) {
			sb.append(e.toString() + EDGE_SEPARATOR);
		}
		sb.deleteCharAt(sb.length() - 1);
		sb.append("]");
		return sb.toString();
	}

	/**
	 * Returns the diameter of this graph. The diameter is defined to be the maximum, over all pairs of 
	 * vertexes <code>u,v</code>, of the length of the shortest path from <code>u</code> to <code>v</code>. 
	 * If the graph is disconnected, the value returned is the the maximum shortest path length over all 
	 * pairs of <b>connected</b> vertexes.
	 */
	public int getDiameter(boolean compute) {
		if (compute) {
			FloydWarshall();
		}
		return diameter.size();
	}

	public int getShorthestDistance(Vertex<T> u, Vertex<T> v) {
		int i = this.vertexes.indexOf(u);
		int j = this.vertexes.indexOf(v);
		return distances[i][j];
	}

	public ArrayList<Vertex<T>> getDiameterPath(boolean compute) {
		if (compute) {
			FloydWarshall();
		}
		return diameter;
	}

	public static <T extends Comparable<? super T>> String toDotFile(AbstractGraph<T> graph, String fileName) {
		String dotFileName = new File(fileName).getAbsolutePath();
		if (fileName.contains(".")) {
			dotFileName = dotFileName.substring(0, dotFileName.length() - 4) + ".dot";
		}
		try {
			PrintWriter writer = new PrintWriter(new File(fileName));
			writer.println(graph.toDOTString());
			writer.close();
		} catch (FileNotFoundException e) {
			System.err.println("Could not create DOT output file with name " + fileName);
			e.printStackTrace();
		}
		return dotFileName;
	}

	/**
	 * This method generates a PDF file from the given DOT file. 
	 * 
	 * @param fileName : the DOT file to convert into a PDF file. 
	 * @return 
	 */
	public static String fromDOTtoPDF(String fileName) {
		File input = new File(fileName);
		try {
			String pdfFileName = input.getAbsolutePath();
			if (fileName.contains(".")) {
				pdfFileName = pdfFileName.substring(0, pdfFileName.length() - 4) + ".pdf";
			}
			Process p = Runtime.getRuntime().exec(new String[] {DOT, "-Tpdf", input.getAbsolutePath(), "-o", pdfFileName});
			final Scanner err = new Scanner(p.getErrorStream());
			new Thread(new Runnable() {

				public void run() {
					while (err.hasNextLine()) {
						System.err.println(err.nextLine());
					}
				}
			}).start();
			return pdfFileName;
		} catch (IOException e) {
			if (e.getMessage().contains("Cannot run program")) {
				System.err.println("Can not run \"" + DOT + "\". Probably, the program is not installed, or it is not on the search path.");
			} else {
				e.printStackTrace();
			}
		}
		return null;
	}

	//
	// /**
	// * This method represents the graph in the DOT format.
	// *
	// * @return a String representation of the graph in DOT format
	// */
	// public String toDOTString() {
	// StringBuilder sb = new StringBuilder("digraph G {\n");
	// sb.append("\tbgcolor=\"white\"\n");
	// sb.append("\tlabel=\"A Constraint Graph Generated by DisChoco\"\n");
	// sb.append("\tfontcolor=\"red\"\n");
	// sb.append("\tnode [fillcolor=\"white\" style=\"filled\" color=\"black\" fontcolor=\"blue\"]\n");
	// sb.append("\tedge [dir=none,color=black];\n");
	// for(Edge<T> edge : this.getEdges()) {
	// sb.append("\t" + edge.toDOT() + ";\n");// [color=blue,style=bold];\n");
	// }
	// sb.append("}\n");
	// return sb.toString();
	// }
	/** 
	 * This method represents the graph in the DOT format. 
	 * 
	 * @return a String representation of the graph in DOT format
	 */
	public String toDOTString() {
		StringBuilder sb = new StringBuilder("digraph G {\n");
		sb.append("\tbgcolor=\"white\"\n");
		// out.append("\tcompound=true;\n");
		sb.append("\tlabel=\"A Constraint Graph Generated by DisChoco\"\n");
		sb.append("\tfontcolor=\"red\"\n");
		sb.append("\tnode [fillcolor=\"white\" style=\"filled\" color=\"black\" fontcolor=\"blue\"]\n");
		if (clusters.isEmpty()) {
			for(Vertex<T> vertex : this.getVertexes()) {
				sb.append("" + vertex + ";\n");
			}
		} else {
			for(Integer cluster : clusters.keySet()) {
				sb.append("\tsubgraph cluster" + cluster + " {\n");
				sb.append("\t\tshape=box\n");
				sb.append("\t\tfillcolor=\"lightgray\"\n");
				sb.append("\t\tlabel = \"Agent " + cluster + "\"\n");
				sb.append("\t\tfontcolor=\"black\"\n");
				sb.append("\t\tstyle=\"filled\"\n");
				HashList<T> vertices = clusters.get(cluster);
				if (vertices.isEmpty())
					sb.append("\t\tempty_" + cluster + " [label=\"\" shape=none];\n");
				for(T v : vertices) {
					sb.append("\t\t\"" + v + "\"");
					// // If this node is connected to at least one node inside a different cluster, fill it
					// for (V neigh : this.graph.get(v)) {
					// if (i != this.clusterOf.get(neigh)) {
					// out.append(" [shape=circle,style=filled,color=lightblue]");
					// //out.append(";\n\t\tcolor=blue");
					// break;
					// }
					// }
					sb.append(";\n");
				}
				sb.append("\t}\n");
			}
		}
		sb.append("\n\tedge [dir=none,color=black];\n");
		for(Edge<T> edge : this.getEdges()) {
			sb.append("\t" + edge.toDOT() + ";\n");// [color=blue,style=bold];\n");
		}
		sb.append("}\n");
		return sb.toString();
	}

	public AbstractTree<T> DFS() {
		Collections.sort(this.vertexes, new MaxDegree());
		for(Vertex<T> v : vertexes) {
			v.setState(VertexState.White);
		}
//		return new AbstractTree<T>(DFS(vertexes.firstElement()));
		if (this.isConnected()) {
			return new AbstractTree<T>(DFS(vertexes.firstElement()));
		} else {
			List<AbstractTree<T>> listDFS = new ArrayList<AbstractTree<T>>();
			for(AbstractGraph<T> subGraph : this.getUnconnectedSubGraphs()) {
				Collections.sort(subGraph.vertexes, new MaxDegree());
				listDFS.add(new AbstractTree<T>(subGraph.DFS(subGraph.vertexes.firstElement())));
			}
			Collections.sort(listDFS, new Comparator<AbstractTree<T>>() {
				@Override
				public int compare(AbstractTree<T> o1, AbstractTree<T> o2) {
					return o2.getDepth() - o1.getDepth();
				}
			});
			AbstractTree<T> dfs = listDFS.remove(0);
			AbstractTreeNode<T> root = dfs.getRoot();
			while(!listDFS.isEmpty()) {
				AbstractTreeNode<T> subRoot = listDFS.remove(0).getRoot();
				//Edge<T> edge = new Edge<T>(root.getNode(), subRoot.getNode(), 0, true, "");
				root.addChild(subRoot);
			}
			return dfs;
		}
	}

	public AbstractTreeNode<T> DFS(Vertex<T> u) {
		u.setState(VertexState.Gray);
		AbstractTreeNode<T> dfs = new AbstractTreeNode<T>(u);
		TreeSet<Vertex<T>> neighbors = new TreeSet<Vertex<T>>(new MaxDegree());
		neighbors.addAll(u.getNeighbors());
		for(Vertex<T> target : neighbors) {
			Edge<T> edge = getConnectingEdge(u, target);
			if (target.getState() == VertexState.White) {
				// dfs.getChildren().add(DFS(target));
				// Edge<T> dircetedEdge = edge.getDirected(u);
				// dfs.getNode().addOutgoingEdge(dircetedEdge);
				dfs.addChild(DFS(target), edge.getDirected(u));
			} else if (target.getState() == VertexState.Black) {
				edge.setBackEdge(true);
				dfs.getNode().addOutgoingEdge(edge);
			}
		}
		u.setState(VertexState.Black);
		return dfs;
	}

	/**
	 * This method computes a DFS-tree from the current graph using and adaptation of the 
	 * Maximum Cardinality Set heuristic proposed by [R. E. Tarjan and M. Yannakakis; 1984]. 
	 * ``Simple linear-time algorithms to test chordality of graphs, test acyclicity 
	 * of hypergraphs, and selectively reduce acyclic hypergraphs''. 
	 * 
	 * At each step, for a given vertex, select a not yet visited neighbor having the maximum 
	 * already visited neighbors. 
	 * 
	 * @return Depth-First Search Tree
	 */
	public AbstractTree<T> MCSDFS() {
		Collections.sort(this.vertexes, new MaxDegree());
		for(Vertex<T> v : vertexes) {
			v.setState(VertexState.White);
		}
//		return new AbstractTree<T>(DFS(vertexes.firstElement(), new MCS()));
		if(this.isConnected()) {
			return new AbstractTree<T>(DFS(vertexes.firstElement(), new MCS()));
		} else {
			List<AbstractTree<T>> listDFS = new ArrayList<AbstractTree<T>>();
			for(AbstractGraph<T> subGraph : this.getUnconnectedSubGraphs()) {
				Collections.sort(subGraph.vertexes, new MaxDegree());
				listDFS.add(new AbstractTree<T>(subGraph.DFS(subGraph.vertexes.firstElement(), new MCS())));
			}
			Collections.sort(listDFS, new Comparator<AbstractTree<T>>() {
				@Override
				public int compare(AbstractTree<T> o1, AbstractTree<T> o2) {
					return o2.getDepth() - o1.getDepth();
				}
			});
			AbstractTree<T> dfs = listDFS.remove(0);
			AbstractTreeNode<T> root = dfs.getRoot();
			while(!listDFS.isEmpty()) {
				AbstractTreeNode<T> subRoot = listDFS.remove(0).getRoot();
				//Edge<T> edge = new Edge<T>(root.getNode(), subRoot.getNode(), 0, true, "");
				root.addChild(subRoot);
			}
			return dfs;
		}
	}

	/**
	 * This method computes a DFS-tree from the current graph using and adaptation of the 
	 * Maximum Cardinality Set heuristic proposed by [R. E. Tarjan and M. Yannakakis; 1984]. 
	 * ``Simple linear-time algorithms to test chordality of graphs, test acyclicity 
	 * of hypergraphs, and selectively reduce acyclic hypergraphs''. 
	 * 
	 * At each step, for a given vertex, select a not yet visited neighbor having the maximum 
	 * already visited neighbors. 
	 * 
	 * @return Depth-First Search Tree
	 */
	public AbstractTree<T> DFS(Comparator<Vertex<T>> comparator) {
		Collections.sort(this.vertexes, comparator);
		for(Vertex<T> v : vertexes) {
			v.setState(VertexState.White);
		}
//		return new AbstractTree<T>(DFS(vertexes.firstElement(), comparator));
		if(this.isConnected()) {
			return new AbstractTree<T>(DFS(vertexes.firstElement(), comparator));
		} else {
			List<AbstractTree<T>> listDFS = new ArrayList<AbstractTree<T>>();
			for(AbstractGraph<T> subGraph : this.getUnconnectedSubGraphs()) {
				Collections.sort(subGraph.vertexes, new MaxDegree());
				listDFS.add(new AbstractTree<T>(subGraph.DFS(subGraph.vertexes.firstElement(), comparator)));
			}
			Collections.sort(listDFS, new Comparator<AbstractTree<T>>() {
				@Override
				public int compare(AbstractTree<T> o1, AbstractTree<T> o2) {
					return o2.getDepth() - o1.getDepth();
				}
			});
			AbstractTree<T> dfs = listDFS.remove(0);
			AbstractTreeNode<T> root = dfs.getRoot();
			while(!listDFS.isEmpty()) {
				AbstractTreeNode<T> subRoot = listDFS.remove(0).getRoot();
				//Edge<T> edge = new Edge<T>(root.getNode(), subRoot.getNode(), 0, true, "");
				root.addChild(subRoot);
			}
			return dfs;
		}
	}

	/**
	 * This method computes a DFS-tree from the current graph using and adaptation of the 
	 * Maximum Cardinality Set heuristic proposed by [R. E. Tarjan and M. Yannakakis; 1984]. 
	 * ``Simple linear-time algorithms to test chordality of graphs, test acyclicity 
	 * of hypergraphs, and selectively reduce acyclic hypergraphs''. 
	 * 
	 * At each step, for a given vertex, select a not yet visited neighbor having the maximum 
	 * already visited neighbors. 
	 * 
	 * @return Depth-First Search Tree
	 */
	public AbstractTree<T> DynamicMaxDegreeDFS() {
		return DFS(new DynamicMaxDegreeDFS());
//		Collections.sort(this.vertexes, new DynamicMaxDegreeDFS());
//		for(Vertex<T> v : vertexes) {
//			v.setState(VertexState.White);
//		}
//		return new AbstractTree<T>(DFS(vertexes.firstElement(), new DynamicMaxDegreeDFS()));
	}

	/**
	 * This method computes a DFS-tree from the current graph using and adaptation of the 
	 * Maximum Cardinality Set heuristic proposed by [R. E. Tarjan and M. Yannakakis; 1984]. 
	 * ``Simple linear-time algorithms to test chordality of graphs, test acyclicity 
	 * of hypergraphs, and selectively reduce acyclic hypergraphs''. 
	 * 
	 * At each step, for a given vertex, select a not yet visited neighbor having the maximum neighbors. 
	 * 
	 * @return Depth-First Search Tree
	 */
	public AbstractTree<T> MaxDegreeDFS() {
		return DFS(new MaxDegree());
//		Collections.sort(this.vertexes, new MaxDegree());
//		for(Vertex<T> v : vertexes) {
//			v.setState(VertexState.White);
//		}
//		return new AbstractTree<T>(DFS(vertexes.firstElement(), new MaxDegree()));
	}

	/**
	 * This method computes a DFS-tree from the current graph using and adaptation of the 
	 * Mid-point of the Longest Shortest Path (MLSP) heuristic proposed by [R. T. Maheswaran, et al. 2005] in 
	 * <i>``Taking DCOP to the Real World: Efficient Complete Solutions for Distributed Multi-Event Scheduling''</i>. 
	 * 
	 * @return Depth-First Search Tree
	 */
	public abstract AbstractTree<T> MLSPTree();

	/**
	 * This method computes a DFS-tree from the given graph {@code graph} rooted at the given root {@code root}. 
	 * The heuristic used is an adaptation of the Mid-point of the Longest Shortest Path (MLSP) heuristic 
	 * proposed by [R. T. Maheswaran, et al. 2005] in 
	 * <i>``Taking DCOP to the Real World: Efficient Complete Solutions for Distributed Multi-Event Scheduling''</i>. 
	 * 
	 * @param root
	 * @param graph
	 * @return Depth-First Search Tree rooted at the given {@code root}.
	 */
	public AbstractTreeNode<T> MLSPTree(Vertex<T> root, AbstractGraph<T> graph) {
		root.setState(VertexState.Gray);
		assert (graph.isConnected()) : " graph shold connected";
		Collection<AbstractGraph<T>> subGraphs = graph.removeFromGraph(root);
		AbstractTreeNode<T> dfs = new AbstractTreeNode<T>(root);
		for(AbstractGraph<T> subGraph : subGraphs) {
			Vertex<T> r = subGraph.getVertexInGraph(root);
			List<Vertex<T>> possibleChildren = new ArrayList<Vertex<T>>(r.getNeighbors());
			subGraph.removeVertex(r);
			ArrayList<Vertex<T>> diameter = subGraph.getDiameterPath(true);
			Vertex<T> midNode = diameter.get(diameter.size() / 2);
			Collections.sort(possibleChildren, new Closest(subGraph, midNode));
			Vertex<T> closestNode = possibleChildren.remove(0);
			Vertex<T> child = subGraph.getVertexInGraph(closestNode);
			Edge<T> edge = graph.getConnectingEdge(r, child);
			dfs.addChild(MLSPTree(child, subGraph), edge.getDirected(r));
			possibleChildren.retainAll(dfs.getAllDescendants());
			for(Vertex<T> neighbor : possibleChildren) {
				Edge<T> backEdge = graph.getConnectingEdge(r, neighbor);
				backEdge.setBackEdge(true);
				dfs.getNode().addOutgoingEdge(backEdge);
			}
		}
		root.setState(VertexState.Black);
		return dfs;
	}

	/**
	 * 
	 * At each step, for a given vertex, select a not yet visited neighbor having the maximum 
	 * already visited neighbors. 
	 * 
	 * @param u
	 * @return
	 */
	public AbstractTreeNode<T> DFS(Vertex<T> u, Comparator<Vertex<T>> comparator) {
		u.setState(VertexState.Gray);
		AbstractTreeNode<T> dfs = new AbstractTreeNode<T>(u);
		List<Vertex<T>> neighbors = new ArrayList<Vertex<T>>(u.getNeighbors());
		while (!neighbors.isEmpty()) {
			Collections.sort(neighbors, comparator);
			Vertex<T> target = neighbors.remove(0);
			Edge<T> edge = getConnectingEdge(u, target);
			if (target.getState() == VertexState.White) {
				dfs.addChild(DFS(target, comparator), edge.getDirected(u));
			} else if (target.getState() == VertexState.Black) {
				edge.setBackEdge(true);
				dfs.getNode().addOutgoingEdge(edge);
			}
		}
		u.setState(VertexState.Black);
		return dfs;
	}

	/**
	 * This method implements the Floyd-Warshall all-pairs shortest path algorithm.
	 * This algorithm assumes that there is no negative cycles in the input graph. 
	 * Reference: "The Floyd-Warshall algorithm on graphs with negative cycles"
	 * @param path
	 */
	public static void FloydWarshall(int[][] path) {
		for(int k = 0 ; k < path.length ; k++) {
			for(int i = 0 ; i < path.length ; i++) {
				for(int j = 0 ; j < path.length ; j++) {
					path[i][j] = Math.min(path[i][j], path[i][k] + path[k][j]);
				}
			}
		}
	}

	/**
	 * This method implements the Floyd-Warshall all-pairs shortest path algorithm.
	 * This algorithm assumes that there is no negative cycles in the input graph. 
	 * Reference: "The Floyd-Warshall algorithm on graphs with negative cycles"
	 * 
	 * @param graph
	 * @return
	 */
	public void FloydWarshall() {
		allShortestPathsDistance = new HashMap<Vertex<T>, Map<Vertex<T>, Integer>>();
		allShortestPaths = new HashMap<Vertex<T>, Map<Vertex<T>, Deque<Vertex<T>>>>();

		int n = vertexes.size();
		// diameter = new ArrayList<Vertex<T>>(n);
		diameter = new ArrayList<Vertex<T>>(n);
		diameter.add(vertexes.get(0));
		distances = new int[n][n];
		predecessor = new int[n][n];

		for(int i = 0 ; i < n ; i++) {
			Arrays.fill(distances[i], Integer.MAX_VALUE);
			Arrays.fill(predecessor[i], -1);
		}
		for(Edge<T> e : getEdges()) {
			int i = vertexes.indexOf(e.getSource());
			int j = vertexes.indexOf(e.getTarget());
			distances[i][j] = 1;
			distances[j][i] = 1;
			predecessor[i][j] = i;
			predecessor[j][i] = j;
		}

		for(int i = 0 ; i < n ; i++) {
			distances[i][i] = 0;
		}

		for(int k = 0 ; k < n ; k++) {
			for(int i = 0 ; i < n ; i++) {
				for(int j = 0 ; j < n ; j++) {
					if (i == j) {
						distances[i][j] = 0;
					} else {
						int ijCost = distances[i][j];
						int ikCost = distances[i][k];
						int kjCost = distances[k][j];
						int summed = (ikCost != Integer.MAX_VALUE && kjCost != Integer.MAX_VALUE)? (ikCost + kjCost) : Integer.MAX_VALUE;
						if (ijCost > summed) {
							distances[i][j] = summed;
							predecessor[i][j] = predecessor[k][j];
						}
					}
				}
			}
		}

		for(int k = 0 ; k < n ; k++) {
			for(int i = 0 ; i < n ; i++) {
				for(int j = 0 ; j < n ; j++) {
					int ijCost = distances[i][j];
					int ikCost = distances[i][k];
					int kjCost = distances[k][j];
					int summed = (ikCost != Integer.MAX_VALUE && kjCost != Integer.MAX_VALUE)? (ikCost + kjCost) : Integer.MAX_VALUE;
					if (ijCost > summed) {
						distances[i][j] = summed;
						predecessor[i][j] = predecessor[k][j];
					}
				}
			}
		}

		for(int i = 0 ; i < n ; i++) {
			Vertex<T> from = vertexes.get(i);
			Map<Vertex<T>, Deque<Vertex<T>>> paths = new HashMap<Vertex<T>, Deque<Vertex<T>>>();
			Map<Vertex<T>, Integer> shortestCost = new HashMap<Vertex<T>, Integer>();
			for(int j = 0 ; j < n ; j++) {
				if (i == j) {
					continue;
				}
				Vertex<T> to = vertexes.get(j);
				int cost = distances[i][j];
				if (cost != Integer.MAX_VALUE) {
					shortestCost.put(to, cost);
					int p = predecessor[i][j];
					Deque<Vertex<T>> path = new ArrayDeque<Vertex<T>>(cost + 1);
					path.push(to);
					while (p != i && p != -1) {
						Vertex<T> pv = vertexes.get(p);
						path.push(pv);
						p = predecessor[i][p];
					}
					path.push(from);
					paths.put(to, path);
					if (diameter.size() < path.size()) { // since diameter contains vertexes and cost is in number of edges
						diameter = new ArrayList<Vertex<T>>(path);
						// System.out.println("new Diameter ("+cost+"), diameter="+diameter+":"+diameter.size()+", path="+path+":"+path.size());
					}
					if (diameter.size() == path.size()) {
						// int degDiameter = 0;
						// for(Vertex<T> v : diameter) {
						// degDiameter+=v.getDegree();
						// }
						// int degPath = 0;
						// for(Vertex<T> v : path) {
						// degPath+=v.getDegree();
						// }
						// if(degPath > degDiameter) {
						// diameter = new ArrayList<Vertex<T>>(path);
						// }
					}
				}
				// allShortestPathsDistance.put(from, map);
				// if(getDiameter() < cost) { // getDiameter() {return (diameter==null)?-1:diameter.size();}
				// diameter = new ArrayList<Vertex<T>>(path);
				// //System.out.println("new Diameter diameter="+diameter+", path="+path);
				// }
				// System.out.println(path);
			}
			allShortestPathsDistance.put(from, shortestCost);
			// System.out.println(from+" : "+allShortestPathsDistance.get(from));
			allShortestPaths.put(from, paths);
			// System.out.println(from+" @ "+shortestCost);
		}
	}

	/**
	 * This method implements the Floyd-Warshall all-pairs shortest path algorithm.
	 * This algorithm assumes that there is no negative cycles in the input graph. 
	 * Reference: "The Floyd-Warshall algorithm on graphs with negative cycles"
	 * 
	 * @param graph
	 * @return
	 */
	public static <T extends Comparable<T>> Map<Vertex<T>, Map<Vertex<T>, Integer>> getAllPairsShortestPaths(AbstractGraph<T> graph) {
		Map<Vertex<T>, Map<Vertex<T>, Integer>> allShortestPaths = new HashMap<Vertex<T>, Map<Vertex<T>, Integer>>();

		List<Vertex<T>> vertexes = graph.getVertexes();
		int n = vertexes.size();
		int[][] sums = new int[n][n];

		for(int i = 0 ; i < n ; i++) {
			for(int j = 0 ; j < n ; j++) {
				sums[i][j] = Integer.MAX_VALUE;
			}
		}

		Collection<Edge<T>> edges = graph.getEdges();
		for(Edge<T> e : edges) {
			int source = vertexes.indexOf(e.getSource());
			int target = vertexes.indexOf(e.getTarget());
			sums[source][target] = 1;
			sums[target][source] = 1;
		}

		for(int k = 0 ; k < n ; k++) {
			for(int i = 0 ; i < n ; i++) {
				for(int j = 0 ; j < n ; j++) {
					if (i == j) {
						sums[i][j] = 0;
					} else {
						int ijCost = sums[i][j];
						int ikCost = sums[i][k];
						int kjCost = sums[k][j];
						int summed = (ikCost != Integer.MAX_VALUE && kjCost != Integer.MAX_VALUE)? (ikCost + kjCost) : Integer.MAX_VALUE;
						if (ijCost > summed)
							sums[i][j] = summed;
					}
				}
			}
		}

		for(int k = 0 ; k < n ; k++) {
			for(int i = 0 ; i < n ; i++) {
				for(int j = 0 ; j < n ; j++) {
					int dd = sums[i][k] + sums[k][j];
					if (sums[i][j] > dd) {
						sums[i][j] = dd;
					}
				}
			}
		}

		for(int i = 0 ; i < n ; i++) {
			Vertex<T> from = vertexes.get(i);
			for(int j = 0 ; j < n ; j++) {
				Vertex<T> to = vertexes.get(j);
				Map<Vertex<T>, Integer> map = allShortestPaths.get(from);
				if (map == null)
					map = new HashMap<Vertex<T>, Integer>();
				int cost = sums[i][j];
				if (cost != Integer.MAX_VALUE)
					map.put(to, cost);
				allShortestPaths.put(from, map);
			}
		}

		return allShortestPaths;
	}

	/**
	 * MCS: Maximum Cardinality Set Comparator: 
	 * 
	 * At each step, for a given vertex, select a not yet visited neighbor having the maximum 
	 * already visited neighbors. 
	 */
	class MCS implements Comparator<Vertex<T>> {

		@Override
		public int compare(Vertex<T> o1, Vertex<T> o2) {
			int nbVisited1 = o1.getNbVisitedNeighbors();
			int nbVisited2 = o2.getNbVisitedNeighbors();
			int cmp = nbVisited2 - nbVisited1;
			if (cmp == 0) {
				return o1.compareTo(o2);
			}
			return cmp;
		}
	}

	/**
	 * MaxDegree: Maximum Connected Node Comparator: Max-degree
	 * 
	 * At each step, for a given vertex, select a not yet visited neighbor having the maximum neighbors. 
	 */
	class MaxDegree implements Comparator<Vertex<T>> {

		@Override
		public int compare(Vertex<T> o1, Vertex<T> o2) {
			int cmp = (o2.getDegree() - o1.getDegree()); // max degree
			if (cmp == 0) {
				return o1.compareTo(o2);
			}
			return cmp;
		}
	}

	/**
	 * MaxDegree: Maximum Connected Node Comparator: Max-degree
	 * 
	 * At each step, for a given vertex, select a not yet visited neighbor having the maximum neighbors. 
	 */
	class DynamicMaxDegreeDFS implements Comparator<Vertex<T>> {

		@Override
		public int compare(Vertex<T> o1, Vertex<T> o2) {
			int dynDeg1 = o1.getNbUnvisitedNeighbors();
			int dynDeg2 = o2.getNbUnvisitedNeighbors();
			int cmp = (dynDeg2 - dynDeg1); // max degree
			if (cmp == 0) {
				return new MaxDegree().compare(o1, o2);
			}
			return cmp;
		}
	}

	/**
	 * MaxDegree: Maximum Connected Node Comparator: Max-degree
	 * 
	 * At each step, for a given vertex, select a not yet visited neighbor having the maximum neighbors. 
	 */
	class Closest implements Comparator<Vertex<T>> {

		Vertex<T> from;
		AbstractGraph<T> graph;

		public Closest(AbstractGraph<T> graph, Vertex<T> vertex) {
			this.graph = graph;
			this.from = vertex;
		}

		@Override
		public int compare(Vertex<T> o1, Vertex<T> o2) {
			Map<Vertex<T>, Integer> distances = graph.allShortestPathsDistance.get(from);
			int sp1 = (distances.containsKey(o1))? distances.get(o1) : Integer.MAX_VALUE;
			int sp2 = (distances.containsKey(o2))? distances.get(o2) : Integer.MAX_VALUE;
			int cmp = (sp1 - sp2); // min Shortest Path
			if (cmp == 0) {
				return new MaxDegree().compare(o1, o2);// .compareTo(o2);
			}
			return cmp;
		}
	}

	public String pretty() {
		StringBuffer sb = new StringBuffer();
		sb.append("Vertexes");
		List<Vertex<T>> vertices = getVertexes();
		sb.append(" (" + vertices.size() + ") ");
		sb.append(" : [");
		sb.append(vertices + "]\n");
		Collection<Edge<T>> edges = getEdges();
		sb.append("Edges");
		sb.append(" (" + edges.size() + ") ");
		sb.append(" : [");
		sb.append(edges);
		sb.append("]");
		return sb.toString();
	}

	public abstract Collection<AbstractGraph<T>> removeFromGraph(Vertex<T> vertex);

	/**
	 * @param varId
	 * @return
	 */
	public HashList<Vertex<T>> getNeighbors(T data) {
		return this.getNeighbors(new Vertex<T>(data));
	}

	/**
	 * @param vertex
	 * @return
	 */
	public HashList<Vertex<T>> getNeighbors(Vertex<T> vertex) {
		return this.getVertexInGraph(vertex).getNeighbors();
	}

	/**
	 * @param varId
	 * @return
	 */
	public HashList<T> getNeighborsData(T data) {
		return this.getNeighborsData(new Vertex<T>(data));
	}

	/**
	 * @param vertex
	 * @return
	 */
	public HashList<T> getNeighborsData(Vertex<T> vertex) {
		return this.getVertexInGraph(vertex).getNeighborsData();
	}

	/**
	 * @return
	 */
	public boolean isConnected() {
		return dfsFromVertex(this).size() == getNbVertexes();
	}

	/**
	 * @param graph
	 * @param vertex
	 * @return 
	 */
	private Set<Vertex<T>> dfsFromVertex(AbstractGraph<T> graph) {
		Stack<Vertex<T>> stack = new Stack<Vertex<T>>();
		stack.push(graph.vertexes.get(0));
		Set<Vertex<T>> visitedVertex = new HashSet<Vertex<T>>();
		while (!stack.isEmpty()) {
			Vertex<T> w = stack.pop();
			for(Vertex<T> u : graph.getVertexInGraph(w).getNeighbors()) {
				if (!visitedVertex.contains(u)) {
					stack.push(u);
				}
			}
			visitedVertex.add(w);
		}
		return visitedVertex;
	}

	/**
	 * @return
	 */
	public int getNbVertexes() {
		return this.vertexes.size();
	}

	/**
	 * @return
	 */
	public int getNbEdges() {
		return this.getEdges().size();
	}

	public int maxDistance(T vertex) {
		return this.maxDistance(new Vertex<T>(vertex));
	}

	public Integer maxDistance(Vertex<T> vertex) {
		if (this.distances == null) {
			FloydWarshall();
		}
		Collection<Integer> values = allShortestPathsDistance.get(vertex).values();
		return Collections.max(new ArrayList<Integer>(values));
	}

	/**
	 * 
	 * @return
	 */
	public boolean isCyclic() {
		if (vertexes == null || vertexes.size() == 0)
			return false;
		for(Vertex<T> vertex : vertexes) {
			vertex.setState(VertexState.White);
		}
		HashSet<Edge<T>> visitedEdges = new HashSet<Edge<T>>();
		for(Vertex<T> vertex : vertexes) {
			if (vertex.getState() == VertexState.White) {
				if (depthFirstSearch(vertex, visitedEdges)) {
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * Search the graph for cycles. In order to detect cycles, we use a modified
	 * depth first search called a colored DFS. All nodes are initially marked
	 * white. When a node is encountered, it is marked grey, and when its
	 * descendants are completely visited, it is marked black. If a grey node is
	 * ever encountered, then there is a cycle.
	 * 
	 * @return the edges that form cycles in the graph. The array will be empty if
	 *         there are no cycles.
	 */
	public boolean depthFirstSearch(Vertex<T> vertex, HashSet<Edge<T>> visitedEdges) {
		if (vertex.getState() == VertexState.White) {
			vertex.setState(VertexState.Gray);
			for(Edge<T> edge : vertex.getOutgoingEdges()) {
				boolean result = false;
				if (!visitedEdges.contains(edge)) {
					visitedEdges.add(edge);
					Vertex<T> opposite = edge.getOppositeVertex(vertex);
					result = depthFirstSearch(opposite, visitedEdges);
				}
				if (result == true)
					return result;
			}
			vertex.setState(VertexState.Black);
		} else { // visited
			return true;
		}
		return false;
	}

	/**
	 * @param source
	 * @param target
	 * @param weight
	 */
	public void setWeight(T source, T target, double weight) {
		this.setWeight(new Vertex<T>(source), new Vertex<T>(target), weight);
	}

	/**
	 * @param source
	 * @param target
	 * @param weight
	 */
	public void setWeight(Vertex<T> source, Vertex<T> target, double weight) {
		Edge<T> edge = getConnectingEdge(source, target);
		if (edge != null) {
			edge.setWeight(weight);
		}
	}

	/**
	 * @param cluster
	 */
	public void createCluster(Integer cluster) {
		this.clusters.put(cluster, new HashList<T>());
	}

	/**
	 * 
	 * @param cluster
	 * @param element
	 */
	public void addToCluster(Integer cluster, T element) {
		this.clusters.get(cluster).add(element);
		this.clusterOf.put(element, cluster);
	}

	/**
	 * @return
	 */
	public abstract Collection<AbstractGraph<T>> getUnconnectedSubGraphs();

}
