/*
DisChoco: Distributed Constraint Reasoning Solver 
Copyright (C) 2007-2013 Redouane Ezzahir & Mohamed Wahbi 

DisChoco is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DisChoco is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

To contact the authors: <http://dischoco.sourceforge.net/>
*/ 
package util.graph;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Stack;
import java.util.Vector;
import util.tree.AbstractTree;

/**
 * 
 * This class provides a simple implementation of a directed old.
 * 
 * @author Mohamed Wahbi
 * 
 */
public class DirectedGraph<T extends Comparable<? super T>> extends AbstractGraph<T> {

	public DirectedGraph(DirectedGraph<T> graph) {
		super(graph);
	}

	public DirectedGraph() {
		super();
	}

	@Override
	public Edge<T> addEdge(Vertex<T> source, Vertex<T> target) {
		return addEdge(source, target, 0);
	}

	/*
	 * (non-Javadoc)
	 * @see util.graph.AbstractGraph#addEdge(util.graph.Vertex, util.graph.Vertex, java.lang.String)
	 */
	@Override
	public Edge<T> addEdge(Vertex<T> source, Vertex<T> target, String label) {
		source = addVertex(source);
		target = addVertex(target);
		Edge<T> edge = new Edge<T>(source, target, label, true);
		source.addOutgoingEdge(edge);
		target.addOutgoingEdge(edge);
		return edge;
	}

	@Override
	public Edge<T> addEdge(Vertex<T> source, Vertex<T> target, double weight) {
		source = addVertex(source);
		target = addVertex(target);
		Edge<T> edge = new Edge<T>(source, target, weight, true);
		source.addOutgoingEdge(edge);
		target.addIncomingEdge(edge);
		return edge;
	}

	@Override
	public Edge<T> addEdge(Vertex<T> source, Vertex<T> target, String label, double weight) {
		target = addVertex(source);
		target = addVertex(target);
		Edge<T> edge = new Edge<T>(source, target, weight, true, label);
		target.addOutgoingEdge(edge);
		target.addIncomingEdge(edge);
		return edge;
	}

	@Override
	public void addEdge(Edge<T> edge) {
		if (!isVertexInGraph(edge.getSource()) || !isVertexInGraph(edge.getTarget())) {
			throw new IllegalArgumentException("Edge end is not in old.");
		}
		if (areConnected(edge.getSource(), edge.getTarget())) {
			throw new IllegalArgumentException("Edge already in old.");
		}
		Vertex<T> source = edge.getSource();
		Vertex<T> target = edge.getTarget();
		edge.setSource(source);
		edge.setTarget(target);
		source.addOutgoingEdge(edge);
		target.addIncomingEdge(edge);
	}

	@Override
	public boolean disconnect(Vertex<T> source, Vertex<T> target) {
		testVertices(source, target);
		Edge<T> edge = getConnectingEdge(source, target);
		if (edge == null) {
			return false;
		}
		source.removeOutgoingEdge(edge);
		target.removeIncomingEdge(edge);
		return true;
	}

	@Override
	public boolean removeVertex(Vertex<T> vertex) {
		if (!isVertexInGraph(vertex)) {
			return false;
		}
		for(Edge<T> edge : vertex.getOutgoingEdges()) {
			Vertex<T> opposite = edge.getOppositeVertex(vertex);
			opposite.removeOutgoingEdge(edge);
			getVertexInGraph(opposite).removeOutgoingEdge(edge);
		}
		for(Edge<T> edge : vertex.getIncomingEdges()) {
			Vertex<T> opposite = edge.getOppositeVertex(vertex);
			opposite.removeOutgoingEdge(edge);
			getVertexInGraph(opposite).removeOutgoingEdge(edge);
		}
		removeVertexFromGraph(vertex);
		return true;
	}

	@Override
	public Edge<T> getConnectingEdge(Vertex<T> source, Vertex<T> target) {
		testVertices(source, target);
		for(Edge<T> edge : source.getOutgoingEdges()) {
			if (edge.getTarget().equals(target)) {
				return edge;
			}
		}
		return null;
	}

	@Override
	public boolean isDirected() {
		return true;
	}

	@Override
	public Collection<AbstractGraph<T>> removeFromGraph(Vertex<T> vertex) {
		DirectedGraph<T> graph = new DirectedGraph<T>(this);
		graph.removeVertex(vertex);
		return DFS(graph);
	}

	public static <T extends Comparable<? super T>> Collection<AbstractGraph<T>> DFS(DirectedGraph<T> graph) {
		Collection<AbstractGraph<T>> subGraphs = new Vector<AbstractGraph<T>>();
		List<Vertex<T>> unvisitedVertex = graph.getVertexes();
		while (!unvisitedVertex.isEmpty()) {
			Vertex<T> v = unvisitedVertex.get(0);
			DirectedGraph<T> subGraph = dfsFromVertex(graph, v);
			subGraphs.add(subGraph);
			unvisitedVertex.removeAll(subGraph.getVertexes());
		}
		return subGraphs;
	}

	public static <T extends Comparable<? super T>> DirectedGraph<T> dfsFromVertex(DirectedGraph<T> graph, Vertex<T> v) {
		Stack<Vertex<T>> stack = new Stack<Vertex<T>>();
		stack.push(v);
		List<Vertex<T>> visitedVertex = new ArrayList<Vertex<T>>();;
		while (!stack.isEmpty()) {
			Vertex<T> w = stack.pop();
			visitedVertex.add(w);
			if (!visitedVertex.contains(w)) {
				for(Edge<T> edge : v.getOutgoingEdges()) {
					stack.push(edge.getTarget());
				}
			}
		}
		DirectedGraph<T> dfs = new DirectedGraph<T>();
		for(Vertex<T> w : visitedVertex) {
			for(Vertex<T> u : visitedVertex) {
				Edge<T> edge = graph.getConnectingEdge(w, u);
				if (edge != null) {
					dfs.addEdge(edge);
				}
			}
		}
		return dfs;
	}

	@Override
	public AbstractTree<T> MLSPTree() {
		// TODO Auto-generated method stub
		return null;
	}

	public static <T extends Comparable<? super T>> boolean hasCycle(DirectedGraph<T> graph) {
		if (graph == null || graph.getNbVertexes() == 0)
			return false;
		Collection<Vertex<T>> vertexCollect = graph.getVertexes();
		Queue<Vertex<T>> q; // Queue will store vertices that have in-degree of zero
		Map<Vertex<T>, Integer> inDegree = new HashMap<Vertex<T>, Integer>();
		long counter = 0;

		/* Calculate the in-degree of all vertices and store on the Vertex */
		for(Vertex<T> v : vertexCollect) {
			inDegree.put(v, 0);
		}
		for(Vertex<T> v : vertexCollect) {
			for(Edge<T> edge : v.getOutgoingEdges()) {
				Vertex<T> target = edge.getTarget();
				int in = inDegree.get(target);
				inDegree.put(target, in++);
			}
		}

		/* Find all vertices with in-degree == 0 and put in queue */
		q = new LinkedList<Vertex<T>>();
		for(Vertex<T> v : vertexCollect) {
			if (inDegree.get(v) == 0)
				q.offer(v);
		}
		if (q.size() == 0) {
			return true; // Cycle found  all vertices have in-bound edges
		}
		/* Traverse the queue counting Vertices visited */
		for(counter = 0 ; !q.isEmpty() ; counter++) {
			Vertex<T> v = q.poll();
			for(Edge<T> e : v.getOutgoingEdges()) {
				Vertex<T> w = e.getTarget();
				if (inDegree.get(w) == 0) {
					q.offer(w);
				}
			}
		}
		if (counter != vertexCollect.size()) {
			return true; // Cycle found
		}
		return false;
	}


	
	@Override
	public Collection<AbstractGraph<T>> getUnconnectedSubGraphs() {
		DirectedGraph<T> graph = new DirectedGraph<T>(this);
		return DFS(graph);
	}
}
