/*
DisChoco: Distributed Constraint Reasoning Solver 
Copyright (C) 2007-2013 Redouane Ezzahir & Mohamed Wahbi 

DisChoco is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DisChoco is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

To contact the authors: <http://dischoco.sourceforge.net/>
*/ 
package util.graph;

import java.lang.reflect.Constructor;
import java.util.StringTokenizer;

/**
 * 
 * This class provides a simple implementation of a old edge, applicable for
 * directed and undirected, labeled and unlabeled, weighted and unweighted graphs.
 * 
 * @author Mohamed Wahbi
 * 
 */
public class Edge<T extends Comparable<? super T>> implements Comparable<Edge<? super T>> {

	/** The source vertex of the edge */
	private Vertex<T> source;

	/** The target vertex of the edge */
	private Vertex<T> target;
	
	/** The label of the edge */
	private String label = null;
	
	/** A boolean showing whether the current edge is directed or not */
	private boolean directed = false;
	
	/** The weight of the current edge */
	private double weight;

	private boolean isBackEdge = false;
	
	/**
	 * Creates an undirected edge between two vertexes with default weight and label. Both vertexes must be non-null.
	 * 
	 * @param source source vertex
	 * @param target target vertex
	 */
	public Edge(Vertex<T> source, Vertex<T> target) {
		this(source, target, 0, false, "");
	}

	/**
	 * Creates an edge between two vertexes with the given weight. Both vertexes
	 * must be non-null.
	 * 
	 * @param source source vertex
	 * @param target target vertex
	 * @param weight edge weight
	 */
	public Edge(Vertex<T> source, Vertex<T> target, double weight) {
		this(source, target, weight, false, "");
	}

	/**
	 * Creates an edge between two vertexes with default weight and label. Both vertexes must be non-null.
	 * 
	 * @param source source vertex
	 * @param target target vertex
	 * @param directed directed edge (true) or undirected one (false)
	 */
	public Edge(Vertex<T> source, Vertex<T> target, boolean directed ) {
		this(source, target, 0, directed, "");
	}

	/**
	 * Creates an edge between two vertexes with default weight and label. Both vertexes must be non-null.
	 * 
	 * @param source source vertex
	 * @param target target vertex
	 * @param label edge label
	 */
	public Edge(Vertex<T> source, Vertex<T> target, String label) {
		this(source, target, 0, false, label);
	}

	/**
	 * Creates an edge between two vertexes with the given weight and type directed. 
	 * Both vertexes must be non-null.
	 * 
	 * @param source source vertex
	 * @param target target vertex
	 * @param weight edge weight
	 * @param directed directed edge (true) or undirected one (false)
	 */
	public Edge(Vertex<T> source, Vertex<T> target, double weight, boolean directed) {
		this(source, target, weight, directed, "");
	}

	/**
	 * Creates an edge between two vertexes with default weight and label. Both vertexes must be non-null.
	 * 
	 * @param source source vertex
	 * @param target target vertex
	 * @param label edge label
	 */
	public Edge(Vertex<T> source, Vertex<T> target, String label, boolean directed) {
		this(source, target, 0, directed, label);
	}

	/**
	 * Creates an edge between two vertexes with the given weight and type directed. 
	 * Both vertexes must be non-null.
	 * 
	 * @param source source vertex
	 * @param target target vertex
	 * @param weight edge weight
	 * @param directed directed edge (true) or undirected one (false)
	 * @param label edge label
	 */
	public Edge(Vertex<T> source, Vertex<T> target, double weight, boolean directed, String label) {
		if (source == null || target == null) {
			throw new IllegalArgumentException("vertexes must be non-null");
		}
		this.source = source;
		this.target = target;
		this.weight = weight;
		this.directed = directed;
		this.label = label;
	}

	/**
	 * Creates an edge between two vertexes with the given weight and type directed. 
	 * Both vertexes must be non-null.
	 * 
	 * @param source source vertex
	 * @param target target vertex
	 * @param label edge label
	 * @param weight edge weight
	 */
	public Edge(Vertex<T> source, Vertex<T> target, String label, double weight) {
		this(source, target, weight, false, label);
	}

	/**
	 * Returns a target vertex of this edge.
	 * 
	 * @return target vertex
	 */
	public Vertex<T> getTarget() {
		return this.target;
	}

	/**
	 * Returns a source vertex of this edge.
	 * 
	 * @return source vertex
	 */
	public Vertex<T> getSource() {
		return this.source;
	}

	/**
	 * Returns a vertex opposite to the given one.
	 * 
	 * @param vertex vertex
	 * @return opposite vertex; null if this edge is not incident to the given vertex
	 */
	public Vertex<T> getOppositeVertex(Vertex<T> vertex) {
		return vertex.equals(this.source) ? this.target : (vertex.equals(this.target) ? this.source : null);
	}

	/**
	 * Returns the weight of this edge.
	 * 
	 * @return edge weight
	 */
	public double getWeight() {
		return this.weight;
	}

	/**
	 * Sets the weight of this edge.
	 * 
	 * @param weight
	 *            edge weight
	 */
	public void setWeight(double weight) {
		this.weight = weight;
	}
	
	/**
	 * Returns the label of this edge
	 * @return the label
	 */
	public String getLabel() {
		return this.label;
	}
	
	/**
	 * Sets the label of this edge
	 * @param label the label to set
	 */
	public void setLabel(String label) {
		this.label = label;
	}
	
	/**
	 * Checks whether this edge is directed or not.
	 * 
	 * @return true if it is; false otherwise
	 */
	public boolean isDirected() {
		return this.directed;
	}
	
	/**
	 * Sets this edge to be directed (true) on undirected (false)
	 * @param directed true to make the edge directed; false otherwise
	 */
	public void setDirected(boolean directed) {
		this.directed = directed;
	}

	/** @see java.lang.Object#equals(Object o) */
	@SuppressWarnings("unchecked")
	@Override
	public boolean equals(Object o) {
		if(o == this) {
			return true;
		}
		if ((o instanceof Edge)) {
			Edge<T> e = (Edge<T>) o;
			T source = (T) e.getSource();
			T target = (T) e.getTarget();
			if(isDirected()) {
				return this.source.equals(source) && this.target.equals(target);
			}
			return ((this.source.equals(source) && this.target.equals(target)) || 
					(this.source.equals(target) && this.target.equals(source)));
		}
		return false;
	}

	/**
	 * Checks if the given vertex belongs to this edge
	 * @param vertex the vertex to check
	 * @return true if this edge contains <code>vertex</code>; false otherwise
	 */
	public boolean contains(Vertex<T> vertex) {
		return this.source.equals(vertex) || this.target.equals(vertex);
	}

	/** @see java.lang.Object#toString() */
	@Override
	public String toString() {
//		StringBuffer sb = new StringBuffer();
//		sb.append(getLabel());
//		return sb.toString();
		if(isDirected()) {
			return "(" + this.getSource() + "," + this.getTarget() + ")";
		} else {
			return "{" + this.source + "," + this.target + "}";
		}
	}
	
	public static <T extends Comparable<? super T>> Edge<T> fromString(String string, Class<T> elementClass) {
		boolean directed = false;
		if (string.startsWith("{")) {
			string = string.substring(string.indexOf("{") + 1, string.indexOf("}")); // remove { and }
		} else if (string.startsWith("(")) {
			string = string.substring(string.indexOf("(") + 1, string.indexOf(")")); // remove ( and )
			directed = true;
		}
		StringTokenizer tok = new StringTokenizer(string, ",");
		Object[] args = new Object[] {tok.nextToken()};
		T source;
		Edge<T> edge = null;
		try {
			Constructor<T> constructor = elementClass.getConstructor(new Class[] {String.class});
			source = constructor.newInstance(args);
			args = new Object[] {tok.nextToken()};
			T target  = constructor.newInstance(args);
			edge = new Edge<T>(new Vertex<T>(source), new Vertex<T>(target), directed);
		} catch (Exception e) {
			e.printStackTrace();
		}
		return edge;
	}
	
	@Override public int hashCode() {
		int hsource = this.source.getData().hashCode();
		int htarget = this.target.getData().hashCode();
		int min = Math.min(htarget, hsource);
		int max = Math.max(htarget, hsource);
		int hash = (max * 1000) + min;
		//System.out.println("--------"+min+" "+max+" hash="+hash+" "+this);
		return hash;
	}

	/**
	 * Returns a string representing this edge in the DOT format
	 * @return
	 */
	public String toDOT() {
		if(isDirected()) {
			StringBuffer sb = new StringBuffer();
			//sb.append("edge [color=blue, style=bold];\n");
			sb.append("\"" + this.getSource() + "\" -> \"" + this.getTarget() + "\"");
			sb.append(" [label=\"" + this.getLabel() + "\"]");
			return sb.toString();
		} else {
			StringBuffer sb = new StringBuffer();
			//sb.append("edge [dir=none, color=red, style=bold];\n");
			sb.append("\"" + this.getSource() + "\" -> \"" + this.getTarget() + "\"");
			sb.append(" [label=\"" + this.getLabel() + "\"]");
			return sb.toString();
		}
	}
	
	public Edge<T> getDirected(Vertex<T> source){
		Vertex<T> target = getOppositeVertex(source);
		return new Edge<T>(source, target, getWeight(), true, getLabel());
	}
	
	/**
	 * @param source the source to set
	 */
	public void setSource(Vertex<T> source) {
		this.source = source;
	}

	/**
	 * @param target the target to set
	 */
	public void setTarget(Vertex<T> target) {
		this.target = target;
	}

	public boolean isBackEdge() {
		return isBackEdge;
	}

	public void setBackEdge(boolean backEdge) {
		this.isBackEdge = backEdge;
	}

	@Override
	public int compareTo(Edge<? super T> o) {
		return this.hashCode() - o.hashCode();
	}

}
