/*
DisChoco: Distributed Constraint Reasoning Solver 
Copyright (C) 2007-2013 Redouane Ezzahir & Mohamed Wahbi 

DisChoco is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DisChoco is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

To contact the authors: <http://dischoco.sourceforge.net/>
*/ 
package util.graph;

import static simulation.parser.InstanceTokens.AGENT;
import static simulation.parser.InstanceTokens.AGENTS;
import static simulation.parser.InstanceTokens.ARITY;
import static simulation.parser.InstanceTokens.CONFLICTS;
import static simulation.parser.InstanceTokens.CONSTRAINT;
import static simulation.parser.InstanceTokens.CONSTRAINTS;
import static simulation.parser.InstanceTokens.DEFAULT_COST;
import static simulation.parser.InstanceTokens.DESCRIPTION;
import static simulation.parser.InstanceTokens.DOMAIN;
import static simulation.parser.InstanceTokens.DOMAINS;
import static simulation.parser.InstanceTokens.ID;
import static simulation.parser.InstanceTokens.INFINITY;
import static simulation.parser.InstanceTokens.MODEL;
import static simulation.parser.InstanceTokens.NAME;
import static simulation.parser.InstanceTokens.NB_AGENTS;
import static simulation.parser.InstanceTokens.NB_CONSTRAINTS;
import static simulation.parser.InstanceTokens.NB_DOMAINS;
import static simulation.parser.InstanceTokens.NB_RELATIONS;
import static simulation.parser.InstanceTokens.NB_TUPLES;
import static simulation.parser.InstanceTokens.NB_VALUES;
import static simulation.parser.InstanceTokens.NB_VARIABLES;
import static simulation.parser.InstanceTokens.PKC;
import static simulation.parser.InstanceTokens.REFERENCE;
import static simulation.parser.InstanceTokens.RELATION;
import static simulation.parser.InstanceTokens.RELATIONS;
import static simulation.parser.InstanceTokens.SCOPE;
import static simulation.parser.InstanceTokens.SEMANTICS;
import static simulation.parser.InstanceTokens.SOFT;
import static simulation.parser.InstanceTokens.TKC;
import static simulation.parser.InstanceTokens.VARIABLE;
import static simulation.parser.InstanceTokens.VARIABLES;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.List;
import java.util.Random;
import java.util.Set;
import java.util.Stack;
import java.util.StringTokenizer;
import java.util.Vector;
import org.jdom2.Element;
import simulation.parser.InstanceTokens;
import util.HashList;
import util.IntArray;
import util.tree.AbstractTree;
import util.tree.AbstractTreeNode;

/**
 * 
 * This class provides a simple implementation of an undirected graph.
 * 
 * @author Mohamed Wahbi
 * 
 */
public class UndirectedGraph<T extends Comparable<? super T>> extends AbstractGraph<T> {
	
	public UndirectedGraph(UndirectedGraph<T> graph) {
		super(graph);
	}

	public UndirectedGraph(List<Edge<T>> edges) {
		
	}
	public UndirectedGraph() {
		super();
	}

	@Override
	public boolean removeVertex(Vertex<T> vertex) {
		if (!isVertexInGraph(vertex)) {
			return false;
		}
		for (Edge<T> edge : vertex.getOutgoingEdges()) {
			Vertex<T> opposite = edge.getOppositeVertex(vertex);
			//if(!opposite.equals(vertex)) { //FIXME
				opposite.removeOutgoingEdge(edge);
				getVertexInGraph(opposite).removeOutgoingEdge(edge);
			//}
			//System.out.println("opposite="+opposite);
		}
		removeVertexFromGraph(vertex);
		return true;
	}

	@Override
	public Edge<T> addEdge(Vertex<T> source, Vertex<T> target) {
		return addEdge(source, target, 0);
	}

	@Override
	public boolean disconnect(Vertex<T> source, Vertex<T> target) {
		testVertices(source, target);
		Edge<T> edge = getConnectingEdge(source, target);
		if (edge == null) {
			return false;
		}
		source.removeOutgoingEdge(edge);
		target.removeOutgoingEdge(edge);
		return true;
	}

	/* (non-Javadoc)
	 * @see util.graph.AbstractGraph#addEdge(util.graph.Vertex, util.graph.Vertex, java.lang.String)
	 */
	@Override
	public Edge<T> addEdge(Vertex<T> source, Vertex<T> target, String label) {
		source = addVertex(source);
		target = addVertex(target);
		Edge<T> edge = new Edge<T>(source, target, label);
		source.addOutgoingEdge(edge);
		target.addOutgoingEdge(edge);
		return edge;
	}

	@Override
	public Edge<T> addEdge(Vertex<T> source, Vertex<T> target, double weight) {
		source = addVertex(source);
		target = addVertex(target);
		Edge<T> edge = new Edge<T>(source, target, weight, false);
		source.addOutgoingEdge(edge);
		target.addOutgoingEdge(edge);
		return edge;
	}

	@Override
	public Edge<T> addEdge(Vertex<T> source, Vertex<T> target, String label, double weight) {
		source = addVertex(source);
		target = addVertex(target);
		Edge<T> edge = new Edge<T>(source, target, weight, false, label);
		source.addOutgoingEdge(edge);
		target.addOutgoingEdge(edge);
		return edge;
	}

	@Override
	public void addEdge(Edge<T> edge) {
		Vertex<T> source = addVertex(edge.getSource());
		Vertex<T> target = addVertex(edge.getTarget());
		if (areConnected(edge.getSource(), edge.getTarget())) {
			throw new IllegalArgumentException("Edge already in old.");
		} else {
			edge.setSource(source);
			edge.setTarget(target);
		}
		source.addOutgoingEdge(edge);
		target.addOutgoingEdge(edge);
	}
	
	@Override
	public Edge<T> getConnectingEdge(Vertex<T> source, Vertex<T> target) {
		testVertices(source, target);
		for (Edge<T> edge : getVertexInGraph(source).getOutgoingEdges()) {
			if (edge.getOppositeVertex(source).equals(target)) {
				return edge;
			}
		}
		return null;
	}

	@Override
	public boolean isDirected() {
		return false;
	}

	@Override
	public Collection<AbstractGraph<T>> removeFromGraph(Vertex<T> vertex) {
		//this.removeVertex(vertex);
		Collection<Edge<T>> edges = vertex.getOutgoingEdges();
		UndirectedGraph<T> graph = new UndirectedGraph<T>(this);
		assert (graph.isConnected()) : "graph must be connected";
		graph.removeVertex(graph.getVertexInGraph(vertex));
		Collection<AbstractGraph<T>> subGraphs = DFS(graph);
		for(AbstractGraph<T> subGraph : subGraphs) {
			for(Edge<T> edge : edges) {
				Vertex<T> opposite = edge.getOppositeVertex(vertex);
				Vertex<T> source = new Vertex<T>(vertex.getData());
				if(subGraph.isVertexInGraph(opposite)) {
					double weight = edge.getWeight();
					String label = edge.getLabel();
					subGraph.addEdge(source, opposite, label, weight);
				}
			}
			assert (subGraph.getVertexInGraph(vertex)!=null) : "\nWe couldn't add edges="+edges
				+"\nto the subGraph="+subGraph+"\nbecause it doesn't contain "+vertex+"\nthis"+this;
		}
		return subGraphs;
	}
	
	@Override
	public Collection<AbstractGraph<T>> getUnconnectedSubGraphs() {
		return DFS(new UndirectedGraph<T>(this));
	}

	public Set<Vertex<T>> simpleDFS(Vertex<T> v) {
		Set<Vertex<T>> visitedVertexes = new HashSet<Vertex<T>>();
		Stack<Vertex<T>> dfs = new Stack<Vertex<T>>();
		dfs.push(v);
		while(!dfs.isEmpty()) {
			Vertex<T> w = dfs.pop();
			if(!visitedVertexes.contains(w)) {
				visitedVertexes.add(w);
				for(Vertex<T> u : w.getNeighbors()) {
					dfs.push(u);
				}
			}
		}
		return visitedVertexes;
	}

	public static <T extends Comparable<? super T>> Collection<AbstractGraph<T>> DFS(UndirectedGraph<T> graph) {
		Collection<AbstractGraph<T>> subGraphs = new Vector<AbstractGraph<T>>();
		List<Vertex<T>> vertexes = graph.getVertexes();
		int N = vertexes.size();
		int nbVisited = 0;
		List<Vertex<T>> unvisited = new ArrayList<Vertex<T>>(vertexes);
		while (nbVisited < N) {
			Vertex<T> v = unvisited.get(0);
			UndirectedGraph<T> subGraph = dfsFromVertex(graph, v);
			subGraphs.add(subGraph);
			List<Vertex<T>> visited = subGraph.getVertexes();
			nbVisited += visited.size();
			unvisited.removeAll(visited);
		}
		return subGraphs;
	}

	public static <T extends Comparable<? super T>> UndirectedGraph<T> dfsFromVertex(UndirectedGraph<T> graph, Vertex<T> v) {
		Stack<Vertex<T>> stack = new Stack<Vertex<T>>();
		stack.push(v);
		Set<Vertex<T>> visitedVertex = new HashSet<Vertex<T>>();
		while(!stack.isEmpty()) {
			Vertex<T> w = stack.pop();
			for(Vertex<T> u : w.getNeighbors()) {
				if(!visitedVertex.contains(u)) {
					stack.push(u);
				}
			}
			visitedVertex.add(w);
		}
		UndirectedGraph<T> dfs = new UndirectedGraph<T>();
		for(Vertex<T> w : visitedVertex) {
			dfs.addVertex(new Vertex<T>(w.getData()));
		}
		for(Edge<T> edge : graph.getEdges()) {
			Vertex<T> s = edge.getSource();
			Vertex<T> t = edge.getTarget();
			if(visitedVertex.contains(s) && visitedVertex.contains(t)) {
				Vertex<T> source = new Vertex<T>(s.getData());
				Vertex<T> taget = new Vertex<T>(t.getData());
				Edge<T> visitedEdge = new Edge<T>(source, taget, edge.getWeight(), edge.isDirected(), edge.getLabel());
				dfs.addEdge(visitedEdge);
			}
		}
		return dfs;
	}

	@Override
	public AbstractTree<T> MLSPTree() {
		UndirectedGraph<T> graph = new UndirectedGraph<T>(this);
		/* 
		//root is selected using max-degree heuristic
		Collections.sort(graph.vertexes, new MaxDegree());
		Vertex<T> root = vertexes.get(vertexes.size()/2);
		*/
		/* 
		//root is the closest node of the neighbors of the max-degree node to the mid-point of the diameter
		Collections.sort(graph.vertexes, new MaxDegree());
		Vertex<T> maxConnected = graph.vertexes.firstElement();
		ArrayList<Vertex<T>> diameter = graph.getDiameterPath(true);
		ArrayList<Vertex<T>> neighbors = new ArrayList<Vertex<T>>(maxConnected.getNeighbors());
		Vertex<T> midNode = diameter.get(diameter.size()/2);
		Collections.sort(neighbors, new Closest(graph, midNode));
		Vertex<T> closestNode = neighbors.remove(0);
		Vertex<T> root = graph.getVertexInGraph(closestNode);
		*/
//		//root is selected to be the midpoint of the diameter
//		ArrayList<Vertex<T>> diameter = graph.getDiameterPath(true);
//		Vertex<T> root = diameter.get(diameter.size()/2);
//		if((diameter.size()%2)==0) {
//			int i = diameter.size()/2;
//			Vertex<T> r1 = diameter.get(i-1);
//			Vertex<T> r2 = diameter.get(i);
//			int cmp = new MaxDegree().compare(r1, r2);
//			if(cmp < 0) {
//				root = r1;
//			}
//		}
//		System.out.println(root+" "+diameter);
		//assert (graph.isConnected()) : "\nThis graph shold connected"+"\n"+graph;
		if(graph.isConnected()) {
			ArrayList<Vertex<T>> diameter = graph.getDiameterPath(true);
			Vertex<T> root = diameter.get(diameter.size()/2);
			return new AbstractTree<T>(MLSPTree(root, graph));
		} else {
			List<AbstractTree<T>> listDFS = new ArrayList<AbstractTree<T>>();
			for(AbstractGraph<T> subGraph : graph.getUnconnectedSubGraphs()) {
				//root is selected to be the midpoint of the diameter
				ArrayList<Vertex<T>> diameter = subGraph.getDiameterPath(true);
				Vertex<T> root = diameter.get(diameter.size()/2);
				listDFS.add(new AbstractTree<T>(MLSPTree(root, subGraph)));
			}
			Collections.sort(listDFS, new Comparator<AbstractTree<T>>() {
				@Override
				public int compare(AbstractTree<T> o1, AbstractTree<T> o2) {
					return o2.getDepth() - o1.getDepth();
				}
			});
			AbstractTree<T> dfs = listDFS.remove(0);
			AbstractTreeNode<T> root = dfs.getRoot();
			while(!listDFS.isEmpty()) {
				root.addChild(listDFS.remove(0).getRoot());
			}
			return dfs;
		}
	}
	
	public HashList<T> neighbors(T vertex) {
		Vertex<T> node = getVertexInGraph(new Vertex<T>(vertex));
		return node.neighbors();
	}

	
	/**
	 * Returns a string representation of the graph
	 * 
	 * @return ConstraintGraph  
	 */
	public static <T extends Comparable<? super T>> UndirectedGraph<T> fromString(String string, Class<T> elementClass) {
		UndirectedGraph<T> graph = new UndirectedGraph<T>();
		if (string.startsWith("[")) {
			string = string.substring(string.indexOf("[") + 1, string.indexOf("]")); // remove [ and ]
		}
		StringTokenizer st = new StringTokenizer(string, EDGE_SEPARATOR);
		while (st.hasMoreTokens()) {
			String edge = st.nextToken();
			graph.addEdge(Edge.fromString(edge, elementClass));
		}
		return graph;
	}
	
	/**
	 * This methods generates a random graph having <tt>n</tt> vertices, 
	 * and a graph density of <tt>density</tt>. Thus, the number of distinct 
	 * edges is: <tt> 0.5*n*(n-1)*density </tt>. 
	 * <pre>
	 * <b><i>n</i></b> : the number of variables
	 * <b><i>density</i></b> : the density of the constraint graph (p_1)
	 * </pre>
	 * @param n : is the number of vertices 
	 * @param density : the density of the graph
	 * @return
	 */
	public static UndirectedGraph<Integer> GenerateRandomGraph(int n, int density, boolean connected) {

		assert n > 1 : "The number of nodes must be > 1";
		assert (density > 0 && density <= 100) : "The graph density must be 0<density<=1";

		UndirectedGraph<Integer> graph = null;
		do {
			graph = new UndirectedGraph<Integer>();
			/*************************************************************************/
			for(int i = 0 ; i < n ; i++) {
				Integer v = new Integer(i);
				graph.createCluster(v);
				graph.addToCluster(v, v);
				graph.addVertex(v);
			}
			// Generates vertices/variables
			for(int v = 0 ; v < n ; v++) {}
			/*************************************************************************/
			/* The maximum number of edges/constraints */
			int maxNbEdges = n * (n - 1) / 2;
			/* Create an array for each possible indexes pair. */
			IntArray indexes = new IntArray(maxNbEdges);
			/*
			 * Initialize the edgesIndexes. Create an index for each possible edge.
			 */
			for(int i = 0 , u = 0 ; u < n - 1 ; u++) {
				for(int v = u + 1 ; v < n ; v++) {
					int edge = u * n + v;
					indexes.add(edge, i++);
				}
			}
			Random rand = new Random();
			int nbEdges = (int) ((maxNbEdges * density) / 100);
			for(int e = 0 ; e < nbEdges ; e++) {
				// Choose a random edge.
				int edge = indexes.removeAt(rand.nextInt(indexes.size()));
				int u = edge / n;
				int v = edge % n;
				graph.addEdge(u, v);
			}
		} while (connected && !graph.isConnected());
		return graph;
	}
	

	/**
	 * This methods generates a random graph having <tt>n</tt> vertices, 
	 * <tt>e</tt> distinct edges, and <tt>p</tt> clusters. 
	 * <pre>
	 * <b><i>nbAgents</i></b> : the number of agents
	 * <b><i>nbVarsPerAgent</i></b> : the number of variable per agent
	 * <b><i>intraDensity</i></b> : the intra-agent density, the connectivity on the agent local variables
	 * <b><i>interDensity</i></b> : the inter-agent density, the connectivity between agent
	 * <b><i>densityOfInterDensity</i></b> : the density on the inter-agent connectivity
	 * </pre>
	 * @param nbAgents : is the number of clusters/agents
	 * @param nbVarsPerAgent : is the number of vertices/variables in each  cluster/agent
	 * @param intraDensity : the intra-agent density, the connectivity on the agent local variables
	 * @param interDensity : the inter-agent density, the connectivity between agent
	 * @param densityOfInterDensity : the density on the inter-agent connectivity
	 * @return
	 */
	public static UndirectedGraph<Integer> GenerateRandomComplexGraph(int nbAgents, int nbVarsPerAgent, int intraDensity, int interDensity,
			int densityOfInterDensity) {

		assert nbAgents > 1 : "The number of clusters/agents must be > 1";
		assert nbVarsPerAgent > 1 : "The number of vertices/variables per cluster/agent must be > 1";

		UndirectedGraph<Integer> graph = new UndirectedGraph<Integer>();

		Random rand = new Random();
		/*************************************************************************/
		for(int a = 0 ; a < nbAgents ; a++) {
			graph.createCluster(a);
			for(int i = 0 ; i < nbVarsPerAgent ; i++) {
				Integer v = new Integer(a * nbVarsPerAgent + i);
				graph.addVertex(v);
				graph.addToCluster(a, v);
			}
		}
		/*************************************************************************/
		// Create intra-agents constraints
		for(int a = 0 ; a < nbAgents ; a++) {
			IntArray indexes = new IntArray(nbVarsPerAgent * (nbVarsPerAgent - 1) / 2);
			for(int i = 0 ; i < nbVarsPerAgent ; i++) {
				indexes.add(a);
			}
			for(int index = 0 , i = 0 ; i < nbVarsPerAgent - 1 ; i++) {
				for(int j = i + 1 ; j < nbVarsPerAgent ; j++) {
					int edge = i * nbVarsPerAgent + j;
					indexes.add(edge, index++);
				}
			}
			int nbEdges = (int) ((indexes.size() * intraDensity) / 100);
			for(int e = 0 ; e < nbEdges ; e++) {
				// Choose a random edge.
				int edge = indexes.removeAt(rand.nextInt(indexes.size()));
				int i = edge / nbVarsPerAgent;
				int j = edge % nbVarsPerAgent;
				int u = a * nbVarsPerAgent + i;
				int v = a * nbVarsPerAgent + j;
				graph.addEdge(u, v);
			}
		}
		/*************************************************************************/
		// Create inter-agents constraints
		IntArray interAgentIndexes = new IntArray(nbAgents * (nbAgents - 1) / 2);
		for(int index = 0 , Ai = 0 ; Ai < nbAgents - 1 ; Ai++) {
			for(int Aj = Ai + 1 ; Aj < nbAgents ; Aj++) {
				interAgentIndexes.add(Ai * nbAgents + Aj, index++);
			}
		}
		int nbConnections = (int) ((interAgentIndexes.size() * interDensity) / 100);
		for(int c = 0 ; c < nbConnections ; c++) {
			// Choose a random connection between two agents Ai and Aj.
			int connection = interAgentIndexes.removeAt(rand.nextInt(interAgentIndexes.size()));
			int Ai = connection / nbAgents;
			int Aj = connection % nbAgents;
			// possible inter-agent constraints
			IntArray indexes = new IntArray(nbVarsPerAgent * nbVarsPerAgent);
			for(int i = 0 ; i < nbVarsPerAgent ; i++) {
				for(int j = 0 ; j < nbVarsPerAgent ; j++) {
					indexes.add(i * nbVarsPerAgent + j);
				}
			}
			int nbEdges = (int) ((indexes.size() * densityOfInterDensity) / 100);
			for(int e = 0 ; e < nbEdges ; e++) {
				// Choose a random edge.
				int edge = indexes.removeAt(rand.nextInt(indexes.size()));
				int i = edge / nbVarsPerAgent; // i is the index of the variable of agent Ai
				int u = Ai * nbVarsPerAgent + i;
				int j = edge % nbVarsPerAgent; // j is the index of the variable of agent Aj
				int v = Aj * nbVarsPerAgent + j;
				graph.addEdge(u, v);
			}
		}
		/*************************************************************************/
		return graph;
	}

	/**
	 * This methods generates <tt>nbTuples</tt> different tuples. 
	 * 
	 * @param domainSize : the domain size
	 * @param nbTuples : the number of tuples 
	 */
	public static String GenerateRandomTuples(int domainSize, int nbTuples) {
		/* Create an array of indexes for each possible tuple. */
		IntArray indexes = new IntArray();
		for(int t = 0 ; t < domainSize * domainSize ; t++) {
			indexes.add(t, t);
		}
		Random rand = new Random();
		String tuples = "";
		for(int t = 0 ; t < nbTuples ; t++) {
			// Choose a random edge.
			int tuple = indexes.removeAt(rand.nextInt(indexes.size()));
			int val1 = tuple / domainSize + 1; // values in domain are ranged 1..domainSize
			int val2 = tuple % domainSize + 1; // values in domain are ranged 1..domainSize
			tuples = tuples.concat(val1 + " " + val2 + "|");
		}
		return tuples.substring(0, tuples.length() - 1);
	}

	/**
	 * This methods generates <tt>nbTuples</tt> different tuples. 
	 * 
	 * @param domainSize : the domain size
	 * @param nbTuples : the number of tuples 
	 * @param minCost : the minimum cost of a permitted tuple
	 * @param maxCost : the maximum cost of a permitted tuple
	 */
	public static String GenerateRandomSoftTuples(int domainSize, int nbTuples, int minCost, int maxCost) {
		/* Create an array of indexes for each possible tuple. */
		StringBuffer sb = new StringBuffer();
		IntArray indexes = new IntArray();
		for(int t = 0 ; t < domainSize * domainSize ; t++) {
			indexes.add(t, t);
		}
		Random rand = new Random();
		for(int t = 0 ; t < nbTuples ; t++) {
			// Choose a random edge.
			int cost  = rand.nextInt(maxCost) + minCost;
			sb.append(cost + InstanceTokens.COST_SEPARATOR);
			int tuple = indexes.removeAt(rand.nextInt(indexes.size()));
			int value1  = tuple / domainSize + 1; // values in domain are ranged 1..domainSize
			int value2  = tuple % domainSize + 1; // values in domain are ranged 1..domainSize
			sb.append(value1 + InstanceTokens.VALUE_SEPARATOR + value2 + InstanceTokens.TUPLES_SEPARATOR);
		}
		return sb.substring(0, sb.length()-1);
	}

	/** The XDisCSP format part */

	/**
	 * This method define the set of agents/clusters in the XDisCSP format
	 * 
	 * @return org.jdom2..Element the agents element
	 */
	public Element getAgents() {
		int nbAgents = this.clusters.size();
		Element agents = new Element(AGENTS);
		agents.setAttribute(NB_AGENTS, "" + nbAgents);
		for(int id = 1 ; id <= nbAgents ; id++) {
			Element agent = new Element(AGENT);
			agent.setAttribute(NAME, "A" + id);
			agent.setAttribute(ID, "" + id);
			agent.setAttribute(DESCRIPTION, AGENT + " " + id);
			agents.addContent(agent);
		}
		return agents;
	}

	/**
	 * This method define the domains in the XDisCSP format. 
	 * One domain with <tt>domainSize</tt> values ranged from 
	 * <tt>1</tt> to <tt>domainSize</tt>.
	 * 
	 * @param domainSize the number of values in the domain
	 * @return org.jdom2..Element the domains element
	 */
	public Element getDomaines(int domainSize) {
		Element domains = new Element(DOMAINS);
		domains.setAttribute(NB_DOMAINS, "1");
		Element domain = new Element(DOMAIN);
		domain.setAttribute(NAME, "D1");
		domain.setAttribute(NB_VALUES, "" + domainSize);
		domain.setText("1.." + domainSize);
		domains.addContent(domain);
		return domains;
	}

	/**
	 * This method define the set of variables/vertices in the XDisCSP format
	 * 
	 * @return org.jdom2..Element the variables element
	 */
	public Element getVariables() {
		int nbVariables = this.getNbVertexes();
		Element variables = new Element(VARIABLES);
		variables.setAttribute(NB_VARIABLES, "" + nbVariables);
		for(int id = 0 ; id < nbVariables ; id++) {
			Element variable = new Element(VARIABLE);
			Integer agId = clusterOf.get(id) + 1;
			Integer index = clusters.get(clusterOf.get(id)).indexOf(id) + 1;
			variable.setAttribute(AGENT, "A" + agId.toString());
			variable.setAttribute(NAME, "X" + agId + "." + index);
			variable.setAttribute(ID, index.toString());
			variable.setAttribute(DOMAIN, "D1");
			variable.setAttribute(DESCRIPTION, VARIABLE + " " + agId + "." + index);
			variables.addContent(variable);
		}
		return variables;
	}

	/**
	 * This method defines the set of constraints/edges in the XDisCSP format
	 * 
	 * @return org.jdom2..Element the constraints element
	 */
	public Element getConstraints() {
		int nbConstraint = this.getNbEdges();
		Element constraints = new Element(CONSTRAINTS);
		constraints.setAttribute(NB_CONSTRAINTS, "" + nbConstraint);
		HashList<Edge<T>> edges = this.getEdges();
		for(int c = 0 ; c < nbConstraint ; c++) {
			Element constraint = new Element(CONSTRAINT);
			constraint.setAttribute(NAME, "C" + c);
			constraint.setAttribute(MODEL, TKC); // Totally Known Constraints
			constraint.setAttribute(ARITY, Integer.toString(2));
			Edge<T> e = edges.get(c);
			T var1 = e.getSource().getData();
			Integer agId1 = clusterOf.get(var1) + 1;
			Integer index1 = clusters.get(clusterOf.get(var1)).indexOf(var1) + 1;
			T var2 = e.getTarget().getData();
			Integer agId2 = clusterOf.get(var2) + 1;
			Integer index2 = clusters.get(clusterOf.get(var2)).indexOf(var2) + 1;
			constraint.setAttribute(SCOPE, "X" + agId1 + "." + index1 + " X" + agId2 + "." + index2);
			constraint.setAttribute(REFERENCE, "R" + c);
			constraints.addContent(constraint);
		}
		return constraints;
	}

	/**
	 * This method defines the set of constraints/edges in the XDisCSP format
	 * 
	 * @return org.jdom2..Element the constraints element
	 */
	public Element getConstraints(String relation) {
		int nbConstraint = this.getNbEdges();
		Element constraints = new Element(CONSTRAINTS);
		constraints.setAttribute(NB_CONSTRAINTS, "" + nbConstraint);
		HashList<Edge<T>> edges = this.getEdges();
		for(int c = 0 ; c < nbConstraint ; c++) {
			Element constraint = new Element(CONSTRAINT);
			constraint.setAttribute(NAME, "C" + c);
			constraint.setAttribute(MODEL, TKC); // Totally Known Constraints
			constraint.setAttribute(ARITY, Integer.toString(2));
			Edge<T> e = edges.get(c);
			T var1 = e.getSource().getData();
			Integer agId1 = clusterOf.get(var1) + 1;
			Integer index1 = clusters.get(clusterOf.get(var1)).indexOf(var1) + 1;
			T var2 = e.getTarget().getData();
			Integer agId2 = clusterOf.get(var2) + 1;
			Integer index2 = clusters.get(clusterOf.get(var2)).indexOf(var2) + 1;
			constraint.setAttribute(SCOPE, "X" + agId1 + "." + index1 + " X" + agId2 + "." + index2);
			constraint.setAttribute(REFERENCE, relation);
			constraints.addContent(constraint);
		}
		return constraints;
	}
	
	/**
	 * This method defines the set of constraints/edges in the XDisCSP format 
	 * when constraints are partially known. Thus for each edge there are two 
	 * constraints. 
	 * 
	 * @param constraintModel : TKC or PKC
	 * @return org.jdom2..Element the constraints element
	 */
	public Element getPartiallyConstrains() {
		int nbConstraint = this.getNbEdges();
		Element constraints = new Element(CONSTRAINTS);
		constraints.setAttribute(NB_CONSTRAINTS, "" + nbConstraint);
		HashList<Edge<T>> edges = this.getEdges();
		for(int c = 0 ; c < nbConstraint ; c++) {
			Edge<T> e = edges.get(c);
			int agId1 = clusterOf.get(e.getSource().getData()) + 1;
			Integer index1 = clusters.get(clusterOf.get(e.getSource().getData())).indexOf(e.getSource().getData()) + 1;
			String label1 = "X" + agId1 + "." + index1;
			Integer agId2 = clusterOf.get(e.getTarget().getData()) + 1;
			Integer index2 = clusters.get(clusterOf.get(e.getTarget().getData())).indexOf(e.getTarget().getData()) + 1;
			String label2 = "X" + agId2 + "." + index2;

			Element constraint1 = new Element(CONSTRAINT);
			constraint1.setAttribute(NAME, "C" + (2 * c));
			constraint1.setAttribute(MODEL, PKC); // Partially Known Constraints
			constraint1.setAttribute(ARITY, "2");
			constraint1.setAttribute(InstanceTokens.OWNER_ID, label1);
			constraint1.setAttribute(SCOPE, label1 + " " + label2);
			constraint1.setAttribute(REFERENCE, "R" + (2 * c));
			constraints.addContent(constraint1);

			Element constraint2 = new Element(CONSTRAINT);
			constraint2.setAttribute(NAME, "C" + ((2 * c) + 1));
			constraint2.setAttribute(MODEL, PKC); // Partially Known Constraints
			constraint2.setAttribute(ARITY, "2");
			constraint1.setAttribute(InstanceTokens.OWNER_ID, label2);
			constraint2.setAttribute(SCOPE, label2 + " " + label1);
			constraint2.setAttribute(REFERENCE, "R" + ((2 * c) + 1));
			constraints.addContent(constraint2);
		}
		return constraints;
	}

	/**
	 * This method defines the set of relations in the XDisCSP format. 
	 * Thus, for each constraint we generate a relation. 
	 * 
	 * @param domainSize : the domain size of each variable
	 * @param tightness : the tightness of the constraint
	 * @return org.jdom2..Element the relations element
	 */
	public Element getHardRelations(int domainSize, int tightness) {
		int nbConstraint = this.getNbEdges();
		int nbTuples = (int) ((domainSize * domainSize * tightness) / 100);
		Element relations = new Element(RELATIONS);
		relations.setAttribute(NB_RELATIONS, "" + nbConstraint);
		/* For each constraint, select T illegal value pairs. */
		for(int c = 0 ; c < nbConstraint ; c++) {
			Element relation = new Element(RELATION);
			relation.setAttribute(NAME, "R" + c);
			relation.setAttribute(ARITY, "2");
			relation.setAttribute(NB_TUPLES, "" + nbTuples);
			relation.setAttribute(SEMANTICS, CONFLICTS);
			String tuples = GenerateRandomTuples(domainSize, nbTuples);
			relation.setText(tuples);
			relations.addContent(relation);
		}
		return relations;
	}

	/**
	 * This method defines the set of relations in the XDisCSP format. 
	 * Thus, for each (2) constraint(s) we generate a (2) relation(s). 
	 * 
	 * @param domainSize : the domain size of each variable
	 * @param tightness : the tightness of the constraint
	 * @return org.jdom2..Element the relations element
	 */
	public Element getPartiallyHardRelations(int domainSize, int tightness) {
		int nbConstraint = this.getNbEdges();
		int nbTuples = (int) ((domainSize * domainSize * tightness) / 100);
		Element relations = new Element(RELATIONS);
		relations.setAttribute(NB_RELATIONS, "" + nbConstraint);
		/* For each constraint, select T illegal value pairs. */
		int T1 = nbTuples / 2;
		int T2 = nbTuples - T1;
		for(int c = 0 ; c < nbConstraint ; c++) {
			IntArray indexes = new IntArray();
			for(int t = 0 ; t < domainSize * domainSize ; t++) {
				indexes.add(t, t);
			}
			Random rand = new Random();
			Element relation1 = new Element(RELATION);
			relation1.setAttribute(NAME, "R" + ((2 * c)));
			relation1.setAttribute(ARITY, "2");
			relation1.setAttribute(NB_TUPLES, "" + T1);
			relation1.setAttribute(SEMANTICS, CONFLICTS);
			relations.addContent(relation1);
			Element relation2 = new Element(RELATION);
			relation2.setAttribute(NAME, "R" + ((2 * c) + 1));
			relation2.setAttribute(ARITY, "2");
			relation2.setAttribute(NB_TUPLES, "" + T2);
			relation2.setAttribute(SEMANTICS, CONFLICTS);

			String tuples1 = "";
			for(int t = 0 ; t < T1 ; t++) {
				// Choose a random edge.
				int tuple = indexes.removeAt(rand.nextInt(indexes.size()));
				int val1 = tuple / domainSize + 1; // values in domain are ranged 1..domainSize
				int val2 = tuple % domainSize + 1; // values in domain are ranged 1..domainSize
				tuples1 = tuples1.concat(val1 + " " + val2 + "|");
			}
			tuples1 = tuples1.substring(0, tuples1.length() - 1);
			relation1.setText(tuples1);
			String tuples2 = "";
			for(int t = 0 ; t < T2 ; t++) {
				// Choose a random edge.
				int tuple = indexes.removeAt(rand.nextInt(indexes.size()));
				int val1 = tuple / domainSize + 1; // values in domain are ranged 1..domainSize
				int val2 = tuple % domainSize + 1; // values in domain are ranged 1..domainSize
				tuples2 = tuples2.concat(val1 + " " + val2 + "|");
			}
			tuples2 = tuples2.substring(0, tuples2.length() - 1);
			relation2.setText(tuples2);
			relations.addContent(relation2);
		}
		return relations;
	}

	// TODO
	// TODO

	/**
	 * This method defines the set of relations in the XDisCSP format. 
	 * Thus, for each constraint we generate a relation. 
	 * 
	 * @param domainSize : the domain size of each variable
	 * @param tightness : the tightness of the constraint
	 * @return org.jdom2..Element the relations element
	 */
	public Element getSoftRelations(boolean maxDisCSP, int domainSize, int tightness, int minCost, int maxCost) {
		int nbConstraint = this.getNbEdges();
		int nbTuples = (int) ((domainSize * domainSize * (100 - tightness)) / 100);
		Element relations = new Element(RELATIONS);
		relations.setAttribute(NB_RELATIONS, "" + nbConstraint);
		/* For each constraint, select T illegal value pairs. */
		for(int c = 0 ; c < nbConstraint ; c++) {
			Element relation = new Element(RELATION);
			relation.setAttribute(NAME, "R" + c);
			relation.setAttribute(ARITY, "2");
			relation.setAttribute(NB_TUPLES, "" + nbTuples);
			relation.setAttribute(SEMANTICS, SOFT);
			relation.setAttribute(DEFAULT_COST, maxDisCSP? Integer.toString(1) : INFINITY);
			String tuples = "";
			if(maxDisCSP) {
				String cost = "0" + InstanceTokens.COST_SEPARATOR;
				tuples = cost + GenerateRandomTuples(domainSize, nbTuples);
			} else {
				tuples = GenerateRandomSoftTuples(domainSize, nbTuples, minCost, maxCost);
			}
			relation.setText(tuples);
			relations.addContent(relation);
		}
		return relations;
	}

	/**
	 * This method defines the set of relations in the XDisCSP format. 
	 * Thus, for each (2) constraint(s) we generate a (2) relation(s). 
	 * 
	 * @param domainSize : the domain size of each variable
	 * @param tightness : the tightness of the constraint
	 * @return org.jdom2..Element the relations element
	 */
	public Element getPartiallySoftRelations(boolean maxDisCSP, int domainSize, int tightness, int minCost, int maxCost) {
		int nbConstraint = this.getNbEdges();
		int nbTuples = (int) ((domainSize * domainSize * tightness) / 100);
		Element relations = new Element(RELATIONS);
		relations.setAttribute(NB_RELATIONS, "" + nbConstraint);
		/* For each constraint, select T illegal value pairs. */
		int T1 = nbTuples / 2;
		int T2 = nbTuples - T1;
		for(int c = 0 ; c < nbConstraint ; c++) {
			IntArray indexes = new IntArray();
			for(int t = 0 ; t < domainSize * domainSize ; t++) {
				indexes.add(t, t);
			}
			Random rand = new Random();
			Element relation1 = new Element(RELATION);
			relation1.setAttribute(NAME, "R" + ((2 * c)));
			relation1.setAttribute(ARITY, "2");
			relation1.setAttribute(NB_TUPLES, "" + T1);
			relation1.setAttribute(SEMANTICS, SOFT);
			relation1.setAttribute(DEFAULT_COST, maxDisCSP? Integer.toString(1) : INFINITY);
			relations.addContent(relation1);
			Element relation2 = new Element(RELATION);
			relation2.setAttribute(NAME, "R" + ((2 * c) + 1));
			relation2.setAttribute(ARITY, "2");
			relation2.setAttribute(NB_TUPLES, "" + T2);
			relation2.setAttribute(SEMANTICS, SOFT);
			relation2.setAttribute(DEFAULT_COST, maxDisCSP? Integer.toString(1) : INFINITY);

			String tuples1 = "";
			for(int t = 0 ; t < T1 ; t++) {
				// Choose a random edge.
				int tuple = indexes.removeAt(rand.nextInt(indexes.size()));
				int val1 = tuple / domainSize + 1; // values in domain are ranged 1..domainSize
				int val2 = tuple % domainSize + 1; // values in domain are ranged 1..domainSize
				tuples1 = tuples1.concat(val1 + " " + val2 + "|");
			}
			tuples1 = tuples1.substring(0, tuples1.length() - 1);
			relation1.setText(tuples1);
			String tuples2 = "";
			for(int t = 0 ; t < T2 ; t++) {
				// Choose a random edge.
				int tuple = indexes.removeAt(rand.nextInt(indexes.size()));
				int val1 = tuple / domainSize + 1; // values in domain are ranged 1..domainSize
				int val2 = tuple % domainSize + 1; // values in domain are ranged 1..domainSize
				tuples2 = tuples2.concat(val1 + " " + val2 + "|");
			}
			tuples2 = tuples2.substring(0, tuples2.length() - 1);
			relation2.setText(tuples2);
			relations.addContent(relation2);
		}
		return relations;
	}
	
}
