/*
DisChoco: Distributed Constraint Reasoning Solver 
Copyright (C) 2007-2013 Redouane Ezzahir & Mohamed Wahbi 

DisChoco is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DisChoco is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

To contact the authors: <http://dischoco.sourceforge.net/>
*/ 
package util.graph;

import java.io.Serializable;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import util.HashList;

/**
 * 
 * This class provides a simple implementation of a old node, applicable for
 * both directed and undirected graphs.
 * 
 * @author Mohamed Wahbi
 * 
 */
//public class Vertex<T extends Comparable<T>> implements Comparable<Vertex<T>> {
@SuppressWarnings("serial")
public class Vertex<T extends Comparable<? super T>> extends Object implements Serializable, Comparable<Vertex<? extends T>> {

	/** The set of outgoing edges */
	private final Set<Edge<T>> outgoingEdges;

	/** The set of incoming edges */
	private final Set<Edge<T>> incomingEdges;
	
	/** The content of the vertex */
	private T data = null;

	private VertexState state = VertexState.White;

	/**
	 * Creates an unconnected vertex.
	 * @param data the content of the vertex
	 */
	public Vertex(T data) {
		this.setData(data);
		this.outgoingEdges = new HashSet<Edge<T>>();
		this.incomingEdges = new HashSet<Edge<T>>();
	}

	/**
	 * Returns the content of the vertex
	 * @return the data
	 */
	public T getData() {
		return data;
	}

	/**
	 * Sets the content of the vertex to the given data
	 * @param data the data to set
	 */
	public void setData(T data) {
		this.data = data;
	}

	/**
	 * Returns a set of leaving edges for directed old and incident edges for undirected old.
	 * 
	 * @return an unmodifiable set of outgoing edges.
	 */
	public Collection<Edge<T>> getOutgoingEdges() {
		return Collections.unmodifiableCollection(this.outgoingEdges);
	}

	/**
	 * Returns a set of entering edges for directed old and an empty set for undirected old.
	 * 
	 * @return an unmodifiable set of outgoing edges.
	 */
	public Collection<Edge<T>> getIncomingEdges() {
		return Collections.unmodifiableCollection(this.incomingEdges);
	}
	
	/**
	 * Removes all edges connecting the current vertex to other vertexes
	 */
	public void removeEdges() {
		this.outgoingEdges.clear();
		this.incomingEdges.clear();
	}

	/**
	 * Adds an edge to the set of outgoing edges
	 * @param edge the edge to add
	 */
	public void addOutgoingEdge(Edge<T> edge) {
		this.outgoingEdges.add(edge);
	}
	
	/**
	 * Removes an edge from the set of outgoing edges
	 * @param edge the edge to remove
	 */
	public void removeOutgoingEdge(Edge<T> edge) {
		this.outgoingEdges.remove(edge);
	}
	
	/**
	 * Adds an edge to the set of incoming edges
	 * @param edge the edge to add
	 */
	public void addIncomingEdge(Edge<T> edge) {
		this.incomingEdges.add(edge);
	}
	
	/**
	 * Removes an edge from the set of incoming edges
	 * @param edge the edge to remove
	 */
	public void removeIncomingEdge(Edge<T> edge) {
		this.incomingEdges.remove(edge);
	}

	/** @see java.lang.Object#equals(Object o) */
	@SuppressWarnings("unchecked")
	public boolean equals(Object o) {
		Vertex<T> vertex = (Vertex<T>) o;
		return this.data.equals(vertex.data);
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return this.data.toString();
	}

	/* (non-Javadoc)
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
	@Override
	public int compareTo(Vertex<? extends T> vertex) {
		return this.getData().compareTo(vertex.getData());
//		int cmp = (vertex.getDegree() - this.getDegree());
//		if(cmp == 0) {
//			return this.data.compareTo(vertex.getData());
//		}
//		return cmp;
	}

	public int getDegree() {
		return this.incomingEdges.size() + this.outgoingEdges.size();
	}

	public HashList<Vertex<T>> getNeighbors() {
		HashList<Vertex<T>> neighbors = new HashList<Vertex<T>>();
		for (Edge<T> edge : getOutgoingEdges()) {
			Vertex<T> opposite = edge.getOppositeVertex(this);
			neighbors.add(opposite);
		}
//		for (Edge<T> edge : getIncomingEdges()) {
//			Vertex<T> opposite = edge.getOppositeVertex(this);
//			neighbors.add(opposite);
//		}
		return neighbors;
	}

	public HashList<T> getNeighborsData() {
		HashList<T> neighbors = new HashList<T>();
		for (Edge<T> edge : getOutgoingEdges()) {
			Vertex<T> opposite = edge.getOppositeVertex(this);
			neighbors.add(opposite.getData());
		}
		for (Edge<T> edge : getIncomingEdges()) {
			Vertex<T> opposite = edge.getOppositeVertex(this);
			neighbors.add(opposite.getData());
		}
		return neighbors;
	}

	public HashList<T> neighbors() {
		HashList<T> neighbors = new HashList<T>();
		for (Edge<T> edge : getOutgoingEdges()) {
			Vertex<T> opposite = edge.getOppositeVertex(this);
			neighbors.add(opposite.getData());
		}
		for (Edge<T> edge : getIncomingEdges()) {
			Vertex<T> opposite = edge.getOppositeVertex(this);
			neighbors.add(opposite.getData());
		}
		return neighbors;
	}
	
	public int getPathDistance(Vertex<T> vertex) {
		HashList<Vertex<T>> visited = new HashList<Vertex<T>>();
		visited.add(this);
		return DFS(vertex, visited).size()-1;
//		double distance = Double.MAX_VALUE;
//		for(Edge<T> edge : this.getOutgoingEdges()) {
//			if(edge.contains(vertex)){
//				distance = 1;
//			} else {
//				if(edge.getSource().equals(this)) {
//					distance = Math.min(distance, edge.getTarget().getPathDistance(vertex));
//				}
//			}
//		}
//		return distance+1;
	}

	public HashList<Vertex<T>> DFS(Vertex<T> vertex, HashList<Vertex<T>> visited){
		for(Edge<T> edge : this.getOutgoingEdges()) {
			Vertex<T> opposite = edge.getOppositeVertex(this);
			visited.add(opposite);
			if(opposite.equals(vertex)) {
				return visited;
			} else if(!visited.contains(opposite)) {
				return opposite.DFS(vertex, visited);
			}
		}
		return visited;
	}

	/**
	 * @return the state
	 */
	public VertexState getState() {
		return state;
	}

	/**
	 * @param state the state to set
	 */
	public void setState(VertexState state) {
		this.state = state;
	}

	public int getNbUnvisitedNeighbors() {
		int nbUnvisited = 0;
		for(Vertex<T> v : getNeighbors()) {
			if(v.getState()==VertexState.White) {
				nbUnvisited++;
			}
		}
		return nbUnvisited;
	}

	public int getNbVisitedNeighbors() {
		int nbVisited = 0;
		for(Vertex<T> v : getNeighbors()) {
			if(v.getState()!=VertexState.White) {
				nbVisited++;
			}
		}
		return nbVisited;
	}
}