/*
 * DisChoco: Distributed Constraint Reasoning Solver
 * Copyright (C) 2007-2013 Redouane Ezzahir & Mohamed Wahbi
 * DisChoco is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * DisChoco is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 * To contact the authors: <http://dischoco.sourceforge.net/>
 */
package util.tree;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;
import kernel.VarIdentifier;
import util.HashList;
import util.graph.AbstractGraph;
import util.graph.Edge;
import util.graph.UndirectedGraph;
import util.graph.Vertex;

public class AbstractTree<T extends Comparable<? super T>> {

	private AbstractTreeNode<T> root;

	public AbstractTree() {}

	public AbstractTree(AbstractTreeNode<T> root) {
		this.root = root;
	}

	public AbstractTreeNode<T> getRoot() {
		return this.root;
	}

	public void setRoot(AbstractTreeNode<T> root) {
		this.root = root;
	}

	public boolean isEmpty() {
		return this.root == null;
	}

	public Vertex<T> getRootNode() {
		return this.root.getNode();
	}

	/**
	 * This method returns the size of the tree, i.e., the number of node in the tree. 
	 * @return <b>int</b> : the size of the tree
	 */
	public int size() {
		return this.root.getNumberOfDescendants() + 1;
	}

	/**
	 * This method computes the depth of this tree. 
	 * The depth of tree is the longest path from root node to leaf node. 
	 * @return int : the depth of tree. 
	 */
	public int getDepth() {
		return this.root.getDepth();
	}

	public int getNumberOfLeaves() {
		return this.root.getNumberOfLeaves();
	}

	/**
	 * This method returns the list of the ancestors of the given node <tt>node</tt>. 
	 * 
	 * @param node : the identifier of the node. 
	 * @return List<T> : the list of the ancestors of the given node.
	 */
	public HashList<Vertex<T>> getAncestors(Vertex<T> node) {
		HashList<Vertex<T>> ancestors = new HashList<Vertex<T>>();
		Vertex<T> parent = this.getParent(node);
		while (parent != null) {
			ancestors.add(parent);
			parent = this.getParent(parent);
		}
		return ancestors;
	}

	/**
	 * This method returns the list of the ancestors of the given node <tt>node</tt>. 
	 * 
	 * @param node : the identifier of the node. 
	 * @return List<T> : the list of the ancestors of the given node.
	 */
	public HashList<Vertex<T>> getAncestors(T node) {
		return this.getAncestors(new Vertex<T>(node));
	}

	/**
	 * This method returns the list of the ancestors of the given node <tt>node</tt>. 
	 * 
	 * @param node : the identifier of the node. 
	 * @return List<T> : the list of the ancestors of the given node.
	 */
	public HashList<T> getAncestorsData(Vertex<T> node) {
		HashList<T> ancestors = new HashList<T>();
		Vertex<T> parent = this.getParent(node);
		while (parent != null) {
			ancestors.addFirst(parent.getData());
			parent = this.getParent(parent);
		}
		return ancestors;
	}

	/**
	 * This method returns the list of the ancestors of the given node <tt>node</tt>. 
	 * 
	 * @param node : the identifier of the node. 
	 * @return List<T> : the list of the ancestors of the given node.
	 */
	public HashList<T> getAncestorsData(T node) {
		Vertex<T> vertex = new Vertex<T>(node);
		return this.getAncestorsData(vertex);
	}

	/**
	 * This method traverses the tree until the given node is found. It returns a pointer 
	 * to the subtree whose root is the parent of the given node. 
	 * 
	 * @param node : the node to locate in the tree
	 * @return AbstractTreeNode<T, E> : the tree rooted at the node <tt>node</tt> 
	 * if the it exists, otherwise, null.
	 */
	public Vertex<T> getParent(T node) {
		return this.getParent(new Vertex<T>(node));
	}

	/**
	 * This method traverses the tree until the given node is found. It returns a pointer 
	 * to the subtree whose root is the parent of the given node. 
	 * 
	 * @param node : the node to locate in the tree
	 * @return AbstractTreeNode<T, E> : the tree rooted at the node <tt>node</tt> 
	 * if the it exists, otherwise, null.
	 */
	public Vertex<T> getParent(Vertex<T> node) {
		if (this.getRoot().getNode().equals(node)) {
			return null;
		}
		return this.root.getParent(node);
	}

	/**
	 * This method traverses the tree until the given node is found. It returns a pointer 
	 * to the subtree whose root is the parent of the given node. 
	 * 
	 * @param node : the node to locate in the tree
	 * @return AbstractTreeNode<T, E> : the tree rooted at the node <tt>node</tt> 
	 * if the it exists, otherwise, null.
	 */
	public T getParentData(T node) {
		return this.getParentData(new Vertex<T>(node));
	}

	/**
	 * This method traverses the tree until the given node is found. It returns a pointer 
	 * to the subtree whose root is the parent of the given node. 
	 * 
	 * @param node : the node to locate in the tree
	 * @return AbstractTreeNode<T, E> : the tree rooted at the node <tt>node</tt> 
	 * if the it exists, otherwise, null.
	 */
	public T getParentData(Vertex<T> node) {
		if (this.getRoot().getNode().equals(node)) {
			return null;
		}
		return this.root.getParent(node).getData();
//		Vertex<T> parent = this.root.getParent(node);
//		return (parent == null)? null : parent.getData();
	}

	/**
	 * This method returns a pointer to the subtree rooted at the node with the given node <code>node</node>.
	 * 
	 * @param v the identifier of the root
	 * @return AbstractTreeNode<T, E> : the sub-tree rooted at the node with the given root <tt>v</tt> 
	 */
	public AbstractTreeNode<T> getSubTree(T node) {
		return this.getSubTree(new Vertex<T>(node));
	}

	/**
	 * This method returns a pointer to the subtree rooted at the node with the given node <code>node</node>.
	 * 
	 * @param v the identifier of the root
	 * @return AbstractTreeNode<T, E> : the sub-tree rooted at the node with the given root <tt>v</tt> 
	 */
	public AbstractTreeNode<T> getSubTree(Vertex<T> node) {
		if (this.isEmpty()) {
			throw new UnsupportedOperationException("This is not a subtree of an empty tree");
		}
		return this.root.getSubTree(node);
	}

	/**
	 * This method adds a subtree <code>subtree</code> to the root node.
	 * @param subtree : the subtree to add 
	 * @param edge : the edge linking the root to root of the subtree. 
	 */
	public void addSubTree(AbstractTreeNode<T> subtree, Edge<T> edge) {
		this.root.addSubTree(subtree, edge);
	}

	/**
	 * This method adds a branch to the current node.
	 * @param subtree : the subtree to add
	 */
	public void addSubTree(AbstractTreeNode<T> subtree) {
		this.addSubTree(subtree, new Edge<T>(this.getRootNode(), subtree.getNode(), true));
	}

	/**
	 * This method checks if a given node <code>node</code> belongs to this tree. 
	 * @return boolean : <i>true</i>, if a such node <code>node</code> is in this tree; <i>false</i>, otherwise. 
	 */
	public boolean isInTree(Vertex<T> node) {
		return this.root.contains(node);
	}

	/**
	 * This method checks if the first given node is <code>child</code> is a descendant of the 
	 * second given node <code>ancestor</code>. 
	 * @param child : a node
	 * @param ancestor : a node 
	 * @return boolean : <code>true</code>, if <code>child</code> is a descendant of <tt>ancestor</tt>; 
	 * <code>false</code>, otherwise. 
	 */
	public boolean isDescendent(Vertex<T> child, Vertex<T> ancestor) {
		return this.getSubTree(ancestor).contains(child);
	}

	/**
	 * This method returns a string representation of the tree
	 * @return String representng the current tree
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append("[");
		if (!this.isEmpty()) {
			AbstractTreeNode<T> root = getRoot();
			sb.append(root.toString());
		}
		sb.append("]");
		return sb.toString();
	}

	/**
	 * This method returns a string representation of the tree
	 * @return String representng the current tree
	 */
	public String pretty() {
		StringBuffer sb = new StringBuffer();
		sb.append("[");
		if (!this.isEmpty()) {
			AbstractTreeNode<T> root = getRoot();
			sb.append(root.toStringVerbose("\t"));
		}
		sb.append("]");
		return sb.toString();
	}

	public static <T extends Comparable<? super T>> AbstractTree<T> fromString(String string, Class<T> elementClass) {
		int beginIndex = string.indexOf("[");
		int endIndex = string.lastIndexOf("]");
		if (beginIndex < 0 || endIndex < 0 || beginIndex + 1 >= endIndex) {
			return new AbstractTree<T>();
		}
		AbstractTreeNode<T> root = AbstractTreeNode.fromString(string.substring(beginIndex + 1, endIndex), elementClass);
		return new AbstractTree<T>(root);
	}

	/** 
	 * This method represents the graph in the DOT format. 
	 * 
	 * @return a String representation of the graph in DOT format
	 */
	public String toDOTString() {
		StringBuilder sb = new StringBuilder("digraph Tree {\n");
		sb.append("\tbgcolor=\"white\"\n");
		sb.append("\tlabel=\"A Constraint Tree Generated by DisChoco\"\n");
		sb.append("\tfontcolor=\"red\"\n");
		sb.append("\tnode [fillcolor=\"white\" style=\"filled\" color=\"black\" fontcolor=\"blue\"]\n");
		sb.append("\n");
		for(Vertex<T> vertex : this.root.getAllDescendants()) {
			sb.append("\t\"" + vertex + "\"; ");
		}
		sb.append("\n");
		sb.append("\tsubgraph Rel1 {\n");
		sb.append("\t\tedge [color=red];\n");
		Collection<Edge<T>> allEdges = this.root.getAllEdges();
		for(Edge<T> edge : allEdges) {
			if (edge.isDirected()) {
				sb.append("\t\t" + edge.toDOT() + ";\n");// [color=blue,style=bold];\n");
			}
		}
		sb.append("\t}\n");
		sb.append("\tsubgraph Rel2 {\n");
		sb.append("\t\tedge [dir=none, color=black];\n");
		for(Edge<T> edge : allEdges) {
			if (!edge.isDirected()) {
				sb.append("\t\t" + edge.toDOT() + ";\n");// [color=blue,style=bold];\n");
			}
		}
		sb.append("\t}\n");
		sb.append("}\n");
		return sb.toString();
	}

	public static <T extends Comparable<? super T>> String toDotFile(AbstractTree<T> tree, String fileName) {
		String dotFileName = new File(fileName).getAbsolutePath();
		if (fileName.contains(".")) {
			dotFileName = dotFileName.substring(0, dotFileName.length() - 4) + ".dot";
		}
		try {
			PrintWriter writer = new PrintWriter(new File(fileName));
			writer.println(tree.toDOTString());
			writer.close();
		} catch (FileNotFoundException e) {
			System.err.println("Could not create DOT output file with name " + fileName);
			e.printStackTrace();
		}
		return dotFileName;
	}

	@SuppressWarnings("unchecked")
	public static <T extends Comparable<? super T>, TreeNode extends AbstractTreeNode<T>> TreeNode simpleDFS(AbstractGraph<T> graph,
			Vertex<T> vertex, Set<Vertex<T>> visited) {
		visited.add(vertex);
		TreeNode dfs = (TreeNode) new AbstractTreeNode<T>(vertex);
		Vertex<T> source = dfs.getNode();
		// Collection<Edge<T>> edges = graph.getEdges(vertex);
		// for (Edge<T> edge : edges) {
		HashList<Vertex<T>> neighbors = vertex.getNeighbors();
		for(Vertex<T> neighbor : neighbors) {
			Edge<T> edge = graph.getConnectingEdge(vertex, neighbor);
			Vertex<T> opposite = edge.getOppositeVertex(vertex);
			if (!visited.contains(opposite)) {
				dfs.getChildren().add(simpleDFS(graph, opposite, visited));
				Edge<T> dircetedEdge = edge.getDirected(source);
				source.addOutgoingEdge(dircetedEdge);
				// dfs.addChild(simpleDFS(graph, opposite, visited), dircetedEdge);
			} else if (dfs.contains(opposite)) {
				edge.setBackEdge(true);
				source.addOutgoingEdge(edge);
			}
		}
		// for (Edge<T> edge : edges) {
		// Vertex<T> opposite = edge.getOppositeVertex(vertex);
		// Edge<T> dircetedEdge = edge.getDirected(source);
		// if (dfs.contains(opposite)) {
		// source.addOutgoingEdge(dircetedEdge);
		// } else {
		// opposite.addIncomingEdge(dircetedEdge);
		// }
		// }
		return dfs;
	}

	public static <T extends Comparable<? super T>> AbstractTree<T> DFS(AbstractGraph<T> graph) {
		Set<Vertex<T>> visited = new HashSet<Vertex<T>>();
		SortedSet<Vertex<T>> vertexes = new TreeSet<Vertex<T>>(graph.getVertexes());
		Vertex<T> root = vertexes.first();
		return new AbstractTree<T>(simpleDFS(graph, root, visited));
	}

	// private void updateEdges() {
	// Vertex<T> source = this.getRootNode();
	// for(Edge<T> edge : source.getOutgoingEdges()) {
	// //.setDirected(true);
	// Vertex<T> target = edge.getTarget();
	// Edge<T> directedEdge = edge.getDirected(source);
	// target.removeOutgoingEdge(edge);
	// target.addIncomingEdge(directedEdge);
	// source.removeOutgoingEdge(edge);
	// source.addOutgoingEdge(directedEdge);
	// }
	// }

	/**
	 * This method returns a set vertexes of this old.
	 * 
	 * @return an unmodifiable set of vertexes.
	 */
	public Collection<Vertex<T>> getVertexes() {
		return this.root.getSubtreeVertexes();
	}

	public boolean areConnected(Vertex<T> u, Vertex<T> v) {
		return (this.isDescendent(u, v) || this.isDescendent(v, u));
	}

	public Collection<AbstractTreeNode<T>> getSubtrees() {
		Collection<AbstractTreeNode<T>> subtrees = new ArrayList<AbstractTreeNode<T>>();
		subtrees.add(root);
		subtrees.addAll(root.getSubtrees());
		return subtrees;
	}

	/**
	 * @param varId
	 * @return
	 */
	public HashList<Vertex<T>> getDescendants(T node) {
		return this.getDescendants(new Vertex<T>(node));
	}

	/**
	 * @param varId
	 * @return
	 */
	public HashList<Vertex<T>> getDescendants(Vertex<T> vertex) {
		return this.root.getSubTree(vertex).getAllDescendants();
	}

	/**
	 * @param varId
	 * @return
	 */
	public HashList<T> getDescendantsData(T node) {
		return this.getDescendantsData(new Vertex<T>(node));
	}

	/**
	 * @param varId
	 * @return
	 */
	public HashList<T> getDescendantsData(Vertex<T> vertex) {
		AbstractTreeNode<T> subtree = this.root.getSubTree(vertex);
		if (subtree == null) {
			return new HashList<T>();
		} else {
			return subtree.getAllDescendantsData();
		}
	}

	/**
	 * @param varId
	 * @return
	 */
	public boolean isRoot(T node) {
		return this.root.getNodeData().equals(node);
	}

	public String toBFSString() {
		StringBuffer sb = new StringBuffer();
		sb.append("[");
		sb.append(this.root.toBFSString());
		sb.append("]");
		return sb.toString();
	}

	public String toDFSString() {
		StringBuffer sb = new StringBuffer();
		sb.append("[");
		sb.append(this.root.toDFSString());
		sb.append("]");
		return sb.toString();
	}

	public static UndirectedGraph<VarIdentifier> getConstraintGraph() {
		int weight = 0;
		UndirectedGraph<VarIdentifier> graph = new UndirectedGraph<VarIdentifier>();
		// let's throw in a clique, too
		for(int i = 1 ; i <= 10 ; i++) {
			for(int j = i + 1 ; j <= 10 ; j++) {
				VarIdentifier xi = new VarIdentifier(i, 1);
				VarIdentifier xj = new VarIdentifier(j, 1);
				weight++;
				graph.addEdge(xi, xj, "C" + weight, weight);
			}
		}

		// and, last, a partial clique
		for(int i = 11 ; i <= 20 ; i++) {
			for(int j = i + 1 ; j <= 20 ; j++) {
				if (Math.random() < 0.2) {
					VarIdentifier xi = new VarIdentifier(i, 1);
					VarIdentifier xj = new VarIdentifier(j, 1);
					weight++;
					graph.addEdge(xi, xj, "C" + weight, weight);
				}
			}
		}

		List<Vertex<VarIdentifier>> index = new ArrayList<Vertex<VarIdentifier>>();
		index.addAll(graph.getVertexes());
		// and one edge to connect them all
		for(int i = 0 ; i < index.size() - 1 ; i++) {
			weight++;
			VarIdentifier xi = (VarIdentifier) index.get(i).getData();
			VarIdentifier xj = (VarIdentifier) index.get(i + 1).getData();
			if (!graph.areConnected(new Vertex<VarIdentifier>(xi), new Vertex<VarIdentifier>(xj))) {
				graph.addEdge(xi, xj, "C" + weight, weight);
			}
		}
		return graph;
	}

	public static UndirectedGraph<VarIdentifier> getSimpleConstraintGraph() {
		UndirectedGraph<VarIdentifier> graph = new UndirectedGraph<VarIdentifier>();
		VarIdentifier x1 = new VarIdentifier(1, 1);
		VarIdentifier x2 = new VarIdentifier(2, 1);
		VarIdentifier x3 = new VarIdentifier(3, 1);
		VarIdentifier x4 = new VarIdentifier(4, 1);
		VarIdentifier x5 = new VarIdentifier(5, 1);
		// VarIdentifier x6 = new VarIdentifier(6, 1);
		// VarIdentifier x7 = new VarIdentifier(7, 1);
		// VarIdentifier x8 = new VarIdentifier(8, 1);
		// VarIdentifier x9 = new VarIdentifier(9, 1);
		graph.addVertex(x1);
		graph.addVertex(x2);
		graph.addVertex(x3);
		graph.addVertex(x4);
		graph.addVertex(x5);
		// graph.addVertex(x6);
		// graph.addVertex(x7);
		// graph.addVertex(x8);
		// graph.addVertex(x9);
		graph.addEdge(x1, x2, "C1", 1.1);
		graph.addEdge(x1, x3, "C2", 2.2);
		graph.addEdge(x1, x5, "C3", 3.9);
		graph.addEdge(x2, x4, "C4", 4.7);
		// graph.addConstraint(x2, x5, "C5", 5.0);
		// graph.addConstraint(x3, x4, "C6", 6.2);
		graph.addEdge(x3, x5, "C7", 7.1);
		return graph;
	}

	public static void main(String[] args) throws Exception {
		AbstractTree<String> tree = new AbstractTree<String>();
		// System.out.println("                    : "+tree.toString()+" isEmpty: "+tree.isEmpty());
		AbstractTreeNode<String> A = new AbstractTreeNode<String>("A");
		AbstractTreeNode<String> B = new AbstractTreeNode<String>("B");
		AbstractTreeNode<String> C = new AbstractTreeNode<String>("C");
		AbstractTreeNode<String> D = new AbstractTreeNode<String>("D");
		AbstractTreeNode<String> E = new AbstractTreeNode<String>("E");
		AbstractTreeNode<String> F = new AbstractTreeNode<String>("F");
		AbstractTreeNode<String> G = new AbstractTreeNode<String>("G");
		tree.setRoot(A);
		A.addSubTree(B);
		A.addSubTree(C);
		A.addSubTree(D);
		C.addChild(E);
		C.addChild(F);
		D.addChild(G);
		String toString = tree.toString();
		// System.out.println("                    : "+tree.toString());
		// System.out.println("Depth               : "+tree.getDepth());
		// System.out.println("NumberOfLeaves      : "+tree.getNumberOfLeaves());
		// System.out.println("getRoot             : "+tree.getRoot());
		// System.out.println("getRootNode         : "+tree.getRootNode());
		// System.out.println("size                : "+tree.size());
		// System.out.println("Parent of G         : "+tree.getParent("G"));
		// System.out.println("sub tree rooted at C: "+tree.getSubTree("C"));
		// System.out.println("All descendants of A: "+tree.getRoot().getAllDescendants());

		System.out.println("\n" + toString);

		// int closingBracket = AbstractTreeNode.findClosingBracket(toString, 0);
		// System.out.println("0 --> "+closingBracket+", length: "+toString.length());
		// closingBracket = AbstractTreeNode.findClosingBracket(toString, 4);
		// System.out.println("4 --> "+closingBracket+", length: "+toString.length());
		// closingBracket = AbstractTreeNode.findClosingBracket(toString, 9);
		// System.out.println("9 --> "+closingBracket+", length: "+toString.length());
		//
		// System.out.println(tree.toString());
		// System.out.println(tree.getRoot().toElegantString());

		// AbstractTree<String, AbstractTreeNode<String>> tr = AbstractTree.fromString(toString);
		// tr = AbstractTree.fromString("[]");
		// System.out.println(tr.toString());
		//
		// String fileName = "./treedot.dot";
		// String dotFileName = AbstractTree.toDotFile(tree, fileName);
		// AbstractGraph.fromDOTtoPDF(dotFileName);

		UndirectedGraph<VarIdentifier> graph = getSimpleConstraintGraph();
		graph = getConstraintGraph();
		String fileName = "./graph.dot";
		String dotFileName = AbstractGraph.toDotFile(graph, fileName);
		AbstractGraph.fromDOTtoPDF(dotFileName);

		// AbstractTree<IVarIdentifier, AbstractTreeNode<IVarIdentifier>> dfs = AbstractTree.DFS(graph);
		AbstractTree<VarIdentifier> dfs = graph.DFS();
		// System.out.println(dfs.pretty());
		String str = dfs.toString();
		System.out.println("DFS: \n" + str + "\n-------------------------------");

		String treeFileName = "./dfs.dot";
		String treeDotFileName = toDotFile(dfs, treeFileName);
		AbstractGraph.fromDOTtoPDF(treeDotFileName);

		dfs = AbstractTree.DFS(graph);
		treeFileName = "./dfs2.dot";
		treeDotFileName = toDotFile(dfs, treeFileName);
		AbstractGraph.fromDOTtoPDF(treeDotFileName);

		AbstractTree<String> tr = fromString(toString, String.class);
		System.out.println(tr.toString());
		tr = fromString("[]", String.class);
		System.out.println(tr.toString());

	}

	/**
	 * @param source
	 * @param target
	 * @param weight
	 */
	public void setWeight(T source, T target, double weight) {
		this.setWeight(new Vertex<T>(source), new Vertex<T>(target), weight);
	}

	/**
	 * @param source
	 * @param target
	 * @param weight
	 */
	public void setWeight(Vertex<T> source, Vertex<T> target, double weight) {
		if (!isInTree(source)) {
			throw new IllegalArgumentException("Source vertex(" + source + ") not in this tree.");
		}
		AbstractTreeNode<T> subtree = this.getSubTree(source);
		if (subtree.contains(target)) {
			Edge<T> edge = subtree.getEdge(target);
			edge.setWeight(weight);
		} else {
			if (!isInTree(target)) {
				throw new IllegalArgumentException("Target vertex(" + target + ") not in this tree.");
			}
			subtree = this.getSubTree(target);
			Edge<T> edge = subtree.getEdge(source);
			edge.setWeight(weight);
		}
	}
}
