/*
DisChoco: Distributed Constraint Reasoning Solver 
Copyright (C) 2007-2013 Redouane Ezzahir & Mohamed Wahbi 

DisChoco is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DisChoco is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

To contact the authors: <http://dischoco.sourceforge.net/>
*/ 
package util.tree;

import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import util.HashList;
import util.graph.Edge;
import util.graph.Vertex;

public class AbstractTreeNode<T extends Comparable<? super T>> {

	public static final char OPENING_BRACKET = '{';
	public static final char CLOSING_BRACKET = '}';
	public static final char SEPARATOR = ';';

	private Vertex<T> node = null;
	private List<AbstractTreeNode<T>> children = null;

	public AbstractTreeNode() {
		super();
		this.children = new ArrayList<AbstractTreeNode<T>>();
	}

	public AbstractTreeNode(Vertex<T> vertex) {
		this();
		this.node = new Vertex<T>(vertex.getData());
		//this.setNode(vertex);
	}
	
	public AbstractTreeNode(T v) {
		this(new Vertex<T>(v));
	}

	public HashList<Vertex<T>> getChildrenNodes() {
		HashList<Vertex<T>> nodes = new HashList<Vertex<T>>();
		for(AbstractTreeNode<T> child : this.children) {
			nodes.add(child.getNode());
		}
		return nodes;
	}

	public HashList<T> getChildrensData() {
		HashList<T> nodes = new HashList<T>();
		for(AbstractTreeNode<T> child : this.children) {
			nodes.add(child.getNode().getData());
		}
		return nodes;
	}

	public List<AbstractTreeNode<T>> getChildren() {
		return this.children;
	}

	public Collection<Edge<T>> getEdges() {
		return this.node.getOutgoingEdges();
	}

	public int getNumberOfChildren() {
		return getChildren().size();
	}

	public boolean hasChildren() {
		return (getNumberOfChildren() > 0);
	}

	public boolean isLeaf() {
		return (getNumberOfChildren() == 0);
	}

	public void setChildren(List<AbstractTreeNode<T>> children) {
		this.children = children;
	}

	public void addChild(AbstractTreeNode<T> child) {
		this.addChild(child, new Edge<T>(this.getNode(), child.getNode(), true));
	}

	public void addChild(AbstractTreeNode<T> child, Edge<T> edge) {
		this.children.add(child);
		this.addEdge(edge);
	}

	public void addEdge(Edge<T> edge) {
		this.node.addOutgoingEdge(edge);
	}

	public void addChildAt(int index, AbstractTreeNode<T> child) throws IndexOutOfBoundsException {
		this.children.add(index, child);
		this.addEdge(new Edge<T>(getNode(), child.getNode(), true));
	}

	public void removeChildren() {
		this.children = new ArrayList<AbstractTreeNode<T>>();
	}

	public void removeChildAt(int index) throws IndexOutOfBoundsException {
		this.children.remove(index);
	}

	public AbstractTreeNode<T> getChildAt(int index) throws IndexOutOfBoundsException {
		return this.children.get(index);
	}

	public Edge<T> getEdgeAt(int index) throws IndexOutOfBoundsException {
		Vertex<T> child = this.children.get(index).getNode();
		return this.getEdge(child);
	}

	public Vertex<T> getNode() {
		return this.node;
	}

	public T getNodeData() {
		return this.node.getData();
	}

	public void setNode(Vertex<T> node) {
		this.node = node;
	}

	public Edge<T> getEdge(Vertex<T> child) {
		for(Edge<T> edge : this.node.getOutgoingEdges()) {
			if(edge.contains(child)) {
				return edge;
			}
		}
		return null;
	}

	public void setEdge(Vertex<T> child, Edge<T> edge) {
		if (!this.isChild(child)) {
			this.addChild(new AbstractTreeNode<T>(child), edge);
		}
		this.addEdge(edge);
	}

	private boolean isChild(Vertex<T> vertex) {
		for(AbstractTreeNode<T> child : this.children) {
			if (child.getNode().equals(vertex)) {
				return true;
			}
		}
		return false;
	}

	public int hashCode() {
		return getNode().hashCode();
	}

	public String toStringVerbose(String tab) {
		StringBuffer sb = new StringBuffer();
		sb.append(getNode().toString());
		for(AbstractTreeNode<T> st : getChildren()) {
			sb.append("\n"+tab+st.toStringVerbose(tab+"\t"));
		}
		return sb.toString();
		// Pattern.DOTALL causes ^ and $ to match. Otherwise it won't. It's retarded.
//		Pattern pattern = Pattern.compile(", $", Pattern.DOTALL);
//		Matcher matcher = pattern.matcher(stringRepresentation);
//
//		stringRepresentation = matcher.replaceFirst("");
//		stringRepresentation += "]";
//
//		return stringRepresentation;
	}

	public int getNumberOfDescendants() {
		if(this.isLeaf()) {
			return 0;
		} else {
			int nbDescendants = 0;
			for(AbstractTreeNode<T> child : getChildren()) {
				nbDescendants += 1 + child.getNumberOfDescendants();
			}
			return nbDescendants;
		}
	}

	public HashList<Vertex<T>> getAllDescendants() {
		HashList<Vertex<T>> descendants = new HashList<Vertex<T>>();
		for(AbstractTreeNode<T> child : getChildren()) {
			descendants.add(child.getNode());
			descendants.addAll(child.getAllDescendants());
		}
		return descendants;
	}

	public HashList<T> getAllDescendantsData() {
		HashList<T> descendants = new HashList<T>();
		for(AbstractTreeNode<T> child : getChildren()) {
			descendants.add(child.getNodeData());
			descendants.addAll(child.getAllDescendantsData());
		}
		return descendants;
	}

	public Collection<Vertex<T>> getSubtreeVertexes() {
		List<Vertex<T>> descendants = new ArrayList<Vertex<T>>();
		descendants.add(this.getNode());
		descendants.addAll(getAllDescendants());
		return Collections.unmodifiableCollection(descendants);
	}

	/**
	 * This method computes the depth of this tree. 
	 * The depth of tree is the longest path from root node to leaf node. 
	 * @return int : the depth of tree. 
	 */
	public int getDepth() {
		int depth = 0;
		for(AbstractTreeNode<T> child : getChildren()) {
			depth = Math.max(depth, child.getDepth());
		}
		return depth + 1;
	}

	public int getNumberOfLeaves() {
		if (isLeaf()) return 1;
		int sum = 0;
		for(AbstractTreeNode<T> child : getChildren())
			sum += child.getNumberOfLeaves();
		return sum;
	}
	
	/**
	 * This method returns a pointer to the subtree rooted at the node with the given identifier.
	 * 
	 * @param node the identifier of the root
	 * @return AbstractTreeNode<T> : the sub-tree rooted at the node with the given root <code>v</code> 
	 */
	public AbstractTreeNode<T> getSubTree(Vertex<T> node) {
		if (getNode().equals(node))
			return this;
		for(AbstractTreeNode<T> st : getChildren()) {
			if (st.getNode().equals(node)) {
				return st;
			} else {
				AbstractTreeNode<T> subtree = st.getSubTree(node);
				if (subtree != null) {
					return subtree;
				}
			}
		}
		return null;
	}

	/**
	 * This tree contains the given node
	 * @param node
	 * @return
	 */
	public boolean contains(Vertex<T> vertex) {
		if (this.getNode().equals(vertex)) {
			return true;
		}
		for(AbstractTreeNode<T> st : getChildren()) {
			if (st.contains(vertex)) {
				return true;
			}
		}
		return false;
	}
	
	/**
	 * This method checks if there any node in the given subtree <code>subtree</node> is already in this tree. 
	 * 
	 * @param subtree : the sub-tree to be traversed. 
	 * @return boolean : <code>true</code>, if a such node <code>node</code> is in this tree; <code>false</code>, otherwise. 
	 **/
	public boolean consistencyCheck(AbstractTreeNode<T> subtree) {
		if (this.contains(subtree.getNode())) {
			return false;
		}
		for(AbstractTreeNode<T> st : subtree.getChildren()) {
			if (!consistencyCheck(st)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * This method adds a branch to the current node.
	 * @param subtree : the subtree to add
	 * @param edge : 
	 */
	public void addSubTree(AbstractTreeNode<T> subtree, Edge<T> edge) {
		if (consistencyCheck(subtree)) {
			this.children.add(subtree);
			this.addEdge(edge);
		} else {
			throw new UnsupportedOperationException("The subtree fails in Consistency Check!!" + "\n " + this.toString() + "\n " + subtree.toString());
		}
	}

	/**
	 * This method adds a branch to the current node.
	 * @param subtree : the subtree to add
	 */
	public void addSubTree(AbstractTreeNode<T> subtree) {
		this.addSubTree(subtree, new Edge<T>(getNode(), subtree.getNode(), true));
	}
	
	/**
	 * This method checks if the given tree node equals to the current one. 
	 * @param st : the tree
	 * @return boolean : true if the current tree equals <code>tree</code> node. False, otherwise. 
	 */
	public boolean equals(AbstractTreeNode<T> tree) {
		return getNode().equals(tree.getNode());
	}

	/**
	 * This method traverses the tree rooted in the current node searching the given node. 
	 * It returns a pointer to the subtree whose root is the parent of the given node. 
	 * 
	 * @param node : the node to locate in the tree
	 * @return AbstractTreeNode<T> : the tree rooted at the node <code>node</code> 
	 * if the it exists, otherwise, null.
	 */
	public Vertex<T> getParent(Vertex<T> node) {
		if (this.isChild(node)) {
			return this.getNode();
		}
		for(AbstractTreeNode<T> st : getChildren()) {
			Vertex<T> parent = st.getParent(node);
			if (parent != null) {
				return parent;
			}
		}
		return null;
	}
	
	/**
	 * This method returns a string representation of the tree rooted at the current node.
	 * 
	 * @return String representing the tree rooted at the current node.
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append(OPENING_BRACKET);
		sb.append(this.node.toString());
		for (AbstractTreeNode<T> st : getChildren()) {
			sb.append(SEPARATOR);
			sb.append(' ');
			sb.append(st.toString());
		}
		sb.append(CLOSING_BRACKET);
		return sb.toString();
	}
	
	public String toElegantString() {
        StringBuilder sb = new StringBuilder();
        sb.append("{").append(getNode().toString()).append(",[");
        int i = 0;
        for (AbstractTreeNode<T> node : getChildren()) {
            if (i > 0) {
                sb.append(",");
            }
            sb.append(node.toElegantString());
            i++;
        }
        sb.append("]").append("}");
        return sb.toString();
    }
	
	public String toBFSString() {
		StringBuffer sb = new StringBuffer();
		LinkedList<AbstractTreeNode<T>> queue = new LinkedList<AbstractTreeNode<T>>();
		queue.addLast(this);
		sb.append(this.getNode());
		while (!queue.isEmpty()) {
			AbstractTreeNode<T> pt = queue.removeFirst();
			for (AbstractTreeNode<T> st : pt.getChildren()) {
				sb.append(" " + st.getNode());
				queue.addLast(st);
			}
		}
		return sb.toString();
	}

	public String toDFSString() {
		StringBuffer sb = new StringBuffer();
		sb.append(getNode());
		for (AbstractTreeNode<T> st : getChildren()) {
			sb.append(" " + st.toDFSString());
		}
		return sb.toString();
	}

	public static <T extends Comparable<? super T>> AbstractTreeNode<T> fromString(String string, Class<T> elementClass) {
		// case where there is one node
		if(string.indexOf(SEPARATOR) < 0) {
			try {
				Object[] args = new Object[] {string.substring(1, string.indexOf(CLOSING_BRACKET)).trim()};
				Constructor<T> constructor = elementClass.getConstructor(new Class[] {String.class});
				T node = constructor.newInstance(args);
				return new AbstractTreeNode<T>(node);
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		// case where there are the current node and its sub-tree
		int separatorIndex = string.indexOf(SEPARATOR);
		T node = null;
		try {
			Object[] args = new Object[] {string.substring(1, separatorIndex).trim()};
			Constructor<T> constructor = elementClass.getConstructor(new Class[] {String.class});
			node  = constructor.newInstance(args);
		} catch (Exception e) {
			e.printStackTrace();
		}
		AbstractTreeNode<T> tree = new AbstractTreeNode<T>(node);
		// look for the index where begins and ends the first subtree string
		int beginIndex = string.indexOf(OPENING_BRACKET, separatorIndex);
		int endIndex = findClosingBracket(string, beginIndex);
		while(beginIndex < endIndex){
			// parse the subtree by recursive calling the method with the substring representing the subtree
			// and then add it the the current tree
			tree.addChild((AbstractTreeNode<T>)fromString(string.substring(beginIndex, endIndex+1), elementClass));
			// compute the next index where begins the next subtree 
			beginIndex = string.indexOf(OPENING_BRACKET, endIndex);
			// if there is no more subtrees to parse return the current tree
			if(beginIndex<0) {
				return tree;
			}
			// otherwise, compute the index where the substring representing the next subtree ends
			endIndex = findClosingBracket(string, beginIndex);
		}
		return tree;
	}
	
	public static int findClosingBracket(String str, int index) {
		int nbOpenedBrace = 0;
		int nbClosedBrace = 0;
		int i = index;
		do {
			char c = str.charAt(i++);
			if (c == CLOSING_BRACKET) {
				nbClosedBrace++;
			} else if (c == OPENING_BRACKET){
				nbOpenedBrace++;
			}
		} while (i < str.length() && nbOpenedBrace > nbClosedBrace );
		return i-1;
	}

	/**
	 * This method returns all the nodes that belong to this tree.
	 * 
	 * @return Collection<DirectedEdge<T>> : all nodes belonging to the tree
	 */
	public Collection<Edge<T>> getAllEdges() {
		Set<Edge<T>> edges = new HashSet<Edge<T>>(this.getEdges());
		for(AbstractTreeNode<T> st : getChildren()) {
			edges.addAll(st.getAllEdges());
		}
		return edges;
	}

	public Collection<AbstractTreeNode<T>> getSubtrees() {
		Collection<AbstractTreeNode<T>> subtrees = new ArrayList<AbstractTreeNode<T>>();
		for(AbstractTreeNode<T> st : getChildren()) {
			subtrees.add(st);
			subtrees.addAll(st.getSubtrees());
		}
		return subtrees;
	}
	
}
